// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.parser.env;


import java.util.*;

import de.uka.ilkd.asmkey.logic.AsmProcedure;
import de.uka.ilkd.asmkey.logic.AsmLemma;
import de.uka.ilkd.asmkey.logic.AsmSort;
import de.uka.ilkd.asmkey.logic.DynamicFunction;
import de.uka.ilkd.key.logic.Named;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.ListOfName;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.NamespaceSet;
import de.uka.ilkd.key.rule.Heuristic;
import de.uka.ilkd.key.rule.Taclet;
import de.uka.ilkd.asmkey.logic.dfinfo.*;



/** Implementation for SortEnvironment, TermEnvironment and
 * DeclarationEnvironment.
 *
 * @author Hubert Schmid
 */

public class DefaultEnvironment implements DeclarationEnvironment {

    /** Creates an approtiate map and puts the elements (instanceof
     * Name) in it. */
    private static Map createMap(Collection c) {
        Map map = new HashMap();
        populateMap (map, c);
        return map;
    }

    private static void populateMap(Map map, Collection c) {
	if (c != null) {
            for (Iterator i = c.iterator(); i.hasNext();) {
                Object element = i.next();
                //assert element instanceof Named;
                map.put(((Named) element).name().toString(), element);
            }
        }
    }

    /** Returns the element with the name 'id' from the map. Throws an
     * EnvironmentException with the text name, if no such element
     * exists.
     *
     * @throws EnvironmentException if no such element exists.
     */
    private static Object get(String name, String id, Map map) throws EnvironmentException {
        if (map.containsKey(id)) {
            return map.get(id);
        } else {
            throw new EnvironmentException(name + " " + id + " not found.");
        }
    }

    /** Creates an unmodifiable list containing all the elements
     * of the map. */
    private static List createList(Map map) {
        return Collections.unmodifiableList(new ArrayList(map.values()));
    }


    /** Meta operators. Mapping from {@link String} to {@link
     * MetaOperator}. */
    private final Map metaOps;

    /** Declared sorts. Mapping from {@link String} to {@link
     * Sort}. */
    private final Map sorts;

    /** Schema variables. Mapping from {@link String} to {@link
     * SchemaVariable}. */
    private final Map vars;

    /** Operations, functions and predications. Mapping from {@link
     * String} to {@link Operator}. */
    private final Map ops;

    /** Named asm rules (procedures). Mapping from {@link String} to
     * {@link AsmProcedure}. */
    private final Map procs;

    /** Declared heuristics. Mapping from {@link String} to {@link
     * Heuristic}. */
    private final Map heurs;

    /** Declared taclets. Mapping from {@link String} (taclet name) to
     * {@link Taclet}. */
    private final Map taclets;

    /** Declared lemmas. Mapping from {@link String} (lemma name) to
     * {@link AsmLemma} */
    private final Map lemmas;

    /** Abbreviations. */
    private final AbbreviationMap abbreviations;

    /** for registering the DFInfo's for access properties. */
    private final DFInfoFactory accessDFInfoFactory;

    /** for registering the DFInfo's for update properties */
    private final DFInfoFactory updateDFInfoFactory;


    /** Create an empty environment. */
    public DefaultEnvironment() {
        this(null, null, null, null, null, null, null, null, null);
    }

    /** Create an empty environment with the given meta operators. */
    public DefaultEnvironment(Collection metaOps) {
        this(metaOps, null, null, null, null, null, null, null, null);
    }

    /** Creates an environment with the given meta operators, sorts,
     * vars, operators (functions, predicates), named asm rules
     * (procedures), heuristics, taclets and abbreviations. Each
     * argument may be null. */
    public DefaultEnvironment(Collection metaOps,
                              Collection sorts,
                              Collection vars,
                              Collection ops,
                              Collection procs,
                              Collection heurs,
                              Collection taclets,
			      Collection lemmas,
                              AbbreviationMap abbreviations) {
        this.metaOps = createMap(metaOps);
        this.sorts = createMap(sorts);
        this.vars = createMap(vars);
        this.ops = createMap(ops);
        this.procs = createMap(procs);
        this.heurs = createMap(heurs);
        this.taclets = createMap(taclets);
	this.lemmas = createMap(lemmas);
        this.abbreviations = abbreviations;
	accessDFInfoFactory = new DFInfoFactory();
	updateDFInfoFactory = new DFInfoFactory();
	this.sorts.put("asm-rule", AsmSort.ASMRULE);
    }

    public boolean containsMetaOp(String id) {
        return metaOps.containsKey(id);
    }

    public MetaOperator getMetaOp(String id) throws EnvironmentException {
        return (MetaOperator) get("Meta operator", id, metaOps);
    }

    public Sort getSort(String id) throws EnvironmentException {
        return (Sort) get("Sort", id, sorts);
    }

    public void addSort(Sort sort) throws EnvironmentException {
        String id = sort.name().toString();
        if (sorts.containsKey(id)) {
            throw new EnvironmentException("Sort " + id + " is already defined.");
        } else {
            sorts.put(id, sort);
        }
    }

    /** Returns all declared (and initial) sorts. */
    public List getSorts() {
        return createList(sorts);
    }

    public boolean containsSchemaVariable(String id) {
        return vars.containsKey(id);
    }

    public SchemaVariable getSchemaVariable(String id) throws EnvironmentException {
        return (SchemaVariable) get("Schema variable", id, vars);
    }

    /** Returns all declared (and initial) schema variables. */
    public List getSchemaVariables() {
        return createList(vars);
    }

    public void addSchemaVariable(SchemaVariable sv) throws EnvironmentException {
        String id = sv.name().toString();
        if (vars.containsKey(id)) {
            throw new EnvironmentException("SchemaVariable " + id + " is already defined.");
        } else {
            vars.put(id, sv);
        }
    }

    public Operator getOperator(String id) throws EnvironmentException {
        return (Operator) get("Operator", id, ops);
    }

    public void addOperator(Operator op) throws EnvironmentException {
        String id = op.name().toString();
        if (ops.containsKey(id)) {
            throw new EnvironmentException("Operator " + id + " is already defined.");
        } else {
            ops.put(id, op);
	    if (op instanceof DynamicFunction) {
		accessDFInfoFactory.register(op.name());
		updateDFInfoFactory.register(op.name());
	    }
        }
    }

    /** Returns all declared (and initial) operators. */
    public List getOperators() {
        return createList(ops);
    }

    public AsmProcedure getProcedure(String id) throws EnvironmentException {
        return (AsmProcedure) get("Procedure", id, procs);
    }

    public void addProcedure(AsmProcedure p) throws EnvironmentException {
        String id = p.getId();
        if (procs.containsKey(id)) {
            throw new EnvironmentException("Procedure " + id + " is already defined.");
        } else {
            procs.put(id, p);
        }
    }

    /** Returns all declared procedures (named ASM rules). */
    public List getProcedures() {
        return createList(procs);
    }

    public Heuristic getHeuristic(String id) throws EnvironmentException {
        if (heurs.containsKey(id)) {
            return (Heuristic) heurs.get(id);
        } else {
            throw new EnvironmentException("Heuristic " + id + " not found.");
        }
    }

    public void addHeuristic(Heuristic h) throws EnvironmentException {
        String id = h.name().toString();
        if (heurs.containsKey(id)) {
            throw new EnvironmentException("Heuristic " + id + " is already defined.");
        } else {
            heurs.put(id, h);
        }
    }

    /** Returns all declared (and initial) heuristics. */
    public List getHeuristics() {
        return createList(heurs);
    }

    public void addTaclet(Taclet t) throws EnvironmentException {
        String id = t.name().toString();
        if (taclets.containsKey(id)) {
            throw new EnvironmentException("Taclet " + id + " is already defined.");
        } else {
            taclets.put(id, t);
        }
    }

    /** Returns all declared (and initial) taclets. */
    public List getTaclets() {
        return createList(taclets);
    }

    public Term getAbbreviatedTerm(String name) throws EnvironmentException {
        if (abbreviations.containsAbbreviation(name)) {
            return abbreviations.getTerm(name);
        } else {
            throw new EnvironmentException("abbreviation " + name + " not found.");
        }
    }
    
    public void addLemma(AsmLemma lemma) throws EnvironmentException {
	String id = lemma.getId();
	if(lemmas.containsKey(id)) {
	    throw new EnvironmentException("Lemma " + id + " is already defined.");
	} else {
	    lemmas.put(id, lemma);
	}
    }

    /** Returns all declared lemmas. */
    public List getLemmas() {
	return createList(lemmas);
    }

    public Map getLemmasMap() {
	return lemmas;
    }

    public AsmLemma getLemma(String id) throws EnvironmentException {
	if (lemmas.containsKey(id)) {
	    return (AsmLemma) lemmas.get(id);
	} else {
	    throw new EnvironmentException("There is no lemma " + id + " declared.");
	}
    }

    /** Create a namespace from a collection a Named objects. The
     * elements of 'c' have to be of type {@link Named}. */
    private static Namespace toNamespace(Collection c) {
        Namespace ns = new Namespace();
        for (Iterator i = c.iterator(); i.hasNext();) {
            ns.add((Named) i.next());
        }
        return ns;
    }
    
    
    public NamespaceSet getEnvAsNamespaceSet() {
       return new NamespaceSet(
          toNamespace(getSchemaVariables()),
          toNamespace(getOperators()), // XXX: filter functions
          toNamespace(getSorts()),
          toNamespace(getHeuristics()),
	  new Namespace(), new Namespace());
    }

    public DFInfo getAccessDFInfo(ListOfName dfs) throws EnvironmentException {
	try {
	    return accessDFInfoFactory.getDFInfo(dfs);
	} catch (DFException e) {
	    throw new EnvironmentException("DFException: " + e.getMessage());
	}
    }

    public DFInfo getUpdateDFInfo(ListOfName dfs) throws EnvironmentException {
	try {
	    return updateDFInfoFactory.getDFInfo(dfs);
	} catch (DFException e) {
	    throw new EnvironmentException("DFException: " + e.getMessage());
	}
    }


}
