// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.parser.tree;


import java.util.Iterator;

import de.uka.ilkd.asmkey.parser.TreeParser;
import de.uka.ilkd.asmkey.parser.ast.*;
import de.uka.ilkd.asmkey.parser.env.*;
import de.uka.ilkd.asmkey.rule.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.SetAsListOfSchemaVariable;
import de.uka.ilkd.key.parser.Location;
import de.uka.ilkd.key.parser.ParserException;
import de.uka.ilkd.key.rule.*;


/** Class to parse taclets declarations.
 *
 * @author Hubert Schmid
 * @author Stanislas Nanchen
 */

final class TacletDeclarationParser {

    private static TacletBuilder createTacletBuilder(Term term) {
        return new RewriteTacletBuilder().setFind(term);
    }

    private static TacletBuilder createTacletBuilder(Sequent sequent, Location location)
        throws ParserException
    {
        if (sequent == Sequent.EMPTY_SEQUENT) {
            return new NoFindTacletBuilder();
        } else if (sequent.antecedent().size() == 1
                   && sequent.succedent().size() == 0) {
            return new AntecTacletBuilder().setFind(sequent.antecedent().get(0).formula());
        } else if (sequent.antecedent().size() == 0
                   && sequent.succedent().size() == 1) {
            return new SuccTacletBuilder().setFind(sequent.succedent().get(0).formula());
        } else {
            throw new ParserException("Invalid find-sequent in " + sequent + ".", location);
        }
    }

    /* Hint: protected access because method is called from inner
     * class. */
    protected static void addGoalReplace(TacletBuilder tb,
                                         String id,
                                         Term term,
                                         Sequent sequent,
                                         Sequent addSequent,
                                         ListOfTaclet addRules,
                                         Location location)
        throws ParserException
    {
        TacletGoalTemplate gt = null;
        if (term == null && sequent == null) {
            gt = new TacletGoalTemplate(addSequent,
                                        addRules,
                                        SLListOfRulevariable.EMPTY_LIST,
                                        SetAsListOfSchemaVariable.EMPTY_SET);
        } else if (tb instanceof NoFindTacletBuilder) {
            // there is a replacewith without a find.
            throw new ParserException("Replacewith without find.", location);
        } else if (tb instanceof SuccTacletBuilder || tb instanceof AntecTacletBuilder) {
            if (sequent != null) {
                gt = new AntecSuccTacletGoalTemplate(addSequent,
                                                     addRules,
                                                     SLListOfRulevariable.EMPTY_LIST,
                                                     sequent,
                                                     SetAsListOfSchemaVariable.EMPTY_SET);
            } else {
                throw new ParserException("Replacewith does not match find. " + term, location);
            }
        } else if (tb instanceof RewriteTacletBuilder) {
            if (term != null) {
                gt = new RewriteTacletGoalTemplate(addSequent,
                                                   addRules,
                                                   SLListOfRulevariable.EMPTY_LIST,
                                                   term,
                                                   SetAsListOfSchemaVariable.EMPTY_SET);
            } else {
                throw new ParserException("Replacewith does not match find. " + sequent, location);
            }
        }
        gt.setName(id);
        tb.addTacletGoalTemplate(gt);
    }

    /* Hint: protected access because method is called from inner
     * class. */
    protected static Term parseNullTerm(TermEnvironment env, AstTerm term)
        throws ParserException
    {
        if (term == null) {
            return null;
        } else {
            return TreeParser.parseTerm(env, term);
        }
    }

    /* Hint: protected access because method is called from inner
     * class. */
    protected static Sequent parseNullSequent(TermEnvironment env, AstSequent sequent)
        throws ParserException
    {
        if (sequent == null) {
            return null;
        } else {
            return TreeParser.parseSequent(env, sequent);
        }
    }

    /* Hint: protected access because method is called from inner
     * class. */
    protected static Sequent parseEmptySequent(TermEnvironment env, AstSequent sequent)
        throws ParserException
    {
        if (sequent == null) {
            return Sequent.EMPTY_SEQUENT;
        } else {
            return TreeParser.parseSequent(env, sequent);
        }
    }

    /* Hint: protected access because method is called from inner
     * class. */
    /** Parses recursively a list of taclets and return these
     * taclets. The taclets are not added to the given environment. */
    protected static ListOfTaclet getListOfTaclet(DeclarationEnvironment env, AstTacletDeclaration[] addRules)
        throws ParserException
    {
        ListOfTaclet list = SLListOfTaclet.EMPTY_LIST;
        if (addRules != null) {
            TacletEnvironment tacletEnv = new TacletEnvironment(env);
            TreeParser.parseDeclarations(tacletEnv, addRules);
            for (Iterator i = tacletEnv.getLocalTaclets().iterator(); i.hasNext();) {
                Taclet taclet = (Taclet) i.next();
                list = list.append(taclet);
            }
        }
        return list;
    }

    /** Parse a taclet in the given environment and add it to this
     * environment. Identifier is the name of the taclet. */
    public static void parse(final DeclarationEnvironment env, Identifier id, AstTaclet taclet) {
        /* Hint: Tthis method is quite long, because it uses a lot of
         * inner classes for the different visitors. The code itself
         * is very simple. */
        try {
            /* create a suitable TacletBuilder */
            final TacletBuilder tb =
                taclet.getTerm() != null
                ? createTacletBuilder(TreeParser.parseTerm(env, taclet.getTerm()))
                : createTacletBuilder(TreeParser.parseSequent(env, taclet.getSequent()),
                                      taclet.getSequent().getLocation());
            tb.setName(new Name(id.getText()));
            /* if sequent */
            if (taclet.getIfSeq() != null) {
                tb.setIfSequent(TreeParser.parseSequent(env, taclet.getIfSeq()));
            }
            /* list of conditions */
            for (Iterator i = taclet.getConditions().iterator(); i.hasNext();) {
                AstCondition cond = (AstCondition) i.next();
                cond.accept(new AstConditionVisitor() {

                        public void visitOperatorEquals(boolean equal, Identifier v, Identifier w) {
                            try {
                                tb.addVariableCondition
                                    (new OperatorEqualsCondition
                                     (equal,
                                      EnvironmentUtil.getSchemaVariable(env, v),
                                      EnvironmentUtil.getSchemaVariable(env, w)));
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
                        }

                        public void visitNotFreeIn(Identifier v, Identifier w) {
                            try {
                                tb.addVarsNotFreeIn
                                    (EnvironmentUtil.getSchemaVariable(env, v),
                                     EnvironmentUtil.getSchemaVariable(env, w));
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
                        }

                        public void visitPure(Identifier v) {
                            try {
                                tb.addVariableCondition
                                    (new PureCondition(EnvironmentUtil.getSchemaVariable(env, v)));
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
                        }

                        public void visitStatic(Identifier v) {
                            try {
                                tb.addVariableCondition
                                    (new StaticCondition(EnvironmentUtil.getSchemaVariable(env, v)));
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
                        }

                        public void visitStaticArgs(Identifier v) {
                            try {
                                tb.addVariableCondition
                                    (new StaticArgsCondition(EnvironmentUtil.getSchemaVariable(env, v)));
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
                        }

                        public void visitDynamic(Identifier v) {
                            try {
                                tb.addVariableCondition
                                    (new DynamicCondition(EnvironmentUtil.getSchemaVariable(env, v)));
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
                        }

			public void visitAtomic(Identifier v) {
			    try {
				tb.addVariableCondition
				    (new AtomicCondition(EnvironmentUtil.getSchemaVariable(env, v)));
			    } catch (ParserException pe) {
				throw new ParserExceptionWrapper(pe);
			    }
			}

                        public void visitCall(Identifier v) {
                            try {
                                tb.addVariableCondition
                                    (new CallCondition(EnvironmentUtil.getSchemaVariable(env, v)));
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
                        }

                        public void visitSkolem(Identifier v, Identifier w) {
                            try {
                                tb.addVarsNewDependingOn
                                    (EnvironmentUtil.getSchemaVariable(env, v),
                                     EnvironmentUtil.getSchemaVariable(env, w));
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
                        }

			public void visitMayUpdate(boolean may, Identifier r, Identifier f) {
			    try {
				tb.addVariableCondition
				    (new MayUpdateCondition
				     (may,
				      EnvironmentUtil.getSchemaVariable(env, r),
				      EnvironmentUtil.getSchemaVariable(env, f)));
			    } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
			}

			public void visitMayAccess(boolean may, Identifier r, Identifier f) {
			    try {
				tb.addVariableCondition
				    (new MayAccessCondition
				     (may,
				      EnvironmentUtil.getSchemaVariable(env, r),
				      EnvironmentUtil.getSchemaVariable(env, f)));
			    } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
			}

                    });
            }

            /* goals */
            for (Iterator i = taclet.getGoals().iterator(); i.hasNext();) {
                AstGoal goal = (AstGoal) i.next();
                final Location location = goal.getLocation();
                goal.accept(new AstGoalVisitor() {

                        public void visitReplace(String name, AstTerm term, AstSequent addSequent,
                                                 AstTacletDeclaration[] addRules) {
                            try {
                                addGoalReplace(tb,
                                               name,
                                               parseNullTerm(env, term),
                                               null,
                                               parseEmptySequent(env, addSequent),
                                               getListOfTaclet(env, addRules),
                                               location);
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
                        }

                        public void visitReplace(String name,
                                                 AstSequent sequent,
                                                 AstSequent addSequent,
                                                 AstTacletDeclaration[] addRules) {
                            try {
                                addGoalReplace(tb,
                                               name,
                                               null,
                                               parseNullSequent(env, sequent),
                                               parseEmptySequent(env, addSequent),
                                               getListOfTaclet(env, addRules),
                                               location);
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
			    }
                        }

			public void visitReplace(String name,
						 AstSequent sequent) {
			    /*try {
                                addGoalReplace(tb,
                                               name,
					       
                                               location);
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
				}*/
			}

                    });
            }

            /* modifiers */
            for (Iterator i = taclet.getModifiers().iterator(); i.hasNext();) {
                AstTacletModifier mod = (AstTacletModifier) i.next();
                mod.accept(new AstTacletModifierVisitor() {

                        public void visitHeuristics(Identifier[] ids) {
                            try {
                                for (int j = 0; j < ids.length; ++j) {
                                    tb.addHeuristic(EnvironmentUtil.getHeuristic(env, ids[j]));
                                }
                            } catch (ParserException pe) {
                                throw new ParserExceptionWrapper(pe);
                            }
                        }

                        public void visitNonInteractive() {
                            tb.setNoninteractive(true);
                        }

                        public void visitRecursive() {
                            tb.setRecursive(true);
                        }

                        public void visitDisplayName(String displayName) {
                            tb.setDisplayName(displayName);
                        }

                        public void visitHelpText(String helpText) {
                            tb.setHelpText(helpText);
                        }

                    });
            }
            env.addTaclet(tb.getTaclet());
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        } catch (EnvironmentException ee) {
            throw new ParserExceptionWrapper
                (new ParserException(ee.getMessage(),
                                     id.getLocation()));
        }

    }

}
