// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.rule;

import java.util.Collection;
import java.util.Iterator;
import de.uka.ilkd.asmkey.logic.*;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.rule.BuiltInRule;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.util.Debug;


/** This class is a builtin rule to make a step from a world into the
 * (deterministic) subsequent world of the Kripke structure.
 *
 * @author Stanislas Nanchen
 */

public final class BigOperatorRule extends OperatorRule {
    
    /** Short refernce to TermFactory. */
    private static final TermFactory tf = TermFactory.DEFAULT;

    /** The MetaBigOperator operator has to know all dynamic function
     * symbols.  This list should be initialized on problem loading,
     * because all dynamic function symbols are declared in the
     * problem file.  The elements of this collection are instances of
     * type 'Function'. */
    private Collection functions;
    private final Name name = new Name("BigOperator");

    /** Set the list of dynamic function symbols that are used by this
     * meta operator. This method can only be called once.  The list
     * of dynamic function symbols is typically known on problem
     * loading. The elements of the collection have to be instances of
     * type 'Function'. */
    public void setDynamicFunctions(Collection c) {
        if (c == null) {
            throw new NullPointerException();
        } else if (functions != null) {
            throw new IllegalStateException("Dynamic functions on ASM_JOIN already set.");
        }
        functions = c;
    }

    public boolean isApplicable(Goal goal, PosInOccurrence pos, Constraint constraint) {
	if (pos == null || pos.posInTerm() != PosInTerm.TOP_LEVEL) {
            return false;
        } else {
            Term term = pos.constrainedFormula().formula();
            if (term.op() instanceof BigOperator) {
		return true;
	    } else {
		return false;
	    }
        }
    }

     /** This method replaces a big operator by a suitable formula */
    protected Term calculate(Term term, Services services) {
	BigOperator big;
	Term formula;
	Collection localFunctions;
	Term result;

        Debug.assertTrue(term.op() instanceof BigOperator);

	if (term.op() instanceof BigOperator) {
	    big = (BigOperator) term.op();
	    localFunctions = big.filter(this.functions);
	    formula = term.sub(0);
	    if (localFunctions == null) {
		throw new IllegalStateException("Dynamic functions are not initialized.");
	    }
	    if (localFunctions.size() == 0) {
		result = tf.createJunctorTerm(Op.TRUE);
	    } else {
		result = null;
		for (Iterator i = localFunctions.iterator(); i.hasNext();) {
		    Function f = (Function) i.next();
		    Term temp;

		    /* We create first n(arity) new logical variable and terms
		     * corresponding to them; and the f(x1, ..., xn) term
		     */
		    LogicVariable[] xs = new LogicVariable[f.arity()];
		    Term[] terms = new Term[f.arity()];
		    Term fterm;
		    VariableFactory factory = new VariableFactory();
		    for(int j = 0; j < xs.length; j++) {
			xs[j] = factory.createNewVariable(term, "x", f.argSort(j));
			terms[j] = tf.createVariableTerm(xs[j]);
		    }
		    fterm = tf.createFunctionTerm(f, terms);
		    
		    /* then we need to transform the formula of the big operator
		     * i.e. transform recursively the %f into:
		     * - a series of all quantors or ex quantors.
		     * - the fterm.
		     */
		    temp = transform_and_subst(formula, big.getVariableName(), xs, fterm, services);
		    
		    /* Now we can collect the results.
		     */
		    if (result == null) {
			result = temp;
		    } else {
			result = big.join(result, temp);
		    }
		}
	    }
	} else {
	    result = term;
	}
	return result;
    }


    public Name name() { return name; }

    public String displayName() {
        return "BigOperator";
    }
}
