// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.casetool.together;

import de.uka.ilkd.key.casetool.ReprModelMethod;
import de.uka.ilkd.key.casetool.together.TogetherReprModelClass;
import de.uka.ilkd.key.casetool.ReprModelClass;
import de.uka.ilkd.key.proof.mgt.JavaModelMethod;

import com.togethersoft.openapi.rwi.RwiMember;
import com.togethersoft.openapi.rwi.RwiModel;
import com.togethersoft.openapi.rwi.RwiDiagram;
import com.togethersoft.openapi.rwi.RwiNode;
import com.togethersoft.openapi.rwi.RwiProperty;
import java.util.Vector;

/** represents a method of the Together model. It is backed by the
 * corresponding Together RWI model element, but it is independent in
 * a sense that changing the together project allows to access all
 * methods unless they are specifically marked.
 */
public class TogetherReprModelMethod extends TogetherReprModel 
    implements ReprModelMethod {

    // which element of the model do I represent
    private final RwiMember orig;     
    private RwiModel currRwiModel;
    private RwiDiagram currRwiDiagram;
    
    private RwiMember origParent;
    private boolean origParentAlreadySet = false;

    private boolean parameterListComputed = false;
    private Vector paraNames = new Vector();
    private Vector paraTypes = new Vector();

    private ReprModelClass containingClass;
    private String retTypeString;
    private String callSig;
    private String callSigOCL;
    private String name;
    private String precond;
    private String postcond;
    private String modifies;
    private String stereotype;
    private boolean constr;
    private boolean priv;
    private int hash = 0;


    public TogetherReprModelMethod(RwiMember theOrig, RwiModel model, 
				   RwiDiagram diag) { 
	orig = theOrig;
	currRwiModel = model;
	currRwiDiagram = diag;
	this.containingClass = createContainingClassRepr();
	retTypeString = orig.getProperty(RwiProperty.RETURN_TYPE);
	callSigOCL=createCallSignature(true);
	callSig=createCallSignature(false);
	name = orig.getProperty(RwiProperty.NAME);
	precond = orig.getProperty("preconditions");
	postcond = orig.getProperty("postconditions");
	modifies = orig.getProperty("modifies");
	stereotype = orig.getProperty("stereotype");	   
	constr = orig.getProperty(RwiProperty.CONSTRUCTOR) != null;
	priv = orig.getProperty(RwiProperty.PRIVATE) != null;
    }

    /** returns the together original object representing this
     * method. Could be invalidated since a new project has been
     * loaded.
     */
    public RwiMember getOrig() {
	return orig;
    }

    public ReprModelClass createContainingClassRepr(){
	RwiNode localContainingClass = orig.getContainingNode();
	return new TogetherReprModelClass(localContainingClass, currRwiModel, 
					  currRwiDiagram);
    }

    public ReprModelClass getContainingClass() {
	return containingClass;
    }

    public String getContainingPackage(){
	return getContainingClass().getContainingPackage();
    }

    public String getContainingClassName(){
	return getContainingClass().getClassName();	
    }

    public boolean isVoid() {
	return (retTypeString==null || "void".equals(retTypeString));
    }

    public String getParameterNameAt(int i) {
	if (!parameterListComputed) 
	    getCallSignature();
	return paraNames.elementAt(i).toString();
    }

    public String getParameterTypeAt(int i) {
	if (!parameterListComputed) 
	    getCallSignature();

	return paraTypes.elementAt(i).toString();
    }

    public String getCallSignature() {
	return getCallSignature(true);
    }
 
   
    public String getCallSignature(boolean translateJavaType2OCLTypes){
	if (translateJavaType2OCLTypes) {
	    return callSigOCL;
	} else {
	    return callSig;
	}
    }

    // returnvalue is adapted to the needs of OCL-parser from dresden
    public String createCallSignature(boolean translateJavaType2OCLTypes){
	String argString = orig.getProperty(RwiProperty.PARAMETERS_TEXT);
	// argString has the form type1 arg1, type2 arg2, ...
	// parser expects of form arg1: type1; arg2: type2
	String parserString = shuffleForParser
	    (argString, translateJavaType2OCLTypes);
	    
	String returnTypeString = orig.getProperty(RwiProperty.RETURN_TYPE);
	String retSuffix = null;
	if (returnTypeString == null){
	    // We have a constructor method ...
	    // Leave it empty (because of the syntax used by the OCL parser ...
	    retSuffix = "";
	} else 
	    if (returnTypeString.equals("void")){
		// Leave it empty (because of the syntax used by the OCL parser ...
		retSuffix = "";
	    } else {
		if (translateJavaType2OCLTypes) 
		    retSuffix = ":" + transformTypeJava2OCL(returnTypeString);
		else 
		    retSuffix = ":" + returnTypeString;
	    }
	return orig.getProperty(RwiProperty.NAME)+"("+ parserString +")" + retSuffix;
    }

    public String getName(){ 
	return name; 
    }


    // besides shuffling the arg with the type we also
    // transform some java-types to ocl-types
    private String shuffleForParser(String argString, 
				    boolean translateJavaType2OCLTypes){
	argString=argString.trim();
	if (argString.equals("")){return "";}
	String resString = "";
	int actPos = 0;
	int nextSpace, nextColon;
	boolean through = false;
	String optColon = "";
	String actArgument, actTypeJava, actTypeOcl;
	while (!through){
	    nextSpace = argString.indexOf(' ', actPos);
	    nextColon = argString.indexOf(',', actPos);
	    if (nextColon == -1){
		// end of argString 
		through = true;
		nextColon = argString.length();
	    }
	    if (actPos > 0) {optColon = "; ";}
	    actArgument = argString.substring(nextSpace+1, nextColon);
	    actTypeJava = argString.substring(actPos, nextSpace);
	    
	    // store parameters in vector
	    if (!parameterListComputed){
		paraNames.addElement(actArgument);
		paraTypes.addElement(actTypeJava);
	    }

	    if (translateJavaType2OCLTypes)
		actTypeOcl = transformTypeJava2OCL(actTypeJava);
	    else
		actTypeOcl = actTypeJava;
	    resString = resString + optColon + actArgument + ": " + actTypeOcl;
	    actPos = nextColon + 2;
	}
	parameterListComputed = true;
	return resString;
    }

    public static String transformTypeJava2OCL(String aJavaType){
	return JavaModelMethod.transformTypeJava2OCL(aJavaType);
    }


    // looks if orig has a parent and set it if needed
    // parent means for methods the next occurrence of the method
    //  in the inheritance hierarchy which has an own pre/post specification
    private RwiMember getOrigParent(){
	if (origParentAlreadySet){
	    return origParent;
	}else{
	    // we ask our containingNode to do the job
	    RwiNode localContainingClass = orig.getContainingNode();
	    TogetherReprModelClass contClassRepr = 
		new TogetherReprModelClass(localContainingClass, currRwiModel, 
					   currRwiDiagram);
	    String myUniqueName = orig.getUniqueName();

	    RwiMember candParent;
 	    // findParentMeth returns the next Meth-Declaration
	    // in the inheritance hierarchy above
	    candParent= contClassRepr.findParentMeth(myUniqueName);
	    if (candParent==null){
		return null;
	    }
	    if ((candParent.getProperty("preconditions") != null || 
		 candParent.getProperty("postconditions") != null)){
		// candParent has its own specification
 		origParent = candParent;
	    }else{
 		// otherwise we ask candParent for its parent
		origParent=
		    (new TogetherReprModelMethod(candParent, currRwiModel, 
						 currRwiDiagram)).getOrigParent();
	    }
	    origParentAlreadySet = true;
	}
	return origParent;
    }




    /** relies on a valid backing RWI model element, i.e. the together
     * project is still valid (opened)
     * 
     */
    public String getMyOrParentPostCond() {
    	String myPre = getMyPreCond();
	String myPost = getMyPostCond();
	if (myPre.equals("") && myPost.equals("")){
	    // ask parent
	    TogetherReprModelMethod parentRepr = 
		new TogetherReprModelMethod(getOrigParent(), currRwiModel, 
					    currRwiDiagram);
	    return parentRepr.getMyPostCond();
	}else{
	    return myPost;
	}
    }


    /** relies on a valid backing RWI model element, i.e. the together
     * project is still valid (opened);
     * we ask the parent only when method does not have neither precond
     * nor postcond.  
     */
    public String getMyOrParentPreCond() {
	String myPre = getMyPreCond();
	String myPost = getMyPostCond();
	if (myPre.equals("") && myPost.equals("")){
	    // ask parent
	    TogetherReprModelMethod parentRepr = 
		new TogetherReprModelMethod(getOrigParent(), currRwiModel, 
					    currRwiDiagram);
	    return parentRepr.getMyPreCond(); 
	}else{
	    return myPre;
	}
    }

    public String getMyPreCond() { 
	return (precond==null) ? "" : precond;
    }

    /** relies on a valid backing RWI model element, i.e. the together
     * project is still valid (opened)
     */
    public void setMyPreCond(String precond){
	this.precond = precond;
	orig.setProperty("preconditions", precond);
    }
 
    /** relies on a valid backing RWI model element, i.e. the together
     * project is still valid (opened)
     */
    public String getParentPreCond() {
	// ask parent
	TogetherReprModelMethod parentRepr = 
	    new TogetherReprModelMethod(getOrigParent(), currRwiModel, 
					currRwiDiagram);
	return parentRepr.getMyPreCond();
    }

    public String getMyPostCond() { 
	return (postcond==null) ? "" : postcond;
    }

    /** relies on a valid backing RWI model element, i.e. the together
     * project is still valid (opened)
     */
    public void setMyPostCond(String postcond){
	this.postcond = postcond;
	orig.setProperty("postconditions", postcond);
    }

    /** relies on a valid backing RWI model element, i.e. the together
     * project is still valid (opened)
     */
    public String getParentPostCond() {
	// ask parent
	TogetherReprModelMethod parentRepr = 
	    new TogetherReprModelMethod(getOrigParent(), currRwiModel, 
					currRwiDiagram);
	return parentRepr.getMyPostCond();
    }

    public String getMyModifClause() {
	return modifies;
    }
    
    /** relies on a valid backing RWI model element, i.e. the together
     * project is still valid (opened)
     */
    public void setMyModifClause(String modifClause) {
	modifies = modifClause;
    	orig.setProperty("modifies", modifClause);
    }

    public boolean isQuery() {
	return ("query".equals(stereotype));
    }

    
    public boolean isConstructor() {
	return constr;
    }


    public boolean isPrivate() {
	return priv;
    }

    public String getMyGFAbs() {
    	String result = orig.getProperty("GFAbstractSyntax");
    	if (result == null) {
	    return "";
	} else {
	    return result;
	}
    }
    
    /** relies on a valid backing RWI model element, i.e. the together
     * project is still valid (opened)
     */
    public void setMyGFAbs(String abs) {
    	orig.setProperty("GFAbstractSyntax", abs);
    }

    public boolean hasOrigParent() {
        return (getOrigParent() != null);
    }
    
    /** relies on a valid backing RWI model element, i.e. the together
     * project is still valid (opened);
     * gets the overridden methods in the superclass if exists else null
     */
    public ReprModelMethod getParentMethod() {
	if (hasOrigParent()) {
	    return new TogetherReprModelMethod(getOrigParent(), currRwiModel, 
					       currRwiDiagram);
	} else
	    return null;
    }
    
    public String toString() {
       return getContainingClass().getClassName() + "::" + getCallSignature();
    }

    public boolean equals(Object cmp) {
	if (cmp==null) return false;
	if (!(cmp instanceof TogetherReprModelMethod)) return false;
	return toString().equals(cmp.toString());
    }

    public int hashCode() {
	return hash==0 ? (hash=toString().hashCode()) : hash;
    }
}
