// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//




/*

Uses code by Hans Muller and Kathy Walrath from
http://java.sun.com/products/jfc/tsc/articles/threads/threads2.html

*/


package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.proof.proofevent.*;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.util.Debug;
import javax.swing.*;
import java.awt.event.*;
import de.uka.ilkd.key.rule.*;

/**
 * Applies rules in an automated fashion in an own thread.
 */
public class ApplyHeuristics {

    private InteractiveProver iaProver;
    /** the proof that is worked with */ 
    private Proof proof;
    /** the maximum of allowed rule applications */
    private int maxApplications;
    /** the java information object */
    private Services services;
    
    /** chooses goals to which rules are applied*/
    private GoalChooser goalChooser;
    
    private SwingWorker worker;
    
    /** number of rules automatically applied */
    private int countApplied = 0;
    private long time;
    
    /** interrupted by the user? */
    private boolean userTerminated = false;
    private boolean autoModeActive = false;

    private ProofListener proofListener = new ProofListener();
    
    private boolean startedAsInteractive;

    /** number of method calls pushed... */
    private int pushCount = 0;
    /** ...and popped (returned) */
    private int popCount = 0;
    /** heuristics should run only until we leave the current meth. frame */
    private boolean justPopFrame = false;


    // Please create this object beforehand and re-use it.
    // Otherwise the addition/removal of the InteractiveProofListener
    // can cause a ConcurrentModificationException during ongoing operation
    public ApplyHeuristics(InteractiveProver iaProver) {
	this.iaProver = iaProver;
        iaProver.addInteractiveProofListener( proofListener );
        this.goalChooser = new GoalChooser();
    }


    private void countMethodCalls(RuleApp app) {
        String s =  app.rule().name().toString();    
        if (s.startsWith("method_call_empty") || 
            s.startsWith("method_call_return_void") ||
            s.startsWith("method_call_ignore_result") ) {
            popCount++;
//            System.err.println("pop "+s);
        }
        else 
        if (s.startsWith("method_call_return_ignore_trail")) ; // ugly naming
        else 
        if (s.startsWith("method_call")||s.startsWith("static_method_call")) {
            pushCount++;
//            System.err.println("push "+s);
        }
    }



    /** applies rules that are chosen by the active strategy 
     * @return true iff a rule has been applied, false otherwise
     */
    private synchronized boolean applyAutomaticRule() {

//        if (justPopFrame && (popCount>pushCount))  return false;

	/** Look for the strategy ... */
	    RuleApp               app = null;
	    Goal                  g;

	    while ( ( g = goalChooser.getNextGoal () ) != null ) {
		app = g.getRuleAppManager().next();

		if ( app == null )
		    goalChooser.removeGoal ( g );
		else
		    break;
	    }

	    if ( app != null ) {

//                if (justPopFrame) countMethodCalls(app);

//		System.out.println ( "Recommendation: " + app );

	        if ( app.rule () instanceof UpdateSimplificationRule )
	            iaProver.applyUpdateSimplificationRule ( g, (BuiltInRuleApp)app );
	        else
	            iaProver.apply ( app, g, false );

//		System.out.println ( "... done" );
		return true;
	    }

	return false;
    }


    void setupStatus() {
	final Main main = Main.getInstance ();
	final MainStatusLine sl = main.getStatusLine ();
	sl.reset ();
	sl.setProgressPanelVisible ( true );
	sl.setProgressBarMaximum ( maxApplications - 1 );
	sl.setStatusText ( "Processing Strategy" );
    }


    void removeStatus() {
	MainStatusLine sl = Main.getInstance ().getStatusLine ();
	sl.reset ();
    }


    void updateStatus() {
	Main.getInstance ().setProgress ( countApplied );
    }
    

    /**
     * applies rules until this is no longer
     * possible or the thread is interrupted.
     */
    Object doWork() {
        time = System.currentTimeMillis();
        try{
	   Debug.out("Strategy started.");
	   while (countApplied < maxApplications &&
	          applyAutomaticRule()) {
               countApplied++;
               updateStatus();
               if (Thread.interrupted()) throw new InterruptedException();
	   }
        } catch (InterruptedException e) {
            return "Interrupted";  // SwingWorker.get() returns this
        } catch (Throwable t) { // treated later in finished()
            System.err.println(t);
            t.printStackTrace();
            return "Error";
	} finally{
	    time = System.currentTimeMillis()-time;
	    Debug.out("Strategy stopped.");
	    Debug.out("Applied ", countApplied);
	    Debug.out("Time elapsed: ", time);
        }
        return "Done.";
    }
    

    private void init(int maxSteps) {
        proof = iaProver.getProof();
	services = mediator().getServices();
        maxApplications = maxSteps;
        countApplied = 0;
        userTerminated = false;
        goalChooser.init(proof);
        pushCount = 0;
        popCount = 0;
    }
    
    private KeYMediator mediator() {
        return iaProver.mediator();
    }
    


    public void start(ListOfGoal goals, int maxSteps) {
        start(goals, maxSteps, false); // justPopFrame = false
    }
    
    
    public void start(ListOfGoal goals, int maxSteps, boolean justPopFrame) {
        init(maxSteps);
        this.justPopFrame = justPopFrame;
        autoModeActive = true;
        setupGoals ( goals );

        worker = new AutoModeWorker();
        if ( ( startedAsInteractive = !mediator().autoMode() ) ) 
           mediator().stopInterface(true);
        iaProver.setInteractive(false);
        proof.setRuleAppIndexToAutoMode ();
	setupStatus ();
        worker.start();
    }

    /**
     * Called by the "Stop" button, interrupts the worker thread 
     * which is running this.doWork().  Note that the doWork() 
     * method handles InterruptedExceptions cleanly.
     */
    public void stop () {
        userTerminated = true;
        worker.interrupt();
    }
    
    
    
    private void setupGoals ( ListOfGoal goals ) {
        goalChooser.setupGoals ( goals );
    }


    /**
     * Starts the automatic mode synchronously (in the same thread).
     * Currently this is done for update-simpl. step after every
     * rule application. This way the caller (e.g. the loader)
     * always sees the final result after update simplification.
     */
    public void startSync(ListOfGoal goals, int maxSteps) {
        init(maxSteps);
        autoModeActive = true;
        setupGoals ( goals );

        worker = new AutoModeWorker() {
           public Object get() { // avoid a Thread.join(); we're same thread
              return getValue();
           }
        };
        if ( ( startedAsInteractive = !mediator().autoMode() ) ) 
           mediator().stopInterface(true);
        iaProver.setInteractive(false);
        proof.setRuleAppIndexToAutoMode ();
//	setupStatus ();
        worker.construct();
        worker.finished();
    }

    

    
    /* Invoking start() on the SwingWorker causes a new Thread
     * to be created that will call construct(), and then
     * finished().  Note that finished() is called even if
     * the worker is interrupted because we catch the
     * InterruptedException in doWork().
     */
    private class AutoModeWorker extends SwingWorker {
    
        public Object construct() {
            Object res = doWork();
	    return res;
        }

        public void finished() {
            autoModeActive = false;
	    Main main = Main.getInstance ();
	    removeStatus ();
	    String timeString = "" + (time/1000)+"."+((time%1000)/100);

            if("Error".equals(get())) {
                if(!Main.batchMode)
                {
	       mediator().startInterface(true);
	       JOptionPane.showMessageDialog(
        	  iaProver.mediator().mainFrame(),
        	  "An exception occured while executing the strategy.",
        	  "Error", JOptionPane.ERROR_MESSAGE);
                }else
                {// Error in batchMode. Terminate with status -1.
                    System.exit(-1);
                }
	    }
            proof.setRuleAppIndexToInteractiveMode ();
            iaProver.setInteractive(true);

            String message;
            int closed = mediator().getNrGoalsClosedByHeuristics();

            // display message in the status bar

            if ( countApplied != 0 ) {
	        message = "Strategy: Applied " + countApplied + " rule";
	        if ( countApplied != 1 )
	            message += "s";
	        if ( closed > 0 ) {
	            message += ", closed " + closed + " goal";
	            if ( closed != 1 )
		        message += "s";			    
	        }
	        message += " (" + timeString + " sec) ";
                message += proof.openGoals().size()+" remaining"; 
	        main.setStatusLine ( message );
            }

            ReuseListener rl = mediator().getReuseListener();
            if (rl != null) rl.recomputeReuse(); 
            if (startedAsInteractive) mediator().startInterface(true);

            if(Main.batchMode)
            {
                main.saveProof(Main.fileNameOnStartUp+".proof"); //Save the proofe before exit.
                if(proof.openGoals().size()==0)
	                System.exit(0); //Says that all Proofs have succeeded
                else
                    System.exit(1); //Says that there is at leat one open Proof
            }

            if ( proof.openGoals().size()==0 &&  startedAsInteractive) {
	        // show message dialog

	        // inform that automatic mode has been stopped
	        message = "Strategy stopped"+
	            (userTerminated ? "by user.\n" : ".\n");
	        if (countApplied == maxApplications) {
	            message = "Maximal number of rule applications" +
		        "has been reached.\n";
	        }

	        // Show message ...

		if (closed>0) {
		    message += closed+
			((closed>1) ? " goals have" : " goal has") +
			" been closed.\n";
		}
		message += "Applied " + countApplied + " rules in " +
		    timeString + " seconds.";
		if (proof.openGoals().size()==0) {
		    message="Proved."+"\nStatistics:\n"+proof.statistics();
		}
		if ((closed>0) || (countApplied == maxApplications) 
		    || (proof.openGoals().size()==0)) {
		    JOptionPane.showMessageDialog
			(iaProver.mediator ().mainFrame(), message, "Note ...", 
			 JOptionPane.INFORMATION_MESSAGE);
		}
	    }
	    mediator().resetNrGoalsClosedByHeuristics();
        }

    }
    
    
    
    
    private class ProofListener implements InteractiveProofListener {

	/** invoked if automatic execution has started
	 */
	public void autoModeStarted(ProofEvent e) {}

	/** invoked if automatic execution has stopped
	 */
	public void autoModeStopped(ProofEvent e) {}
    
	/** invoked when a rule has been applied */
	public void ruleApplied(ProofEvent e) {
            if (!autoModeActive) return;
	    RuleAppInfo rai = e.getRuleAppInfo ();
	    if ( rai == null )
		return;

	    synchronized ( ApplyHeuristics.this ) {
		ListOfGoal                newGoals = SLListOfGoal.EMPTY_LIST;
		IteratorOfNodeReplacement it       = rai.getReplacementNodes ();
		Node                      node;
		Goal                      goal;

		while ( it.hasNext () ) {
		    node = it.next ().getNode ();
		    goal = proof.getGoal ( node );
		    if ( goal != null )
			newGoals = newGoals.prepend ( goal );
		}

                goalChooser.updateGoalList ( rai.getOriginalNode (), newGoals );
	    }
	}
	
    }    

}
