// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.gui;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Metavariable;
import java.util.*;

/**
 * An InitialPositionTable is a PositionTable that describes the
 * beginning of the element/subelement relationship. Thus, an
 * InitialPositionTable describes the information on where the
 * semisequents of a sequent are located. It is the root of the tree of
 * PositionTables and may be asked for a PosInSequent for a given index
 * position and a given Sequent.
 *
 * <p>For simplicity, the an InitialPositionTable has only one row.
 * The various constrained formulae of a sequent are located 
 * one level below.  In other words the whole sequent is not 
 * represented by an empty position list but by the list [0].
 */
public class InitialPositionTable extends PositionTable{

    /**
     * creates a new Initial PositionTable.
     */
    public InitialPositionTable() {
	super(1);
    }

    /**
     * Returns the PosInSequent for a given char position in a 
     * sequent.
     * @param index the char position that points to the wanted
     *              position in sequent
     * @param filter the sequent print filter from that was used to
     *              print the sequent
     *
     */
    public PosInSequent getPosInSequent(int index,
					SequentPrintFilter filter) {
	if ( index < startPos[0] || index >= endPos[0]) {
	    return null;
	}

	ListOfInteger posList = pathForIndex(index);

	PosInSequent pis = getTopPIS(posList,filter);

	Range r = rangeForIndex(index);
	pis.setBounds(r);
	Range firstStatement 
	    = firstStatementRangeForIndex(index);
	if ( firstStatement!=null ) {
	    pis.setFirstJavaStatementRange(firstStatement);
	}
	return pis;
    }

    /** Returns a PosInSequent for a given position list, 
     * but without filling in the bounds.  It is assumed
     * that this is the top level position table for a sequent.
     * @param posList the position list that navigates through
     *                the position tables.
     * @param filter  the sequent print filter from that was used to
     *                print the sequent
     */
    private PosInSequent getTopPIS(ListOfInteger posList,
				   SequentPrintFilter filter) {
	if (posList.isEmpty() || posList.tail().isEmpty()) {
	    return PosInSequent.createSequentPos();	
	} else {
	    return child[0].getSequentPIS(posList.tail(),filter);
	}
    }


    /** Returns the path for a given PosInOccurrence.  This is 
     * built up from the initial 0, the number of the 
     * ConstrainedFormula in the sequent, the position in the 
     * constrained formula, and possibly inside a Metavariable
     * instantiation. */
    public ListOfInteger pathForPosition(PosInOccurrence pio,
					 SequentPrintFilter filter) {
	ListOfInteger p = SLListOfInteger.EMPTY_LIST;
	
	p = prependPathBelowMV(p,pio);
	p = prependPathInFormula(p,pio);
	p = p.prepend(new Integer(indexOfCfma(pio.constrainedFormula(),
					      filter)));
	p = p.prepend(new Integer(0));
	return p;
    }

    private ListOfInteger prependPathBelowMV(ListOfInteger p,
					    PosInOccurrence pio) {
	if (pio.posInTermBelowMetavariable() != null) {
	    IntIterator pit = pio
		.posInTermBelowMetavariable()
		.reverseIterator();
	    while (pit.hasNext()) {
		p = p.prepend(new Integer(pit.next()));
	    }
	}
	return p;
    }

    private ListOfInteger prependPathInFormula(ListOfInteger p,
					       PosInOccurrence pio) {
	IntIterator pit = pio.posInTerm().reverseIterator();
	while (pit.hasNext()) {
	    p = p.prepend(new Integer(pit.next()));
	}
	return p;
    }
    

    /** Returns the index of the constrained formula in the sequent
     * as printed. */
    private int indexOfCfma(ConstrainedFormula cfma,
			    SequentPrintFilter filter) {
	ListOfSequentPrintFilterEntry list =
	    filter.getAntec().append(filter.getSucc());
	int k;
	for ( k=0 ; !list.isEmpty(); k++,list = list.tail() ) {
	    if (list.head().getOriginalFormula()==cfma) {
		return k;
	    }
	}
	return -1;
    }

    /**
     * Returns the character range of the `lowest' subtable that
     * includes <code>index</code> in its range.
     * @param index the character index to search for.
     */
    public Range rangeForIndex(int index) {
	return rangeForIndex(index,endPos[0]);
    }

    /** Returns the character range for the subtable indicated
     * by the given integer list.
     */
    public Range rangeForPath(ListOfInteger path) {
	return rangeForPath(path,endPos[0]);
    }


}


