// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.gui;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;
import java.util.HashSet;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.RuleAppIndex;
import de.uka.ilkd.key.proof.SLListOfGoal;
import de.uka.ilkd.key.proof.ListOfNode;
import de.uka.ilkd.key.proof.SLListOfNode;
import de.uka.ilkd.key.proof.TacletFilter;
import de.uka.ilkd.key.proof.proofevent.NodeChangeJournal;
import de.uka.ilkd.key.proof.proofevent.RuleAppInfo;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.rule.*;
import org.apache.log4j.Logger;

public class InteractiveProver implements java.io.Serializable {

    /** the proof the interactive prover works on */
    private Proof proof;

    /** the user focused goal */
    private Goal focusedGoal;

    /** true iff in interactive mode */

    private boolean interactive = true;

    /**
     * Roots of subtrees containing all nodes to which rules have been
     * applied; this is used when auto mode is active
     */
    private ListOfNode modifiedSubtrees      = null;
    private HashSet    modifiedSubtreesCache = null;

    /** the central heuristics processor including GUI signalling */
    private ApplyHeuristics applyHeuristics;

    // list of proof listeners and interactive proof listeners
    private List listenerList = Collections.synchronizedList(new ArrayList(10));


    /** listens to the current selected proof and node */
    private KeYSelectionListener selListener;

    /** the mediator */
    private KeYMediator mediator;
    
    private boolean autoHeuristicsOn = false;

    /** 
     * mode of automatic update simplification: true: runs in implicit mode,
     * this means no new proof nodes are created for simplification steps
     */
    private boolean implicitUpdateSimplification = ProofSettings.THE_SETTINGS.
	getSimultaneousUpdateSimplifierSettings().implicitExecution();
    
    
    private static Logger threadLogger = Logger.getLogger("key.threading");
    
    

    /** creates a new interactive prover object 
     */
    public InteractiveProver(KeYMediator mediator) {
	selListener = new InteractiveProverKeYSelectionListener();
	this.mediator = mediator;
	mediator.addKeYSelectionListener(selListener);
	applyHeuristics = new ApplyHeuristics(this);
    }

    /** returns the KeYMediator */
    KeYMediator mediator() {
	return mediator;
    }

    /** sets up a new proof 
     * @param p a Proof that contains the root of the proof tree
     */
    public void setProof(Proof p) {
	proof = p;
    }

    /** 
     * sets the update simplification mode   
     * @param implicit a boolean that if it is true, inhibits the creation of
     * proof nodes for update simplification steps
     */
    public void setUpdateSimplificationMode(boolean implicit) {
	this.implicitUpdateSimplification = implicit;
    }

    /**
     * @return true iff updates should be simplified implicitly
     */
    public boolean getUpdateSimplificationMode () {
	return implicitUpdateSimplification;
    }

    /**
     * In auto mode, add a node which has been modified in a way
     * leading to structural changes of the proof tree
     */
    private void addModifiedNode ( Node p_node ) {
	try {
	    if ( modifiedSubtrees != SLListOfNode.EMPTY_LIST ) {
		Node n = p_node;
		while ( true ) {
		    if ( modifiedSubtreesCache.contains ( n ) )
			return;
		    if ( n.root () )
			break;
		    n = n.parent ();
		}
	    }

	    modifiedSubtrees = modifiedSubtrees.prepend ( p_node );	
	} finally {
	    modifiedSubtreesCache.add ( p_node );
	}
    }

    /** adds an interactive ProofListener
     * @param p the InteractiveProofListener to be added
     */
    public synchronized void addInteractiveProofListener
	(InteractiveProofListener p) { 
	synchronized(listenerList) {
	    listenerList.add(p);
	}
    }

    /** removes an interactive ProofListener
     * @param p the InteractiveProofListener to be removed
     */
    public synchronized void removeInteractiveProofListener
	(InteractiveProofListener p) { 
	synchronized(listenerList) {	
	    listenerList.remove(p);
	}
    }

    /** fires the event that automatic execution has started */
    protected synchronized void fireAutoModeStarted(ProofEvent e) {
	if ( modifiedSubtrees == null ) {
	    modifiedSubtrees      = SLListOfNode.EMPTY_LIST;
	    modifiedSubtreesCache = new HashSet ();
	}

	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((InteractiveProofListener)it.next()).
		    autoModeStarted(e);
	    }
	}
    }

    /** fires the event that automatic execution has stopped */
    private synchronized void fireAutoModeStopped(ProofEvent e) {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((InteractiveProofListener)it.next()).
		    autoModeStopped(e);
	    }
	}
    }
    

    protected synchronized void fireAutoModeStopped () {
	ProofEvent e;
	if (proof == null) // loader failed
	    e = new ProofEvent(this);
	else if ( modifiedSubtrees != null )
	    e = new ProofEvent ( this,
				 modifiedSubtrees );
	else
	    e = new ProofEvent ( this,
				 SLListOfNode.EMPTY_LIST.prepend
				 ( getProof ().root () ) );

	fireAutoModeStopped ( e );

	modifiedSubtrees      = null;
	modifiedSubtreesCache = null;	
    }


    public void setAutoHeuristics(boolean b) {
       autoHeuristicsOn = b;
    }
 

    public boolean autoHeuristics() {
	return autoHeuristicsOn;
    }

    public boolean interactiveMode() {
	return interactive;
    }

    public void setInteractive(boolean interact) {
	interactive=interact;
    }

    /** fires the event that a rule has been applied */
    protected synchronized void fireRuleApplied( ProofEvent p_e ) {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((InteractiveProofListener)it.next()).
		    ruleApplied(p_e);
	    }
	}
    }    

    /** The rule application is performed on the given goal
     * @param app the RuleApp to apply 
     * @param goal the goal where to apply the rule application
     */
    public void apply(RuleApp app, Goal goal) {
	apply ( app, goal, true,  interactive );
    }

    public void apply(RuleApp app, Goal goal, boolean applyAutoHeuristics) {
	apply( app, goal, applyAutoHeuristics, interactive );
    }


    public void apply(RuleApp app, Goal goal, boolean applyAutoHeuristics,
                      boolean interactive) {
	Debug.out(" >>>>> INTERACTIVE APPLICATION <<<< ", app);
	Debug.outIfEqual("auto_insert_eq applied", 
			 app.rule().name(), "auto_insert_eq");

	Services services = mediator().getServices();

	if (app instanceof TacletApp) {
	    app = ((TacletApp)app).instantiateWithMV ( goal );
	    app = ((TacletApp)app).createDependingFunctions 
		( mediator ().func_ns () );
	}

	if ( modifiedSubtrees != null )
	    addModifiedNode ( goal.node () );

	NodeChangeJournal journal     = new NodeChangeJournal ( getProof (),
								goal );

	ListOfGoal        goalList    = app.execute(goal, services); 

	if (app instanceof BuiltInRuleApp) {
	    // A goal specific rule application does not always
	    // return a result
	    BuiltInRuleApp gsApp = (BuiltInRuleApp)app;
	    if (!gsApp.success()) { 
		if (!"".equals(gsApp.errorMessage())) {
		    mediator().popupInformationMessage
			(gsApp.errorMessage(), "Note...");
		}
		return;
	    }
	}
	// application was successful

	if ( goalList == SLListOfGoal.EMPTY_LIST ) {
	    proof.closeGoal ( goal, app.constraint () );	    	 
	}
	else {
	    proof.replace   ( goal, goalList );
	    if ( app instanceof TacletApp &&
		 ((TacletApp)app).taclet ().closeGoal () )
		// the first new goal is the one to be closed
		proof.closeGoal ( goalList.head (),
				  app.constraint () );
	}

    
	if ( getUpdateSimplificationMode() ) {
	    // in this case the simplification of the updates is regarded as
	    // part of the rule application
	    if ( interactive && !( applyAutoHeuristics && autoHeuristicsOn ) )
	        applyUpdateSimplifier ( goalList );
	}
    

	final RuleAppInfo ruleAppInfo = journal.getRuleAppInfo ( app );

        threadLogger.warn("Applying rule"); // shows curr. thread
	fireRuleApplied ( new ProofEvent ( this, ruleAppInfo ) );
        

	if (interactive) {
            if ( !getProof().closed () ) {
// all heuristics if both are true, otherwise simplifier only
                if ( !( applyAutoHeuristics && autoHeuristicsOn ) ) {
                    // only apply the update simplifier following an
                    // interactive rule application
                    applyUpdateSimplifier ( goalList );
                } else
                    startAutoMode ();
// this is a patchwork solution, we need better indication of who is
// calling apply: user, loader, reuse, etc....                                     
            } else {
	        mediator().indicateNoReuse();
	        if (javax.swing.SwingUtilities.isEventDispatchThread())
		    fireAutoModeStopped
			(new ProofEvent(this,
					SLListOfNode.EMPTY_LIST.prepend
					(proof.root ())));
	    }
        }
    }

    
    /**
     * HACK for the application of simplification rules after each interactive
     * rule application
     */
    private void applyUpdateSimplifier (Goal goal, boolean antec) {
        final Node node = goal.node ();
        final Sequent seq = node.sequent ();
        final Constraint userConstraint =
            node.proof ().getUserConstraint ().getConstraint ();

        IteratorOfConstrainedFormula it = ( antec ? seq.antecedent ()
                                                  : seq.succedent () ).iterator ();

        final BuiltInRule rule = UpdateSimplificationRule.INSTANCE;

        while ( it.hasNext () ) {
            final ConstrainedFormula cfma = it.next ();
            final PosInOccurrence pos = new PosInOccurrence ( cfma,
                                                              PosInTerm.TOP_LEVEL,
                                                              seq,
                                                              antec );

            if ( rule.isApplicable ( goal, pos, userConstraint ) ) {
                BuiltInRuleApp app = new BuiltInRuleApp ( rule,
                                                          pos,
                                                          userConstraint );
                applyUpdateSimplificationRule ( goal, app );
                
                // recursive call to find further applications
                applyUpdateSimplifier ( goal, antec );
                return;
            }
        }
    }

    private void applyUpdateSimplifier (Goal goal) {
        applyUpdateSimplifier ( goal, true );
        applyUpdateSimplifier ( goal, false );
    }

    private void applyUpdateSimplifier (ListOfGoal goalList) {
        final IteratorOfGoal it = goalList.iterator ();
        while ( it.hasNext() )
            applyUpdateSimplifier ( it.next () );
    }

    public void applyUpdateSimplificationRule (Goal goal, BuiltInRuleApp rapp) {
        if ( getUpdateSimplificationMode () ) {
            PosInOccurrence pio = rapp.posInOccurrence ();
            ConstrainedFormula conFormula = pio.constrainedFormula ();
            ConstrainedFormula simplified =
                goal.simplifier ().simplify ( conFormula,
                                              mediator ().getUserConstraint ()
                                                         .getConstraint () );
            goal.changeFormula ( simplified, pio );
        } else
            apply ( rapp, goal, false, false );
    }

    /** returns the proof the interactive prover is working on
     * @return the proof the interactive prover is working on
     */
    public Proof getProof() {
	return proof;
    }    

    /** starts the execution of rules with active strategy 
     * @param maxApplicationOfHeuristics maximal allowed automatic
     * applications of Taclets
     */
    public void startAutoMode () {
	startAutoMode( proof.openGoals () );
    }


    /** starts the execution of rules with active strategy. The
     * strategy will only be applied on the goals of the list that
     * is handed over and on the new goals an applied rule adds
     */
    public void startAutoMode ( ListOfGoal goals ) {
	/*
         * if (activeHeuristics() == SLListOfHeuristic.EMPTY_LIST) {
         * mediator().popupWarning("No heuristics selected", "Oops..."); return; }
         */
        if ( goals == SLListOfGoal.EMPTY_LIST ) {
            if(Main.batchMode)
            {
                System.exit(0); //Everything is allready proved.
                //Noting to be saved. Exit successfuly.
            } else {
		mediator ().popupWarning ( "No (more) goals", "Oops..." );
		return;
            }
        }
	if(Main.batchMode) {
	    interactive = false;
            //Allow much more Steps in batchMode then in regular mode.
	    applyHeuristics.start(goals, 
				  mediator ().getMaxAutomaticSteps ()*100);
	} else {
	    applyHeuristics.start(goals, mediator ().getMaxAutomaticSteps ());
	}
    }
    
    /** stops the execution of rules */
    public void stopAutoMode () {
        applyHeuristics.stop();
    }


    /**
     * collects all built-in rules that are applicable at the
     * given sequent position 'pos'.
     * @param pos the PosInSequent where to look for applicable rules 
     * @param userConstraint the user defined constraint
     */
    public ListOfBuiltInRule getBuiltInRule(PosInSequent pos, 
						 Constraint   userConstraint) {
	ListOfBuiltInRule rules = SLListOfBuiltInRule.EMPTY_LIST;
	IteratorOfRuleApp it = 
	    getInteractiveRuleAppIndex ().getBuiltInRule
	    (focusedGoal, pos.getPosInOccurrence(), userConstraint).iterator();

	while (it.hasNext()) {
	    BuiltInRule r = (BuiltInRule) it.next().rule();
	    if (!rules.contains(r)) {
		rules = rules.prepend(r);
	    }
	}
	return rules;
    }

    /**
     * @return the <code>RuleAppIndex</code> of the goal currently focussed,
     *         after setting the index to unrestricted (non-automated mode)
     */
    private RuleAppIndex getInteractiveRuleAppIndex () {
        final RuleAppIndex index = focusedGoal.ruleAppIndex ();
        index.autoModeStopped ();
        return index;
    }

    /**
     * collects all built-in rule applications for the given rule that are
     * applicable at position 'pos' and the current user constraint
     * 
     * @param rule
     *            the BuiltInRule for which the applications are collected
     * @param pos
     *            the PosInSequent the position information
     * @param userConstraint
     *            the user defined constraint
     * @return a SetOfRuleApp with all possible rule applications
     */
    public SetOfRuleApp getBuiltInRuleApp(BuiltInRule rule, 
					  PosInSequent     pos,
					  Constraint       userConstraint) {
	
	SetOfRuleApp result = SetAsListOfRuleApp.EMPTY_SET;
	IteratorOfRuleApp it = getInteractiveRuleAppIndex ().
	    getBuiltInRule(focusedGoal, 
			   pos.getPosInOccurrence(),
			   userConstraint).iterator();

	while (it.hasNext()) {
	    RuleApp app = it.next();
	    if (app.rule() == rule) {
		result = result.add(app);
	    }
	}

	return result;
    }
    
    
    /** collects all applicable NoFindTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable NoFindTaclets
     */
    ListOfTacletApp getNoFindTaclet() {	
	return filterTaclet(getInteractiveRuleAppIndex ().
		       getNoFindTaclet(TacletFilter.TRUE,
				       mediator.getServices(),
				       mediator.getUserConstraint().getConstraint()));
    }    

    /** collects all applicable FindTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable FindTaclets
     */

    ListOfTacletApp getFindTaclet(PosInSequent pos) {
	if (!pos.isSequent() && (focusedGoal!=null) && (pos!=null)) {
	    return filterTaclet(getInteractiveRuleAppIndex ().
			      getFindTaclet(TacletFilter.TRUE,
	 	                            pos.getPosInOccurrence(),
		                            mediator.getServices(),
					    mediator.getUserConstraint().getConstraint()));
	}
	return SLListOfTacletApp.EMPTY_LIST;
    }
    
    /** collects all applicable RewriteTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable RewriteTaclets
     */
    ListOfTacletApp getRewriteTaclet(PosInSequent pos) {
	if (!pos.isSequent())  {
	    return filterTaclet(getInteractiveRuleAppIndex ().
		   getRewriteTaclet(TacletFilter.TRUE,
				    pos.getPosInOccurrence(),
				    mediator.getServices(),
				    mediator.getUserConstraint().getConstraint())); 
	}

	return SLListOfTacletApp.EMPTY_LIST;
    }



    /** collects all Taclet applications at the given position of the specified
     * taclet
     * @param pos the PosInOccurrence describing the position
     * @param taclet the Taclet
     * @return a list of all found rule applications of the given rule at
     * position pos  
     */
    protected SetOfTacletApp getAppsForName(Goal goal, String name, 
            PosInOccurrence pos) {
        return getAppsForName(goal, name, pos, TacletFilter.TRUE);
    }
    
    protected SetOfTacletApp getAppsForName(Goal goal, String name, 
                                            PosInOccurrence pos,
                                            TacletFilter filter) {


	SetOfTacletApp result = SetAsListOfTacletApp.EMPTY_SET;
        ListOfTacletApp fittingApps = SLListOfTacletApp.EMPTY_LIST;
        final RuleAppIndex index          = goal.ruleAppIndex();
	final Constraint   userConstraint =
            mediator.getUserConstraint().getConstraint();
	if ( pos == null ) {
            final IteratorOfNoPosTacletApp it =
                index.getNoFindTaclet ( filter,
                                        mediator.getServices(),
                                        userConstraint ).iterator ();
            while ( it.hasNext () )
                fittingApps = fittingApps.prepend ( it.next () );
        } else 
            fittingApps = index.getTacletAppAt ( filter,
                                                 pos, 
			                         mediator.getServices(),
					         userConstraint );

	IteratorOfTacletApp it = fittingApps.iterator();
	// filter fitting applications
	while (it.hasNext()) {
	    TacletApp app = it.next();
	    if (app.rule().name().toString().equals(name) ){
		result = result.add(app);
	    }
	}
//if (result.size()==0) System.err.println("Available was "+fittingApps);
	return result;
    }
    
    /** listener to KeYSelection Events in order to be informed if the
     * current proof changed
     */
    private class InteractiveProverKeYSelectionListener 
	implements KeYSelectionListener {

	/** focused node has changed */
	public void selectedNodeChanged(KeYSelectionEvent e) {
	    focusedGoal = e.getSource().getSelectedGoal();
	}

	/** the selected proof has changed (e.g. a new proof has been
	 * loaded) */ 
	public void selectedProofChanged(KeYSelectionEvent e) {
	    Debug.out("InteractiveProver: initialize with new proof");
            Proof newProof = e.getSource().getSelectedProof();
	    setProof(newProof); // this is null-safe
            if (newProof != null) {
	        focusedGoal = e.getSource().getSelectedGoal();
            } else {
                focusedGoal = null;
            }
	}

    }

    /**
     * takes NoPosTacletApps as arguments and returns a duplicate free list of
     * the contained TacletApps
     */
    private ListOfTacletApp filterTaclet(ListOfNoPosTacletApp tacletInstances) {
		java.util.HashSet applicableRules = new java.util.HashSet();
		ListOfTacletApp result = SLListOfTacletApp.EMPTY_LIST;
		IteratorOfNoPosTacletApp it = tacletInstances.iterator();		
		while (it.hasNext()) {
		    TacletApp app = it.next ();
	            
	            if (mediator().stupidMode()) {
	                ListOfTacletApp ifCandidates = app.findIfFormulaInstantiations(
	                   mediator().getSelectedGoal().sequent(),
	                   mediator().getServices(),
	                   mediator().getUserConstraint().getConstraint());
	                if (ifCandidates.size() == 0) continue; // skip this app
	            }
	
		    // for the moment, just remove taclets which are
		    // inconsistent with user constraint
		    if ( mediator ().getUserConstraint ().getConstraint ()
			 .join ( app.constraint () ).isSatisfiable () ) {
				Taclet taclet = app.taclet();
				if (!applicableRules.contains(taclet)) {
				    applicableRules.add(taclet);
				    result = result.prepend(app);
				}
		    }
		}
	
       	return result;
    }
    
}
