// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.util.pp.*;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.util.Debug;

import java.util.Stack;
import java.util.StringTokenizer;
import java.io.IOException;
import java.util.ArrayList;
import java.util.ListIterator;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.io.StringWriter;

import org.apache.log4j.Logger;


public class OCLPrinter {

    public static final int LINE_WIDTH = 55;
    private LogicPrinter logicPrinter;
    protected Layouter layouter;
    static Logger logger = Logger.getLogger(OCLPrinter.class.getName());

    public OCLPrinter(LogicPrinter lp) {
	logicPrinter = lp;
	layouter = lp.layouter;
    }


    public void printOCLExpression(String name, Term t) throws IOException {
	if (name.equals("forAll")  
	    || name.equals("iterate")
	    || name.equals("asSet")
	    || name.equals("includes")
	    || name.equals("excludes")
	    || name.equals("count")
	    || name.equals("includesAll")
	    || name.equals("excludesAll")
	    || name.equals("isEmpty")
	    || name.equals("notEmpty")
	    || name.equals("sum")
	    || name.equals("exists")
	    || name.equals("isUnique")
	    || name.equals("sortedBy")
	    || name.equals("any")
	    || name.equals("one")
	    || name.equals("union")
	    || name.equals("intersection")
	    || name.equals("including")
	    || name.equals("excluding")
	    || name.equals("symmetricDifference")
	    || name.equals("select")
	    || name.equals("reject")
	    || name.equals("collect")
	    || name.equals("asSequence")
	    || name.equals("asBag")
	    || name.equals("append")
	    || name.equals("prepend")
	    || name.equals("subSequence")
	    || name.equals("at")
	    || name.equals("first")
	    || name.equals("last")
	    || name.equals("size")) {
	    printArrowExpr(name, t);
	} else if (name.equals("allSubtypes")
		   || name.equals("allInstances")
		   || name.equals("name")
		   || name.equals("attributes")
		   || name.equals("associationEnds")
		   || name.equals("operations")
		   || name.equals("supertypes")
		   || name.equals("allSupertypes")
		   || name.equals("oclIsKindOf")
		   || name.equals("oclIsTypeOf")
		   || name.equals("oclAsType")
		   || name.equals("oclIsNew")
		   || name.equals("evaluationType")
		   || name.equals("abs")
		   || name.equals("floor")
		   || name.equals("round")
		   || name.equals("max")
		   || name.equals("min")
		   || name.equals("div")
		   || name.equals("mod")
		   || name.equals("concat")
		   || name.equals("toUpper")
		   || name.equals("toLower")
		   || name.equals("substring")) {
	    printDotExpr(name, t); 
	} else if (name.equals("and") 
		   || name.equals("or")
		   || name.equals("xor")
		   || name.equals("implies")) {
	    logicPrinter.printInfixTerm(t.sub(0), 10, name, t.sub(1), 10);
	} else if (name.equals("equals")) {
	    logicPrinter.printInfixTerm(t.sub(0), 10, "=", t.sub(1), 10);
	} else if (name.equals("notEquals")) {
	    logicPrinter.printInfixTerm(t.sub(0), 10, "<>", t.sub(1), 10);
	} else if (name.equals("lessThan")) {
	    logicPrinter.printInfixTerm(t.sub(0), 10, "<", t.sub(1), 10);
	} else if (name.equals("greaterThan")) {
	    logicPrinter.printInfixTerm(t.sub(0), 10, ">", t.sub(1), 10);
	} else if (name.equals("lessThanOrEqual")) {
	    logicPrinter.printInfixTerm(t.sub(0), 10, "<=", t.sub(1), 10);
	} else if (name.equals("greaterThanOrEqual")) {
	    logicPrinter.printInfixTerm(t.sub(0), 10, ">=", t.sub(1), 10);
	} else if (name.equals("add")) {
	    logicPrinter.printInfixTerm(t.sub(0), 10, "+", t.sub(1), 10);
	} else if (name.equals("subtract")) {
	    logicPrinter.printInfixTerm(t.sub(0), 10, "-", t.sub(1), 10);
	} else if (name.equals("multiply")) {
	    logicPrinter.printInfixTerm(t.sub(0), 10, "*", t.sub(1), 10);
	} else if (name.equals("divide")) {
	    logicPrinter.printInfixTerm(t.sub(0), 10, "/", t.sub(1), 10);
	} else if (name.equals("if")) {
	    printIfThenElseExpr(name, t);
	} else if (name.equals("OclWrapper")) {
	    logicPrinter.startTerm(t.arity());
	    logicPrinter.markStartSub();
	    logicPrinter.printTerm(t.sub(0));
	    logicPrinter.markEndSub();
	} else if (name.equals("cons")) {
	    layouter.print("Set{");
	    printCollectionExpr(t);
	    layouter.print("}");
	} else if (name.equals("nil")) {
	    layouter.print("Set{}");
	} else {
	    logicPrinter.printFunctionTerm(name, t);
	}
    }		  

    private void printArrowExpr(String name, Term t) throws IOException {
	logicPrinter.startTerm(t.arity());
	layouter.beginC(1);
	logicPrinter.markStartSub();
	logicPrinter.printTerm(t.sub(0));
	logicPrinter.markEndSub();
	layouter.print("->").brk(0,1).print(name);
	layouter.print("(").beginC(0);
	if (t.varsBoundHere(1).size() != 0) {
	    ArrayOfQuantifiableVariable vars = t.varsBoundHere(1);
	    QuantifiableVariable v = vars.getQuantifiableVariable(0);		    
	    layouter.print("").print(v.name().toString());
	    layouter.print(":").print(v.sort().toString()).print(" | ");
	    logicPrinter.markStartSub();
	    logicPrinter.printTerm(t.sub(1));
	    logicPrinter.markEndSub();
	}
	else if (t.varsBoundHere(2).size() != 0) {
	    ArrayOfQuantifiableVariable vars = t.varsBoundHere(2);
	    QuantifiableVariable v1 = vars.getQuantifiableVariable(0);
	    QuantifiableVariable v2 = vars.getQuantifiableVariable(1);
	    layouter.print("").print(v1.name().toString());
	    layouter.print(":").print(v1.sort().toString()).print("; ");		
	    layouter.print(v2.name().toString());
	    layouter.print(":").print(v2.sort().toString()).print(" = ");
	    logicPrinter.markStartSub();
	    logicPrinter.printTerm(t.sub(1));
	    logicPrinter.markEndSub();
	    layouter.print(" | ");
	    logicPrinter.markStartSub();
	    logicPrinter.printTerm(t.sub(2));
	    logicPrinter.markEndSub();
	}
	layouter.print(")").end().end();
    }

    private void printDotExpr(String name, Term t) throws IOException {
	logicPrinter.startTerm(t.arity());
	layouter.beginC(0);
	logicPrinter.markStartSub();
	logicPrinter.printTerm(t.sub(0));
	logicPrinter.markEndSub();
	layouter.print(".").print(name);
	layouter.print("(").beginC(0);
	for (int i=1;i<t.arity();i++) {
	    logicPrinter.markStartSub();
	    logicPrinter.printTerm(t.sub(i));
	    logicPrinter.markEndSub();
		
	    if (i<t.arity()-1) {
		layouter.print(",").brk(1,0);
	    }
	}
	layouter.print(")").end().end();
    }

    private void printIfThenElseExpr(String name, Term t) throws IOException {
	logicPrinter.startTerm(3);
	layouter.beginC(0);
	layouter.brk(1,3).print("if ");
	logicPrinter.markStartSub();
	logicPrinter.printTerm(t.sub(0));
	logicPrinter.markEndSub();
	layouter.brk(1,4).print("then ");
	logicPrinter.markStartSub();
	logicPrinter.printTerm(t.sub(1));
	logicPrinter.markEndSub();
	layouter.brk(1,4).print("else ");
	logicPrinter.markStartSub();
	logicPrinter.printTerm(t.sub(2));
	logicPrinter.markEndSub();
	layouter.brk(1,3).print("endif").end();
    }

    private void printCollectionExpr(Term t) throws IOException {
	logicPrinter.startTerm(2);
	logicPrinter.markStartSub();
	logicPrinter.printTerm(t.sub(0));
	logicPrinter.markEndSub();
	if (t.sub(1).op().name().toString().equals("$nil")) {
	    logicPrinter.markStartSub();
	    layouter.print("");
	    logicPrinter.markEndSub();
	} else if (t.sub(1).op().name().toString().equals("$cons")) {
	    layouter.print(", ");
	    logicPrinter.markStartSub();
	    printCollectionExpr(t.sub(1));
	    logicPrinter.markEndSub();
	} else {
	    layouter.print(", ");
	    logicPrinter.markStartSub();
	    logicPrinter.printTerm(t.sub(1));
	    logicPrinter.markEndSub();
	}
    }

    /*
    private void printCollectionExpr(Term t) throws IOException {
	List elems = extractCollectionElements(new LinkedList(), t);
	logicPrinter.startTerm(elems.size());
	Iterator iter = elems.iterator();
	layouter.print("Set{");
	while (iter.hasNext()) {
	    Term elem = (Term)iter.next();
	    logicPrinter.markStartSub();
	    logicPrinter.printTerm(elem);
	    //layouter.print(elem.op().name().toString().substring(1));
	    logicPrinter.markEndSub();
	    if (iter.hasNext()) {
		layouter.print(", ");
	    }
	    else {
		layouter.print("}");
	    }
	}
    }

    private List extractCollectionElements(List list, Term t) {
	if (t.op().name().toString().equals("$nil")) {
	    return list;
	} else {
	    list.add(t.sub(0));
	    return extractCollectionElements(list, t.sub(1));
	}
    }
    */
}
