// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.gui;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringBufferInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;
import java.util.Properties;
import javax.swing.JMenuItem;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.KeYResourceManager;

/** This class is used to load and save settings for proofs such as
 * which data type models are used to represent the java types. Which
 * heuristics have to be loaded and so on.
 * The class loads the file proofsettings.config from the place where
 * you started key. If the file is not available standard settings are
 * used.
 * The loaded file has the following structure:
 * <code>
 * // KeY-Configuration file
 * ActiveHeuristics=simplify_prog , simplify
 * MaximumNumberOfHeuristcsApplications=400
 * number  = IntegerLDT.class 
 * boolean = BooleanLDT.class
 * </code>
 */
public class ProofSettings {


    public static final File PROOF_CONFIG_FILE;
    public static final URL PROOF_DEFAULT_CONFIG;
    public static final ProofSettings THE_SETTINGS;

    static {
	PROOF_CONFIG_FILE = new File
	    (System.getProperty("user.home")+
	     File.separator+".key"+File.separator+"proof-settings.props");
	PROOF_DEFAULT_CONFIG =
	    KeYResourceManager.getManager ().getResourceFile
	    ( ProofSettings.class, "default-proof-settings.props" );
	THE_SETTINGS = new ProofSettings();
    }

    private boolean initialized=false;


    /** all setting objects in the following order: heuristicSettings */
    private Settings[] settings;

    /** the default listener to settings */
    private ProofSettingsListener listener = new ProofSettingsListener();
    /** is set to true if a setting object has changed but the proof
     * settings have not been saved yet
     */
    private boolean changed;

    /** the item displayed in the main options menu */
    private JMenuItem proofSettingsItem = new JMenuItem("Save Settings");
    
    /** create a proof settings object */
    public ProofSettings() {       	
	setChanged(false);
	settings = new Settings[] {
            new StrategySettings(),
	    new LDTSettings(),
	    new OCLRefinementSettings(),
	    new ModelSourceSettings(),
	    new SimultaneousUpdateSimplifierSettings(),
            new GeneralSettings(),
	    new ChoiceSettings(),
	    new DecisionProcedureSettings(),
	    new ViewSettings(),
	};
	for (int i = 0; i < settings.length; i++) { 
	    settings[i].addSettingsListener(listener);
	}
    }

    public void ensureInitialized() {
	if (!initialized) {
	    loadSettings();
	    initialized=true;	
	}
    }

    
    /** returns the item displayed in the main options menu */
    public JMenuItem item() {
	return proofSettingsItem;
    }
    
    /** sets the changed variable */
    private void setChanged(boolean chg) {
	changed = chg;
	proofSettingsItem.setEnabled(changed);
    }
	/** Used by saveSettings() and settingsToString()
    */
    public void settingsToStream(Settings[] s,OutputStream out) {
    try {
        Properties result = new Properties();
	    for (int i = 0; i < s.length; i++) {
	    s[i].writeSettings(result);
	    }
	    result.store(out, "Proof-Settings-Config-File");
	    setChanged(false);
	} catch (IOException e){
	    System.err.println("Warning: could not save proof-settings.");
	    System.err.println(e);
	    Debug.out(e);
	}
    }
    /**
     * Saves the current settings in this dialog into a configuration file.
     */
    public void saveSettings() {
	try {
	    if (!PROOF_CONFIG_FILE.exists()) {
		new File(System.getProperty("user.home")+File.separator
			 +".key"+File.separator).mkdir();
	    }
	    FileOutputStream out = 
		new FileOutputStream(PROOF_CONFIG_FILE);
        settingsToStream(settings,out);
	} catch (IOException e){
	    System.err.println("Warning: could not save proof-settings.");
	    System.err.println(e);
	    Debug.out(e);
	}
    }

	public String settingsToString() {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        settingsToStream(settings,out);
        return new String(out.toByteArray());
    }

    /** Used by loadSettings() and loadSettingsFromString(...) */
    public void loadSettingsFromStream(InputStream in) {
    Properties defaultProps = new Properties ();

    if ( PROOF_DEFAULT_CONFIG == null )
	    System.err.println
	    ("Warning: default proof-settings file could not be found.");
	else {
	    try {
	    defaultProps.load ( PROOF_DEFAULT_CONFIG.openStream () );
	    } catch (IOException e){
	    System.err.println
	        ("Warning: default proof-settings could not be loaded.");
	    Debug.out(e);
	    }
	}

    Properties props = new Properties ( defaultProps );
    try {
        props.load(in);
	} catch (IOException e){
            System.err.println
	    ("Warning: no proof-settings could be loaded, using defaults");
	    Debug.out(e);
	}

    for (int i = settings.length-1; i>=0 ;i--) { 
	    settings[i].readSettings(props);  
	}

    setChanged(false);

    }
    /**
     * Loads the the former settings from configuration file.
     */
    public void loadSettings(){
	try {
	    FileInputStream in = new FileInputStream(PROOF_CONFIG_FILE);
	    loadSettingsFromStream(in);
	} catch (IOException e){
            System.err.println
		("Warning: no proof-settings could be loaded, using defaults");
	    Debug.out(e);
	}
    }

    /** Used to load Settings from a .key file*/
    public void loadSettingsFromString(String s) {
        StringBufferInputStream in = new StringBufferInputStream(s);
        loadSettingsFromStream(in);
    }

    /** returns the StrategySettings object
     * @return the StrategySettings object
     */
    public StrategySettings getStrategySettings() {
        ensureInitialized();
        return (StrategySettings) settings[0];
    }

    /** returns the ChoiceSettings object
     * @return the ChoiceSettings object
     */
    public ChoiceSettings getChoiceSettings() {
	ensureInitialized();
	return (ChoiceSettings) settings[6];
    }

    /** returns the DecisionProcedureSettings object
     * @return the DecisionProcedureSettings object
     */
    public DecisionProcedureSettings getDecisionProcedureSettings() {
	ensureInitialized();
	return (DecisionProcedureSettings) settings[7];
    }

    /** sets the LDTSettings */
    public void setLDTSettings(LDTSettings ldts){
	settings[1]=ldts;
    }

    /** returns the LDTSettings object
     * @return the LDTSettings object
     */
    public LDTSettings getLDTSettings() {
	ensureInitialized();
	return (LDTSettings) settings[1];
    }

    /** returns the OCLRefinementSettings object
     * @return the OCLRefinementSettings object
     */
    public OCLRefinementSettings getOCLRefinementSettings() {
	ensureInitialized();
	return (OCLRefinementSettings) settings[2];
    }


    public ModelSourceSettings getModelSourceSettings() {
	ensureInitialized();
	return (ModelSourceSettings) settings[3];
    }

    public SimultaneousUpdateSimplifierSettings 
	getSimultaneousUpdateSimplifierSettings() {
	ensureInitialized();
	return (SimultaneousUpdateSimplifierSettings) settings[4];    
    }

    public GeneralSettings getGeneralSettings() {
	ensureInitialized();
	return (GeneralSettings) settings[5];
    }

    public ViewSettings getViewSettings() {
	ensureInitialized();
	return (ViewSettings) settings[8];
    }

    
    /** returns true iff the settings have not been saved yet but a
     * setting object has been changed 
     * @return  true iff the settings have not been saved yet but a
     * setting object has been changed
     */
    public boolean changed() {
	return changed;
    }


    private class ProofSettingsListener implements SettingsListener {
	
	ProofSettingsListener() {
	}

	/** called by the Settings object to inform the listener that its
	 * state has changed
	 * @param e the Event sent to the listener
	 */
	public void settingsChanged(GUIEvent e) {	    
	    setChanged(true);
	}

    }
    
}
