// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.rule.*;
import java.util.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.util.Debug;
import org.apache.log4j.Logger;


public class ReuseListener implements InteractiveProofListener,
                                      KeYSelectionListener {

   private KeYMediator medi;
   private Logger reuseLogger = Logger.getLogger("key.proof.reuse");
   
   public ReuseListener(KeYMediator medi) {
      this.medi = medi;
   }
   
   public void selectedNodeChanged(KeYSelectionEvent e) {
      if (medi.reuseInProgress()) return; // we are still in the loop
      Node target = e.getSource().getSelectedNode();
      if (medi.getProof().getGoal(target) == null) return; // inner node
      Node targetParent = target.parent();
      if (targetParent == null) return;
      ReusePoint rp = targetParent.getReuseSource();
      if (rp == null) return;
      int siblingNr = targetParent.getChildNr(target);
      Node source = rp.source().child(siblingNr);
      try{
         Main.getInstance().setStatusLine("Template proof continues with: "+
                                          source.name());
//         source.getAppliedRuleApp().rule().name());
      } catch (NullPointerException npe) {
	  // there is no ruleApp at this point in the source proof (=open goal)
	  // forget about the hint then
	  Debug.out("Exception thrown by class ReuseListener at getInstance()");
      }
   
   }

   public void autoModeStarted(ProofEvent e) {}
   public void autoModeStopped(ProofEvent e) {}

   public void selectedProofChanged(KeYSelectionEvent e) {
      recomputeReuse();
   }

   public void ruleApplied(ProofEvent e) {
//      recomputeReuse();
   }



   public void recomputeReuse() {
      if (Node.reuseCandidatesNumber()==0) { // optimize the stuff below!
         medi.indicateNoReuse();
         return;
      }
   
      Vector reusePoints = new Vector(10,10);

      IteratorOfGoal it = medi.getProof().openGoals().iterator();
      reuseLogger.info("***********************************************");         
      while (it.hasNext()) { // iterate over goals, more or less
         analyzeCandidates(it.next(), reusePoints);
         //if (reusePoints.size() > 0) break; no, get all!
      }
      pickBest(reusePoints);
   }


   
   private void pickBest(Vector reusePoints) {
      reuseLogger.warn(">Possible applications detected: "+reusePoints.size());
      if (reusePoints.size() > 0) {
         java.util.Collections.sort(reusePoints);
         ReusePoint best = (ReusePoint) reusePoints.firstElement();
         if (best.notGoodEnough()) { 
            medi.indicateNoReuse();
            return;
         }
         medi.indicateReuse(best);
      } else {
         medi.indicateNoReuse(); // no reuse points
      }
   }

   
   public void nextStepCadidates(Node used) {
      used.unmarkReuseCandidate();
      IteratorOfNode it = used.childrenIterator();
      while (it.hasNext()) {
         Node n = it.next();
         if ("Closed goal".equals(n.name())) return; // this is ugly
         if ("OPEN GOAL".equals(n.name())) return;
         n.markReuseCandidate();
      }
   }
   
   
   void analyzeCandidates(Goal target, Vector result) {
   reuseLogger.info("***** (next goal) candidate nodes in system: "+
   Node.reuseCandidatesNumber());
      Iterator it = Node.reuseCandidatesIterator();
      while (it.hasNext()) { // iterate candidates
         ReusePoint blank = null;
         FindTaclet t; // other taclets?
         try {
            blank = new ReusePoint((Node)it.next(), target);
            t = (FindTaclet) blank.getApp().rule();
         } catch(NullPointerException e) {
            continue; // open goal, no app
         } catch(ClassCastException e) {
            String ruleName = (blank == null) ? 
                "null" : blank.getApp().rule().getClass().toString();
            reuseLogger.error("Cannot re-use "+ruleName);
            continue; // not a FindTaclet
         }
         
         applicableWhere(blank, medi.getServices(),
           medi.getUserConstraint().getConstraint(), result);
      }         
   }
   

    public void applicableWhere(ReusePoint blank, Services services, 
                                Constraint userC, Vector result) {
       Goal g = blank.target();
       Rule rule = blank.getApp().rule();

       IteratorOfConstrainedFormula formIter;
       if (rule instanceof AntecTaclet) 
          formIter = g.sequent().antecedent().iterator();
       else
       if (rule instanceof SuccTaclet) 
          formIter = g.sequent().succedent().iterator();
       else
          formIter = g.sequent().iterator();
       // but only if pos taclet
       while (formIter.hasNext()) {
          ConstrainedFormula cfma = formIter.next();
          PosInOccurrence start = 
             new PosInOccurrence(cfma, PosInTerm.TOP_LEVEL, g.sequent());
          recMatch(start, cfma.formula(), rule, result, blank, services, userC);
       }
    }

    private void recMatch(PosInOccurrence pos, Term t, Rule rule, Vector result,
                          ReusePoint blank, Services services, Constraint userC) {
       if (isApplicable(rule, t, services, userC)) {
          ReusePoint p = blank.initialize(pos, medi);
          if (p != null) { // RuleApp can be transferred
             result.add(p);
          }
       }
       // Succ- and Antec- are only applicable at top-level of formula
       if (!(rule instanceof RewriteTaclet)) return;
       int i;
       for (i=0; i<t.arity(); i++) {
          recMatch(pos.down(i), t.sub(i), rule, result, blank, services, userC);
       }
    }
    
    
    private boolean isApplicable(Rule rule, Term t, 
                                 Services services, Constraint userC) {
       if (rule instanceof FindTaclet)
          return
              ((FindTaclet)rule).matchFind
                  (t, MatchConditions.EMPTY_MATCHCONDITIONS, services, userC)
              .success();
       else return false;
       
    }
    
    
   


}
