// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.gui;

/** This class displays a DL-formula. It    
* (a) highlights the subpart of the sequent that is nearest to the mouse    
* cursor    
* (b) pops up a menu with all applicable Taclets at the highlighted    
* position    
*/

import java.awt.Font;
import java.awt.Color;
import java.awt.Point;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*;

import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.util.Debug;


public class SequentView extends JEditorPane {

    // the current tag of the highlight
    private Object currentHighlight;

    private Object additionalHighlight;
    private Highlighter.HighlightPainter additionalHighlightPainter;

    private Object additionalHighlight2;
    private Highlighter.HighlightPainter additionalHighlightPainter2;

    // the used sequentprinter
    private LogicPrinter printer;

    // the used sequent filter
    private SequentPrintFilter filter;

    // the used sequent
    private Sequent seq;

    // the mediator
    private KeYMediator mediator;

    // the mouse/mouseMotion listener
    protected SequentViewListener listener;
    
    // an object that detects opening and closing of an Taclet instantiation dialog
    private GUIListener guiListener;

    // enables this component to be a Drag Source
    DragSource dragSource = null;

    /** 
     * creates a viewer for a sequent 
     * @param printer the Sequentprinter to use
     */
    public SequentView(KeYMediator mediator) {
	/* setting this to text/html causes the position tables to go wrong */
	super("text/plain","");
	setMediator(mediator);
	// view cannot be edited
	setEditable(false);
	// disables selection
	setSelectionColor(getBackground());
	// sets the painter for the highlightning
	setHighlighter(new DefaultHighlighter());	
	//sets initial highlight (not visible) and stores its tag
	try {
	    // REMARK: THE ORDER FOR ADDING HIGHLIGHTS IS CRUCIAL
	    // WHEN USING OVERLAPPING HIGHLIGHTS:
	    // Adding highlight H1 before highlight H2 ensures that 
	    // H1 can overwrite overlapping parts of H2 !!!

	    additionalHighlightPainter = 
		new DefaultHighlighter.DefaultHighlightPainter(Color.lightGray);
	    additionalHighlight = getHighlighter().
		addHighlight(0, 0, additionalHighlightPainter); 
	    
	   
	    currentHighlight = getHighlighter().
		addHighlight(0, 0,
			     new DefaultHighlighter.
				 DefaultHighlightPainter(Color.yellow)); 

	   
	} catch(BadLocationException badLocation) {
	    System.err.println("SequentView tried to highlight an area"
			       +"that is not existing.");
	    System.err.println("Exception:"+badLocation);
	}	

	Config.DEFAULT.addConfigChangeListener(
	    new ConfigChangeListener() {
		    public void configChanged(ConfigChangeEvent e) {
			updateUI();
		    }
		});
	setSequentViewFont();

	listener = new SequentViewListener(this, mediator());
	addMouseListener(listener);
	addMouseMotionListener(listener);
	addMouseListener(listener);

	
	guiListener = new GUIListener(){
		/** invoked if a frame that wants modal access is opened */
		public void modalDialogOpened(GUIEvent e){
		 

		    if (e.getSource() instanceof ApplyTacletDialog){
			// enable drag'n'drop ...
			listener.setModalDragNDropEnabled(true);
			listener.setRefreshHighlightning(true);

		    } else {
			// disable drag'n'drop and highlightning ...
			listener.setModalDragNDropEnabled(false);
			listener.setRefreshHighlightning(false);
		    }
		}
		
		/** invoked if a frame that wants modal access is closed */
		public void modalDialogClosed(GUIEvent e){
		    if (e.getSource() instanceof ApplyTacletDialog){
			// disable drag'n'drop ...
			listener.setModalDragNDropEnabled(false);
			listener.setRefreshHighlightning(true);
		    } else {
			listener.setRefreshHighlightning(true);
		    }
		}
		
		/** invoked if the user wants to abort the proving session */
		public void shutDown(GUIEvent e){
		}	
	    };
	
	
	mediator().addGUIListener(guiListener); 

	// and here comes the drag'n'drop stuff that allows the user to copy
	// the currently highlighted subterm/formula by mere drag'n'dop actions
	
	dragSource = new DragSource();
	
	dragSource.createDefaultDragGestureRecognizer( this, 
						       DnDConstants.ACTION_COPY,
						       listener );

	// And now the Drag'n'drop stuff ...
	DropTarget aDropTarget = 
	    new DropTarget(this, new QuantifierInstDropTargetListener(this));
	
        
	this.setDropTarget(aDropTarget);
	

    }

    protected SequentViewListener getSeqViewListener() {
	return listener;
    }

    protected DragSource getDragSource() {
	return dragSource;
    }
    
    public void updateUI() {
	super.updateUI();
	setSequentViewFont();
    }

    public void setSequentViewFont() {
	Font myFont = UIManager.getFont(Config.KEY_FONT_CURRENT_GOAL_VIEW);
	if (myFont != null) {
	    setFont(myFont);
	} else {
	    Debug.out("KEY_FONT_CURRENT_GOAL_VIEW not available. Use standard font.");
	}
    }


    /** sets the text being printed */
    public synchronized void printSequent() {
	if ( SwingUtilities.isEventDispatchThread () ) {
	    printSequentImmediately ();
	} else {
	    Runnable sequentUpdater = new Runnable() {
		    public void run() {
			printSequentImmediately ();
		    }
		};
	    SwingUtilities.invokeLater(sequentUpdater);
	}
    }

    /** sets the text being printed */
    public synchronized void printSequentImmediately() {
	removeMouseMotionListener(listener);
	removeMouseListener(listener);
	printer.update(seq, filter);
	boolean errorocc;
	do {
	    errorocc=false;
	    try {
		setText(printer.toString());
	    } catch (Error e) {
		System.err.println("Error occurred while printing Sequent!");
		errorocc=true;
	    }
	} while (errorocc);
	repaint();
	addMouseMotionListener(listener);
	addMouseListener(listener);
    }
        
    /** sets the LogicPrinter to use
     * @param sp the LogicPrinter that is used
     */
    public void setPrinter(LogicPrinter sp, Sequent s) {
    	setPrinter ( sp, null, s );
    }

    // xxx remove?
    protected SequentPrintFilter getSequentPrintFilter() {
    	return filter;
    }

    /** sets the LogicPrinter to use
     * @param sp the LogicPrinter that is used
     * @param f the SequentPrintFilter that is used
     */
    public void setPrinter(LogicPrinter sp, SequentPrintFilter f, Sequent s) {
		printer = sp;
		filter  = f;
		seq = s;
    }

    /** return used LogicPrinter
     * @returns the LogicPrinter that is used
     */
    public LogicPrinter printer() {
    	return printer;
    }

    /** sets the mediator
     * @param mediator the KeYMediator used to communicate with other
     * components 
     */
    private void setMediator(KeYMediator mediator) {
		Debug.assertTrue(mediator != null, "Mediator must be set");
		this.mediator = mediator;
    }
    

    /** the startposition and endposition+1 of the string to be
     * highlighted
     * @param p the mouse pointer coordinates
     */
    public void paintHighlights(Point p) {
	try {	
	    Range firstStatement = getFirstStatementRange(p);
	    // Change highlight for additional Java statement ...
	    if (firstStatement != null) {
		getHighlighter()
		    .changeHighlight(additionalHighlight, 
				    firstStatement.start(),
				    firstStatement.end());
	    } else {
		getHighlighter()
		    .changeHighlight(additionalHighlight, 0, 0);
	    }


	    Range highlight = getHighlightRange(p);
	    // Change Highlighter for currently selected sequent part 
	    if (highlight!=null) {
		getHighlighter()
		    .changeHighlight(currentHighlight, 
				     highlight.start(),
				     highlight.end());
	    } else {
		getHighlighter()
		    .changeHighlight(currentHighlight, 0, 0);
	    }

	} catch(BadLocationException badLocation) {
	    System.err.println("SequentView tried to highlight an area"
			       +"that is not existing.");
	    System.err.println("Exception:"+badLocation);
	}	
    }


    /** returns the mediator of this view
     * @return the KeYMediator
     */
    public KeYMediator mediator() {
	return mediator;
    }

    /** selected rule to apply 
     * @param taclet the selected Taclet
     * @param pos the PosInSequent describes the position where to apply the 
     * rule 
     */
    void selectedTaclet(TacletApp taclet, PosInSequent pos) {
    	mediator().selectedTaclet(taclet, pos);
    }

    /**
     * Enables drag'n'drop of highlighted subterms in the sequent window.
     * @param allowDragNDrop enables drag'n'drop iff set to true.
     */
    public synchronized void setModalDragNDropEnabled(boolean allowDragNDrop){
	    listener.setModalDragNDropEnabled(allowDragNDrop);
	}
    
    /**
     * Checks whether drag'n'drop of highlighted subterms in the sequent window
     * currently is enabled..
     * @param return true iff drag'n'drop is enabled.
     */	
    public synchronized boolean modalDragNDropEnabled(){
	return listener.modalDragNDropEnabled();
    }
    
    public String getHighlightedText() {
       String s;
       try{
	   PosInSequent pos = listener.getMousePos();
	   s = getText(pos.getBounds().start(),
		       pos.getBounds().length());
       } catch (Exception ble) { // whatever it is -- forget about it
	   s="";
       }
       return s;
    }
    
	
    public PosInSequent getMousePosInSequent() {
       return listener.getMousePos();
    }


    /** Get a PosInSequent object for a given coordinate of the 
     * displayed sequent. */
    protected synchronized PosInSequent getPosInSequent(Point p) {
	String seqText = getText();
	if (seqText.length() > 0) {
	    int characterIndex = correctedViewToModel(p);
	    
	    return printer().getPositionTable().
		getPosInSequent(characterIndex, filter);
	} else {
	    return null;
	}
    }

    /** Get the character range to be highlighted for the given
     * coordinate in the displayed sequent. */
    protected synchronized Range getHighlightRange(Point p) {
	String seqText = getText();
	if (seqText.length() > 0) {
	    int characterIndex = correctedViewToModel(p);
	    
	    return printer().getPositionTable().
		rangeForIndex(characterIndex);
	} else {
	    return null;
	}
    }

    /** Get the character range to be highlighted for the first
     * statement in a java block at the given
     * coordinate in the displayed sequent.  Returns null
     * if there is no java block there.*/
    protected synchronized Range getFirstStatementRange(Point p) {
	String seqText = getText();
	if (seqText.length() > 0) {
	    int characterIndex = correctedViewToModel(p);
	    
	    return printer().getPositionTable().
		firstStatementRangeForIndex(characterIndex);
	} else {
	    return null;
	}
    }

    /** Return the character index for a certain coordinate.  The
     * usual viewToModel is focussed on inter-character spaces, not
     * characters, so it returns the correct index in the
     * left half of the glyph but one too many in the right half.
     * Therefore, we get width of character before the one given
     * by viewToModel, substract it from the given x value, and get
     * the character at the new position. That is the correct one.  
     */
    private int correctedViewToModel(Point p) {
	String seqText = getText();

	int cursorPosition = viewToModel(p);
	
	cursorPosition -= ( cursorPosition > 0 ? 1 : 0);
	
	cursorPosition = (cursorPosition >= seqText.length() 
			  ? seqText.length()-1 :  
			  cursorPosition);
	cursorPosition = ( cursorPosition >= 0 ? cursorPosition : 0);
	int previousCharacterWidth =
	    getFontMetrics(getFont()).charWidth
	    (seqText.charAt(cursorPosition));	    	    
	
	int characterIndex = viewToModel
	    (new Point( (int)p.getX() - (previousCharacterWidth/2), 
			    (int)p.getY())); 

	return characterIndex;
    }
    
}






