// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.java;

import de.uka.ilkd.key.java.reference.IExecutionContext; 
import de.uka.ilkd.key.util.ExtList;
import java.io.IOException;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.logic.op.SchemaVariable; 

/** 
 * In the DL-formulae description of Taclets the program part can have
 * the following form < pi alpha;...; omega > Phi where pi is a prefix
 * consisting of open brackets, try's and so on and omega is the rest
 * of the program. Between the prefix pi and the postfix omega there
 * can stand an arbitrary program. This pattern is realized using this
 * class.
 */

public class ContextStatementBlock extends StatementBlock {

    /** contains the middle part of the context term */    
    private final StatementBlock body;

    /** is true if the context terms has a prefix and postfix */
    private boolean hasPrefix;   

    /** 
     * the last execution context of the context term 
     */
    private IExecutionContext executionContext;

    /** creates a ContextStatementBlock 
     * @param body an Extlist  representing the middle part of
     * the ContextStatementBlock
     * @param hasPrefix a boolean indicating if the ContextStatementBlock
     * has a pi and an omega
     */
    public ContextStatementBlock(ExtList children) {
	super(children);
	body = new StatementBlock(children); 
    }

    /** creates a ContextStatementBlock 
     * @param children the body of the context term
     * @param executionContext the required execution context
     */
    public ContextStatementBlock(ExtList children, 
		       IExecutionContext executionContext) {
	super(children);
	body = new StatementBlock(children); 
	this.executionContext = executionContext;
    }


    /** returns true if the prefix of the ContextStatementBlock is not empty
     * this means if one exists
     * @return true if the prefix of the ContextStatementBlock is not empty
     * this means if one exists
     */
    public boolean hasPrefix() {
	return hasPrefix;
    }

    public boolean requiresExplicitExecutionContextMatch() {
	return executionContext != null;
    }

    public IExecutionContext getExecutionContext() {
	return executionContext;
    }

    /** returns the middle part( the body) of the context term
     * @return the body of the context term
     */
    public StatementBlock getMidBody() {	
	return body;
    }

    public int getChildPositionCode(ProgramElement child) {
        // role 1 (IDX): statements
        if (executionContext != null) {
	    if (child == executionContext) {
		return 0;
	    }
	}
	if (body != null) {
	    if (child == body) {
		return 1;
	    }
        }
        return -1;
    }

    /**
     * Returns the number of children of this node.
     * @return an int giving the number of children of this node
    */

    public int getChildCount() {
	int count = 0;
	if (executionContext != null) count++;
	if (body != null) count++;
        return count;
    }

    /**
     * Returns the child at the specified index in this node's "virtual"
     * child array
     * @param index an index into this node's "virtual" child array
     * @return the program element at the given position
     * @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     * of bounds
     */
    public ProgramElement getChildAt(int index) {
	if (executionContext != null) {
	    if (index == 0) {
		return executionContext;
	    } 
	    index--;
	}
	if (body != null) {
	    if (index == 0) {
		return body;
	    }
	    index--;
	}
	throw new ArrayIndexOutOfBoundsException();
    }


    /**
     *      Get the number of statements in this container.
     *      @return the number of statements.
     */

    public int getStatementCount() {
        return (body != null) ? 1 : 0;
    }

    /*
      Return the statement at the specified index in this node's
      "virtual" statement array.
      @param index an index for a statement.
      @return the statement with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */

    public Statement getStatementAt(int index) {
	return body;
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnContextStatementBlock(this);
    }

    public void prettyPrint(PrettyPrinter w) throws IOException {
	w.printContextStatementBlock(this);
    }
    
    /* toString */
    public String toString() {
	StringBuffer result = new StringBuffer();
	result.append((hasPrefix ? ".." : ""));
	result.append(body.toString());
	result.append("\n");
	result.append((hasPrefix ? "..." : ""));
	return result.toString();
    }
    
    public ArrayOfStatement getBody() {
        throw new UnsupportedOperationException(getClass()+
            ": We are not quite a StatementBlock");
    }
    
    public boolean isEmpty() {
        return body.isEmpty();
    }

    public int getTypeDeclarationCount() {
        throw new UnsupportedOperationException(getClass()+
            ": We are not quite a StatementBlock");
    }

    public de.uka.ilkd.key.java.declaration.TypeDeclaration 
                                            getTypeDeclarationAt(int index) {
        throw new UnsupportedOperationException(getClass()+
            ": We are not quite a StatementBlock");
    }

    public SourceElement getFirstElement() {
        return body.getFirstElement();
    }
    
}
