// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.java;

import java.util.HashMap;
import java.math.BigInteger;
import de.uka.ilkd.key.util.Debug;

/** 
 * This class is used for the equals modulo renaming method in
 * SourceElement. The purpose of this class is to abstract from
 * names. Therefore it represents a mapping o1 x o2 -> abstractName
 * where o1, o2 are of the same type (Label or ProgramVariable). The
 * objectif is that the comparing method uses this new name for o1 and
 * o2 instead of their real name. For this comparision a method is
 * offered so that the assigned name is not given outside.
 */
public class NameAbstractionTable {

    /** 
     * an internal counter to ensure unique names. Is used as the
     * abstract name of the elements added to this table. If the
     * equality of two elements is tested we use object identity for the
     * abstract name instances.
     */
    private BigInteger nameCounter;

    /** map containing entries like (el, name) */
    private HashMap nameMap;

    
    /** creates a new NameAbstractionTable
     */
    public NameAbstractionTable() {
	nameCounter = BigInteger.ZERO;
	nameMap = new HashMap();
    }

    /** 
     * adds the given two elements to the table and assigns them the
     * same abstract name. A precondition of this method is that the
     * elements have not been added before. 
     * @param pe1 SourceElement to be added
     * @param pe2 SourceElement to be added     
     */
    public void add(SourceElement pe1, SourceElement pe2) {
	Debug.assertTrue(!nameMap.containsKey(pe1) && !nameMap.containsKey(pe2),
			 "nameabstractiontable: Element has been assigned "+
			 "a name before.");
	BigInteger name = nameCounter;
	nameMap.put(pe1, name);
	nameMap.put(pe2, name);
	nameCounter = nameCounter.add(BigInteger.ONE);
    }

    /** 
     * tests if the given elements have been assigned to the same
     * abstract name. 
     * @param pe1 SourceElement 
     * @param pe2 SourceElement 
     * @returns true if the pe1 and pe2 have been assigned to the same
     * name
     */
    public boolean sameAbstractName(SourceElement pe1, SourceElement pe2) {
	return (nameMap.containsKey(pe1) && nameMap.containsKey(pe2) && 
		nameMap.get(pe1) == nameMap.get(pe2));
    }

    /** 
     * returns true if an abstract name has been assigned to this
     * element 
     * @return true if an abstract name has been assigned to this
     * element 
     */
    public boolean contains(SourceElement pe) {
	return nameMap.containsKey(pe);
    }

    /** 
     * toString
     */
    public String toString() {
	return "NameAbstractionTable: "+nameMap;
    }

}
