// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.java;

import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceServiceConfiguration;
import de.uka.ilkd.key.java.recoderext.SchemaCrossReferenceServiceConfiguration;
import de.uka.ilkd.key.util.Debug;

/**
 * this is a collection of common services to the KeY prover. Services
 * include information on the underlying Java model and a converter to
 * transform Java program elements to logic (where possible) and back.
 */
public class Services{


    /** used to determine whether an expression is a compile-time 
     * constant and if so the type and result of the expression
     */
    private ConstantExpressionEvaluator cee;

    /** used to convert types, expressions and so on to logic elements
     * (in special into to terms or formulas)
     */
    private TypeConverter typeconverter;

    /**
     * the information object on the Java model
     */
    private JavaInfo javainfo;

    /**
     * creates a new Services object with a new TypeConverter and a new
     * JavaInfo object with no information stored at none of these.
     */

    public Services(){
	cee = new ConstantExpressionEvaluator(this);
        typeconverter = new TypeConverter(this);
        javainfo = new JavaInfo(new KeYProgModelInfo(typeconverter), this);
    }

    private Services(KeYCrossReferenceServiceConfiguration crsc, 
		     KeYRecoderMapping rec2key) {
	cee = new ConstantExpressionEvaluator(this);
	typeconverter = new TypeConverter(this);
	javainfo = new JavaInfo
	    (new KeYProgModelInfo(crsc, rec2key, typeconverter), this);
    }


     /** THIS CONSTRUCTOR IS ONLY FOR TESTS!
      * creates a Services object that contains the given JavaInfo object, a
      * type converter is created
      * @param ji the JavaInfo to use
      */
     public Services(JavaInfo ji){
         javainfo = ji;
         typeconverter = new TypeConverter(this);
     }

    /**
     * Returns the TypeConverter associated with this Services object.
     */
    public TypeConverter getTypeConverter(){
        return typeconverter;
    }

    private void setTypeConverter(TypeConverter tc) {
	typeconverter=tc;
    }


    /**
     * Returns the ConstantExpressionEvaluator associated with this Services object.
     */
    public ConstantExpressionEvaluator getConstantExpressionEvaluator(){
        return cee;
    }

    private void setConstantExpressionEvaluator(ConstantExpressionEvaluator ce) {
	cee = ce;
    }

    /**
     * Returns the JavaInfo associated with this Services object.
     */
    public JavaInfo getJavaInfo(){
        return javainfo;
    }


    private void setJavaInfo(JavaInfo ji) {
	javainfo=ji;
	if (ji.getServices()!=this) {
	    throw new IllegalStateException("Services are inconsistent.");
	}
    }

    /**
     * creates a new services object containing a copy of the java info of
     * this object and a new TypeConverter (shallow copy)
     * @return the copy
     */
    public Services copy() {
	Debug.assertTrue
	    (!(getJavaInfo().getKeYProgModelInfo().getServConf() 
	       instanceof SchemaCrossReferenceServiceConfiguration),
	     "services: tried to copy schema cross reference service config.");
	Services s = new Services
	    (getJavaInfo().getKeYProgModelInfo().getServConf(),
	     getJavaInfo().getKeYProgModelInfo().rec2key().copy());
	s.setTypeConverter(getTypeConverter().copy(s));
	return s;
    }

    /**
     * creates a new service object with the same ldt information 
     * as the actual one
     */
    public Services copyPreservesLDTInformation() {
	Debug.assertTrue
	    (!(javainfo.getKeYProgModelInfo().getServConf() 
	       instanceof SchemaCrossReferenceServiceConfiguration),
	     "services: tried to copy schema cross reference service config.");
	Services s = new Services();
	s.setTypeConverter(getTypeConverter().copy(s));
	return s;
    }
	
}





