// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.java.recoderext;

import recoder.abstraction.*;
import recoder.abstraction.Package;
import recoder.java.*;
import recoder.java.declaration.*;
import recoder.java.expression.*;
import recoder.java.expression.operator.*;
import recoder.java.reference.*;
import recoder.list.*;
import recoder.util.*;
import recoder.*;
import recoder.service.*;

import recoder.convenience.*;
import recoder.io.*;
import recoder.java.statement.*;
import recoder.java.declaration.*;
import recoder.java.expression.literal.*;
import recoder.java.expression.operator.*;
import recoder.java.reference.*;
import recoder.kit.*;

import java.io.*;
import java.util.HashMap;

import de.uka.ilkd.key.util.Debug;


public class KeYCrossReferenceSourceInfo 
    extends DefaultCrossReferenceSourceInfo {

    private HashMap names2vars = null;

    private final PrimitiveType recoderTypeSVType;

    /**
       Strict checking. Does not allow "broken links" during reference
       resolution.
     */
    public int STRICT = 0;
    
    /**
       Sloppy checking. Allows "broken links" during reference resolution.
     */
    public int SLOPPY = 1;


    public KeYCrossReferenceSourceInfo(ServiceConfiguration config) {
	super(config);
	recoderTypeSVType = new PrimitiveType("TypeSV", this);
    }

    public void setNames2Vars(HashMap names2vars){
	this.names2vars = names2vars;
    }
    
    /**
       Called by the service configuration indicating that all services
       are known. Services may now start communicating or linking among
       their configuration partners. The service configuration can be
       memorized if it has not been passed in by a constructor already.
       @param cfg the service configuration this services has been assigned to.
     */
    public void initialize(ServiceConfiguration cfg) {
	super.initialize(cfg);
	cfg.getChangeHistory().
	    removeChangeHistoryListener((ChangeHistoryListener)this);
	cfg.getChangeHistory().addChangeHistoryListener(this);
    }

    /**
	Returns the class type that contains the given program element.
	@param  context a program element.
	@return the type to which the given program element belongs
	(may be <CODE>null</CODE>).
    */
    public ClassType getContainingClassType(ProgramElement context){
	if (context instanceof TypeDeclaration) {
	    context = context.getASTParent();
	}
	do {
	    if (context instanceof ClassType) {
		return (ClassType)context;
	    }else if(context instanceof MethodCallStatement){
		return (ClassType) getType(((MethodCallStatement) context).
					      getExecutionContext()
					      .getTypeReference());
	    }
	    context = context.getASTParent();
	} while (context != null);
	return null; 
    }


    public Variable getVariable(String name, ProgramElement context) {
	updateModel();
	// look for the next variable scope equals to or parent of context
	ProgramElement pe = context;
	while (pe != null && !(pe instanceof VariableScope) &&
	       !((pe instanceof MethodCallStatement) && !(context instanceof ExecutionContext) &&
		 !(context.equals(((MethodCallStatement) pe).getResultVariable())))
		   ) {
	    context = pe;
	    pe = pe.getASTParent();
	    
	}
	if (pe == null) {
	    // a null scope can happen if we try to find a variable
	    // speculatively (for URQ resolution)
	    return null;	    
	}
	if(pe instanceof MethodCallStatement && !(context instanceof ExecutionContext) &&
	    !(context.equals(((MethodCallStatement) pe).getResultVariable()))){
	    pe = getTypeDeclaration((ClassType) getType(((MethodCallStatement) pe).
							   getExecutionContext()
							   .getTypeReference()));
	}
	VariableScope scope = (VariableScope)pe;
	Variable result;
	do {
	    result = scope.getVariableInScope(name);
	    if (result != null) {
		// must double check this result - rare cases of confusion
		// involving field references before a local variable of the
		// same name has been specified
		if (scope instanceof StatementBlock) {
		    StatementContainer cont = (StatementBlock)scope;	    
		    // we need the topmost var-scope including context,
		    // or context itself if the found scope is the topmost one
		    VariableDeclaration def = 
			((VariableSpecification)result).getParent();
		    for (int i = 0; true; i += 1) {
			Statement s = cont.getStatementAt(i);
			if (s == def) {
			    // Debug.log(">>> Not ignored: " +
			    // Format.toString("%c \"%s\" @%p", result)
			    // + " for context " +
			    // Format.toString("@%p", context));

			    // stop if definition comes first
			    break;
			}
			if (s == context) {
			    // tricky: reference before definition - must
			    // ignore the definition :(

			    // Debug.log(">>> Ignored: " +
			    // Format.toString("%c \"%s\" @%p", result)
			    // + " for context " +
			    // Format.toString("@%p", context));

			    result = null;
			    break;
			}
		    }
		}
		if (result != null) {
		    // leave _now_
		    break;
		}
	    }
	    if (scope instanceof TypeDeclaration) { 
		result = getInheritedField(name, (TypeDeclaration)scope);
		if (result != null) {
		    break;
		}
		// might want to check for ambiguity of outer class fields!!!
	    }
	    pe = scope.getASTParent();
	    while (pe != null && !(pe instanceof VariableScope) && 
		   !((pe instanceof MethodCallStatement) && !(context instanceof ExecutionContext))) {
		context = pe;  // proceed the context
		pe = pe.getASTParent();
	    }
	    if(pe instanceof MethodCallStatement && !(context instanceof ExecutionContext) && 
		!(context.equals(((MethodCallStatement) pe).getResultVariable()))){
		pe = getTypeDeclaration((ClassType) getType(((MethodCallStatement) pe).
							    getExecutionContext()
							   .getTypeReference()));
	    }
	    scope = (VariableScope)pe;
	} while (scope != null);
	// we were at the compilation unit scope, leave for good now
	if(result == null && names2vars!=null){
	    return (recoder.abstraction.Variable) names2vars.get(name);
	}
	return result;
    }


}
