// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.java.statement;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.logic.op.IProgramVariable;
import de.uka.ilkd.key.logic.op.ProgramMethod;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.reference.TypeRef;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.expression.operator.CopyAssignment;
import de.uka.ilkd.key.java.expression.literal.NullLiteral;

/**
 *  A shortcut-statement for a method body.
 */
public class CatchAllStatement extends JavaNonTerminalProgramElement
    implements Statement, NonTerminalProgramElement, Desugarable {

    private StatementBlock body;
    private ParameterDeclaration paramdecl;

    public CatchAllStatement(StatementBlock body, ParameterDeclaration paramdecl) {
	this.body = body;
	this.paramdecl = paramdecl;
    }
    
    public Statement getBody() {
	return body;
    }

    public ParameterDeclaration getParameterDeclaration() {
	return paramdecl;
    }

    /**
     *      Returns the number of children of this node.
     *      @return an int giving the number of children of this node
     */
    public int getChildCount() {
	int i=0;
	if (body!=null) i++;
	if (paramdecl!=null) i++;
	return i;
    }

    /**
     *      Returns the child at the specified index in this node's "virtual"
     *      child array.
     *      @param index an index into this node's "virtual" child array
     *      @return the program element at the given position
     *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *                 of bounds
     */
    public ProgramElement getChildAt(int index) {
	if (index==0) {
	    return paramdecl;
	}
	if (index==1) {
	    return body;
	}	
 	throw new ArrayIndexOutOfBoundsException();
    }
    
    /**
     *        Returns the index of the given child, or <CODE>-1</CODE> if there is 
     *        no such child. The child is searched for by identity:
     *        <CODE>getChildAt(getIndexOfChild(x)) == x</CODE>.
     *        @param child the exact child to look for.
     *        @return the index of the given child, or <CODE>-1</CODE>.
     */
    public int getIndexOfChild(ProgramElement child) {
	//  	if (child == methRef) return 0;
 	return -1;
    }

    /**
     *        Returns the positional code of the given child, or
     *        <CODE>-1</CODE> if there is no such child. The result
     *        contains an encoding of the relative position of the
     *        child as well as the role it has been playing in this
     *        parent element. This information is required for proper
     *        undo of transformations and is to be delivered to the
     *        detached method of the ChangeHistory.
     *        @see recoder.service.ChangeHistory#detached
     *        @param child the exact child to look for.
     *        @return the positional code of the given child, or <CODE>-1</CODE>.
     */
    public int getChildPositionCode(ProgramElement child) {
	// needs to be implemented
 	return getIndexOfChild(child);
    }
    
    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnCatchAllStatement(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
	p.printCatchAllStatement(this);
    }

    public Object desugar() {	
	IProgramVariable pv = getParameterDeclaration()
	    .getVariableSpecification().getProgramVariable();
	LocalVariableDeclaration lvd = new LocalVariableDeclaration
	    (new TypeRef(pv.getKeYJavaType()),
	     new VariableSpecification(pv, 0, NullLiteral.NULL, 
				       pv.getKeYJavaType()));
	ProgramVariable paramExc = new ProgramVariable
	    (new ProgramElementName("e"),
	     pv.getKeYJavaType());
	CopyAssignment ass = new CopyAssignment((Expression)pv, paramExc);
	ParameterDeclaration parDecl 
	    = new ParameterDeclaration(new Modifier[0],
				       new TypeRef(pv.getKeYJavaType()),
				       new VariableSpecification(pv),
				       false);
	Catch catchBranch = new Catch(parDecl, new StatementBlock(ass));
	Try tryBlock = new Try(body, new Branch[]{catchBranch});
	return new StatementBlock(new Statement[]{lvd, tryBlock});
    }

        
}
