// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.java.statement;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.IProgramVariable;
import de.uka.ilkd.key.logic.op.ProgramMethod;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.reference.TypeReference;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.java.reference.MethodReference;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.java.expression.*;

/**
 *  A shortcut-statement for a method body.
 */
public class MethodBodyStatement extends JavaNonTerminalProgramElement
    implements Statement, NonTerminalProgramElement {

    private final ReferencePrefix invocationTarget;

    private final IProgramVariable resultVar;
    
    private final ProgramMethod method;

    private final ArrayOfExpression arguments;
    

    /**
     *      Construct a method body shortcut
     *      @param pm called ProgramMethod 
     *      @param bodySource exact class where the body is declared
     */
     public MethodBodyStatement(ProgramMethod method, 
                                ReferencePrefix invocationTarget,
				IProgramVariable resultVar,
				ArrayOfExpression arguments) {
         this.method = method;
	 this.invocationTarget = invocationTarget;
	 this.resultVar = resultVar;
	 this.arguments = arguments;
     }
    
    /**
     *      MethodBodyStatement.
     *      @param list the ExtList with the children
     */
    public MethodBodyStatement(ExtList list, ArrayOfExpression args) {
	this.method = (ProgramMethod) list.get(ProgramMethod.class);
	this.resultVar = (IProgramVariable) list.get(IProgramVariable.class);
	if (resultVar!=null) {
	    list.remove(resultVar);
	}
        this.invocationTarget = (ReferencePrefix) list.get(ReferencePrefix.class);
	this.arguments = args;
    }


    /**
     *      Get method body.
     *      @return the Statement
     */
    public Statement getBody(Services services) {
	return (Statement) method.getBody();
    }

    /**
     *      Returns the number of children of this node.
     *      @return an int giving the number of children of this node
     */
    public int getChildCount() {
 	int i = arguments.size()+1;
	if (invocationTarget!=null) i++;
	if (resultVar!=null) i++;
	return i;
    }

    /**
     *      Returns the child at the specified index in this node's "virtual"
     *      child array.
     *      @param index an index into this node's "virtual" child array
     *      @return the program element at the given position
     *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *                 of bounds
     */
    public ProgramElement getChildAt(int index) {
	int i = 0;
	if (index==0) {
	    return method;
	}
	if (invocationTarget!=null) i++;
	if (index==1 && invocationTarget!=null) {
	    return invocationTarget;
	}	
 	if (index-i-1<arguments.size()) {
 	    return arguments.getExpression(index-1-i);
  	}
	if (index-i-1 == arguments.size() && resultVar!=null) {
	    return resultVar;
	}	
 	throw new ArrayIndexOutOfBoundsException();
    }
    
    /**
     *        Returns the index of the given child, or <CODE>-1</CODE> if there is 
     *        no such child. The child is searched for by identity:
     *        <CODE>getChildAt(getIndexOfChild(x)) == x</CODE>.
     *        @param child the exact child to look for.
     *        @return the index of the given child, or <CODE>-1</CODE>.
     */
    public int getIndexOfChild(ProgramElement child) {
	//  	if (child == methRef) return 0;
 	return -1;
    }

    /**
     *        Returns the positional code of the given child, or <CODE>-1</CODE> if 
     *        there is no such child. The result contains an encoding of the
     *        relative position of the child as well as the role it has been
     *        playing in this parent element. This information is required for
     *        proper undo of transformations and is to be delivered to the
     *        detached method of the ChangeHistory.
     *        @see recoder.service.ChangeHistory#detached
     *        @param child the exact child to look for.
     *        @return the positional code of the given child, or <CODE>-1</CODE>.
     */
    public int getChildPositionCode(ProgramElement child) {
	// needs to be implemented
 	return getIndexOfChild(child);
    }


    
    public boolean isStatic(Services services) {
        return method.isStatic();
    }


    /**
     * The symbolic execution context of this method call frame
     * @return prefix
     */
    public ReferencePrefix getDesignatedContext() {
        return invocationTarget;
    }

    public KeYJavaType getDesignatedContextType(Services services) {
	return services.getTypeConverter().getKeYJavaType
	    ((Expression)getDesignatedContext());
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnMethodBodyStatement(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
	p.printMethodBodyStatement(this);
    }
    
    public IProgramVariable getResultVariable() {
	return resultVar;
    }

    public ArrayOfExpression getArguments() {
	return arguments;
    }

    public KeYJavaType getBodySource() {
	return method.getContainedInKeYJavaType();
    }

    public ProgramMethod getProgramMethod() {
	return method;
    }

    public String reuseSignature(Services services) {
       return super.reuseSignature(services)+"("+getBodySource().getName()+")";
    }
    
}
