// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.java.statement;

import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.logic.ProgramPrefix;

/**
 *  Try.
 *  @author <TT>AutoDoc</TT>
 */

public class Try extends BranchStatement 
    implements StatementContainer, ProgramPrefix {

    /**
 *      Body.
     */

    private final StatementBlock body;

    /**
 *      Branches.
     */

    private final ArrayOfBranch branches;


    /**
 *      Try.
     */

    public Try() {
        this.body=null;
	this.branches=null;
    }

    /**
 *      Try.
 *      @param body a statement block.
     */

    public Try(StatementBlock body) {
        this.body=body;
	this.branches=null;
    }

    /**
 *      Try.
 *      @param body a statement block.
 *      @param branches a branch array.
     */

    public Try(StatementBlock body, Branch[] branches) {
        this.body=body;
	this.branches=new ArrayOfBranch(branches);
    }

    /**
 *      Try.
 *      @param body a statement block.
 *      @param branches a branch array.
     */

    public Try(StatementBlock body, ArrayOfBranch branches) {
        this.body=body;
	this.branches = branches;
    }

    /**
 *      Try.
 *      @param children a list with all children
     */

    public Try(ExtList children) {
        super(children);
	this.body = (StatementBlock)children.get(StatementBlock.class);
	this.branches=new
	    ArrayOfBranch((Branch[])children.collect(Branch.class));
    }



    public SourceElement getLastElement() {
        return getChildAt(getChildCount() - 1).getLastElement();
    }

    /**
 *      Returns the number of children of this node.
 *      @return an int giving the number of children of this node
    */

    public int getChildCount() {
        int result = 0;
        if (body     != null) result++;
        if (branches != null) result += branches.size();
        return result;
    }

    /**
 *      Returns the child at the specified index in this node's "virtual"
 *      child array
 *      @param index an index into this node's "virtual" child array
 *      @return the program element at the given position
 *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
 *                 of bounds
    */

    public ProgramElement getChildAt(int index) {
        if (body != null) {
            if (index == 0) return body;
            index--;
        }
        if (branches != null) {
            return branches.getBranch(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }


    public int getChildPositionCode(ProgramElement child) {
        // role 0: body
        // role 1 (IDX): branch
        if (body == child) {
            return 0;
        }
        if (branches != null) {
            int index = getArrayPos(branches,child);
            if (index >= 0) {
                return (index << 4) | 1;
            }
        }
        return -1;
    }


    /**
 *      Get body.
 *      @return the statement block.
     */

    public StatementBlock getBody() {
        return body;
    }

    /**
 *      Get the number of statements in this container.
 *      @return the number of statements.
     */

    public int getStatementCount() {
        return (body != null) ? 1 : 0;
    }

    /*
      Return the statement at the specified index in this node's
      "virtual" statement array.
      @param index an index for a statement.
      @return the statement with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */

    public Statement getStatementAt(int index) {
        if (body != null && index == 0) {
            return body;
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /**
 *      Get the number of branches in this container.
 *      @return the number of branches.
     */

    public int getBranchCount() {
        return (branches != null) ? branches.size() : 0;
    }

    /*
      Return the branch at the specified index in this node's
      "virtual" branch array.
      @param index an index for a branch.
      @return the branch with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */

    public Branch getBranchAt(int index) {
        if (branches != null) {
            return branches.getBranch(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /* Return the branch array wrapper
     * @return the array wrapper of the branches
     */
    public ArrayOfBranch getBranchList() {
	return branches;
    }

    public Branch containsAsBranch(Class cl) {
	ArrayOfBranch br=getBranchList();
	for (int i=0; i<br.size(); i++) {
	    if (cl.isInstance(br.getBranch(i))) {
		return branches.getBranch(i);
	    }
	}
	return null;
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnTry(this);
    }


    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
        p.printTry(this);
    }
}
