// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
//
//
package de.uka.ilkd.key.java.visitor;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.expression.literal.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.declaration.modifier.*;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.util.SimpleStackOfExtList;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.abstraction.Type;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.soundness.ProgramSVProxy;

/**
 * Walks through a java AST in depth-left-fist-order. 
 */
public abstract class CreatingASTVisitor extends JavaASTVisitor {

    protected static final Boolean CHANGED = new Boolean(true);

    /**  */
    protected SimpleStackOfExtList stack = new SimpleStackOfExtList();

    /** create the  CreatingASTVisitor
     * @param root the ProgramElement where to begin
     */
    public CreatingASTVisitor(ProgramElement root) {
	super(root);
    }

    /** the action that is performed just before leaving the node the
     * last time 
     */
    protected void doAction(ProgramElement node) {
	node.visit(this);
    }
    
    protected void walk(ProgramElement node) {
        ExtList l = new ExtList();
        l.add(node.getPositionInfo());
	stack.push(l);
	super.walk(node);
    }

    public String toString() {
	return stack.peek().toString();
    }

    /** 
     * called if the program element x is unchanged
     */
    protected void doDefaultAction(SourceElement x) {
	addChild(x);
    }
    

    public void performActionOnEmptyStatement(EmptyStatement x) {
	doDefaultAction(x);
    }

    public void performActionOnStatementBlock(StatementBlock x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new StatementBlock(changeList);
		}
	    };
	def.doAction(x);
    }

//eee
    public void performActionOnWhile(While x)     {
	ExtList changeList = stack.peek();
	if (changeList.getFirst() == CHANGED) {
	    changeList.removeFirst();
            PositionInfo pos = (PositionInfo) changeList.
                removeFirstOccurrence(PositionInfo.class);
	    Expression guard = ((Guard)changeList.removeFirst()).getExpression();
	    Statement body = (Statement) (changeList.isEmpty() ? 
					  null :
					  changeList.removeFirst());
	    addChild(new While(guard, body, pos));

	    changed();
	} else {
	    doDefaultAction(x);
	}
    }

//eee
    public void performActionOnDo(Do x)     {
	ExtList changeList = stack.peek();
	if (changeList.getFirst() == CHANGED) {
	    changeList.removeFirst();
            PositionInfo pos = (PositionInfo) changeList.
                removeFirstOccurrence(PositionInfo.class);
	    Statement body = (Statement) (changeList.isEmpty() ?
					  null :
					  changeList.removeFirst());
	    Expression guard = ((Guard) changeList.removeFirst()).getExpression();

	    addChild(new Do(guard, body, pos));	

	    changed();
	} else {
	    doDefaultAction(x);
	}
    }


    public void performActionOnIf(If x)     {

	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new If(changeList);
		}
	    };
	def.doAction(x);

    }

    public void performActionOnThen(Then x)     {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Then(changeList);
		}
	    };
	def.doAction(x);
    }

//eee
    public void performActionOnVariableSpecification
	(VariableSpecification x)     {
	ExtList changeList = stack.peek();
	if (changeList.getFirst() == CHANGED) {	    
	    changeList.removeFirst();	    
            PositionInfo pos = (PositionInfo) changeList.
                removeFirstOccurrence(PositionInfo.class);
	    Expression expr=null;
	    if (changeList.size()>1) {
		expr=(Expression)changeList.get(1);
	    }
	    IProgramVariable pv=(IProgramVariable)changeList.get(0);
	    addChild(new VariableSpecification(pv,
					       x.getDimensions(),
					       expr,         
					       pv.getKeYJavaType()));	
	    changed();
	} else {
	    doDefaultAction(x);
	} 

    }

//eee
    public void performActionOnFieldReference(FieldReference x)     {
	ExtList changeList = stack.peek();
	if (changeList.getFirst() == CHANGED) {	    
	    changeList.removeFirst();	    
            PositionInfo pos = (PositionInfo) changeList.
                removeFirstOccurrence(PositionInfo.class);
	    if (x.getReferencePrefix() != null) {
		Expression field = (Expression)changeList.get(1);
		if (field instanceof ProgramVariable) {
		    addChild(new FieldReference
			((ProgramVariable)field, 
			 (ReferencePrefix)changeList.get(0)));
		} else {
		    addChild(new FieldReference
			((ProgramVariable)((FieldReference)field).
			 getProgramVariable(), 
			 (ReferencePrefix)changeList.get(0)));
		}
	    } else {
		addChild(new FieldReference
		    ((ProgramVariable)changeList.get(0),
		     null));		
	    }
	    changed();
	} else {
	    doDefaultAction(x);
	}    
	
    }

    public void performActionOnSchematicFieldReference
	(SchematicFieldReference sfr) {
	performActionOnFieldReference(sfr);
    }


//eee
    public void performActionOnMethodReference(MethodReference x)     {
	ExtList changeList = stack.peek();
	if (changeList.getFirst() == CHANGED) {	    
	    changeList.removeFirst();	    
            PositionInfo pos = (PositionInfo) changeList.
                removeFirstOccurrence(PositionInfo.class);
	    ReferencePrefix rp=null;			    
	    if (x.getReferencePrefix()!=null) {
		rp = (ReferencePrefix)changeList.get(0);
	    }
	    changeList.remove(rp);
	    MethodName name = (MethodName) changeList.get(MethodName.class);
	    MethodReference mr = new MethodReference(changeList, name, rp);
	    addChild(mr);
	    changed();
	} else {
	    doDefaultAction(x);
	}    
    }    

    public void performActionOnTypeReference(final TypeReference x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new TypeRef(changeList, 
				       x.getKeYJavaType(), 
				       x.getDimensions());
		}
	    };
	def.doAction(x);
    }

    public void performActionOnElse(Else x)     {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Else(changeList);
		}
	    };
	def.doAction(x);
    }

//eee
    public void performActionOnCase(Case x)     {
	Expression e = null;
	ExtList changeList = stack.peek();
	if (changeList.getFirst() == CHANGED) {	    
	    changeList.removeFirst();	    
            PositionInfo pos = (PositionInfo) changeList.
                removeFirstOccurrence(PositionInfo.class);
	    if (x.getExpression() != null) {
		e = (Expression) changeList.removeFirst();
	    }
	    addChild(new Case(changeList, e, pos));
	    changed();
	} else {
	    doDefaultAction(x);
	}    
    }

    public void performActionOnCatch(Catch x)     {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Catch(changeList);
		}
	    };
	def.doAction(x);
    }

    public void performActionOnThrow(Throw x)     {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Throw(changeList);
		}
	    };
	def.doAction(x);
    }


    public void performActionOnTry(Try x)     {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Try(changeList);
		}
	    };
	def.doAction(x);
    }



    public void performActionOnDefault(Default x)     {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Default(changeList);
		}
	    };
	def.doAction(x);
    }


    public void performActionOnFinally(Finally x)     {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Finally(changeList);
		}
	    };
	def.doAction(x);
    }


    public void performActionOnInstanceof(Instanceof x)     {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Instanceof(changeList);
		}
	    };
	def.doAction(x);
    }


    public void performActionOnBreak(Break x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Break(changeList);
		}
	    };
	def.doAction(x);
    } 

    public void performActionOnContinue(Continue x)     {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Continue(changeList);
		}
	    };
	def.doAction(x);
    }
 
    public void performActionOnFor(For x) {
	DefaultAction def=new DefaultAction() {
		ProgramElement createNewElement(ExtList changeList) {
		    return new For(changeList);
		}
	    };
	def.doAction(x);
    }


//eee
    public void performActionOnLabeledStatement(LabeledStatement x) {
	Label l = null;
	ExtList changeList = stack.peek();
	if (changeList.getFirst() == CHANGED) {	    
	    changeList.removeFirst();	    
            PositionInfo pos = (PositionInfo) changeList.
                removeFirstOccurrence(PositionInfo.class);
	    if (x.getLabel() != null) {
		l = (Label) changeList.removeFirst();
	    }
	    addChild(new LabeledStatement(changeList, l, pos));
	    changed();
	} else {
	    doDefaultAction(x);
	}    
    } 
//eee
    public void performActionOnMethodFrame(MethodFrame x) {
	ExtList changeList = stack.peek();
	if (!changeList.isEmpty() && changeList.getFirst() == CHANGED) {	    
	    changeList.removeFirst();
            PositionInfo pos = (PositionInfo) changeList.
                removeFirstOccurrence(PositionInfo.class);

	    if (x.getChildCount() == 3) {
		addChild
		    (new MethodFrame((IProgramVariable)changeList.get(0), 
				     (IExecutionContext)changeList.get(1),
				     (StatementBlock)changeList.get(2)));

	    } else if (x.getChildCount() == 2) {
		addChild
		    (new MethodFrame(null, 
				     (IExecutionContext)changeList.get(0), 
				     (StatementBlock)changeList.get(1)));
	    } else {
		throw new IllegalStateException
		    ("Methodframe has not allowed number of children.");
	    }
	    changed();
	} else {
	    doDefaultAction(x);
	}	    
    }

    public void performActionOnMethodBodyStatement(final MethodBodyStatement x) {
	DefaultAction def=new DefaultAction() {
		ProgramElement createNewElement(ExtList changeList) {
		    PositionInfo posInfo = (PositionInfo)changeList.removeFirst();
		    changeList.removeFirst(); //program method
		    ReferencePrefix invocTarget = null;
		    if (x.getDesignatedContext()!=null) {
			invocTarget = (ReferencePrefix) changeList.removeFirst();
		    }
		    IProgramVariable resultVar = null;
		    if (x.getResultVariable()!=null) {
			resultVar = (IProgramVariable)changeList.removeLast();
		    }
		    Expression[] exps = new Expression[changeList.size()];
		    exps = (Expression[]) changeList.toArray(exps);
 		    return new MethodBodyStatement(x.getProgramMethod(), 
						   invocTarget, resultVar, 
						   new ArrayOfExpression(exps));
		}
	    };
	def.doAction(x);
    } 


    public void performActionOnSynchronizedBlock(SynchronizedBlock x)     {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new SynchronizedBlock(changeList);
		}
	    };
	def.doAction(x);
    }
 

    public void performActionOnCopyAssignment(CopyAssignment x) {
	DefaultAction def=new DefaultAction() {			
		ProgramElement createNewElement(ExtList changeList) {
		    return new CopyAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnPreIncrement(PreIncrement x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new PreIncrement(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnPostIncrement(PostIncrement x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new PostIncrement(changeList);
		}
	    };
	def.doAction(x);	
    }


    public void performActionOnPlus(Plus x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Plus(changeList);
		}
	    };
	def.doAction(x);	
    }


    public void performActionOnTimes(Times x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Times(changeList);
		}
	    };
	def.doAction(x);	
    }


    public void performActionOnMinus(Minus x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Minus(changeList);
		}
	    };
	def.doAction(x);	
    }


    public void performActionOnEquals(Equals x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Equals(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnNotEquals(NotEquals x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new NotEquals(changeList);
		}
	    };
	def.doAction(x);	
    }


    public void performActionOnReturn(Return x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Return(changeList);
		}
	    };
	def.doAction(x);	
    }


    public void performActionOnLessThan(LessThan x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new LessThan(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnParenthesizedExpression(ParenthesizedExpression x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ParenthesizedExpression(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnPassiveExpression(PassiveExpression x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new PassiveExpression(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnTypeCast(TypeCast x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new TypeCast(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnGreaterThan(GreaterThan x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new GreaterThan(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnBinaryAnd(BinaryAnd x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new BinaryAnd(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnBinaryOr(BinaryOr x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new BinaryOr(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnBinaryXOr(BinaryXOr x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new BinaryXOr(changeList);
		}
	    };
	def.doAction(x);	
    }


    public void performActionOnBinaryNot(BinaryNot x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new BinaryNot(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnBinaryAndAssignment(BinaryAndAssignment x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new BinaryAndAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnBinaryOrAssignment(BinaryOrAssignment x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new BinaryOrAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }


    public void performActionOnBinaryXOrAssignment(BinaryXOrAssignment x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new BinaryXOrAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnDivideAssignment(DivideAssignment x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new DivideAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnMinusAssignment(MinusAssignment x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new MinusAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnModuloAssignment(ModuloAssignment x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ModuloAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnPlusAssignment(PlusAssignment x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new PlusAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnPostDecrement(PostDecrement x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new PostDecrement(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnPreDecrement(PreDecrement x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new PreDecrement(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnShiftLeftAssignment(ShiftLeftAssignment x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ShiftLeftAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnShiftRightAssignment(ShiftRightAssignment x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ShiftRightAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnTimesAssignment(TimesAssignment x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new TimesAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnConditional(Conditional x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Conditional(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnUnsignedShiftRightAssignment(UnsignedShiftRightAssignment x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new UnsignedShiftRightAssignment(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnDivide(Divide x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Divide(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnNewArray(NewArray x) {
	DefaultAction def=new DefaultAction(x) {		
		NewArray y = (NewArray)pe;		
		ProgramElement createNewElement(ExtList children) {
		    ArrayInitializer arrInit=(ArrayInitializer)
			children.get(ArrayInitializer.class);
		    children.remove(arrInit);	
		    return new NewArray(children, y.getKeYJavaType(), 
					arrInit, y.getDimensions());     
		}
	    };
	def.doAction(x);	
    }
    
//ppp    
    public void performActionOnNew(New x) {
	DefaultAction def=new DefaultAction(x) {		
		ProgramElement createNewElement(ExtList children) {
                    PositionInfo pos = (PositionInfo) children.
                        removeFirstOccurrence(PositionInfo.class);
		    New y = (New)pe;
		    ReferencePrefix rp=null;
		    int rpPos=getPosition(y,y.getReferencePrefix());
		    if (rpPos!=-1) {
			rp=(ReferencePrefix)children.get(rpPos);
			children.remove(rpPos);
		    }
                    children.add(pos); // bendover necessary since 0.989
		    return new New(children, rp);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnLogicalNot(LogicalNot x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new LogicalNot(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnLogicalAnd(LogicalAnd x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new LogicalAnd(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnLogicalOr(LogicalOr x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new LogicalOr(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnModulo(Modulo x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Modulo(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnNegative(Negative x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Negative(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnPositive(Positive x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Positive(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnShiftLeft(ShiftLeft x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ShiftLeft(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnShiftRight(ShiftRight x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ShiftRight(changeList);
		}
	    };
	def.doAction(x);	
    }
    
//ppp
    public void performActionOnArrayReference(ArrayReference x) {
	DefaultAction def=new DefaultAction(x) {		
		ProgramElement createNewElement(ExtList children) {
                    PositionInfo pos = (PositionInfo) children.
                        removeFirstOccurrence(PositionInfo.class);
		    ArrayReference y=(ArrayReference)pe;
		    ReferencePrefix prefix=null;		    
		    int prefixPos=getPosition(y,y.getReferencePrefix());
		    if (prefixPos!=-1) {
			prefix=(ReferencePrefix)children.get(prefixPos);
			children.remove(prefixPos);
		    }
                    children.add(pos); // bendover necessary since 0.989
		    return new ArrayReference(children, prefix);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnMetaClassReference(MetaClassReference x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new MetaClassReference(changeList);
		}
	    };
	def.doAction(x);	
    }


//ppp
    public void performActionOnSuperConstructorReference(SuperConstructorReference x) {
	DefaultAction def=new DefaultAction(x) {		
		ProgramElement createNewElement(ExtList children) {
                    PositionInfo pos = (PositionInfo) children.
                        removeFirstOccurrence(PositionInfo.class);
		    SuperConstructorReference y = (SuperConstructorReference)pe;
		    ReferencePrefix prefix = null;
		    int prefixPos = getPosition(y,y.getReferencePrefix());   
		    if (prefixPos!=-1) {
			prefix=(ReferencePrefix)children.get(prefixPos);
			children.remove(prefixPos);
		    }		    
                    children.add(pos); // bendover necessary since 0.989
		    return new SuperConstructorReference(children, prefix);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnThisConstructorReference(ThisConstructorReference x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ThisConstructorReference(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnExecutionContext(ExecutionContext x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ExecutionContext(changeList);
		}
	    };
	def.doAction(x);	
    }



    public void performActionOnSuperReference(final SuperReference x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new SuperReference(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnThisReference(final ThisReference x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ThisReference(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnArrayLengthReference(ArrayLengthReference x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ArrayLengthReference(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnSwitch(Switch x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new Switch(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnClassInitializer(ClassInitializer x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ClassInitializer(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnArrayInitializer(ArrayInitializer x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new ArrayInitializer(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnPackageReference(PackageReference x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new PackageReference(changeList);
		}
	    };
	def.doAction(x);	
    }
    public void performActionOnPackageSpecification(PackageSpecification x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new PackageSpecification(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnLessOrEquals(LessOrEquals x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new LessOrEquals(changeList);
		}
	    };
	def.doAction(x);	
    }



    public void performActionOnGreaterOrEquals(GreaterOrEquals x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new GreaterOrEquals(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnLocalVariableDeclaration
	(LocalVariableDeclaration x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {		  
		    return new LocalVariableDeclaration(changeList);
		}
	    };
	def.doAction(x);	
    }

    public void performActionOnParameterDeclaration(ParameterDeclaration x)     {
	DefaultAction def=new ParameterDeclarationAction(x);
	def.doAction(x);
    }

    private class ParameterDeclarationAction extends DefaultAction {
	ParameterDeclaration x;
	ParameterDeclarationAction(ParameterDeclaration x) {
	    this.x = x;
	}
	ProgramElement createNewElement(ExtList changeList) {		  
	    return new ParameterDeclaration
		(changeList, x.parentIsInterfaceDeclaration ());
	}
    }

//eee
    public void performActionOnForUpdates
	(ForUpdates x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {		  
                    PositionInfo pos = (PositionInfo) changeList.
                        removeFirstOccurrence(PositionInfo.class);
		    return new ForUpdates(changeList, pos);
		}
	    };
	def.doAction(x);	
    }


    public void performActionOnGuard
	(Guard x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {		  
		    return new Guard(changeList);
		}
	    };
	def.doAction(x);	
    }


//eee
    public void performActionOnLoopInit
	(LoopInit x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {		  
                    PositionInfo pos = (PositionInfo) changeList.
                        removeFirstOccurrence(PositionInfo.class);
		    return new LoopInit(changeList, pos);
		}
	    };
	def.doAction(x);	
    }


    public void performActionOnProgramSVProxy
	(ProgramSVProxy x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {		  
		    return new ProgramSVProxy(changeList);
		}
	    };
	def.doAction(x);	
    }


    /**
     * returns the position of pe2 in the virtual child array of pe1
     */
    protected static int getPosition(NonTerminalProgramElement pe1, 
				     ProgramElement pe2) {
	int n=pe1.getChildCount();
	int i=0;
	while ((i<n) && (pe1.getChildAt(i)!=pe2)) i++;
	return (i==n) ? -1 : i; 
    }



    protected void changed() {
	ExtList list = stack.peek();
	if (list.isEmpty() || list.getFirst()!=CHANGED) {
	    list.addFirst(CHANGED);
	}
    }

    protected void addToTopOfStack(SourceElement x) {
	if (x != null) {
	    ExtList list =  stack.peek();
	    list.add(x);
	}
    }

    protected void addChild(SourceElement x) {
	stack.pop();
	addToTopOfStack(x);
    }
    
    protected void addChildren(ArrayOfProgramElement arr) {	
	stack.pop();
	for (int i=0; i<arr.size(); i++) {
	    addToTopOfStack((SourceElement)arr.getProgramElement(i));
	}
    }
    
    protected abstract class DefaultAction {
	abstract ProgramElement createNewElement(ExtList changeList);
	protected ProgramElement pe;
	protected DefaultAction() {}
	protected DefaultAction(ProgramElement pe) {
	    this.pe=pe;
	}

	protected void addNewChild(ExtList changeList) {
	    addChild(createNewElement(changeList));
	    changed();	    
	}
	
	public void doAction(ProgramElement x) {
	    ExtList changeList = stack.peek();
	    if (changeList.size()==0) {
		doDefaultAction(x);
		return;
	    }
	    if (changeList.getFirst() == CHANGED) {	    
		changeList.removeFirst();	    
		addNewChild(changeList);
	    } else {
		doDefaultAction(x);
	    }	    
	}
    }
}







