// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.java.visitor;

import java.util.Stack;
import java.util.HashMap;
import java.util.Map;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.op.IProgramVariable;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.ProgramElementName;

/**
 * Walks through a java AST in depth-left-first-order. This visitor
 * replaces a number of program variables by others or new ones.
 */
public class ProgVarReplaceVisitor extends CreatingASTVisitor {

    private ProgramElement result=null;

    // indicates if ALL program variables are to be replace by new
    // variables with the same name
    protected boolean replaceallbynew=true;

    /**
     * stores the program variables to be replaced as keys and the new
     * program variables as values
     */
    protected Map replaceMap;

    /**
     * creates a visitor that replaces the program variables in the given
     * statement by new ones with the same name
     * @param st the statement where the prog vars are replaced
     * @param oldPVs the program variables that are replaced
     */
    public ProgVarReplaceVisitor(Statement st, ProgramVariable[] oldPVs) {
	super(st);
	replaceMap=new HashMap();
	for (int i=0; i<oldPVs.length; i++) {
	    replaceMap.put(oldPVs[i], copy(oldPVs[i]));		
	}
    }

    /**
     * creates a visitor that replaces the program variables in the given
     * statement by new ones with the same name
     * @param st the statement where the prog vars are replaced
     * @param map the HashMap with the replacements
     */
    public ProgVarReplaceVisitor(Statement st, Map map) {
	super(st);
	replaceMap = map;
    }

    /**
     * creates a visitor that replaces the program variables in the given
     * statement by new ones with the same name
     * @param st the statement where the prog vars are replaced
     * @param oldPVs the program variables that are replaced
     */
    public ProgVarReplaceVisitor(Statement st) {
	this(st, new HashMap());
    }

    //%%% HACK: there should be a central facility for introducing new program
    // variables; this method is also used in <code>MethodCall</code> to
    // create copies of parameter variables
    public static ProgramVariable copy(ProgramVariable pv) {
	return copy(pv, "");
    }

    //%%% HACK: there should be a central facility for introducing new program
    // variables; this method is also used in <code>MethodCall</code> to
    // create copies of parameter variables
    public static ProgramVariable copy(ProgramVariable pv, String postFix) {
	return new ProgramVariable
	    (new ProgramElementName
		(pv.getProgramElementName().toString()+postFix), 
	     pv.getKeYJavaType());
    }


    protected void walk(ProgramElement node) {
	if (node instanceof LocalVariableDeclaration && replaceallbynew) {
	    LocalVariableDeclaration vd= (LocalVariableDeclaration)node;
	    ArrayOfVariableSpecification vspecs=vd.getVariableSpecifications();
	    for (int i=0; i<vspecs.size(); i++) {
		ProgramVariable pv 
		    = (ProgramVariable) 
		         vspecs.getVariableSpecification(i).getProgramVariable();
		if (!replaceMap.containsKey(pv)) {
		    replaceMap.put(pv, copy(pv));
		}
	    }
	}
	super.walk(node);
    }

    /** the action that is performed just before leaving the node the
     * last time 
     */
    protected void doAction(ProgramElement node) {
	node.visit(this);
    }
    
    /** starts the walker*/
    public void start() {	
	stack.push(new ExtList());		
	walk(root());
	ExtList el=(ExtList)stack.peek();
	int i=0;
	while (!(el.get(i) instanceof ProgramElement)) {
	    i++;
	}
	result=(ProgramElement) ((ExtList)stack.peek()).get(i);
    }

    public ProgramElement result() { 	
	return result;
    }
    
    public void performActionOnProgramVariable(ProgramVariable pv) {
	ProgramElement newPV = (ProgramElement) replaceMap.get(pv);
	if (newPV!=null) {
	    addChild(newPV);
	    changed();
	} else {
	    doDefaultAction(pv);
	}
    }


}
