// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.java.visitor;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.expression.literal.*;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import junit.framework.*;
import de.uka.ilkd.key.java.abstraction.PrimitiveType;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.rule.*;
import java.io.*;
import de.uka.ilkd.key.logic.*;

public class TestJavaIterator  extends TestCase {

    // some non sense java blocks with lots of statements and expressions
    private static String[] jblocks=new String[]{
	"{ {int d = 0; {d++; d = 5;}} { int e=1; e--;} } ",
	"{ int d=7; int e; d = ++e; while (d == e) { d = (++e) + (++e); } } ",

	"{ int e; if (e == 4) { e++;} else { e--;} } ",
	"{ int e; if (e == 1) e++; else e--; ;}"
    };

    static JavaBlock[] test_block = new JavaBlock[jblocks.length];
    private Recoder2KeY c2k;

    public TestJavaIterator(String name) {
	super(name);
	c2k=new Recoder2KeY(TacletForTests.services(), new NamespaceSet());
    }


    public void setUp() {
	for (int i = 0; i < jblocks.length; i++) {
	    test_block[i] = c2k.readBlockWithEmptyContext(jblocks[i]);
	}

    }


    public void testSkipSubTree2() {
	StatementBlock stmntBlock = (StatementBlock) test_block[3].program();
	JavaIterator it = new JavaIterator(stmntBlock);
	it.next(); // declaration	
	it.skip();
	assertTrue("If expected but "+it.element(), 
	       it.element() instanceof If);
	it.next(); // if condition
	assertTrue("Then expected.",it.skip() instanceof Then);
	assertTrue("PostIncrement expected.", it.next() instanceof PostIncrement);
	assertTrue("Else expected.", it.skip() instanceof Else);
	assertTrue("EmptyStatement expected.", it.skip() instanceof EmptyStatement);
	assertNull("Null expected.", it.skip());
    }

    public void testSkipSubTree() {
	StatementBlock stmntBlock = (StatementBlock) test_block[2].program();
	JavaIterator it = new JavaIterator(stmntBlock);
	assertSame("StatementBlock expected.",it.element(), stmntBlock);
	it.skip();
	assertNull("Skipping root must return null, but: "+it.element(), it.element());
	it.back();
	assertTrue("ProgramVariable expected, but is "+it.element(), 
		   it.element() instanceof ProgramVariable);
	// new game -- new fun
	it = new JavaIterator(stmntBlock);
	it.next(); // declaration	
	it.skip();
	assertTrue("If expected but "+it.element(), 
		   it.element() instanceof If);
	it.next(); // if condition
	assertTrue("Then expected.",it.skip() instanceof Then);
	assertTrue("Parent should be If.",it.parent() instanceof If);
	assertTrue("StatementBlock expected.", it.next() instanceof StatementBlock);
	assertTrue("Parent should be Then.",it.parent() instanceof Then);
	assertTrue("Else expected.", it.skip() instanceof Else);
	assertNull("Null expected.", it.skip());
	assertNull("Parent should be null.",it.parent());
	assertTrue("ProgramVariable expected.", it.back() instanceof ProgramVariable);	
    }

    public void testJavaIteratorBlock2() {
	StatementBlock stmntBlock = (StatementBlock) test_block[1].program();
	JavaIterator it = new JavaIterator(stmntBlock);
	ProgramElement child1_pe = it.element();	
	assertSame("Start element should be the statement block, but is "+child1_pe, 
		   child1_pe, stmntBlock);
	it.next();
	child1_pe = it.back();
	assertSame("One next, one back, is start state but is "+child1_pe, 
		   child1_pe, stmntBlock);
    }

    /** 
     */
    public void testJavaIterator() {
	StatementBlock stmntBlock = (StatementBlock) test_block[0].program();
	JavaIterator it = new JavaIterator(stmntBlock);
	assertSame("First return should be the statement block itself.",
		   it.element(), stmntBlock);
	ProgramElement child1_pe = it.next();
	ProgramElement child2_pe = it.nextSibling();
	assertTrue("The first child should be a statement block", 
		   child1_pe instanceof StatementBlock);
	assertTrue("The second sibling should be a statement block, but is "+child2_pe.getClass(), 
		   child2_pe instanceof StatementBlock);

	StatementBlock child1 = (StatementBlock) child1_pe;
	StatementBlock child2 = (StatementBlock) child2_pe;

	IntIterator intIt = it.position().iterator();
	assertTrue("The position should be [1] but is "+it.position(),
		     intIt.next() == 1);
	assertTrue("The position should be [1] but is "+it.position(),
		   !intIt.hasNext());

	ProgramElement five = it.back();
	assertTrue("After going back there should be the 5 but is "+five,
		   (five instanceof IntLiteral) && 
		   (((IntLiteral)five).getValue().equals("5")));	
	ProgramElement pe = it.back();
	assertTrue("After going back there should be the d reference but is "+pe,
		   pe instanceof ProgramVariable && 
		   ((ProgramVariable)pe).name().equals(new Name("d")));	
	ProgramElement fiveAgain = it.next();
	assertTrue("After next there should be the >>5<< again but is "+fiveAgain,
		   (fiveAgain instanceof IntLiteral) && 
		   (((IntLiteral)fiveAgain).getValue().equals("5")));
	
	while (it.hasNext()) {
	    it.next();
	}
	ProgramElement nullElement = it.next();
	assertNull("Next should be null but is "+nullElement, nullElement);
	nullElement = it.next();
	assertNull("Next next should be null also but is"+nullElement, nullElement);
	ProgramElement prevElement = it.back();
	assertTrue("But the previous element should be >>e<< but is " + prevElement,
		   prevElement instanceof ProgramVariable);


    }

    /** same test like in testJavaIterator but with some hasNext() between */
    public void testJavaIteratorHasNext() {
	StatementBlock stmntBlock = (StatementBlock) test_block[0].program();
	JavaIterator it = new JavaIterator(stmntBlock);
	assertSame("First return should be the statement block itself.",
		   it.element(), stmntBlock);
	ProgramElement child1_pe = it.next();
	ProgramElement child2_pe = it.nextSibling();
	assertTrue("The first child should be a statement block", 
		   child1_pe instanceof StatementBlock);
	assertTrue("The second sibling should be a statement block", 
		   child2_pe instanceof StatementBlock);
	StatementBlock child1 = (StatementBlock) child1_pe;
	StatementBlock child2 = (StatementBlock) child2_pe;
	IntIterator intIt = it.position().iterator();
	//	assertTrue(" The position should be [1] but is "+it.position(),
	//	           intIt.next().equals(new Integer(1)));
	intIt.next();
	//	assertTrue("The position should be [1] but is "+it.position(),
	       //	           !intIt.hasNext());
	ProgramElement five = it.back();
	assertTrue("After going back there should be the 5 but is "+five,
		   (five instanceof IntLiteral) && 
		   (((IntLiteral)five).getValue().equals("5")));
	ProgramElement pe = it.back();
	assertTrue("After going back there should be the d reference but is "+pe,
		   pe instanceof ProgramVariable && 
		   ((ProgramVariable)pe).name().equals(new Name("d")));
	ProgramElement fiveAgain = it.next();
	assertTrue("After next there should be the 5 again but is "+fiveAgain,
		   (fiveAgain instanceof IntLiteral) && 
		   (((IntLiteral)fiveAgain).getValue().equals("5")));

	while (it.hasNext()) {
	    it.next();
	}
	ProgramElement nullElement = it.next();
	assertNull("Next should be null but is " + nullElement, nullElement);
	nullElement = it.next();
	assertNull("Next next should be null also but is"+nullElement, nullElement);
	ProgramElement prevElement = it.back();
	assertTrue("But the previous element should be >>e<< but is " + prevElement,
		   prevElement instanceof ProgramVariable);
	
    }

}
