// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;
import de.uka.ilkd.key.util.Debug;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Stack;

/**
 * This class describes the deletion strategy used by the simultaneous update
 * simplifier. The deletion strategy removes updates (or parts of a simultaneouseous
 * one) if e.g. there exists an update like {x:=x} or the location in the update
 * is not refered to during the rest of the term behind the update.
 */
public class DefaultDeletionStrategy extends Visitor 
    implements DeletionStrategy {

    private static final TermFactory tf = TermFactory.DEFAULT;

    /** 
     * Stack used to store information of the visited subterms,
     * e.g. the used programvariables in a sub term and in particular
     * the sub terms itself, which may change while applying this
     * strategy.
     */
    private Stack termStack = new Stack();
    {
	termStack.ensureCapacity(200);
    }

    /**
     * if set unused local variables will be deleted 
     */
    private boolean deleteUnused  = true;
    /**
     * if set trivial updates are tried to be removed
     */
    private boolean deleteSimple  = true;
            
    
    public DefaultDeletionStrategy() {
	termStack.push(new StackElement());		
    }

    public DefaultDeletionStrategy(boolean deleteSimple, 
				   boolean deleteUnused) {

	this.deleteSimple  = deleteSimple;
	this.deleteUnused  = deleteUnused;

	termStack.push(new StackElement());		
    }

    public DeletionStrategy copy() {
	return new DefaultDeletionStrategy(deleteSimple, 
					   deleteUnused);
    }

    public Object clone() {
	return copy();
    }

    /**
     * returns true if the given operator denotes an is an attribute
     * (static or none static)
     */
    private final boolean isAttribute(Operator op) {
	return op instanceof AttributeOp || 
	    (op instanceof ProgramVariable && 
	     ((ProgramVariable)op).isStatic());
    }


    /** 
     * returns true if the subformula contains a CTRL operator, necessary for
     * specification computing 
     */
    protected boolean deletionForbidden() {
	return containsCtrl();
    }

    protected boolean isUnusedUpdate(Term loc) {	
	return !usesVar(loc);
    }

    protected boolean isSimpleUpdate(Term loc, Term ass) {
	return loc.equals(ass);
    }    

    protected boolean 
	isSimpleUpdateDeletionSafe(HashSet accessed, Term location) {
	if (isAttribute(location.op())) {
	    return !accessed.contains(location.op());
	} else if (location.op() == Op.ARRAY) {
	    final Sort locSort = location.sort();
	    final Iterator it = accessed.iterator();
	    while (it.hasNext()) {
		Object key = it.next();
		if (key instanceof Sort &&
		    locSort.extendsTrans((Sort)key)) {
		    return false;
		}
	    }
	}
	return true;
    }

    protected boolean delete() {
	final StackElement se  = peek();
	HashSet accessed = new HashSet();
	boolean deleted  = false;
	int max = se.size() - 1;
	for (int i = 0; i < max ; i+=2) {
	    if (!deletionForbidden()) {		
		final Term location = se.getSubTerm(i);
		final Term value    = se.getSubTerm(i+1);
		if ((deleteSimple && isSimpleUpdate(location, value) &&
		     isSimpleUpdateDeletionSafe(accessed, location)) || 
		    (deleteUnused && isUnusedUpdate(location))) {		    
		    se.removeSubTerm(i+1);
		    se.removeSubTerm(i);
		    deleted = true;
		    // keep track of removed elements
		    i   -= 2;
		    max -= 2;
		} else if (isAttribute(location.op())) {
		    accessed.add(location.op());
		} else if (location.op() == Op.ARRAY) {
		    accessed.add(location.sort());
		} 
	    }
	}
	return deleted;	
    }

    protected void collectUsedVar(Term visited) {
	if (visited.op() instanceof ProgramVariable && 
	    (parent() == null ||
	    !(parent().op() instanceof SimultaneousUpdateOp) ||	     
	     relativePos() % 2 != 0 || 
	     relativePos() == parent().arity()-1))  {	    
	    	    
	    addVar((ProgramVariable)visited.op());	    

	} else if (visited.javaBlock() != JavaBlock.EMPTY_JAVABLOCK) {
	    ProgramVariableCollector pvc = 
		new ProgramVariableCollector(visited.javaBlock().program());
	    pvc.start();
	    addVar(pvc.result());

	}
    }


    // VISITOR RELATED METHODS

    public void subtreeEntered(Term t) {
	// put subterm list on stack
	termStack.push(new StackElement(peek(), t));
	// collect vars in t
	collectUsedVar(t);
    }


    public void visit(Term visited) {
	Term result = visited;	

	if (visited.op() instanceof SimultaneousUpdateOp) {
	    if (delete()) {
		markChanged();
	    }
	} 

	StackElement se = pop();
	if (se.changed()) { //rebuild
	    Term[] subs = new Term[se.size()];

	    ArrayOfQuantifiableVariable[] vars = new 
		ArrayOfQuantifiableVariable[subs.length];

	    if (visited.op() instanceof SimultaneousUpdateOp) {
		for (int i = 0; i < subs.length; i++) {
		    subs[i] = se.getSubTerm(i);
		}		
		result = (subs.length == 1 ? subs[0] : 
			  tf.createUpdateTerm(subs));	    
	    } else {
		// default behaviour, except SimultaneousUpdateOp
		for (int i = 0; i < subs.length; i++) {
		    subs[i] = se.getSubTerm(i);
		    vars[i] = visited.varsBoundHere(i);
		}
	    		
		result = tf.createTerm(visited.op(), subs, vars, 
				       visited.javaBlock(), visited.ruleVar());
	    }
	    markChanged();	    
	}
	pushOnStack(result);
    }    

    // ENTRANCE METHOD
    public Term delete(Term t) {
	t.execPostOrder(this);
	return peek().getSubTerm(0);
    }

    // TERMSTACK OPERATIONS
     
    private StackElement pop() {
	return (StackElement) termStack.pop();
    }

    private StackElement peek() {
	return (StackElement) termStack.peek();
    }

    private void pushOnStack(Term t) {
	peek().addSubTerm(t);
    }

    private void markChanged() {
	peek().changed(true);
    }

    public boolean usesVar(Term loc) {
	// at the moment no attributes are deleted, a partial decision
	// is possible, but later when needed 
	// arrays are also not deleted, but here once they are
	// implemented we can be stronger
	return loc.op() instanceof AccessOp || 
	    isAttribute(loc.op())  ? true : 
	    peek().usesVar(loc.op());
    }

    public boolean containsCtrl() {
	return peek().containsCtrl(); 
    }

    public void addVar(ProgramVariable var) {
	peek().addVar(var);
    }

    public void addVar(HashSet vars) {
	peek().addVar(vars);
    }

    public Term parent() {
	return peek().parent();
    }

    public Term attached() {
	return peek().attached();
    }

    public int relativePos() {
	StackElement se = peek().parentSE();
	return  se != null ? se.size() : -1;
    }

    // inner container class

    /**
     * This class models the elements on the stack. It keeps track of
     * the sub terms and propagates informations from the sub terms to
     * the parent sub terms, for example the used program variables.
     */    
    private static class StackElement {	

	/** 
	 * the parent stack element
	 */
	private StackElement parentSE;
	private Term attached;
	/** 
	 * the subterms of attached (they may have changed,
	 * e.g. deleted) 
	 */
	private ArrayList subterms = new ArrayList(10);
	/**
	 * the used local variable by this term
	 */
	private HashSet usedVar = new HashSet(10);
	private boolean changed = false;
	private boolean containsCtrl = false;
	

	/**
	 * creates an empty stack element, usually as top element of the
	 * stack
	 */
	public StackElement() {
	}

	/**
	 * creates a stackelement for the <code>attached</code> term,
	 * which is a subterm of the term attached to 
	 * <code>parentSE</code>
	 */
	public StackElement(StackElement parentSE, Term attached) {
	    this.parentSE = parentSE;
	    this.attached = attached;	    

	    if (attached.op() == Op.CTRL) {
		propagateContainsCtrl();
	    }
	}

	/**
	 * returns the parent stack element
	 */
	public StackElement parentSE() {
	    return parentSE;
	}

	/**
	 * returns the term of which the attached term is a subterm or
	 * null if <code>attached</code> is top level
	 */
	public Term parent() {
	    return parentSE != null ? parentSE.attached() : null;
	}

	/**
	 * returns the attached term to this stack element
	 */
	public Term attached() {
	    return attached;
	}
	
	/**
	 * returns the may be altered sub terms of the attached term
	 */
	public Term getSubTerm(int idx) {
	    return (Term)subterms.get(idx);
	}

	/**
	 * returns adds a new sub term to the list of subterms
	 */
	public void addSubTerm(Term term) {
	    subterms.add(term);
	}

	/**
	 * returns remove the sub term at position idx from the list of
	 * subterms
	 */
	public void removeSubTerm(int idx) {
	    subterms.remove(idx);
	}

	/**
	 * returns the number of sub terms
	 */
	public int size() {
	    return subterms.size();
	}

	/**
	 * returns if the sub terms has changed (implying that the
	 * <code>attached</code> has to be created)
	 */
	public boolean changed() {
	    return changed;
	}

	/**
	 * sets the flag indicating that the <code>attached</code> has
	 * to be created
	 */
	public void changed(boolean b) {
	    changed = b;
	}

	/**
	 * returns true if the attached term contains the control
	 * operator 
	 */
	public boolean containsCtrl() {
	    return containsCtrl;
	}

	/**
	 * returns true if the given variable is used in one of sub
	 * terms 
	 */
	public boolean usesVar(Operator var) {
	    return usedVar.contains(var);
	}

	public void addVar(HashSet vars) {
	    Iterator it = vars.iterator();
	    while (it.hasNext()) {
		addVar((ProgramVariable) it.next());
	    }
	}
	
	private void propagateContainsCtrl() {
	    containsCtrl = true;
	    if ( parentSE != null ) {
		parentSE.propagateContainsCtrl();
	    }
	}
	
	private void addVarHelp(ProgramVariable var) {
	    if (attached != null && 
		attached.op() instanceof SimultaneousUpdateOp &&
		attached.arity() - 1 == size()) {
		usedVar.add(var);
	    }	    
	}

	public void addVar(ProgramVariable var) {	    
	    if (parentSE != null) {
		parentSE.addVarHelp(var);
		parentSE.addVar(var);		
	    }
	}

    }
    

}
