// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

/**
 * this class describes a position in an occurrence of a term. A
 * ConstrainedFormula and a PosInTerm determine an object of this class exactly.
 */
package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.IteratorOfNode;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.op.Metavariable;

public class PosInOccurrence implements java.io.Serializable {

    private final ConstrainedFormula cfma;
    private final PosInTerm posInTerm;

    // Descend into metavariables instantiations
    private final Term      metaTerm;
    private final PosInTerm metaPosInTerm;

    private final Sequent seq;
    private final Proof proof;

    /**
     * is true iff there had been a computation of the isAntec
     * bool previously. 
     */    
    private boolean inAntecDetermined;

    /**
     * is true iff the position is in the antecedent of a sequent. Only valid
     * if isAntecDetermined is true
     */
    private boolean inAntec;

    /**
     * The subterm this object points to, or <code>null</code>
     */
    private Term subTermCache=null;

    private final int hashCode;
    
    /**
     * constructs a new PosInOccurrence object with the given
     * ConstrainedFormula and the given PosInTerm determining the position in
     * an occurrence. The Proof is given in order to be able to determine if
     * the ConstrainedFormula occurrs in the antecedent or in the succedent.
     * @param posInTerm a position in a term
     * @param cfma a formula with constraint
     * @param proof a Proof with a sequent containing the cfma.
     */
    public PosInOccurrence(ConstrainedFormula cfma, 
			   PosInTerm posInTerm, 
			   Proof proof) {
        this(cfma, posInTerm, null, null, null, proof, false, false);
    }

    /**
     * constructs a new PosInOccurrence object with the given
     * ConstrainedFormula and the given PosInTerm determining the position in
     * an occurrence. A Sequent is given to facilitate the query
     * determining if the ConstrainedFormula is in the antecedent or
     * the succedent. 
     * @param posInTerm a position in a term
     * @param cfma a formula with constraint
     */
    public PosInOccurrence(ConstrainedFormula cfma, PosInTerm
			   posInTerm, Sequent seq) {
        this(cfma, posInTerm, null, null, seq, null, false, false);
    }  

    public PosInOccurrence(ConstrainedFormula cfma, PosInTerm
			   posInTerm, Sequent seq,
			   boolean inAntec) {
        this(cfma, posInTerm, null, null, seq, null, inAntec, true);
    }

    public PosInOccurrence(ConstrainedFormula cfma, PosInTerm
			   posInTerm, Sequent seq,
			   Term metaTerm, PosInTerm metaPosInTerm) {
        this(cfma, posInTerm, metaTerm, metaPosInTerm, seq, null, false, false);
    }  

    private PosInOccurrence(ConstrainedFormula cfma, PosInTerm 
                            posInTerm, Sequent seq, Proof proof,
                            boolean inAntec, boolean 
                            inAntecDetermined) {
        this(cfma, posInTerm, null, null, seq, proof, inAntec, inAntecDetermined);
    }

    private PosInOccurrence(ConstrainedFormula cfma, PosInTerm 
                            posInTerm,
                            Term metaTerm, PosInTerm metaPosInTerm,
                            Sequent seq, Proof proof,
                            boolean inAntec, boolean 
                            inAntecDetermined) {
	this.inAntecDetermined=inAntecDetermined;
	this.inAntec=inAntec;
	this.cfma=cfma;
	this.proof=proof;
	this.seq=seq;
	this.posInTerm=posInTerm;
	this.metaTerm=metaTerm;
	this.metaPosInTerm=metaPosInTerm;
	this.hashCode = computeHash ();
    }

    /**
     * creates a new PosInOccurrence that has exactly the same state as this
     * object except the PosInTerm is new and walked down the specified
     * subterm, as specified in method down of 
     * {@link de.uka.ilkd.key.logic.PosInTerm}.
     */
    public PosInOccurrence down(int i) {
	if ( metaTerm == null )
	    return new PosInOccurrence(cfma, posInTerm.down(i), seq,
				       proof, inAntec, inAntecDetermined);
	else
	    return new PosInOccurrence(cfma, posInTerm, metaTerm, metaPosInTerm.down(i), seq,
				       proof, inAntec, inAntecDetermined);
    }
    
    
    /**
     * @return Depth of the represented position within a formula; top-level
     *         positions (<code>isTopLevel()</code> have depth zero
     */
    public int depth () {
        int res = posInTerm ().depth ();

        if ( termBelowMetavariable () != null )
            res += posInTermBelowMetavariable ().depth ();

        return res;
    }
    
    
    /**
     * Ascend to the parent node
     */
    public PosInOccurrence up() {
	Debug.assertFalse( isTopLevel(),
			   "not possible to go up from top level position" );
	if ( metaTerm == null || metaPosInTerm == PosInTerm.TOP_LEVEL )
	    return new PosInOccurrence(cfma, posInTerm.up(), seq,
				       proof, inAntec, inAntecDetermined);
	else
	    return new PosInOccurrence(cfma, posInTerm, metaTerm, metaPosInTerm.up(), seq,
				       proof, inAntec, inAntecDetermined);	    
    }

    /**
     * Ascend to the top node of the formula this object points to 
     */
    public PosInOccurrence topLevel () {
	return new PosInOccurrence(cfma, PosInTerm.TOP_LEVEL, seq,
				   proof, inAntec, inAntecDetermined);    	
    }

    public boolean isTopLevel () {
	return
	    posInTerm == PosInTerm.TOP_LEVEL &&
	    ( metaTerm == null || metaPosInTerm == PosInTerm.TOP_LEVEL );
    }
    
    /**
     * Change the sequent in which the formula this object points to
     * is contained.
     * The proof is not changed
     * @param p_seq
     * @return PosInOccurrence object that points to the same formula
     * as this object, but in the sequent <code>p_seq</code>
     */
    public PosInOccurrence replaceSequent(Sequent p_seq) {
        if ( p_seq == seq )
            return this;
        return new PosInOccurrence(cfma, posInTerm, metaTerm, metaPosInTerm, 
                                   p_seq, proof, inAntec, inAntecDetermined);
    }

    /**
     * If "this" points to a metavariable, a term can be inserted at
     * this position, allowing to walk down the term further "inside"
     * the metavariable.
     */
    public PosInOccurrence setTermBelowMetavariable ( Term p_metaTerm ) {
	Debug.assertTrue ( metaTerm == null &&
			   ( constrainedFormula().formula().subAt(posInTerm).op()
			     instanceof Metavariable ) &&
			   p_metaTerm.sort().extendsTrans
			   ( constrainedFormula().formula().subAt(posInTerm).sort() ),
			   "Illegal position (no metavariable) or incompatible sorts" );

	return new PosInOccurrence(cfma, posInTerm, p_metaTerm, PosInTerm.TOP_LEVEL,
				   seq, proof, inAntec, inAntecDetermined);
    }

    /**
     * returns the ConstrainedFormula that determines the occurrence of
     * this PosInOccurrence 
     */
    public ConstrainedFormula constrainedFormula() {
	return cfma;
    }

    /**
     * Replace the formula this object points to with the new formula given
     * 
     * @param p_newFormula
     *                the new formula
     * @param p_newSeq
     *                the new sequent
     * @return a <code>PosInOccurrence</code> object that points to the same
     *         position within the formula <code>p_newFormula</code> as this
     *         object does within the formula <code>constrainedFormula()</code>.
     *         It is not tested whether this position exists within <code>p_newFormula</code>
     */
    public PosInOccurrence replaceConstrainedFormula (ConstrainedFormula p_newFormula,
                                                      Sequent p_newSeq) {
        final PIOPathIterator it = iterator ();
        Term newTerm = p_newFormula.formula ();
        PosInTerm newPosInTerm = PosInTerm.TOP_LEVEL;
        int depth = 0;

        while ( true ) {
            final int subNr = it.next ();


            if ( newTerm.op () instanceof Metavariable && depth == posInTerm ().depth() ) {
                // It is necessary to insert a term below the metavariable,
                // as the iterator still holds further elements

                return new PosInOccurrence ( p_newFormula,
                                             posInTerm (),
                                             termBelowMetavariable (),
                                             posInTermBelowMetavariable (),
                                             p_newSeq,
                                             proof,
                                             inAntec,
                                             inAntecDetermined );
            }

            if ( subNr == -1 ) break;

            newPosInTerm = newPosInTerm.down( subNr );
            newTerm = newTerm.sub ( subNr );
            ++depth;
        }

        return new PosInOccurrence ( p_newFormula,
                                     newPosInTerm,
                                     p_newSeq,
                                     proof,
                                     inAntec,
                                     inAntecDetermined );
    }

    /**
     * returns the subterm this object points to
     */
    public Term subTerm () {
	if ( subTermCache == null ) {
	    if ( metaTerm == null )
		subTermCache = constrainedFormula().formula().subAt(posInTerm);
	    else
		subTermCache = metaTerm.subAt ( metaPosInTerm );
	}
	return subTermCache;
    }

    public Proof proof () {
	return this.proof;
    }


    /**
     * returns true iff the occurrence is in the
     * antecedent of a sequent. 
     */
    public boolean isInAntec() {
	if (!inAntecDetermined) {
	    determineAntecSucc();
	}
	return inAntec;
    }

    /** 
     * compares this PosInOccurrence with another PosInOccurrence
     * and returns true if both describe the same occurrence 
     */
    public boolean eqEquals(Object obj) {
	if (!(obj instanceof PosInOccurrence)) {
	    return false;
	}
	final PosInOccurrence cmp = (PosInOccurrence)obj;

	if ( !constrainedFormula ().equals ( cmp.constrainedFormula () ) ) {
	    return false;
        }
    
	return equalsHelp ( cmp );
    }

    private boolean equalsHelp (final PosInOccurrence cmp) {
        if ( isInAntec () == cmp.isInAntec () ) {
            if ( !posInTerm ().equals ( cmp.posInTerm () ) ) return false;

            if ( termBelowMetavariable () == null )
                return cmp.termBelowMetavariable () == null;
            else
                return termBelowMetavariable ()
                               .equals ( cmp.termBelowMetavariable () )
                       && posInTermBelowMetavariable ()
                               .equals ( cmp.posInTermBelowMetavariable () );
        }
        return false;
    }

    /**
     * Contrary to <code>eqEquals</code>, this method returns true only for pio
     * objects that point to the same (identical) formula
     * @param obj
     * @return
     */
    public boolean equals (Object obj) {
	if (!(obj instanceof PosInOccurrence)) {
	    return false;
	}
	final PosInOccurrence cmp = (PosInOccurrence)obj;

	// NB: the class <code>NonDuplicateAppFeature</code> depends on the usage
	// of <code>!=</code> in this condition
	if ( constrainedFormula() != cmp.constrainedFormula() ) {
	    return false;
	}
    
	return equalsHelp ( cmp );
    }

    public int hashCode () {
    	return hashCode;
    }     

    private int computeHash () {
        int res = constrainedFormula().hashCode() * 13 + posInTerm().hashCode();
	if ( termBelowMetavariable() != null )
    	    res += termBelowMetavariable().hashCode() * 13 +
    	        posInTermBelowMetavariable().hashCode();
    	return res;
    }

    /**
     * determines the 'inAntec' information depending on what has been given
     * when constructing the PosInOccurrence: Sequence or Proof.
     */
    private void determineAntecSucc() {
	if (seq==null) {
	    determineAntecSucc(proof);
	} else {
	    determineAntecSucc(seq);
	}
    }

    /**
     * sets the attributes 'isAntec' and 'isAntecDetermined': The given
     * sequent is searched for the ConstrainedFormula of this
     * PosInOccurrence. If it is found in the antecedent isAntec is set to
     * true; if it is found in the succedent isAntec is set to false. If one
     * of these two cases hold isAntecDetermined is set to true, otherwise
     * to false.
     * @param seq the sequent that determines if the ConstrainedFormula is in
     * the succedent or in the antecedent.
     */
    private void determineAntecSucc(Sequent seq) {
	if (seq.antecedent().contains(cfma)) {
	    inAntec=true;
	    inAntecDetermined=true;
	} else if (seq.succedent().contains(cfma)) {
	    inAntec=false;
	    inAntecDetermined=true;
	} else {
	    inAntecDetermined=false;
	}
    }

    /**
     * sets the attributes 'isAntec' and 'isAntecDetermined': The given
     * proof is searched for the ConstrainedFormula of this
     * PosInOccurrence. If it is found in the antecedent of a sequent in the
     * proof isAntec is set to
     * true; if it is found in the succedent isAntec is set to false. If one
     * of these two cases hold isAntecDetermined is set to true, otherwise
     * to false.
     * @param proof the proof that determines if the ConstrainedFormula is in
     * the succedent or in the antecedent of a sequent.
     */
    private void determineAntecSucc(Proof proof) {
	determineAntecSuccAfterNode(proof.root());
    }

    /**
     * calls determineAntecSucc for each sequence of the nodes after (and
     * inclusive) the given node until all nodes have been visited or the
     * ConstrainedFormula could be found.
     * @param node the start node of the search
     */
    private void determineAntecSuccAfterNode(Node node) {
	determineAntecSucc(node.sequent());
	if (!inAntecDetermined) {
	    IteratorOfNode it=node.childrenIterator();
	    while (it.hasNext() && !inAntecDetermined) {
		determineAntecSuccAfterNode(it.next());
	    }
	}
    }
    
    
    public static PosInOccurrence findInSequent(Sequent seq, 
                                                int formulaNumber, 
                                                PosInTerm pos) {
       return new PosInOccurrence(seq.getFormulabyNr(formulaNumber),  pos, seq);
    }
    
    
    public int formulaNumberInSequent() {
       return seq.formulaNumberInSequent(isInAntec(), cfma);
    }


    /**
     * List all subterms between the root and the position this
     * objects points to; the first term is the whole term
     * <code>constrainedFormula().formula()</code>, the last one
     * <code>subTerm()</code>
     * @return an iterator that walks from the root of the term to the
     * position this <code>PosInOccurrence</code>-object points to
     */
    public PIOPathIterator iterator () {
	return new PIOPathIteratorImpl ();
    }



    /**
     * IF POSSIBLE, YOU SHOULD USE THE METHODS ABOVE AND NOT THE
     * FOLLOWING THREE ONES
     *
     * returns the position in the formula of the ConstrainedFormula of
     * this PosInOccurrence. 
     */
    public PosInTerm posInTerm() {
	return posInTerm;
    }


    public Term termBelowMetavariable () {
	return metaTerm;
    }


    public PosInTerm posInTermBelowMetavariable () {
	return metaPosInTerm;
    }



    private class PIOPathIteratorImpl implements PIOPathIterator {
	IntIterator       currentPathIt;
	Term              currentSubTerm    = null;
	int               child;
	int               count             = 0;
	boolean           belowMetavariable = false;
	
	private PIOPathIteratorImpl               () {
	    currentPathIt = posInTerm ().iterator ();
	}

	public boolean         hasNext            () {
	    return currentPathIt != null;
	}

	/**
	 * @return the number of the next child on the path, or
	 * <code>-1</code> if no further child exists
	 */
	public int         next               () {
	    int res;

	    if ( currentSubTerm == null )
		currentSubTerm = cfma.formula ();
	    else
		currentSubTerm = currentSubTerm.sub ( child );

	    if ( currentPathIt.hasNext () )
		res = currentPathIt.next ();
	    else if ( !belowMetavariable &&
		      termBelowMetavariable () != null ) {
		belowMetavariable = true;
		currentPathIt     = posInTermBelowMetavariable ().iterator ();
		currentSubTerm    = termBelowMetavariable ();

		if ( currentPathIt.hasNext () )
		    res = currentPathIt.next ();
		else {
		    res           =  -1 ;
		    currentPathIt = null;
		}
	    } else {
		res           =  -1 ;
		currentPathIt = null;
	    }

	    child = res;
	    ++count;
	    return res;
	}

	/**
	 * @return the current position within the term
	 * (i.e. corresponding to the latest <code>next()</code>-call)
	 */
	public PosInOccurrence getPosInOccurrence () {
	    // the object is created only now to make the method
	    // <code>next()</code> faster

	    final PosInOccurrence pio;
	    if ( belowMetavariable ) {
	        pio = new PosInOccurrence ( cfma,
                                            posInTerm,
                                            metaTerm,
                                            firstN ( metaPosInTerm,
                                                     count - posInTerm.depth () - 1 ),
                                            seq,
                                            proof,
                                            inAntec,
                                            inAntecDetermined );
	    } else {
	        pio = new PosInOccurrence ( cfma,
                                            firstN ( posInTerm, count - 1 ),
                                            seq,
                                            proof,
                                            inAntec,
                                            inAntecDetermined );            
	    }
        
	    return pio;
	}

	private PosInTerm firstN ( PosInTerm p_pit,
				   int       p_n ) {
	    IntIterator it  = p_pit.iterator ();
	    PosInTerm         res = PosInTerm.TOP_LEVEL;

	    while ( p_n-- != 0 )
		res = res.down ( it.next () );

	    return res;
	}

	/**
	 * @return the current subterm this object points to
	 * (i.e. corresponding to the latest <code>next()</code>-call);
	 * this method satisfies
	 * <code>getPosInOccurrence().subTerm()==getSubTerm()</code>
	 */
	public Term            getSubTerm         () {
	    return currentSubTerm;
	}

	/**
	 * @return the number of the next child on the path, or
	 * <code>-1</code> if no further child exists (this is the number
	 * that was also returned by the last call of <code>next()</code>)
	 */
	public int             getChild           () {
	    return child;
	}
    }

    
    /** toString */
    public String toString() {
	String res = "Term "+
	    posInTerm()+" of "+ constrainedFormula();
	if ( termBelowMetavariable() != null )
	    res +=  " / Term "+
		posInTermBelowMetavariable()+" of "+ termBelowMetavariable();
	return res;
//            formulaNumberInSequent()+" th formula";

    }
	
}
