// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.soundness.StateDependingObject;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.util.Debug;

import java.util.HashSet;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;

/**
 * Performs simplification of updates. It is parametrised with a
 * deletion strategy and a boolean flag that indicates if simultaneous
 * updates shall be created. As default simultaneous updates and the
 * default deletion strategy are used.  
 */
public class SimultaneousUpdateSimplifier {

    /**
     * integers encoding alias information
     */
    public static final int FALSE = -1;
    public static final int UNKNOWN = 0;
    public static final int TRUE = 1;

    /**
     * shortcut to access the default termfactory
     */
    private final TermFactory tf = TermFactory.DEFAULT;

    /** 
     * The deletion strategy performs simplifications that delete parts of
     * existing updates, making use of static analysis.
     */
    private DeletionStrategy deletionStrategy;

    /** 
     * if true then consecutive updates are merged to a single simultaneous
     * update
     */
    private boolean useSimultaneousUpdates = true;


    /** 
     * the contraint under which the formula is simplified 
     */
    private Constraint formulaConstraint = Constraint.BOTTOM;

    /** 
     * creates a new simplifier that makes use of simultaneous updates and the
     * DefaultDeletionStrategy 
     * 
     */
    public SimultaneousUpdateSimplifier() {
	this.deletionStrategy = new DefaultDeletionStrategy();
    }

    /**
     * creates a new simplifier 
     * @param deletionStrategy the DeletionStrategy to use 
     * @param useSimultaneousUpdates a boolean that allows the use of simultaneous
     * updates
     */
    public SimultaneousUpdateSimplifier(DeletionStrategy deletionStrategy,
					boolean useSimultaneousUpdates) {

	this.useSimultaneousUpdates = useSimultaneousUpdates;
	this.deletionStrategy = deletionStrategy;	
    }

    // helpers

    /**
     * creates an if-else term, if necessary or resolves the term
     * if possible
     */
    private final Term createConditionalTerm(Operator op, Term[] subs) {
	if (op == Op.IF_ELSE) {
	    return createIfElseTerm(subs);
	} else if (op instanceof ConjCond) {
	    return createConjCondTerm(subs);
	}
	Debug.fail();
	return null;
    }

    /**
     * creates an if-else term, if necessary or resolves the term
     * if possible
     */
    private final Term createIfElseTerm(Term[] subs) {
	if (subs[0].equals(subs[1])) {	    
	    // case like (o?=o) then : else
	    // return then part
	    return subs[2];
	} else if (subs[2].equals(subs[3])) {
	    // same branches
	    return subs[2];
	} 
	return tf.createIfElseTerm(Op.IF_ELSE, subs);
    }

    /**
     * creates a conj. conditional term, if necessary or resolves the
     * term as far as possible
     */
    private final Term createConjCondTerm(Term[] subs) {
	Term[] result = subs;
	boolean sameBranches = true;
	for (int i = 3; i < subs.length; i+=3) {
	    if (subs[0].equals(subs[i]) && 
		subs[1].equals(subs[i+1])) {
		if (i == 3) {
		    return subs[5];
		} else {
		    result = new Term[i];
		    System.arraycopy(subs, 0, result, 0, i);
		    result[2] = subs[i+2];		    
		    break;
		}     
	    }
	    sameBranches = sameBranches && subs[2].equals(subs[i+2]);
	}
	if (sameBranches) {
	    return subs[2];
	} else {
	    return tf.createConjCondTerm(result);
	}
    }


    /**
     * Assumes that uLoc and tLoc are arrays and creates a conditional
     * term which evaluates to valueIfAlias if uLoc and tLoc are aliased
     * otherwise to uLoc used to express 
     *        {uLoc:=valueIfAlias} tLoc 
     * in cases where it can not be decided if uLoc and tLoc are aliased
     * It takes care if array tLoc is shadowed or not.
     */
    private final Term createArrayConditionalTerm(Term uLoc, 
						  Term tLoc, 
						  Term valueIfAlias) {
	final Term result;
	if (tLoc.op() instanceof ShadowedOperator) {
	    result = createCondTermForShadowedArray(uLoc, tLoc, 
						    uLoc, valueIfAlias);
	} else {
	    result = createConjCondTerm(new Term[]{
					    uLoc.sub(0), uLoc.sub(1), 
					    valueIfAlias, 
					    tLoc.sub(0), tLoc.sub(1),
					    uLoc});					
	}
	return result;	
    }


    /**
     * Assumes that uLoc and tLoc are non-static attributes and creates
     * a conditional creates a conditional term which evaluates to
     * valueIfAlias if uLoc and tLoc are aliased otherwise to uLoc
     * used to express 
     *        {uLoc:=valueIfAlias} tLoc 
     * in cases where it can not be decided if uLoc and tLoc are aliased
     * It takes care if the tLoc is shadowed or not.
     */
    private final Term createAttributeConditionalTerm(Term uLoc, 
						      Term tLoc, 
						      Term valueIfAlias) {
	final Term result;
	if (tLoc.op() instanceof ShadowedOperator) {
	    result = createCondTermForShadowedAttribute(uLoc, tLoc, 
							uLoc, valueIfAlias);
	} else {
	    result = createIfElseTerm(new Term[]{
					  uLoc.sub(0), tLoc.sub(0), 
					  uLoc, valueIfAlias});	
	}
	return result;	
    }


 
    // Basic Simplification Rules

    /**
     * an update is applied on a programm variable      
     * @preconditions var.op().oclIsTypeOf(ProgramVariable)    
     */
    private Term applyOnProgramVariable(Term update, 
					Term var) {
	
	final ProgramVariable pv = (ProgramVariable)var.op();
	for (int i = update.arity()-3; i >= 0; i-=2) {
	    if (update.sub(i).op() == pv) {
		return update.sub(i+1);
	    }
	}
	return var;
    }


    /**
     * a list of updates is applied on a programm variable      
     * @preconditions var.op().oclIsTypeOf(ProgramVariable)
     */
    private Term applyOnProgramVariable(ListOfTerm updates, Term var) {
	ListOfTerm unusedUpdates = updates;	
	Term result = var;	
	while (unusedUpdates != SLListOfTerm.EMPTY_LIST &&
	       result.op() instanceof ProgramVariable) {	    

	    result = applyOnProgramVariable(unusedUpdates.head(), result);
	    unusedUpdates = unusedUpdates.tail();
	}

	return unusedUpdates == SLListOfTerm.EMPTY_LIST ? 
	    result : simplify(unusedUpdates, result);
    }


    private Term createAttributeTerm(AttributeOp op, 
				     Term reference,				     
				     Term transactionNumber) {
	return transactionNumber != null ? 
	    tf.createShadowAttributeTerm((ShadowAttributeOp)op, 
					 reference, transactionNumber) :
	    tf.createAttributeTerm(op, reference);
    }


    private Term createArrayTerm(Term reference,
				 Term index, 
				 Term transactionNumber) {	
	return transactionNumber != null ?
	    tf.createArrayTerm(Op.SHADOW_ARRAY, 
			       reference, index, transactionNumber) : 
	    tf.createArrayTerm(Op.ARRAY, reference, index);
    }


    /**
     * returns the transaction number of a shadowed term or null if term
     * is not shadowed
     */
    private final Term getTransactionNumber(Term shadowed) {
	if (!(shadowed.op() instanceof ShadowedOperator)) {
	    return null;
	}
	return ((ShadowedOperator)shadowed.op()).getTransactionNumber(shadowed);
    }

    /**
     * the simultaneous update of the given term <code>update</code>
     * is applied on the non-static attribute term <code>attr</code>
     * @param update the Term whose top operator is the simultaneous
     * update to be applied on the array term (we are here only
     * interested in its <code>update.arity()-1</code> subterms
     * @param attribute the Term that is evaluated in the state
     * described by the simultaneous update of the term
     * <code>update</code>
     * @return the simplified term 
     *
     * @preconditions attribute.op().oclIsTypeOf(AttributeOp)    
     */
    private Term applyOnNoneStaticAttribute(Term update,
					    Term attr) {
	
	final boolean shadowed = (attr.op() instanceof ShadowedOperator);

	// simplified reference: Be U={l1:=t1,...,ln:=tn}. U o.a --> then (U o)
	// is the simplified reference

	final Term simplifiedReference         = simplify(update, attr.sub(0));	
	final Term simplifiedTransactionNumber = shadowed ? 
	    simplify(update, getTransactionNumber(attr)) : null;
	
	// the composed (Uo).a (if shadowed (Uo)^{Utn}.a)
	final Term simplifiedAttribute = 
	    (simplifiedReference         == attr.sub(0) && 
	     simplifiedTransactionNumber == getTransactionNumber(attr)) ?
	    attr : createAttributeTerm((AttributeOp)attr.op(),
				       simplifiedReference,       
				       simplifiedTransactionNumber);

	Term result = simplifiedAttribute;

	/* 
	   collect all attributes with same attribute variable as the given
	   attribute (lefthandside) and the terms they are updated with
	   (righthandside).
	*/
	for (int i = 0; i < update.arity()-1 ; i+=2) {

	    final Term location  = update.sub(i);
	    final Operator locOp = update.sub(i).op();
	    final Term value     = update.sub(i+1);

	    int aliasOrShadow = FALSE;

	    if (shadowed                  && 
		locOp instanceof AccessOp &&
		!(locOp instanceof ShadowedOperator)) {
		aliasOrShadow = isShadowOf(location, simplifiedAttribute);
	    } else {
		aliasOrShadow = potentialAliased 
		    (location, simplifiedAttribute, false);
	    }

	    switch (aliasOrShadow) {
	    case TRUE:
		result = value;
		break;
	    case UNKNOWN:
		if (!shadowed || !(locOp instanceof ShadowAttributeOp)) {
		    result = createIfElseTerm(new Term[]{
						  simplifiedReference,
						  location.sub(0), 
						  value, result});
		} else {
		    // in this case the update operator 
		    // has to be shadowed as well
		    result = createConjCondTerm
			(new Term[]{
			    simplifiedReference, 
			    simplifiedTransactionNumber,
			    result, 
			    location.sub(0),
			    getTransactionNumber(location), 
			    value});
		}
		break;
	    case FALSE:
		break;
	    default:
		Debug.fail();
	    }
	}
	return result;
    }

    /**
     * a list of updates is applied on an attribute
     * @preconditions var.op().oclIsTypeOf(ProgramVariable)
     */
    private Term applyOnAttribute(ListOfTerm updates,
				  Term attr) {
	Term result = attr;
	ListOfTerm unusedUpdates = updates;	
	while (unusedUpdates != SLListOfTerm.EMPTY_LIST && 
	       result.op() instanceof AttributeOp) {
	    result = applyOnNoneStaticAttribute(unusedUpdates.head(), result);
	    unusedUpdates = unusedUpdates.tail();
	}

	return unusedUpdates == SLListOfTerm.EMPTY_LIST ? 
	    result : simplify(unusedUpdates, result);
    }
    
    /**
     * creates a conjunction conditional out of a list of arrays
     * therefore the conditional part is split up into the conjunctions,
     * e.g. (a[0], a[j], a[4], a[0], 3, 4) becomes to
     * a*0 : a[0]?= (a * j:3 , a * 4 : 4 ) 
     */ 
    private Term createConjCond(List cond, List val) {
	Term result = (Term) val.get(0);
	Term arrayReference = ((Term)cond.get(0)).sub(0);
	Term arrayIndex     = ((Term)cond.get(0)).sub(1);
	for (int i = cond.size()-1; i>=1; i--) {
	    Term candidate = (Term) cond.get(i);
	    result = createConjCondTerm
		(new Term[]{arrayReference, 
			    arrayIndex,
			    result, 
			    candidate.sub(0), 
			    candidate.sub(1), 
			    (Term)val.get(i)});
	}
        return result;
    }


    private Term applyOnArray(ListOfTerm updates, Term array) {
	ListOfTerm updatesToApply = updates;
	Term result = array;

	while (updatesToApply != SLListOfTerm.EMPTY_LIST && 
	       result.op() instanceof ArrayOp) {
	    result         = applyOnArray(updatesToApply.head(), 
					  result);
	    updatesToApply = updatesToApply.tail();
	}

	return updatesToApply == SLListOfTerm.EMPTY_LIST ? result : 
	    simplify(updatesToApply, result);
    }

    /**
     * Applies the simultaneous update of the term <code>update</code>
     * on the given array term <code>array</code>. 
     * @param update the Term whose top operator is the simultaneous
     * update to be applied on the array term (we are here only
     * interested in its <code>update.arity()-1</code> subterms
     * @param array the Term that is evaluated in the state described
     * by the simultaneous update of the term <code>update</code>
     * @return the simplified term 
     */
    private Term applyOnArray(Term update, Term array) {
	
	// U a[i] where U = {l_1 := t_1, ..., l_n := t_n}	
	/*
	  First: we compute the simplified reference Ua and the
	  simplified index Ui
	*/	 
	final boolean shadowed    = (array.op() == Op.SHADOW_ARRAY);
	final Term arrayReference = simplify(update, array.sub(0));
	final Term arrayIndex     = simplify(update, array.sub(1));
	final Term transactionNumber = shadowed ? 
	    simplify(update, 
		     getTransactionNumber(array)) : 
	    null;

	/*
	  In order to save memory resources we test if the simplified
	  array (Ua)[(Ui)] differs from the original array a[i] and
	  only if positive a new array term is created
	*/
	final Term simplifiedArray = 
	    arrayIndex != array.sub(1) || arrayReference != array.sub(0) 
	    || (transactionNumber != null && 
		transactionNumber != getTransactionNumber(array)) ?
	    createArrayTerm(arrayReference, 
			    arrayIndex, 
			    transactionNumber) : array;

	Term result = simplifiedArray;

	/*
	  now we collect potential candidates of assignment pairs
	  which may effect the value of (U a)[(Ui)]. These are all
	  pairs whose left side is an array of compatible sort
	*/
	final List candidates = new ArrayList(update.arity()<<1);	

	// Collect potential location candidates. A potential candidate has the
	// form r_i[J_i], where r is compatible to arr
	for (int i = 0; i < update.arity() - 1; i += 2) {	  	    
	    if (potentialAliased(simplifiedArray, update.sub(i), 
				 true) > FALSE) {
		// add assignment pair (location, value)
		candidates.add(update.sub(i)); 
		candidates.add(update.sub(i+1));		    
	    }
	}

        if (candidates.size()==0) { 
	    // no candidates, so we can return (Ua)[(Ui)]
	    return result;
	}

	/*
	  Several candidates exists. We have now to build up
	  conjunctive conditional terms paying respect to the
	  last-one-wins semantic.
	 */
	final List conditionals = new ArrayList(1+candidates.size());
	final List values       = new ArrayList(1+conditionals.size());

	conditionals.add(result);
	values.add(result);

	for (int i = candidates.size()-2; i>=0; i-=2) {
	    final Term candidateArray = (Term)candidates.get(i);
	    // the local variable alias or shadow is assigned to
	    // 'TRUE'   : if the candidate is aliased to
	    // or a shadowed version of the update it is applied to
	    // 'FALSE'  : if it is definitely NOT aliased or a shadow
	    // 'UNKNOWN': otherwise
	    int aliasOrShadow = FALSE;

	    if (shadowed && 
		!((AccessOp)candidateArray.op()).isShadowed()) {
		aliasOrShadow = 
		    isShadowOf(candidateArray, simplifiedArray);
	    } else {
		aliasOrShadow = potentialAliased
		    (candidateArray, simplifiedArray, false);
	    }
	    
	    if (aliasOrShadow == TRUE) {
		/*
		  The considered candidate is aliased to (Ua)[(Ui)]. We
		  can stop here, as none of the updates considered
		  before is aliased, and hence, the current assignment
		  pair has to be taken.
		 */
		values.set(0, (Term)candidates.get(i+1));
		break; 
	    } else {
		conditionals.add(candidateArray);
		values.add(candidates.get(i+1));
	    }
	}
	// create new array reference		       
	return createConjCond(conditionals, values);
    }
 
    // rules for complex term structure 

    /**
     * application of a list of updates on a conditional term. Therefore the
     * updates are succesively applied on the target term.
     */
    private Term applyOnConditionalTerm(ListOfTerm updates, Term target) {
	ListOfTerm unusedUpdates = updates;
	Term result = target;

	while (unusedUpdates != SLListOfTerm.EMPTY_LIST && 
	       result.op() instanceof IfElse) {	    
	    result = applyOnConditionalTerm(unusedUpdates.head(), result);
	    unusedUpdates = unusedUpdates.tail();
	}

	return unusedUpdates == SLListOfTerm.EMPTY_LIST ? result : 
	    simplify(unusedUpdates, result);
    }

    /**
     * application of an update on a conditional (ifelse) term.
     * If possible the conditional term is simplified, too.
     */
    private Term applyOnConditionalTerm(Term update, Term target) {
		
	boolean changed = false;
	Term[] subs = new Term[target.arity()];

	for (int i = 0; i<subs.length; i++) {
	    subs[i] = simplify(update, target.sub(i));
	    changed = changed || (subs[i] != target.sub(i));
	}
	
	return (changed ? createConditionalTerm(target.op(), subs) : target);
    }


    private Term createOperatorTerm(Term oldTerm, Term[] subs) {

	final ArrayOfQuantifiableVariable[] vbh =
	    new ArrayOfQuantifiableVariable[oldTerm.arity()];
	
	for (int i = 0; i < oldTerm.arity(); i++) {
	    vbh[i] = oldTerm.varsBoundHere(i);
	}
	
	return tf.createTerm(oldTerm.op(), subs, vbh, 
			    JavaBlock.EMPTY_JAVABLOCK, null);
	
    }
    
    /**
     * application of a list of updates on a target term. updates are handed
     * over to the subterms.
     */
    private Term applyOnOperatorTerm(ListOfTerm updates, Term target) {
	Term[] subs = new Term[target.arity()];
	boolean changed = false;
	for (int i = 0; i < target.arity(); i++) {
	    subs[i] = simplify(updates, target.sub(i));
 	    changed = changed || (subs[i] != target.sub(i));
	}
	return (changed ? createOperatorTerm(target, subs) : target);					
    }

    /**
     * concatenation of updates in front of a given term
     */
    private Term concatenate(ListOfTerm updates, Term target) {
	Term result = target;
	IteratorOfTerm it = updates.iterator();
	while ( it.hasNext() ) {		
	    Term upd = it.next();
  	    if (upd.sub(upd.arity()-1) != result) {
		Term[] subs = new Term[upd.arity()];
		for (int i = 0; i < upd.arity()-1; i++) {
		    subs[i] = upd.sub(i);
		}
		subs[subs.length-1] = result;
		result = tf.createUpdateTerm(subs);
  	    } else {
  		result = upd;
  	    }
	}
	return result;
    }

    /**
     * application of a list of updates on a program term. That is the point
     * where updates are merged and the deletion strategy is used.
     * @preconditions target.op().oclIsTypeOf(Modality)     
     */
    private Term applyOnProgramTerm(ListOfTerm updates, 
				      Term target) {


	Term result = simplify(target.sub(0));
	result = (result != target.sub(0)) ?
	    tf.createProgramTerm
	    ((Modality)target.op(), target.javaBlock(), result) : target;

	result = useSimultaneousUpdates ? 
		  merge(updates, result) : concatenate(updates, result);

	return delete(result);
    }
            
    /**
     * @preconditions target.op().oclIsTypeOf(SimultaneousUpdateOp)
     */
    private Term applyOnSimultaneousUpdateTerm(ListOfTerm updates, 
					  Term target) {
	return simplify(updates.prepend(target), target.sub(target.arity()-1));
    }

    
    private Term delete(Term t) {
	return deletionStrategy.copy().delete(t);
    }


    
    private Term merge(ListOfTerm updates, Term target) {

	final IteratorOfTerm updateIterator = updates.iterator();
	Term result = updateIterator.next();

	if (result == null) {
	    return target;
	} else if (updates.size() == 1 && 
		   result.sub(result.arity()-1) == target) {
	    return result;
	}

	while (updateIterator.hasNext()) {
	    result = mergeSimultaneousUpdates(updateIterator.next(), result);
	}
	
	final Term[] subs = new Term[result == null ? 1 : result.arity()];

	for (int i = 0; i < subs.length-1; i++) {
	    subs[i] = result.sub(i);
	    
	}

	subs[subs.length-1] = target;

	return subs.length == 1 ? subs[0] : tf.createUpdateTerm(subs) ;
    }

    /**
     * applies the simultaneous update of the term <code>update</code>
     * on the location described by term <code>loc</code>. 
     * @param update the Term whose top operator is the simultaneous
     * update that is applied on location <code>loc</code>
     * @param loc the Term
     * @return the updated location 
     */
    private Term applyUpdateOnLocation (Term update, Term loc) {
	Term result = loc;	
	if (loc.op() instanceof AccessOp) {
	    
	    Term simplifiedTransactionNumber = null;

	    boolean transactionNumberChanged = false;
	    if (loc.op() instanceof ShadowedOperator) {
		final Term oldTransactionNumber = getTransactionNumber(loc);

		simplifiedTransactionNumber = 
		    simplify(update, oldTransactionNumber);
		transactionNumberChanged    = 
		    simplifiedTransactionNumber != oldTransactionNumber;
	    }

	    final Term simplifiedReference = simplify(update, loc.sub(0));
	    

	    if (loc.op() instanceof AttributeOp) {	    
		if (simplifiedReference != loc.sub(0) || transactionNumberChanged) {
		    result = createAttributeTerm((AttributeOp)loc.op(), 
						 simplifiedReference,
						 simplifiedTransactionNumber);	    
		}
	    } else {
		final Term simplifiedIndex = simplify(update, loc.sub(1));
		if ( simplifiedIndex != loc.sub(1) || 
		     simplifiedReference != loc.sub(0) ||
		     transactionNumberChanged) {
		    result = createArrayTerm(simplifiedReference, 
					     simplifiedIndex,
					     simplifiedTransactionNumber);
		}
	    }
	}
	return result;
    }


    /**
     * merging to simultaneous updates takes place here
     * @preconditions target.op().oclIsTypeOf(SimultaneousUpdateOp)
     */
    private Term mergeSimultaneousUpdates(Term update, 
					    Term target) {

	/*
	  The simultaneouseous update in Term update be refered as 
	       U={l_1:=t_1,...,l_n:=t_n}, 
	  the simultaneouseous update in Term target be refered as 
	       U'={l'_1:=t'_1,...,l'_s:=t'_s}
	  U U' Phi ~~~~> {U*, l''_1:=U t'_1,...,U l''_s:= U t'_s} Phi
	  where U* = U\{l_i: ex j : l_i=l'_j} and l''_i = l'_i, 
          if l'i = program variable or  (U r'_i).a if (l'_i = r'_i.a) or
	  (Ur'_i)[Uj] if l'_i:=r'_i[j]  
	 */
	
	ListOfTerm subs = SLListOfTerm.EMPTY_LIST;

	HashSet pv_locs = new HashSet(target.arity());
	HashMap locsWithAccessors = new HashMap(target.arity());

	boolean changed = false;

	for (int i = 0; i < target.arity()-1; i += 2) {
	    
	    // compute U l'_i
	    Term loc_i = applyUpdateOnLocation(update, target.sub(i));
	    if (loc_i.op() instanceof ProgramVariable) {
		pv_locs.add(loc_i.op());
	    } else {
		/*
		  For the computation of U*, we cache here information
		  of potential aliased locations l_i. 
		  At the moment we consider attributes references to
		  be potential aliased, if they have the same
		  attribute variable, arrays are potential aliased if
		  they have compatible sorts. 
		  For this reason we use the attribute variable as key 
		  for the list of potential aliases and the sort for
		  potential aliased arrays.
		  %% should introduce separate keys for shadowed access
		  operators
		*/
		Object accessorKey = (loc_i.op() instanceof AttributeOp ?
				      (Object) loc_i.op() : 
				      (Object) loc_i.sub(0).sort());
		
		ListOfTerm sameAccessors = 
		    (ListOfTerm)locsWithAccessors.get(accessorKey);
		if (sameAccessors == null) { 
		    sameAccessors = SLListOfTerm.EMPTY_LIST;
		}
		locsWithAccessors.put(accessorKey, sameAccessors.prepend(loc_i));
	    }
	    // compute t'_i
	    Term val_i = simplify(update, target.sub(i+1));
	    changed = changed || (val_i != target.sub(i+1));
	    subs = subs.append(loc_i).append(val_i);
	}

	// compute U*
	ListOfTerm u_star = SLListOfTerm.EMPTY_LIST;

	for (int i = 0; i < update.arity()-1; i+=2) {

	    final Term location = update.sub(i);
	    Term value = update.sub(i+1);
	    if (location.op() instanceof ProgramVariable) {
		value = pv_locs.contains(location.op()) ? null : value;
	    } else {
		ListOfTerm potentialEqualLocations = SLListOfTerm.EMPTY_LIST;
		AccessOp accessOp = (AccessOp) location.op();
		if (accessOp instanceof AttributeOp) {
		    if (locsWithAccessors.containsKey(accessOp)) {
			potentialEqualLocations = 
			    (ListOfTerm)locsWithAccessors.get(accessOp);
		    }
		} else { // attribute operator
		    Sort referenceSort = location.sub(0).sort();
		    if (locsWithAccessors.containsKey(referenceSort)) {
			potentialEqualLocations = 
			    (ListOfTerm)locsWithAccessors.get(referenceSort);
		    }
		}

		IteratorOfTerm it = potentialEqualLocations.iterator();
		
		while (it.hasNext()) {
		    final Term locInTarget = it.next();		    
		    final int aliased = 
                           potentialAliased (location, locInTarget, false);

		    switch (aliased) {
		    case TRUE:
			value = null;
			break;				
		    case UNKNOWN:
			Debug.assertTrue(!(locInTarget.op() instanceof ProgramVariable), 
					 "Aliasing on static attributes can be decided.");
			value = (locInTarget.op() instanceof ArrayOp) ? 			    			    
			    createArrayConditionalTerm(location, locInTarget, value) 
			    : 
			    createAttributeConditionalTerm(location, locInTarget, value);
			break;
		    case FALSE:
			break;
		default:
		    Debug.fail();
		}
	    }
	}
	changed = changed || (value != null);
	    u_star = (value != null ? 
		      u_star.append(location).append(value) :
		      u_star);
	    
	} 
	return changed ? tf.createUpdateTerm
	    (subs.append(target.sub(target.arity()-1)).
	     prepend(u_star)) : target;
    }

    private final Term createCondTermForShadowedArray
	(Term fst, Term snd, Term thenBranch, Term elseBranch) {
	/*
	  {fst,snd}.sub(0) reference to the array object
	  {fst,snd}.sub(1) denotes the array index
	*/
	return createConjCondTerm
	    (new Term[] { 
		fst.sub(0), fst.sub(1), elseBranch,
		snd.sub(0), snd.sub(1), 
		createIfElseTerm(new Term[]
		    {getTransactionNumber(fst), getTransactionNumber(snd),
		     elseBranch, thenBranch})
	    });
    }

    private final Term createCondTermForShadowedAttribute
	(Term fst, Term snd, Term thenBranch, Term elseBranch) {
	/*
	  {fst,snd}.sub(0) reference to the object 
	                   whose attribute is accessed
	  {fst,snd}.sub(1) denotes the transaction number 
	*/
	return createIfElseTerm
	    (new Term[]{
		fst.sub(0), snd.sub(0),
		createIfElseTerm(new Term[]{
		    fst.sub(1), snd.sub(1),
		    thenBranch, elseBranch}), 
		elseBranch});
    }

    private final Term applyOnMetavariable(ListOfTerm updates, Term target) {
	return target;
    }

    /**
     * application on a state depending object 
     */
    private final Term applyOnStateDependingObject(ListOfTerm updates, 
						   Term target) {
	return useSimultaneousUpdates ? 
	    merge(updates, target) : concatenate(updates, target);    	
    }

    private final Term applyOnRigidTerm (ListOfTerm updates, Term target) {
	return delete ( target );
    }


    /**
     * the updates are applied on the target term. Therefore the
     * correct application method with respect to the target terms top
     * level operator is invoked. 
     * @param update ListOfTerm containing terms whose top level
     * simultaneous update is applied to the target term
     * @param target the Term that is evaluated in the state described
     * by the given updates
     */
    private Term simplify(ListOfTerm updates, Term target) {
	final Operator choiceOp = target.op();
	if (choiceOp == Op.CTRL) {
	    return applyOnStateDependingObject(updates, target);	
	} else if (target.isRigid()) {
	    return applyOnRigidTerm(updates, target);
	} else if (choiceOp  instanceof Equality ||	    
		   choiceOp  instanceof Junctor  ||
		   choiceOp  instanceof Quantifier) {
	    return applyOnOperatorTerm(updates, target);
	} else if (choiceOp instanceof StateDependingObject) {
	    return applyOnStateDependingObject(updates, target);
	} else if (choiceOp instanceof ProgramVariable) {
	    return applyOnProgramVariable(updates, target);
	} else if (choiceOp instanceof Modality) {
	    return applyOnProgramTerm(updates, target);
	} else if (choiceOp instanceof AttributeOp) {
	    return applyOnAttribute(updates, target);
	} else if (choiceOp instanceof ArrayOp) {
	    return applyOnArray(updates, target);
	} else if (choiceOp instanceof IfElse) {
	    return applyOnConditionalTerm(updates, target);
	} else if (choiceOp instanceof SimultaneousUpdateOp) {
	    return applyOnSimultaneousUpdateTerm(updates, target);
	} else if (choiceOp instanceof Metavariable) {
	    // should not be reachable as metavariables should be rigid
	    return applyOnMetavariable(updates, target);
	} else if (choiceOp instanceof TermSymbol || 
		   choiceOp instanceof Op) {
	    return applyOnOperatorTerm(updates, target);
	} 

        /* 
	 * Only print an error message if 'updates' is not empty.
         * e.g. the ASM-KeY version introduces a lot of new kinds of
         * terms, but doesn't use simultaneous updates at all, -hs. 
	 */
        if (updates.size() > 0) {
            Debug.assertTrue(false, "simultaneousupdatesimplifier: unrecognized term");
            Debug.out("simultaneousupdatesimplifier: unrecognized term:", target.op());
        }
        return target;
    }


    /**
     * Convenience method if only one update is applied.
     */
    public Term simplify(Term update, Term target) {
	return simplify(SLListOfTerm.EMPTY_LIST.prepend(update), target);
    }

    /**
     * starts simplification
     */
    public Term simplify(Term target) {
	formulaConstraint = Constraint.BOTTOM;
	if (target.op() instanceof SimultaneousUpdateOp) {
	    return simplify(target, target.sub(target.arity()-1));
	}

	return simplify(SLListOfTerm.EMPTY_LIST, target);
    }

    /**
     * starts simplification for a contrained formula
     */
    public ConstrainedFormula simplify(ConstrainedFormula target, 
				       Constraint userConstraint) {
	
	Term result;
	Term targetTerm = target.formula();
	formulaConstraint = target.constraint();
	if (targetTerm.op() instanceof SimultaneousUpdateOp) {
	    result = simplify(targetTerm, targetTerm.sub(targetTerm.arity()-1));
	    
	} else {
	    result = simplify(SLListOfTerm.EMPTY_LIST, targetTerm);
	}

	return target.constraint()==formulaConstraint && 
	    result == targetTerm ? target : 
	    new ConstrainedFormula(result, formulaConstraint);
    }


    // some helpers
    /** returns true iff the two given sorts are compatible this means if one of
     * them is a subsort of the other one 
     * @param s1 first Sort to be tested for compatibility with the second sort s2
     * @param s2 second Sort to be tested for compatibility with the first sort s1
     * @return true iff s1 is subsort of s2 or vice versa
     */
    private boolean compatible(Sort s1, Sort s2) {
	return s1.extendsTrans(s2) || s2.extendsTrans(s1);
    }

    
    /**
     * determines if the given attribute terms are potential
     * aliases. The answer is three valued: 1 aliased,  0 unknown, -1
     * not aliased 
     * @param fst, the first Term 
     * @param snd, the first Term 
     * @return an int encoding the available aliased information
     */
    private int potentialAliasedAttributes(Term fst, Term snd) {
	if (fst.op() instanceof ProgramVariable && 
	    ((ProgramVariable)fst.op()).isStatic()) {
	    if (fst.op() == snd.op()) {
		return TRUE;
	    }
	} else if (fst.op() == snd.op()) {
	    if (compatible(fst.sub(0).sort(), snd.sub(0).sort())) {
		if (equalsWrtFormulaConstraint(fst.sub(0), snd.sub(0))) {
		    return TRUE;
		} else {
		    return UNKNOWN;
		}
	    } 
	}
	return FALSE;
    }


    /**
     * determines if the given attribute terms are potential
     * aliases. The answer is three valued: 
     * TRUE, UNKNOWN (i.e. unknown), FALSE
     * @param fst, the first Term 
     * @param snd, the first Term 
     * @return an int encoding the available aliased information
     */
    private int potentialAliasedArrays(Term fst, Term snd) {
	if (compatible(fst.sub(0).sort(), snd.sub(0).sort())) {
	    if (equalsWrtFormulaConstraint(fst.sub(0), snd.sub(0)) &&
		equalsWrtFormulaConstraint(fst.sub(1), snd.sub(1))) {
		return TRUE;
	    } else {
		return UNKNOWN;
	    }
	}
	return FALSE;
    }

    /**
     * determines if the given terms are potential aliases. The answer
     * is three valued: 1 aliased,  0 unknown, -1 not aliased
     * @param fst, the first Term 
     * @param snd, the first Term 
     * @param ignore, boolean if true the treat shadowed operators as if
     * they were not shadowed
     * @return an int encoding the available aliased information
     */
    public int potentialAliased
	(Term fst, Term snd, boolean ignore) {
	
	int weaken = TRUE; // no value greater than weaken must be returned  
	
	if (!ignore) {
	    if (fst.op() != snd.op()) {
		return FALSE;
	    } else if (fst.op() instanceof ShadowedOperator) {
		if (!equalsWrtFormulaConstraint(getTransactionNumber(fst), 
						getTransactionNumber(snd))) {
		    weaken = UNKNOWN;
		}
	    }
	}

	int result = FALSE;

	if ((fst.op() instanceof AttributeOp && 
	     snd.op() instanceof AttributeOp) ||
	    // the next condition is for static attributes
	    (fst.op() instanceof ProgramVariable &&
	     snd.op() instanceof ProgramVariable)) {
	    result =  potentialAliasedAttributes (fst, snd);
	} else if (fst.op() instanceof ArrayOp &&
		   snd.op() instanceof ArrayOp) {
	    result = potentialAliasedArrays(fst, snd);
	}
	return result > weaken ? weaken : result;
    }


    /** 
     * Determines if the snd term is a shadow of the first one. 
     * The results can be: FALSE, UNKNOWN, TRUE
     * @param fst, the first Term 
     * @param snd, the first Term 
     * @return an int encoding the the validity degree of the isShadowOf
     * relation
     */
    private int isShadowOf(Term fst, Term snd) {
	if (!(fst.op() instanceof ShadowedOperator) && 
	      snd.op() instanceof ShadowedOperator) {
	    return potentialAliased(fst, snd, true);
	} else {
	    return FALSE;
	}
    }

    /**
     * compares both terms if they are equal with respect to the instantiations
     * of the formula (not user) constraint. This is done by unifying both terms
     * and returning true if the received constrained is subsumed by the formula
     * constraint
     * @param t1 the Term compared to t2
     * @param t2 the Term compared to t1
     * @return true if both terms are equal wrt to the formula constrained
     */
    private boolean equalsWrtFormulaConstraint(Term t1, Term t2) {
	return Constraint.BOTTOM.unify(t1, t2).isAsWeakAs(formulaConstraint);
    }
}
