// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.logic.op.QuantifiableVariable;
import de.uka.ilkd.key.logic.op.ArrayOfQuantifiableVariable;
import de.uka.ilkd.key.logic.sort.Sort;
/**
 * A SubstitutionTerm is an object that contains an Operator two subterms. 
 * and one variable
 * Instances should never be accessed via 
 * this interface, use the interface of the superclass Term and construct 
 * instances only via
 * a TermFactory instead.
 */

class SubstitutionTerm extends Term {

    /** variable to substitute
     * @link aggregation
     * @supplierCardinality * 
     */
    private final ArrayOfQuantifiableVariable substVar;
    
    /** 
     * The two sub terms of the substitution term, where subs[0] is the term
     * with which the variable is substituted and subs[1] is the term on
     * which the substitution is applied. <code> {x/subs[0]}subs[1]
     */
    private final ArrayOfTerm subs;

    /** depth of the term */
    private final int depth;

    //For OCL simplification
    /** arity of the term */
    private int arity;
    

    /** creates a substitution term
     * @param op the substitution operator
     * @param substVar the variable to be substituted
     * @param subs subs[0] replaces variable substVar 
     *        in subs[1]
     */
    public SubstitutionTerm(Operator op, 
			    QuantifiableVariable substVar, 
			    Term[] subs) {
	super(op, op.sort(subs));
	
	this.subs = new ArrayOfTerm(subs);
	this.substVar  = new ArrayOfQuantifiableVariable
	    (new QuantifiableVariable[] {substVar});
	this.depth = 1+(subs[0].depth() > subs[1].depth() ? 
			subs[0].depth() : subs[1].depth());

	arity = 2;

	fillCaches();
    }


    //For OCL simplification

    /** creates a substitution term
     * @param op the substitution operator
     * @param substVar1 the 1st variable to be substituted
     * @param substVar2 the 2nd variable to be substituted
     * @param subs subs[0] replaces variable substVar1 and 
     *        subs[1] replaces variable substVar2 in subs[2]
     */
    public SubstitutionTerm(Operator op, 
			    QuantifiableVariable substVar1,
			    QuantifiableVariable substVar2,
			    Term[] subs) {
	super(op, op.sort(subs));
	
	this.subs = new ArrayOfTerm(subs);
	this.substVar  = new ArrayOfQuantifiableVariable
	    (new QuantifiableVariable[] {substVar1, substVar2});
	this.depth = 1+(subs[0].depth() > subs[1].depth() ? 
			subs[0].depth() : subs[1].depth());
	arity = 3;

	fillCaches();
    }
    //

    /** @return n-th subterm ( e.g. {x s} t (0:s ,  1: t) */    
    public Term sub(int n) {
	return subs.getTerm(n);
    }	

    /**@return depth of the term */
    public int depth() {
	return depth;
    }
   
    /** @return arity of the substitution term as int */
    public int arity() {
	return arity;
    } 

    /** @return the bound variables of the n-th subterm */
    public ArrayOfQuantifiableVariable varsBoundHere(int n) {
	//The normal case
	if (arity == 2) {
	    return (n==1 ? substVar : EMPTY_VAR_LIST);
	} 
	//For OCL simplification
	else {
	    return (n==2 ? substVar : EMPTY_VAR_LIST);
	}
	//
    }

}
