// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.logic.op.QuantifiableVariable;
import de.uka.ilkd.key.logic.op.ArrayOfQuantifiableVariable;
import de.uka.ilkd.key.logic.sort.Sort;
/**
 * A TernaryBindingTerm is an object that contains an Operator, three subterms, 
 * and two variables.
 * Instances should never be accessed via 
 * this interface, use the interface of the superclass Term and construct 
 * instances only via
 * a TermFactory instead.
 */

class TernaryBindingTerm extends Term {

    /** iteration variable and accumulation variable
     * @link aggregation
     * @supplierCardinality * 
     */
    private final ArrayOfQuantifiableVariable substVar;
    
    /** 
     * The three sub terms of this term, where subs[0] is a collection,
     * subs[1] is the initial value of the accumulator, and
     * subs[2] is the expression that is evaluated for all elements
     * in the collection, and the expr. in which the variables are bound.
     */
    private final ArrayOfTerm subs;

    /** depth of the term */
    private final int depth;

    /** creates a term with three subterms
     * @param substVar1 the iteration variable (bound in the 3rd subterm)
     * @param substVar1 the accumulation variable (bound in the 3rd subterm)
     * @param subs the subterms
     */
    public TernaryBindingTerm(Operator op, 
			      QuantifiableVariable substVar1,
			      QuantifiableVariable substVar2,
			      Term[] subs) {
	super(op, op.sort(subs));
	
	this.subs = new ArrayOfTerm(subs);
	this.substVar  = new ArrayOfQuantifiableVariable
	    (new QuantifiableVariable[] {substVar1, substVar2});
	this.depth = 1+(subs[0].depth() > subs[1].depth() ? 
			subs[0].depth() : subs[1].depth());

	fillCaches();
    }

    /** @return n-th subterm */    
    public Term sub(int n) {
	return subs.getTerm(n);
    }	

    /**@return depth of the term */
    public int depth() {
	return depth;
    }
   
    /** @return arity of the substitution term 1 as int */
    public int arity() {
	return 3;
    } 

    /** @return the bound variables of the n-th subterm */
    public ArrayOfQuantifiableVariable varsBoundHere(int n) {
	return (n==2 ? substVar : EMPTY_VAR_LIST);
    }
}
