// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.logic.ldt;

import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.expression.Literal;
import de.uka.ilkd.key.java.expression.literal.BooleanLiteral;
import de.uka.ilkd.key.java.expression.operator.BinaryAnd;
import de.uka.ilkd.key.java.expression.operator.BinaryOr;
import de.uka.ilkd.key.java.expression.operator.LogicalNot;
import de.uka.ilkd.key.java.abstraction.Type;
import de.uka.ilkd.key.java.abstraction.ListOfType;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.util.Debug;
import java.io.File;
import de.uka.ilkd.key.java.expression.Operator;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.proof.RuleSource;


/** This class inherits from LDT and implements all method that are
 * necessary to handle the primitive type boolean.
 */
public class BooleanLDT extends LDT {

    /** the boolean literals as function symbols and terms */
    private Function bool_true;
    private Term term_bool_true;
    private Function bool_false;
    private Term term_bool_false;

    public BooleanLDT() {
	super(new Name("boolean"));
        ruleSource = RuleSource.initRuleFile("booleanRules.key");
    }

    public void init(Namespace func_ns, Sort targetSort, ListOfType type) {
	super.init(func_ns, targetSort, type);

	// the literals
	bool_true  = addFunction((Function) func_ns.lookup(new Name("TRUE")));
	term_bool_true  = TermFactory.DEFAULT.createFunctionTerm(bool_true);
	bool_false = addFunction((Function) func_ns.lookup(new Name("FALSE")));
	term_bool_false  = TermFactory.DEFAULT.createFunctionTerm(bool_false);
	
    }

    /** returns the name of the target sort of the LDT 
     * is needed to find the right Sort when constructing the LDT
     * @return the Name of the target sort
     */
    public Name getSortName() {
	return targetSort() == null ? new Name("boolean") : targetSort().name();
    }

    /** returns true if the LDT is responsible for this kind of literals 
     * @param lit the Literal 
     * @return true if the LDT is responsible for this kind of literals
     */
    public boolean isResponsible(Literal lit) {
	return (lit instanceof BooleanLiteral);
    }

    /** returns true if the LDT offers an operation for the given java
     * operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param subs the logic subterms of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterms 
     */
    public boolean isResponsible
	(de.uka.ilkd.key.java.expression.Operator op, Term[] subs) {
	if (subs.length == 1) {
	    return isResponsible(op, subs[0]);
	} else if (subs.length == 2) {
	    return isResponsible(op, subs[0], subs[1]);	
	}
	return false;	
    }

    /** returns true if the LDT offers an operation for the given
     * binary java operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param left the left subterm of the java operator
     * @param right the right subterm of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterms 
     */
    public boolean isResponsible
	(de.uka.ilkd.key.java.expression.Operator op, Term left, Term right) {
	return false;

    }

    /** returns true if the LDT offers an operation for the given
     * unary java operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param sub the logic subterms of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterm
     */
    public boolean isResponsible
	(de.uka.ilkd.key.java.expression.Operator op, Term sub) {
	return false;
    }

    /** translates a given literal to its logic counterpart 
     * @param lit the Literal to be translated
     * @return the Term that represents the given literal in its logic
     * form
     */ 
    public Term translateLiteral(Literal lit) {
	if (lit instanceof BooleanLiteral) {
	    return (((BooleanLiteral)lit).getValue() ? 
		    term_bool_true : term_bool_false);
	}
	Debug.fail("booleanLDT: Wrong literal", lit);	
	return null;
    }

    /** returns the function symbol for the given operation 
     * @return  the function symbol for the given operation 
     */
    public Function getFunctionFor
	(de.uka.ilkd.key.java.expression.Operator op) {	
	return null;
    }   

    public boolean hasLiteralFunction(Function f) {
	return containsFunction(f) && f.arity()==0;
    }

    public Expression translateTerm(Term t, ExtList children) {
	if (t.op()==bool_true) return BooleanLiteral.TRUE;
	if (t.op()==bool_false) return BooleanLiteral.FALSE;
	throw new RuntimeException("BooleanLDT: Cannot convert term to program: "
				   +t);
    }


} 
