// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.ProgramElementName;

/** A factory class for different Schema Variables 
 */
public class SchemaVariableFactory {

    private SchemaVariableFactory() {}
    
    /**
     * creates a SchemaVariable representing quantified variables 
     * @param name the Name of the SchemaVariable
     * @param sort the Sort of the variable the SchemaVariable will be
     * used to represent
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of quantified variables
     * @return the SchemaVariable
     */
    public static SchemaVariable createVariableSV(Name name,
						  Sort sort,
						  boolean listSV) {
	return new VariableSV(name, sort, listSV);
    }

    /** 
     * creates a SchemaVariable representing a term but not a formula
     * @param name the Name of the SchemaVariable
     * @param sort the Sort of the term the SchemaVariable will be
     * used to represent
     * @param rigidness true iff this SV may only match rigid
     * terms/formulas
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of terms
     * @return the SchemaVariable
     */
    public static SchemaVariable createTermSV(Name    name,
					      Sort    sort,
					      boolean listSV,
					      boolean rigidness) {
	return new TermSV(name, sort, listSV, rigidness);
    }


    /**
      * creates a SchemaVariable representing modal operator variable
      * @param name the Name of the SchemaVariable
      * @param arity the arity of the modal operators represented by this SV
      * param modalitylist the list of actual modalities to match this SV (right now box & diamond)
      * @return the SchemaVariable
      */
    public static SchemaVariable createModalOperatorSV(Name name,
						       int arity,
						       Modality[] modalities) {
	return new ModalOperatorSV(name, arity, modalities);
    }

    public static SchemaVariable createTermSV(Name    name,
					      Sort    sort,
					      boolean listSV) {
	return createTermSV(name, sort, listSV, false);
    }

    /** creates a SchemaVariable representing a formula 
     * @param name the Name of the SchemaVariable
     * @param rigidness true iff this SV may only match rigid
     * terms/formulas
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of formulas
     * @return the SchemaVariable
     */
    public static SchemaVariable createFormulaSV(Name    name,
						 boolean listSV,
						 boolean rigidness) {
	return new FormulaSV(name, listSV, rigidness);
    }

    public static SchemaVariable createFormulaSV(Name    name,
						 boolean listSV) {
	return createFormulaSV(name, listSV, false);
    }

    /** creates a SchemaVariable representing a program construct
     * @param name the ProgramElementName of the SchemaVariable
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of program constructs
     * @return the SchemaVariable
     */
    public static SchemaVariable createProgramSV(ProgramElementName name, 
						 ProgramSVSort s,
						 boolean listSV){
	return new ProgramSV(name, s, listSV);
    }


    /** 
     * creates a SchemaVariable representing a skolem term
     * @param name the Name of the SchemaVariable
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of skolem terms
     * @return the SchemaVariable
     */
    public static SchemaVariable createDependingSV(Name name, 
						   Sort s,
						   boolean listSV){
	return new DependingSV(name, s, listSV);
    }
}
