// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
//
//

/** @author Kristofer Johannisson */

package de.uka.ilkd.key.ocl.gf;

import java.util.HashMap;
import java.lang.String;
import java.util.Iterator;
import java.io.PrintStream;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import java.io.FileNotFoundException;


import de.uka.ilkd.key.casetool.UMLOCLClassifier;
import de.uka.ilkd.key.casetool.UMLOCLStructuralFeature;
import de.uka.ilkd.key.casetool.UMLOCLBehaviouralFeature;
import de.uka.ilkd.key.casetool.UMLOCLFeature;
import de.uka.ilkd.key.casetool.UMLOCLAssociation;
import de.uka.ilkd.key.casetool.UMLOCLMultiplicity;
import de.uka.ilkd.key.casetool.HashMapOfClassifier;
import de.uka.ilkd.key.casetool.HashMapOfFeatures;
import de.uka.ilkd.key.casetool.HashMapOfAssociations;
import de.uka.ilkd.key.casetool.JavaOCLTypeMap;

import tudresden.ocl.check.types.Type;
import tudresden.ocl.check.OclTypeException;
import org.apache.log4j.Logger;


/** export UML/OCL type information to a file, for use in 
  * typechecking and with GF 
  */
public class ModelExporter
{
	/** key: package name "java::lang", val: string of class signatures */
	private HashMap packClass;
	/** key: package name "java::lang", val: string of association signatures */
	private HashMap packAssoc;
	
	private HashMapOfClassifier model;

	private Logger logger;
	
	ModelExporter(HashMapOfClassifier model)
	{
		logger = Logger.getLogger("key.ocl.gf");
		packClass = new HashMap();
		packAssoc = new HashMap();
		this.model = model;
	}
	
	/* @param name package-qualified name, with dots e.g. "java.lang.Object"
	 *
	 * @return package in OCL notation, where dots are replaced with double
	 * colons (last part discarded), e.g. "java::lang"
	 */
	private String name2package(String name)
	{
		// inefficient way of doing it, with lots of concats:
		int last = 0;
		int next = name.indexOf(".");
		boolean addSeparator = false;
		String result = "";
		
		while (next != -1) {
			if (addSeparator) {
				result = result.concat("::");
			}
			result = result.concat(name.substring(last, next));
			last = next + 1;
			next = name.indexOf(".", last);
			addSeparator = true;
		}
		
		return result.equals("") ? "NOPACKAGE" : result;
	}
	
	private String qualifiedName(UMLOCLClassifier c)
	{
		String fullName = c.getFullName();
		String pack = name2package(fullName);
		if (pack.equals("")) {
			return c.getName();
		} else {
			return pack + "::" + c.getName();
		}
	}
	
	/** @param fullname package qualified name, e.g. "java.lang.Object"
	  * @param currentClass the class currently being defined, e.g. "java.lang.Exception"
	  * @return class name, with package if necessary, also fixes array types
	  */
	private String class2str(String fullname, String currentClass)
	{
		String pack = name2package(fullname);
		
		if (pack.equals(name2package(currentClass))) {
			return fixArrays(fullname);
		}
		else {
			String shortName = fullname.substring(fullname.lastIndexOf("."), fullname.length());
			return fixArrays(shortName);
		}
	}
	
	/** complete Java to OCL conversion for arrays. E.g. type[] -> Sequence(type) 
	  */
	private String fixArrays(String clss)
	{
		if (clss.endsWith("[]")) { // arrays -> Sequence
			String javaType = clss.substring(0, clss.length()-2);
			String oclType;
			try {
				oclType =  JavaOCLTypeMap.getBasicTypeFor(javaType).toString();
			} catch (OclTypeException ote) {
				oclType = javaType;
			}
			return "Sequence (" + oclType + ")";
		} else {
			return clss;
		}
	}
	
	private String structFeature2str(UMLOCLStructuralFeature attr)
	{
		return attr.getName() + " : " + fixArrays(attr.getType().toString()) + ";\n";
	}

	private boolean inhFromObject(UMLOCLFeature oper)
	{
		String n = oper.getName();
		Type r = oper.getType();
		String rS;
		if (r != null) {
			rS = r.toString();
		} else {
			rS = "";
		}
		
		return 
		  (n.equals("wait()") && r == null) ||
		  (n.equals("wait(long)") && r == null) ||
		  (n.equals("wait(long,int)") && r == null) ||
		  (n.equals("getClass()") && r == null) ||
		  (n.equals("toString()") && rS.equals("String")) ||
		  (n.equals("equals(java.lang.Object)") && rS.equals("Boolean")) ||
		  (n.equals("hashCode()") && rS.equals("Integer")) ||
		  (n.equals("notifyAll()") && r == null) ||
		  (n.equals("finalize()")  && r == null) ||
		  (n.equals("clone()") && rS.equals("Object")) ||
		  (n.equals("notify()") && r == null);			
	}
	
	/** operation names contain parameters, this method cleans up the name,
	  * i.e. it drops everything after (and including) the first '(' */
	private String dropArguments(String operName)
	{
		int x = operName.indexOf("(");
		if (x != -1) {
			return operName.substring(0,x);
		} else {
			return operName;
		}
	
	}

	private String behavFeature2str(UMLOCLBehaviouralFeature oper)
	{
		String result;
		String name = dropArguments(oper.getName());
		
		Type[] args = oper.getParameters();
		Type returns = oper.getType();
		
		// note that isQuery currently always returns true
		if (oper.isQuery() && returns != null) { 
			result = "{query} ";
		} else {
			result = "";
		}
		
		result += name + "(";
		for (int i=0; i < args.length; i++) {
			if (i != 0) { result += ", ";}
			result += fixArrays(args[i].toString());
		}
		result += ")";
		
		if (returns != null) {
			result += " : " + fixArrays(returns.toString());
		}
	
		result += ";\n";
		return result;
	}
	
	private String mult2str(UMLOCLMultiplicity mult)
	{
		int min = mult.getMin();
		int max = mult.getMax();

		
		if (min == 1  &&  max == 1) {
			return "[1]";
		} else { 
			String minS = "0"; // default in case of no match below
			String maxS = "*"; // default in case of no match below

			if (min == 0) {
				minS = "0";
			} else if (min == 1) {
				minS = "1";
			}
			if (max == 1) {
				maxS = "1";
			} else if (max == UMLOCLMultiplicity.INFINITE) {
				maxS = "*";
			}
			
			return "[" + minS + ".." + maxS + "]";
		}
	}
	
	private String assoc2str(UMLOCLAssociation assoc)
	{	
		// we cannot determine if an UMLOCLAssociation is ordered or not
	
		String sourceClass = fixArrays(assoc.getSource().getName());
		String targetClass = fixArrays(assoc.getTarget().getName());
		String sourceMult = mult2str(assoc.getSourceMultiplicity());
		String targetMult = mult2str(assoc.getSourceMultiplicity());
		String sourceRole = assoc.getSourceRole();		
		String targetRole = assoc.getTargetRole();
		
		String source, target;

		if (sourceRole != null) {
			source = sourceRole + " : " + sourceClass + " " + sourceMult;
		} else {
			source = sourceClass + " " + sourceMult;
		}

		if (targetRole != null) {
			target = targetRole + " : " + targetClass + " " + targetMult;
		} else {
			target = targetClass + " " + targetMult;
		}
		
		return source + " <-> " + target +";\n";
	}
	
	private String supertypes2str(HashMapOfClassifier supertypes)
	{
		String result = "";
		
		boolean addSeparator = false;
		Iterator i = supertypes.keySet().iterator();
		while (i.hasNext()) {
			if (addSeparator) {
				result += ", ";
			}
			UMLOCLClassifier classif = ((UMLOCLClassifier) supertypes.get(i.next()));
			result += qualifiedName(classif);
			addSeparator = true;
		}
		
		return result;
	}
	
	private boolean bracketed(String s)
	{
	   return s.startsWith("<") && (
	       s.endsWith(">") || s.indexOf(">(") != -1
	   );
	}
	
	private void addClassifier(UMLOCLClassifier classif)
	{
		// array types are for some reason included by Together/KeY
		// do not include these
		if (classif.getName().endsWith("[]")) { 
			return;
		}
	
		HashMapOfClassifier supertypes = classif.getSupertypes();
		HashMapOfFeatures features = classif.features();
		HashMapOfAssociations assocs = classif.getAssociations();

		String pack = name2package(classif.getFullName());

		String supers = supertypes2str(supertypes);

		String attrs = "";
		String opers = "";
		Iterator i = features.keySet().iterator();
		while (i.hasNext()) {
			UMLOCLFeature f = (UMLOCLFeature) features.get(i.next());
			if (! bracketed(f.getName())) {
                if (f instanceof UMLOCLStructuralFeature) {
                    attrs += structFeature2str( (UMLOCLStructuralFeature) f);
                }
                else if (f instanceof UMLOCLBehaviouralFeature && (! inhFromObject(f)) ) {
                    opers += behavFeature2str( (UMLOCLBehaviouralFeature) f);
                }
			}
		}
		
		String assocsDef = "";
		i = assocs.keySet().iterator();
		while (i.hasNext()) {
			assocsDef += assoc2str( (UMLOCLAssociation) assocs.get(i.next()));
		}
		if (! assocsDef.equals("")) {
			String assocsDefs = (String) packAssoc.get(pack);
			if (assocsDefs != null) {
				packAssoc.put(pack, assocsDefs + assocsDef);
			} else {
				packAssoc.put(pack, assocsDef);
			}
		}
		
		String classDef = "<class> " + classif.getName()	+ "\n" +
			(supers.equals("") ? "" : "<super> " + supers + " </super>\n") +
			(attrs.equals("")  ? "" : "<attributes> " + attrs + " </attributes>\n") +
			(opers.equals("")  ? "" :"<opers> " + opers + " </opers>\n") +
			"</class>\n";
		
		String classDefs = (String) packClass.get(pack);
		if (classDefs != null) {
			classDefs = classDefs + "\n" + classDef;
		} else {
			classDefs = classDef;
		}
		packClass.put(pack, classDefs);
	}

	/** export model information to a file
	  * @param path path for what file to write to 
	  */
	public void export(String path)
	{
		try {
			PrintStream stream = 
				new PrintStream(new BufferedOutputStream(new FileOutputStream(path)));
			export(stream);
			stream.flush();
			stream.close();
		} catch (FileNotFoundException e) {
			logger.error("Can not open file " + path + " for writing model information.");
		}
		
	}
	
	/** export model information
	  * @param ps export on this PrintStream
	  */
	public void export(PrintStream ps)
	{
		Iterator i = model.keySet().iterator();
		while (i.hasNext()) {
			addClassifier( (UMLOCLClassifier) model.get(i.next()));
		}
		
		i = packClass.keySet().iterator();
		String pkg;

		while (i.hasNext()) {
			pkg = (String) i.next();
			ps.println("<package> " + pkg);
			
			ps.println( (String) packClass.get(pkg));
			ps.println();
			
			String pAssocs = (String) packAssoc.get(pkg);
			if (pAssocs != null) {
				ps.println("<associations>");
				ps.println(pAssocs);
				ps.println("</associations>");
			}
			
			ps.println("</package>");
			ps.println();
		}
	}
	
}
