// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.ocl.ocldltranslator;

/**
 * The transformation of an OCL constraint in logical
 * formulas of JavaCard-DL as used in the KeY-System.
 */

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.ProverFacade;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import  de.uka.ilkd.key.java.abstraction.Type; 

import de.uka.ilkd.key.java.abstraction.KeYJavaType;

import tudresden.ocl.*;
import tudresden.ocl.parser.node.*;
import tudresden.ocl.normalize.*;

import java.util.Set;
import java.util.Map;
import java.util.Iterator;
import java.util.Vector;

import de.uka.ilkd.key.ocl.ocldltranslator.control.*;
import de.uka.ilkd.key.ocl.ocldltranslator.normalize.*;
import de.uka.ilkd.key.ocl.ocldltranslator.transform.*;
import de.uka.ilkd.key.logic.op.Junctor;
import de.uka.ilkd.key.casetool.ReprModelMethod;
import de.uka.ilkd.key.java.statement.MethodBodyStatement;
import de.uka.ilkd.key.java.JavaNonTerminalProgramElement;
public class KeyOclConstraintTranslator extends BasicOclConstraintTranslator 
                                        implements TranslationConstants {
    
    protected OclTree tree;
    private Vector invTranslations;
    private Vector preTranslations;
    private Vector postTranslations;
    private Vector globalConstraintVarTranslations;
    protected ReprModelMethod contextMethod;

    private TermFactory tf = TermFactory.DEFAULT;
    private ProverFacade proverFacade;
    private NewVariableDeliverer newVariableDeliverer;
    private NewFunctionNameDeliverer newFunctionNameDeliverer;
    private ProgramVariable calledObjectVar;
    private KeyOclExpressionTranslator exprTranslator;
    
    /** A set containing all the nonrigid function symbols
	that have been created during the last translation of some
	OCL constraint.
     */
    private Set nonrigidFunctionSymbols;

    /** A map mapping all '@pre'-tagged function symbols
	for attributes of some model type onto the corresponding
	non tagged logic operator.
	that occur in the translation of the constraint.
     */
    private Map attributeAtPreFunctions;
    
    /** A map mapping all '@pre'-tagged function symbols
	for attributes or methodcalls of some model type
	onto the corresponding non tagged logic operator.
	that occur in the translation of the constraint.
    */
    private Map methodAtPreFunctions;

    private boolean useDiamondForm;

    private NameCreator keyNameCreator;
    private boolean logEnabled;

    private KeyAtPreTransformer atPreTransformer;
    private DefaultTranslationControl translationControl;


    protected ProgramVariable[] methodParameters;

    private KeyOclConstraintTranslator(){
	super();
	invTranslations = new Vector();
	preTranslations = new Vector();
	postTranslations = new Vector();
	nonrigidFunctionSymbols = null;
	attributeAtPreFunctions = null;
	methodAtPreFunctions = null;
	globalConstraintVarTranslations = null;
	newVariableDeliverer = null;
	newFunctionNameDeliverer = null;
	exprTranslator = null;
	keyNameCreator = new NameCreator("key");
	useDiamondForm = false;
	logEnabled =  false;
    }

    public KeyOclConstraintTranslator(ProverFacade facade){
	this();
        proverFacade = facade;
	atPreTransformer = new SimpleAtPreTransformer();
	translationControl = null;
    }

    public KeyOclConstraintTranslator(ProverFacade facade,
				      ReprModelMethod method){
	this();
	contextMethod = method;
        proverFacade = facade;
	atPreTransformer = new SimpleAtPreTransformer();
	translationControl = null;
    }

    public ProgramVariable[] getMethodParameters() {
	return methodParameters;
    }


    /**
     * A template method to easier change the used expression translator
     * in subclasses
     *
     * @param predefinedVarNames the variables that are defined
     * @param predefinedVarTypes the types of variables that are defined
     * @param tControl the control object to support modified translations
     * @return the translator object for expressions to be used
     */    
    protected KeyOclExpressionTranslator 
	getKeyOclExpressionTranslator(Vector predefinedVarNames,
				      Vector predefinedVarTypes,
				      DefaultTranslationControl tControl){
	return new KeyOclExpressionTranslator(tree, this,
					      predefinedVarNames,
					      predefinedVarTypes,
					      tControl);
    }
    

    public KeyOclExpressionTranslator exprTransl() {
	return exprTranslator;
    }
        
    // ================= Some method to translate a complete OCL constraint =============

    public TranslationResult getTranslationFor(OclTree oclTree){
	return getTranslationFor(oclTree,true,false);
    }



    public void inAOperationContext(AOperationContext oc) {
	constrainedType=oc.getTypeName().toString().trim();

	predefinedVarNames.add("self");
	predefinedVarTypes.add(constrainedType);

	constrainedOperation=
	    oc.getName().toString().trim();

	String param;
	String callSignature = contextMethod.getCallSignature(false);
	String params = callSignature.substring(callSignature.indexOf("(")+1,
						callSignature.indexOf(")"));
	Vector paramsVector = new Vector();
	// parse parameter list of the form "var1:type1; var2:type2; ..."
	while (params.indexOf(";") != -1) {
	    paramsVector.add(params.substring(0,params.indexOf(";")));
	    params = params.substring(params.indexOf(";")+1,
				      params.length());
	}
	// last paramter...
	if (params.indexOf(":") != -1) {
	    paramsVector.add(params);
	}

	String dummy;
	for (int i=0; i<paramsVector.size(); i++) {
	    dummy = paramsVector.elementAt(i).toString();
	    // add var to predefined types: 1st arg is var-name, 2nd its type
	    predefinedVarNames.add(contextMethod.getContainingClass().getClassName()+dummy.substring
				   (0, dummy.indexOf(":")).trim());
	    predefinedVarTypes.add(dummy.substring(dummy.indexOf(":")+1, 
				  dummy.length()).trim());

	    methodParameterNames.add(dummy.substring
				   (0, dummy.indexOf(":")).trim());
	    methodParameterTypes.add(dummy.substring(dummy.indexOf(":")+1, 
				  dummy.length()).trim());
	    

	}
	ProgramVariable result = null;
	if (!contextMethod.isVoid()) {
	    String returnType=callSignature.
		substring(callSignature.indexOf(")")+2, callSignature.length());
	    predefinedVarNames.add("result");
	    predefinedVarTypes.add(returnType);
	    methodReturnType = returnType;
	}	
// 	for (int i=0; i<predefinedVarNames.size(); i++)
// 	    System.out.println("predefined vars "+predefinedVarNames.elementAt(i)+":"+
// 			       predefinedVarTypes.elementAt(i));
// 	for (int i=0; i<methodParameterNames.size(); i++)
// 	    System.out.println("method params "+methodParameterNames.elementAt(i)+":"+
// 			       methodParameterTypes.elementAt(i));
    }


    /**
     * Returns the translation of the given ocl constraint (in tree form).
     * @param useDiamondForm - if true a DLformulas of the form
     * "F_pre -> < result = self.m(p1,..,pn) > F_post" 
     * will be generated for pre/post constraints.
     * if false - the diamond form wil be generated and afterwards the
     * diamond will be transformed away.
     * @param ignorePost - if true only the F_pre part of the formula
     * "F_pre -> < result = self.m(p1,..,pn) > F_post" 
     *will be returned. The part after the implication will be omitted.
     */

    public TranslationResult getTranslationFor(OclTree oclTree, 
					       boolean useDiamondForm, 
					       boolean ignorePost){
	
	return getTranslationFor(oclTree, useDiamondForm, ignorePost, false);
    }


    private void ensuresCalledObjectVarSet() {
	if (calledObjectVar==null) {
	    KeYJavaType calledObjectVarType = exprTranslator.
		getKeYJavaTypeForNonCollectionType(constrainedType);
	    
	    calledObjectVar = proverFacade.
		getProgramVariable("self",calledObjectVarType);
	}
    }

    public ProgramVariable[] collectMethodParameters(String methodContext) {
	ProgramVariable[] methodParams = 
	    new ProgramVariable[methodParameterNames.size()];
	for(int i=0;i<methodParameterNames.size();i++){
		
	    methodParams[i] = 
		proverFacade.getProgramVariable
		(methodParameterNames.elementAt(i).toString(), 
		 exprTranslator.getKeYJavaTypeForNonCollectionType
		 (methodParameterTypes.elementAt(i).toString()));
	}

	return methodParams;
    }

    
    public JavaBlock getMethodCallJavaBlock(ProgramVariable[] argProgVars) {

	// Now we have to build a diamond term of the form ...
	// ForPre -> < res = methodName(p1,...,pn); > ForPost
	
	// First create the program variables res, p1, ..., pn and the
	// corresponding JavaBlock ... 
	
	boolean voidMethodCall = (methodReturnType == null);
	JavaBlock methodCallJavaBlock = null;
	ensuresCalledObjectVarSet();
	
	if (!voidMethodCall) {
	    // we do have a specified return type for the method ...
		    
	    // Create a program variable for the result of the method ...
	    KeYJavaType methodResultType = exprTranslator.
		getKeYJavaTypeForNonCollectionType(methodReturnType);
	    ProgramVariable resultProgVar = proverFacade.
		getProgramVariable("result", methodResultType);
		    
	    methodCallJavaBlock = proverFacade.getMethodCall(calledObjectVar,
							       constrainedOperation,
							       resultProgVar,
							       argProgVars);
	} else {
	    // we do not have a specified return type for the method ...
	    methodCallJavaBlock = proverFacade.getVoidMethodCall(calledObjectVar,
								   constrainedOperation,
								   argProgVars);
	}
	return methodCallJavaBlock;
    }

    public Term calledObjectNotNull(JavaBlock jb) {
	ensuresCalledObjectVarSet();
	ProgramVariable called=calledObjectVar;
	Term eq = tf.createEqualityTerm(tf.createFunctionTerm(called), 
					proverFacade.getNullTerm());
	return tf.createJunctorTerm(Op.NOT, eq);
    }

    private boolean staticCall(JavaBlock jb) {
	JavaNonTerminalProgramElement o 
	    = (JavaNonTerminalProgramElement)jb.program();
	o = (JavaNonTerminalProgramElement) o.getChildAt(0);
	if (o instanceof MethodBodyStatement) {
	    MethodBodyStatement mbs = (MethodBodyStatement) o;
	    return mbs.getProgramMethod().isStatic();	    
	}
	return false;
    }

    public TranslationResult getTranslationFor
	(OclTree oclTree, boolean useDiamondForm, boolean ignorePost, 
	 boolean oclTreeIsSimpleExpression){
	log("Initialize translation state ...");
	// init state ...
	invTranslations = new Vector();
	preTranslations = new Vector();
	postTranslations = new Vector();
	globalConstraintVarTranslations = null;
	newVariableDeliverer = null;
	newFunctionNameDeliverer = null;
	nonrigidFunctionSymbols = new java.util.HashSet();
	attributeAtPreFunctions = new java.util.HashMap();
	methodAtPreFunctions = new java.util.HashMap();
	this.useDiamondForm = useDiamondForm;

	// Set Name Generator for OCL tree ...
	oclTree.setNameCreator(keyNameCreator);

	// create new ProverFacade if none was given ...
	log("Created NEW facade object ...");
	if (proverFacade == null) {
	    
	    try {
		proverFacade = new ProverFacadeImpl();
	    } catch(RuntimeException ioe){
		ioe.printStackTrace();
	    }
	    
	}

	// Create new variable deliverer ...
	newVariableDeliverer = new NewVariableDeliverer();
        
	// Create new function name deliverer ...
	newFunctionNameDeliverer = new NewFunctionNameDeliverer();

	log("Constraint to be translated : " + oclTree);

	log("Applying Key-Default-Normalizations ...");

	applyKeyDefaultNormalizations(oclTree);

	tree = oclTree;

	// Start the translation ...
	log("Starting translation ...");

	log("oclTree for startTranslationTo is " + oclTree);

	startTranslationOf(oclTree);
	
	log("... finished ! Building result formula ...");

	// Now we have to build the result for the whole constraint ...
	
	Term resultTerm, t;
	KeyTranslationResult transResult = new KeyTranslationResult();

	if (!oclTreeIsSimpleExpression) {
	    // We are translating a real ocl constraint ...

	    // There are only Invariants XOR Pre-/Postconditions. They cannot be mixed !
	    
	    // Handle invariants ...
	    if (invTranslations.size() > 0){
		KeyTranslationResult nextInvariant = (KeyTranslationResult) invTranslations.elementAt(0);
		resultTerm = nextInvariant.getTerm();
		
		transResult.addAxioms(nextInvariant.getAxioms());
		
		for(int i=1; i<invTranslations.size();i++){
		    nextInvariant = (KeyTranslationResult) invTranslations.elementAt(i);
		    t = nextInvariant.getTerm();
		    resultTerm = tf.createJunctorTerm(Op.AND,resultTerm,t); // append conjunctive ... 
		    transResult.addAxioms(nextInvariant.getAxioms());
		    
		}
		
		transResult.setTerm(resultTerm);
		resultTerm = convertTranslationResultToFormula(transResult, Op.IMP);
		
		log("CONVERT-TERM == " + resultTerm);
		
		// Now create All-CLOSURE-closure of resultTerm ...
		resultTerm = createAllClosureOfTerm(resultTerm);
		
		log("INTO (all closure)  == " + resultTerm);
		
		transResult = new KeyTranslationResult(resultTerm);
		
		
		
	    } else {
		// Handle pre/post-conditions ...
		
		KeyTranslationResult preTranslation, postTranslation;
		
		// Handle preconditions ...
		if (preTranslations.size() > 0){
		    transResult = new KeyTranslationResult();
		    KeyTranslationResult nextPreCondition = (KeyTranslationResult) preTranslations.elementAt(0);
		    resultTerm = nextPreCondition.getTerm();
		    transResult.addAxioms(nextPreCondition.getAxioms());
		    
		    for(int i=1; i<preTranslations.size();i++){
			nextPreCondition = (KeyTranslationResult) preTranslations.elementAt(i);
			t = nextPreCondition.getTerm();
			resultTerm = tf.createJunctorTermAndSimplify(Op.AND,resultTerm,t); // append conjunctive ... 
			transResult.addAxioms(nextPreCondition.getAxioms());
			
		    }
		    
		    transResult.setTerm(resultTerm);
		    
		    preTranslation = transResult;	       
		    
		} else {
		    // we have no precondition ...
		    // => Use 'true' ...
		    resultTerm = tf.createJunctorTerm(Op.TRUE);
		    preTranslation = new KeyTranslationResult(resultTerm);
		}
		
		
		// Handle postconditions ...
		if (postTranslations.size() > 0){
		    transResult = new KeyTranslationResult();
		    KeyTranslationResult nextPostCondition = (KeyTranslationResult) postTranslations.elementAt(0);
		    resultTerm = nextPostCondition.getTerm();

		    transResult.addAxioms(nextPostCondition.getAxioms());
		    
		    for(int i=1; i<postTranslations.size();i++){
			nextPostCondition = (KeyTranslationResult) postTranslations.elementAt(i);
			t = nextPostCondition.getTerm();
			resultTerm = tf.createJunctorTermAndSimplify(Op.AND,resultTerm,t); // append conjunctive ... 
			transResult.addAxioms(nextPostCondition.getAxioms());
			
		    }
		    
		    transResult.setTerm(resultTerm);
		    
		    postTranslation = transResult;	       
		    
		} else {
		    // we have no postcondition ...
		    // => Use 'true' ...
		    resultTerm = tf.createJunctorTerm(Op.TRUE);
		    postTranslation = new KeyTranslationResult(resultTerm);
		}

		// Now we check, whether we have to add additional axioms for 
		// an 'allInstances@pre' feature, that has been used in the post-Part ...
		// This has to be done, because the translation itself can't add the
		// necessary axiom at the right place (in the pre-Part!).
		AllInstancesCollector collector = new AllInstancesCollector(proverFacade);

		// First, get all allInstances symbols that occur in @pre-form in the post-Part ...
		
		log("postPart term  == " + preTranslation.getTerm());


		Set allInstancesSymbolsWithoutAxiom = collector.collectAllInstancesSymbols(postTranslation.getTerm(), true);

		log("allInstances without axiom in postPart == " + allInstancesSymbolsWithoutAxiom );

		// Second, get all allInstances symbols that occur in the pre-Part ...
		// For these symbols an appropriate axiom already exists !

		log("prePart term  == " + preTranslation.getTerm());

		Set allInstancesSymbolsInPrePart = collector.collectAllInstancesSymbols(preTranslation.getTerm(), false);
		
		log("allInstances in prePart == " + allInstancesSymbolsInPrePart );
		
		
		Function currentAllInstancesSymbol;
		for(Iterator it = allInstancesSymbolsWithoutAxiom.iterator(); it.hasNext();){
		    currentAllInstancesSymbol = (Function) it.next();
		    if (!allInstancesSymbolsInPrePart.contains(currentAllInstancesSymbol)){
			// We don't already have an appropriate axiom, therefore we have to add one ...
			
		       
			Term  allInstancesTerm = tf.createFunctionTerm(currentAllInstancesSymbol);
			Term newAxiom;

			// Since the symbol is non-ridgid, but might not be marked as such a symbol, mark it now ...
			nonrigidFunctionSymbols.add(currentAllInstancesSymbol);
			    
			    
			// Create an additional axiom constraining the interpretation of the new constant ...
			CollectionSort collSort = (CollectionSort) currentAllInstancesSymbol.sort();
			Sort elemSort = proverFacade.getElementSortForCollectionSort(collSort); 
			LogicVariable predicateVar = newVariableDeliverer.getNewVariableForSort(elemSort);
			Term predicateVarTerm = tf.createVariableTerm(predicateVar);
			    
			    
			Function setContainsFunction = 
			    proverFacade.getCollectionFunctionForFeature(CONTAINS,
									   new Sort[]{collSort, elemSort},
									   Sort.FORMULA);
			    
			
			
			Term containsTerm = tf.createFunctionTerm(setContainsFunction, allInstancesTerm, predicateVarTerm);
			newAxiom = exprTranslator.getQuantifiedTermWith(Op.ALL, predicateVar , containsTerm);
			    
			log("Adding defining axiom for "+currentAllInstancesSymbol+"@pre in prePart:" + newAxiom.toString());
			    
			preTranslation.addAxiom(newAxiom);

		    }
		}

		// Now we have to build a diamond term of the form ...
		JavaBlock methodCallJavaBlock = null; 

		// REAMRK: THIS  is not correct for package classes in packages prefix ...
		boolean methodIsConstructor = (constrainedType.equals(constrainedOperation)); 
		
		log("Checking whether specified method is a construtor : IsConstructor == " 
		    + methodIsConstructor);
		
		ProgramVariable resultProgVar=null;
		
		String methodContext = exprTranslator.
		    getKeYJavaTypeForNonCollectionType(constrainedType).getJavaType().getName();
		methodParameters = collectMethodParameters(methodContext);

		if (!methodIsConstructor) {
		    // Non constructor case ...
		    
		    // ForPre -> < res = methodName(p1,...,pn); > ForPost
		    
		    // First create the program variables res, p1, ..., pn and the corresponding JavaBlock ...

		    methodCallJavaBlock = getMethodCallJavaBlock(methodParameters);
		
		} else {
		    // Constructor case ...
		    
		    // ForPre -> < self = new ClassName(p1,...,pn); > ForPost
		    
		    // First create the program variables res, p1, ..., pn and the corresponding JavaBlock ...
		    
		    KeYJavaType calledObjectVarType = 
			exprTranslator.getKeYJavaTypeForNonCollectionType(constrainedType);
		    resultProgVar = proverFacade.getProgramVariable("self",calledObjectVarType);
		    
		    methodCallJavaBlock = proverFacade.getConstructorCall(calledObjectVarType,
									    resultProgVar,
									    methodParameters);
		   
		
		}

		log("preTranslation before @pre-resolving   == " + preTranslation);
		
		// Take also care of the generated axioms for the post translation ...
		// Remark the axioms for the post translation should be handeled here (AFTER
		// the diamond term) because there are possible occurences of ...@pre in some
		// generated axioms !!!!
		
		Term postTerm = convertTranslationResultToFormula(postTranslation, Op.IMP);
		
		Term preTerm = convertTranslationResultToFormula(preTranslation, Op.AND);
		
		if (ignorePost){
		    resultTerm = preTerm;
		} else  
		    if (useDiamondForm){
			
		      		
			// and we have to handle the @pre-tagged function symbols in the postTerm ...
			// this may change the translation term as well as the set of additional axioms ...
			handleAtPreFunctionSymbols(preTranslation,postTranslation);

			// we create the result in the diamond form as :
			// preFormula -> < res = obj.m(p1,...,pn)> postFormula ...
			preTerm = convertTranslationResultToFormula(preTranslation, Op.AND);
			postTerm = convertTranslationResultToFormula(postTranslation, Op.IMP);		     

			//Term diamondTerm = tf.createDiamondTerm(methodCallJavaBlock,postTerm);
			Modality modality = KeyOclCorrectnessMode.getInstance().getModality();
			Term modalityTerm = tf.createProgramTerm(modality,methodCallJavaBlock,new Term[]{postTerm});
			
			if (!staticCall(methodCallJavaBlock)) {
			    preTerm = tf.createJunctorTerm
				(Op.AND, preTerm, 
				 calledObjectNotNull(methodCallJavaBlock));
			}

			resultTerm = tf.createJunctorTerm(Op.IMP, preTerm, 
							  modalityTerm);
			
			

		    } else {
			log("Created result formula without DIAMOND !!!");
			
			// we create a repesentation without diamond (by applying some
			// functionsymbol renamings) ...
			// They can be applied e.g. for horizontal verification !
			resultTerm = createNonDiamondRepresentationFor(preTerm, postTerm);
		    }
		
		log("CONVERT-TERM == " + resultTerm);

		// Now create All-CLOSURE-closure of resultTerm ...
		resultTerm = createAllClosureOfTerm(resultTerm);
		
		log("INTO (all closure)  == " + resultTerm);
		
		transResult = new KeyTranslationResult(resultTerm);
		
	    }

	} else {
	    // We translate an arbitrary OCL expression (not neccessary with result type boolean) ...
	    // We only want the result of the translation of THIS EXPRESSION and not of the CONSTRAINT !

	    log("Generating the translation of a single OCL expression (no constraint !!!)");

	    KeyTranslationResult constraintTranslation;
	     if (invTranslations.size() > 0){
		constraintTranslation = (KeyTranslationResult) invTranslations.elementAt(0);
	     } else if (preTranslations.size() > 0){
		 constraintTranslation= (KeyTranslationResult) preTranslations.elementAt(0);
	     } else {
		 constraintTranslation= (KeyTranslationResult) postTranslations.elementAt(0);
	     }

	     
	     transResult = constraintTranslation;
	       
	} 
	
    
	   return transResult;

    }

    // translate a given OCL-expression of a OCL constraint body ...
    public TranslationResult translateOclExpression(Node cb, AExpression oclExpression,
						       String stereoType,
						       String exprName, 
						       Vector predefinedVarNames,
						       Vector predefinedVarTypes){
	log("translateOclExpression ...");
	log("Translating OCL-Expression: " + exprName + " ( " + stereoType + " )");
	
	DefaultTranslationControl tControl;
	if (translationControl != null){
	    tControl = translationControl; 
	    
	} else {
	    // tControl = new GeneralTranslationControl();
	    tControl = new KeyTranslationControl();
	}
	
	log("TranslationControl is done by : " + tControl.getClass() + " !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" );
	
	exprTranslator = getKeyOclExpressionTranslator(predefinedVarNames,
						       predefinedVarTypes,
						       tControl);

	
	if (this.logEnabled) {
	    exprTranslator.setLoggingEnabled(true);
	}

	// Set initial variable transformations for 'self' and all parameters ...
	String varName,typeName;
	TranslationResult varTranslation;

	if (globalConstraintVarTranslations == null){
	    // Global constraint variable were not translated before ...
	    globalConstraintVarTranslations = new Vector();
	    
	    ProgramVariable lv;
	    Function pf;
	    Term varTerm;
	    Sort varSort;
	    for(int i=0;i<predefinedVarNames.size();i++){
		varName = (String) predefinedVarNames.elementAt(i);
		typeName = (String) predefinedVarTypes.elementAt(i);		

		varSort = exprTranslator.getDLSortForNonCollectionType(typeName);
		if (varSort != Sort.FORMULA) {
		    KeYJavaType kjt = exprTranslator.getKeYJavaTypeForNonCollectionType(typeName);
		    

			// proverFacade.getKeYJavaTypeForClass(typeName);
// 		    if (kjt==null) {
// 			kjt=TypeConverter.DEFAULT.getPrimitiveKeYJavaType(typeName);
// 		    }

		    lv = proverFacade.getProgramVariable
			(varName, proverFacade.getKeYJavaTypeForClass(typeName));//new LogicVariable(new Name(varName),varSort);

		    varTerm = tf.createVariableTerm(lv);

		} else {
		    // We are not allowed to create a logical variable of 
		    // sort FORMULA
		    // since we could possibly quantify over PL-1 formulas, so
		    // that we would have a formula which is NOT PL-1
		    // => Therfore we create a new predicate symbol ...
		    pf = new Function(new Name(varName),varSort, new Sort[0]);
		    varTerm = tf.createFunctionTerm(pf);

		}

 		varTranslation = new KeyTranslationResult(varTerm);
 		globalConstraintVarTranslations.add(varTranslation);
		
	    }
	} 
	
	for(int j=0;j<globalConstraintVarTranslations.size();j++){
	    varName = (String) predefinedVarNames.elementAt(j);
	    typeName = (String) predefinedVarTypes.elementAt(j);
	    varTranslation = (KeyTranslationResult) globalConstraintVarTranslations.elementAt(j);
	    exprTranslator.setTranslationForVariable(varName,typeName,varTranslation);
	}
	

	log("getting translation for the given expression ...");
	KeyTranslationResult result = 
	    (KeyTranslationResult) exprTranslator.getTranslationFor(oclExpression);
	
	log("result of constraintTranslation of ("+exprName+") is : " + result);
	
	
	if (stereoType.toLowerCase().equals("inv")){
	    invTranslations.add(result);
	} else if (stereoType.toLowerCase().equals("pre")){
	    preTranslations.add(result); 
	} if (stereoType.toLowerCase().equals("post")){
	    postTranslations.add(result); 
	}

	// Now keep track of all occuring non-rigid function symbols in the
	// translation of the constraint ...
	
	this.nonrigidFunctionSymbols.addAll(exprTranslator.getOccuringNonRigidFunctionSymbols());
	this.attributeAtPreFunctions.putAll(exprTranslator.getOccuringAttributeAtPreFunctions());
	this.methodAtPreFunctions.putAll(exprTranslator.getOccuringMethodAtPreFunctions());

	return (result);
	
    }

    /**
     * Creates the ALL-Closure of the given term t.
     * @param t - the term for which the all closure should be computed.
     * @return a closed term that means a term without free variables.
     */
     public Term createAllClosureOfTerm(Term t){
	Term result, t2;
	SetOfQuantifiableVariable freeVariables = t.freeVars();
	int freeVarsCnt = freeVariables.size();
	if (freeVarsCnt > 0){
	    LogicVariable[] qVars = new LogicVariable[freeVarsCnt];
	    IteratorOfQuantifiableVariable varIt = freeVariables.iterator();
	    LogicVariable currentVar;
	    int i=0;
	    Term andTerm = null;
	    Term varTerm, isCreatedTerm;
	    Function isCreatedFunction;
	    while (varIt.hasNext()){
		currentVar = (LogicVariable) varIt.next();
		qVars[i] = currentVar;
		i++;

	    }
	    result=t;
	    for(int j=0;j<qVars.length;j++){
		result = 
		    exprTranslator.getQuantifiedTermWith(Op.ALL,qVars[(qVars.length - j)-1],result);
	    }

	} else {
	    // There are no free variables therefore leave the term unchanged ...
	    result = t;
	}
	return result;
 
    }

    /** 
     * @return the ProverFacade which should be used
     * for building logical terms and formulas.
     */
    public ProverFacade getProverFacade(){
	return proverFacade;
    }

    /**
     * @return the variable deliverer for the translation of the constraint 
     */
   public NewVariableDeliverer getNewVariableDeliverer(){
	return newVariableDeliverer;
    }

    /**
     * @return the function name deliverer for the translation of the constraint 
     */
    public NewFunctionNameDeliverer getNewFunctionNameDeliverer(){
	return newFunctionNameDeliverer;
    }

    /**
     * Applies several normalization steps on the given OCL tree.
     * The tree will be normalized in a way that is suitable for
     * the KeY system, e.g. the context element is always given explicitly.
     * This method will be called before the actual translation is started. 
     */
    
     protected void applyKeyDefaultNormalizations(OclTree tree){
	CompoundNormalizer cn=new CompoundNormalizer();
	NormalizerPass np1=new NormalizerPass();
	NormalizerPass np2=new NormalizerPass();
	
	// First a normalizer (of our own) that makes set usage of instances of classes explicit ...
	
	np1.add( new InstanceAsSetNormalizer() );

	// Then some other  useful normalizations ...

	np1.add( new CollectShorthandExpansion() );
	np1.add( new IteratorInsertion() );
	cn.add(np1);

	// OLD :
	cn.add(new MultipleIteratorSolving());

	// NEW: Do not use let-abbreviations for resolving multiple iterators.
	// Use syntactical replacement instead.
	// This causes problems with the syntactical replacement when
	// no (!) context element has been specified !
	// Insertion of an appropriate context element in KeYMultipleIteratorSolving has to be done
	// POSSIBLY TO DO IN THE FUTURE !!!!!!!!!!!!!
	// cn.add(new KeyMultipleIteratorSolving());


	np2.add( new DefaultContextInsertion() );
	np2.add( new ConstraintNaming() );
	np2.add( new TypeInformationInsertion() );
	cn.add(np2);

	// cn.add(new VariableClarification());
	
	cn.normalize( tree );
    }

    /**
     * @return a term that has the form
     *  (ax-1 and ax-2 and ... and ax-n) JUNC (trans-term)
     */
    public Term convertTranslationResultToFormula(KeyTranslationResult transResult, Junctor junc){
	Term result;
	Term axTerm,nextTerm;
	Term transTerm = transResult.getTerm();
	Set axioms = transResult.getAxioms();
	if (axioms.size() > 0){
	    
	    Iterator it = axioms.iterator();
	    axTerm = (Term) it.next();
	    while(it.hasNext()){
		nextTerm = (Term) it.next();
		// append axiom conjunctive ...
		axTerm = tf.createJunctorTerm(Op.AND, axTerm, nextTerm);
	    }
	    
	    result = tf.createJunctorTerm(junc,axTerm,transTerm);

	} else {
	    result = transTerm;
	}
	
	return result;
    }



    /**
     * Removes the diamond construct in the translation of a pre/post constraint
     * by renaming non-@pre-tagged function symbols in postTerm with new function
     * symbols, and @pre-tagged function symbols in postTerm with the corresponding
     * function symbols out of the preTerm.
     */
    protected Term createNonDiamondRepresentationFor(Term preTerm, Term postTerm){
	Term resultTerm;
	
	FunctionNameFunctionSymbolMapper fSymbolMapper 
	    = new FunctionNameFunctionSymbolMapper();
	preTerm.execPostOrder(fSymbolMapper);
	// This map contains to each function symbol name in the preTerm the 
	// corresponding function symbol object, which is essential for the KeY-System.
	Map prefSymbolMap = fSymbolMapper.getNameSymbolMapping();
	
	log("Transforming POST term for diamond elimination ...");
	log("NonRigid func symbols == " + nonrigidFunctionSymbols);
	PostFormulaTransformer postTermTransformer = 
	    new PostFormulaTransformer(nonrigidFunctionSymbols,prefSymbolMap);
	
	log("Start transformation ...");
	
	postTerm.execPostOrder(postTermTransformer);
	Term newPostTerm = postTermTransformer.getTerm();

	log("Translation of POST formula == " + newPostTerm);

	resultTerm = tf.createJunctorTerm(Op.IMP,preTerm, newPostTerm);
	
	return resultTerm;
    }


    protected void handleAtPreFunctionSymbols(KeyTranslationResult preTranslation, 
					      KeyTranslationResult postTranslation){
	Term result = null;

	Term preTerm = convertTranslationResultToFormula(preTranslation, Op.AND);
	Term postTerm = convertTranslationResultToFormula(postTranslation, Op.IMP);

	FunctionNameFunctionSymbolMapper fSymbolMapper 
	    = new FunctionNameFunctionSymbolMapper();
	preTerm.execPostOrder(fSymbolMapper);
	postTerm.execPostOrder(fSymbolMapper);

	// This map contains to each function symbol name in the prePostTermWithDiamond the 
	// corresponding function symbol object, which is essential for the KeY-System.
	Map prepostfSymbolMap = fSymbolMapper.getNameSymbolMapping();
	
	log("Transforming @pre's in DIAMOND-IMPL term ...");
	log("NonRigid func symbols == " + nonrigidFunctionSymbols);

	// Reuse the atPreTransformer which is currently set ...
	KeyAtPreTransformer transformer = this.atPreTransformer;
	log("Init @pre-transformer ...");
	transformer.initTransformer(nonrigidFunctionSymbols, prepostfSymbolMap,
				    attributeAtPreFunctions, methodAtPreFunctions, this);
				
	log("Start transformation ...");

	transformer.transformAtPre(preTranslation, postTranslation);
	
	log("Finushed transformation of @pre's with (new) preTranslation : " + preTranslation);
	
    }


    /**
     * Sets the transformer to be used for handling @pre-tagged function symbols
     * If no special transformer is set a very simple default transfomer will be used.
     */
    public void setAtPreTransformer(KeyAtPreTransformer transformer){
	this.atPreTransformer = transformer;
    }

    /**
     * Sets the heuristic to be used for controlling the translation of OCL expressions in a given OCL
     * constraint.
     * If no special transformer is set a very simple default transfomer will be used.
     */
    public void setTranslationControl(DefaultTranslationControl tControl){
	this.translationControl = tControl;
    }


 // ================= Some helper methods ============================================
    
    public void setLoggingEnabled(boolean log){
	logEnabled = log;
    }

    protected void log(String msg){
	// do nothing ...
	if (logEnabled) {
	    super.log(msg);
	}
    }
    

}
