// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
//
//

package de.uka.ilkd.key.ocl.ocldltranslator;

import java.io.*;

import de.uka.ilkd.key.util.*;
import de.uka.ilkd.key.logic.NamespaceSet;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.expression.operator.CopyAssignment;
import de.uka.ilkd.key.java.expression.operator.New;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.expression.literal.BooleanLiteral;
import de.uka.ilkd.key.gui.*;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.abstraction.*;
import java.util.HashMap;

import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.op.ProgramMethod;
import de.uka.ilkd.key.logic.op.ProgramVariable;


public class ProverFacadeImpl implements de.uka.ilkd.key.ProverFacade {

    private static final VarAndType[] emptyVarAndTypeArray = new VarAndType[0];
    private HashMap assocVars=new HashMap();

    public Namespace sorts      = new Namespace();
    private Namespace functions  = new Namespace();
    private Namespace heuristics = new Namespace();
    private Namespace programVariables = new Namespace();
    private TacletIndex rules    = new TacletIndex();
    private Services services;
    private ProofSettings proofSettings;
    

    public ProverFacadeImpl() {
	this(new Services(), new NamespaceSet());
    }


    /**
     * Constructor for tests only
     * @param cus
     */
    protected ProverFacadeImpl (String[] cus) {
	this(createServicesForTest(cus), new NamespaceSet());
    }

    /**
     * constructor used only for test cases
     */
    private static Services createServicesForTest(String cus[]) {
	Services servicesForTest = new Services ();
	JavaInfo ji = servicesForTest.getJavaInfo ();
	Recoder2KeY r2k = new Recoder2KeY(servicesForTest, 
					  new NamespaceSet());
	r2k.readCompilationUnitsAsFiles(cus);
	return servicesForTest;	
    }

    static ProverFacadeImpl createFacadeForTest(ReprModelClass mc) {
	String[] cus = mc.getClassesInPackage();
	return new ProverFacadeImpl(cus);
    }

    public ProverFacadeImpl(Services services, NamespaceSet ns) {
	this.services=services;
	sorts=ns.sorts();
	heuristics=ns.heuristics();
	functions=ns.functions();
	this.programVariables=ns.programVariables();
    }


    public JavaInfo javaInfo() {
	return services ().getJavaInfo ();
    }

    public Services services() {
	return services;
    }

    public KeYJavaType getKeYJavaTypeForClass(String name){
	return javaInfo ().getKeYJavaTypeByClassName(name);
    }

    public Sort getSortForClass(String name) {
	return getKeYJavaTypeForClass(name).getSort();
    }

    public boolean isSortForClass(Sort s) {
	return false;
    }

    
    public Sort getElementSortForCollectionSort(CollectionSort collSort){
	return collSort.elementSort();
    }


    public KeYJavaType getKeYJavaTypeForOCLInteger() {
	return javaInfo ().getKeYJavaType(PrimitiveType.JAVA_INT);
    } 

    public KeYJavaType getKeYJavaTypeForOCLReal() {
	return getKeYJavaTypeOfName("Real");
    }

    public KeYJavaType getKeYJavaTypeForOCLString() {
	return getKeYJavaTypeOfName("String");
    }

    public KeYJavaType getKeYJavaTypeForOCLBoolean() {
	return javaInfo ().getKeYJavaType(PrimitiveType.JAVA_BOOLEAN);//getKeYJavaTypeOfName("boolean");
    }

    public Sort getNumberSort() {
	return getSortOfName("numbers");
    } 

    public Sort getSortForOCLInteger() {
	return getKeYJavaTypeForOCLInteger().getSort();
    } 

    public Sort getSortForOCLReal() {
	return getKeYJavaTypeForOCLReal().getSort();
    }

    public Sort getSortForOCLString() {
	return getKeYJavaTypeForOCLString().getSort();
    }

    public Sort getSortForOCLBoolean() {
	return getKeYJavaTypeForOCLBoolean().getSort();
    }

    public Sort getSortForOCLType() {
	return getSortOfName("OclType");
    }
    
    public Sort getSortForOCLSet(NonCollectionSort elem) {
	return elem.getSetSort();
    }

    public Sort getSortForOCLSequence(NonCollectionSort elem) {
	return elem.getSequenceSort();
    }

    public Sort getSortForOCLBag(NonCollectionSort elem) {
	return elem.getBagSort();
    }

    public Term getNullTerm() {
	return javaInfo ().getNullTerm();
    }

    public Namespace programVariables() {
	return programVariables;
    }

    public void setProgramVariables(Namespace ns) {
	programVariables = ns;
    }
    
    public void doNotAddProgramVariables() {
	setProgramVariables(new Namespace(programVariables()));
    }
    
    public ProgramVariable getProgramVariable(String name, KeYJavaType s) {
	ProgramElementName n = new ProgramElementName(name);
	ProgramVariable pv = (ProgramVariable)(programVariables().lookup(n));
	if (pv == null) {
	    pv = new ProgramVariable(n,s);
	    programVariables().add(pv);
	} else if ( pv.getKeYJavaType() != s ) {
	    System.err.println("Warning: ProgramVariable "+name+
			       " previously declared with KeYJavaType "
			       +pv.getKeYJavaType()+" Type now: "+s);
	    return new ProgramVariable(n,s);
	}
	return pv;
    }

    public ProgramVariable getProgramVariable(String name) {
	ProgramElementName n = new ProgramElementName(name);
	ProgramVariable pv = (ProgramVariable)(programVariables().lookup(n));
	return pv;
    }
    
    public Function getFunctionForFeature(String attr,
					  Sort[] argSorts,
					  Sort retSort) {
 	Name n = null;
 	String className;
  	Function f = null;
 	if (argSorts.length > 0){
 	    className = argSorts[0].name().toString();
 	    n = new Name(attr);
 	    f = (Function)(functions.lookup(n));
 	    if (f == null) {
 		n = new Name(className+"::"+attr);
 		f = (Function)(functions.lookup(n));
 	    }
 	} else {
 	    className = retSort.name().toString();
 	    n = new Name(attr);
 	    f = (Function)(functions.lookup(n));
 	    if (f == null) {
 		n = new Name(className+"::"+attr);
 		f = (Function)(functions.lookup(n));
 	    }
 	}
 	f = (Function)(functions.lookup(n));
 	if ( f == null ) {
 	    f = new Function(n,retSort,argSorts);
 	    functions.add(f);
	    //	    throw new RuntimeException("getFunctionForFeature(): "+attr+" "+retSort);
 	} else {
 	    if (retSort != f.sort()) {
 		throw new RuntimeException("Function "+n+
 					   " previously declared with"+
 					   " return type "+ f.sort());
 	    }
 	    if ( argSorts.length != f.arity() ) {
 		throw new RuntimeException("Function "+n+
 					   " previously declared with"+
 					   + f.arity() + " args.");
 	    }
 	    for ( int i = 0; i<f.arity(); i++ ) {
 		if (argSorts[i] != f.argSort(i)) {
 		    throw new RuntimeException("Function "+n+
 					       " previously declared with"+
 					       " arg nr."+i+
 					       " of type "+f.argSort(i));
 		}
 	    }
 	}
 	return f;
    }

    public ProgramMethod getQueryForFeature(String method, Term[] argterms) {
	if (argterms.length==0 || !(argterms[0].op() instanceof ProgramInLogic)) {
	    throw new IllegalArgumentException("Arguments are empty or "
					       +"cannot be converted to "
					       +"programs "+method+"  "
					       +argterms.length+"   "
					       +argterms[0]+"   "
					       +argterms[0].getClass());
	}
	TypeConverter tc=services ().getTypeConverter();
	KeYJavaType kjt=tc.getKeYJavaType(argterms[0]);
	ListOfType typelist=SLListOfType.EMPTY_LIST;
	for (int i=argterms.length-1; i>=1; i--) {
	    typelist=typelist.prepend(tc.getKeYJavaType(argterms[i]));
	}
	ProgramMethod pm = javaInfo ().getProgramMethod(kjt, method, typelist, 
							kjt);
	if (pm==null) {
	    throw new IllegalArgumentException("No Query "+method
					       +" found on type "+kjt);
	}
	return pm;
    }

    public JavaBlock getMethodCall(ProgramVariable reference,
				   String method,
				   ProgramVariable result,
				   ProgramVariable[] args, 
				   VarAndType[] pvexc) {
	int exc=pvexc.length;
	ExtList stmtLst = new ExtList();
	for (int i=0; i<exc; i++) {
	    KeYJavaType kjt = getKeYJavaTypeForOCLBoolean();
	    VariableSpecification varSpec
		=new VariableSpecification(pvexc[i].getProgramVariable(), 
					   new BooleanLiteral(false), kjt);
	    LocalVariableDeclaration decl = new LocalVariableDeclaration
		(new Modifier[0], 
		 javaInfo().createTypeReference(kjt),
		 new VariableSpecification[]{varSpec});
	    stmtLst.add(decl);
	}
	ExtList lst = makeExtListFromArray(args);
	ProgramMethod meth=javaInfo ().
	    getProgramMethod(reference, method, args, 
			     reference.getKeYJavaType());
	if (meth==null) {
	    String params = "";
	    for (int i=0; i<args.length-1; i++) {
		params += args[i].name()+":"+args[i].sort()+", ";
	    }
	    params += args[args.length-1].name()+
		":"+args[args.length-1].sort();
	    throw new RuntimeException("Method "+method+"("+
				       params+")"+
				       " not found!");
	}
	if (meth.isStatic()) {
	    reference=null;
	}
	MethodBodyStatement call = 
	    new MethodBodyStatement(meth, (ReferencePrefix) reference,
				    result, new ArrayOfExpression(args));
	Statement statement = call;
	StatementBlock sb = new StatementBlock(statement);
	
	if (exc>0) {
	    // pack it into a try:

	    // catch branch:
	    KeYJavaType kjt = getKeYJavaTypeForClass("Throwable");
	    ProgramVariable pvexc0 = new ProgramVariable(new ProgramElementName("thrownExc"), kjt);
	    VariableSpecification varSpec = new VariableSpecification(pvexc0, kjt);	    
	    ParameterDeclaration param 
		= new ParameterDeclaration(new Modifier[0], 
					   javaInfo().createTypeReference(kjt),
					   varSpec, false);
	    // catch body:
	    ExtList catchStmtList=new ExtList();
	    for (int i=0; i<exc; i++) {
		TypeReference typeRef=javaInfo().createTypeReference(pvexc[i].getKeYJavaType());
		Instanceof io=new Instanceof(pvexc0, typeRef);
		catchStmtList.add(new CopyAssignment(pvexc[i].getProgramVariable(), io));
	    } 
	    Branch branch = new Catch(param, new StatementBlock(catchStmtList));

	    Try tryStat = new Try(sb, new Branch[]{branch});
	    stmtLst.add(tryStat);
	    sb = new StatementBlock(stmtLst);
	} 
	return JavaBlock.createJavaBlock(sb);
    }
    
    public JavaBlock getMethodCall(ProgramVariable reference,
				   String method,
				   ProgramVariable result,
				   ProgramVariable[] args) {
	return getMethodCall(reference, method, result, args, emptyVarAndTypeArray);
    }

    public JavaBlock getMethodCall(String method,
				   ProgramVariable result,
				   ProgramVariable[] args) {
	return getMethodCall(null, method, result, args);
    }

    public JavaBlock getVoidMethodCall(ProgramVariable reference,
				       String method,
				       ProgramVariable[] args) {
	return getMethodCall(reference, method, null, args);
    }

    public JavaBlock getVoidMethodCall(String method,
				       ProgramVariable[] args) {
	return getVoidMethodCall(null,method,args);
    }


    public JavaBlock getConstructorCall(KeYJavaType cls,
					ProgramVariable result,
					ProgramVariable[] args) {
	

	TypeReference clsRef = new TypeRef(cls);
	
	ExtList lst = makeExtListFromArray(args);
	
	lst.add(0,clsRef); 
	lst.add(0,null); // we have no anonymous class ...
	
	New newCall = new New(lst,null);

	Statement statement;
	
	lst = new ExtList();
	lst.add(result);
	lst.add(newCall);
	statement = new CopyAssignment(lst);
	
	lst = new ExtList();
	lst.add(statement);
	StatementBlock sb = new StatementBlock(lst);//meth.getBody();
	return JavaBlock.createJavaBlock(sb);
    }


    public JavaBlock getVoidMethodCall(ProgramVariable reference,
				       String method,
				       ProgramVariable[] args, 
				       VarAndType[] excProgVars) {
	return getMethodCall(reference, method, null, args, excProgVars);
    }

  
    public ProgramVariable getAttribute(String name, 
					KeYJavaType classReference) {
	ProgramVariable pv;
	try{	   
	    pv=javaInfo ().getAttribute(name, classReference);
	    if (pv==null) { // for private attributes, though OCL has
			    // no visibilities
		pv=javaInfo ().getAttribute
		    (classReference.getSort().toString()+"::"+name, 
		     classReference);
	    }
        } catch(RuntimeException e) { // can this happen ??? %RB
           System.err.println("Attribute "+name+" in "+
                              classReference+" not found.");
	   
	   pv=null;
        }
	if (pv!=null) {
	    return pv; //Java Attribute
	} 
	return null;
    }

    

    public Term getAttributeTerm(ProgramVariable pvar, 
				 Term obj) {
	return pvar.isStatic() ? TermFactory.DEFAULT.createVariableTerm(pvar) : 
	    TermFactory.DEFAULT.createAttributeTerm(pvar, obj);
    }

    
    public Function getCollectionFunctionForFeature(String attr,
					  Sort[] argSorts,
					  Sort retSort) {
	Name n = null;
	String className;
 	Function f = null;
	if (argSorts.length > 0){
	    className = argSorts[0].name().toString();
	    n = new Name(attr);
	    f = (Function)(functions.lookup(n));
	    if (f == null) {
		n = new Name(className+"::"+attr);
		f = (Function)(functions.lookup(n));
	    }
	} else {
	    className = retSort.name().toString();
	    n = new Name(attr);
	    f = (Function)(functions.lookup(n));
	    if (f == null) {
		n = new Name(className+"::"+attr);
		f = (Function)(functions.lookup(n));
	    }
	}

	if ( f == null ) {
	    System.err.println("CREATE NEW FUNCTION "+className+" :: "+attr+"   "+functions);
	    f = new Function(n,retSort,argSorts);
	    functions.add(f);
	} else {
	    if (retSort != f.sort()) {
		throw new RuntimeException("Function "+n+
					   " previously declared with"+
					   " return type "+ f.sort()+
                                           " not with type "+retSort);
	    }
	    if ( argSorts.length != f.arity() ) {
		throw new RuntimeException("Function "+n+
					   " previously declared with"+
					   + f.arity() + " args.");
	    }
	    for ( int i = 0; i<f.arity(); i++ ) {
		if (argSorts[i] != f.argSort(i)) {
		    if (Debug.ENABLE_DEBUG==true) {
			throw new RuntimeException("Function "+n+
			   " previously declared with arg nr."+i+
			   " of type "+f.argSort(i)+"instead of "+argSorts[i]);
		    } else {
			n = new Name(className+"::"+attr);
			f = (Function)(functions.lookup(n));
			if (f == null) {
			    n = new Name(className+"::"+attr);
			    f = new Function(n,retSort,argSorts);
			    functions.add(f);
			}	
		    }
		}
	    }
	}
	return f;
    }



    public KeYJavaType getKeYJavaTypeOfName(String name) {
	return javaInfo ().getKeYJavaTypeByClassName(name);
    }


    // --------------- private utilities ---------------

    Sort getSortOfName(String name) {
	Name n = new Name(name);
	Sort s = (Sort)(sorts.lookup(n));
	if ( s == null ) {
	    Debug.out("WARNING: Sorts should all be defined in advance! "
		      +"Create one..."+n);
	    s = new PrimitiveSort(n);
	    sorts.add(s);
	}
	return s;
    }


    ExtList makeExtListFromArray(Object[] a) {
	ExtList res = new ExtList();
	for(int i = 0; i<a.length; i++) {
	    res.add(a[i]);
	}
	return res;
    }

}
