// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.ocl.ocldltranslator.control;

import tudresden.ocl.*;
import tudresden.ocl.parser.analysis.*;
import java.util.Vector;
import java.util.Set;
import java.util.Map;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Enumeration;

import tudresden.ocl.parser.node.*;
import tudresden.ocl.check.types.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.heuristics.*;

/**
 * An abstract class that serves for the management of
 * the different heuristics that might be applied on
 * some OCL constraint that will be translated.
 *
 * Each instances manages a list of registered heuristic.
 *
 * This class is the base class for such classes that
 * control the way some constraint will be treated by
 * the OCL translation.
 */

public abstract class BasicTranslationControl {

    /** All the heuristics that have been registered and therefore will be applied */
    protected Vector registeredHeuristics;
    

    // This map maps nodes in the AST to a translation type which specifies the way the logical translation
    // of that node will be generated when executing
    // the translation.
    // It uses Nodes as keys and delivers translation type objects representing the kind of translation 
    // to be performed for that node translation ...
    protected Map nodeTranslationTypeMap;
    

    /** Initilizes a translation control object with an empty set of heuristics  */
    public BasicTranslationControl(){
	this.registeredHeuristics = new Vector();
	this.nodeTranslationTypeMap = new HashMap();
    }



    // basic operations for the handling of translation type for nodes ...
    
    /**
     * Determines how the given node in the tree representation of the constraint
     * should be treated by the translation process itself.
     * This method should be called after <code>executeHeuristics</code>.
     * @return the TranslationType object that describes how the given node
     *         should be processed when translating the corresponding OCL expression.
     */
    protected TranslationType getTranslationTypeForNode(Node n){
	
	if(n == null){ 
	    error("Trying to get a translation type for a 'null' node !");
	}
	
	if (!nodeTranslationTypeMap.containsKey(n)) {
	   throw new RuntimeException("Translation type for node ("+ n +") required but not found !!!");
	    
	}
	return (TranslationType) nodeTranslationTypeMap.get(n); 
    }


    /** 
     * Adds a new heuristic to the translation control if it isn't already present.
     * The element will be added as the last element in the list.
     */
    public void  addHeuristic(BasicTranslationHeuristic th){
	if (!registeredHeuristics.contains(th)){
	    registeredHeuristics.add(th);
	}
    }

    /** 
     * Removes a heuristic from the translation control if it has been registered before.
     */
    public void removeHeuristic(BasicTranslationHeuristic th){
	if (registeredHeuristics.contains(th)){
	    registeredHeuristics.remove(th);
	}
    }

    /** 
     * @return the list of all registered heuristics.
     * Each such object is an instance of BasicTranslationHeuristic or some sublcass.
     */
    public Vector getRegisteredHeuristics(){
	return registeredHeuristics;
    }

    
    /**
     * Executes all the registered heuristics in the order they have been
     * registered on the given OCL expression that belongs the the specified
     * constraint tree.
     * @param constraintTree - the tree representation of the OCL constraint to be handled.
     * @param expression - the OCL expression to be translated and therefore marked by the
     *                     registered heuristics.
     * @param exprTranslator - the translator object to be used for the overall translation.
     */
    public void executeHeuristics(OclTree constraintTree, AExpression expression, 
				  KeyOclExpressionTranslator exprTranslator){
	this.nodeTranslationTypeMap = new HashMap();
	for(int i=0; i < registeredHeuristics.size(); i++){
	    BasicTranslationHeuristic currentHeuristic = 
		(BasicTranslationHeuristic) registeredHeuristics.elementAt(i);
	    
	    // Start current heuristic ...
	    currentHeuristic.determineTranslationTypesFor(constraintTree, expression, 
							  nodeTranslationTypeMap, exprTranslator);
	    
	    // Get new markings of the constraint tree ...
	    nodeTranslationTypeMap = currentHeuristic.getMarkingOfConstraintTree();
	}


    }

     
    protected void log(String text){
	System.err.println(text);
	System.err.println("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
    }
    
    // a method for constructing error messages ...
    void error(String msg)  {
	throw (new TranslationException(msg));
    }
    
    void error(String msg, Node node)  {
	throw(new TranslationException(msg,node));
    }





}


