// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.ocl.ocldltranslator.heuristics;

import tudresden.ocl.*;
import tudresden.ocl.parser.analysis.*;
import java.util.*;
import tudresden.ocl.parser.node.*;
import tudresden.ocl.check.types.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.control.*;

/**
 * This translation heuristic determines which boolean 
 * expressions of a constraint have to be handled by
 * an ADT 'Boolean' instead of formulas.
 * All other boolean expressions are implicitly determined to be
 * translated by use of some formula.
 *
 * A boolean term will be marked for a represenation by a Boolean term
 * when it is a subexpression of a :
 * - Collection-Literal of Boolean values
 * - A collect-expression that creates a Boolean collection.
 */

public class BooleanExpressionsAsFormulasHeuristic extends KeyTranslationHeuristic {

    // indicator sets for a useful usage of a translation by some term ...
    protected Set useTermModeNodes;

    /**
     * @return the name of this translation heuristic.
     */
    public String getHeuristicName(){
	return "Boolean Expressions As Formulas - Heuristic";
    }

 
    // ================ The main part of the heuristic ===================================
    // -> Care about OCL path expression ...

    
    // =========================================================================================================
    // ================================= The feature calls on OCL type Set[T] ==================================
    // =========================================================================================================


    protected void translateFeatureCallonSet(Node n, 
					     String featureName, 
					     Node appliedTo, 
					     Vector parameterNodes, Vector iteratorNames, 
					     Type iteratorType, boolean  featureCallWithAtPre){
	log("Feature call on Set : " + featureName);
	log("Applied on node :" + appliedTo.toString());
	
	if(featureName.equals("collect")){
	    Node expressionNode = (Node) parameterNodes.elementAt(0);
	    Type expressionType = tree.getNodeType(expressionNode);
	    log("Expression Type of the current collect expression == " +  expressionType);

	    boolean createsBooleanCollection  = ( expressionType == Basic.BOOLEAN);
	   
	    if (createsBooleanCollection) {
		useTermModeNodes.add(n);
		log("Marked current collect expression node to be represented by a BOOLEAN TERM!");	    
	    }
	}
    }


     // =========================================================================================================
    // ================================= The feature calls on OCL type Bag[T] ==================================
    // =========================================================================================================


    protected void translateFeatureCallonBag(Node n, 
					     String featureName, 
					     Node appliedTo, 
					     Vector parameterNodes, Vector iteratorNames, 
					     Type iteratorType, boolean  featureCallWithAtPre){
	log("Feature call on Bag : " + featureName);
	log("Applied on node :" + appliedTo.toString());
	
	if(featureName.equals("collect")){
	    Node expressionNode = (Node) parameterNodes.elementAt(0);
	    Type expressionType = tree.getNodeType(expressionNode);
	    log("Expression Type of the current collect expression == " +  expressionType);

	    boolean createsBooleanCollection  = ( expressionType == Basic.BOOLEAN);
	   
	    if (createsBooleanCollection) {
		useTermModeNodes.add(n);
		log("Marked current collect expression node to be represented by a BOOLEAN TERM!");	    
	    }
	}
    }


    // =========================================================================================================
    // ============================ The feature calls on OCL type Sequence[T] ==================================
    // =========================================================================================================


    protected void translateFeatureCallonSequence(Node n, 
						  String featureName, 
						  Node appliedTo, 
						  Vector parameterNodes, Vector iteratorNames, 
						  Type iteratorType, boolean  featureCallWithAtPre){
	log("Feature call on Sequence : " + featureName);
	log("Applied on node :" + appliedTo.toString());
	
	if(featureName.equals("collect")){
	    Node expressionNode = (Node) parameterNodes.elementAt(0);
	    Type expressionType = tree.getNodeType(expressionNode);
	    log("Expression Type of the current collect expression == " +  expressionType);

	    boolean createsBooleanCollection  = ( expressionType == Basic.BOOLEAN);
	   
	    if (createsBooleanCollection) {
		useTermModeNodes.add(n);
		log("Marked current collect expression node to be represented by a BOOLEAN TERM!");	    
	    }
	}
    }


    /**
     * Checks whether the expression structure indicates that it would be useful to
     * represent the given  Node (corresponding to some boolean expression) by a Boolean term
     * instead of a formula, e.g. some construct in the constraint tree can be handled more
     * easily (without the introduction of a new function symbol)
     * @return true iff the tree structure gives rise to a useful application of a term representation
     *                  for the OCL expression corresponding to the given node.
     */
    protected boolean usageOfTermTranslationSetFromAbove(Node n){
	boolean termIndicatorFound = useTermModeNodes.contains(n);
	boolean finished = termIndicatorFound;
	Node currentNode = n;
	Node parentNode;
	while(!finished){
	    parentNode = currentNode.parent();
	    if (parentNode != null) {
		termIndicatorFound = useTermModeNodes.contains(parentNode);
		finished = termIndicatorFound;
		currentNode = parentNode;
	    } else {
		finished = true;
	    }

	}
	
	return termIndicatorFound;

    }



    protected void translateFeatureCallonOclType(Node n, 
						 String featureName, Node appliedTo, 
						 Vector parameterNodes, Vector iteratorNames, 
						 boolean featureCallWithAtPre, boolean featureCallWithArrow, 
						 boolean isMethodCall ){
	log("Feature call on OclType : " + featureName);
	log("Applied on node :" + n.toString());
	
	// Check for static feature ...
	
	boolean isStaticFeature = ( !featureName.equals("allInstances") && 
				    !featureName.equals("name") && 
				    !featureName.equals("attributes") &&
				    !featureName.equals("associationEnds") &&
				    !featureName.equals("operations") &&
				    !featureName.equals("supertypes") &&
				    !featureName.equals("allSupertypes") );
	
	log("Current feature call on type object is a static call :" + isStaticFeature);


	Type resultType = tree.getNodeType(n);
	if ((resultType == Basic.BOOLEAN) && isStaticFeature) {
	    // For KeY: Always mark static boolean class features to be represented by boolean terms ...
	    useTermModeNodes.add(n);
	    log("Marked static boolean class feature node to be represented by a BOOLEAN TERM!");	
	    
	}

    }
    
    protected void translateFeatureCallonModelType(Node n, 
						   String featureName, Node appliedTo, 
						   Vector parameterNodes, Vector iteratorNames, 
						   boolean  featureCallWithAtPre, 
						   boolean featureCallWithArrow, boolean isMethodCall ){
	log("Feature call on UML Model type : " + featureName);
	log("Applied on node :" + appliedTo.toString());
	
	Type resultType = tree.getNodeType(n);
	if (resultType == Basic.BOOLEAN) {
	    // For KeY: Always mark boolean model type features to be represented by boolean terms ...
	    useTermModeNodes.add(n);
	    log("Marked boolean model type feature node to be represented by a BOOLEAN TERM!");	
	    
	}
    }



    // =, <>, <, <=, >, >=
    // Use predicates for '=' and '<>' on sets if possible ...
    protected void translateRelationalExpression(Node re, 
						 String relationalOperatorName, 
						 Node leftOperand, Node rightOperand){
	log("Translating relational expression with Operator '"+ relationalOperatorName  +"'.");
	
	Type operandTypeLeft = tree.getNodeType(leftOperand);
	Type operandTypeRight = tree.getNodeType(rightOperand);
	
	if( (operandTypeLeft == Basic.BOOLEAN) && (operandTypeRight == Basic.BOOLEAN) ){
	    
	    if (relationalOperatorName.equals("=") || relationalOperatorName.equals("<>")) {
		boolean useTermTranslation = usageOfTermTranslationSetFromAbove(re);
		if(useTermTranslation){
		    useTermModeNodes.add(re);
		    log("Marked boolean relational expression to be represented by a BOOLEAN TERM!");	
		}
		
	    }
	}
    }

    // and, or, xor, implies
    protected void translateLogicalExpression(Node le, 
					      String logicalOperatorName, 
					      Node leftOperand, 
					      Node rightOperand){

	log("XX Translating logical expression with Operator '"+ logicalOperatorName  +"'.");
	
	boolean useTermTranslation = usageOfTermTranslationSetFromAbove(le);
	
	log("Logical expression as Boolean term is indicated from above == " + useTermTranslation);

	if(useTermTranslation){
	    useTermModeNodes.add(le);
	    log("Marked logical expression to be represented by a BOOLEAN TERM!");	
	}
	
    }
    
    // true, false
    protected void translateBooleanLiteral(Node sl, String lit){
	log("Translating Boolean literal: " + lit);
	
	boolean useTermTranslation = usageOfTermTranslationSetFromAbove(sl);
	if(useTermTranslation){
	    useTermModeNodes.add(sl);
	    log("Marked boolean literal to be represented by a BOOLEAN TERM!");	
	}

    }

    // -(...), not
    protected void  translateUnaryExpression(Node ue,
					     String unaryOperatorName,
					     Node operand){
	log("Translating unary expression with Operator '"+ unaryOperatorName  +"'.");
	
	if(unaryOperatorName.equals("not")){
	    boolean useTermTranslation = usageOfTermTranslationSetFromAbove(ue);
	    if(useTermTranslation){
		useTermModeNodes.add(ue);
		log("Marked boolean unary expression to be represented by a BOOLEAN TERM!");	
	    } 
	}

    }

    
     protected void translateIfThenElseExpression(Node ie, Node ifExpression, Node thenExpression,
						 Node elseExpression){
	log("Translating if-then-else-construct.");
	
	Type thenExpressionType = tree.getNodeType(thenExpression);
	Type elseExpressionType = tree.getNodeType(elseExpression);
	
	if( (thenExpressionType == Basic.BOOLEAN) && (elseExpressionType == Basic.BOOLEAN)) {
	    boolean useTermTranslation = usageOfTermTranslationSetFromAbove(ie);
	    if(useTermTranslation){
		useTermModeNodes.add(ie);
		log("Marked boolean literal to be represented by a BOOLEAN TERM!");	
	    }    
	}
     }



    // Now handle variables ...
    // This is necessary for the handling of the the translation types for expressions defined by let constructs.
    // The letVariable is used to let the translation types of the in-expression affect 
    // those of the defined expression !

    
    protected void translateVariable(Node n, String varName , Type varType){
	String varTypeString = varType.toString();
	log("Translating variable '"+ varName +"' of type '"+ varTypeString +"'.");
	
	if (varType == Basic.BOOLEAN) {
	    log("Check whether the variablename should be marked (from above) ...");

	    // Check wheter the variable has already been marked by some other subexpression (e.g. b in
	    // let b : Boolean = exp in (not b and c->collect(b and ...)->forAll(...))
	    // In the case that b has been marked for a termlike translation, no other mark may be set.
	    // REMARK: ALL occurences of some let variable b will therefore be translated in the same way.
	    // It may be better to allow DIFFERENT translations for DIFFERENT occurences, but things would
	    // get more complicated in the translation code. Because the benefits for such a more general
	    // approach are not clear at the moment, we currently omit this possibility.
	    TranslationType previouslySetTranslationType =  getTranslationTypeForVariable(varName, varTypeString);
	    boolean useTermTrans =  
		((previouslySetTranslationType.containsMark(TranslationMark.AS_BOOLEAN_TERM) ) || 
		 usageOfTermTranslationSetFromAbove(n));
	
	    TranslationType transType = new TranslationType();
	    if (useTermTrans){
		// mark the corresponding variable for term translation ...
		transType.addMark(TranslationMark.AS_BOOLEAN_TERM);
	    } else {
		transType.addMark(TranslationMark.DEFAULT);
	    }

	    setTranslationTypeForVariable(varName, varTypeString, transType);
	    log("Marked variable '"+ varName +"' of type '"+ varTypeString +"' with translation type :" + transType);
	}

    }


    // Handling let-expression correctly ...
    protected void translateLetExpression(Node le, 
					  String letVarName,Type letVarType, 
					  Node defExpression, Node inExpression ){
	String letVarTypeString = letVarType.toString();
	log("Translating LET-construct with Variable '" + letVarName + "'.");
	TranslationType transType;
	
	if (letVarType == Basic.BOOLEAN) {
	    // Get the type of translation that should be applied to the defined expression  ...
	    transType = getTranslationTypeForVariable(letVarName, letVarTypeString);
	    
	    log("Let-Variable '"+ letVarName +"' of type '"+ letVarTypeString 
		+"' was set to be translated as : " + transType + "\nMarking the defined expression in that way : " 
		+ defExpression + " (hashCode = "+ defExpression.hashCode() +")");

	    if (transType.containsMark(TranslationMark.AS_BOOLEAN_TERM)){
		useTermModeNodes.add(defExpression);
		log("Marked abbreviated boolean let expression be represented by a BOOLEAN TERM!");	
	    } 
	}
	

	// We don't need the variable mark any longer ...
	removeTranslationTypeForVariable(letVarName, letVarTypeString);
	
    }

    
    protected void translateCollectionLiteralWithExpression(Node cl, Type collType, Vector expressions){
	log("Translating Collection literal with expressions for type: " + collType.toString());

	Node currentExpression; 
	Type currentType;
	for(int i=0; i<expressions.size(); i++){
	    currentExpression = (Node) expressions.elementAt(i);
	    currentType = tree.getNodeType(currentExpression);
	    
	    if (currentType == Basic.BOOLEAN){
		// The collection literal uses some boolean expression, therefore
		// we have to use a term representation for such nodes ...
		useTermModeNodes.add(currentExpression);
		log("Marked element expression ind collection literal to be represented by a BOOLEAN TERM: " 
		    + currentExpression);	
	    }
	   

	}

		
    }
    
  


    // ================ Start the computation of the translation types =====================
	    
    public void determineTranslationTypesFor(OclTree constraintTree, 
					     AExpression expression, 
					     Map currentMarking,
					     KeyOclExpressionTranslator exprTranslator){
		
	this.tree = constraintTree;
	this.expression = expression;
	this.exprTranslator = exprTranslator;

	this.varMap = new HashMap();
	nodeTranslationTypeMap = currentMarking;
	
	useTermModeNodes = new HashSet();
	
	expression.apply(this);

	// NOW, mark the computed node set accordingly ...
	Node currentNode;

	TranslationType currentType;

	// For the termlike nodes ...
	for (Iterator it=useTermModeNodes.iterator(); it.hasNext(); ){
	    currentNode = (Node) it.next();
	    currentType =  getTranslationTypeForNode(currentNode);
	    currentType.addMark(TranslationMark.AS_BOOLEAN_TERM);
	    setTranslationTypeForNode(currentNode, currentType);
	}

	
    }
      
}







