// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.ocl.ocldltranslator.heuristics;

/**
 * Implements a simple heuristic for translation control:
 * Only path expression in OCL will be handeled by the heuristic.
 * Such path expression will be translated in a predicative way when
 * the following holds :
 * 
 * It is possible to describe the collection (enitity) represented by 
 * the path expression by use of a formula without introducing any new
 * axioms.
 * For details please refer to my diploma thesis.
 */
import tudresden.ocl.*;
import tudresden.ocl.parser.analysis.*;
import java.util.Vector;
import java.util.Set;
import java.util.Map;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Enumeration;
import tudresden.ocl.parser.node.*;
import tudresden.ocl.check.types.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.control.*;


public class CollectionsAsSetsHeuristic extends KeyTranslationHeuristic {

    protected boolean usePredicativeTranslationMode;
    
    // indicator sets for a useful usage of predicate or term like translation ...
    protected Set usePredicativeModeNodes;
    protected Set useTermModeNodes;

    // indicator for variables. These are used for handling let-expressions ...
    protected Set usePredicativeModeVariables;
    protected Set useTermModeVariables;

    /**
     * @return the name of this translation heuristic.
     */
    public String getHeuristicName(){
	return "Collections as Sets - Heuristic";
    }

 
    // ================ The main part of the heuristic ===================================
    // -> Care about OCL path expression ...

    
    public void inAPostfixExpression(APostfixExpression pe) {
	
	Vector featureCalls = new Vector();

	if (!pe.getPostfixExpressionTail().isEmpty()) {
	    // we have a 'real' ocl path expression ...
	    
	    ListIterator iter = pe.getPostfixExpressionTail().listIterator();
	    AFeatureCall currentFeatureCall;
	    APostfixExpressionTail currentPET;
	    while (iter.hasNext()){
		currentPET = (APostfixExpressionTail) iter.next();
		currentFeatureCall = (AFeatureCall) currentPET.getFeatureCall();
		featureCalls.add(currentFeatureCall);
	    } 
	
	    handleStartOfOCLPathExpression(pe,featureCalls);
	}	
	
    }



    protected void handleStartOfOCLPathExpression(APostfixExpression pe, Vector featureCalls){
	// Check whether the type of the last feature call is boolean ...
	if (!featureCalls.isEmpty()){
	    // AFeatureCall lastFeatureCall = (AFeatureCall) featureCalls.lastElement();
	   
	   
	    Type pathExpressionType = tree.getNodeType(pe);

	    log("Return Type of the current path expression == " + tree.getNodeType(pe));

	    boolean lastFeatureCallReturnsBoolean = ( pathExpressionType == Basic.BOOLEAN);
	    boolean lastFeatureCallReturnsCollectionType = (pathExpressionType instanceof Collection);


	    // Check whether there are some critical featurecalls in the given path expression ...
	    boolean allFeatureCallsUncritical = true;
	    AFeatureCall fc;
	    Node prevNode =  pe.getPrimaryExpression();
	    for(Enumeration e = featureCalls.elements(); allFeatureCallsUncritical && e.hasMoreElements();){
		fc = (AFeatureCall) e.nextElement();
		allFeatureCallsUncritical = uncriticalFeatureCall(fc, prevNode);
		prevNode = fc;
	    }

	    log("All feature calls uncritical  == " + allFeatureCallsUncritical);

	    if (lastFeatureCallReturnsBoolean && allFeatureCallsUncritical) {
		log("Got a OCL path expression (of length > 1) with return type BOOLEAN : " + pe.toString()+ 
		    "\n\nSet predicative translation for that expression !");
	
		usePredicativeModeNodes.add(pe);
	    } else {
		if (lastFeatureCallReturnsCollectionType && allFeatureCallsUncritical) {
		    // If the expression containing the current OCL path expression indicates that a predicative
		    // representation of the resulting collection would be useful, do it ...
		    if (usageOfPredicativeTranslationSetFromAbove(pe)){
			log("Got a OCL path expression (of length > 1) with return type SET/BAG/SEQUENCE : " + pe.toString()+ 
			    "\n\nSet predicative translation for that expression because of indicator from conaining expression !");
			
			usePredicativeModeNodes.add(pe);
		    }
		} else {
		    // especially when the return type of the path expression is Integer, Real or more general T (for a Type T) ...
		    log("Got a OCL path expression (of length > 1) with return type OTHER than BOOLEAN : " + pe.toString()+ 
			"\n\nSet term translation for that expression !");
		    
		    useTermModeNodes.add(pe);
		}
		
	    }
	    
	}
    }

    
    /**
     * Checks whether the given featureCall is uncritical in the sense that is can be handeled 
     * by predicates.
     */
    boolean uncriticalFeatureCall(AFeatureCall fc, Node appliedTo){
	boolean result = false;
	String featureName=fc.getPathName().toString().trim();
	Type appliedToType = tree.getNodeType(appliedTo);

	if (appliedToType == null){ 
	    // Use the right node in the AST in order to determine the type information !
	    appliedToType = tree.getNodeType(appliedTo.parent());
	}
	
       
	boolean featureIsCollectionFeature = (appliedToType instanceof Collection);

	
	boolean featureIsCritical = 
	    featureIsCollectionFeature && (
	    featureName.equals("at") ||
	    featureName.equals("first") ||
	    featureName.equals("last") ||
	    featureName.equals("subSequence") ||
	    featureName.equals("iterate") 
	    );
	
	result = !featureIsCritical;
	return result;
	
	
    }
    
    

    /**
     * Checks whether the expression structure indicates that it would be useful to translate the given node
     * (or more precisely its corresponding OCL expression) in a predicative way.
     * @return true iff the tree structure gives rise to a useful application af a predicative representation
     *                  for the OCL expression corresponding to the given node.
     */
    protected boolean usageOfPredicativeTranslationSetFromAbove(Node n){
	boolean termIndicatorFound = false;
	boolean predicateIndicatorFound = false;
	boolean finished = false;
	boolean result = false;
	Node currentNode = n;
	Node parentNode;
	while(!finished){
	    parentNode = currentNode.parent();
	    if (parentNode != null) {
		termIndicatorFound = useTermModeNodes.contains(parentNode);
		predicateIndicatorFound = usePredicativeModeNodes.contains(parentNode);
		finished = termIndicatorFound || predicateIndicatorFound;
		currentNode = parentNode;
	    } else {
		finished = true;
	    }

	}
	
	if (termIndicatorFound) {
	    result = false;
	} else {
	    result = predicateIndicatorFound;
	}

	return result;

    }

    /**
     * Checks whether the expression structure indicates that it would be useful to translate the given node
     * (or more precisely its corresponding OCL expression) in as a term.
     * @return true iff the tree structure gives rise to a useful application of a term representation
     *                  for the OCL expression corresponding to the given node.
     */
    protected boolean usageOfTermTranslationSetFromAbove(Node n){
	boolean termIndicatorFound = false;
	boolean predicateIndicatorFound = false;
	boolean finished = false;
	boolean result = false;
	Node currentNode = n;
	Node parentNode;
	while(!finished){
	    parentNode = currentNode.parent();
	    if (parentNode != null) {
		termIndicatorFound = useTermModeNodes.contains(parentNode);
		predicateIndicatorFound = usePredicativeModeNodes.contains(parentNode);
		finished = termIndicatorFound || predicateIndicatorFound;
		currentNode = parentNode;
	    } else {
		finished = true;
	    }

	}
	
	if (termIndicatorFound) {
	    result = true;
	} else {
	    result = !(predicateIndicatorFound);
	}

	return result;

    }



    
    // REMARK: In the following methods the nodes which should be translated in a predicative way will be marked accordingly.
    // BUT: If ALL those nodes will be translated in this way will be decided by the translation routine. This means
    // That more nodes will be marked here for predicative translation than actually translated in this way by the
    // translation routine. This is no problem but leads to simpler program code in this class. 
    // SO please read the given marks for the nodes as CAN (proposals) instead of MUST (commands) for the
    // translation routine.

    protected void translateFeatureCallonSet(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType, boolean  featureCallWithAtPre){
	log("Feature call on Set : " + featureName);
	log("Applied on node :" + appliedTo.toString());

	boolean usePredicativeTranslation = usageOfPredicativeTranslationSetFromAbove(n);
	if(usePredicativeTranslation){
	    // Mark also appliedTo node if not already done ...
	    if (!nodeTranslationTypeMap.containsKey(appliedTo)) {
		usePredicativeModeNodes.add(appliedTo);
		log("Marked node as predicative :" + appliedTo.toString());
	    }

	    // finally mark the current node itself ...
	    usePredicativeModeNodes.add(n);
	    log("Marked node as predicative :" + n.toString());
	}
	
    }

    protected void translateFeatureCallonBag(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType , boolean  featureCallWithAtPre){
	log("Feature call on Bag : " + featureName);
	log("Applied on node :" + n.toString());


	boolean usePredicativeTranslation = usageOfPredicativeTranslationSetFromAbove(n);
	if(usePredicativeTranslation){
	    // Mark also appliedTo node if not already done ...
	    if (!nodeTranslationTypeMap.containsKey(appliedTo)) {
		usePredicativeModeNodes.add(appliedTo);
		log("Marked node as predicative :" + appliedTo.toString());
	    }

	    // finally mark the current node itself ...
	    usePredicativeModeNodes.add(n);
	    log("Marked node as predicative :" + n.toString());
	}

    }

    protected void translateFeatureCallonSequence(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType, boolean  featureCallWithAtPre){
	log("Feature call on Sequence : " + featureName);
	log("Applied on node :" + n.toString());
	
	boolean usePredicativeTranslation = usageOfPredicativeTranslationSetFromAbove(n);
	if(usePredicativeTranslation){
	    // Mark also appliedTo node if not already done ...
	    if (!nodeTranslationTypeMap.containsKey(appliedTo)) {
		usePredicativeModeNodes.add(appliedTo);
		log("Marked node as predicative :" + appliedTo.toString());
	    }

	    // finally mark the current node itself ...
	    usePredicativeModeNodes.add(n);
	    log("Marked node as predicative :" + n.toString());
	}


    }
    
    protected void translateFeatureCallonOclType(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow, boolean isMethodCall ){
	log("Feature call on OclType : " + featureName);
	log("Applied on node :" + n.toString());
	
	boolean usePredicativeTranslation = usageOfPredicativeTranslationSetFromAbove(n);
	if(usePredicativeTranslation){
	    // Mark also appliedTo node if not already done ...
	    if (!nodeTranslationTypeMap.containsKey(appliedTo)) {
		usePredicativeModeNodes.add(appliedTo);
		log("Marked node as predicative :" + appliedTo.toString());
	    }

	    // finally mark the current node itself ...
	    usePredicativeModeNodes.add(n);
	    log("Marked node as predicative :" + n.toString());
	}


    }
    
    protected void translateFeatureCallonModelType(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow, boolean isMethodCall ){
	log("Feature call on UML Model type : " + featureName);
	log("Applied on node :" + appliedTo.toString());
	
	boolean usePredicativeTranslation = usageOfPredicativeTranslationSetFromAbove(n);
	if(usePredicativeTranslation){
	    // Mark also appliedTo node if not already done ...
	    if (!nodeTranslationTypeMap.containsKey(appliedTo)) {
		usePredicativeModeNodes.add(appliedTo);
		log("Marked node as predicative :" + appliedTo.toString());
	    }

	    // finally mark the current node itself ...
	    usePredicativeModeNodes.add(n);
	    log("Marked node as predicative :" + n.toString());
	}


    }



    // =, <>, <, <=, >, >=
    // Use predicates for '=' and '<>' on sets if possible ...
    protected void translateRelationalExpression(Node re, String relationalOperatorName, Node leftOperand, Node rightOperand){
	log("Translating relational expression with Operator '"+ relationalOperatorName  +"'.");
	
	Type operandTypeLeft = tree.getNodeType(leftOperand);
	Type operandTypeRight = tree.getNodeType(rightOperand);

	if(!operandTypeLeft.equals(operandTypeRight)){
	    // We can not handle different types for relational expressions yet !
	    // error("Different types for the subexpressions of a RELATIONAL-expression can't be handeled YET !!");
	}

	if (relationalOperatorName.equals("=") || relationalOperatorName.equals("<>")) {
	    Type t = operandTypeLeft;
	    
	    if (t instanceof Collection) {
		Collection c=(Collection) t;
		if (c.getCollectionKind() == Collection.SET){
		    // We have equality on sets, therefore predicates would be good (if possible)  ...
		    usePredicativeModeNodes.add(re);
		    log("Marked node as predicative :" + re.toString());
		    
		} 

	    }
	}
    }
    

    
    // Now handle variables ...
    // This is necessary for the handling of the the translation types for expressions defined by let constructs.
    // The letVariable is used to let the translation types of the in-expression affect those of the defined expression !

    
    protected void translateVariable(Node n, String varName , Type varType){
	String varTypeString = varType.toString();
	log("Translating variable '"+ varName +"' of type '"+ varTypeString +"'.");
	
	log("Check whether the variablename should be marked (from above) ...");

	
	// Check wheter the variable has already been marked by some other subexpression (e.g. p in
	// let p : Set(T) = exp in (p->size = 1 and p->forAll(b))
	// In the case that p has been marked for a termlike translation, no other mark may be set.
	// REMARK: ALL occurences of some let variable p will therefore be translated in the same way.
	// It may be better to allow DIFFERENT translations for DIFFERENT occurences, but things would
	// get more complicated in the translation code. Because the benefits for such a more general
	// approach are not clear at the moment, we currently omit this possibility.
	TranslationType previouslySetTranslationType =  getTranslationTypeForVariable(varName, varTypeString);
	
	boolean useTermTrans =  (  previouslySetTranslationType.containsMark(TranslationMark.AS_TERM) ||
				   usageOfTermTranslationSetFromAbove(n) );

	
	TranslationType transType = new TranslationType();
	if (useTermTrans){
	    // mark the corresponding variable for term translation ...
	    transType.addMark(TranslationMark.AS_TERM);
	
	} else {
	    boolean usePredTrans =  usageOfPredicativeTranslationSetFromAbove(n);
	    if (usePredTrans){
		// mark the corresponding variable for predicative translation ...
		transType.addMark(TranslationMark.AS_PREDICATE);
	    }
	}

	setTranslationTypeForVariable(varName, varTypeString, transType);
	log("Marked variable '"+ varName +"' of type '"+ varTypeString +"' with translation type :" + transType);
	
    }


    // Handling let-expression correctly ...
    protected void translateLetExpression(Node le, String letVarName,Type letVarType, Node defExpression, Node inExpression ){
	String letVarTypeString = letVarType.toString();
	log("Translating LET-construct with Variable '" + letVarName + "'.");
	TranslationType transType;
	
	// Get the type of translation that should be applied to the defined expression  ...
	transType = getTranslationTypeForVariable(letVarName, letVarTypeString);

	log("Let-Variable '"+ letVarName +"' of type '"+ letVarTypeString 
	    +"' was set to be translated as : " + transType + "\nMarking the defined expression in that way : " 
	    + defExpression + " (hashCode = "+ defExpression.hashCode() +")");

	if (transType.containsMark(TranslationMark.AS_PREDICATE)){
	    usePredicativeModeNodes.add(defExpression);
	} else if (transType.containsMark(TranslationMark.AS_TERM)){
	    useTermModeNodes.add(defExpression);
	}
	
	// We don't need the variable mark any longer ...
	removeTranslationTypeForVariable(letVarName, letVarTypeString);
	
    }

    
    protected void translateCollectionLiteralWithExpression(Node cl, Type collType, Vector expressions){
	log("Translating Collection literal with expressions for type: " + collType.toString());

	boolean usePredicativeTranslation = usageOfPredicativeTranslationSetFromAbove(cl);
	if(usePredicativeTranslation){
	    // mark the current node itself ...
	    usePredicativeModeNodes.add(cl);
	    log("Marked node as predicative :" + cl.toString());
	}
	
    }
    
    
    protected void translateCollectionLiteralWithRange(Node cl, Type collType, Node lowerBound,
						       Node upperBound){
	log("Translating Collection literal with range for type: " + collType.toString());

	boolean usePredicativeTranslation = usageOfPredicativeTranslationSetFromAbove(cl);
	if(usePredicativeTranslation){
	    // mark the current node itself ...
	    usePredicativeModeNodes.add(cl);
	    log("Marked node as predicative :" + cl.toString());
	}
    }

    


    // ================ Start the computation of the translation types =====================
	    
    public void determineTranslationTypesFor(OclTree constraintTree, 
					     AExpression expression, 
					     Map currentMarking,
					     KeyOclExpressionTranslator exprTranslator){
		
	this.tree = constraintTree;
	this.expression = expression;
	this.exprTranslator = exprTranslator;

	this.varMap = new HashMap();
	nodeTranslationTypeMap = currentMarking;
	
	usePredicativeModeNodes = new HashSet();
	useTermModeNodes = new HashSet();
	

	expression.apply(this);

	// NOW, mark the computed node sets accordingly ...
	Node currentNode;

	TranslationType currentType;

	// For the predicative nodes ...
	for (Iterator it=usePredicativeModeNodes.iterator(); it.hasNext(); ){
	    currentNode = (Node) it.next();
	    currentType =  getTranslationTypeForNode(currentNode);
	    currentType.addMark(TranslationMark.AS_PREDICATE);
	    setTranslationTypeForNode(currentNode, currentType);
	}
	
	// For the termlike nodes ...
	for (Iterator it=useTermModeNodes.iterator(); it.hasNext(); ){
	    currentNode = (Node) it.next();
	    currentType =  getTranslationTypeForNode(currentNode);
	    currentType.addMark(TranslationMark.AS_TERM);
	    setTranslationTypeForNode(currentNode, currentType);
	}

	
    }

    
}







