// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.ocl.ocldltranslator.test;

/*
 * Class is a default implementation of a SAX document handler. 
 */

import java.net.URL;
import java.util.*;
import org.xml.sax.*;


import tudresden.ocl.*;
import tudresden.ocl.parser.analysis.*;
import tudresden.ocl.parser.node.*;
import tudresden.ocl.check.types.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.*;
import java.io.*;

import  org.apache.xerces.framework.*;
import  org.apache.xerces.parsers.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.heuristics.*;
import de.uka.ilkd.key.ocl.ocldltranslator.control.*;
import de.uka.ilkd.key.ocl.ocldltranslator.normalize.*;
import de.uka.ilkd.key.ocl.ocldltranslator.transform.*;

/**
 * A class for reading a test suite file for testing the transformation of OCL constraints
 * in DL formulas as used in the KeY-System.
 */
public class TestSuiteReader implements DocumentHandler {
    
    // Some constants for XML tags ...
    public static final String E_TESTSUITE = "TESTSUITE";
    public static final String E_NAME = "NAME";
    public static final String E_DESCRIPTION = "DESCRIPTION";
    public static final String E_XMI_MODEL = "XMI-MODEL";
    public static final String E_IMPLEMENTATION_SOURCES = "IMPLEMENTATION-SOURCES";
    public static final String E_TESTCASE = "TESTCASE";
    public static final String E_CONSTRAINT = "CONSTRAINT";
    public static final String E_EXPRESSION = "EXPRESSION";
    public static final String E_OPTION = "OPTION";
    public static final String A_NAME = "name";
    public static final String A_VALUE = "value";
    public static final String A_FILE = "file";
    public static final String A_DIRECTORY = "directory";
    public static final String A_CONTEXTELEMENT = "contextelement";
    public static final String A_TYPE = "type";

    // If set to true all incoming events will be logged on the given output stream ...
    protected boolean _outputEnabled = false;

    // Writer for the output information ...
    protected  PrintWriter _outputWriter = null;

    protected String testSuiteFileName;
    protected InputSource testSuiteInputData;
    protected  org.apache.xerces.parsers.SAXParser xmlParser;

    protected TransformationTestSuite testSuite;


    // Some state variables for during the parsing process ...
    protected TransformationTestSuite currentTestSuite;
    protected TransformationTestCase currentTestCase;
    protected String currentText;

    protected String currentExprType;
    protected String currentExprContextElement;

    public TestSuiteReader() {
	
	xmlParser = new  org.apache.xerces.parsers.SAXParser();
	
	// Register as document handler ...
	xmlParser.setDocumentHandler(this);

    }
    

    public void readTestSuite(String fileName){
	
	// synchronized(waitingObject){
	    
	    testSuiteFileName = fileName;
	    testSuiteInputData = getFileInputSourceFor(testSuiteFileName);
	    
	    // start parser on the given input source ...
	    // System.err.println("Start parsing test suite in file : " + fileName);
	    try {
		xmlParser.parse(testSuiteInputData);
		
		// System.err.println("Finished parsing test suite successfully !");
	    } catch (SAXException se) {
		System.err.println("An error occured during processing of file : " + fileName);
		se.printStackTrace();
	    } catch (IOException ioe){
		System.err.println("An error occured during processing of file : " + fileName);
		ioe.printStackTrace();
	    } finally {
		try {
		    xmlParser.reset();
		} catch (Exception e){
		    System.err.println("An error occured while reseting the XML-parser ...");
		    e.printStackTrace();
		}
	    }


	    // }

    }

    
    
    public TransformationTestSuite getTestSuite(){
	return testSuite;
    }

    

    // ======================== Some logging methods ========================================


    /**
     * Forces the instance to log information about the received SAX events on the given
     * output stream
     */
    public void enableOutputLogging(OutputStream os){
	if (os != null) {
	    _outputEnabled = true;
	    _outputWriter = new PrintWriter(os,true);
	}
    }

    public void enableOutputLogging(PrintWriter pw){
	if (pw != null) {
	    _outputEnabled = true;
	    _outputWriter = pw;
	}
    }
    
    public void disableOutputLogging(){
	_outputEnabled = false;
	_outputWriter = null;
    }



    
    // ======================  Implementation of the SAX interface ================================

    public void characters(char[] ch, int start, int length) throws org.xml.sax.SAXException {
	// Receive notification of character data.
	// do optional output ...
	if(_outputEnabled){
	    _outputWriter.println("-- Got CHARACTERS :");
	    _outputWriter.println(new String(ch,start,length));
	    _outputWriter.flush();
	}

	currentText = currentText + new String(ch,start,length);
	
	
    }
    
    public void endDocument() throws org.xml.sax.SAXException { 
	// Receive notification of the end of a document.

	// do optional output ...
	if(_outputEnabled){
	    _outputWriter.println("---------------- END of XML-Document reached ----------------");
	    _outputWriter.flush();
	    _outputWriter.close();
	
	}
	
	testSuite = currentTestSuite;
	
    }

    public void endElement(java.lang.String name) throws org.xml.sax.SAXException {
	// Receive notification of the end of an element.
	// do optional output ...
	if(_outputEnabled){
	    _outputWriter.println("-- GOT ENDELEMENT " );
	    _outputWriter.println("NAME = </" + name + ">" );
	    _outputWriter.flush();
	}
	// Now we interpret the received data ...
	
	if (name.equals(E_DESCRIPTION)) {
	    if (currentTestCase == null) {
		// No test case opened, therefore the description belongs to the test suite ...
		currentTestSuite.setDescription(currentText);
	    } else {
		// the description belongs to the current test case ...
		currentTestCase.setDescription(currentText);
	    }
	} else if (name.equals(E_TESTCASE)) {
	    currentTestSuite.addTestCase(currentTestCase);
	    currentTestCase = null;

	} else if (name.equals(E_CONSTRAINT)) {
	    // First we have to replace all entity-references &lt; for '<' in the
	    // string. This is because of the syntax of XML tags :-(
	    // This has to be done in the future ...

	    currentTestCase.setConstraint(currentText);
	} else if (name.equals(E_EXPRESSION)) {
	    String constraintText;
	    constraintText = "context "+ currentExprContextElement + " " + 
		currentExprType + ":\n" + currentText;

	    currentTestCase.setConstraint(constraintText);
	    currentTestCase.setExpression(currentText);
	    
	    
	} 
	
	

	
    }
  
    public void ignorableWhitespace(char[] ch, int start, int length) throws org.xml.sax.SAXException {
	// Receive notification of ignorable whitespace in element content.
	// do optional output ...
	if(_outputEnabled){
	    _outputWriter.println("-- Got IGNORABLE WHITESPACE :");
	    _outputWriter.println(new String(ch,start,length));
	    _outputWriter.flush();
	}
	
    }

    
    
    public void startDocument() throws org.xml.sax.SAXException {
	// Receive notification of the beginning of a document.
	
	// do optional output ...
	if (_outputEnabled) {
	    _outputWriter.println("---------------- START of XML-Document ------------------");
	    _outputWriter.flush();
	}
 
    }
  
    public void startElement(java.lang.String name, AttributeList atts) throws org.xml.sax.SAXException { 
	// Receive notification of the beginning of an element. 
	
	// do optional output ...
	if(_outputEnabled){
	    _outputWriter.println("-- GOT STARTELEMENT " );
	    _outputWriter.println("NAME = <" + name + ">" );
	    _outputWriter.println("ATTRIBUTES ... ");

	    String aName,aType,aValue;
	    for (int i = 0; i < atts.getLength(); i++) {
		aName = atts.getName(i);
		aType = atts.getType(i);
		aValue = atts.getValue(i);
		_outputWriter.println("["+i+"] "+aName+ " = " +aValue+ " (Type : "  + aType +" )");
	    } // for

	    _outputWriter.flush();
	    
	}
	
	
	// Now we interpret the received data ...

	if (name.equals(E_TESTSUITE)) {

	    currentTestSuite =  new TransformationTestSuite();

	} else if (name.equals(E_NAME)) {

	    String currentName = getAttributeValueOf(A_NAME,atts);
	    if (currentTestCase == null) {
		// We don't process a test case tag ...
		currentTestSuite.setName(currentName);
	    } else {
		// The name belongs to the currently opened testcase ...
		currentTestCase.setName(currentName);
	    }

	} else if (name.equals(E_DESCRIPTION)) {
	    currentText = "";

	} else if (name.equals(E_XMI_MODEL)) {
	    String currentFileName = getAttributeValueOf(A_FILE,atts);
	    currentTestSuite.setXMIModelFile(currentFileName);

	} else if (name.equals(E_IMPLEMENTATION_SOURCES)) {
	    String currentDirName = getAttributeValueOf(A_DIRECTORY,atts);
	    currentTestSuite.setImplementationSourcesDirectory(currentDirName); 

	} else if (name.equals(E_TESTCASE)) {

	    currentTestCase = new TransformationTestCase(currentTestSuite);

	} else if (name.equals(E_CONSTRAINT)) {
	    currentText = "";
	    
	} else if (name.equals(E_EXPRESSION)) {
	    currentText = "";
	    currentExprType = getAttributeValueOf(A_TYPE,atts);
	    currentExprContextElement = getAttributeValueOf(A_CONTEXTELEMENT,atts);
	    
	    
	} else if (name.equals(E_OPTION)) {
	    String optName = getAttributeValueOf(A_NAME,atts);
	    String optValue = getAttributeValueOf(A_VALUE,atts);
	    currentTestCase.setOption(optName, optValue);
	}
	
		  
    }


     public void processingInstruction(java.lang.String target, java.lang.String data) throws org.xml.sax.SAXException  {
	// Receive notification of a processing instruction.
	// do optional output ...
	if(_outputEnabled){
	    _outputWriter.println("-- Got a PROCESSING INSTRUCTION :");
	    _outputWriter.println("TARGET = " + target);
	    _outputWriter.println("DATA = " + data); 
	    _outputWriter.flush();
	}
	
    }
  
    public void setDocumentLocator(Locator locator)  {
	// Receive an object for locating the origin of SAX document events.
	// do optional output ...
	if(_outputEnabled){ 
	    _outputWriter.println("-- SET DOCUMENT LOCATOR TO : " + locator);
	    _outputWriter.flush();
	}
	
    }



    // =============================================================================================
    // ============================= Some helper methods ===========================================
    // =============================================================================================

    /**
     * extract an InputSource out of a string. Maybe deleted in the future
     */
    public InputSource getFileInputSourceFor(String fileName) {
  	File file = new File(fileName);
	String path = file.getAbsolutePath();
	String fSep = System.getProperty("file.separator");
	if (fSep != null && fSep.length() == 1)
	    path = path.replace(fSep.charAt(0), '/');
	if (path.length() > 0 && path.charAt(0) != '/')
	    path = '/' + path;
	try {
  	    URL url = new URL("file", "", path);
	    return new InputSource(url.toString());
	}
	catch (java.net.MalformedURLException e) {
	    // cannot happen
	}
	catch (java.io.IOException e) {
	    // cannot happen
	}
	return null;
	}


    public String getAttributeValueOf(java.lang.String name, AttributeList atts) { 
	String result = null;;
	for (int i = 0; i < atts.getLength(); i++) {
	    if (name.equals(atts.getName(i))){
		result = atts.getValue(i);
		break;
	    }
	   
	} // for
	
	return result;
	
    }




    public static void main(String[] args){
	if (args.length == 1){
	    String inputFileName = args[0];
	    TestSuiteReader aReader = new TestSuiteReader();
	    aReader.enableOutputLogging(System.err);
	    aReader.readTestSuite(inputFileName);
	} else {
	    System.err.println("Wrong syntax ! Use 'TestSuiteReader  fileName'");
	}
    }
    


}

