// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.parser;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.*;
import junit.framework.*;
import java.io.*;
import de.uka.ilkd.key.logic.op.TermSymbol;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.java.*;

/** class tests the parser for Taclets
*/


public class TestTacletParserOCL extends TestCase {


    public NamespaceSet nss;
    public Services services;
    public Choice defaultChoice;

    TermFactory tf=TermFactory.DEFAULT;

    Sort ocl_sort;
 
    Function forAll, iterate, and, true_, nil;

    SortedSchemaVariable x, y, acc, coll, expr, init, e1, e2;

    Term nilTerm, trueTerm;

    public TestTacletParserOCL(String name) {
	super(name);
    }

    //
    //  set up
    //

    public void setUp() {
	nss =new NamespaceSet();
	services = TacletForTests.services();
	defaultChoice = new Choice(new Name("Default"));
	nss.setFunctions(defaultChoice.funcNS());

	parseDecls("sorts { ocl OclAny; }\n" +
		   "schema variables {\n" +
		   "   OclAny variables x, y, acc;\n" +
		   "   OclAny coll, expr, init, e1, e2;\n" +
		   "}\n" +
		   "functions { \n" +
		   "   OclAny $iterate(OclAny, OclAny, OclAny);\n" +
		   "   OclAny $true;\n" +
		   "   OclAny $false;\n" +
		   "   OclAny $cons(OclAny, OclAny);\n" +
		   "   OclAny $nil;\n" +
		   "   OclAny $and(OclAny, OclAny);\n" +
		   "   OclAny $forAll(OclAny, OclAny);\n" +
		   "}\n" +
		   "predicates {\n" +
		   "   OclWrapper(OclAny);\n" +
		   "}\n"
		   );

	ocl_sort = lookup_sort("OclAny");

	x = (SortedSchemaVariable)lookup_var("x");
	y = (SortedSchemaVariable)lookup_var("y");
	acc = (SortedSchemaVariable)lookup_var("acc");
	coll = (SortedSchemaVariable)lookup_var("coll");
	expr = (SortedSchemaVariable)lookup_var("expr");
	init = (SortedSchemaVariable)lookup_var("init");
	e1 = (SortedSchemaVariable)lookup_var("e1");
	e2 = (SortedSchemaVariable)lookup_var("e2");

	forAll = lookup_func("$forAll");
	iterate = lookup_func("$iterate");
	and = lookup_func("$and");
	true_ = lookup_func("$true");
	nil = lookup_func("$nil");

	nilTerm = tf.createFunctionTerm(nil);
	trueTerm = tf.createFunctionTerm(true_);
    }

    //
    // Utility methods for setUp:
    //
    
    TermSymbol lookup_var(String name) {
	return (TermSymbol)nss.variables().lookup(new Name(name));
    }

    Sort lookup_sort(String name) {
	return (Sort)nss.sorts().lookup(new Name(name));
    }

    Function lookup_func(String name) {
	return (Function)defaultChoice.funcNS().lookup(new Name(name));
    }

    private DeclParser stringDeclParser(String s) {
	return new DeclParser(new KeYLexer(new StringReader(s)),
			      "No file. parser/TestTacletParser.stringDeclParser("+s+")",  
			       services, nss, defaultChoice);
    }

    public void parseDecls(String s) {
	try {
	    DeclParser p = stringDeclParser(s);
	    p.decls();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }

    //
    // Utility Methods for test cases.
    //

    private TacletParser stringTacletParser(String s) {
	return new TacletParser
	    (new KeYLexer(new StringReader(s)), 
	     "No file. parser/TestTacletParser.stringTacletParser("+s+")",  
	     tf, services, nss, defaultChoice);
    }

    public Term parseTerm(String s) {
	try {
	    TacletParser p = stringTacletParser(s);
	    return p.term();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }


    public Term parseFma(String s) {
	try {
	    TacletParser p = stringTacletParser(s);
	    
	    return p.formula();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }

    Taclet parseTaclet(String s) {
	try {
	    TacletParser p = stringTacletParser(s);
	    
	    return p.taclet(SetAsListOfChoice.EMPTY_SET);
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }

    //
    // Test cases.
    //

    //Tests OCL operations where one variable gets bound (forAll, exists, select, ...)
    //Only for the problem clause of the .key file, where bound vars are treated as
    //logical vars (not schema vars)
    public void test1() {
	Term tParse = parseTerm("$forAll($nil, \\s:OclAny/ $and($true, s))");
	LogicVariable this_s 
	    = (LogicVariable)tParse.varsBoundHere(1).getQuantifiableVariable(0);
	Term tS = tf.createVariableTerm(this_s);
	Term tAnd = tf.createFunctionTerm(and, trueTerm, tS);
	Term term = tf.createBinaryBindingTerm(forAll, this_s, new Term[]{nilTerm, tAnd});
	assertEquals("parse $forAll($nil, \\s:OclAny/ $and($true, s))",
		     term,
		     tParse);
    }

    //Tests the OCL operation where two variables get bound -- iterate
    //Only for the problem clause of the .key file, where bound vars are treated as
    //logical vars (not schema vars)
    public void test2() {
	Term tParse = parseTerm("$iterate($nil," +
				"$true, \\s:OclAny\\acc:OclAny/ $and(s, acc))");
	LogicVariable this_s 
	    = (LogicVariable)tParse.varsBoundHere(2).getQuantifiableVariable(0);
	LogicVariable this_acc 
	    = (LogicVariable)tParse.varsBoundHere(2).getQuantifiableVariable(1);
	Term tS = tf.createVariableTerm(this_s);
	Term tAcc = tf.createVariableTerm(this_acc);
	Term tAnd = tf.createFunctionTerm(and, tS, tAcc);
	Term term = tf.createTernaryBindingTerm(iterate, this_s, this_acc, 
					       new Term[]{nilTerm, trueTerm, tAnd});
	assertEquals("parse $iterate($nil," +
			             "$true, \\s:OclAny\\acc:OclAny/ $and(s, acc))",
		     term,
		     tParse);
    }

    //Tests OCL operations where one variable gets bound (forAll, exists, select, ...)
    //Only for the rules clause of the .key file, where bound vars are treated as
    //schema vars (not logical vars)
    public void test3() {
	Term tExpr = tf.createVariableTerm(expr);
	Term tColl = tf.createVariableTerm(coll);
	Term term = tf.createBinaryBindingTerm(forAll, x, new Term[]{tColl, tExpr});
	assertEquals("parse $forAll(coll, \\x/ expr)",
		     term,
		     parseTerm("$forAll(coll, \\x/ expr)"));
    }

    //Tests the OCL operation where two variables get bound -- iterate
    //Only for the rules clause of the .key file, where bound vars are treated as
    //schema vars (not logical vars)
    public void test4() {
	Term tExpr = tf.createVariableTerm(expr);
	Term tColl = tf.createVariableTerm(coll);
	Term tInit = tf.createVariableTerm(init);
	Term term = tf.createTernaryBindingTerm(iterate, x, acc, 
						new Term[]{tColl, tInit, tExpr});
	assertEquals("parse $iterate(coll, init, \\x\\acc/ expr)",
		     term,
		     parseTerm("$iterate(coll, init, \\x\\acc/ expr)"));
    }

    //Tests the "old" syntax of the substitution operator (one substitution)
    public void test5() {
	Term tExpr = tf.createVariableTerm(expr);
	Term tE1 = tf.createVariableTerm(e1);
	Term term = tf.createSubstitutionTerm(Op.SUBST, x, new Term[]{tE1, tExpr});
	assertEquals("parse {x e1}expr",
		     term,
		     parseTerm("{x e1}expr"));
    }

    //Tests the syntax for substitution used for OCL taclets (one substitution)
    public void test6() {
	Term tExpr = tf.createVariableTerm(expr);
	Term tE1 = tf.createVariableTerm(e1);
	Term term = tf.createSubstitutionTerm(Op.SUBST, x, new Term[]{tE1, tExpr});
	assertEquals("parse {x/e1}expr",
		     term,
		     parseTerm("{x/e1}expr"));
    }
  
    //Tests the syntax for substitution used for OCL taclets (two substitutions)
    public void test7() {
	Term tExpr = tf.createVariableTerm(expr);
	Term tE1 = tf.createVariableTerm(e1);
	Term tE2 = tf.createVariableTerm(e2);
	Term term = tf.createTernarySubstTerm(Op.SUBST, x, y, new Term[]{tE1, tE2, tExpr});
	assertEquals("parse {x/e1, y/e2}expr",
		     term,
		     parseTerm("{x/e1, y/e2}expr"));
    }
}
