// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.parser;

import junit.framework.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.java.*;
import java.io.*;
import de.uka.ilkd.key.gui.AbbrevMap;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.op.LogicVariable;
import de.uka.ilkd.key.rule.TacletForTests;
import de.uka.ilkd.key.rule.SetAsListOfTaclet;


public class TestTermParserOCL extends TestCase {
    
    private TermFactory tf = TermFactory.DEFAULT;

    NamespaceSet nss;

    Choice defaultChoice;

    Sort ocl_sort;
 
    Function iterate, true_, false_, cons, nil, and, forAll, oclWrapper;

    LogicVariable s, i;

    Term nilTerm, trueTerm, falseTerm;

    public TestTermParserOCL(String name) {
	super(name);
    }


    public void setUp() {
	defaultChoice = new Choice(new Name("Default"));
	nss = new NamespaceSet();

	parseDecls("sorts { ocl OclAny; }\n" +
		   "functions { \n" +
		   "   OclAny $iterate(OclAny, OclAny, OclAny);\n" +
		   "   OclAny $true;\n" +
		   "   OclAny $false;\n" +
		   "   OclAny $cons(OclAny, OclAny);\n" +
		   "   OclAny $nil;\n" +
		   "   OclAny $and(OclAny, OclAny);\n" +
		   "   OclAny $forAll(OclAny, OclAny);\n" +
		   "}\n" +
		   "predicates {\n" +
		   "   OclWrapper(OclAny);\n" +
		   "}\n"
		   );

	ocl_sort = lookup_sort("OclAny");

	forAll = lookup_func("$forAll");
	and = lookup_func("$and");
	cons = lookup_func("$cons");
	nil = lookup_func("$nil");
	iterate = lookup_func("$iterate");
	true_ = lookup_func("$true");
	false_ = lookup_func("$false");
	oclWrapper = lookup_func("OclWrapper");

	nilTerm = tf.createFunctionTerm(nil);
	trueTerm = tf.createFunctionTerm(true_);
	falseTerm = tf.createFunctionTerm(false_);
    }

    Sort lookup_sort(String name) {
	return (Sort)nss.sorts().lookup(new Name(name));
    }
    
    Function lookup_func(String name) {
	return (Function)defaultChoice.funcNS().lookup(new Name(name));
    }

    LogicVariable declareVar(String name,Sort sort) {
	LogicVariable v = new LogicVariable(new Name(name),sort);
	nss.variables().add(v);
	return v;
    }
    

    private DeclParser stringDeclParser(String s) {
	return new DeclParser(new KeYLexer(new StringReader(s)),
			      "No file. Call of parser from parser/TestTermParser.java",
			      new Services(), nss, defaultChoice);
    }

    public void parseDecls(String s) {
	try {
	    DeclParser p = stringDeclParser(s);
	    p.decls();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }

    public Term parseProblem(String s) {
	try {
	    ProblemParser p = new ProblemParser
		(new KeYLexer(new StringReader(s)),
		 "No file. Call of parser from parser/TestTermParser.java",
		 new ParserConfig(new Services(), nss, defaultChoice),
		 new ParserConfig(new Services(), nss, defaultChoice),
		 null, SetAsListOfTaclet.EMPTY_SET);
	    return p.problem();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}	
    }

    private TermParser stringTermParser(String s) {
	return new TermParser(new KeYLexer(new StringReader(s)), 
			      "No file. Call of parser from parser/TestTermParser.java",
			      tf, 
			      new Recoder2KeY(TacletForTests.services(), nss),
			      TacletForTests.services(), nss, 
                  new AbbrevMap());
    }

    public Term parseTerm(String s) {
	try {
	    TermParser p = stringTermParser(s);
	    return p.term();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }


    public Term parseFma(String s) {
	try {
	    TermParser p = stringTermParser(s);
	    
	    return p.formula();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }

    //Tests OCL operators/operations without any bound variables
    public void test1() {
	Term term = tf.createFunctionTerm(and, trueTerm, falseTerm);
	assertEquals("parse $and($true, $false)",
		     term,
		     parseTerm("$and($true, $false)"));
    }

    //Tests OCL operations where one variable gets bound (forAll, exists, select, ...)
    //Only for the problem clause of the .key file, where bound vars are treated as
    //logical vars (not schema vars)
    public void test2() {
	Term tParse = parseTerm("$forAll($cons($true, $cons($false, $nil))," +
				"\\s:OclAny/ $and($true, s))");
	LogicVariable this_s 
	    = (LogicVariable)tParse.varsBoundHere(1).getQuantifiableVariable(0);
	Term tS = tf.createVariableTerm(this_s);
	Term t1Cons = tf.createFunctionTerm(cons, new Term[]{falseTerm, nilTerm});
	Term tCons = tf.createFunctionTerm(cons, new Term[]{trueTerm, t1Cons});
	Term tAnd = tf.createFunctionTerm(and, trueTerm, tS);
	Term term = tf.createBinaryBindingTerm(forAll, this_s, new Term[]{tCons, tAnd});
	assertEquals("parse $forAll($cons($true, $cons($false, $nil))," +
				"\\s:OclAny/ $and($true, s))",
		     term,
		     tParse);
    }

    //Tests the OCL operation where two variables get bound -- iterate
    //Only for the problem clause of the .key file, where bound vars are treated as
    //logical vars (not schema vars)
    public void test3() {
	Term tParse = parseTerm("$iterate($cons($true, $cons($false, $nil))," +
				"$false, \\s:OclAny\\acc:OclAny/ $and(s, acc))");
	LogicVariable this_s 
	    = (LogicVariable)tParse.varsBoundHere(2).getQuantifiableVariable(0);
	LogicVariable this_acc 
	    = (LogicVariable)tParse.varsBoundHere(2).getQuantifiableVariable(1);
	Term tS = tf.createVariableTerm(this_s);
	Term tAcc = tf.createVariableTerm(this_acc);
	Term t1Cons = tf.createFunctionTerm(cons, new Term[]{falseTerm, nilTerm});
	Term tCons = tf.createFunctionTerm(cons, new Term[]{trueTerm, t1Cons});
	Term tAnd = tf.createFunctionTerm(and, tS, tAcc);
	Term term = tf.createTernaryBindingTerm(iterate, this_s, this_acc, 
					       new Term[]{tCons, falseTerm, tAnd});
	assertEquals("parse $iterate($cons($true, $cons($false, $nil))," +
			             "$false, \\s:OclAny\\acc:OclAny/ $and(s, acc))",
		     term,
		     tParse);
    }
}
