// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof;

import javax.swing.table.AbstractTableModel;
import javax.swing.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.gui.AbbrevMap;
import java.io.*;
import java.util.Vector;
import de.uka.ilkd.key.rule.Taclet;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.rule.SetOfTacletApp;
import de.uka.ilkd.key.rule.IteratorOfTacletApp;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.rule.IfFormulaInstantiation;
import de.uka.ilkd.key.rule.ListOfIfFormulaInstantiation;
import de.uka.ilkd.key.rule.SLListOfIfFormulaInstantiation;
import de.uka.ilkd.key.rule.IfFormulaInstSeq;
import de.uka.ilkd.key.rule.MatchConditions;

public class ApplyTacletDialogModel {
	
    /** the model used for the instantiation of the if-sequent */
    private IfChoiceModel[] ifChoiceModel;
    private static final IfChoiceModel[] EMPTY_IF_CHOICES = new IfChoiceModel [ 0 ];
    /** the model used for the instantiation of the schemavariables */
    private TacletInstantiationsTableModel tableModel;

    /** 
     * the application of the Taclet of which this model is used to
     * complete the instantiations of the schemavariables and/or
     * if-sequent
     */
    private TacletApp app;
    /** the sequent the application above is applied */
    private Sequent seq;

    /** listeners of this model */
    private Vector listeners = new Vector();
    /** the change event that is sent */
    private final ModelEvent changeEvent = new ModelEvent(this);

    /** namespace of variables */
    private NamespaceSet nss;
    /** the java service class */
    private Services services ;
    /** user constraint */
    private Constraint userConstraint;
    /** abbrevmap */
    private AbbrevMap scm;
    
    /**
     * create new data model for the apply Taclet dialog wrapping a combo box
     * model and a table model
     */
    public ApplyTacletDialogModel(TacletApp app, Sequent seq, 
				  Services services,
				  Constraint userConstraint,
				  NamespaceSet nss, AbbrevMap scm) {
	this.app = app;
	this.seq = seq;
	this.services = services;
	this.userConstraint = userConstraint;
	this.nss = nss;
	this.scm = scm;
	
	initIfChoiceModels();	

	tableModel = new TacletInstantiationsTableModel(app, services, nss, scm);
    }

    public void addModelChangeListener(ModelChangeListener l) {
	listeners.add(l);
    }

    public void removeModelChangeListener(ModelChangeListener l) {
	listeners.remove(l);
    }
	
    public IfChoiceModel ifChoiceModel(int i) {
	return ifChoiceModel[i];
    }

    public int ifChoiceModelCount() {
	return ifChoiceModel.length;
    }

    public TacletInstantiationsTableModel tableModel() {
	return tableModel;
    }

    public TacletApp application() {
	return app;
    }

    public Taclet taclet() {
	return app.taclet();
    }

    public TacletApp tacletApp() {
	return app;
    }

    public Term ifFma(int i) {
	return ifChoiceModel(i).ifFma();
    }

    private void initIfChoiceModels() {
 	Sequent ifseq   = taclet().ifSequent();
	int     asize   = ifseq.antecedent().size();
	int     size    = asize + ifseq.succedent().size();

	if ( size > 0 ) {
	    ListOfIfFormulaInstantiation antecCand =
		IfFormulaInstSeq.createList ( seq, true );
	    ListOfIfFormulaInstantiation succCand  =
		IfFormulaInstSeq.createList ( seq, false );

	    IteratorOfConstrainedFormula it        = ifseq.iterator();
	    Term                         ifFma;
	    MatchConditions              matchCond = app.matchConditions ();

	    ifChoiceModel                          = new IfChoiceModel[size];

	    for (int i=0; i<size; i++) {
		ifFma            = it.next ().formula ();
		ifChoiceModel[i] = 
		    new IfChoiceModel ( ifFma,
					taclet ().matchIf ( ( i < asize ?
							      antecCand.iterator () :
							      succCand .iterator () ),
							    ifFma,
							    matchCond,
							    services,
							    userConstraint ).getFormulas (),
					services, nss, scm );
	    }
	} else
	    ifChoiceModel = EMPTY_IF_CHOICES;
    }

    /** creates the data model to be displayed in the dataTable */
    private TacletInstantiationsTableModel buildDataModel
	(SetOfTacletApp set) {
	de.uka.ilkd.key.util.Debug.assertTrue(set.size()<=1, 
					      "Too many instantiations");
	IteratorOfTacletApp it = set.iterator();
	return new 
	    TacletInstantiationsTableModel(it.next(), services, nss, scm);
    }
    

    private TacletApp createTacletAppFromIfs(TacletApp app) 
	throws IfMismatchException, 
	       SVInstantiationParserException,
	       MissingInstantiationException,
	       SortMismatchException {

	ListOfIfFormulaInstantiation instList = SLListOfIfFormulaInstantiation.EMPTY_LIST;

	int i = ifChoiceModel.length;
	while ( i-- != 0 ) {	   
	    instList = instList.prepend ( ifChoiceModel[i].getSelection (i) );
	}

	try {
	    app = app.setIfFormulaInstantiations
		( instList, services, userConstraint );
	} catch ( SortException e ) {
	    throw new SortMismatchException ( "'if'-sequent", null, 0, 0 );
	}

	if ( app == null ) {
 	    throw new IfMismatchException
		("Mismatch of 'if'-formulas.\n"+
		 "Reasons may be: ambigous instantiation "+
		 "of schemavariables or unsatisfiable constraints.");
	}

	return app;
    }

    public String getStatusString() {
	TacletApp rapp = app;
	try {	    
	    rapp = createTacletApp();
	} catch (SVInstantiationException ime) {
	    return "Rule is not applicable.\n Detail:" + 
		ime.getMessage();
	}   
	if (rapp == null) {
	    return "Rule is not applicable."; 
	}
	if ( rapp.sufficientlyComplete () )
	    return "Instantiation is OK.";
	else
	    return "Instantiations are incomplete."; 
	
    }

    public TacletApp createTacletApp() throws SVInstantiationException {
	return createTacletAppFromIfs(tableModel.createTacletAppFromVarInsts());
    }
    
    
    private void informListenerAboutModelChange() {
	for (int i = 0; i < listeners.size(); i++) {
	    if (listeners.get(i) != null) {
		((ModelChangeListener)listeners.get(i)).modelChanged(changeEvent);
	    }
	}
    }

    /** sets the manual if-input */
    public void setManualInput(int i, String s) {
	ifChoiceModel(i).setInput(s);
	informListenerAboutModelChange();
    }

    /**
     * replaces the TacletApp of this ApplyTacletDialogModel by an TacletApp
     * where all name conflicts are resolved and thus the parser is enabled
     * to accept variables from the context or the prefix of the Taclet.
     *
     */
    public void prepareUnmatchedInstantiation() {
	app = app.prepareUserInstantiation();	
    }

    public Namespace var_ns() {
	return nss.variables();
    }

}
