// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof;

import java.io.*;
import de.uka.ilkd.key.gui.AbbrevMap;
import de.uka.ilkd.key.gui.PresentationFeatures;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.proof.init.*;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.gui.Main;
import de.uka.ilkd.key.gui.KeYMediator;
import de.uka.ilkd.key.gui.SwingWorker;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.util.Debug;
import javax.swing.JOptionPane;
import java.util.*;

public class ProblemLoader implements Runnable {

   File file;
   Main main;
   KeYMediator mediator;
   
   Proof proof = null;
   IteratorOfNode children = null;

   Node currNode = null;
   Goal currGoal = null;
   String currTacletName = null;
   int currFormula = 0;
   PosInTerm currPosInTerm = PosInTerm.TOP_LEVEL;
   Stack stack = new Stack();
   LinkedList loadedInsts = null;
   ListOfIfFormulaInstantiation ifSeqFormulaList = 
      SLListOfIfFormulaInstantiation.EMPTY_LIST;

   private SwingWorker worker;

   public ProblemLoader(File file, Main main) {
      this.main = main;
      mediator  = main.mediator();
      this.file = file;
   }

   public void run() {
        /* Invoking start() on the SwingWorker causes a new Thread
         * to be created that will call construct(), and then
         * finished().  Note that finished() is called even if
         * the worker is interrupted because we catch the
         * InterruptedException in doWork().
         */
        worker = new SwingWorker() {
            public Object construct() {
                Object res = doWork();
		return res;
            }
            public void finished() {
                mediator.startInterface(true);
	        Main main = Main.getInstance();
                String msg = (String) get();
                if(!"".equals(msg)) {
                    if(Main.batchMode){
                    System.exit(-1);
                    } else {
                    JOptionPane.showMessageDialog(
                    main, msg, "Load Error",
                    JOptionPane.ERROR_MESSAGE);
                    }
                } else {
	           PresentationFeatures.
                       initialize(init.initConfig().
                                  createNamespacesForActivatedChoices().
				  functions(), 
			          mediator.getNotationInfo(),
			          mediator.getSelectedProof());
                }
                if (Main.batchMode) {
                    //System.out.println("Proof: " +proof.openGoals());
		    if(proof.openGoals().size()==0) {
		    	System.out.println("proof.openGoals.size="+proof.openGoals().size());
		        System.exit(0);
		    }
		        mediator.startAutoMode();
                }
            }
        };
        mediator.stopInterface(true);
        worker.start();
   }


   ProblemInitializer init;


   protected Object doWork() {
      KeYUserProblemFile problemFile = 
         new KeYUserProblemFile(file.getName(), file, main);
      String status = "";
      try{
	  init = new ProblemInitializer(main.proofSettings(), main);
          main.setStatusLine("Loading problem");
	  init.setUpProof(problemFile);
	  proof = mediator.getSelectedProof();
          main.setStatusLine("Loading proof");
          currNode = proof.root(); // initialize loader
          children = currNode.childrenIterator(); // --"--
          init.tryReadProof(this);
          main.setStandardStatusLine();
      } catch (Throwable e) {
          main.setStatusLine("Failed to load problem/proof");
	  status =  e.toString();
          e.printStackTrace();
      }
      finally {
	  if (problemFile!=null) problemFile.close();
          return status;
      }
   }




// note: Expressions without parameters only emit the endExpr signal
   public void beginExpr(char id, String s) {
//System.out.println("start "+id+"="+s);
      switch (id) {
         case 'b' : 
            stack.push(children);
            if (children.hasNext()) currNode = children.next();
            break;

         case 'r' : 
            if (currNode == null) currNode = children.next();
            // otherwise we already fetched the node at branch point
            currGoal      = proof.getGoal(currNode);
//System.err.println((currNode == null)+"<-N G->"+  (currGoal == null));

//if (currGoal == null) System.err.println(currNode);
            Debug.assertTrue(currGoal != null, 
			     main.proofSettings()
			     .getSimultaneousUpdateSimplifierSettings()
			     .implicitExecution()?
	       "Selected node not a goal in proof!\nAlthough the \"implicit "+
	       "execution\" option of the update simplifier is on!"
	      : "Selected node not a goal in proof!\n" +
                "Note that you need to switch on \"implicit execution\"\n" +
		"in the dialog of the Options/Update Simplifier menu.\n" +
		"Otherwise, you cannot load proofs.");

            currTacletName= s;
            // set default state
            currFormula   = 0;
            currPosInTerm = PosInTerm.TOP_LEVEL; 
            loadedInsts   = null;
            ifSeqFormulaList = SLListOfIfFormulaInstantiation.EMPTY_LIST;
            break;

         case 'f' :
            currFormula   = Integer.parseInt(s);
            break;

         case 't' :
            currPosInTerm = PosInTerm.parseReverseString(s);
            break;
            
         case 'i' :
            if (loadedInsts == null) loadedInsts = new LinkedList();
            loadedInsts.add(s);
            break;
            
         case 'h' :
//             Debug.fail("Detected use of heuristics!");
             break;
            
         case 'q' : // ifseqformula
//            Namespace varNS = mediator.var_ns();
//            varNS = currGoal.getVariableNamespace(varNS);
//            Term f = parseTerm(s, varNS);
            Sequent seq = currGoal.sequent();
            ifSeqFormulaList = ifSeqFormulaList.append(
               new IfFormulaInstSeq(seq, Integer.parseInt(s)));
            break;
         case 'u' : //UserLog
            if(proof.userLog==null)
                proof.userLog = new Vector();
            proof.userLog.add(s);
            break;
         case 'v' : //Version log
            if(proof.keyVersionLog==null)
                proof.keyVersionLog = new Vector();
            proof.keyVersionLog.add(s);
            break;
         case 's' : //ProofSettings
            //System.out.println("------------------------\n" + s + "------------\n");
            main.proofSettings().loadSettingsFromString(s);
            break;
      }
   }


   public void endExpr(char id) {
//System.out.println("end "+id);
      switch (id) {
         case 'b' : 
            children = (IteratorOfNode) stack.pop();
            break;
         case 'a' :
             // do nothing; attribute interactive was modified here
             break;
         case 'r' : 
            try{
               mediator.apply(constructApp(), currGoal);
               children = currNode.childrenIterator();
               currNode = null;
            } catch(AppConstructionException e) {
//               System.err.println(e.getMessage());
               throw new RuntimeException(e.getMessage());
            }
            break;
      }
   }
   
   
   private TacletApp constructApp() throws AppConstructionException {

      TacletApp ourApp = null;

      PosInOccurrence pos = null;
      if (currFormula != 0) // otherwise we have no pos
         pos = PosInOccurrence.findInSequent(currGoal.sequent(),
                                             currFormula,
                                             currPosInTerm);

//System.err.print("Want to apply "+currTacletName+" at "+currGoal);
      SetOfTacletApp applics;
      try {
         applics=mediator.getTacletApplications(currGoal, currTacletName, pos);
      } catch(ArrayIndexOutOfBoundsException e) { // ???
         System.err.println(currTacletName+":"+pos);
         e.printStackTrace();
         throw new AppConstructionException(e.toString());
      }
      if (applics.size() != 1) {
         throw new AppConstructionException(currTacletName+": found "+
            applics.size()+" applications. Don't know what to do!\n"+
            "@ "+pos);
      }
      ourApp = applics.iterator().next();
      ourApp = constructInsts(ourApp);

      ourApp = ourApp.setIfFormulaInstantiations(ifSeqFormulaList,
         mediator.getServices(), mediator.getUserConstraint().getConstraint ());

      ourApp = ourApp.createDependingFunctions ( mediator.func_ns () );

      return ourApp;
   }
   
   
   private TacletApp constructInsts(TacletApp app) {
      if (loadedInsts == null) return app;
      SetOfSchemaVariable uninsts = app.uninstantiatedVars();

      // first pass: add variables
      Iterator it = loadedInsts.iterator();
      while (it.hasNext()) {
         String s = (String) it.next();
         int eq = s.indexOf('=');
         String varname = s.substring(0, eq);
         String value = s.substring(eq+1, s.length());
         
         SchemaVariable sv = lookupName(uninsts, varname);
         
         if (sv.isVariableSV()) {
	    LogicVariable lv=new LogicVariable(new Name(value),
					       app.getRealSort(sv));
	    Term instance = TermFactory.DEFAULT.createVariableTerm(lv);
	    app = app.addCheckedInstantiation(sv, instance, true);
         }
      }

      // second pass: add everything else
      it = loadedInsts.iterator();
      while (it.hasNext()) {
         String s = (String) it.next();
         int eq = s.indexOf('=');
         String varname = s.substring(0, eq);
         String value = s.substring(eq+1, s.length());
         
         SchemaVariable sv = lookupName(uninsts, varname);
         
         if (sv.isVariableSV()) {
         // ignore -- already done
	 } else if (sv.isProgramSV()) {
	    ProgramElement pe = 
               TacletInstantiationsTableModel.getProgramElement(
                  app, value, sv, mediator.getServices());
	    app = app.addInstantiation(sv, pe, true);
	 } else if ( sv.isDependingSV() ) {
	     app = app.createDependingConstant ( value,
						 sv,
						 true );
	 } else {
            Namespace varNS = mediator.var_ns();
            varNS = currGoal.getVariableNamespace(varNS);
            varNS = app.extendVarNamespaceForSV(varNS, sv);
            Term instance = parseTerm(value, proof, varNS);
            app = app.addCheckedInstantiation(sv,instance, true); 
         }
      }
      return app;
   }
   
   
   private static Term parseTerm(String value, Proof proof, Namespace varNS) {
         try {
            return TermParserFactory.createInstance().
		parse(new StringReader(value), null,
		      proof.getServices(), 
		      varNS, 
		      proof.getNamespaces().functions(), 
		      proof.getNamespaces().sorts(),
		      new AbbrevMap());
         } catch(ParserException e) {
              throw new RuntimeException("While parsing "+value+"\n"+
                                         varNS+"\n"+e);         
         }
   }
   
   public static Term parseTerm(String value, Proof proof) {
      return parseTerm(value, proof, proof.getNamespaces().variables());
   }
   
   
   private SchemaVariable lookupName(SetOfSchemaVariable set, String name) {
      IteratorOfSchemaVariable it = set.iterator();
      while (it.hasNext()) {
         SchemaVariable v = it.next();
         if (v.name().toString().equals(name)) return v;
      }
      return null; // handle this better!
   }
   
   
   private class AppConstructionException extends Exception {

       AppConstructionException(String s) {
          super(s);
       }
   }
   
}
