// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.logic.op.SimultaneousUpdateOp;
import de.uka.ilkd.key.logic.op.Metavariable;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.gui.ConstraintSequentPrintFilter;

/**
 * Class whose objects represent an index of taclet apps for one particular
 * position within a formula, and that also contain references to the indices of
 * direct subformulas
 */
public class TermTacletAppIndex {
    /** the term for which NoPosTacletApps are kept in this index node */
    private final Term term;
    /** NoPosTacletApps for this term */
    private final ListOfNoPosTacletApp localTacletApps;
    /** indices for subterms */
    private final ListOfTermTacletAppIndex subtermIndices;
    /** the constraint that is used to extend the considered term below
     *  metavariables */
    private final Constraint displayConstraint;
    /** */
    private final RuleFilter ruleFilter;

    /**
     * Create a TermTacletAppIndex
     */
    private TermTacletAppIndex( Term term,
				ListOfNoPosTacletApp localTacletApps,
				ListOfTermTacletAppIndex subtermIndices,
				Constraint displayConstraint,
				RuleFilter ruleFilter ) {
	this.term              = term;
	this.subtermIndices    = subtermIndices;
	this.localTacletApps   = localTacletApps;
	this.displayConstraint = displayConstraint;
	this.ruleFilter        = ruleFilter;
    }


    private TermTacletAppIndex getSubIndex ( int subterm ) {
	return subtermIndices.take ( subterm ).head ();
    }


    /** 
     * collects all RewriteTacletInstantiations for the given
     * heuristics in a subterm of the constrainedFormula described by a
     * PosInOccurrence
     * @param pos the PosInOccurrence to focus 
     * @param services the Services object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    public static ListOfNoPosTacletApp
	getRewriteTaclet(PosInOccurrence pos,
			 RuleFilter      filter,
			 Services        services,
			 Constraint      userConstraint,
			 TacletIndex     tacletIndex) { 

	ListOfNoPosTacletApp result = SLListOfNoPosTacletApp.EMPTY_LIST;

	Constraint c = pos.constrainedFormula().constraint();

	if ( pos.termBelowMetavariable () != null ) {
	    c = c.unify ( pos.constrainedFormula ().formula ().subAt ( pos.posInTerm () ),
			  pos.termBelowMetavariable () );
	    if ( !c.isSatisfiable () )
		return SLListOfNoPosTacletApp.EMPTY_LIST;
	}
	
	// check location conditions
	// rewrite taclets, whose find match on a location are only applicable if
	// and only if they have no rewrite part

	IteratorOfNoPosTacletApp rewriteIterator = 
	    tacletIndex.getRewriteTaclet(pos, c,
					 filter,
					 services, 
					 userConstraint).iterator();


	while (rewriteIterator.hasNext()) {
	    NoPosTacletApp tacletApp = rewriteIterator.next();
	    if (!(tacletApp.taclet().hasReplaceWith() && 		
		  isLocation(pos))) {
		result=result.prepend(tacletApp);
	    } 
	}
	return result;
    }


    /**
     * looks if the enclosing term is a simultaneous update and if the current
     * position describes a location
     * @param pos the PosInOccurrence describing the potential
     * location
     * @return true if the described position is a location on the left side of
     * an assignment pair of an update
     */
    private static boolean isLocation(PosInOccurrence pos) {
	PIOPathIterator it     = pos.iterator();
	Operator        op     = null;
	int             subPos = -1;

	it.next ();
	while ( it.hasNext () ) {
	    subPos = it.getChild   ();
	    op     = it.getSubTerm ().op ();
	    it.next ();
	}

	return (op != null && 
		op instanceof SimultaneousUpdateOp && 
		((SimultaneousUpdateOp)op).isLocation(subPos));
    }

    /** 
     * collects all FindTaclets with instantiations for the given
     * heuristics and position
     * @param pos the PosInOccurrence to focus
     * @param services the Services object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    public static ListOfNoPosTacletApp
	getFindTaclet(PosInOccurrence pos,
		      RuleFilter      filter,
		      Services        services,
		      Constraint      userConstraint,
		      TacletIndex     tacletIndex) { 
	ListOfNoPosTacletApp
	    tacletInsts=SLListOfNoPosTacletApp.EMPTY_LIST;
	if (pos.isTopLevel()) {
	    if (pos.isInAntec()) {
		tacletInsts = tacletInsts.prepend
		    (antecTaclet(pos.constrainedFormula(),
			         filter, services,
				 userConstraint, tacletIndex)); 
	    } else {
		tacletInsts = tacletInsts.prepend
		    (succTaclet(pos.constrainedFormula(),
				filter,
			        services, userConstraint,
				tacletIndex)); 
	    } 
	} else {	    
	    tacletInsts = tacletInsts.prepend
		(getRewriteTaclet(pos, filter, services,
				  userConstraint, tacletIndex));
	}
	return tacletInsts;
    }


    /** 
     * collects all AntecedentTaclet instantiations for the given
     * heuristics and ConstrainedFormula
     * @param conFml the ConstrainedFormula the rules have to be
     * connected to
     * @param services the Services object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    private static ListOfNoPosTacletApp
        antecTaclet(ConstrainedFormula conFml,
		    RuleFilter         filter,
		    Services           services,
		    Constraint         userConstraint,
		    TacletIndex        tacletIndex) { 
	return tacletIndex.getAntecedentTaclet(conFml.formula(),
					       conFml.constraint(),
					       filter,
					       services,
					       userConstraint); 
    }
    
    /** 
     * collects all SuccedentTaclet instantiations for the given
     * heuristics and ConstrainedFormula
     * @param conFml the ConstrainedFormula the rules have to be
     * connected to
     * @param services the Services object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    private static ListOfNoPosTacletApp
	succTaclet(ConstrainedFormula conFml,
		   RuleFilter         filter,
		   Services           services,
		   Constraint         userConstraint,
		   TacletIndex        tacletIndex) {
	return tacletIndex.getSuccedentTaclet(conFml.formula(),
					      conFml.constraint(),
					      filter,
					      services,
					      userConstraint);    
    }


    /**
     * Check whether the given term is a metavariable, and replace it
     * with a concrete term provided that such a term is determined by
     * the user constraint
     * @return A <code>PosInOccurrence</code> object in which
     * eventually the metavariable has been replaced with a term as
     * given by the user constraint. In any case the object points to
     * the same position of a term as the <code>pos</code> parameter
     */
    private static PosInOccurrence handleDisplayConstraint
	( PosInOccurrence pos, Constraint displayConstraint ) {

	Term localTerm = pos.subTerm ();

	if ( localTerm.op () instanceof Metavariable ) {
	    if ( pos.termBelowMetavariable () == null ) {
		Term metaTerm = displayConstraint
		    .getInstantiation ( (Metavariable)localTerm.op () );
		if ( metaTerm.op () != localTerm.op () )
		    return pos.setTermBelowMetavariable ( metaTerm );
	    }
	}

	return pos;
    }


    /**
     * Descend and create indices for each of the direct subterms of
     * the given term
     * @param pos pointer to the term/formula for whose subterms
     * indices are to be created
     * @return list of the index objects
     */
    private static ListOfTermTacletAppIndex
        createSubIndices (PosInOccurrence pos,
		          Services        services,
		          Constraint      userConstraint,
		          Constraint      displayConstraint,
		          TacletIndex     tacletIndex,
			  NewRuleListener listener,
			  RuleFilter      filter) {
	ListOfTermTacletAppIndex result = SLListOfTermTacletAppIndex.EMPTY_LIST;
	Term localTerm = pos.subTerm ();
	int i = localTerm.arity();

	while ( i-- != 0 )
	    result = result.prepend ( createHelp ( pos.down(i),
						   services,
						   userConstraint,
						   displayConstraint,
						   tacletIndex,
						   listener,
						   filter ) );

	return result;
    }


    /**
     * Create an index for the given term
     * @param pos Pointer to the term/formula for which an index is to
     * be created. <code>pos</code> has to be a top-level term
     * position
     * @return the index object
     */
    public static TermTacletAppIndex create(PosInOccurrence pos,
					    Services        services,
					    Constraint      userConstraint,
					    TacletIndex     tacletIndex,
					    NewRuleListener listener,
					    RuleFilter      filter) {
	Debug.assertTrue ( pos.isTopLevel (),
			   "Someone tried to create a term index for a real " +
                           "subterm" );
	// This depends on the active implementation of
	// <code>SequentPrintFilter</code>, and has to be compatible
	// with the displayed sequent
	final Constraint localDisplayConstraint =
	    ConstraintSequentPrintFilter.determineDisplayConstraint
	    ( pos.constrainedFormula (), userConstraint );

	return createHelp ( pos,
			    services,
			    userConstraint,
			    localDisplayConstraint,
			    tacletIndex,
			    listener,
			    filter );
    }


    private static TermTacletAppIndex createHelp(PosInOccurrence pos,
						 Services        services,
						 Constraint      userConstraint,
						 Constraint      displayConstraint,
						 TacletIndex     tacletIndex,
						 NewRuleListener listener,
						 RuleFilter      filter) {
	pos       = handleDisplayConstraint ( pos, displayConstraint );
	Term localTerm = pos.subTerm ();

	ListOfNoPosTacletApp     localApps  = getFindTaclet( pos,
	                                                     filter,
							     services,
							     userConstraint,
							     tacletIndex );

	ListOfTermTacletAppIndex subIndices = createSubIndices( pos,
								services,
								userConstraint,
								displayConstraint,
								tacletIndex,
								listener,
								filter );

	fireRulesAdded ( listener, localApps, pos );

	return new TermTacletAppIndex ( localTerm,
	                                localApps,
	                                subIndices,
	                                displayConstraint,
					filter );
    }


    /**
     * Create a new tree of indices that additionally contain the taclets that
     * are selected by <code>filter</code>
     * @param filter The taclets that are supposed to be added
     * @param pos Pointer to the term/formula for which an index is to
     * be created. <code>pos</code> has to be a top-level term
     * position
     * @return the index object
     */
    public TermTacletAppIndex addTaclets(RuleFilter      filter,
					 PosInOccurrence pos,
					 Services        services,
					 Constraint      userConstraint,
					 TacletIndex     tacletIndex,
					 NewRuleListener listener) {
	RuleFilter effectiveFilter = new AndRuleFilter ( filter, ruleFilter );

	return addTacletsHelp 
	    ( effectiveFilter, pos, services, userConstraint, tacletIndex, listener );
    }

    private TermTacletAppIndex addTacletsHelp(RuleFilter      filter,
					      PosInOccurrence pos,
					      Services        services,
					      Constraint      userConstraint,
					      TacletIndex     tacletIndex,
					      NewRuleListener listener) {
	pos = handleDisplayConstraint ( pos, displayConstraint );

	final ListOfTermTacletAppIndex newSubIndices =
	    addTacletsSubIndices ( filter,
				   pos,
				   services,
				   userConstraint,
				   tacletIndex,
				   listener );
	
	final ListOfNoPosTacletApp additionalApps =
	    getFindTaclet ( pos,
			    filter,
			    services,
			    userConstraint,
			    tacletIndex );

	fireRulesAdded ( listener, additionalApps, pos );
			    
	return new TermTacletAppIndex ( term,
					localTacletApps.prepend ( additionalApps ),
					newSubIndices,
					displayConstraint,
					ruleFilter );
    }


    private ListOfTermTacletAppIndex
	addTacletsSubIndices( RuleFilter      filter,
			      PosInOccurrence pos,
			      Services        services,
			      Constraint      userConstraint,
			      TacletIndex     tacletIndex,
			      NewRuleListener listener ) {
        ListOfTermTacletAppIndex result = SLListOfTermTacletAppIndex.EMPTY_LIST;
	IteratorOfTermTacletAppIndex subIt = subtermIndices.iterator();
        
        int i = 0;
        while ( subIt.hasNext () ) {
            final TermTacletAppIndex oldSubIndex = subIt.next ();
	    final TermTacletAppIndex newSubIndex = oldSubIndex.addTacletsHelp
		( filter, pos.down(i), services, userConstraint, tacletIndex, listener );
	    result = result.append ( newSubIndex );
            ++i;
        }
        
        return result;
    }


    /**
     * Recursively update the term index, starting at <code>this</code> and
     * descending along the given path iterator to the term position below which
     * a modification was performed
     * @param pathToModification an iterator that walks from the root of the
     * formula to the position of modification
     * @return
     */
    private TermTacletAppIndex updateHelp(PIOPathIterator pathToModification,
					  Services        services,
					  Constraint      userConstraint,
					  TacletIndex     tacletIndex,
					  NewRuleListener listener ) {
	pathToModification.next ();

	final PosInOccurrence pos = pathToModification.getPosInOccurrence ();
	
	if ( !pathToModification.hasNext() ) {
	    // Below the position of modification everything has to be rebuilt
	    return createHelp ( pos,
				services,
				userConstraint,
				displayConstraint,
				tacletIndex,
				listener,
				ruleFilter );
	} else {
	    final Term subterm = pathToModification.getSubTerm ();
	    final ListOfTermTacletAppIndex newSubIndices =
		updateSubIndices ( pathToModification,
				   services,
				   userConstraint,
				   tacletIndex,
				   listener );

	    ListOfNoPosTacletApp localApps = getFindTaclet( pos,
		                                            ruleFilter,
						            services,
							    userConstraint,
							    tacletIndex );

	    fireRulesAdded ( listener, localApps, pos );

	    return new TermTacletAppIndex ( subterm,
					    localApps,
					    newSubIndices,
					    displayConstraint,
					    ruleFilter);
	}
    }


    /**
     * Update the subtree of indices the given iterator
     * <code>pathToModification</code> descends to 
     */
    private ListOfTermTacletAppIndex
	updateSubIndices ( PIOPathIterator pathToModification,
			   Services        services,
			   Constraint      userConstraint,
			   TacletIndex     tacletIndex,
			   NewRuleListener listener ) {
        ListOfTermTacletAppIndex subindices = subtermIndices;
        ListOfTermTacletAppIndex stack      =
	    SLListOfTermTacletAppIndex.EMPTY_LIST;
        
	int subterm = pathToModification.getChild();
        while ( subterm > 0 ) {
	    subterm--;
	    stack = stack.prepend ( subindices.head () );
	    subindices = subindices.tail ();
        }
        
        final TermTacletAppIndex oldIndex = subindices.head ();
        
        subindices = subindices.tail ().prepend
	    ( oldIndex.updateHelp ( pathToModification,
				    services,
				    userConstraint,
				    tacletIndex,
				    listener ) );
        
        while ( stack != SLListOfTermTacletAppIndex.EMPTY_LIST ) {
	    subindices = subindices.prepend ( stack.head () );
	    stack = stack.tail ();
        }

        return subindices;
    }


    /**
     * Updates the TermTacletAppIndex after a change at or below 
     * position <code>pos</code>
     * @param pos Pointer to the term/formula where a change occurred
     * @return the updated index object
     */
    public TermTacletAppIndex update ( PosInOccurrence pos,
				       Services        services,
				       Constraint      userConstraint,
				       TacletIndex     tacletIndex,
				       NewRuleListener listener ) {
	PIOPathIterator it = pos.iterator ();
	
	return updateHelp (it,
	                   services,
	                   userConstraint,
	                   tacletIndex,
	                   listener);
    }


    /**
     * @return the sub-index for the given position
     */
    private TermTacletAppIndex descend ( PosInOccurrence pos ) {
    	if ( pos.isTopLevel () )
    	    return this;
    	    
	PIOPathIterator    it  = pos.iterator ();
	TermTacletAppIndex res = this;
	
	while ( true ) {
	    int child = it.next ();

	    if ( child == -1 )
		return res;

	    res = res.getSubIndex ( child );
	}
    }


    /**
     * @return all taclet apps for the given position
     */
    public ListOfNoPosTacletApp getTacletAppAt ( PosInOccurrence pos,
						 RuleFilter p_filter ) {
	TermTacletAppIndex index = descend ( pos );
	return filter ( p_filter, index.localTacletApps );
    }


    /**
     * @return all taclet apps for or below the given position
     */
    public ListOfTacletApp getTacletAppAtAndBelow ( PosInOccurrence pos,
						    RuleFilter filter ) {
	return descend ( pos ).collectTacletApps ( pos, filter );
    }
    
    /**
     * Class that is used to convert <code>NoPosTacletApp</code>s to
     * <code>PosTacletApp</code>s via the method
     * <code>reportTacletApps</code>
     */
    private static class CollectTacletAppListener implements NewRuleListener {
        private ListOfTacletApp res = SLListOfTacletApp.EMPTY_LIST;
        private final RuleFilter filter;
        
        public CollectTacletAppListener ( RuleFilter p_filter ) {
            filter = p_filter;
        }
        
        public ListOfTacletApp getResult () {
            return res;
        }
        
        public void ruleAdded ( RuleApp app, PosInOccurrence pos ) {
            if ( filter.filter( ( app.rule() ) ) ) {
                TacletApp tacletApp = 
                    TacletAppIndex.createTacletApp( (NoPosTacletApp) app, pos );
                if ( tacletApp != null ) {
                    res = res.prepend ( tacletApp );
                }
            }
        }
    }
    
    /**
     * Collect all taclet apps that are stored by <code>this</code> (and by
     * the sub-indices of <code>this</code>). <code>NoPosTacletApp</code>s are
     * converted to <code>PosTacletApp</code>s using the parameter
     * <code>pos</code>
     * @param pos The position of this index
     * @return a list of all taclet apps
     */
    private ListOfTacletApp collectTacletApps ( PosInOccurrence pos,
						RuleFilter p_filter ) {
	pos = handleDisplayConstraint ( pos, displayConstraint );
    
	final CollectTacletAppListener listener = 
	    new CollectTacletAppListener ( p_filter );
	
	reportTacletApps ( pos, listener );
    
	return listener.getResult();
    }
    
    /**
     * Report all <code>NoPosTacletApp</code> s that are stored by
     * <code>this</code> (and by the sub-indices of <code>this</code>).
     * 
     * @param pos
     *            The position of this index
     * @param listener
     *            The listener to which the taclet apps found are supposed to be
     *            reported
     * @return a list of all taclet apps
     */
    public void reportTacletApps ( PosInOccurrence pos,
                                   NewRuleListener listener ) {
        pos = handleDisplayConstraint ( pos, displayConstraint );
    
        ListOfNoPosTacletApp rest = localTacletApps;
        while ( !rest.isEmpty() ) {
            listener.ruleAdded( rest.head(), pos );
            rest = rest.tail();
        }
        
        final IteratorOfTermTacletAppIndex it = subtermIndices.iterator ();
        int subterm = 0;
        while ( it.hasNext () ) {
            it.next ().reportTacletApps ( pos.down ( subterm ),
                                          listener );
            ++subterm;
        }
    }

    private static void fireRulesAdded ( NewRuleListener      listener,
				         ListOfNoPosTacletApp taclets,
				         PosInOccurrence       pos ) {
	IteratorOfNoPosTacletApp it = taclets.iterator();

	while ( it.hasNext () ) {
	    listener.ruleAdded ( it.next(), pos );
	}
    }


    /**
     * @return filtered list
     */
    public static ListOfNoPosTacletApp filter (RuleFilter p_filter,
                                               ListOfNoPosTacletApp taclets) {
        ListOfNoPosTacletApp result = SLListOfNoPosTacletApp.EMPTY_LIST;

        while ( taclets != SLListOfNoPosTacletApp.EMPTY_LIST ) {
            final NoPosTacletApp app = taclets.head ();
            taclets = taclets.tail ();
            if ( p_filter.filter ( app.taclet () ) )
            result = result.prepend ( app );
        }

        return result;
    }
}
