// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.proof.mgt.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.parser.*;
import java.io.*;
import de.uka.ilkd.key.gui.Main;
import java.util.Vector;
import java.util.ArrayList;
import java.util.HashMap;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.java.ConvertException;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.proof.mgt.Contract;
import de.uka.ilkd.key.proof.mgt.SpecExtractor;
import de.uka.ilkd.key.proof.mgt.OCLSpecExtractor;
import de.uka.ilkd.key.rule.SetOfTaclet;
import de.uka.ilkd.key.gui.ProofSettings;
import java.util.Iterator;
import de.uka.ilkd.key.collection.SetOfString;
import de.uka.ilkd.key.collection.SetAsListOfString;
import de.uka.ilkd.key.util.Debug;

/** represents an input from a .key problem file producing environment as
 * well as a proof obligation.
 */
public class KeYFile implements EnvInput {

    /** the RuleSource delivering the input stream for the file.
     */
    protected RuleSource file = null;
    
    protected InputStream input = null;

    /** the initial configuration to be used (and modified) while reading.
     */
    protected InitConfig initConfig = null;

    private String name;

    private String javaPath;
    private boolean javaPathAlreadyParsed=false;

    protected ProblemParser problemParser=null;

    /** the graphical entity to notify on the state of reading.
     */
    protected Main main;

    /** creates a new representation for a given file by indicating a name
     * and a file representing the physical source of the .key file.
     */
    public KeYFile(String name, File file) {
	this(name, RuleSource.initRuleFile(file), null);
    }

    /** creates a new representation for a given file by indicating a name
     * and a file representing the physical source of the .key file.
     * @param main the graphical entity to notify on the state of reading
     */
    public KeYFile(String name, File file, Main main) {
	this(name, RuleSource.initRuleFile(file), main);
    }

    /** creates a new representation for a given file by indicating a name
     * and a RuleSource representing the physical source of the .key file.
     * @param main the graphical entity to notify on the state of reading
     */
    public KeYFile(String name, RuleSource file, Main main) {
	this.file=file;
	this.name=name;
	this.main=main;
    }


    /** returns the initial configuration that is used to read and
     * that may be modified by reading.
     */
    public InitConfig getInitConfig() {
	return initConfig;
    }

    /** returns the number of chars in the .key file.
     */
    public int getNumberOfChars() {
	return file.getNumberOfChars();
    }
    

    protected InputStream getNewStream() {
        try {
            if (input != null) input.close();
        } catch(IOException ioe) {
            System.err.println("WARNING: Cannot close stream "+file+"\n"+ioe);
        }
        input = file.getNewStream();
        return input;
    }
    

    /** reads the whole .key file and modifies the initial configuration
     * assigned to this object according to the given modification strategy. 
     * Throws an exception if no initial configuration has been set yet.
     */
    public void read(ModStrategy mod) throws ProofInputException {
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	try {
            Debug.out("Reading KeY file", file);
	    CountingBufferedInputStream cinp = 
		new CountingBufferedInputStream
		    (getNewStream(),main,getNumberOfChars()/100);
	    ParserConfig pc = new ParserConfig
		(initConfig.getServices(), 
		 initConfig.namespaces().copy(), 
		 initConfig.getDefaultChoice().copy());
	    problemParser = new ProblemParser
		(new KeYLexer(cinp), file.toString(), pc, pc, initConfig.
		 getTaclet2Builder(), initConfig.getTaclets()); 
            problemParser.problem(); 
	    initConfig.setDefaultChoice(problemParser.getDefaultChoice());
	    initConfig.addCategory2DefaultChoices(problemParser.
						  getCategory2Default());
	    SetOfTaclet st = problemParser.getTaclets();
	    initConfig.setTaclets(st);
	    initConfig.add(problemParser.namespaces(), mod);
	    initConfig.add(problemParser.getHeuristicPrioritization(), mod);
	    if (initConfig.getProofEnv()!=null) {
		initConfig.getProofEnv().addContracts
		    (problemParser.getContracts());
	    }
            Debug.out("Read KeY file   ", file);
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	}
    }

    public Includes readIncludes() throws ProofInputException{
	try {
	    ParserConfig pc = new ParserConfig
		(new Services(), 
		 new NamespaceSet(), 
		 new Choice(new Name("Default")));
	    problemParser = new ProblemParser
		(new KeYLexer(getNewStream()), file.toString(), 
		 pc, pc, null, null); 
            problemParser.parseIncludes(); 
	    return problemParser.getIncludes();
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
        }
    }

    public void readActivatedChoices() throws ProofInputException{
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	try {
	    ParserConfig pc = new ParserConfig
		(new Services(), 
		 initConfig.namespaces(), 
		 initConfig.getDefaultChoice());
	    problemParser = new ProblemParser
		(new KeYLexer(getNewStream()), file.toString(), 
		     pc, pc, null, null); 
	    problemParser.parseWith(); 
	    initConfig.setActivatedChoices
		(problemParser.getActivatedChoices());
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	    
	}
    }

    public SetOfString getActivatedChoices() throws ProofInputException{
	Iterator it
	    = ProofSettings.THE_SETTINGS.getChoiceSettings()
	    .getDefaultChoices().values().iterator();
	SetOfString result = SetAsListOfString.EMPTY_SET;
	while (it.hasNext()) {
	    result = result.add((String)it.next());
	}
	return result;
    }

    private DeclParser createDeclParser() {
	return new DeclParser(new KeYLexer(getNewStream()),
			      file.toString(), 
			      initConfig.getServices().copy(),
			      initConfig.namespaces().copy(),
			      initConfig.getDefaultChoice());
    }

    /** reads the sorts declaration of the .key file only, 
     * modifying the sort namespace
     * of the initial configuration if allowed in the given 
     * modification strategy.
     */
    public void readSorts(ModStrategy mod) throws ProofInputException {
	try {
	    DeclParser p=createDeclParser();
	    p.parseSorts();
	    initConfig.addCategory2DefaultChoices(p.getCategory2Default());
	    initConfig.add(p.namespaces(), mod);
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	}
    }
    
    /** reads the functions and predicates declared in the .key file only, 
     * modifying the function namespaces of the respective taclet options. 
     */
    public void readFuncAndPred(ModStrategy mod) throws ProofInputException {
	try {
	    DeclParser p=createDeclParser();
	    p.parseFuncAndPred();
	    initConfig.add(p.namespaces(), mod);
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	}
    }

   /** reads the rules and problems declared in the .key file only, 
     * modifying the set of rules 
     * of the initial configuration if allowed in the given 
     * modification strategy.
     */
    public void readRulesAndProblem(ModStrategy mod) throws ProofInputException {
	CountingBufferedInputStream cinp = 
	    new CountingBufferedInputStream
	    (getNewStream(),main,getNumberOfChars()/100);
	/*
	  two namespace sets which share all namespace except the
	  variable namespace 	     	      
	*/
	NamespaceSet schema = initConfig.namespaces().copy();
	// some namespaces have to be shared
	NamespaceSet normal = new NamespaceSet();
	normal.setHeuristics(schema.heuristics());
	normal.setFunctions(schema.functions());
	normal.setChoices(schema.choices());
	ParserConfig schemaConfig = 
	    new ParserConfig(initConfig.getServices(), schema, 
			     initConfig.getDefaultChoice());
	ParserConfig normalConfig = 
	    new ParserConfig(initConfig.getServices(), normal, 
			     initConfig.getDefaultChoice());
	problemParser = new ProblemParser(new KeYLexer(cinp), 
					  file.toString(),
					  schemaConfig,
					  normalConfig,
					  initConfig.getTaclet2Builder(),
					  initConfig.getTaclets()); 
	try {
	    problemParser.parseTacletsAndProblem();
	    initConfig.add(schemaConfig.namespaces(), mod);
	    initConfig.add(normalConfig.namespaces(), mod);
	    SetOfTaclet st = problemParser.getTaclets();
	    initConfig.setTaclets(st);
	    initConfig.add(problemParser.getHeuristicPrioritization(), mod);
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	}
    }


    public void readProblem(ModStrategy mod) throws ProofInputException {
	readRulesAndProblem(mod);
    }


    /** defines the initial configuration that is to be used to read the .key
     * file and that the read declarations are added to. The set initial 
     * configuration is overwritten by a call to setConfig().
     */
    public void addTo(InitConfig conf) {
	this.initConfig=conf;
    }

    /** defines the initial configuration that is to be used to read the .key
     * file and that the read declarations are added to. The set initial 
     * configuration is overwritten by a call to addTo().
     */
    public void setConfig(InitConfig conf) {
	this.initConfig=conf;
    }

    /** reads a saved proof of a .key file
     */
    public void readProof(ProblemLoader prl) throws ProofInputException {
        try {
            problemParser.proof(prl);
        } catch(antlr.ANTLRException e) {
            throw new ProofInputException(e);
        }
    }

    /** returns the name of the .key file
     */
    public String name() {
	return name;
    }

    private Vector getClasses(String f) throws ProofInputException  {
	File cfile = new File(f);
	Vector v=new Vector();
	if (!cfile.isAbsolute()) { // test relative pathname
	    File parent=file.file().getParentFile();
	    cfile=new File(parent,f).getAbsoluteFile();	
	}
	if (cfile.isDirectory()) {
	    String[] list=cfile.list();
	    for (int i=0; i<list.length; i++) {
		String fullName = cfile.getPath()+File.separator+list[i];
		File n=new File(fullName);
		if (n.isDirectory()) {		    
		    v.addAll(getClasses(fullName));
		} else if (list[i].endsWith(".java")) {
		    v.add(fullName);	
		}
	    }
	    return v;
	} else {
	    throw new ProofInputException("Java model path "+f+" not found.");
	}
	
    }

    private String[] getClassesInPath(String f) throws ProofInputException {
	return (String[]) getClasses(f).toArray(new String[]{});
    }

    public String getJavaPath() throws ProofInputException {
	if (javaPathAlreadyParsed) return javaPath;
	problemParser = new ProblemParser(new KeYLexer(getNewStream()),
					  file.toString(), 
					  null, null, null, null);
        try { 	    
	    javaPath=problemParser.java(); 
	    javaPathAlreadyParsed=true;
	    return javaPath;
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
        }
    }

    public String getAbsoluteJavaPath() throws ProofInputException {
	String f=null;
	f = getJavaPath();
	if (f==null) return null;
	File cfile=new File(f);
	if (!cfile.isAbsolute()) { // test relative pathname
	    File parent=file.file().getParentFile();
	    cfile=new File(parent,f).getAbsoluteFile();	
	}
	return cfile.getAbsolutePath();
    }


    public String readJava() throws ProofInputException {
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	// no java is parsed at all just the java path
	javaPath = getJavaPath();
	if (javaPath != null) {
	    Recoder2KeY r2k = new Recoder2KeY(initConfig.getServices(),
					      initConfig.namespaces());

	    String[] cus = getClassesInPath(javaPath);
	    try { 
		r2k.readCompilationUnitsAsFiles(cus);
	    } catch (ConvertException ce) {
		throw new ProofInputException(ce);
	    } 
	}

        return javaPath;
    }

    public void close() {
        try {
            if (input != null) { 
		input.close();
	    }
        } catch(IOException ioe) {
            System.err.println("WARNING: Cannot close stream "+file+"\n"+ioe);
        }
    }

    public String toString() {
	return name()+" "+file.toString();
    }
    
    public boolean equals(Object o){
	if(!(o instanceof KeYFile)) return false;
	KeYFile kf = (KeYFile) o;
	return kf.file.file().getAbsolutePath().
	    equals(file.file().getAbsolutePath());
    }

    public int hashCode(){
	return file.file().getAbsolutePath().hashCode();
    }

    public void readSpecs(){
	// library classes specifications
	SpecExtractor spex = null;
	try {
	    String jpath = getJavaPath();
	    if (jpath==null) return;
	    spex = new OCLSpecExtractor(initConfig.getServices(),
					jpath);
	    initConfig.getProofEnv().addContracts(spex.getSpecsFromAllClasses());
	} catch (ProofInputException e) {
	    // do nothing
	    Debug.out("Exception thrown by class KeYFile at SpecExtraction");
	}
    }

    public Object getObjectOfContract() {
	return null;
    }

    public boolean fitsToContract(Contract ct) {
	return false;
    }

    public void startProtocol() {
	// do nothing
    }

}
