// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.ArrayOfExpression;
import de.uka.ilkd.key.java.TypeConverter;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.reference.MethodReference;
import de.uka.ilkd.key.java.reference.FieldReference;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.java.visitor.ProgramTypeCollector;
import de.uka.ilkd.key.java.declaration.ClassDeclaration;
import java.util.*;
import de.uka.ilkd.key.proof.mgt.Contract;
import de.uka.ilkd.key.collection.SetOfString;
import de.uka.ilkd.key.collection.SetAsListOfString;

/** represents a proof obligation originating from a given Modifies Clause
 * as input for the prover.
 */
public class ModifiesCheckProofOblInput implements ProofOblInput {

    private ReprModelMethod methRepr;
    private String name;
    private String xmifile;

    private InitConfig initConfig;
    
    private Term proofObl;

    private Set modifiesElements;
    private Set occuringClasses;
    private Set currentClassAttributes;

    private JavaInfo javaInfo;
    private ProgramMethod m;
    private String methodName;

    private ProgramVariable self;
    //private ProgramVariable pv;
    private Vector pvVector;

    //private Vector parameterNames = new Vector();
    private Vector parameterTypes;
    
    private Namespace ns = new Namespace();

    /** creates a new representation of a proof obligation from a Modifies 
     * Clause with the given name, originating from a representation of a model
     * method and having the given String as text of the Modifies Clause.
     * The given XMI file name can be used to read the underlying model.
     */
    public ModifiesCheckProofOblInput(String name, ReprModelMethod aMethRepr, 
				      String xmifile) {
	this.name=name+" "+aMethRepr.getName();
	this.methRepr=aMethRepr;
	this.xmifile=xmifile;
    }

    /** reads the underlying Java model and modifies the initial configuration
     * of this object. If the latter is not yet defined an exception is
     * thrown.
     */
    public String readJava() {
	if (initConfig==null) {
	    throw new IllegalStateException("InitConfig not set.");
	}
	Recoder2KeY r2k = new Recoder2KeY(initConfig.getServices(), 
					  initConfig.namespaces());
	String[] cus = methRepr.getContainingClass().getClassesInPackage();
	r2k.readCompilationUnitsAsFiles(cus);
	Debug.out("Initialize JavaInfo");
	return methRepr.getContainingClass().getRootDirectory();
    }

    public String getJavaPath() {
        return methRepr.getContainingClass().getRootDirectory();
    }


    /** returns false, that is the input never asks the user which
     * environment he prefers
     */
    public boolean askUserForEnvironment() {
	return false;
    }
    
    /** generates the proof obligation
     * 
     */
    public void read(ModStrategy mod) {
	if (initConfig==null) {
	    throw new IllegalStateException("InitConfig not set.");
	}
        
	javaInfo = initConfig.getServices().getJavaInfo();
	
	getAllFromModParsHelper();
	
	// extracts the signature as a list
	ListOfType signatureList = createSignatureList();
	
	// finds the current program method
	m = javaInfo.getProgramMethod
	    (javaInfo.getKeYJavaType(methRepr.getContainingClassName()),
	     methodName,
	     signatureList,
	     javaInfo.getKeYJavaType(methRepr.getContainingClassName()));
	//last argument added. hope it is right. /AR
	
	// puts all occuring classes into the set occuringClasses
	collectOccuringClasses();
	
	this.proofObl = createProof();
    }

    public void readProblem(ModStrategy mod) {
	read(mod);
    }

    private void getAllFromModParsHelper(){
	ModifiesParserHelper modParserH = 
	    new ModifiesParserHelper(methRepr, initConfig);
	
	// starts the internal computation of the helper class
	modParserH.parseModifiesClauseToHashSet();
	
	modifiesElements = modParserH.getModifiesElementsHashSet();
	methodName = modParserH.getMethodName();
	self = modParserH.getPVSelf();
	pvVector = modParserH.getPVVector();
	parameterTypes = modParserH.getParameterTypesVector();
    }

    private ListOfType createSignatureList(){
	ListOfType signatureList = SLListOfType.EMPTY_LIST;
	int index = parameterTypes.size()-1;
	while (index >= 0){
	    signatureList = signatureList.prepend((Type) parameterTypes.
	    				  elementAt(index));
	    index = index - 1;
	}
	return signatureList;
    }

    private void collectOccuringClasses() {
	ProgramTypeCollector mCollector = 
	    new ProgramTypeCollector(m, javaInfo);
	mCollector.start();
	occuringClasses = mCollector.result();
	occuringClasses.add(javaInfo.getKeYJavaType
				   (methRepr.getContainingClassName()));

	// get the transitive closure of the occuring classes in the
	// sense that for each class the type of all its' attributes
	// is in the resulting Set

	Set occuringClassesToAdd = (HashSet) ((HashSet) occuringClasses).clone();
	Iterator iterate;
	KeYJavaType current;
	while (!(occuringClassesToAdd.isEmpty())){
	    iterate = occuringClassesToAdd.iterator();
	    current = (KeYJavaType) iterate.next();
	    
	    Type currClassType = current.getJavaType();
	    if (!(currClassType instanceof PrimitiveType)){
	    	// get all attributes of the given class
		ListOfField allAttributes = javaInfo.getAllFields
		    ((ClassDeclaration) currClassType);
		IteratorOfField iterateAttributes = allAttributes.iterator();
		KeYJavaType addToOccCla = null;
		
		while (iterateAttributes.hasNext()){
		    ProgramVariable att = (ProgramVariable) iterateAttributes.next()
			.getProgramVariable();
		    if (!att.isImplicit()){
			addToOccCla = (KeYJavaType) att.getKeYJavaType();
			if (!(occuringClasses.contains(addToOccCla))){
			    occuringClasses.add(addToOccCla);
			    occuringClassesToAdd.add(addToOccCla);
			}
		    }
		}
	    }
	    occuringClassesToAdd.remove(current);
	}
    }
    
    
    private Term createProof() {
	Term formula = TermFactory.DEFAULT.createJunctorTermAndSimplify
	    (Op.IMP, 
	     TermFactory.DEFAULT.createJunctorTerm
	     (Op.NOT,
	      TermFactory.DEFAULT.createJunctorTerm
	      (Op.EQUALS,
	       TermFactory.DEFAULT.createVariableTerm(self),
	       TermFactory.DEFAULT.createFunctionTerm(Op.NULL))), 
	     createProofStart());
	return formula;
    }
    
    private Term createProofStart() {
	Iterator iterateClasses = occuringClasses.iterator();
	Term accumulateClassFormulas = 
	    TermFactory.DEFAULT.createJunctorTerm(Op.TRUE);
	KeYJavaType currClass;
	while (iterateClasses.hasNext()){
	    currClass = (KeYJavaType) iterateClasses.next();
	    
	    accumulateClassFormulas = 
		TermFactory.DEFAULT.createJunctorTermAndSimplify
		(Op.AND,
		 createClassFormula(currClass),
		 accumulateClassFormulas);
	}
	
	return accumulateClassFormulas;
    }
    
    private Term createClassFormula(KeYJavaType currClass){
	LogicVariable oprime  = new LogicVariable
	    (new Name("o"), 
	     currClass.getSort());
	
	// generate formula starting with \forall o': currClass
	Term classFormula = TermFactory.DEFAULT.createQuantifierTerm
	    (Op.ALL,
	     oprime ,
	     createClassSubFormula(currClass, oprime));
	
	return classFormula;
    }
    
    private Term createClassSubFormula(KeYJavaType currClass, 
				      LogicVariable oprime){
	Type currClassType = currClass.getJavaType();
	// stop primitive types from getting checked for attributes, there are
	// none anyway
	if (currClassType instanceof PrimitiveType)
	    {return TermFactory.DEFAULT.createJunctorTerm(Op.TRUE);}
	
	// get all attributes of the given class
	ListOfField allAttributes = javaInfo.getAllFields
	    ((ClassDeclaration) currClassType);
	IteratorOfField iterateAttributes = allAttributes.iterator();
	
	// some initial value
	Term accumulateAttributeFormulas = 
	    TermFactory.DEFAULT.createJunctorTerm(Op.TRUE);
	
	ProgramVariable currAttribute = null;

	while (iterateAttributes.hasNext()){
	    currAttribute = (ProgramVariable) 
		iterateAttributes.next().getProgramVariable();
	    
	    accumulateAttributeFormulas = 
		TermFactory.DEFAULT.createJunctorTermAndSimplify
		(Op.AND, 
		 createAttributeFormula(currAttribute, currClass, oprime),
		 accumulateAttributeFormulas);
	}

	
	return accumulateAttributeFormulas;
    }
    
    
    private Term createAttributeFormula(ProgramVariable currAttribute,
					KeYJavaType currClass,
					LogicVariable oprime) {
	if (!currAttribute.isImplicit()){

	    // create conjunction of o' != o111, o' != o112, ...
	    
	    // some initial value
	    Term ante = 
		TermFactory.DEFAULT.createJunctorTerm(Op.TRUE);
	    
	    Iterator modifiesIterator = modifiesElements.iterator();
	    ProgramElement modElm;
	    ProgramElement modProgVar;
	    ProgramElement modProgAttrib;
	    TypeConverter typeConv = initConfig.getServices().getTypeConverter();
	    while (modifiesIterator.hasNext()) {
		modElm = (ProgramElement) modifiesIterator.next();
		
		//check whether it is appropriate to add this to the ante
		modProgVar = modElm;
		
		if (! (modProgVar instanceof ProgramVariable)){
		    modProgAttrib = 
			(ProgramVariable) ((FieldReference) modProgVar)
			.getChildAt(((FieldReference) modProgVar)
				    .getChildCount()-1);
		    modProgVar = 
			((FieldReference) modProgVar).getReferencePrefix();
		    
		    // Check if the Class Types match
		    if (((Expression) modProgVar).getKeYJavaType(initConfig
								 .getServices(), null)
			== currClass) {
			// Check if the attributes match
			if (((ProgramVariable) modProgAttrib) == currAttribute){
			    // build formula
			    ante = TermFactory.DEFAULT.
				createJunctorTermAndSimplify
				(Op.AND,
				 TermFactory.DEFAULT.
				 createJunctorTerm
				 (Op.NOT,
				  TermFactory.DEFAULT.createEqualityTerm
				  (Op.EQUALS,
				   TermFactory.DEFAULT.
				   createVariableTerm(oprime),
				   (Term) typeConv.convertToLogicElement
				   ((ProgramElement) modProgVar))),
				 ante);
			}
		    }
		}
		else {
		    // The modifies element has been a single ProgramVariable,
		    // meaning no dereferencing has happened. In this case there
		    // is nothing that can be done because this shouldn't happen
		    // Only attributes can be changed and thus only those are 
		    // relevant in the modifies clause.
		    // THROW EXCEPTION HERE !!!
		    System.out.println("Error - Element with no dereferencing "+
				       "in the modifies clause: " + modProgVar);
		}	
	    }
	    
	    // create part after the implication arrow
	    LogicVariable x  = new LogicVariable
		(new Name("x"), 
		 currAttribute.getKeYJavaType().getSort());
	    
	    // Create the MethodReference belonging to the method whose modifies
	    // clause is being checked here.
	    Expression[] p = new Expression[pvVector.size()];
	    for (int i = 0; i < pvVector.size(); i = i+1) {
		p[i] = (ProgramVariable) pvVector.elementAt(i);
	    }
	    MethodReference mRef  = new MethodReference
		(new ArrayOfExpression(p),
		 new ProgramElementName(methodName),
		 self);
	    StatementBlock mBlock = new StatementBlock(mRef);
	    
	    
	    Term post = TermFactory.DEFAULT.createQuantifierTerm
		(Op.ALL, x, 
		 TermFactory.DEFAULT.createJunctorTermAndSimplify
		 (Op.IMP,
		  TermFactory.DEFAULT.createEqualityTerm
		  (Op.EQUALS,
		   TermFactory.DEFAULT.createVariableTerm(x),
		   TermFactory.DEFAULT.createAttributeTerm
		   (currAttribute, 
		    TermFactory.DEFAULT.createVariableTerm(oprime))), 
		  TermFactory.DEFAULT.createDiamondTerm
		  (JavaBlock.createJavaBlock(mBlock), 
		   TermFactory.DEFAULT.createEqualityTerm
		   (Op.EQUALS, 
		    TermFactory.DEFAULT.createVariableTerm(x), 
		    TermFactory.DEFAULT.createAttributeTerm
		    (currAttribute, 
		     TermFactory.DEFAULT.createVariableTerm(oprime))))));
	    
	    return TermFactory.DEFAULT.
		createJunctorTermAndSimplify(Op.IMP, ante, post);
	}
	else {
	    return TermFactory.DEFAULT.createJunctorTerm(Op.TRUE);
	}
    }
    

    

    public ProofList getPO() {
        if (proofObl == null) throw 
           new IllegalStateException("No proofObl term");
        String s = createProofHeader(methRepr.getContainingClass()
				     .getRootDirectory());
	return new ProofList(
           new Proof(name(), proofObl, s, 
		     initConfig.createTacletIndex(),
		     initConfig.createBuiltInRuleIndex()), name());
    }

    /** creates declarations necessary to save/load proof in textual form */
    String createProofHeader(String javaPath) {
        String s;
        s = "java {\""+javaPath+"\"}\n\n";
        s += "sorts {\n"; // create declaration header for the proof
        IteratorOfNamed it;
        it = initConfig.sortNS().getProtocolled();
        while (it.hasNext()) {
	    String n=it.next().toString();
	    int i;
	    if ((i=n.indexOf("."))<0 || 
		initConfig.sortNS().lookup(new Name(n.substring(0,i)))==null) {
		//the line above is for inner classes.
		//KeY does not want to handle them anyway...
		s = s+"object "+n+";\n";
	    }
	}

        s+="}\n\nprogram variables {\n";
        it = initConfig.funcNS().getProtocolled();
        while (it.hasNext()) 
           s = s+((ProgramVariable)(it.next())).proofToString();

        s+="}\n\nfunctions {\n";
        it = initConfig.funcNS().getProtocolled();
        while (it.hasNext()) s = s+((Function)(it.next())).proofToString();

        s+="}\n\n";
        return s;
    }

    /** sets the initial configuration that is used to read the OCL
     * input and that is used to be modified during reading.
     */
    public void setConfig(InitConfig conf) {
	this.initConfig=conf;
        initConfig.sortNS().startProtocol();
        initConfig.funcNS().startProtocol();
        initConfig.progVarNS().startProtocol();
    }

    /** returns the name of the modifies check proof obligation input.
     */
    public String name() {
	return name;
    }

    public void readActivatedChoices() throws ProofInputException {
	initConfig.setActivatedChoices(SetAsListOfChoice.EMPTY_SET);
    }

    public SetOfString getActivatedChoices() 
	throws ProofInputException{
	return SetAsListOfString.EMPTY_SET;
    }
    
    /** reads the include section and returns an Includes object.  
     */
    public Includes readIncludes() throws ProofInputException {
	RuleSource standard = RuleSource.initRuleFile("standardRules.key");
	Includes includes = new Includes();
	includes.put("standardRules", standard);
	return includes;
    }

    public void readSpecs(){
	// nothing here
    }

    public Term getPOTerm() {
	return proofObl;
    }
    
    
    public Set getModifiesElements() {
	return modifiesElements;
    }

    public Object getObjectOfContract() {
	return methRepr;
    }

    public boolean fitsToContract(Contract ct) {
	return false; //%%
    }

    public void startProtocol() {
	// do nothing
    }

    
}
