// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof.init;

import java.io.File;

import de.uka.ilkd.key.gui.Main;
import de.uka.ilkd.key.logic.IteratorOfNamed;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.NamespaceSet;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.TacletIndex;
import de.uka.ilkd.key.rule.IteratorOfTacletApp;
import de.uka.ilkd.key.rule.SetOfNoPosTacletApp;
import de.uka.ilkd.key.rule.SetAsListOfNoPosTacletApp;
import de.uka.ilkd.key.rule.NoPosTacletApp;
import de.uka.ilkd.key.rule.SetOfTaclet;
import de.uka.ilkd.key.rule.SetAsListOfTaclet;
import de.uka.ilkd.key.rule.IteratorOfTaclet;
import de.uka.ilkd.key.rule.soundness.POBuilder;
import de.uka.ilkd.key.rule.soundness.POSelectionDialog;
import de.uka.ilkd.key.proof.mgt.LemmaSpec;
import de.uka.ilkd.key.proof.mgt.ProofEnvironment;
import de.uka.ilkd.key.proof.mgt.RuleJustificationByMethodSpec;
import de.uka.ilkd.key.proof.mgt.Contract;
import de.uka.ilkd.key.proof.ProofList;
import de.uka.ilkd.key.logic.Name;

public class TacletSoundnessPO extends KeYUserProblemFile {

    private ProofList proof;
    private NoPosTacletApp app;
    
    public TacletSoundnessPO (String name, File file, Main main) {
        super ( name, file, main );
    }
    
    /** returns the proof obligation term as result of the proof obligation
     * input. If there is still no input available because nothing has
     * been read yet null is returned.
     */
    public ProofList getPO () {
        return proof;
    }

    public NoPosTacletApp getTaclet () {
        return app;
    }

    /** starts reading the input and modifies the InitConfig of this 
     * object with respect to the given modification
     * strategy. 
     */
    public void read(ModStrategy mod) throws ProofInputException {

	InitConfig old = initConfig;
	initConfig = new InitConfig();
	initConfig.setProofEnvironment(old.getProofEnv());
	initConfig.setServices(old.getServices());
	initConfig.namespaces().setHeuristics(old.heurNS());
	SetOfTaclet newTaclets=null;
	try {
	    super.read ( ModStrategy.NO_VARS_FUNCS); // actually this
						     // reads the
						     // complete
				// problem, which is not really
				// needed; could be optimized
	    newTaclets = initConfig.getTaclets();
	} finally {
	    initConfig=old;
	}        
	// this ensures that necessary Java types are loaded
        initConfig.getServices().getJavaInfo().readJavaBlock("{}");
        
	IteratorOfTaclet it = newTaclets.iterator();
	SetOfNoPosTacletApp newTacApps = SetAsListOfNoPosTacletApp.EMPTY_SET;
	while (it.hasNext()) {
	    newTacApps = newTacApps.add
		(NoPosTacletApp.createNoPosTacletApp(it.next()));
	}

        final POSelectionDialog dialog = new POSelectionDialog 
	    ( Main.getInstance().mediator (),
	      newTacApps);

        app = dialog.getSelectedTaclet ();

        if ( app == null )
            throw new ProofInputException ( "No taclet was selected" );

        final POBuilder pob = new POBuilder ( app, initConfig.getServices() );
        pob.build ();

        updateNamespaces ( pob );
        
        proof = new ProofList
	    (new Proof [] { new Proof ( name (),
					pob.getPOTerm (),
					getProblemHeader (),
					initConfig.createTacletIndex (),
					initConfig.createBuiltInRuleIndex (),
					initConfig.getActivatedChoices () ) }, 
	     name());

	LemmaSpec lemmaSpec = new LemmaSpec(getTaclet());
	ProofEnvironment env = initConfig.getProofEnv();
	env.addContract(lemmaSpec);
	env.registerProof(this, proof);
	env.registerRule(getTaclet(), 
			 new RuleJustificationByMethodSpec(lemmaSpec));
	env.addToAllProofs(getTaclet());
    }

    public void readProblem(ModStrategy mod) throws ProofInputException {
	read(mod);
    }

    private void updateNamespaces (POBuilder p_pob) {
        NamespaceSet globalNss = initConfig.namespaces();
        Namespace funcNs = globalNss.functions ();

        {
            final IteratorOfNamed it =
                        p_pob.getFunctions ().allElements ().iterator ();
            while ( it.hasNext () )
                funcNs.add ( it.next () );
        }

//         {
//             final IteratorOfTacletApp it = p_pob.getTaclets ().iterator ();
//             while ( it.hasNext () )
//                 p_tacletIndex.add((NoPosTacletApp)it.next ());
//         }
    }

    /**
     * reads the Java model that belongs to this input.
     */
    public String readJava() throws ProofInputException {
        return null;
    }


    /** set the initial configuration used to read an input. It may become
     * modified during reading depending on the modification strategy used
     * for reading.
     */
    public void setConfig(InitConfig i) {
        initConfig = i;
    }

    public void readActivatedChoices() throws ProofInputException {}
    
    /** reads the include section and returns an Includes object.  
     */
    public Includes readIncludes() throws ProofInputException {
        return new Includes ();
    }
    
    /** returns the name of the proof obligation input.
     */
    public String name() {
	if (app==null) return "Taclet proof obligation";
        return "Proof obligation for "+app.taclet().name();
    }

    public Object getObjectOfContract() {
	return app;
    }

    public boolean fitsToContract(Contract ct) {
	if (!(ct instanceof LemmaSpec)) {
	    return false;
	}
	LemmaSpec lct = (LemmaSpec)ct;
	return getObjectOfContract().equals(lct.getObjectOfContract());
    }


}
