// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof.mgt;

import de.uka.ilkd.key.java.ArrayOfExpression;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.JavaProgramElement;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.NonTerminalProgramElement;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.ContextStatementBlock;
import de.uka.ilkd.key.java.declaration.MethodDeclaration;
import de.uka.ilkd.key.java.expression.Assignment;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.reference.ConstructorReference;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.java.statement.MethodBodyStatement;
import de.uka.ilkd.key.java.statement.CatchAllStatement;
import de.uka.ilkd.key.java.statement.Throw;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.proof.init.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.util.*;
import java.util.*;

/** 
 * This abstract class implements the transformation function used to
 * generate a method lemma taclet from the method's specification. The
 * transformation is described in detail in the minor thesis of Bastian
 * Katz for OCL specifications.
 * The generated taclet is of similar to :
 *  <code>
 *     find(==> <{t::mbs(params);}> #postcond 
 *     add(==> pre);
 *     add(==>\forall modifies_schema_variables 
 *                 (\forall transformation_vars (pi1 -> pi2_with_updates) 
 *                     -> #postcond_with_updates))
 *  </code>
 * where <ul>
 *    <li><code>pre</code> is the specified precondition </li>
 *    <li><code>postcond</code> is the specified postcondition </li>
 *    <li><code>pi1, pi2, modifies_*, </code> are connected with the
 *    modifies clause and the elimination of <code>@pre</code> like operators</li>
 * </ul>
 */
public abstract class AbstractMethodLemmaFactory {
    

    private final TermFactory tF = TermFactory.DEFAULT;
    
    private Term[] modSUTArr; 
    private Term schVarPsiTerm;
    private Statement transformedStatement;
    private Term preCondition;
    private Term postCondition;
    private Services services;
    private SchemaVariable schVarPsi;

    private static int uniqueNameCounter = 1;
    
    /** 
     * precondition of the method whose specification is processed
     * (called "PRE" in the minor thesis of Bastian Katz) 
     */
    protected abstract Term getPreCond(); 
    
    /** 
     * quantifiable variables used to store the values of non-rigid
     * functions in order to access them in the postcondition of the
     * method (called "transformation_vars" in Bastian Katz's minor
     * thesis)
     */
    protected abstract ArrayOfQuantifiableVariable getExtraVars(); 

    /** 
     * additional precondition necessary to eliminate
     * <code>@pre</code> like operators (called "pi1" in the minor
     * thesis of Bastian Katz)
     */
    protected abstract Term getExtraPre();
    
    /** 
     * post condition of the method specification
     * (called "pi2" in the minor thesis of Bastian Katz)
     */
    protected abstract Term getPostCond();

    /** modifies elements as a set of terms */
    protected abstract SetOfTerm getModifies(); 
    
    /** 
     * statement before the transformation so we can get the real
     * variables (no schema variables) 
     */
    protected abstract ProgramElement getProofStatement();

    protected abstract Modality getModality();
    
    /** 
     * string which is printed out a couple of times and given in the
     * constructor of the classes that extend this abstract class 
     */
    protected abstract String getSuffix();

    /** heuristic the taclet belongs to */
    protected abstract Heuristic getHeuristic();

    protected AbstractMethodLemmaFactory() {}

    protected AbstractMethodLemmaFactory(Services services) {
	this.services = services;
    }

    /**
     * creates and returns the generated taclet for the given method
     * specification
     * @return the taclet that can be used to insert the given method
     * specification rather than the concrete method implementation
     */
    public Taclet getTaclet() {	
	if ( getModifies() == null ) {
	    return null;
	}
	final int modifiesSize = getModifies().size();
	Map schemaVariablesMap = new HashMap();
	
	createSpecialSVMappings(schemaVariablesMap);

	// special treatment to specify thrown exceptions
	ProgramVariable catchExceptionsAs = null;
	Statement unwrappedProofStatement 
	    = unwrapStatementBlocks((Statement)getProofStatement());
	if (unwrappedProofStatement instanceof CatchAllStatement) {
	    CatchAllStatement catchAll 
		= (CatchAllStatement) unwrappedProofStatement;
	    unwrappedProofStatement = catchAll.getBody();
	    catchExceptionsAs 
		= (ProgramVariable) catchAll.getParameterDeclaration()
		      .getVariableSpecification().getProgramVariable();	
	    SchemaVariable catchExcAsSV = SchemaVariableFactory.
		createProgramSV(new ProgramElementName("#thrownException"), 
				ProgramSVSort.VARIABLE, false);
	    schemaVariablesMap.put(catchExceptionsAs, catchExcAsSV);
	}

	// substitute program variables by schema variables
	final TermSVReplacer tSVR = new TermSVReplacer(schemaVariablesMap);
	transformedStatement = tSVR.inst(unwrappedProofStatement);
	preCondition = tSVR.inst(getPreCond()); // PRE
	Term extraPreCondition = tSVR.inst(getExtraPre()); // pi1
	postCondition = tSVR.inst(getPostCond()); // pi2

	// create the updated specification postcondition and the
	// logic variables therefore
	final QuantifiableVariable[] logicQuantVarsArr
	    = new QuantifiableVariable[modifiesSize];
	Term pi2Upd  = getPi2Upd(logicQuantVarsArr, tSVR);
	
	//now start to build the taclet...
	SuccTacletBuilder stb = new SuccTacletBuilder();

	Map lV2SV = new HashMap();
	// this is the second add clause for the postcondition
	Term addFormula = createNormalCaseFormula(pi2Upd, extraPreCondition,
						  catchExceptionsAs,
						  logicQuantVarsArr, 
						  schemaVariablesMap, lV2SV,
						  modSUTArr);

        stb = stb.setFind(createFindTerm());		
	
	// add one add part to the taclet builder
	stb.addTacletGoalTemplate(createTGTfromTerm(preCondition, 
						    "Precondition"));
	
	// add the other add part to the taclet builder
	stb.addTacletGoalTemplate(createTGTfromTerm(addFormula, 
						    "Postcondition"));
	// add third (exceptional) case if necessary
	if (catchExceptionsAs!=null) {
	    Term excFormula = createExceptionalCaseFormula
		(pi2Upd, extraPreCondition, catchExceptionsAs, 
		 logicQuantVarsArr, schemaVariablesMap, lV2SV, modSUTArr);
	    stb.addTacletGoalTemplate(createTGTfromTerm(excFormula,
							"Exception thrown"));
	    stb.addVarsNew
		((SchemaVariable)schemaVariablesMap.get(catchExceptionsAs), 
		 catchExceptionsAs.sort());
	}

	//add `not free in' variable conditions
	Iterator it = lV2SV.values().iterator();
	while (it.hasNext()) {
	    stb.addVarsNotFreeIn((SchemaVariable)it.next(), 
				 (SchemaVariable)schVarPsiTerm.sub(0).op());
	}

	//add `new' variable conditions
	it = tSVR.getLocallyDeclaredVars().iterator(); 
	while (it.hasNext()) {
	    ProgramVariable pv = (ProgramVariable) it.next();
	    stb.addVarsNew((SchemaVariable)schemaVariablesMap.get(pv), 
			   pv.sort());	    
	}

	//add names and rule sets
	stb.setDisplayName(getTacletDisplayName(transformedStatement, 
						((catchExceptionsAs==null) ? "" 
						 : "exceptional_")+getSuffix()));
	stb.setName(getTacletName(transformedStatement));
	if (getHeuristic()!=null) {
	    stb.addHeuristic(getHeuristic());
	}

	Taclet t = stb.getSuccTaclet();
	return t;
    }

    private StringBuffer getNameBody(Statement transformedStatement) {
	final StringBuffer stbName = new StringBuffer();
	if (transformedStatement instanceof MethodBodyStatement) {
	    stbName.append
		(((MethodBodyStatement) transformedStatement).
		 getProgramMethod().name().toString());	    
	}
	return stbName;
    }

    private String getTacletDisplayName(Statement transformedStatement,
					String suffix) {
	return getNameBody(transformedStatement) + "_"+suffix;
    }

    private Name getTacletName(Statement transformedStatement) {
	final StringBuffer stbName = getNameBody(transformedStatement);
	if (transformedStatement instanceof MethodBodyStatement) { 
	    final MethodDeclaration meth = 
		((MethodBodyStatement) transformedStatement).
		getProgramMethod().getMethodDeclaration();

	    for (int i = 0; i<meth.getParameterDeclarationCount(); i++) {
		stbName.append("_");
		stbName.append(meth.getParameterDeclarationAt(i).
			       getVariableSpecification().getType().getName());
	    }
	} 
	stbName.append("_");
	stbName.append(getSuffix());	
	stbName.append("_"+(uniqueNameCounter++));
	return new Name(stbName.toString());
    }


    private Term getPi2Upd(QuantifiableVariable[] logicQuantVarsArr,
			   TermSVReplacer tSVR) {
	int modifiesSize = logicQuantVarsArr.length;
	modSUTArr = new Term[2*modifiesSize+1]; 
	SetOfTerm modifiesSchema = SetAsListOfTerm.EMPTY_SET;	
	final IteratorOfTerm modElmsAsTermsIt = getModifies().iterator();
	for (int counter = 0; counter<modifiesSize; counter++) {
	    final Term nextModifierTerm = modElmsAsTermsIt.next();
	    logicQuantVarsArr[counter] = getLogicVariable(nextModifierTerm);
	    Term schemaModTerm = tSVR.inst(nextModifierTerm);
	    modSUTArr[2*counter] = schemaModTerm;
	    modSUTArr[2*counter+1] 
		= tF.createVariableTerm((LogicVariable)logicQuantVarsArr[counter]);
	    modifiesSchema = modifiesSchema.add(schemaModTerm);
	}
	if (modifiesSize != 0) {
	    modSUTArr[modSUTArr.length-1] = postCondition;
	    return tF.createUpdateTerm(modSUTArr);
	} else {
	    return postCondition; 
	}	
    }

    private LogicVariable getLogicVariable(Term nextModifierTerm) {
	final Name varName;
	if (nextModifierTerm.op() instanceof AttributeOp) {
	    varName = new Name
		("c_"+ ((AttributeOp)nextModifierTerm.op()).attribute());
	} else if (nextModifierTerm.op() instanceof ProgramVariable) {
	    varName = new Name("c_"+ nextModifierTerm.op().name());
	} else if (nextModifierTerm.op() instanceof ArrayOp) {
	    varName = new Name("c_"+ nextModifierTerm.sub(0).op().name()+"_"
			       +nextModifierTerm.sub(1).op().name());
	} else {
	    varName = new Name("c_"+ nextModifierTerm.op().name());
	}
	return new LogicVariable(varName, nextModifierTerm.sort());
    }

    private Term createPsiUpdated(Statement st, Term[] modSUTArr) {
	if (schVarPsi==null) {
	    schVarPsi = SchemaVariableFactory.createFormulaSV
	    (new ProgramElementName("#psi"), false);
	}

	final Term psiUpd;

	ExtList statementList = new ExtList();
	if (st!=null) {
	    statementList.add(st);
	} 

	schVarPsiTerm = tF.createProgramTerm
	    (getModality(),
	     JavaBlock.createJavaBlock(new ContextStatementBlock(statementList)), 
	     tF.createVariableTerm(schVarPsi));
	
	if (getModifies().size() == 0) {
	    psiUpd = schVarPsiTerm;
	} else {
	    final Term[] modSUTrArr = new Term[modSUTArr.length];

	    System.arraycopy(modSUTArr, 0, modSUTrArr, 0, modSUTArr.length-1);
	    modSUTrArr[modSUTrArr.length - 1] = schVarPsiTerm;

	    psiUpd = tF.createUpdateTerm(modSUTrArr);	
	}

	return psiUpd;
    }
    
    private static SortedSchemaVariable getSV(QuantifiableVariable lv, 
					      Map logic2Schema) {
	if (!logic2Schema.containsKey(lv)) {
	    logic2Schema.put(lv, 
			     (SortedSchemaVariable)SchemaVariableFactory.
			     createVariableSV(lv.name(), lv.sort(), false));
	}
	return (SortedSchemaVariable)logic2Schema.get(lv);
    }

    /**
     * replaces all logic variables in t with schema variables. The used
     * mapping logic variable to schema variable is stored in the
     * logic2Schema map 
     * @param t the Term whose logic variables have to be replaced
     * @param logic2Schema a Map associating each logic variable to a
     * schema variable so that the same logic variable is replaced with
     * the same term
     * @return a term with all logic variables replaced by schema
     * variables 
     */
    private static Term bindLogicVarsAsSchemaVars(Term t, Map logic2Schema) {
	final Term[] subs = new Term[t.arity()];
	final ArrayOfQuantifiableVariable[] a
	    = new ArrayOfQuantifiableVariable[t.arity()];

	boolean changed = false;
	for (int i=0; i<subs.length; i++) {

	    subs[i] = bindLogicVarsAsSchemaVars(t.sub(i), logic2Schema);
	    if (subs[i]!=t.sub(i)) changed = true;

	    final ArrayOfQuantifiableVariable boundVars = t.varsBoundHere(i);
	    final int boundVarsSize = boundVars.size();
	    final QuantifiableVariable[] qs = 
		new QuantifiableVariable[boundVarsSize];

	    changed = (changed || boundVarsSize > 0);
	    for (int j=0; j<boundVarsSize; j++) {
		qs[j] = getSV(boundVars.getQuantifiableVariable(j), 
			      logic2Schema);
	    }
	    a[i] = new ArrayOfQuantifiableVariable(qs);		    
	}

	if (t.op() instanceof LogicVariable) {
	    return TermFactory.DEFAULT.createVariableTerm
		(getSV((LogicVariable)t.op(), logic2Schema));
	} else if (changed){
	    return TermFactory.DEFAULT.createTerm(t.op(), subs, a, 
						  t.javaBlock(), null);
	} else {
	    return t;
	}
    }

    private static ArrayOfQuantifiableVariable getSVs
	(ArrayOfQuantifiableVariable qvs, Map m) {

	final QuantifiableVariable[] a = new QuantifiableVariable[qvs.size()];
	for (int i = 0; i<a.length; i++) {
	    a[i] = (QuantifiableVariable)m.get(qvs.getQuantifiableVariable(i));
	}
	return new ArrayOfQuantifiableVariable(a);
    }

    private Term createAddFormula(Term pi2Upd, Term psiUpd, 
				  Term extraPreCondition,
				  QuantifiableVariable[] logicQuantVarsArr,
				  Map lV2SV) {
	Term addFormula = tF.createJunctorTermAndSimplify(Op.AND, 
							  extraPreCondition, 
							  pi2Upd);
	
	addFormula = bindLogicVarsAsSchemaVars(addFormula, lV2SV);
	psiUpd     = bindLogicVarsAsSchemaVars(psiUpd, lV2SV);

	final ArrayOfQuantifiableVariable proofQuantVars = getExtraVars();
	if (proofQuantVars.size() != 0) {
	    addFormula = tF.createQuantifierTerm
		(Op.EX, getSVs(proofQuantVars, lV2SV), addFormula);
	}
	addFormula = tF.createJunctorTermAndSimplify(Op.IMP, 
						     addFormula, 
						     psiUpd);
	if (logicQuantVarsArr.length != 0) {
	    addFormula = tF.createQuantifierTerm
		(Op.ALL, 
		 getSVs(new ArrayOfQuantifiableVariable(logicQuantVarsArr), 
			lV2SV), 
		 addFormula);	    
	}	
	return addFormula;
    }  

    private Term createNormalCaseFormula(Term pi2Upd, Term extraPreCondition,
					 ProgramVariable catchExceptionsAs,
					 QuantifiableVariable[] 
					 logicQuantVarsArr, 
					 Map schemaVariablesMap,
					 Map lV2SV,
					 Term[] modSUTArr) {
	Term extraPre = extraPreCondition;
	if (catchExceptionsAs!=null) {
	    extraPre = tF.createJunctorTermAndSimplify
		(Op.AND, extraPreCondition,
		 tF.createEqualityTerm
		 (tF.createFunctionTerm
		  ((TermSymbol)schemaVariablesMap.get(catchExceptionsAs)), 
		  services.getJavaInfo().getNullTerm()));
	}	
	return createAddFormula(pi2Upd, createPsiUpdated(null, modSUTArr), 
				extraPre, logicQuantVarsArr, lV2SV);
    }

    private Term createExceptionalCaseFormula(Term pi2Upd, 
					      Term extraPreCondition,
					      ProgramVariable catchExceptionsAs, 
					      QuantifiableVariable[] 
					      logicQuantVarsArr, 
					      Map schemaVariablesMap,
					      Map lV2SV,
					      Term[] modSUTArr) {
	SchemaVariable catchExcAsSV 
	    = (SchemaVariable)schemaVariablesMap.get(catchExceptionsAs);
	Term preexc = tF.createJunctorTermAndSimplify
	    (Op.AND, extraPreCondition, 
	     tF.createJunctorTerm
	     (Op.NOT,
	      tF.createEqualityTerm(tF.createFunctionTerm
				    ((TermSymbol)catchExcAsSV), 
				    services.getJavaInfo().getNullTerm())));
	ExtList args = new ExtList();
	args.add(catchExcAsSV);
	Term psiUpd = createPsiUpdated(new Throw(args), modSUTArr);
	return createAddFormula(pi2Upd, psiUpd, preexc,
				logicQuantVarsArr, lV2SV);	
    }

    private Statement unwrapStatementBlocks(Statement s) {
	while (s instanceof StatementBlock 
	       && ((StatementBlock)s).getChildCount()==1) {
	    s = (Statement)((StatementBlock)s).getChildAt(0);
	}
	return s;
    }

    private Term createFindTerm() {
	transformedStatement = unwrapStatementBlocks(transformedStatement);
	ExtList children = new ExtList();
	if (transformedStatement instanceof StatementBlock 
	    && ((StatementBlock)transformedStatement).getChildCount()>1) {
	    for (int i=0; i<((StatementBlock)transformedStatement).getChildCount();
		 i++) {
		children.add(((StatementBlock)transformedStatement).getChildAt(i));
	    }
	} else {
	    children.add(transformedStatement);
	}
	return tF.createProgramTerm
	    (getModality(),
	     JavaBlock.createJavaBlock(new ContextStatementBlock(children)), 
	     schVarPsiTerm.sub(0));
    }

    private TacletGoalTemplate createTGTfromTerm(Term add, String name) {
	final Semisequent semSeq = Semisequent.EMPTY_SEMISEQUENT.
	    insertFirst(new ConstrainedFormula(add)).semisequent();
	final TacletGoalTemplate tgt =
	    new AntecSuccTacletGoalTemplate(Sequent.EMPTY_SEQUENT,
					    SLListOfTaclet.EMPTY_LIST, 
					    SLListOfRulevariable.EMPTY_LIST,
					    Sequent.createSuccSequent(semSeq), 
					    SetAsListOfSchemaVariable.EMPTY_SET);
	tgt.setName(name);
	return tgt;
    }

    private void createSpecialSVMappings(Map schemaVariablesMap) {
	//the next case is to allow non-simple expressions like self.a
	//as callee in a mbs. nothing wrong can happen since mbs exist
	//only if there is no null pointer exception thrown
	Statement sta 
	    = (Statement)unwrapStatementBlocks((Statement)getProofStatement());
	if (sta instanceof MethodBodyStatement) {
	    MethodBodyStatement mbs = (MethodBodyStatement) sta;
	    SchemaVariable methodPrefixSV = null;
	    if (!mbs.getProgramMethod().isStatic()) {
 		methodPrefixSV = SchemaVariableFactory.
		    createProgramSV(new ProgramElementName("#self"), 
				    ProgramSVSort.EXPRESSION, false);
 		schemaVariablesMap.put(mbs.getDesignatedContext(), 
 				       methodPrefixSV);
 	    }
	}
    }
    
}
