// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


// CvsRunner.java
// $Id: CvsRunner.java 1.4.1.9.1.1 Thu, 14 Oct 2004 15:24:56 +0200 klebanov $  
// (c) COPYRIGHT MIT and INRIA, 1997.
// Please first read the full copyright statement in file COPYRIGHT.html
// http://dev.w3.org/cvsweb/java/classes/org/w3c/cvs/CvsRunner.java

// Contains code from com.ice.jcvslet.JCVSletLog by Tim Endres, time@gjt.org



package de.uka.ilkd.key.proof.mgt;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FilterInputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.Reader;

import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;

import de.uka.ilkd.key.util.Debug;



public class CvsRunner {
//    private static final File tmpdir = new File("/tmp");
    public static boolean debug = true;
    
    private static final String DOT_KEY = System.getProperty("user.home")+
       File.separator+".key";
    private static final String REP_ROOT = DOT_KEY+File.separator+"CVS-PO-REP"+
       File.separator;
    private static final String ROOT_OPT = " -d "+REP_ROOT;

    private Logger mgtLogger = Logger.getLogger("key.proof.mgt");
    

    /**
     * Dump the given string into a temporary file.
     * This is used for th <code>-f</code> argument of the cvs commit command.
     * This method should only be used from a synchronized method.
     * @param string The string to dump.
     */

    File getTemporaryFile (String string) 
	throws CvsException
    {
	// Create a pseudo-random temporary filename
	String fn = "cvs-" + System.currentTimeMillis()+"-"+string.length();
	File temp = null;
	try {
	    temp = File.createTempFile (fn, "-"+string.length()) ;
	    temp.deleteOnExit();
	    PrintStream out  = new PrintStream (new FileOutputStream(temp)) ;
	    out.print(string) ;
	    out.close() ;
	    return temp ;
	} catch (IOException ex) {
	    error ("temporaryFile"
		   , "unable to create/use temporary file: " 
		   + temp.getAbsolutePath()) ;
	}
	return temp ;
    }

    /**
     * Emit an error.
     * Some abnormal situation occured, emit an error message.
     * @param mth The method in which the error occured.
     * @param msg The message to emit.
     * @exception CvsException The exception that will be thrown as a 
     *     result of the error.
     */

    protected void error (String mth, String msg) 
	throws CvsException
    {
	String emsg = this.getClass().getName()+"["+mth+"]: "+msg ;
	throw new CvsException (emsg) ;
    }

    /**
     * Read the given input stream as a text.
     * @param in The input stream to read from.
     * @param into The StringBuffer to fill.
     * @return The provided StringBuffer, filled with the stream content.
     */

    private StringBuffer readText(InputStream procin, StringBuffer into) 
	throws IOException 
    {
	BufferedReader in   = new BufferedReader(new InputStreamReader(procin));
	String         line = null ;

	while ((line = in.readLine()) != null) {
	    if ( into != null )
		into.append (line+"\n") ;
	}
	return into ;
    }

    /**
     * Get a filename between quote contained in this String.
     * @return the filename of null.
     */
    protected String getQuotedFilename(String line) {
	int idx = line.indexOf('\'');
	if (idx == -1)
	    return null;
	char ch;
	StringBuffer buffer = new StringBuffer();
	try {
	    while ( (ch = line.charAt(idx++)) != '\'' )
		buffer.append(ch);
	} catch (ArrayIndexOutOfBoundsException ex) 
	    {Debug.out("Exception thrown by class CsvRunner at buffer.append()");}
	return buffer.toString();
    }

    /**
     * Parse the error message and throw an exception if necessary.
     * @exception UpToDateCheckFailedException if file was not up to date.
     * @exception CvsAddException if an error ocurs during add.
     * @exception IOException if an IO error occurs.
     */
    protected StringBuffer readError(InputStream procin, StringBuffer into) 
	throws IOException, CvsException
    {
	BufferedReader in   = new BufferedReader(new InputStreamReader(procin));
	String         line = null ;

	while ((line = in.readLine()) != null) {
	    if ( into != null )
		into.append (line+"\n") ;
	}
	return into ;
    }

    /**
     * Wait for the underlying CVS process to finish.
     * Once the process is terminated, all relevant streams are closed, and
     * an exception if potentially thrown if the process indicated failure.
     * @param proc The CVS process.
     * @param ccode Should we expect a zero status from the child process.
     * @exception CvsException If a zero status is expected, and the CVS
     * @return true iff operation successful
     * process exit status is not zero.
     */

    protected synchronized boolean waitForCompletion 
	(Process proc, boolean ccode)
	throws CvsException 
    {
	while ( true ) {
	    try {
		CvsException exception = null;
		// Try reading the input stream:
		StringBuffer outputlog = new StringBuffer();
		// Try reading the error stream:
		StringBuffer errorlog = new StringBuffer() ;
		try {
		    outputlog = readText(proc.getInputStream(), outputlog) ;
		    errorlog  = readError(proc.getErrorStream(), errorlog) ;
		} catch (IOException ex) {
		    ex.printStackTrace();
		} catch (CvsException cvs_ex) {
		    exception = cvs_ex;
		} finally {
		    // Close all streams, just to make sure...
		    try { proc.getInputStream().close(); } 
		    catch (Exception ex) {Debug.out("Exception thrown by class CvsRunner at IO-close");}
		    try { proc.getOutputStream().close(); } 
		    catch (Exception ex) {Debug.out("Exception thrown by class CvsRunner at IO-close");}
		    try { proc.getErrorStream().close(); } 
		    catch (Exception ex) {Debug.out("Exception thrown by class CvsRunner at IO-close");}
		    // Check ecode if requested to do so:
		    proc.waitFor() ;
		}
		//no exception thrown, that's an unknown exception.
		int ecode = proc.exitValue() ;
		if ( ecode != 0 ) {
		    String msg = ("Process exited with error code: " + ecode
				  + " error ["+errorlog+"]");
		    if ( debug )
			mgtLogger.error(msg) ;
		    if ( ccode ) {
			if (exception == null)
			    throw new CvsException (msg) ;
			else
			    throw exception;
		    }
		    return false;
		} 
		return true;
	    } catch (InterruptedException e) 
		{ Debug.out("Exception thrown by class CvsRunner at IO");}

	}
    }




    public boolean cvsImport(String moduleName, String workingDir,
                          String vendorTag, String releaseTag)
	                                   throws CvsException {	
        initRep();
	String command = "cvs " + ROOT_OPT + " import -d "+
             " -I*.tpr -I*.tws -I*.df* -I*.txv* " +
             " -m your.message.here "+ 
             moduleName+ " " + vendorTag + " " + releaseTag;
	// Run it:
	try {
	    mgtLogger.info("Executing "+command+" in "+workingDir);
	    Process proc =  Runtime.getRuntime().exec(command, null, 
                                                      new File(workingDir));
	    return waitForCompletion(proc, false);
	} catch (IOException ex) {
	    ex.printStackTrace();
	    throw new CvsException(ex.getMessage());
	}
    }
    
    
    public void cvsLog(String moduleName) throws CvsException {
	String command = "/usr/bin/cvs " + ROOT_OPT + " rlog -h "+ moduleName;
	// Run it:
	try {
	    mgtLogger.info("Executing "+command);
	    Process proc =  Runtime.getRuntime().exec(command);
            parseLogOutput(proc.getInputStream());
	    waitForCompletion(proc, false);
	} catch (IOException ex) {
	    ex.printStackTrace();
	    throw new CvsException(ex.getMessage());
	}
    }



    public String cvsDiff(String moduleName, String tag1, String tag2) 
                                                     throws CvsException {
	String command = "/usr/bin/cvs " + ROOT_OPT + " rdiff -u "+
             " -r " + tag1 + " -r " + tag2 + " " + moduleName;
        String s = "";
	// Run it:
	try {
	    mgtLogger.info("Executing "+command);
	    Process proc =  Runtime.getRuntime().exec(command, null, 
               new File(System.getProperty("user.home")));
	    BufferedReader in   = new BufferedReader(
               new InputStreamReader(proc.getInputStream()));
            String         line = null ;
            while ((line = in.readLine()) != null) {
                // Make sure the line isn't empty:
                if ( line.length() <= 0 ) continue;
                s += line+"\n";
            }
	    waitForCompletion(proc, true);
	} catch (IOException ex) {
	    ex.printStackTrace();
	    throw new CvsException(ex.getMessage());
	}
        mgtLogger.debug("Diff:\n"+s);        
        return s;
    }


    private void parseLogOutput(InputStream procin) 
        throws IOException, CvsException {
        
        boolean collectSymbols = false;
    
	BufferedReader in   = new BufferedReader(new InputStreamReader(procin));
        String         line = null ;
        while ((line = in.readLine()) != null) {
            // Make sure the line isn't empty:
            if ( line.length() <= 0 ) continue;


            if ( collectSymbols ) {
                if ( line.startsWith( "\t" ) ) {
                    String symStr = line.substring(1);
                    int colonPos = symStr.indexOf(':');
                    symStr = symStr.substring(0,colonPos).trim();
                    mgtLogger.debug("CVS Symbols: "+ symStr );
                } else {
                    collectSymbols = false;
                }
            } else 
               if (line.startsWith( "symbolic names:" )) collectSymbols = true;

            
        }
    }
    
    
    private void initRep() throws CvsException {
       File rep = new File(REP_ROOT+File.separator+"CVSROOT");
       File anchor = new File(DOT_KEY+File.separator+"CVS_ANCHOR_DIR"+
                              File.separator+"KEY_CVS_ANCHOR");
                           
       try {
           if (!rep.exists()) {
              rep.mkdirs();
	      String command = "cvs " + ROOT_OPT + " init";
	      // Run it:
	          mgtLogger.info("Executing "+command);
	          Process proc =  Runtime.getRuntime().exec(command);
	          waitForCompletion(proc, false);
           }

//           if (!anchor.exists()) anchor.createNewFile();
       } catch (IOException ex) {
	   ex.printStackTrace();
	   throw new CvsException(ex.getMessage());
       }
    
    }






}
