// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof.mgt;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.java.JavaProgramElement;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.ArrayOfExpression;
import de.uka.ilkd.key.java.TypeConverter;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.util.*;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.reference.ConstructorReference;
import de.uka.ilkd.key.java.expression.Assignment;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.init.*;
import java.util.*;
import de.uka.ilkd.key.java.statement.MethodBodyStatement;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.java.declaration.MethodDeclaration;
import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;

public class OCLMethodLemmaFactory extends AbstractMethodLemmaFactory {
    
    private ReprModelMethod repr;
    private ModifiesParserHelper modPH = null;
    private InitConfig initConfig;
    private String pre;
    private String post;
    private String suffix;
    private ProgramElement proofStatement;
    private Modality modality;

    private static final ArrayOfQuantifiableVariable emptyvars 
	= new ArrayOfQuantifiableVariable();

    public OCLMethodLemmaFactory(ReprModelMethod repr, InitConfig initConfig,
				 String pre, String post, String suffix) {
	this.repr = repr;
	this.initConfig = initConfig;
	this.pre = pre;
	this.post = post;
	this.suffix = suffix;
	preprocessing();
    }


    protected Term getPreCond() {
	return proofPrecondition;
    }

    protected ArrayOfQuantifiableVariable getExtraVars() {
	return proofQuantVars;
    }

    protected Term getExtraPre() {
	return proofPurePrefix;
    }
    
    protected Term getPostCond() {
	return proofSuffix;
    }

    protected SetOfTerm getModifies(){
	return modElms;
    }

    protected String getSuffix() {
	return suffix;
    }
    

    protected Heuristic getHeuristic() {
	return (Heuristic)getInitConfig().heurNS().lookup
	    (new Name("use_method_specs"));
    }

    protected ProgramElement getProofStatement() {
	return proofStatement;
    }

    protected Modality getModality() {
	return modality;
    }

    private InitConfig getInitConfig() {
	return initConfig;
    }


    //-------- below the stuff to create a taclet that inserts this method spec 

    private OCLPostPO postPO = null;
    private ArrayOfQuantifiableVariable proofQuantVars = null;
    private Term proofPrecondition;
    private Term proofPurePrefix;
    private Term proofSuffix;
    private Term pi2Upd;
    private Term[] modSUTArr; 
    private QuantifiableVariable[] logicQuantVarsArr;
    private SetOfTerm modElms;
    private ArrayOfQuantifiableVariable logicQuantVars;
    private Term psiUpd = null;
    private Term addFormula = null;
    private Term schVarPsiTerm;
    private Term findTerm;
    private Map lV2SV;


    public void preprocessing() {

	if (repr.getMyModifClause()==null && !repr.isQuery()) {
	    return;
	}
	
	// gets the OCL-DL translated proof which is also already transformed
	// by the method described in Bastian's minor thesis so it can be
	// used in the way his formula construction needs it
	Term proof = createOCLtoDLTransformedProof();
	
	// there is an error in the pre or postcondition and therefore no
	// taclet can be created by the given rule
	if (proof == null) {
	    return;
	}
	
	// puts the eliminated quantifiers into proofQuantVars which
	// is an ArrayOfQuantifiableVariable
	Term nonQuantifiedProof = 
	    eliminateQuantificationRetainsQuantifyingVariables(proof);
	
	// partition the now unquantified proof
	final Term proofPrefix = nonQuantifiedProof.sub(0);
	proofPrecondition = proofPrefix.sub(0);//PRE,i.e.Phi-Pre before map
	proofPurePrefix   = proofPrefix.sub(1);  //Pi1
	
	proofSuffix = nonQuantifiedProof.sub(1).sub(0); //Pi2
	
	modality = (Modality) nonQuantifiedProof.sub(1).op();

	JavaBlock proofJavaBlock = nonQuantifiedProof.sub(1).javaBlock();
	createProofStatement(proofJavaBlock);
	
	// finds all modifies elements and returns a Set of them 
	modElms = extractModifiesElements();
    }


    private Term createOCLtoDLTransformedProof() {
	// get the OCL-DL translated Term including pre- and postconditions
	postPO = new OCLPostPO(repr.getName(), repr, pre,
			       post, repr.getActXmifile());
	postPO.setConfig(initConfig);
	NamespaceSet nss = initConfig.namespaces();
	Namespace oldProgVarsNS = nss.programVariables();
	nss.setProgramVariables(new Namespace());
	try{
	    postPO.read(ModStrategy.NO_VARS);
	}
	// in case there is an exception we couldn't parse the given pre- or
	// postcondition and thus can't create a rule, this gets taken care of 
	// just after the call to this method in the main part
	catch (de.uka.ilkd.key.ocl.ocldltranslator.OCLTranslationException e) {
	    System.out.println("Warning: Syntax error in OCL specification of "
			       +repr.getName());
	    return null;
	} finally {
	    nss.setProgramVariables(oldProgVarsNS);
	}
	Term dlProof = postPO.getPOTerm();

	// transform it to the form which is needed for the taclet generation 
	// according to Bastian's minor thesis
	Term proof = MethodSpecTransformation.getTransformedHoare(dlProof);
	
	return proof;
    }
    

    private Term eliminateQuantificationRetainsQuantifyingVariables(Term proof){
	Term nonQuantifiedProof = proof;
	ArrayOfQuantifiableVariable proofQuantVarsHelpArray = null;
	Vector proofQuantVarsVector = new Vector();
	while (nonQuantifiedProof.op() instanceof Quantifier){
	    proofQuantVarsHelpArray = nonQuantifiedProof.varsBoundHere(0);
	    proofQuantVarsVector.add(proofQuantVarsHelpArray.
				     getQuantifiableVariable(0));
	    nonQuantifiedProof = nonQuantifiedProof.sub(0);
	}
	
	Iterator proofQuantVarsVectorIterator = proofQuantVarsVector.iterator();
	QuantifiableVariable[] proofQuantVarsArray = 
	    new QuantifiableVariable[proofQuantVarsVector.size()];
	int count = 0;
	while (proofQuantVarsVectorIterator.hasNext()) {
	    proofQuantVarsArray[count] = 
		(QuantifiableVariable) proofQuantVarsVectorIterator.next();
	    count = count + 1;
	}
	proofQuantVars = new ArrayOfQuantifiableVariable(proofQuantVarsArray);
	
	return nonQuantifiedProof;
    }

    private void createProofStatement(JavaBlock proofJavaBlock) {
	StatementBlock proofStatementBlock = 
	    (StatementBlock) proofJavaBlock.program();
	proofStatement = proofStatementBlock.getChildAt(0);
    }

    private SetOfTerm extractModifiesElements() {
	ProgramVariableCollector pvc 
	    = new ProgramVariableCollector(getProofStatement());
	pvc.start();
	Namespace ns = new Namespace();
	Iterator it = pvc.result().iterator();	    
	while (it.hasNext()) {
	    ns.add((Named)it.next());
	}
	modPH = new ModifiesParserHelper(repr, initConfig);
	try {
	    modPH.parseModifiesClauseToHashSet(ns);
	}
	// in case there is an exception we couldn't parse the given modifies
	// clause and thus can't create a rule, this gets taken care of just
	// after the call to this method in the main part
	catch (Exception e) { 
	    System.out.println("Exception while parsing modifies: "+repr+"    "+ns);
	    e.printStackTrace();
	    return null;
	}
	Set localModElms = modPH.getModifiesElementsHashSet(); 
	//"ProgramElement" members
	
	// in case there is no modifies set return null
	if (localModElms == null) {
	    return null;
	}
	
	SetOfTerm localModElmsAsTerms = SetAsListOfTerm.EMPTY_SET;
	Iterator localModElmsIt = localModElms.iterator();
	Term t = null;
	while (localModElmsIt.hasNext()) {
	    t = (Term) getInitConfig().getServices().getTypeConverter()
		.convertToLogicElement((ProgramElement) localModElmsIt.next());
	    localModElmsAsTerms = localModElmsAsTerms.add(t);
	}
	
	return localModElmsAsTerms;
    }
   
}
