// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
package de.uka.ilkd.key.proof.mgt;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.logic.op.Modality;
import de.uka.ilkd.key.proof.init.ProofOblInput;
import de.uka.ilkd.key.proof.init.OCLPostPO;
import de.uka.ilkd.key.rule.*;
import java.util.Map;
import java.util.HashMap;
import de.uka.ilkd.key.proof.init.InitConfig;
import java.util.Set;
import java.util.Iterator;
import java.util.*;
import org.apache.log4j.Logger;

/** The unique environment a proof is performed in. The environment
 * consists of a java model, specifications, and a set of justified
 * rules. Since the starting point of the proofs contained in the
 * environment is equal, there is an InitConfig contained to be used
 * to start proofs of this environment.
 */
public class ProofEnvironment {

    private JavaModel jModel;
    private RuleConfig ruleConfig;
    private RuleJustificationInfo justifInfo = new RuleJustificationInfo();
    private InitConfig initConfig;
    private SpecificationRepository specRepos = new SpecificationRepository(); 
    private Set proofs = new HashSet(); //of ProofList
    private int number=0;

    private Logger mgtLogger = Logger.getLogger("key.proof.mgt");

    /** constructs a proof environment with the given initial
     * configuration of the proofs contained in the environment.
     */
    public ProofEnvironment(InitConfig initConfig) {
	this.initConfig = initConfig;
    }

    /** sets the initial configuration of the proofs contained in the
     * environment.
     */
    public void setInitConfig(InitConfig initConfig) {
	this.initConfig = initConfig;
    }

    /** retrieves the java model underlying this environment.
     */
    public JavaModel getJavaModel() {
	return jModel;
    }

    public RuleConfig getRuleConfig() {
	return ruleConfig;
    }

    /** sets the java model underlying this environment. Only to be
     * called by the {@link de.uka.ilkd.key.proof.init.ProblemInitializer}.
     */
    public void setJavaModel(JavaModel m) {
	jModel=m;
    }

    public void setRuleConfig(RuleConfig rc) {
	ruleConfig=rc;
    }

    /** adds a contract to the specifications of this environment.
     */
    public void addContract(Contract ct) {
	specRepos.add(ct);
	ct.setProofEnv(this);		
    }

    /** adds contracts to the specifications of this environment.
     */
    public void addContracts(ContractSet ct) {
	Iterator it = ct.iterator();
	while (it.hasNext()) {
	    addContract((Contract) it.next());
	}
    }

    /** retrieves an iterator on all specifications contained in this
     * environment.
     */
    public Iterator getSpecs() {
	return specRepos.getSpecs();
    }

    /** returns the object managing the rules in this environement and
     * their justifications. The object is unique to this environment. 
     */
    public RuleJustificationInfo getJustifInfo() {
	return justifInfo;
    }

    /** returns the initial configuration to be used to load proofs in
     * this environment. 
     */
    public InitConfig getInitConfig() {
	return initConfig;
    }

    /** registers a proof loaded with the given {@link 
     * de.uka.ilkd.key.proof.init.ProofObligationInput}. The proof is 
     * represented by the given {@link de.uka.ilkd.key.proof.ProofList}
     * which contains a list of proofs for this input. The method adds
     * the proof list to the suitable contract if one is found. In all
     * cases the proof is added to the proofs of this environment and the
     * proofs are marked to belong to this environment.
     */
    public void registerProof(ProofOblInput input, ProofList pl) {
	Object objOfContract = input.getObjectOfContract();
	ContractSet cset = specRepos.getContract(objOfContract);
	if (cset!=null) {
	    Iterator it = cset.iterator();
	    boolean found = false;
	    while (it.hasNext()) {
		Contract ct = (Contract) it.next();
		if (input.fitsToContract(ct)) {
		    ct.addProofList(pl);
		    found=true;
		}
	    }
 	    if (!found) { //can and should this happen?
		System.out.println(objOfContract);
 		throw new IllegalStateException("No appropriate Spec. found.");
 	    }
	}
	pl.setProofEnv(this);
	proofs.add(pl);
    }

    /** registers a rule with the given justification at the
     * justification managing {@link RuleJustification} object of this
     * environment. 
     */
    public void registerRule(Rule r, RuleJustification j) {
	justifInfo.addJustification(r, j);
    }

    public void registerRuleIntroducedAtNode(RuleApp r, Node node) {
	RuleJustification just = new RuleJustificationByAddRules(node);
	justifInfo.addJustification(r, just);
    }

    public void registerRule(RuleApp r, RuleJustification j) {
	justifInfo.addJustification(r, j);
    }

    public void addToAllProofs(NoPosTacletApp r) {
	Iterator it = proofs.iterator();
	while (it.hasNext()) {
	    ProofList pl = (ProofList)it.next();
	    for (int i=0; i<pl.size(); i++) {
		Proof p=pl.get(i);
		IteratorOfGoal goalIt = p.getSubtreeGoals(p.root()).iterator();
		while (goalIt.hasNext()) {
		    Goal g = goalIt.next();
		    g.addNoPosTacletApp(r);
		    g.updateRuleAppIndex();
		}
	    }
	}
    }

    /** registers a set of rules with the given justification at the
     * justification managing {@link RuleJustification} object of this
     * environment. All rules of the set are given the same
     * justification. 
     */
    public void registerRules(SetOfTaclet s, RuleJustification j) {
	IteratorOfTaclet it = s.iterator();
	while (it.hasNext()) {
	    registerRule(it.next(), j);
	}
    }

    /** registers a list of rules with the given justification at the
     * justification managing {@link RuleJustification} object of this
     * environment. All rules of the list are given the same
     * justification. 
     */
    public void registerRules(ListOfBuiltInRule s, RuleJustification j) {
	IteratorOfBuiltInRule it = s.iterator();
	while (it.hasNext()) {
	    Rule r=it.next();
	    registerRule(r, j);
	}
    }

    /** retrieves all proofs registered at this environment 
     */
    public Set getProofs() {
	return proofs;
    }

    public void removeProofList(ProofList pl) {
	proofs.remove(pl);
	specRepos.removeProofList(pl);
    }


    /** returns true iff the java model equals those of the argument
     * proof environment. TODO: extend to available rules and specs.
     */
    public boolean equals(Object cmp) {
 	if (!(cmp instanceof ProofEnvironment)) {
 	    return false;
 	}
 	ProofEnvironment pe = (ProofEnvironment) cmp;
 	return pe.getJavaModel().equals(getJavaModel())&&
	    pe.getRuleConfig().equals(getRuleConfig()) &&
	    pe.getNumber()==(getNumber());
    }

    public int hashCode() {
	int result = 5;
	result = result*17+ getJavaModel().hashCode();
	result = result*17+ getRuleConfig().hashCode();
	result = result*17+ getNumber();
	return result;
    }

    /** returns a short String description of the environment.
     */
    public String description() {
	return "Env. with "+getJavaModel().description()+" #"+getNumber();
    }

    public void updateProofStatus() {
	Set allProofs = getProofs();
	Iterator allProofsIt = allProofs.iterator();
	ProofList  pl = null;
	while (allProofsIt.hasNext()) {
	    pl = (ProofList) allProofsIt.next();
	    pl.updateProofStatus();
	}
    }

    public String getPureDiff(ProofEnvironment pe) {
	CvsRunner cvs = new CvsRunner();
	String diff="";
	if (!getJavaModel().isEmpty()) {
	    try{
		diff = cvs.cvsDiff(getJavaModel().getCVSModule(),
				   pe.getJavaModel().getModelTag(), 
				   getJavaModel().getModelTag());
	    } catch(CvsException cvse) {
		mgtLogger.error("Diffing models in CVS failed: "+cvse);
	    }
	}
	return diff;
    }

    public String getRuleDiff(ProofEnvironment pe) {
	return "Rules: \n    Earlier: " +pe.getRuleConfig().description()
	    +"\n    Now: "+getRuleConfig().description();
    }

    public String getDiffUserInfo(ProofEnvironment pe) {
	String base = "Comparing "+description()+" with "+pe.description()+":\n";
	if (!getJavaModel().getModelDir().equals
	    (pe.getJavaModel().getModelDir())) {
	    return base+"No Diff for different model directories: \n"
		+getJavaModel().getModelDir()+" \n and "
		+pe.getJavaModel().getModelDir();
	}
	return base+getPureDiff(pe) +"\n"+ getRuleDiff(pe);
    }

    /** sets a number that distinguishes two proof environments
     * with equal java model and rule config from the user perspective
     */
    public void setNumber(int number) {
	this.number = number;
    }

    /** returns a number that distinguishes two proof environments
     * with equal java model and rule config from the user perspective
     */
    public int getNumber() {
	return number;
    }

    public String toString() {
	return description();
    }    

}
