// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/**
 * tests if match checks the variable conditions in Taclets. 
 */
package de.uka.ilkd.key.proof.mgt;


import de.uka.ilkd.key.util.Debug;

import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.init.*;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.statement.MethodBodyStatement;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.rule.*;


import junit.framework.*;

import java.io.*;

public class TestAbstractMethodLemmaFactory extends TestCase {

    private class TestMethodLemmaFactory extends AbstractMethodLemmaFactory {
	public TestMethodLemmaFactory() {}
	
        /** precondition which is called "PRE" by Sebastian
	 * Katz's Studienarbeit */
	protected Term getPreCond(){
	    Function pre_func = new Function(new Name("pre"), Sort.FORMULA, new Sort[0]);
	    Term pre = tF.createFunctionTerm(pre_func);
	    return pre;
	}
	
	/** extra quantifiable variables created by the transformation necessary
	 * after the Studienarbeit of Sebastian Katz "transformation_vars" */
	protected ArrayOfQuantifiableVariable getExtraVars(){
	    LogicVariable a = new LogicVariable(new Name("logic_var_a"), Sort.NULL);
	    LogicVariable b = new LogicVariable(new Name("logic_var_b"), Sort.NULL);
	    QuantifiableVariable[] qva = new QuantifiableVariable[2];
	    qva[0] = a;
	    qva[1] = b;
	    ArrayOfQuantifiableVariable qVarArr = new ArrayOfQuantifiableVariable(qva);
	    return qVarArr;
	}
	
	/** extra precondtion created by the transformation necessary
	 * after the Studienarbeit of Sebastian Katz 
	 * there this is called "pi1" */
	protected Term getExtraPre(){
	    Function extraPre_func = new Function(new Name("extra_pre"), Sort.FORMULA, new Sort[0]);
	    Term extraPre = tF.createFunctionTerm(extraPre_func);
	    return extraPre;
	}
	
	/** post condition of the method spec
	 * this is called "pi2" in Sebastian Katz's Studienarbeit */
	protected Term getPostCond(){
	    Function post_func = new Function(new Name("post"), 
					      Sort.FORMULA, new Sort[0]);
	    Term post = tF.createFunctionTerm(post_func);
	    return post;
	}
	
	/** modifies elements as a set of terms */
	protected SetOfTerm getModifies(){
	    KeYJavaType kjt = new KeYJavaType(Sort.NULL);
	    ProgramVariable modifies_var1 = 
		new ProgramVariable(new ProgramElementName("mod_elm_1"), 
				    kjt, kjt, false, true);
	    Term modifies1 = tF.createVariableTerm(modifies_var1);
	    ProgramVariable modifies_var2 = 
		new ProgramVariable(new ProgramElementName("mod_elm_2"), 
				    kjt, kjt, false, true);
	    Term modifies2 = tF.createVariableTerm(modifies_var2);
	    
	    SetOfTerm sot = SetAsListOfTerm.EMPTY_SET;
	    sot = sot.add(modifies1);
	    sot = sot.add(modifies2);
	    return sot;
	}
	
	/** statement before the transformation so we can get the real
	 * variables (no schema variables) */
	protected ProgramElement getProofStatement(){
	    return new MethodBodyStatement
		(new ProgramMethod
		 (new MethodDeclaration(new Modifier[0], null, null, 
					(ArrayOfParameterDeclaration) null,
					null, null, true), 
		  new KeYJavaType(Sort.NULL), new KeYJavaType(Sort.NULL), null ),
		 new ProgramVariable(new ProgramElementName("x"), (Sort) null), 
		 new ProgramVariable(new ProgramElementName("y"), (Sort) null), 
		 new ArrayOfExpression());
	}
	
	/** string which is printed out a couple of times and given in the
	 * constructor of the classes that extend this abstract class */
	protected String getSuffix(){
	    return new String("suffix");
	}
	
	/** initconfig is needed here */
	protected InitConfig getInitConfig(){
	    InitConfig iC = new InitConfig();
	    iC.setNamespaces(new NamespaceSet(null, null, null, new Namespace(), null, null));
	    return iC;
	}

	protected Heuristic getHeuristic() {
	    return (Heuristic)getInitConfig().heurNS().lookup
		(new Name("use_method_specs"));
	}

	protected Modality getModality() {
	    return Op.DIA;
	}
	
    };
    
    
    public TestAbstractMethodLemmaFactory(String name) {
	super(name);
    }

    public TestAbstractMethodLemmaFactory(String name, boolean b) {
	super(name);
	Debug.ENABLE_DEBUG = b;
    }



    private TestMethodLemmaFactory tmlf;
    private TermFactory tF = TermFactory.DEFAULT;
    


    public void setUp() {	
	tmlf = new TestMethodLemmaFactory();

    }

    public void testFind() {
	Taclet tac = tmlf.getTaclet();
	assertTrue("Taclet is instance of SuccTaclet by construction", 
		   tac instanceof SuccTaclet);
	Term findTerm = ((SuccTaclet) tac).find();
	assertTrue("Diamond operator as toplevel operator required", 
		   findTerm.op() == Op.DIA);
	assertTrue("Post condition is a schema variable", 
		   findTerm.sub(0).op() instanceof SchemaVariable);
	
	
	JavaBlock findJB = findTerm.javaBlock();
	StatementBlock findSB = ((ContextStatementBlock) findJB.program()).getMidBody();
	ProgramElement findPE = findSB.getChildAt(0);
	
	assertTrue("Inner term of the findTerm is a mbs", 
		   findPE instanceof MethodBodyStatement);
	
    }

    public void testReplaceWith() {
	Taclet tac = tmlf.getTaclet();
	ListOfTacletGoalTemplate ltgt = tac.goalTemplates();
	AntecSuccTacletGoalTemplate firstTGT = (AntecSuccTacletGoalTemplate) ltgt.head();
	AntecSuccTacletGoalTemplate secondTGT = (AntecSuccTacletGoalTemplate) ltgt.tail().head();
	// as there are only 2 TGTs this is enough.
	Term first = firstTGT.replaceWith().succedent().getFirst().formula();
	Term second = secondTGT.replaceWith().succedent().getFirst().formula();

	// some tests necessary to be put here!
    }
    
    
    public static void main(String args[]) {
	
    }
    
}
