// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.SortedSchemaVariable;
import de.uka.ilkd.key.logic.op.SetOfSchemaVariable;
import de.uka.ilkd.key.logic.op.IteratorOfSchemaVariable;
import de.uka.ilkd.key.logic.op.MapFromSchemaVariableToTacletPrefix;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import java.util.Vector;

/** 
 * An abstract class to represent Taclets with a find part. This means, they
 * have to be attached to a formula or term of the sequent. This class is
 * extended by several subclasses to distinct between taclets that have to
 * attached to a top level formula of the antecedent ({@link AntecTaclet}),
 * to the succedent ({@link SuccTaclet}) or to an arbitrary term that matches
 * the find part somewhere in the sequent ({@link RewriteTaclet}).  
 */ 
public abstract class FindTaclet extends Taclet {

    /** contains the find term */
    protected Term find;

    /** Set of schemavariables of the if and the (optional) find part */
    private SetOfSchemaVariable ifFindVariables = null;

    /** creates a FindTaclet 
     * @param name the Name of the taclet
     * @param applPart the TacletApplPart that contains the if-sequent, the
     * not-free and new-vars conditions 
     * @param goalTemplates a ListOfTacletGoalTemplate that contains all goaltemplates of
     * the taclet (these are the instructions used to create new goals when
     * applying the Taclet)
     * @param heuristics a ListOfHeuristic that contains all heuristics the Taclet
     *      is atteached to
     * @param constraint the Constraint of the Taclet (has to be fulfilled in
     * order to achieve this Taclet)
     * @param attrs the TacletAttributes encoding if the Taclet is non-interactive,
     * recursive or something like that
     * @param find the Term that is the pattern that has to be found in a
     * sequent and the places where it matches the Taclet can be applied
     * @param prefixMap a MapFromSchemaVariableToTacletPrefix that contains the
     * prefix for each SchemaVariable in the Taclet
     */
    public FindTaclet(Name name, TacletApplPart applPart,  
		      ListOfTacletGoalTemplate goalTemplates, 
		      ListOfHeuristic heuristics,
		      Constraint constraint, TacletAttributes attrs,
		      Term find,
		      MapFromSchemaVariableToTacletPrefix prefixMap,
		      SetOfChoice choices){
	super(name, applPart, goalTemplates, 
	      heuristics, constraint, attrs,
	      prefixMap, choices);
	this.find = find;
    }
    
    public Term find() {
	return find;
    }

    /** this method is used to determine if top level updates are
     * allowed to be ignored. This is the case if we have an Antec or
     * SuccTaclet but not for a RewriteTaclet
     * @return true if top level updates shall be ignored 
     */
    protected abstract boolean ignoreTopLevelUpdates();

    /** checks if term and rule fit in some particular conditions 
     * @param services the Services object encapsulating information
     * about the java datastructures like (static)types etc.
     *
     * PRECONDITION: matchCond.getConstraint ().isSatisfiable ()
     */
    public MatchResult matchFind(Term term,
				 MatchConditions matchCond,
				 Services services,
				 Constraint userConstraint) {
        matchCond = match(term, find(), 
        		  ignoreTopLevelUpdates(),
        		  matchCond, services, userConstraint);
        return ((matchCond != null) ? new MatchResult(matchCond) : 
        	MatchResult.FAILED); 
    }

    /** checks if a given schema variable is used in the rule this implies
     * in the ifseq, findexp, rwexp, addseq or addrule
     * @param var the SchemaVariable to be looked for
     * @return true iff SchemaVariable var occurs in the Taclet
     */
    protected boolean lookup(SchemaVariable var) {
	return (lookupInTerm(var,find()) || lookupImpl(var));
    }

    /** checks if a variable occurrs maximal in one instantiation area
     * @param var the SchemaVariable whose instantiation is checked
     * @return true iff instantiation is correct
     */
    protected boolean correctInstantiationAreaOccurrence(SchemaVariable var) {
	// here for findexp == sequent, overwritten in RewriteTaclet
	// ifseq, findexp, all addseq and all rwexp are one area
	Boolean correct = areaCache.get(var);
	if ( correct != null ) {
	    return correct.booleanValue();
	}

	boolean found = false;
	if ( lookupInTerm ( var, find() ) || lookupInAreaOne( var ) ) {
	    found = true;
	}

	if  ( !lookupInAddRulesArea ( var, found ) ) {
	    areaCache.put ( var, new Boolean ( true ) );
	    return true;
	}

	areaCache.put ( var, new Boolean ( false ) );	    
	return false;
    }


    /** CONSTRAINT NOT USED 
     * applies the replacewith part of Taclets
     * @param gt TacletGoalTemplate used to get the replaceexpression 
     * in the Taclet
     * @param goal the Goal where the rule is applied
     * @param posOfFind the PosInOccurrence belonging to the find expression
     * @param services the Services encapsulating all java information
     * @param svInst a SVInstantiations object containing the necessary
     * instantiations, so that the find term matches 
     */
    protected abstract void applyReplacewith(TacletGoalTemplate gt, Goal goal,
					     PosInOccurrence posOfFind,
					     Services services,
					     MatchConditions matchCond);


    /**
     * adds the sequent of the add part of the Taclet to the goal sequent
     * @param add the Sequent to be added
     * @param goal the Goal to be updated
     * @param posOfFind the PosInOccurrence describes the place where to add
     * the semisequent 
     * @param services the Services encapsulating all java information
     * @param svInst a SVInstantiations object containing the necessary
     * instantiations, so that the find term matches 
     */
    protected abstract void applyAdd(Sequent add, Goal goal,
				     PosInOccurrence posOfFind,
				     Services services,
				     MatchConditions matchCond);


    /**  
     * the rule is applied on the given goal using the
     * information of rule application. 
     * @param goal the goal that the rule application should refer to.
     * @param services the Services encapsulating all java information
     * @param ruleApp the taclet application that is executed.
     * @param proveIfSequent a boolean that is true if a goal has to
     * be created in order to prove that the sequent in the if part of
     * the Taclet can be derived and therefore the application of the
     * rule is correct (normally true when the if-sequent does not match)
     */
    public ListOfGoal apply(Goal     goal,
			    Services services,
			    RuleApp  ruleApp) {

	// Number without the if-goal eventually needed
	int                          numberOfNewGoals = goalTemplates().size();

	TacletApp                    tacletApp        = (TacletApp) ruleApp;
	MatchConditions              mc               = tacletApp.matchConditions ();

	// Restrict introduced metavariables to the subtree
	setRestrictedMetavariables ( goal, mc );

	ListOfGoal                   newGoals         =
	    checkIfGoals ( goal,
			   tacletApp.ifFormulaInstantiations (),
			   mc,
			   numberOfNewGoals );
	
	IteratorOfTacletGoalTemplate it               = goalTemplates().iterator();
	IteratorOfGoal               goalIt           = newGoals.iterator();

	while (it.hasNext()) {
	    TacletGoalTemplate gt          = it    .next();
	    Goal               currentGoal = goalIt.next();
	    // add first because we want to use pos information that
	    // is lost applying replacewith
	    applyAdd(         gt.sequent(),
			      currentGoal,
			      tacletApp.posInOccurrence(),
			      services,
			      mc );

	    applyReplacewith( gt,
			      currentGoal,
			      tacletApp.posInOccurrence(),
			      services,
			      mc );

	    applyAddrule(     gt.rules(),
			      currentGoal,
			      services,
			      mc );

	    applyAddProgVars( gt.addedProgVars(),
			      currentGoal, 
			      mc);
                              
            currentGoal.setBranchLabel(gt.name());
	}

	return newGoals;
    }

    /**
     * returns true iff the given variable schema variable is bound either 
     * in the ifseq or in
     * any part of the TacletGoalTemplates or in the find part.
     * @param v the bound variable to be searched 
     */
    protected boolean varIsBound(SchemaVariable v) {
	BoundVarsVisitor bvv=new BoundVarsVisitor();
	find().execPostOrder(bvv);
	if (bvv.getBoundVariables().contains((SortedSchemaVariable)v)) {
	    return true;
	}	    
	return super.varIsBound(v);
    }


    /**
     * returns pairs of variables (x,y) where y is bound in this Taclet and x
     * occurs in the scope of y; (y,y) is part of the set for all bound ys
     * with no x bound in the scope of y.
     * @return a set of variable pairs as specified above
     */
    protected SetOfInScopePair getVarInScope() {
	VarInScopeCollector bc=new VarInScopeCollector();
	find().execPostOrder(bc);
	return super.getVarInScope().union(bc.getVarInScope());
    }


    StringBuffer toStringFind(StringBuffer sb) {
	return sb.append("find(").
	    append(find().toString()).append(")\n");
    }


    /**
     * returns a representation of the Taclet with find part as String
     * @return string representation
     */
    public String toString() {
	if (tacletAsString == null) {
	    StringBuffer sb = new StringBuffer();
	    sb = sb.append(name()).append(" {\n");
	    sb = toStringIf(sb);
	    sb = toStringFind(sb);
	    sb = toStringVarCond(sb);
	    sb = toStringGoalTemplates(sb);
	    sb = toStringHeuristics(sb);
	    sb = toStringAttribs(sb); 
	    tacletAsString = sb.append("}").toString();
	}
	return tacletAsString;
    }


    /**
     * @return Set of schemavariables of the if and the (optional)
     * find part
     */
    public SetOfSchemaVariable getIfFindVariables () {
	if ( ifFindVariables == null ) {
	    TacletVariableCollector svc = new TacletVariableCollector ();
	    find ().execPostOrder ( svc );
	    
	    ifFindVariables             = getIfVariables ();
	    IteratorOfSchemaVariable it = svc.varIterator ();
	    while ( it.hasNext () )
		ifFindVariables = ifFindVariables.add ( it.next () );
	}

	return ifFindVariables;
    }


    protected Taclet setName(String s, FindTacletBuilder b) {
	return super.setName(s, b); 
    }

    /**
     * Determine whether a replacewith-part is really supposed to modify
     * concerned formulas when applying a taclet (otherwise copies of the
     * formulas are created). Copies are created whenever the constraint of the
     * new formulas is stronger than the conjunction of the original
     * find-formula constraint and the user constraint. (Taking the user
     * constraint into account at this point ensures that the existence of the
     * user constraint is transparent to the user, and is a temporary solution
     * to cope with the absence of disunification constraints).
     * 
     * @param goal
     *            the goal to which the taclet is applied
     * @param posOfFind
     *            position of the find-formula (which is the formula whose
     *            constraint is considered as the original formula constraint)
     * @param matchCond
     *            results of matching the taclet, in particular the new
     *            constraint that is attached to formulas added by replacewith
     * @return true iff replacewith is supposed to create copies of concerned
     *         formulas
     */
    protected boolean createCopies (Goal goal,
                                    PosInOccurrence posOfFind,
                                    MatchConditions matchCond) {
        final Constraint userConstraint =
            goal.node ().proof ().getUserConstraint ().getConstraint ();
        final Constraint oriConstraint =
            posOfFind.constrainedFormula ().constraint ();
        final Constraint combinedConstraint =
            userConstraint.join ( oriConstraint );
        final Constraint newConstraint = matchCond.getConstraint ();
        return !newConstraint.isAsWeakAs ( combinedConstraint );
    }
    
} 
