// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.logic.op.SVSubstitute;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.SortedSchemaVariable;
import de.uka.ilkd.key.logic.sort.GenericSort;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.rule.inst.GenericSortCondition;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.rule.inst.SortException;
import de.uka.ilkd.key.util.Debug;


/**
 * Variable condition that enforces a given generic sort to be instantiated with
 * the sort of a program expression a schema variable is instantiated with
 * (currently only schema variables <code>ProgramSVSort.EXPRESSION</code> are
 * supported)
 */
public class JavaTypeToSortCondition implements VariableCondition {

    private final SchemaVariable exprSV;
    private final GenericSort    sort;
    
    public JavaTypeToSortCondition (final SchemaVariable exprSV,
                                    final GenericSort sort) {
        this.exprSV = exprSV;
        this.sort = sort;
        
        if ( !( exprSV instanceof SortedSchemaVariable
                && ( (SortedSchemaVariable)exprSV ).sort () ==
                   ProgramSVSort.EXPRESSION ) ) {
            throw new RuntimeException
                ( "Expected a program schemavariable for expressions" ); 
        }
    }
    
    /**
     * checks if the condition for a correct instantiation is fulfilled
     * 
     * @param var
     *            the template Variable to be instantiated
     * @param matchCond
     *            the MatchCondition with the current matching state and in
     *            particular the SVInstantiations that are already known to be
     *            needed
     * @param services
     *            the program information object
     * @return modified match results if the condition can be satisfied, or
     *         <code>null</code> otherwise
     */
    public MatchConditions check (SchemaVariable var,
                                  SVSubstitute svSubst,
                                  MatchConditions matchCond,
                                  Services services) {
        if ( var != exprSV )
            return matchCond;
        
        Debug.assertTrue ( svSubst instanceof Expression );
        
        final Expression expr = (Expression)svSubst;
        final SVInstantiations inst = matchCond.getInstantiations ();
        final KeYJavaType type =
            expr.getKeYJavaType ( services, inst.getExecutionContext () );
        
        try {
            return matchCond.setInstantiations ( inst.add
                ( GenericSortCondition.createIdentityCondition
                                                  ( sort, type.getSort () ) ) );
        } catch ( SortException e ) {
            return null;
        }        
    }

    public String toString () {
        return "" + exprSV + " hasSort " + sort;
    }
}
