// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule;

import java.util.HashMap;


/**
 * General interface for prioritization of the available heuristics
 */
public interface Prioritization {

    /**
     * Give the objecs an ordering (namely a strict half ordering)
     *
     * @return true iff object <code>p_a</code> should be considered
     * earlier than object <code>p_b</code>
     */
    boolean precedes ( PrioritizedObject p_a,
			      PrioritizedObject p_b );

    /**
     * Sort the given list of objects topologically, using the
     * relation <code>precedes</code>; remove objects occurring more
     * than once in <code>p_hs</code>
     */
    ListOfPrioritizedObject sortTopologically
	( ListOfPrioritizedObject p_hs );

    /**
     * Add the priority information <code>p</code> contains to the
     * information of this object
     */
    void add ( Prioritization p );

    Object clone ();


    /**
     * A simple implementation, using numerical priorities for the
     * objects, thus ordering totally (quasi)
     */
    class StaticPrioritization
	implements Prioritization {

	/**
	 * Priorities of the objects
	 *
	 * Keys: <code>PrioritizedObject</code>
	 *
	 * Values: <code>Integer</code>
	 */
	private HashMap priorities;

	private StaticPrioritization ( HashMap p_priorities ) {
	    priorities = p_priorities;
	}

	public StaticPrioritization () {
	    this ( new HashMap () );
	}

	public void    setPriority ( PrioritizedObject p_object,
				     int               p_priority ) {
	    priorities.put ( p_object, new Integer ( p_priority ) );
	}

	public boolean precedes    ( PrioritizedObject p_a,
				     PrioritizedObject p_b ) {
	    Integer a = (Integer)priorities.get ( p_a );
	    Integer b = (Integer)priorities.get ( p_b );
	    
	    // PrioritizedObjects not having a priority are always the
	    // least important
	    if ( a == null )
		return false;
	    else
		return b == null || a.intValue () < b.intValue ();
	}

	/**
	 * This method should really be part of an abstract superclass
	 */
	public ListOfPrioritizedObject sortTopologically ( ListOfPrioritizedObject p_hs ) {
	    ListOfPrioritizedObject     res     = SLListOfPrioritizedObject.EMPTY_LIST;
	    IteratorOfPrioritizedObject it;
	    PrioritizedObject           curMax;
	    PrioritizedObject           tMax;
	    ListOfPrioritizedObject     tList;

	    // This could be implemented more efficiently (for this
	    // special kind of prioritization) by using bucket sort;
	    // instead we implement a general algorithm for sorting
	    // topologically that works for all strict half orderings
	    while ( p_hs != SLListOfPrioritizedObject.EMPTY_LIST ) {
		// search for a least important element
		it      = p_hs.iterator ();
		curMax  = it.next ();

		if ( res.contains ( curMax ) ) {
		    p_hs = p_hs.tail ();
		    continue;
		}

		tList   = SLListOfPrioritizedObject.EMPTY_LIST;

		while ( it.hasNext () ) {
		    tMax = it.next ();
		    if ( !res.contains ( tMax ) ) {
			if ( precedes ( curMax, tMax ) ) {
			    tList  = tList.prepend ( curMax );
			    curMax = tMax;
			} else
			    tList  = tList.prepend ( tMax );
		    }
		}

		res     = res.prepend ( curMax );
		p_hs    = tList;
	    }

	    return res;
	}

	public void add ( Prioritization p ) {
	    if ( p instanceof StaticPrioritization )
		priorities.putAll ( ((StaticPrioritization)p).priorities );
	}

	public Object clone () {
	    return new StaticPrioritization ( (HashMap)priorities.clone () );
	}

    }

}
