// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;

/** this class inherits from TacletGoalTemplate. It is used if there is a
 * replacewith in the ruleGoals that replaces a term by another
 * term. For a sequent {@see de.uka.ilkd.key.proof.AnteSuccTacletGoalTemplate}
 */
public class RewriteTacletGoalTemplate extends TacletGoalTemplate {
    /** term that replaces another one */
    private Term replacewith;

    /** creates new Goaldescription 
     *@param addedSeq new Sequent to be added
     *@param addedRules ListOfTaclet contains the new allowed rules
     * at this branch 
     *@param ruleVars ListOfRulevariable contains used rulevariables 
     *@param replacewith the Term that replaces another one
     */
    public RewriteTacletGoalTemplate(Sequent             addedSeq,
				     ListOfTaclet        addedRules,
				     ListOfRulevariable  ruleVars,
				     Term                replacewith,
				     SetOfSchemaVariable pvs) {
	super(addedSeq, addedRules, ruleVars, pvs);
	TacletBuilder.checkContainsFreeVarSV(replacewith, "replacewith term");
	this.replacewith = replacewith;
    }

    public RewriteTacletGoalTemplate(Sequent addedSeq,
			   ListOfTaclet addedRules,
			   ListOfRulevariable ruleVars,
			   Term replacewith) {
	this(addedSeq, addedRules, ruleVars, replacewith,
	     SetAsListOfSchemaVariable.EMPTY_SET);
    }


    /** a Taclet may replace a Term by another. The new Term is returned.     
     * @return Term being paramter in the rule goal replacewith(Seq)
     */

    public Term replaceWith() {
	return replacewith;
    }
    
    /**
     * @return Term being paramter in the rule goal replacewith(term)
     */
    Object replaceWithExpressionAsObject() {
	return replacewith;
    }


   /**
     * returns true iff the given schema variable 
     * is bound in the sequent part of
     * this goal sequent or in an added rule or in the replacewith sequent.
     * @param v the bound variable to be searched
     */
    boolean varIsBound(SchemaVariable v) {
	if (!(v.isVariableSV())) {
	    return false;
	}
	BoundVarsVisitor bvv=new BoundVarsVisitor();
	replaceWith().execPostOrder(bvv);
	if (bvv.getBoundVariables().contains((SortedSchemaVariable)v)) {
	    return true;	    
	}
	return super.varIsBound(v);
    } 



    /**
     * returns pairs of variables (x,y) where y is bound in this
     * TacletGoalTemplate and x 
     * occurs in the scope of y; (y,y) is part of the set for all bound ys
     * with no x bound in the scope of y. x is a LogicVariableRepresentative 
     * while y is a QuantifiableVariable (as it is bound)
     * @return a set of variable pairs as specified above
     */
    protected SetOfInScopePair getVarInScope() {
	SetOfInScopePair result
	    =SetAsListOfInScopePair.EMPTY_SET;
	VarInScopeCollector bc=new VarInScopeCollector();
	replaceWith().execPostOrder(bc);
	return super.getVarInScope().union(bc.getVarInScope());
    }

    public boolean equals(Object o) {
	if ( ! ( o instanceof RewriteTacletGoalTemplate ) )
	    return false;
	RewriteTacletGoalTemplate other = (RewriteTacletGoalTemplate) o;

	return super.equals(other)
	    && replacewith.equals(other.replacewith);
    }
    
    public int hashCode(){
    	int result = 17;
    	result = 37 * result * super.hashCode();
    	result = 37 * result * replacewith.hashCode();
    	return result;
    }

    
    /** toString */
    public String toString() {
	String result=super.toString();
	result+="replacewith("+replaceWith()+") ";       
	return result;
    }

}
