// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


/**
 * visitor for <t> execPostOrder </t> of 
 * {@link de.uka.ilkd.key.logic.Term}. Called with that method
 * on a term, the visitor builds a new term replacing SchemaVariables with their
 * instantiations that are given as a SVInstantiations object.
 */
package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.java.reference.ExecutionContext;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.SortDefiningSymbols;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.java.visitor.ProgramContextAdder;
import de.uka.ilkd.key.java.visitor.ProgramReplaceVisitor;
import de.uka.ilkd.key.logic.Visitor;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.java.TypeConverter;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.Constraint;
import de.uka.ilkd.key.logic.op.Metavariable;
import de.uka.ilkd.key.logic.MetavariableDeliverer;
import de.uka.ilkd.key.logic.op.LogicVariable;
import de.uka.ilkd.key.logic.Name;

//import de.uka.ilkd.key.util.SimpleStack;
import java.util.Stack;

public class SyntacticalReplaceVisitor extends Visitor { 	

    private final SVInstantiations svInst;
    private final Constraint metavariableInst;
    private MapFromSchemaVariableToTerm newInstantiations =
                                MapAsListFromSchemaVariableToTerm.EMPTY_MAP;
    private final boolean forceSVInst;
    private final Name svInstBasename;
    private Services services;
    private Term computedResult = null;
    private TypeConverter typeConverter = null;
    private final boolean allowPartialReplacement;
    private final boolean resolveSubsts;

    /**
     * the stack contains the subterms that will be added in the next step of
     * execPostOrder in Term in order to build the new term. A boolean value
     * between or under the subterms on the stack indicate that a term using
     * these subterms should build a new term instead of using the old one,
     * because one of its subterms has been built, too.
     */
    private final Stack subStack; //of Term (and Boolean)
    private final TermFactory tf = TermFactory.DEFAULT;
    private final Boolean newMarker = new Boolean(true);


    /**
     * @param forceSVInst iff true instantiate uninstantiated SVs by
     * creating new metavariables or new bound logicvariables
     */
    public SyntacticalReplaceVisitor(Services              services, 
				     SVInstantiations      svInst,
				     Constraint            metavariableInst,
				     boolean               forceSVInst,

				     Name                  svInstBasename,
				     boolean               allowPartialReplacement,
				     boolean               resolveSubsts) { 
	this.services         = services;	
	this.svInst           = svInst;
	this.metavariableInst = metavariableInst;
	this.forceSVInst      = forceSVInst;
	this.svInstBasename   = svInstBasename;
	this.allowPartialReplacement = allowPartialReplacement;
	this.resolveSubsts    = resolveSubsts;
	subStack = new Stack(); // of Term
    }

    public SyntacticalReplaceVisitor(Services              services, 
				     SVInstantiations      svInst,
				     Constraint            metavariableInst,
				     boolean               forceSVInst,
				     Name                  svInstBasename) {
	this ( services, 
	       svInst,
	       metavariableInst,
	       forceSVInst,
	       svInstBasename,
	       false, true );
    }

    public SyntacticalReplaceVisitor(Services services, 
				     SVInstantiations svInst) { 
	this ( services, svInst, Constraint.BOTTOM, false, null, false, true );
    }

    public SyntacticalReplaceVisitor(Services services, 
				     Constraint metavariableInst) { 
	this ( services,
	       SVInstantiations.EMPTY_SVINSTANTIATIONS,
	       metavariableInst, 
	       false, null, false, true );
    }

    public SyntacticalReplaceVisitor(Constraint metavariableInst) { 
	this ( null, metavariableInst );
    }

    private JavaProgramElement addContext(StatementBlock pe) {
	final ContextInstantiationEntry cie = 
	    svInst.getContextInstantiation();
	if (cie == null) {
	    throw new IllegalStateException("Context should also "
					    +"be instantiated");
	}	

	if (cie.prefix() != null) {
	    return ProgramContextAdder.INSTANCE.start
		((JavaNonTerminalProgramElement)cie.contextProgram(), 
		 pe,
		 (ContextStatementBlockInstantiation)cie.getInstantiation());
	} 

	return pe;
    }

    private Services getServices () {
	if ( services == null )
	    services = new Services ();
	return services;
    }

    private TypeConverter getTypeConverter () {
	if ( typeConverter == null )
	    typeConverter = getServices ().getTypeConverter();
	return typeConverter;
    }

    private JavaBlock replacePrg(SVInstantiations svInst, JavaBlock jb) {
	if ( svInst == SVInstantiations.EMPTY_SVINSTANTIATIONS ) {
	    return jb;
	}
	ProgramReplaceVisitor trans;
	ProgramElement result = null;

	if (jb.program() instanceof ContextStatementBlock) {
	    trans = new ProgramReplaceVisitor
		(((ContextStatementBlock)jb.program()).getChildAt(0),
		 getServices (),
		 svInst,
		 allowPartialReplacement);
	    trans.start();
	    result = addContext((StatementBlock)trans.result());
	} else {
	    trans = new ProgramReplaceVisitor(jb.program(),
					      getServices (),
					      svInst,
					      allowPartialReplacement);
	    trans.start();
	    result = trans.result();
	}
	return (result==jb.program()) ? 
            jb : JavaBlock.createJavaBlock((StatementBlock)result);
    }

    private Term[] neededSubs(int n) {
	boolean newTerm = false;
	Term[] result   = new Term[n];
	for (int i = n-1; i >= 0; i--) {
	    Object top = subStack.pop();
	    if (top == newMarker){
		newTerm = true; 
		top     = subStack.pop();
	    }
	    result[i] = (Term) top;
	}
	if (newTerm && (subStack.empty() || 
			subStack.peek() != newMarker) ) {
	    subStack.push(newMarker);
	}
	return result;
    }

    private void pushNew(Object t) {
	if (subStack.empty() || subStack.peek() != newMarker) {
	    subStack.push(newMarker);
	}
	subStack.push(t);	
    }

    private Term toTerm(Object o) {
	if (o instanceof Term) {
	    return (Term)o;
	} else if (o instanceof ProgramElement) {	    
	    ExecutionContext ec
		= (svInst.getContextInstantiation()==null)
		? null 
		: svInst.getContextInstantiation()
		               .activeStatementContext();
	    return getTypeConverter().
		convertToLogicElement((ProgramElement)o, ec);
	}
	de.uka.ilkd.key.util.Debug.fail("Wrong instantiation in SRVisitor");
	return null;
    }

    public void visit(Term visited) {
	// Sort equality has to be ensured before calling this method
	if (visited.op() instanceof SortedSchemaVariable && 
	    svInst.isInstantiated((SchemaVariable)visited.op())) {
	    pushNew(toTerm(svInst.getInstantiation
			       ((SchemaVariable)visited.op())));
	} else if ( forceSVInst &&
		    visited.op() instanceof SortedSchemaVariable && 
		    ((SchemaVariable)visited.op()).isTermSV () &&
		    instantiateWithMV ( visited ) ) {
	    // then we are done ...	
	} else if ( ( visited.op() instanceof Metavariable ) &&
		    metavariableInst
		    .getInstantiation ( (Metavariable)visited.op () ) != visited.op() ) {
	    pushNew ( metavariableInst
		      .getInstantiation ( (Metavariable)visited.op () ) );
	}  else if (visited.op() instanceof ExpressionOperator) {
	    ExpressionOperator exprOp = (ExpressionOperator) visited.op();
	    pushNew(exprOp.resolveExpression(svInst, getServices()));
	} else {
	    Operator op = visited.op();

	    if ( op instanceof SortDependingSymbol )
                op = handleSortDependingSymbol ( op );

	    boolean operatorInst=false;
	    
	    if (op instanceof ModalOperatorSV){
		ModalOperatorSV modalsv = (ModalOperatorSV)op;
		op = (Modality) svInst.getInstantiation(modalsv);
		Debug.assertTrue(op != null, "No instantiation found for " + op);
		operatorInst = true;
	    } else if (op instanceof AttributeOp &&
		       ((AttributeOp)op).attribute() instanceof SchemaVariable) {

		final IProgramVariable attribute = 
		    (ProgramVariable)svInst.getInstantiation
		    ((SchemaVariable)((AttributeOp)op).attribute());

		if (attribute == null) {
		    // illegal inst. exception required for matchMV to fail
		    throw new IllegalInstantiationException("No instantiation found for " + op);
		}

		if (op instanceof ShadowedOperator) {
		    op = ShadowAttributeOp.getAccessOp(attribute);
		} else {
		    op = AttributeOp.getAccessOp(attribute);
		}
		operatorInst = true;
	    }
	    

	    boolean jblockChanged = false;
	    JavaBlock jb          = visited.javaBlock();

	    if (jb != JavaBlock.EMPTY_JAVABLOCK) {
		jb = replacePrg(svInst, jb);
		if (jb!=visited.javaBlock()) {
		    jblockChanged=true;
		}
	    }

	    boolean varsChanged=false;

	    ArrayOfQuantifiableVariable boundVars = null;
	    if (visited.varsBoundHere(0).size() == 0) {
		if (visited.varsBoundHere(1).size() == 0) {
		    //For OCL simplification
		    boundVars = visited.varsBoundHere(2);
		    //
		} else {
		    boundVars = visited.varsBoundHere(1);
		} 
	    } else {
		boundVars = visited.varsBoundHere(0);
	    }
	    	    
	    final QuantifiableVariable[] vars = 
		new QuantifiableVariable[boundVars.size()];		

	    for (int i = 0; i < vars.length; i++) {		    		   
		final QuantifiableVariable boundVar = boundVars.getQuantifiableVariable(i);
		if (boundVar instanceof SchemaVariable) {
			final SortedSchemaVariable boundSchemaVariable = 
			    (SortedSchemaVariable)boundVar;
			if ( svInst.isInstantiated(boundSchemaVariable) ) {
			    vars[i] = (QuantifiableVariable)
				( (Term) svInst.getInstantiation
				  (boundSchemaVariable) ).op();
			} else {
			    vars[i] = createTemporaryLV ( boundSchemaVariable );
			}
			varsChanged = true;
		} else {
		    vars[i] = boundVar;
		}
	    }
	
	    final Term[] neededsubs = neededSubs(visited.arity());
	    if (varsChanged || jblockChanged || operatorInst ||
		(!subStack.empty() &&  subStack.peek() == newMarker) ||
		op != visited.op () ) {

		pushNew(resolveSubst(tf.createTerm(op, 
						   neededsubs,
						   vars,
						   jb,
						   visited.ruleVar())));
	    } else {
		final Term t = resolveSubst(visited);
		if ( t == visited )
		    subStack.push(t);
		else
		    pushNew(t);
	    }
	}
    }
    
    /**
     * @param boundSchemaVariable
     * @return
     */
    private LogicVariable createTemporaryLV (SortedSchemaVariable boundSchemaVariable) {
        final Term t = newInstantiations.get ( boundSchemaVariable );
        if ( t != null ) return (LogicVariable)t.op ();
        
        final Sort realSort = svInst.getGenericSortInstantiations ()
                                .getRealSort ( boundSchemaVariable.sort () );
        final LogicVariable v = new LogicVariable ( new Name ( "TempLV" ),
                                                    realSort );
        
        newInstantiations = newInstantiations.put ( boundSchemaVariable,
                                                    tf.createVariableTerm ( v ) );
        return v;
    }

    private Operator handleSortDependingSymbol (Operator op) {
        final SortDependingSymbol depOp = (SortDependingSymbol)op;
        final Sort depSort = depOp.getSortDependingOn ();
        final SortDefiningSymbols realDepSort =
            (SortDefiningSymbols)svInst.getGenericSortInstantiations ()
                                       .getRealSort ( depSort );
        final Operator res = (Operator)depOp.getInstanceFor ( realDepSort );
        Debug.assertFalse ( res == null,
                            "Did not find instance of symbol "
                            + op + " for sort " + realDepSort );
        return res;
    }

    private Term resolveSubst(Term t) {
	if (resolveSubsts && t.op() instanceof SubstOp)
	    return ((SubstOp)t.op ()).apply ( t );
	return t;
    }


    /**
     * PRECONDITION: visited.op() instanceof SchemaVariable &&
     *               ((SchemaVariable)visited.op()).isTermSV ()
     * Instantiate the given TermSV with a new metavariable
     * @return true iff the instantiation succeeded
     */
    private boolean instantiateWithMV ( Term visited ) {
	if ( newInstantiations == null ) return false;

	final SchemaVariable sv = (SchemaVariable)visited.op ();

	Term t = newInstantiations.get ( sv );

	if ( t == null ) {
	    final Sort realSort = svInst.getGenericSortInstantiations ()
	                                .getRealSort ( visited.sort () );
            final Metavariable mv = MetavariableDeliverer.createNewMatchVariable
                                        ( svInstBasename.toString (), realSort );

            if ( mv == null ) {
                newInstantiations = null;
                return false;
            } else {
                t = tf.createFunctionTerm ( mv );
                newInstantiations = newInstantiations.put ( sv, t );
            }
        }

	pushNew ( t );
	return true;
    }


    /**
     * delivers the new built term
     */
    public Term getTerm() {
	if (computedResult==null) {
	    Object o=null;
	    do {
		o=subStack.pop();
	    } while (o==newMarker);
	    Term t = (Term) o;
// 	    CollisionDeletingSubstitutionTermApplier substVisit
// 		= new CollisionDeletingSubstitutionTermApplier();
// 	    t.execPostOrder(substVisit);	    
// 	    t=substVisit.getTerm();	
	    computedResult=t;
	}	
	return computedResult;
    }


    public SVInstantiations getSVInstantiations () {
	return svInst;
    }


    /**
     * @return introduced metavariables for instantiation of schema
     * variables, or null if some schema variables could not be
     * instantiated
     */
    public MapFromSchemaVariableToTerm getNewInstantiations () {
	return newInstantiations;
    }

    /**
     * this method is called in execPreOrder and execPostOrder in class Term
     * when leaving the subtree rooted in the term subtreeRoot.
     * Default implementation is to do nothing. Subclasses can
     * override this method 
     * when the visitor behaviour depends on informations bound to subtrees.
     * @param subtreeRoot root of the subtree which the visitor leaves.
     */

    public void subtreeLeft(Term subtreeRoot){
	if (subtreeRoot.op() instanceof MetaOperator) {
	    MetaOperator mop = (MetaOperator) subtreeRoot.op();
	    pushNew(mop.calculate((Term)subStack.pop(),svInst, getServices()));
	} 
   }


}
