// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/**
 * Tests matching and application of some OCL taclets (for OCL simplification). 
 */
package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.util.Debug;

import junit.framework.*;

import java.io.*;

public class TestOCLTaclets extends TestCase {

    RuleAppIndex index;
    Goal goal;

    RewriteTaclet ocl_iterate_step;
    RewriteTaclet ocl_iterate_base;
    RewriteTaclet ocl_if;
    RewriteTaclet ocl_equals;


    public TestOCLTaclets(String name) {
	super(name);
    }

    public TestOCLTaclets(String name, boolean b) {
	super(name);
	Debug.ENABLE_DEBUG = b;
    }

    public void setUp() {	
	TacletForTests.setStandardFile(System.getProperty("key.home")+
				       java.io.File.separator+"system"+java.io.File.separator+
				       "proofExamples"+java.io.File.separator+"ocl"+
				       java.io.File.separator+"ocl_taclets.key");
	
	TacletForTests.parse();

	ocl_iterate_step = (RewriteTaclet)
	    TacletForTests.getTaclet("ocl_iterate_step").taclet();

	ocl_iterate_base = (RewriteTaclet)
	    TacletForTests.getTaclet("ocl_iterate_base").taclet();

	ocl_if = (RewriteTaclet)
	    TacletForTests.getTaclet("ocl_if").taclet();

	ocl_equals = (RewriteTaclet)
	    TacletForTests.getTaclet("ocl_equals").taclet();


	SetOfTaclet tacletSet = SetAsListOfTaclet.EMPTY_SET;
	tacletSet = tacletSet.add(ocl_iterate_step);
	tacletSet = tacletSet.add(ocl_iterate_base);
	tacletSet = tacletSet.add(ocl_equals);
	tacletSet = tacletSet.add(ocl_if);
	TacletIndex tacletIndex = new TacletIndex(tacletSet);
	TacletAppIndex tacletAppIndex = new TacletAppIndex(tacletIndex);
	BuiltInRuleIndex biri = new BuiltInRuleIndex();
	BuiltInRuleAppIndex birai = new BuiltInRuleAppIndex(biri);
	index = new RuleAppIndex(tacletAppIndex, birai);
    }

    public void test1() {
	Term t = TacletForTests.parseTerm("$equals(set, set)");
	ListOfTacletApp list = findMatchingTaclets(t);
	TacletApp tacletApp = list.head();
	assertNotNull("Find ocl_equals", tacletApp);
	ListOfGoal goalList = tacletApp.execute(goal, TacletForTests.services());
	goal = goalList.head();
	assertNotNull("Successfully apply ocl_equals", goal);
	Term result = goal.node().sequent().succedent().getFirst().formula().sub(0);
	Term expected = TacletForTests.parseTerm("$true");
	assertEquals("ocl_equals applied to: $equals(set, set)", result, expected);
    }

    public void test2() {
	Term t = TacletForTests.parseTerm("$if($true, $false, $true)");
	ListOfTacletApp list = findMatchingTaclets(t);
	TacletApp tacletApp = list.head();
	assertNotNull("Find ocl_if", tacletApp);
	ListOfGoal goalList = tacletApp.execute(goal, TacletForTests.services());
	goal = goalList.head();
	assertNotNull("Successfully apply ocl_if", goal);
	Term result = goal.node().sequent().succedent().getFirst().formula().sub(0);
	Term expected = TacletForTests.parseTerm("$false");
	assertEquals("ocl_if applied to: $if($true, $false, $true)", result, expected);
    }
   
    public void test3() {
	Term t = TacletForTests.parseTerm("$iterate($cons($true, $nil), $false, " +
					  "\\elem:OclAny\\acc:OclAny/ $and(elem, acc))");

	LogicVariable this_elem 
	    = (LogicVariable)t.varsBoundHere(2).getQuantifiableVariable(0);
	LogicVariable this_acc 
	    = (LogicVariable)t.varsBoundHere(2).getQuantifiableVariable(1);

	ListOfTacletApp list = findMatchingTaclets(t);
	TacletApp tacletApp = list.head();
	assertNotNull("Find ocl_iterate_step", tacletApp);
	ListOfGoal goalList = tacletApp.execute(goal, TacletForTests.services());
	goal = goalList.head();
	assertNotNull("Successfully apply ocl_iterate_step", goal);
	Term result = goal.node().sequent().succedent().getFirst().formula().sub(0);

	//Build the expected result term:
	//$and($true, 
	//     $iterate($nil, 
	//              $false, 
	//              \elem:OclAny\acc:OclAny/ $and(elem, acc)))
	Term nilT = TacletForTests.parseTerm("$nil");
	Term falseT = TacletForTests.parseTerm("$false");
	Term trueT = TacletForTests.parseTerm("$true");
	Term elemT = TermFactory.DEFAULT.createVariableTerm(this_elem);
	Term accT = TermFactory.DEFAULT.createVariableTerm(this_acc);
	TermSymbol and = TacletForTests.funcLookup("$and");
	Term andTerm2 = TermFactory.DEFAULT.createFunctionTerm(and, 
							       new Term[]{elemT, accT});
	TermSymbol iterate = TacletForTests.funcLookup("$iterate");
	Term iterateTerm 
	    = TermFactory.DEFAULT.createTernaryBindingTerm(iterate,
							   this_elem, this_acc,
							   new Term[]{nilT, falseT, andTerm2});
	Term expected = TermFactory.DEFAULT.createFunctionTerm(and, 
							       new Term[]{trueT, iterateTerm});
    
	assertEquals("ocl_iterate_step applied to: $iterate($cons($true, $nil), $false, " +
					     "\\elem:OclAny\\acc:OclAny/ $and(elem, acc)", 
		     result, expected);
    }

    public void test4() {
	Term t = TacletForTests.parseTerm("$iterate($nil, $false, " +
					  "\\elem:OclAny\\acc:OclAny/ $and(elem, acc))");
	ListOfTacletApp list = findMatchingTaclets(t);
	TacletApp tacletApp = list.head();
	assertNotNull("Find ocl_iterate_base", tacletApp);
	ListOfGoal goalList = tacletApp.execute(goal, TacletForTests.services());
	goal = goalList.head();
	assertNotNull("Successfully apply ocl_iterate_base", goal);
	Term result = goal.node().sequent().succedent().getFirst().formula().sub(0);
	Term expected = TacletForTests.parseTerm("$false");
	assertEquals("ocl_iterate_base applied to: $iterate($nil, $false, " +
					     "\\elem:OclAny\\acc:OclAny/ $and(elem, acc)", 
		     result, expected);
    }

    private ListOfTacletApp findMatchingTaclets(Term t) {
	TermSymbol oclWrapper = TacletForTests.funcLookup("$OclWrapper");
	Term term = TermFactory.DEFAULT.createFunctionTerm(oclWrapper, t);
	ConstrainedFormula cf = new ConstrainedFormula(term);
	Semisequent succ = Semisequent.EMPTY_SEMISEQUENT;
	SemisequentChangeInfo info = succ.insertFirst(cf);
	succ = info.semisequent();
	Sequent seq = Sequent.createSuccSequent(succ);
	Proof proof = new Proof();
	Node node = new Node(proof, seq);
	proof.setRoot(node);
	proof.setSimplifier(new SimultaneousUpdateSimplifier());
	goal = new Goal(node, index);

	//Matching of taclets
	PosInOccurrence pos = new PosInOccurrence(cf, PosInTerm.TOP_LEVEL.down(0), proof);
	ListOfTacletApp list 
	    = index.getTacletAppAt(//SLListOfHeuristic.EMPTY_LIST,
				   //true,
				   TacletFilter.TRUE,
				   pos,
				   TacletForTests.services(),
				   Constraint.BOTTOM);
	return list;
    }
}
