// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.metaconstruct;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.ArrayOfStatement;
import de.uka.ilkd.key.java.ArrayOfExpression;
import de.uka.ilkd.key.java.NonTerminalProgramElement;
import de.uka.ilkd.key.java.expression.operator.CopyAssignment;
import de.uka.ilkd.key.java.statement.MethodFrame;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.reference.MethodReference;
import de.uka.ilkd.key.java.declaration.Modifier;
import de.uka.ilkd.key.java.declaration.MethodDeclaration;
import de.uka.ilkd.key.java.declaration.LocalVariableDeclaration;
import de.uka.ilkd.key.java.declaration.VariableSpecification;
import de.uka.ilkd.key.java.declaration.ParameterDeclaration;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.expression.operator.*; 
import de.uka.ilkd.key.java.expression.literal.*;
import de.uka.ilkd.key.rule.metaconstruct.ProgramMetaConstruct;
import de.uka.ilkd.key.java.visitor.ProgVarReplaceVisitor;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.util.Debug;
import java.util.Vector;
 

/** 
 * Symbolically executes a method invocation
 */ 
public class MethodCall extends ProgramMetaConstruct {


    private static long COUNTER = 0;
    

    private final SchemaVariable resultVar;
    
    private MethodReference methRef;
    private ProgramMethod pm;
    private ReferencePrefix newContext;
    private ProgramVariable paramVars[];
    private Services services;
    private ProgramVariable pvar;
    private IExecutionContext execContextSV;
    private ExecutionContext execContext;
    private ArrayOfExpression args;

    /** creates the methodcall-MetaConstruct 
     * @param body the ProgramElement contained by the meta construct 
     */
    public MethodCall(ProgramElement body) {
	this(null, null, body);
    }

    
    /** creates the methodcall-MetaConstruct 
     * @param result the SchemaVariable that is used to keep the result
     * @param body the ProgramElement contained by the meta construct 
     */
    public MethodCall(SchemaVariable result, ProgramElement body) {
	this(null, result, body);
    }

    /** creates the methodcall-MetaConstruct 
     * @param result the SchemaVariable that is used to keep the result
     * @param body the ProgramElement contained by the meta construct 
     */
    public MethodCall(SortedSchemaVariable ec, SchemaVariable result,
		      ProgramElement body) {
	super(new Name("method-call"), body);
	this.resultVar = result;
	this.execContextSV = ec;
    }

    

    /** gets an array of expression and returns a list of types */
    private ListOfKeYJavaType getTypes(Services services, 
				       ArrayOfExpression args) {
	ListOfKeYJavaType result = SLListOfKeYJavaType.EMPTY_LIST; 
	for (int i = args.size()-1; i >= 0 ; i--) {
	    Expression argument = args.getExpression(i);
	    result = result.prepend
		(services.getTypeConverter().getKeYJavaType(argument));
	}
	return result;
    }


    private ProgramMethod assertImplementationPresent(ProgramMethod pm,
						      KeYJavaType t) {
	if (pm == null) {	    
	    Debug.fail("methodcall:No implementation available for ", pm);
	}
	return pm;
    }


    private KeYJavaType getStaticPrefixType(ReferencePrefix refPrefix, 
					    Services services) {
	if (refPrefix==null || refPrefix instanceof ThisReference) {
	    return execContext.getTypeReference().getKeYJavaType();
	} else if (refPrefix instanceof TypeRef) {
	    KeYJavaType t = ((TypeRef)refPrefix).getKeYJavaType();
	    if (t == null) { //%%%
		Debug.fail();
	    }
	    return t;
	} else if (refPrefix instanceof ProgramVariable) {
	    return ((ProgramVariable)refPrefix).getKeYJavaType();
	} else if (refPrefix instanceof FieldReference) {
	    return ((FieldReference)refPrefix).getProgramVariable()
		.getKeYJavaType();
	} else if (refPrefix instanceof SuperReference) {;
	    return services.getJavaInfo().getSuperclass
		(execContext.getTypeReference().getKeYJavaType()); 	
	} else {
	    throw new de.uka.ilkd.key.util.NotSupported
		("Unsupported method invocation mode\n"+
		 refPrefix.getClass());
	}			    
    }

    private ProgramMethod getMethod(KeYJavaType prefixType, MethodReference mr) {
	if (execContext != null){
	    return mr.method(services, prefixType, execContext);
	} else {
	    return mr.method(services, prefixType, prefixType);
	}
    }

    private ProgramMethod getSuperMethod(ExecutionContext ex,
					 MethodReference mr) {
	return mr.method(services, getSuperType(ex), ex);
    }

    private KeYJavaType getSuperType(ExecutionContext ex) {
	return services.getJavaInfo().getSuperclass
	    (ex.getTypeReference().getKeYJavaType());
    }
    

    /** performs the program transformation needed for symbolic
     * program execution 
     * @param services the Services with all necessary information 
     * about the java programs
     * @param svInst the instantiations esp. of the inner and outer label 
     * @return the transformed program
     */
    public ProgramElement symbolicExecution(ProgramElement pe,
					    Services services,
					    SVInstantiations svInst) {

	Debug.out("method-call: called for ", pe);
	this.services=services;

	if (resultVar != null) {
	    pvar = (ProgramVariable) svInst.getInstantiation(resultVar);
	}

	if (execContextSV!=null) {
	    execContext
		= (ExecutionContext) svInst.getInstantiation
		((SortedSchemaVariable)execContextSV);
	} else {
	    execContext = (ExecutionContext) svInst.
		getContextInstantiation().activeStatementContext();
	}

	methRef = (MethodReference) pe;

	ReferencePrefix refPrefix = methRef.getReferencePrefix();
	if (refPrefix == null) {
	    if (execContext.getRuntimeInstance() == null) {
		refPrefix = execContext.getTypeReference();
	    } else {
		refPrefix = execContext.getRuntimeInstance();
	    }
	}
	
	KeYJavaType staticPrefixType 
	    = getStaticPrefixType(methRef.getReferencePrefix(), services);
	if(execContext != null){
	    pm = assertImplementationPresent
		(methRef.method(services, staticPrefixType, execContext),
		 staticPrefixType);
	}else{
	    pm = assertImplementationPresent
		(methRef.method(services, staticPrefixType, staticPrefixType), 
		 staticPrefixType);	    
	}
        newContext = methRef.getReferencePrefix();
	if (newContext == null || newContext instanceof ThisReference) {
	    newContext = execContext.getRuntimeInstance();
	} else if (newContext instanceof FieldReference) {
	    FieldReference fieldContext = (FieldReference) newContext;
	    if ((fieldContext.getReferencePrefix() == null || 
		 fieldContext.getReferencePrefix() instanceof ThisReference) 
		&& !fieldContext.getProgramVariable().isStatic()) {
		newContext = fieldContext.setReferencePrefix
		    (execContext.getRuntimeInstance());
	    }
	}

	VariableSpecification[] paramSpecs = createParamSpecs();
	Statement[] paramDecl = createParamAssignments(paramSpecs);


        args = getVariables(paramSpecs);

	// make hiliting save and disabled in automatic mode%%%
	//        if (!de.uka.ilkd.key.gui.Main.standalone) {
        //    de.uka.ilkd.key.casetool.together.ModelUtil.hiliteMethod
	//	(staticPrefixType, pm);
	//}

	Statement result = null;
	
	if (pm.isStatic()) {	// Static invocation mode
	    Debug.out("method-call: invocation of static method detected");
            newContext = null;
	    ProgramMethod staticMethod = getMethod(staticPrefixType, methRef);
	    result = new MethodBodyStatement(staticMethod, newContext,
					     pvar, args); 
	} else if (refPrefix instanceof SuperReference) {
	    Debug.out("method-call: super invocation of method detected." + 
		      "Requires static resolving.");
	    ProgramMethod superMethod = getSuperMethod(execContext,
						       methRef);
	    result = new MethodBodyStatement
		(superMethod, execContext.getRuntimeInstance(), pvar,
		 args);
	} else {    // Instance invocation mode
	    if (pm.isPrivate()) { // private methods are bound statically
		Debug.out("method-call: invocation of private method detected." + 
			  "Requires static resolving.");
		result = makeMbs(staticPrefixType);
	    } else {
		Debug.out("method-call: invocation of non-private"
			  +" instance method detected." 
			  +"Requires dynamic resolving.");
		ListOfKeYJavaType imps = 
		    services.getJavaInfo().getKeYProgModelInfo()
		    .findImplementations
		    ((KeYJavaType)staticPrefixType, methRef.getName(),
		     getTypes(services, args));
		result = makeIfCascade(imps);
	    }
	}
	return insertStatementInBlock(paramDecl, new StatementBlock(result));
    }


    //***************** Dynamic Binding Construction Utilities ***************

    private If ifCascadeTop = null;
    private If ifCascadeBottom = null;
    private Vector ifBodys = null;
    private Vector ifConds = null;


    private Statement makeMbs(KeYJavaType t) {
	ProgramMethod meth = getMethod(t, methRef);
	return new MethodBodyStatement(meth, newContext,
				       pvar, args);
    }


    public Expression makeIOf(Type t) {
	Debug.assertTrue(newContext!=null);
	return new Instanceof((Expression) newContext, 
			      new TypeRef((KeYJavaType)t));
    }


    private Statement makeIfCascade(ListOfKeYJavaType imps) {
        KeYJavaType currType = imps.head();
        if (imps.size()==1) 
           return makeMbs(currType);
        else return new If(makeIOf(currType),
                           new Then(makeMbs(currType)),
                           new Else(makeIfCascade(imps.tail())));
    }


    public VariableSpecification[] createParamSpecs(){
	
	MethodDeclaration methDecl    = pm.getMethodDeclaration();
	int params                    = methDecl.getParameterDeclarationCount();
	paramVars = new ProgramVariable[params];
	VariableSpecification[] varSpecs = new VariableSpecification[params];
	for (int i = 0; i < params; i++) {
	    ParameterDeclaration parDecl = 
		methDecl.getParameterDeclarationAt(i);
	    VariableSpecification originalSpec = 
		parDecl.getVariableSpecification();	    	    
	    final ProgramVariable originalParamVar =
	        (ProgramVariable)originalSpec.getProgramVariable ();
	    final IProgramVariable paramVar =
	        ProgVarReplaceVisitor.copy ( originalParamVar, "_" + COUNTER++);
	    varSpecs[i] = 
		new VariableSpecification
		    (paramVar, 
		     originalSpec.getDimensions(), 
		     methRef.getArgumentAt(i),
		     originalSpec.getType());
	} 
	return varSpecs;
    }

    public Statement[] createParamAssignments(VariableSpecification[] specs) {
	MethodDeclaration methDecl    = pm.getMethodDeclaration();
	Statement[] paramDecl = new Statement[specs.length];
	for (int i=0; i<specs.length; i++) {
	    ParameterDeclaration parDecl = 
		methDecl.getParameterDeclarationAt(i);
	    paramDecl[i] = new LocalVariableDeclaration
		(parDecl.getModifiers(), parDecl.getTypeReference(), specs[i]);
	}
	return paramDecl;
    }

    private ArrayOfExpression getVariables(VariableSpecification[] varspecs) {
	Expression[] vars = new Expression[varspecs.length];
	for (int i=0; i<varspecs.length; i++) {
	    vars[i] = (Expression) varspecs[i].getProgramVariable();
	}
	return new ArrayOfExpression(vars);
    }
	


    /** inserts the given statements at the begin of the block 
     * @param stmnt array of Statement those have to be inserted
     * @param b the Statementblock where to insert
     */
    private StatementBlock insertStatementInBlock(Statement[] stmnt, 
						  StatementBlock b) {
	
	Statement[] block = new Statement[b.getStatementCount()+
					  stmnt.length];
	System.arraycopy(stmnt, 0, block, 0, stmnt.length);
	b.getBody().arraycopy(0, block, stmnt.length, b.getStatementCount());
	return new StatementBlock(new ArrayOfStatement(block));	
    }


}
