// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.rule.metaconstruct;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.op.AbstractMetaOperator;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.ProgramMethod;
import de.uka.ilkd.key.logic.op.ArrayOfQuantifiableVariable;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.java.ArrayOfExpression;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.statement.MethodBodyStatement;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.declaration.LocalVariableDeclaration;
import de.uka.ilkd.key.java.declaration.VariableSpecification;
import de.uka.ilkd.key.java.reference.TypeRef;
import de.uka.ilkd.key.logic.JavaBlock;

public class ResolveQuery extends AbstractMetaOperator {


    private Term[] addUpdates = null;
    private JavaBlock addJavaBlock = null;
    private JavaBlock addDecls = null;
    private Services services;
    private TermFactory tf = TermFactory.DEFAULT;
    private boolean stop = false;    
    
    public ResolveQuery() {
	super(new Name("#resolveQuery"), 1);
    }


    /**
     * checks whether the top level structure of the given @link Term
     * is syntactically valid, given the assumption that the top level
     * operator of the term is the same as this Operator. The
     * assumption that the top level operator and the term are equal
     * is NOT checked.  
     * @return true iff the top level structure of
     * the @link Term is valid.
     */
    public boolean validTopLevel(Term term) {
	// a meta operator accepts almost everything
	return  term.arity()==arity();
    }

    private ArrayOfExpression createArgumentPVs(Term t) {
	boolean staticMethod = ((ProgramMethod)t.op()).isStatic();
	Expression[] result = new Expression[t.arity()-(staticMethod ? 0 : 1)];
	for (int i=0; i<result.length; i++) {
	    result[i]=createPV(t.sub(i + (staticMethod ? 0 : 1)));
	}
	return new ArrayOfExpression(result);
    }

    private ProgramVariable createPV(Term t) {
	String pvstring = t.op().name().toString()+"_prg";
	if (pvstring.indexOf("::")>0) {
	    pvstring=pvstring.substring(pvstring.lastIndexOf("::")+2);
	}
	if (pvstring.startsWith(".")) pvstring=pvstring.substring(1);
	ProgramElementName pvname 
	    = new ProgramElementName(pvstring);
	return new ProgramVariable
	    (pvname, services.getTypeConverter().getKeYJavaType(t));
    }

    private Term createProgramMethodSubstitute(Term t) {
	ProgramMethod pm = (ProgramMethod) t.op();
	addUpdates = new Term[2*t.arity() +1];
	// the last slot is for the subterm of the update term (to
	// be set on a higher recursion level)
	ArrayOfExpression argPVs = createArgumentPVs(t);
	ProgramVariable callerPV = pm.isStatic() ? null : createPV(t.sub(0));
	// create updates
	for (int i=0; i<argPVs.size(); i++) {
	    addUpdates[2*i] = tf.createVariableTerm
		((ProgramVariable)argPVs.getExpression(i));
	    addUpdates[2*i+1] = t.sub(i + (pm.isStatic() ? 0 : 1));
	}
	if (!pm.isStatic()) {
	    addUpdates[addUpdates.length-3] = tf.createVariableTerm(callerPV);
	    addUpdates[addUpdates.length-2] = t.sub(0);
	}	
	// create java block
	ProgramVariable res = createPV(t);
	MethodBodyStatement mbs
	    = new MethodBodyStatement(pm, callerPV, res, argPVs);
	addJavaBlock = JavaBlock.createJavaBlock(new StatementBlock(mbs));
	// create local var declaration java block
	LocalVariableDeclaration[] vardecls 
	    = new LocalVariableDeclaration[t.arity()+1]; 
	vardecls[0] = new LocalVariableDeclaration
	    (new TypeRef(res.getKeYJavaType()), 
	     new VariableSpecification(res)); 
	for (int i=0; i<argPVs.size(); i++) {
	    ProgramVariable pv = (ProgramVariable)argPVs.getExpression(i);
	    vardecls[i+1] = new LocalVariableDeclaration
		(new TypeRef(pv.getKeYJavaType()), 
		 new VariableSpecification(pv));
	}
	if (!pm.isStatic()) {
	    vardecls[vardecls.length-1] = new LocalVariableDeclaration
		(new TypeRef(callerPV.getKeYJavaType()), 
		 new VariableSpecification(callerPV));
	}
	addDecls = JavaBlock.createJavaBlock(new StatementBlock(vardecls));
	return tf.createVariableTerm(res);
    }

    /** returns a formula equivalent to the given one. The returned
     * formula has replaced the first occurrence (in depth-first order) 
     * with a top-operand ProgramMethod with a non ProgramMethod term.
     */
    private Term replaceFirstProgramMethodTerm(Term t) {	
	if (!stop && (t.op() instanceof ProgramMethod)) {
	    stop = true;	    
	    return createProgramMethodSubstitute(t);
	}
	Term[] subs = new Term[t.arity()];
	boolean changed = false;
	for (int i=0; i<t.arity(); i++) {
	    subs[i] = replaceFirstProgramMethodTerm(t.sub(i));	   
	    if (subs[i]!=t.sub(i)) {
		changed = true;
	    }
	}
	if (changed) {
	    ArrayOfQuantifiableVariable v 
		= (t.varsBoundHere(0).size() == 0) 
		? t.varsBoundHere(1)
		: t.varsBoundHere(0);
	    Term result = tf.createTerm(t.op(), subs, v, 
					t.javaBlock(), t.ruleVar());
	    if (result.sort()==Sort.FORMULA && addJavaBlock!=null) {
		result = tf.createDiamondTerm(addJavaBlock, result);
		addJavaBlock = null;
		if (addUpdates!=null && addUpdates.length>1) {
		    addUpdates[addUpdates.length-1] = result;
		    result = tf.createUpdateTerm(addUpdates);
		    addUpdates = null;
		}
		if (addDecls!=null) {
		    result = tf.createDiamondTerm(addDecls, result);
		    addDecls=null;
		}
	    }
	    return result;
	} else { // unchanged, share old term
	    return t;
	}
    }

    /** calculates the resulting term. */
    public Term calculate(Term term, SVInstantiations svInst, 
			  Services services) {
	this.services = services;
	stop=false;	
	return replaceFirstProgramMethodTerm(term.sub(0));	
    }


}
