// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.strategy;

import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Named;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.NamespaceSet;
import de.uka.ilkd.key.proof.IHTacletFilter;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.TacletFilter;
import de.uka.ilkd.key.proof.UpdateSimplificationRuleFilter;
import de.uka.ilkd.key.rule.Heuristic;
import de.uka.ilkd.key.rule.ListOfHeuristic;
import de.uka.ilkd.key.rule.SLListOfHeuristic;
import de.uka.ilkd.key.strategy.feature.*;
import de.uka.ilkd.key.util.Debug;


/**
 *
 */
public abstract class AbstractFeatureStrategy implements Strategy {

    private final Proof proof;

    public AbstractFeatureStrategy (Proof proof) {
        this.proof = proof;
    }

    /**
     * @return Returns the proof.
     */
    protected Proof getProof () {
        return proof;
    }

    /**
     * @param heuristics
     * @param thenFeature
     * @return
     */
    protected Feature ifHeuristics (String[] heuristics, Feature thenFeature) {
        return ConditionalFeature.createConditional (
                getFilterFor ( heuristics ),
                thenFeature );
    }

    /**
     * @param heuristics
     * @param thenFeature
     * @param elseFeature
     * @return
     */
    protected Feature ifHeuristics (String[] heuristics,
                                    Feature thenFeature,
                                    Feature elseFeature) {
        return ConditionalFeature.createConditional ( getFilterFor ( heuristics ),
                                                      thenFeature,
                                                      elseFeature );
    }

    protected Feature longConst (long a) {
        return ConstFeature.createConst ( c ( a ) );
    }

    protected Feature inftyConst () {
        return ConstFeature.createConst ( infty () );
    }

    protected Feature add (Feature a, Feature b) {
        return SumFeature.createSum ( new Feature[] { a, b} );
    }

    protected Feature ifZero (Feature cond, Feature thenFeature) {
        return ShannonFeature.createConditionalBinary ( cond, thenFeature );
    }

    protected Feature ifZero (Feature cond, Feature thenFeature, Feature elseFeature) {
        return ShannonFeature.createConditionalBinary ( cond,
                                                        thenFeature,
                                                        elseFeature );
    }

    /**
     * @param names
     * @param priority
     * @return
     */
    protected Feature ifHeuristics (String[] names, int priority) {
        return ConditionalFeature.createConditional ( getFilterFor ( names ),
                c ( priority ), c ( 0 ) );
    }

    /**
     * @param priority
     * @return
     */
    protected Feature selectSimplifier (long priority) {
        return ConditionalFeature.createConditional(
                UpdateSimplificationRuleFilter.INSTANCE, 
                longConst ( priority ) );
    }

    private RuleAppCost c (long p) {
        return new LongRuleAppCost ( p );
    }

    private RuleAppCost infty () {
        return TopRuleAppCost.INSTANCE;
    }

    protected TacletFilter getFilterFor (String[] p_names) {
        ListOfHeuristic heur = SLListOfHeuristic.EMPTY_LIST;
        for ( int i = 0; i != p_names.length; ++i )
            heur = heur.prepend ( getHeuristic ( p_names[i] ) );
        return new IHTacletFilter ( false, heur );
    }

    private Heuristic getHeuristic (String p_name) {
        final NamespaceSet nss = getProof ().getNamespaces ();
        
        Debug.assertFalse ( nss == null, "Rule set namespace not available." ); 
        
        final Namespace ns = nss.heuristics ();
        final Named h = ns.lookup ( new Name ( p_name ) );
    
        Debug.assertFalse ( h == null, "Did not find the rule set " + p_name );
    
        return (Heuristic)h;
    }
}
