// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.ocl.ocldltranslator;

/**
 * The transformation of OCL expression in logical formulas
 * as used in the KeY project !
 */

import tudresden.ocl.*;
import tudresden.ocl.parser.analysis.*;
import java.util.Set;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Vector;
import tudresden.ocl.parser.node.*;
import tudresden.ocl.check.types.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.*;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import java.io.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.heuristics.*;
import de.uka.ilkd.key.ocl.ocldltranslator.control.*;
import de.uka.ilkd.key.ocl.ocldltranslator.normalize.*;
import de.uka.ilkd.key.ocl.ocldltranslator.transform.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.LogicVariable;
import de.uka.ilkd.key.logic.op.Quantifier;

public class KeyOclExpressionTranslator extends BasicOclExpressionTranslator implements TranslationConstants {

    KeyOclConstraintTranslator constraintTranslator;
    DefaultTranslationControl translationControl;

    protected TermFactory tf = TermFactory.DEFAULT;
    protected ProverFacade proverFacade;
    protected NewVariableDeliverer newVariableDeliverer;
    NewFunctionNameDeliverer newFunctionNameDeliverer;
    Set occuringNonrigidFunctionSymbols;
    Map occuringAttributeAtPreFunctions;
    Map occuringMethodAtPreFunctions;

    // The nodes in the AST which will be translated by predicative formulas ...
    Set nodesForPredicativTranslation;


    boolean logEnabled = false;

    public KeyOclExpressionTranslator(OclTree constraintTree, 
				      KeyOclConstraintTranslator constraintTranslator){
	super(constraintTree);
	this.constraintTranslator = constraintTranslator;
	// this.translationControl = new GeneralTranslationControl();
	this.translationControl = new KeyTranslationControl();
    }

    public KeyOclExpressionTranslator(OclTree constraintTree, 
				      KeyOclConstraintTranslator constraintTranslator,
				      DefaultTranslationControl tControl){
	super(constraintTree);
	this.constraintTranslator = constraintTranslator;
	this.translationControl = tControl;
    }

    public KeyOclExpressionTranslator(OclTree constraintTree,
				      KeyOclConstraintTranslator constraintTranslator,
				      Vector predefinedVarNames,
				      Vector predefinedVarTypes){
	super(constraintTree, predefinedVarNames, predefinedVarTypes);
	this.constraintTranslator = constraintTranslator;
	// this.translationControl = new GeneralTranslationControl();
	this.translationControl = new KeyTranslationControl();
    }
    
    public KeyOclExpressionTranslator(OclTree constraintTree,
				      KeyOclConstraintTranslator constraintTranslator,
				      Vector predefinedVarNames,
				      Vector predefinedVarTypes ,
				      DefaultTranslationControl tControl){
	super(constraintTree, predefinedVarNames, predefinedVarTypes);
	this.constraintTranslator = constraintTranslator;
	this.translationControl = tControl;
    }

    // for test purposes ...
    public TranslationResult getTranslationForNode(Node n){
	
	try {
	    return(super.getTranslationForNode(n));
	} catch(Exception e){
	    log("Ignored unfound translation result for node : " + n);
	    return(new KeyTranslationResult(null));
	}

	
    }


    public ProverFacade proverFacade() {
	return proverFacade;
    }

    public TermFactory termFactory() {
	return tf;
    }

    public KeyOclConstraintTranslator constraintTranslator() {
	return constraintTranslator;
    }

    public NewVariableDeliverer newVariableDeliverer() {
	return newVariableDeliverer;
    }

    public void setTranslationForNode(Node n, TranslationResult translation){
	KeyTranslationResult trans = (KeyTranslationResult) translation;
	if (trans != null) {
	    if (trans.getTerm() != null) {
		
		if (trans.isPredicativeTranslation()){
		    log("Setting PREDICATIVE translation for Node : " + n);
		    
		}
		nodeTranslationResultMap.put(n, trans);
	    } else  {
		error("Trying to set 'null' as the term in the translation of the node :" + n );
	    }
	} else {
	    error("Trying to set 'null' as the translation of the node :" + n );
	}
    }
    

    /**
     * Specifies the way in which the given node should be translated according to the used translation 
     * contol heuristic
     */
    protected TranslationType getTranslationTypeForNode(Node n){
	return translationControl.getTranslationTypeForNode(n);
    }


    public TranslationResult getTranslationFor(AExpression expression){
	this.proverFacade = constraintTranslator.getProverFacade();
	this.newVariableDeliverer = constraintTranslator.getNewVariableDeliverer();
	this.newFunctionNameDeliverer = constraintTranslator.getNewFunctionNameDeliverer();
	this.occuringNonrigidFunctionSymbols = new java.util.HashSet();
	this.occuringAttributeAtPreFunctions = new java.util.HashMap();
	this.occuringMethodAtPreFunctions = new java.util.HashMap();
	this.nodesForPredicativTranslation = new java.util.HashSet();

	// First start the heuristic which determines for every node the type of translation 
	// to be performed for this node ...
	log("Determine the way in which the nodes of the tree should be translated !!!");
	log("Using for TranslationControl the class : " +translationControl.getClass().toString());

	translationControl.setLoggingEnabled(this.logEnabled);
	translationControl.executeHeuristics(tree,expression, this);

	// Now start the main translation ...
	return (super.getTranslationFor(expression));
    }


    // ============= And now the translations ==============================

    


    // and, or, xor, implies
    protected TranslationResult translateLogicalExpression(Node le, String logicalOperatorName, TranslationResult leftOperand, TranslationResult rightOperand){
	log("Translating logical expression with Operator '"+ logicalOperatorName  +"'.");
	
	// Get the kind of translation to be performed for the given node ...
	TranslationType translationType = getTranslationTypeForNode(le);
	log("Current logical junctor should be translated in the following way : " + translationType.toString());
	

	KeyTranslationResult translation;
	Term resultTerm = null;
	KeyTranslationResult lOp = (KeyTranslationResult) leftOperand;
	KeyTranslationResult rOp = (KeyTranslationResult) rightOperand;

	Term lOpTerm = lOp.getTerm();
	Term rOpTerm = rOp.getTerm();


	if (!translationType.containsMark(TranslationMark.AS_BOOLEAN_TERM)) {
	    // Use formulas to represent logical OCL expressions in the returned translation result ...

	    // First of all convert all operand terms to formulas ...
	    Term lOpFormula = convertTermToFormula(lOpTerm);
	    Term rOpFormula = convertTermToFormula(rOpTerm);
	    
	
	    if(logicalOperatorName.equals("and")){
		// create AND junctor term  ...
		resultTerm = tf.createJunctorTermAndSimplify(Op.AND,lOpFormula,rOpFormula); 
	    } else if (logicalOperatorName.equals("or")){
		// create OR junctor term  ...
		resultTerm = tf.createJunctorTermAndSimplify(Op.OR,lOpFormula,rOpFormula); 
	    } if(logicalOperatorName.equals("implies")){
		// create IMPLIES junctor term  ...

		resultTerm = tf.createJunctorTermAndSimplify(Op.IMP,lOpFormula,rOpFormula); 
	    } if(logicalOperatorName.equals("xor")){
		// create term for xor [P xor Q as : not (P <-> Q)]
		Term t = tf.createJunctorTerm(Op.EQV,lOpFormula,rOpFormula); 
		resultTerm = tf.createJunctorTerm(Op.NOT,t); 
	    }
	    
	} else{
	    // Create some term over the ADT (or the JAVA-Type) 'boolean' instead of formulas.
	    
	    Function junctorSymbol;
	    Sort boolSort =  proverFacade.getSortForOCLBoolean();

	    // First of all convert all operand formulas to terms of sort boolean ...
	    lOpTerm = convertFormulaToTerm(lOpTerm, lOp);
	    rOpTerm =  convertFormulaToTerm(rOpTerm, rOp);


	     if(logicalOperatorName.equals("and")){
		// create AND symbol  ...
		  junctorSymbol =   proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_AND,
									 new Sort[]{boolSort,boolSort},
									 boolSort);


		 resultTerm = tf.createFunctionTerm(junctorSymbol ,lOpTerm,rOpTerm); 

	    } else if (logicalOperatorName.equals("or")){
		// create OR symbol  ...
		junctorSymbol =   proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_OR,
								       new Sort[]{boolSort,boolSort},
								       boolSort); 
		resultTerm = tf.createFunctionTerm(junctorSymbol ,lOpTerm,rOpTerm); 

	    } if(logicalOperatorName.equals("implies")){
		// create IMPLIES symbol  ...
		junctorSymbol =   proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_IMPLIES,
								       new Sort[]{boolSort,boolSort},
								       boolSort); 
		resultTerm = tf.createFunctionTerm(junctorSymbol ,lOpTerm,rOpTerm); 
 
	    } if(logicalOperatorName.equals("xor")){
		// create term for xor [P xor Q as : not (P = Q)]
		junctorSymbol =   proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_EQUALS_TO_BOOLEAN,
								       new Sort[]{boolSort,boolSort},
								       boolSort); 


		Term t = tf.createFunctionTerm(junctorSymbol ,lOpTerm,rOpTerm);
		junctorSymbol =   proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_NOT,
								       new Sort[]{boolSort},
								       boolSort); 

		resultTerm = tf.createFunctionTerm(junctorSymbol,t); 
	    }

	   
	     
	}
	
	log("INTO : " + resultTerm.toString());

	translation = new KeyTranslationResult(resultTerm);
	
	Set lAxioms = lOp.getAxioms();
	Set rAxioms = rOp.getAxioms();
       
	// Now add all axioms of all subexpressions/subtranslations ...
	translation.addAxioms(lAxioms);
	translation.addAxioms(rAxioms);

	return translation;
	
    }
    
    
    protected void translateBooleanLiteral(Node sl, String lit){
	log("Translating Boolean literal: " + lit);
	
	KeyTranslationResult translation;
	Term resultTerm;
	
	// Get the kind of translation to be performed for the given node ...
	TranslationType translationType = getTranslationTypeForNode(sl);
	log("Current boolean literal should be translated in the following way : " + translationType.toString());
	
	if (!translationType.containsMark(TranslationMark.AS_BOOLEAN_TERM)) {
	    // Use formulas to represent logical OCL expressions in the returned translation result ...
	   
	    if(lit.equals("true")){
		// create TRUE formula ...
		resultTerm = tf.createJunctorTerm(Op.TRUE); 
		
	    } else {
		// create FALSE formula ...
		resultTerm = tf.createJunctorTerm(Op.FALSE);
	    }
  
	} else {
	    // Create some term over the ADT (or the JAVA-Type) 'boolean' instead of formulas.

	    Function literalSymbol;
	    Sort boolSort =  proverFacade.getSortForOCLBoolean();

	    if(lit.equals("true")){
		// create TRUE constant of sort 'boolean' ...
		literalSymbol = proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_TRUE,
								       new Sort[0],
								       boolSort);
		
	    } else {
		// create FALSE constant of sort 'boolean' ...
		literalSymbol = proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_FALSE,
								       new Sort[0],
								       boolSort);
	    }

	    resultTerm = tf.createFunctionTerm(literalSymbol);
	    
  
	
	}

	log("INTO : " + resultTerm.toString());

	translation = new KeyTranslationResult(resultTerm);
	setTranslationForNode(sl,translation);
	
    }


    // +, -
    protected TranslationResult translateAdditiveExpression(Node ae, 
							    String additiveOperatorName,
							    TranslationResult leftOperand,
							    TranslationResult rightOperand){
	
	log("Translating additive expression with Operator '"+ additiveOperatorName  +"'.");
	KeyTranslationResult result = new KeyTranslationResult();
	Term resultTerm;
	String opName;
	if (additiveOperatorName.equals("+")){
	    opName = PLUS;
	} else {
	    opName = BINARY_MINUS;
	}
	
	Term leftOperandTerm = ((KeyTranslationResult)leftOperand).getTerm();
	Term rightOperandTerm =((KeyTranslationResult) rightOperand).getTerm();

	Sort operandSortLeft = leftOperandTerm.sort();
	Sort operandSortRight = rightOperandTerm.sort();
	if(!operandSortLeft.equals(operandSortRight)){
	     // We can not handle different types for additive expressions yet !
	    // error("Different types for the subexpressions of a ADDITIVE-expression can't be handeled YET !!");
	 }

	Sort opSort = leftOperandTerm.sort();
	Sort retSort = opSort;

	Function additveFunction =
	    proverFacade.getFunctionForFeature(opName,
						 new Sort[]{opSort,opSort},
						 retSort);
	    
	resultTerm = tf.createFunctionTerm(additveFunction,
					   leftOperandTerm,
					   rightOperandTerm);
	result.setTerm(resultTerm);
	result.addAxioms(((KeyTranslationResult)leftOperand).getAxioms());
	result.addAxioms(((KeyTranslationResult)rightOperand).getAxioms());
	return result;
    }

    // *, /
    protected TranslationResult translateMultiplicativeExpression(Node me,
								  String multiplicativeOperatorName,
								  TranslationResult leftOperand,
								  TranslationResult rightOperand){
	log("Translating multiplicative expression with Operator '"+ multiplicativeOperatorName  +"'.");
	KeyTranslationResult result = new KeyTranslationResult();
	Term resultTerm;
	String opName;
	
	
	Term leftOperandTerm = ((KeyTranslationResult)leftOperand).getTerm();
	Term rightOperandTerm =((KeyTranslationResult) rightOperand).getTerm();

	Sort operandSortLeft = leftOperandTerm.sort();
	Sort operandSortRight = rightOperandTerm.sort();
	if(!operandSortLeft.equals(operandSortRight)){
	     // We can not handle different types for additive expressions yet !
	    // error("Different types for the subexpressions of a MULTIPLICATIVE-expression can't be handeled YET !!");
	 }

	Sort opSort = leftOperandTerm.sort();
	Sort retSort = opSort;

	if (multiplicativeOperatorName.equals("*")){
	    opName = MULTIPLICATION;
	    retSort = opSort;
	} else {
	    opName = DIVISION;
	    retSort = getDLSortForType(Basic.REAL);
	}
	

	Function multiplicativeFunction =
	    proverFacade.getFunctionForFeature(opName,
						 new Sort[]{opSort,opSort},
						 retSort);
	    
	resultTerm = tf.createFunctionTerm(multiplicativeFunction,
					   leftOperandTerm,
					   rightOperandTerm);
	result.setTerm(resultTerm);
	result.addAxioms(((KeyTranslationResult)leftOperand).getAxioms());
	result.addAxioms(((KeyTranslationResult)rightOperand).getAxioms());
	return result;

    }


    // =, <>, <, <=, >, >=
    protected void translateRelationalExpression(Node re, String relationalOperatorName, Node leftOperand, Node rightOperand){
	log("Translating relational expression with Operator '"+ relationalOperatorName  +"'.");

	 Type operandTypeLeft = tree.getNodeType(leftOperand);
	 Type operandTypeRight = tree.getNodeType(rightOperand);
	 
	 if(!operandTypeLeft.equals(operandTypeRight)){
	     // We can not handle different types for relational expressions yet !
	     // error("Different types for the subexpressions of a RELATIONAL-expression can't be handeled YET !!");
	 }

	 log("OperandType left : " + operandTypeLeft );
	 log("OperandType right : " + operandTypeRight );
	 
	 KeyTranslationResult leftOpTranslation = (KeyTranslationResult) getTranslationForNode(leftOperand);
	 KeyTranslationResult rightOpTranslation = (KeyTranslationResult) getTranslationForNode(rightOperand);
	 
	 Type t = operandTypeLeft;
	 

	 KeyTranslationResult result;
	 Term resultTerm = null;
	 Term aTerm;

	if(relationalOperatorName.equals("=") || relationalOperatorName.equals("<>") ){
	    // Translate the equality or inequality ...
	    Term leftOp = leftOpTranslation.getTerm();
	    Term rightOp = rightOpTranslation.getTerm();

	    log("opTermSortLeft == " + leftOp.sort());
	    log("opTermSortRight == " + rightOp.sort());

	    // Get the kind of translation to be performed for the given node ...
	    TranslationType translationType = getTranslationTypeForNode(re);
	    log("Current equality expression should translated in the following way : " + translationType.toString());
	

	    // First translate as equality ...
	    
	    // We have to take care of the different possible sorts ...
	    Sort opSort;
	    Equality eqFunction = null;
	    Term eqFunctionTerm;
	    if (t instanceof Basic) {
		if ((t==Basic.INTEGER) || (t==Basic.REAL) || (t==Basic.STRING) ){
		    opSort = getDLSortForType(t);
		    eqFunction = opSort.getEqualitySymbol();
		}
		else if (t==Basic.BOOLEAN) {

		    if (!translationType.containsMark(TranslationMark.AS_BOOLEAN_TERM)) {
			// Use formulas to represent logical OCL expressions in the returned translation result ...
		       
			opSort = Sort.FORMULA;
				
			eqFunction = opSort.getEqualitySymbol();
			leftOp = convertTermToFormula(leftOp);
			rightOp = convertTermToFormula(rightOp);

		    } else {
			// Create some term over the ADT (or the JAVA-Type) 'boolean' instead of formulas.

			Function eqSymbol;
			Sort boolSort =  proverFacade.getSortForOCLBoolean();

			eqSymbol =   proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_EQUALS_TO_BOOLEAN,
									  new Sort[]{boolSort,boolSort},
									  boolSort); 

			leftOp = convertFormulaToTerm(leftOp, leftOpTranslation);
			rightOp = convertFormulaToTerm(rightOp, rightOpTranslation);

			resultTerm = tf.createFunctionTerm(eqSymbol, leftOp, rightOp);
		    
		    }

		}
		else throw new RuntimeException("illegal basic type");

		if (resultTerm == null) {
		    resultTerm = tf.createEqualityTerm(eqFunction, leftOp, rightOp);
		}

	    } else if (t instanceof Collection) {
		Collection c=(Collection) t;
		Sort elemSort = getElementSortForCollectionType(c);
		
		switch (c.getCollectionKind()) {
		case Collection.SET: 
		    // Now we translate the equality on sets ...
		    
		//      // ============ TRANSLATION AS TERM ============================

//  		    opSort = getDLSortForType(t);
//  		    eqFunction = opSort.getEqualitySymbol();
		    
//  		    resultTerm = tf.createEqualityTerm(eqFunction,
//  						       leftOpTranslation.getTerm(),
//  						       rightOpTranslation.getTerm());
		

		    //  log("leftOpTranslation is predicative == " + leftOpTranslation.isPredicativeTranslation());
//  		    log("rightOpTranslation is predicative == " + rightOpTranslation.isPredicativeTranslation());

		    //================== OLD PREDICTAIVE ALTERNATIVE ================
		    //  Convert the two translations to predicative translations ...
		    leftOpTranslation = getPredicativeTranslationOfTranslation(leftOpTranslation,c);
		    rightOpTranslation = getPredicativeTranslationOfTranslation(rightOpTranslation,c);
		    
		    // Get quantor variable ...
		    LogicVariable newQuantorVariable = leftOpTranslation.getPredicateVariable();
		    // Rename the predicate variables according to the quantor variable ...
		    LogicVariable varToRename = rightOpTranslation.getPredicateVariable();
    
		    
		    aTerm = renameVariableInTerm(rightOpTranslation.getTerm(), varToRename, newQuantorVariable);
		    aTerm = tf.createJunctorTerm(Op.EQV,leftOpTranslation.getTerm(), aTerm);

		    // OLD:  resultTerm = tf.createQuantifierTerm(Op.ALL, newQuantorVariable, aTerm);
		    // NEW:
		    resultTerm = getQuantifiedTermWith(Op.ALL, newQuantorVariable, aTerm);
		   

		    break;

		case Collection.BAG:
		   
		    Collection collType = (Collection) operandTypeLeft;
		    Term opTermLeft = leftOpTranslation.getTerm();
		    Term opTermRight = rightOpTranslation.getTerm();
 	
		    elemSort = getElementSortForCollectionType(collType);
		    Sort collSort = getDLSortForType(collType);
		    Sort intSort =  proverFacade.getSortForOCLInteger();
	   
		    LogicVariable predicateVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		    Term predicateVarTerm = tf.createVariableTerm(predicateVar);

	    
		    Function bagCountFunction = proverFacade.getCollectionFunctionForFeature(COUNT,
										       new Sort[]{collSort,elemSort},
										       intSort);
	    

	   
		    Term countTerm1 = tf.createFunctionTerm(bagCountFunction,opTermLeft,predicateVarTerm);
		    Term countTerm2 = tf.createFunctionTerm(bagCountFunction,opTermRight,predicateVarTerm);	

		    Term eqTerm = tf.createEqualityTerm(countTerm1,countTerm2);
		    // resultTerm = tf.createQuantifierTerm(Op.ALL,predicateVar, eqTerm);
		    // NEW:
		    resultTerm = getQuantifiedTermWith(Op.ALL, predicateVar, eqTerm);
		    break;

		case Collection.SEQUENCE:
		    // This is ok, since the term representation of 'equal' sequences is identical ...
		    
		    opSort = getDLSortForType(t);
		    eqFunction = opSort.getEqualitySymbol();
		    
		    resultTerm = tf.createEqualityTerm(eqFunction,
						       leftOpTranslation.getTerm(),
						       rightOpTranslation.getTerm());
		

		    

		    break;
		default: throw new RuntimeException("illegal collection type");
		}
	    } else if (t instanceof OclType) {
		error("'=' or '<>' on OCLTYPE not handled yet !!!!!!!!!!!!!!!!");
	    } else {
		// we have an UML model type ...
		opSort = getDLSortForType(t);
		eqFunction = opSort.getEqualitySymbol();
		resultTerm = tf.createEqualityTerm(eqFunction, leftOpTranslation.getTerm(),
						   rightOpTranslation.getTerm() );
	    }




	    // Finally, if we have an inequality , negate the translation with equality ...
	    if(relationalOperatorName.equals("<>") ){

		if (!translationType.containsMark(TranslationMark.AS_BOOLEAN_TERM)) {
		    // Use formulas to represent logical OCL expressions in the returned translation result ...
		    resultTerm = tf.createJunctorTerm(Op.NOT,resultTerm);

		} else {
		    // Create some term over the ADT (or the JAVA-Type) 'boolean' instead of formulas.
		    
		    Function notSymbol;
		    Sort boolSort =  proverFacade.getSortForOCLBoolean();
	 
		    notSymbol =   proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_NOT,
								       new Sort[]{boolSort},
								       boolSort); 

		    resultTerm = tf.createFunctionTerm(notSymbol, resultTerm);
		}
	    }

	    result = new KeyTranslationResult(resultTerm);
	    result.addAxioms(leftOpTranslation.getAxioms());
	    result.addAxioms(rightOpTranslation.getAxioms());
	    
	    // Now set the translation result ...
	    setTranslationForNode(re, result);
	    
	} else if(relationalOperatorName.equals("<") || relationalOperatorName.equals(">=") ){
	    Sort opSort;
	    Function relOpFunction;
	    opSort = getDLSortForType(operandTypeLeft);

	    String functionName = (relationalOperatorName.equals("<")) ? LESS_THAN : GREATER_OR_EQUAL_THAN;
	    relOpFunction = proverFacade.getFunctionForFeature(functionName,
								 new Sort[]{opSort,opSort},
								 Sort.FORMULA);

	    resultTerm = tf.createFunctionTerm(relOpFunction,
					       leftOpTranslation.getTerm(),
					       rightOpTranslation.getTerm());


	   
	    result = new KeyTranslationResult(resultTerm);
	    result.addAxioms(leftOpTranslation.getAxioms());
	    result.addAxioms(rightOpTranslation.getAxioms());
	    
	    // Now set the translation result ...
	    setTranslationForNode(re, result);
	
	} else if(relationalOperatorName.equals(">") || relationalOperatorName.equals("<=") ){
	    Sort opSort;
	    Function relOpFunction;
	    opSort = getDLSortForType(operandTypeLeft);
	    String functionName = (relationalOperatorName.equals(">")) ? GREATER_THAN : LESS_OR_EQUAL_THAN;
	    relOpFunction = proverFacade.getFunctionForFeature(functionName,
								 new Sort[]{opSort,opSort},
								 Sort.FORMULA);

	    resultTerm = tf.createFunctionTerm(relOpFunction,
					       leftOpTranslation.getTerm(),
					       rightOpTranslation.getTerm());


	    result = new KeyTranslationResult(resultTerm);
	    result.addAxioms(leftOpTranslation.getAxioms());
	    result.addAxioms(rightOpTranslation.getAxioms());
	    
	    // Now set the translation result ...
	    setTranslationForNode(re, result);
	
	} else  {
	    error("Relational Operator '" + relationalOperatorName +"' not supported yet !!!");
	}

	
    }

    // -(...), not
    protected TranslationResult translateUnaryExpression(Node ue,
					    String unaryOperatorName,
					    TranslationResult operand){
	log("Translating unary expression with Operator '"+ unaryOperatorName  +"'.");
	KeyTranslationResult opTranslation = (KeyTranslationResult) operand;
	if(unaryOperatorName.equals("not")){
	    Term resultTerm, opTerm;
	    
	    opTerm = opTranslation.getTerm();
	    
	    // Get the kind of translation to be performed for the given node ...
	    TranslationType translationType = getTranslationTypeForNode(ue);
	    log("Current unary boolean expression should translated in the following way : " + translationType.toString());
	
	    if (!translationType.containsMark(TranslationMark.AS_BOOLEAN_TERM)) {
		// Use formulas to represent logical OCL expressions in the returned translation result ...

		Term t = convertTermToFormula(opTerm);
		resultTerm = tf.createJunctorTerm(Op.NOT,t);
	 
	    } else {
		// Create some term over the ADT (or the JAVA-Type) 'boolean' instead of formulas.

		Function notSymbol;
		Sort boolSort =  proverFacade.getSortForOCLBoolean();

		notSymbol = proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_NOT,
								       new Sort[]{boolSort},
								       boolSort);
		resultTerm = tf.createFunctionTerm(notSymbol,opTerm);
	    }


	    log("INTO : " + resultTerm.toString());

	    KeyTranslationResult translation = new KeyTranslationResult(resultTerm);
	    // add all axioms for subtranslations ...
	    translation.addAxioms(opTranslation.getAxioms());
	    
	    return translation;

	} else if(unaryOperatorName.equals("-")){
	    KeyTranslationResult result = new KeyTranslationResult();
	    Term resultTerm, opTerm;
	    opTerm = opTranslation.getTerm();

	    Sort opSort = opTerm.sort();
	    Sort retSort = opSort;
	    
	    Function monadicMinusFunction =
		proverFacade.getFunctionForFeature(UNARY_MINUS,
						     new Sort[]{opSort},
						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(monadicMinusFunction,opTerm);
	    
	    result.setTerm(resultTerm);
	    result.addAxioms(opTranslation.getAxioms());
	    result.addAxioms(opTranslation.getAxioms());
	    return result;
	    
	}

	error("Unary Operator '" + unaryOperatorName + "' NOT supported yet ...");
	return null;
    }

    

    

    // the translations for the features of OCL types ...
        

   // =========================================================================================================
   // ================================= The feature calls on OCL type Set[T] ==================================
   // =========================================================================================================


    protected void translateFeatureCallonSet(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType, boolean  featureCallWithAtPre){
	log("Feature call on Set : " + featureName);
	log("Applied on node :" + appliedTo.toString());
	KeyTranslationResult result = null;
	Term resultTerm = null;
	
	// Get the kind of translation to be performed for the given node ...
	TranslationType translationType = getTranslationTypeForNode(n);
	log("Current feature call should be translated in the following way : " + translationType.toString());
	
	Sort iteratorSort;

	if(featureName.equals("including") || featureName.equals("excluding") ){
	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult elementTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term setTerm = setTranslation.getTerm();

 	    Term elementTerm = elementTranslation.getTerm();
	    
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = collSort;

	    if(elemSort == proverFacade.getSortForOCLBoolean()){
		// Ensure the representation of the expression to be a boolean term (not a formula) ...
		elementTerm = convertFormulaToTerm(elementTerm, elementTranslation);
	    }

	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		String funcName = (featureName.equals("including")) ? SET_INSERT : SET_REMOVE; 
		
		Function setFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(funcName,
								   new Sort[]{collSort,elemSort},
								   retSort);
		
		resultTerm = tf.createFunctionTerm(setFeatureFunction,setTerm,elementTerm);
		
		log("INTO :" + resultTerm.toString());
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(elementTranslation.getAxioms());
		result.addAxioms(setTranslation.getAxioms());

	    } else {
		// ==================== PREDICATIVE TRANSLATION ==========================

		// use a predicative translation for the set on which 'including' is applied ...
		setTranslation = 
		    getPredicativeTranslationOfTranslation(setTranslation,collType);
		
		Term setPredicate = setTranslation.getTerm();
		
		LogicVariable predicateVar = setTranslation.getPredicateVariable();
		Term varTerm = tf.createVariableTerm(predicateVar);
		Term eqTerm = null;
		
		if (featureName.equals("including")){
		    // Create new predicate for including ...
		    eqTerm = tf.createEqualityTerm(varTerm, elementTerm);
		    resultTerm = tf.createJunctorTermAndSimplify(Op.OR, setPredicate, eqTerm);
		} else {
		    // Create new predicate for excluding ...
		    eqTerm = tf.createEqualityTerm(varTerm, elementTerm);
		    eqTerm = tf.createJunctorTerm(Op.NOT,eqTerm);
		    resultTerm = tf.createJunctorTermAndSimplify(Op.AND, setPredicate, eqTerm);
		}
		
		log("INTO :" + resultTerm.toString());
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(elementTranslation.getAxioms());
		result.addAxioms(setTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);

	    }

	} else if(featureName.equals("forAll") ||featureName.equals("exists") ){
	    // Translate forAll and exists on Sets ...

	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult formulaTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    Collection collType = (Collection) tree.getNodeType(appliedTo);

	    // use a predicative translation for the set on which 'forAll' is applied ...
 	    setTranslation = 
		getPredicativeTranslationOfTranslation(setTranslation,collType);
	    
	    Term setPredicate = setTranslation.getTerm();
	    Term formulaTerm = convertTermToFormula(formulaTranslation.getTerm());
	    LogicVariable predicateVar = setTranslation.getPredicateVariable();
	    
	    log(featureName + "formulaTerm : " + formulaTerm.toString());

	    

	    // We have exactely one iterator variable here !!!!! ...
	    String iteratorVarName = (String)iteratorNames.elementAt(0);
	    String iteratorTypeName = iteratorType.toString().trim();

	    // Get the translation of the iteratorVariable ...
	    KeyTranslationResult iteratorTranslation = 
		(KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								 iteratorTypeName);
	    
	    Term iteratorTerm = iteratorTranslation.getTerm();
	    LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
	    

	    formulaTerm = renameVariableInTerm(formulaTerm, iteratorVar, predicateVar);

	    log(featureName + "formulaTerm after renaming: " + formulaTerm.toString());
	    Term aTerm = null;

	    if (featureName.equals("forAll")){
		// Create formula for forAll ...
		aTerm = tf.createJunctorTermAndSimplify(Op.IMP,setPredicate,formulaTerm);
		// resultTerm = tf.createQuantifierTerm(Op.ALL,predicateVar,aTerm);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);

	    } else {
		// Create formula for exists ...
		aTerm = tf.createJunctorTermAndSimplify(Op.AND,setPredicate,formulaTerm);
		// resultTerm = tf.createQuantifierTerm(Op.EX,predicateVar,aTerm);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , aTerm);
	    }

	    
	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(formulaTranslation.getAxioms());
	    result.addAxioms(setTranslation.getAxioms());

	} else if(featureName.equals("isEmpty") ||featureName.equals("notEmpty") ){
	    // Translate isEmpty and notEmpty on Sets ...

	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);

	    // use a predicative translation for the set on which 'forAll' is applied ...
 	    setTranslation = 
		getPredicativeTranslationOfTranslation(setTranslation,collType);
	    
	    Term setPredicate = setTranslation.getTerm();
	    LogicVariable predicateVar = setTranslation.getPredicateVariable();
	    
	    Term aTerm;


	    if (featureName.equals("isEmpty")){
		// Create formula for isEmpty ...
		aTerm = tf.createJunctorTerm(Op.NOT,setPredicate);
		// resultTerm = tf.createQuantifierTerm(Op.ALL,predicateVar,aTerm);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);
	    } else {
		// Create formula for notEmpty ...
		// resultTerm = tf.createQuantifierTerm(Op.EX,predicateVar,setPredicate);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , setPredicate);
	    }

	    
	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(setTranslation.getAxioms());

	} else if(featureName.equals("includes")|| (featureName.equals("excludes") )){
	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult elementTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term setTerm = setTranslation.getTerm();
 	    Term elementTerm = elementTranslation.getTerm();

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = proverFacade.getSortForOCLBoolean();
	   
	    Function setFeatureFunction =
 		proverFacade.getCollectionFunctionForFeature(CONTAINS,
 						     new Sort[]{collSort,elemSort},
 						     Sort.FORMULA);

	  
	    resultTerm = tf.createFunctionTerm(setFeatureFunction,setTerm,elementTerm);
		
	    if(featureName.equals("excludes") ){
		// we have to negate the formula ...
		resultTerm = tf.createJunctorTerm(Op.NOT,resultTerm);
	    }
 	 

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(setTranslation.getAxioms());
	    result.addAxioms(elementTranslation.getAxioms());
	   
	    
	}  else if(featureName.equals("includesAll") || (featureName.equals("excludesAll") )){
	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);

	    Node collArgNode = (Node) parameterNodes.elementAt(0);

	    KeyTranslationResult collTranslation =
		(KeyTranslationResult) getTranslationForNode(collArgNode);
	    
	    Collection collType1 = (Collection) tree.getNodeType(appliedTo);
	    Collection collType2 = (Collection) tree.getNodeType(collArgNode);
	    
	    // use a predicative translation for the set on which the feature is applied ...
 	    setTranslation = 
		getPredicativeTranslationOfTranslation(setTranslation,collType1);
	    collTranslation = 
		getPredicativeTranslationOfTranslation(collTranslation,collType2);

	    Term setPredicate = setTranslation.getTerm();
 	    Term collPredicate = collTranslation.getTerm();

	    Sort elemSort1 = getElementSortForCollectionType(collType1);
	    Sort collSort1 = getDLSortForType(collType1);
	    Sort elemSort2 = getElementSortForCollectionType(collType2);
	    Sort collSort2 = getDLSortForType(collType2);

	    // Sort retSort = proverFacade.getSortForOCLBoolean();
	   
	    if (elemSort1 != elemSort2) {
		log("includesAll/excludesAll on collections with DIFFERENT element types is currently not supported !!!!");
	    }

	    LogicVariable predicateVar1 = setTranslation.getPredicateVariable();
	    LogicVariable predicateVar2 = collTranslation.getPredicateVariable();
	    LogicVariable predicateVar = predicateVar1;

	    collPredicate = renameVariableInTerm(collPredicate, predicateVar2, predicateVar);
	    
	    // log(featureName + "collPredicate after renaming: " + collPredicate.toString());
	    Term aTerm = null;
	    
	    if (featureName.equals("excludesAll")){
		// Create formula for excludesAll ...
		aTerm = tf.createJunctorTerm(Op.NOT,setPredicate);
	    }  else {
		aTerm = setPredicate;
	    }
	    
	    aTerm = tf.createJunctorTermAndSimplify(Op.IMP,collPredicate,aTerm);
	    // resultTerm = tf.createQuantifierTerm(Op.ALL,predicateVar,aTerm);
	    // NEW:
	    resultTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(setTranslation.getAxioms());
	    result.addAxioms(collTranslation.getAxioms());

	} else if(featureName.equals("size") ){
	    
	    // ==================== TRANSLATION AS TERM ==============================
	    
	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Term setTerm = setTranslation.getTerm();

 	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = proverFacade.getSortForOCLInteger();

	    Function setFeatureFunction =
		proverFacade.getCollectionFunctionForFeature(SIZE,
						     new Sort[]{collSort},
						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(setFeatureFunction,setTerm);
		

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(setTranslation.getAxioms());
	    
	        
	} else if(featureName.equals("count")){
	    
	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult elementTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term setTerm = setTranslation.getTerm();
 	    Term elementTerm = elementTranslation.getTerm();

	    // ==================== TRANSLATION AS TERM ==============================
	    
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = proverFacade.getSortForOCLInteger();
	   
	    if(elemSort == proverFacade.getSortForOCLBoolean()){
		// Ensure the representation of the expression to be a boolean term (not a formula) ...
		elementTerm = convertFormulaToTerm(elementTerm, elementTranslation);
	    }

	    Function setFeatureFunction =
 		proverFacade.getCollectionFunctionForFeature(COUNT,
 						     new Sort[]{collSort,elemSort},
 						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(setFeatureFunction,setTerm,elementTerm);

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(setTranslation.getAxioms());
	    result.addAxioms(elementTranslation.getAxioms());
	   
	    
	} else if(featureName.equals("sum") ){
	    
	    // ==================== TRANSLATION AS TERM ==============================
	    
	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Term setTerm = setTranslation.getTerm();

 	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Sort collSort = getDLSortForType(collType);
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort retSort = elemSort;

	    if ( (elemSort != proverFacade.getSortForOCLInteger()) && 
		 (elemSort != proverFacade.getSortForOCLReal()) ){
		// We don't allow the sum feature for other element types than OclInteger or OclReal ...
		error("The feature sum on collections is only allowed for collections with elements of type Integer or Real !!!!");
	    }

	    Function setFeatureFunction =
		proverFacade.getCollectionFunctionForFeature(SUM,
						     new Sort[]{collSort},
						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(setFeatureFunction,setTerm);
		

	    log("INTO :" + resultTerm.toString());

	
	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(setTranslation.getAxioms());
	        
	} else if(featureName.equals("union") || (featureName.equals("intersection") )){
	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);

	    Node collArgNode = (Node) parameterNodes.elementAt(0);

	    KeyTranslationResult collTranslation =
		(KeyTranslationResult) getTranslationForNode(collArgNode);
	    
	    Collection collType1 = (Collection) tree.getNodeType(appliedTo);
	    Collection collType2 = (Collection) tree.getNodeType(collArgNode);
	    
	    
	    Term setTerm = setTranslation.getTerm();
 	    Term collTerm = collTranslation.getTerm();

	    Sort elemSort1 = getElementSortForCollectionType(collType1);
	    Sort collSort1 = getDLSortForType(collType1);
	    Sort elemSort2 = getElementSortForCollectionType(collType2);
	    Sort collSort2 = getDLSortForType(collType2);

	    Sort retSort;

	    if (elemSort1 != elemSort2) {
		log("union/intersection on collections with DIFFERENT element types is currently not supported !!!!");
	    }
	    
	    if(!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
	    
		String funcName;
		if (featureName.equals("union")){
		    retSort = collSort2;
		    funcName = SET_UNION;
		} else {
		    retSort = collSort1;
		    funcName = SET_INTERSECTION;
		}
		
	       
		Function setFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(funcName,
								   new Sort[]{collSort1,collSort2},
								   retSort);
		
		resultTerm = tf.createFunctionTerm(setFeatureFunction,setTerm,collTerm);
		
		log("INTO :" + resultTerm.toString());
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(setTranslation.getAxioms());
		result.addAxioms(collTranslation.getAxioms());
	    } else {
		
		// ==================== TRANSLATION AS PREDICATE  ==============================
		setTranslation = 
		    getPredicativeTranslationOfTranslation(setTranslation,collType1);
		collTranslation = 
		    getPredicativeTranslationOfTranslation(collTranslation,collType2);

		Term setPredicate = setTranslation.getTerm();
		Term collPredicate = collTranslation.getTerm();

		retSort = proverFacade.getSortForOCLBoolean();
	   
		LogicVariable predicateVar1 = setTranslation.getPredicateVariable();
		LogicVariable predicateVar2 = collTranslation.getPredicateVariable();
		LogicVariable predicateVar = predicateVar1;

		collPredicate = renameVariableInTerm(collPredicate, predicateVar2, predicateVar);
	    
		// log(featureName + "collPredicate after renaming: " + collPredicate.toString());
		Term aTerm = null;
	    
		if (featureName.equals("union")){
		    // Create formula for union ...
		    aTerm = tf.createJunctorTermAndSimplify(Op.OR,setPredicate, collPredicate);
		}  else {
		    // Create formula for intersection ...
		    aTerm = tf.createJunctorTermAndSimplify(Op.AND,setPredicate, collPredicate);
		}
	    
	   
		resultTerm = aTerm;

		log("INTO :" + resultTerm.toString());

		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(setTranslation.getAxioms());
		result.addAxioms(collTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    
	    
	    }
	    
	}  else if(featureName.equals("select") ||featureName.equals("reject") ){
	    // Translate select and reject on Sets ...

	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult formulaTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));

	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term setTerm = setTranslation.getTerm();
 	    Term formulaTerm = convertTermToFormula(formulaTranslation.getTerm());

	    // ==================== TRANSLATION AS TERM ==============================
	    
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = collSort;
	    
	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
	   
		// Get the iterator variable (and the corresponding term) ...

		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		log("IteratorVarName == " + iteratorVarName);
		log("IteratorTypeName == " + iteratorTypeName);

		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		log("IteratorTranslation == " + iteratorTranslation);
		
		Term iteratorTerm = iteratorTranslation.getTerm();

		log("iteratorVar == " + iteratorTerm.op());

		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		

		String functionName = newFunctionNameDeliverer.getNewFunctionNameForName("select");
		
		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable freeVarsExpr = formulaTerm.freeVars();
		SetOfQuantifiableVariable paramsForNewFunction = freeVarsExpr.remove(iteratorVar);
		
	
		Term[] paramTerms;
		Term[] variableParamTerms;
		Sort[] paramSorts;
		Sort[] variableParamSorts;
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt + 1];
		paramSorts = new Sort[freeVarsCnt + 1];
		
		variableParamTerms = new Term[freeVarsCnt];
		variableParamSorts = new Sort[freeVarsCnt];

		// The first parameter is the collection the feature is applied to ...
		paramTerms[0] = setTerm;
		paramSorts[0] = collSort;

		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=1;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    variableParamTerms[i-1] = paramTerms[i];
		    variableParamSorts[i-1] = paramSorts[i];
		    i++;
		}
		    
		
		Function setFeatureFunction =
 		proverFacade.getCollectionFunctionForFeature(functionName,
							       paramSorts,
							       retSort);
		
		resultTerm = tf.createFunctionTerm(setFeatureFunction,paramTerms);
		
		log("INTO :" + resultTerm.toString());
		
		// Create additional axioms for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2, newAxiom3;
		Term[] currentArgs;
		Sort[] currentSorts;

		// Now get the corresponding constant for the empty collection ...
		Function emptyCollFunction = proverFacade.getCollectionFunctionForFeature(SET_EMPTYSET,
											    new Sort[0],
											    retSort);
		
		Function collInsertFunction = proverFacade.getCollectionFunctionForFeature(SET_INSERT,
											     new Sort[]{retSort,elemSort},
											     retSort);
	
		Term emptyCollTerm = tf.createFunctionTerm(emptyCollFunction);

		
		
		// Create necessary variables and variable terms ...
		LogicVariable elementVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		Term elementVarTerm = tf.createVariableTerm(elementVar);
		
		LogicVariable collVar = newVariableDeliverer.getNewVariableForSort(collSort);
		Term collVarTerm = tf.createVariableTerm(collVar);
		
		
		// Set up the formuls term with the 'right' iteratorvariable ...
		
		log(featureName + "formulaTerm : " + formulaTerm.toString());

		formulaTerm = renameVariableInTerm(formulaTerm, iteratorVar, elementVar);
		
		log(featureName + "formulaTerm after renaming: " + formulaTerm.toString());
		
		// Create the boolean expression to be used for the test ...
		Term expressionTerm, negExpressionTerm; 
		if (featureName.equals("select")) {
		    expressionTerm = formulaTerm;
		    negExpressionTerm = tf.createJunctorTerm(Op.NOT,formulaTerm);
		} else {
		    expressionTerm = tf.createJunctorTerm(Op.NOT,formulaTerm);
		    negExpressionTerm = formulaTerm;
		}
		
		// Now we can build the additional axioms ...
		
		// No.1 ...
	
		currentArgs = prependTerm(emptyCollTerm,variableParamTerms);
		
		Term selectTerm = tf.createFunctionTerm(setFeatureFunction, currentArgs);
		Term eqTerm0 = tf.createEqualityTerm(selectTerm, emptyCollTerm);
		newAxiom1 = constraintTranslator.createAllClosureOfTerm(eqTerm0); 
		
		// No.2 ...
		
		Term insertTerm1 =  tf.createFunctionTerm(collInsertFunction, collVarTerm, elementVarTerm);
		currentArgs = prependTerm(insertTerm1,variableParamTerms);
		Term selectTerm1 =  tf.createFunctionTerm(setFeatureFunction, currentArgs);
		currentArgs = prependTerm(collVarTerm,variableParamTerms);
		Term selectTerm2 =  tf.createFunctionTerm(setFeatureFunction, currentArgs);
		Term insertTerm2 =  tf.createFunctionTerm(collInsertFunction, selectTerm2, elementVarTerm);
		Term eqTerm = tf.createEqualityTerm(selectTerm1,insertTerm2);
		
		Term impTerm = tf.createJunctorTermAndSimplify(Op.IMP, expressionTerm, eqTerm);
		newAxiom2 = constraintTranslator.createAllClosureOfTerm(impTerm);
		
		// No.3 ...
		eqTerm = tf.createEqualityTerm(selectTerm1,selectTerm2);
		
		impTerm = tf.createJunctorTermAndSimplify(Op.IMP, negExpressionTerm, eqTerm);
		newAxiom3 = constraintTranslator.createAllClosureOfTerm(impTerm);
		
	    
		log("Adding new axiom (1)  for " + functionName + " : " + newAxiom1);
		log("Adding new axiom (2)  for " + functionName + " : " + newAxiom2);
		log("Adding new axiom (3)  for " + functionName + " : " + newAxiom3);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(setTranslation.getAxioms());
		result.addAxioms(formulaTranslation.getAxioms());
		result.addAxiom(newAxiom1);
		result.addAxiom(newAxiom2);
		result.addAxiom(newAxiom3);

	    } else {
		// ==================== TRANSLATION AS PREDICATE  ==============================
		setTranslation = 
		    getPredicativeTranslationOfTranslation(setTranslation,collType);
		
		Term setPredicate = setTranslation.getTerm();
		
		LogicVariable predicateVar = setTranslation.getPredicateVariable();
		
		// Set up the formuls term with the 'right' iteratorvariable for the formula term ...
		
		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		
		formulaTerm = renameVariableInTerm(formulaTerm, iteratorVar, predicateVar);
	    
		Term aTerm = null;
	    
		if (featureName.equals("select")){
		    // Create filter formula for select  ...
		    aTerm = formulaTerm;
		}  else {
		    // Create filter formula for reject  ...
		    aTerm = tf.createJunctorTerm(Op.NOT,formulaTerm);
		}
	    
	   
		resultTerm = tf.createJunctorTermAndSimplify(Op.AND, setPredicate, aTerm);

		log("INTO :" + resultTerm.toString());

		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(setTranslation.getAxioms());
		result.addAxioms(formulaTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    	    
	    }



	}  else if(featureName.equals("collect")){
	    // Translate collect on sets ...

	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult expressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term setTerm = setTranslation.getTerm();

	    Type expressionType = tree.getNodeType((Node) parameterNodes.elementAt(0));
 	    Term expressionTerm;

	    // Convert the representation of the expression to be computed in
	    // some suitable form ...
	    if (expressionType == Basic.BOOLEAN) {
		// We have to ensure to use boolean terms (instead of formulas) ...
		expressionTerm = convertFormulaToTerm(expressionTranslation.getTerm(), 
						      expressionTranslation);
	    } else {
		expressionTerm = expressionTranslation.getTerm();
	    }
	    
	   
	    
	    boolean flatteningRequired = false;
	    if (expressionType instanceof Collection) {
		log("Working on a collect producing collections => Flattening is required ...");
		flatteningRequired = true;
	    }

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);

	    Sort exprSort = expressionTerm.sort();
	    NonCollectionSort exprElementSort;
	    Sort retSort;
	    
	    if (!flatteningRequired){
		// the given collect expression has a BASIC type and therefore no flattening is required ...
		exprElementSort = (NonCollectionSort) exprSort;
		retSort = proverFacade.getSortForOCLBag(exprElementSort);
	    } else {
		exprElementSort = getElementSortForCollectionType((Collection) expressionType);
		retSort = proverFacade.getSortForOCLBag(exprElementSort);
	    }

	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		
	    
		// Get the iterator variable (and the corresponding term) ...

		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		

		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable freeVarsExpr =  expressionTerm.freeVars();
		SetOfQuantifiableVariable paramsForNewFunction = freeVarsExpr.remove(iteratorVar);
		
	
		Term[] paramTerms;
		Term[] variableParamTerms;
		Sort[] paramSorts;
		Sort[] variableParamSorts;
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt + 1];
		paramSorts = new Sort[freeVarsCnt + 1];
		
		variableParamTerms = new Term[freeVarsCnt];
		variableParamSorts = new Sort[freeVarsCnt];

		// The first parameter is the collection the feature is applied to ...
		paramTerms[0] = setTerm;
		paramSorts[0] = collSort;

		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=1;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    variableParamTerms[i-1] = paramTerms[i];
		    variableParamSorts[i-1] = paramSorts[i];
		    i++;
		}
		    
		String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);
		
		Function setFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(functionName,
								   paramSorts,
								   retSort);

		
		resultTerm = tf.createFunctionTerm(setFeatureFunction,paramTerms);
		
		log("INTO :" + resultTerm.toString());
		
		
		// Create additional axioms for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2;
		Term[] currentArgs;
		Sort[] currentSorts;

		// Now get the corresponding constant for the empty collection ...
		Function emptySetFunction = proverFacade.getCollectionFunctionForFeature(SET_EMPTYSET,
											   new Sort[0],
											   collSort);
		
		Function emptyBagFunction = proverFacade.getCollectionFunctionForFeature(BAG_EMPTYBAG,
											   new Sort[0],
											   retSort);
		
		Function setInsertFunction = proverFacade.getCollectionFunctionForFeature(SET_INSERT,
											    new Sort[]{collSort,elemSort},
											    collSort);

		Function setRemoveFunction = proverFacade.getCollectionFunctionForFeature(SET_REMOVE,
											    new Sort[]{collSort,elemSort},
											    collSort);
		
	
		Term emptySetTerm = tf.createFunctionTerm(emptySetFunction);
		Term emptyBagTerm = tf.createFunctionTerm(emptyBagFunction);
		
		
		// Create necessary variables and variable terms ...
		LogicVariable elementVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		Term elementVarTerm = tf.createVariableTerm(elementVar);
		
		LogicVariable collVar = newVariableDeliverer.getNewVariableForSort(collSort);
		Term collVarTerm = tf.createVariableTerm(collVar);
		
		
		// Set up the formulas term with the 'right' iteratorvariable ...
		
		log(featureName + "expressionTerm : " + expressionTerm.toString());
			
		expressionTerm = renameVariableInTerm(expressionTerm, iteratorVar, elementVar);
		
		log(featureName + "expressionTerm after renaming: " + expressionTerm.toString());
		

		// Now we can build the additional axioms ...
		
		// No.1 ...
		currentArgs = prependTerm(emptySetTerm,variableParamTerms);

		Term collectTerm = tf.createFunctionTerm(setFeatureFunction, currentArgs);
		Term eqTerm0 = tf.createEqualityTerm(collectTerm, emptyBagTerm);
		newAxiom1 = constraintTranslator.createAllClosureOfTerm(eqTerm0);
 
		// No.2 ...
		Term insertTerm1 =  tf.createFunctionTerm(setInsertFunction, collVarTerm, elementVarTerm);
		Term removeTerm = tf.createFunctionTerm(setRemoveFunction, collVarTerm, elementVarTerm);
		currentArgs = prependTerm(insertTerm1, variableParamTerms);
		Term collectTerm1 =  tf.createFunctionTerm(setFeatureFunction, currentArgs);
		currentArgs = prependTerm(removeTerm, variableParamTerms);
		Term collectTerm2 =  tf.createFunctionTerm(setFeatureFunction, currentArgs);
		
		Term insertOrUnionTerm;
		
		if (!flatteningRequired){
		    
		    Function bagInsertFunction = 
			proverFacade.getCollectionFunctionForFeature(BAG_INSERT,
								       new Sort[]{retSort,
										  exprSort},
								       retSort);
		    
		    insertOrUnionTerm = tf.createFunctionTerm(bagInsertFunction,collectTerm2,expressionTerm);
		    
		} else {
		    Function bagUnionFunction = 
			proverFacade.getCollectionFunctionForFeature(BAG_UNION,
								       new Sort[]{retSort,
										  exprSort},
								       retSort);
		    insertOrUnionTerm = tf.createFunctionTerm(bagUnionFunction,collectTerm2,expressionTerm);
		}
		Term eqTerm = tf.createEqualityTerm(collectTerm1,insertOrUnionTerm);
		
		newAxiom2 = constraintTranslator.createAllClosureOfTerm(eqTerm);
		
		
		log("Adding new axiom (1)  for " + functionName + " : " + newAxiom1);
		log("Adding new axiom (2)  for " + functionName + " : " + newAxiom2);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(setTranslation.getAxioms());
		result.addAxioms(expressionTranslation.getAxioms());
		result.addAxiom(newAxiom1);
		result.addAxiom(newAxiom2);
		
	    } else {
		// ==================== TRANSLATION AS PREDICATE  ==============================
		
		setTranslation = 
		    getPredicativeTranslationOfTranslation(setTranslation,collType);
			
		Term setPredicate = setTranslation.getTerm();
		
		LogicVariable predicateVar = setTranslation.getPredicateVariable();
		
		// Set up the formuls term with the 'right' iteratorvariable for the formula term ...
		
		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		
		
		// Create necessary new variables and variable terms ...
		LogicVariable newPredicateVar  = newVariableDeliverer.getNewVariableForSort(exprElementSort);
		Term newPredicateVarTerm = tf.createVariableTerm(newPredicateVar);
		
		Term aTerm = null;
	    
		if (!flatteningRequired){
		    // the given collect expression has a BASIC type and therefore no flattening is required ...
		   
		    expressionTerm = renameVariableInTerm(expressionTerm, iteratorVar, predicateVar);
		    
		    Term eqTerm = tf.createEqualityTerm(newPredicateVarTerm, expressionTerm);
		    aTerm = tf.createJunctorTermAndSimplify(Op.AND, setPredicate, eqTerm);
		    
		    resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , aTerm);


		} else {
		    // the given collecting expression has a COLLECTION type => Flattening is required ...
		   
		    if (expressionTranslation.isPredicativeTranslation()) {
			expressionTranslation = 
			    getPredicativeTranslationOfTranslation(expressionTranslation, (Collection) expressionType);
		    } else {
			expressionTerm = renameVariableInTerm(expressionTerm, iteratorVar, predicateVar);
			expressionTranslation.setTerm(expressionTerm);
			expressionTranslation = 
			    getPredicativeTranslationOfTranslation(expressionTranslation, (Collection) expressionType);
		    }
			
		    Term expressionPredicate = expressionTranslation.getTerm();
		    LogicVariable expressionPredicateVar = expressionTranslation.getPredicateVariable();
	       
		    expressionPredicate = renameVariableInTerm(expressionPredicate, expressionPredicateVar, newPredicateVar);
		
		    aTerm = tf.createJunctorTermAndSimplify(Op.AND, setPredicate, expressionPredicate);
		    
		    resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , aTerm);

		    
		}

		log("INTO :" + resultTerm.toString());

		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(setTranslation.getAxioms());
		result.addAxioms(expressionTranslation.getAxioms());
		result.setPredicateVariable(newPredicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    
		
	    }

	} else if(featureName.equals("asBag")){
	    // Translate asBag on sets ...
 
	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term setTerm = setTranslation.getTerm();
 	  
	    NonCollectionSort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort =  proverFacade.getSortForOCLBag(elemSort);
	    Sort intSort =  proverFacade.getSortForOCLInteger();
	    String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);

	     if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		 
		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable paramsForNewFunction = setTerm.freeVars();
	
		Term[] paramTerms;
		Sort[] paramSorts;		
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt];
		paramSorts = new Sort[freeVarsCnt];
		
		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=0;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    i++;
		}
		    

		 Function setFeatureFunction =
		     proverFacade.getCollectionFunctionForFeature(functionName,
								    paramSorts,
								    retSort);

		 resultTerm = tf.createFunctionTerm(setFeatureFunction,paramTerms);	
		 log("INTO :" + resultTerm.toString());
		 
		 // Create an additional axiom for the interpretation of the new function symbol ...
		 Term newAxiom1, newAxiom2;
		 
		 // use a predicative translation for the set on which the feature is applied ...
		 setTranslation = 
		     getPredicativeTranslationOfTranslation(setTranslation,collType);
		 
		 Term setPredicate = setTranslation.getTerm();
		 LogicVariable predicateVar = setTranslation.getPredicateVariable();
		 Term predicateVarTerm = tf.createVariableTerm(predicateVar);
		 
		 
		 Function countFunction = proverFacade.getCollectionFunctionForFeature(COUNT,
											 new Sort[]{retSort,elemSort},
											 intSort);
		 
		 
		 Function bagContainsFunction =  proverFacade.getCollectionFunctionForFeature(CONTAINS,
												new Sort[]{retSort,elemSort},
												Sort.FORMULA);
		 
		 Function leqFunction =  proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN,
									      new Sort[]{intSort,intSort},
									      Sort.FORMULA);
		 
		 Term oneTerm = this.computeTranslationForIntegerLiteral("1");
		 Term countTerm = tf.createFunctionTerm(countFunction,resultTerm,predicateVarTerm);
		 Term leqTerm =  tf.createFunctionTerm(leqFunction,countTerm,oneTerm);
		 Term eqTerm = tf.createFunctionTerm(countFunction,countTerm,oneTerm);	
		 Term aTerm = tf.createJunctorTerm(Op.EQV,setPredicate, eqTerm);

		 aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);
		 newAxiom1 = constraintTranslator.createAllClosureOfTerm(aTerm);
		 
		 log("Adding new axiom for " + functionName + " : " + newAxiom1);

		 aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , leqTerm);
		 newAxiom2 = constraintTranslator.createAllClosureOfTerm(aTerm);
		 
		 log("Adding new axiom for " + functionName + " : " + newAxiom2);

		 result = new KeyTranslationResult();
		 result.setTerm(resultTerm);
		 result.addAxioms(setTranslation.getAxioms());
		 result.addAxiom(newAxiom1);
		 result.addAxiom(newAxiom2);
	     
	     } else {
		 // ==================== TRANSLATION AS A PREDICATE ==============================
		 
		 setTranslation = 
		     getPredicativeTranslationOfTranslation(setTranslation,collType);
			
		 Term setPredicate = setTranslation.getTerm();
		 LogicVariable predicateVar = setTranslation.getPredicateVariable();

		 log("INTO :" + setPredicate.toString());
		
		 result = new KeyTranslationResult();
		 result.setTerm(setPredicate);
		 result.addAxioms(setTranslation.getAxioms());
		 result.setPredicateVariable(predicateVar);
		 result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	     }
	    
	    
	}  else if(featureName.equals("asSequence")){
	    // Translate asSequence on sets ...
 
	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term setTerm = setTranslation.getTerm();
 	  
	    
	    NonCollectionSort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort =  proverFacade.getSortForOCLSequence(elemSort);
	    Sort intSort =  proverFacade.getSortForOCLInteger();
	    String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);
	    
	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
	    
		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable paramsForNewFunction = setTerm.freeVars();
	
		Term[] paramTerms;
		Sort[] paramSorts;		
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt];
		paramSorts = new Sort[freeVarsCnt];
		
		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=0;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    i++;
		}
		    
		Function setFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(functionName,
								   paramSorts,
								   retSort);
		
		resultTerm = tf.createFunctionTerm(setFeatureFunction,paramTerms);	
		log("INTO :" + resultTerm.toString());
		
		// Create an additional axiom for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2;
		
		// use a predicative translation for the set on which the feature is applied ...
		setTranslation = 
		    getPredicativeTranslationOfTranslation(setTranslation,collType);
		
		Term setPredicate = setTranslation.getTerm();
		LogicVariable predicateVar = setTranslation.getPredicateVariable();
		Term predicateVarTerm = tf.createVariableTerm(predicateVar);
		
		Function countFunction = proverFacade.getCollectionFunctionForFeature(COUNT,
											new Sort[]{retSort,elemSort},
											intSort);
		
		
		Function seqContainsFunction =  proverFacade.getCollectionFunctionForFeature(CONTAINS,
											       new Sort[]{retSort,elemSort},
											       Sort.FORMULA);
		
		Function leqFunction =  proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN ,
									     new Sort[]{intSort,intSort},
									     Sort.FORMULA);
		
		Term oneTerm = this.computeTranslationForIntegerLiteral("1"); 
		Term countTerm = tf.createFunctionTerm(countFunction,resultTerm,predicateVarTerm); 
		Term leqTerm =  tf.createFunctionTerm(leqFunction,countTerm,oneTerm);
		Term containsTerm = tf.createFunctionTerm(seqContainsFunction, resultTerm, predicateVarTerm);
		Term aTerm = tf.createJunctorTerm(Op.EQV,setPredicate,containsTerm);
	
		
		aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);
		newAxiom1 = constraintTranslator.createAllClosureOfTerm(aTerm);
		
		log("Adding new axiom for " + functionName + " : " + newAxiom1);

		aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , leqTerm);
		newAxiom2 = constraintTranslator.createAllClosureOfTerm(aTerm);
		
		log("Adding new axiom for " + functionName + " : " + newAxiom2);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(setTranslation.getAxioms());
		result.addAxiom(newAxiom1);
		result.addAxiom(newAxiom2);
		
	    }  else {
		// ==================== TRANSLATION AS A PREDICATE ==============================
		
		setTranslation = 
		    getPredicativeTranslationOfTranslation(setTranslation,collType);
		
		Term setPredicate = setTranslation.getTerm();
		LogicVariable predicateVar = setTranslation.getPredicateVariable();

		log("INTO :" + setPredicate.toString());
		
		result = new KeyTranslationResult();
		result.setTerm(setPredicate);
		result.addAxioms(setTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    }
	    

	    
	} else if(featureName.equals("isUnique")){
	    // Translate isUnique on sets ...

	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult expressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term setTerm = setTranslation.getTerm();
 	    Term expressionTerm = expressionTranslation.getTerm();

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort exprSort = expressionTerm.sort();
	    
	    // use a predicative translation for the set on which 'isUnique' is applied ...
 	    setTranslation = 
		getPredicativeTranslationOfTranslation(setTranslation,collType);
	    
	    Term setPredicate = setTranslation.getTerm();
	    LogicVariable predicateVar = setTranslation.getPredicateVariable();
	    
	    // Create necessary variables and variable terms ...
	    LogicVariable elementVar1 = newVariableDeliverer.getNewVariableForSort(elemSort);
	    Term elementVarTerm1 = tf.createVariableTerm(elementVar1);
	    LogicVariable elementVar2 = newVariableDeliverer.getNewVariableForSort(elemSort);
	    Term elementVarTerm2 = tf.createVariableTerm(elementVar2);

	    // We have exactely one iterator variable here !!!!! ...
	    String iteratorVarName = (String)iteratorNames.elementAt(0);
	    String iteratorTypeName = iteratorType.toString().trim();

	    // Get the translation of the iteratorVariable ...
	    KeyTranslationResult iteratorTranslation = 
		(KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								 iteratorTypeName);
	    
	    Term iteratorTerm = iteratorTranslation.getTerm();
	    LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
	    
	    Term containsTerm1 = renameVariableInTerm(setPredicate, predicateVar, elementVar1);
	    Term containsTerm2 = renameVariableInTerm(setPredicate, predicateVar, elementVar2);
	    Term exprTerm1 = renameVariableInTerm(expressionTerm, iteratorVar, elementVar1);
	    Term exprTerm2 = renameVariableInTerm(expressionTerm, iteratorVar, elementVar2);
	    Term eqTerm1 = tf.createEqualityTerm(exprTerm1,exprTerm2);
	    Term andTerm = tf.createJunctorTermAndSimplify(Op.AND, containsTerm1, containsTerm2);
	    andTerm = tf.createJunctorTermAndSimplify(Op.AND, andTerm, eqTerm1);
	    Term eqTerm2 =  tf.createEqualityTerm(elementVarTerm1,elementVarTerm2);
	    Term impTerm =  tf.createJunctorTermAndSimplify(Op.IMP, andTerm, eqTerm2);
	    
	    // resultTerm = tf.createQuantifierTerm(Op.ALL,new LogicVariable[]{elementVar1,elementVar2},impTerm);
	    // NEW:
	    resultTerm = getQuantifiedTermWith(Op.ALL, elementVar2 , impTerm);
	    resultTerm = getQuantifiedTermWith(Op.ALL, elementVar1 , resultTerm);

	    log("INTO :" + resultTerm.toString());

	
	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(setTranslation.getAxioms());
	    result.addAxioms(expressionTranslation.getAxioms());
	  
	    
	}  else if(featureName.equals("sortedBy")){
	    // Translate sortedBy on sets ...
 
	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	     KeyTranslationResult expressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term setTerm = setTranslation.getTerm();
	    Term expressionTerm = expressionTranslation.getTerm();

	    NonCollectionSort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort =  proverFacade.getSortForOCLSequence(elemSort);
	    Sort intSort =  proverFacade.getSortForOCLInteger();

	    Sort exprSort = expressionTerm.sort();
	    if ( (exprSort != proverFacade.getSortForOCLInteger()) && 
		 (exprSort != proverFacade.getSortForOCLReal()) ){
		// We don't allow the sortedBy expression for other element types than OclInteger or OclReal ...
		error("The expression used to describe the order for the 'sortedBy' feature is only allowed to have Integer or Real as its type !!!!");
	    }

	     
	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
	    
		String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);
	    
		// Collect all the necessary parameters for the new function symbol:
		// These are all free variables in the translation of the collection and the expression ...

		// except of the iteratorvariable ...
		
		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		// This is the variable we have to care about within the expression (which specifies the order !) ...
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
	    
	    
		
		SetOfQuantifiableVariable freeVarsSet = setTerm.freeVars();
		SetOfQuantifiableVariable freeVarsExpr = expressionTerm.freeVars();
	 
		SetOfQuantifiableVariable paramsForNewFunction = freeVarsSet.union(freeVarsExpr.remove(iteratorVar));
		
		log("FreeVars for new sorted collection '" + functionName + "' == " + paramsForNewFunction );
		
		Term[] paramVars;
		Sort[] paramSorts;
		
		int freeVarsCnt = paramsForNewFunction.size();
		if ( freeVarsCnt > 0){
		    paramVars = new Term[freeVarsCnt];
		    paramSorts = new Sort[freeVarsCnt];
		    
		    IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		    LogicVariable currentVar;
		    int i=0;
		    
		    while (varIt.hasNext()){
			currentVar = (LogicVariable) varIt.next();
			paramVars[i] = tf.createVariableTerm(currentVar);
			paramSorts[i] = currentVar.sort();
			i++;
		    }
		    
		} else {
		    paramVars = new Term[0];
		    paramSorts = new Sort[0];
		}
		
		
		// Now we can create a new function symbol for the sorted collection ...
		
		
		Function setFeatureFunction =
		    proverFacade.getFunctionForFeature(functionName,
							 paramSorts,
							 retSort);
		
		
		// We take this new function term as the result of the translation ...
		resultTerm = 
		    tf.createFunctionTerm(setFeatureFunction, paramVars);
		
		
		// Create additional axioms for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2;
		
		
		// 1st axiom ... The resulting sequence has the same elements as the original collection !
		
		// use a predicative translation for the set on which the feature is applied ...
		setTranslation = 
		    getPredicativeTranslationOfTranslation(setTranslation,collType);
		
		Term setPredicate = setTranslation.getTerm();
		LogicVariable predicateVar = setTranslation.getPredicateVariable();
		Term predicateVarTerm = tf.createVariableTerm(predicateVar);
		
		Function countFunction = proverFacade.getCollectionFunctionForFeature(COUNT,
											new Sort[]{retSort,elemSort},
											intSort);
		
		
		Function seqContainsFunction =  proverFacade.getCollectionFunctionForFeature(CONTAINS,
											       new Sort[]{retSort,elemSort},
											       Sort.FORMULA);
		
		Function leqFunction =  proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN ,
									     new Sort[]{intSort,intSort},
									     Sort.FORMULA);
		
		Term oneTerm = this.computeTranslationForIntegerLiteral("1");
		Term countTerm = tf.createFunctionTerm(countFunction,resultTerm,predicateVarTerm);
		Term leqTerm =  tf.createFunctionTerm(leqFunction,countTerm,oneTerm);
		Term containsTerm = tf.createFunctionTerm(seqContainsFunction,resultTerm,predicateVarTerm);	
		Term aTerm = tf.createJunctorTerm(Op.EQV,setPredicate,containsTerm);
		aTerm = tf.createJunctorTermAndSimplify(Op.AND,aTerm,leqTerm);
		//  aTerm = tf.createQuantifierTerm(Op.ALL,predicateVar,aTerm);
		// NEW:
		aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);
		
		newAxiom1 = constraintTranslator.createAllClosureOfTerm(aTerm);
		
		
		// 2nd axiom ... The resulting sequence is ordered according to the given expression ...
	
		LogicVariable intVar1 = new LogicVariable(new Name("i"), intSort );
		LogicVariable intVar2 = new LogicVariable(new Name("j"), intSort );
		
		Term intVarTerm1 = tf.createVariableTerm(intVar1);
		Term intVarTerm2 = tf.createVariableTerm(intVar2);
		
		Function intLessOrEqualThanFunction = proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN ,
											   new Sort[]{intSort, intSort},
											   Sort.FORMULA);

		Function exprLessOrEqualThanFunction = proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN ,
											    new Sort[]{exprSort, exprSort},
											    Sort.FORMULA);
		
		Function collSizeFunction = 
		    proverFacade.getFunctionForFeature(SIZE,
							 new Sort[]{retSort},
							 intSort);
		
		Function collAtFunction = 
		    proverFacade.getFunctionForFeature(SEQUENCE_AT,
							 new Sort[]{retSort, intSort},
							 elemSort);
		
		Term sizeTerm = tf.createFunctionTerm(collSizeFunction, resultTerm);
		Term lessTerm1 = tf.createFunctionTerm(intLessOrEqualThanFunction, oneTerm, intVarTerm1);
		Term lessTerm2 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm1, sizeTerm);
		Term lessTerm3 = tf.createFunctionTerm(intLessOrEqualThanFunction, oneTerm, intVarTerm2);
		Term lessTerm4 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm2, sizeTerm);
		Term lessTerm5 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm1, intVarTerm2);
		
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, lessTerm1 , lessTerm2 );
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm3 );
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm4 );
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm5 );
		
		Term atTerm1 = tf.createFunctionTerm(collAtFunction, resultTerm, intVarTerm1);
		Term atTerm2 = tf.createFunctionTerm(collAtFunction, resultTerm, intVarTerm2);
	    
		// Now we want to build terms like : expression(at(i))
		// Therefore we have to substitute occurences of  the iteratorvariable with the 'at(i)' term ...
		
		ClashFreeSubst subst1 = new ClashFreeSubst(iteratorVar, atTerm1);
		Term exprAtTerm1 = subst1.apply(expressionTerm);
		ClashFreeSubst subst2 = new ClashFreeSubst(iteratorVar, atTerm2);
		Term exprAtTerm2 = subst2.apply(expressionTerm);
		
		Term lessTerm6 = tf.createFunctionTerm(exprLessOrEqualThanFunction,exprAtTerm1,exprAtTerm2);
		
		Term impTerm = tf.createJunctorTermAndSimplify(Op.IMP, aTerm , lessTerm6 );
		
		Term allTerm = tf.createQuantifierTerm(Op.ALL, new LogicVariable[]{intVar1, intVar2}, impTerm);
		
		newAxiom2 = constraintTranslator.createAllClosureOfTerm(allTerm);
		
		
		log("Adding new axiom (1) for " + functionName + " : " + newAxiom1);
		log("Adding new axiom (2) for " + functionName + " : " + newAxiom2);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(setTranslation.getAxioms());
		result.addAxioms(expressionTranslation.getAxioms());
		result.addAxiom(newAxiom1);
		result.addAxiom(newAxiom2);
	    }  else {
		// ==================== TRANSLATION AS A PREDICATE ==============================
		
		setTranslation = 
		     getPredicativeTranslationOfTranslation(setTranslation,collType);
		
		Term setPredicate = setTranslation.getTerm();
		LogicVariable predicateVar = setTranslation.getPredicateVariable();

		log("INTO :" + setPredicate.toString());

		result = new KeyTranslationResult();
		result.setTerm(setPredicate);
		result.addAxioms(setTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    }
	    
	    
	} else if(featureName.equals("iterate")){
	    // Translate iterate on sets ...

	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);

	    // The first parameter node belongs to the OCL expression that is used for initializing
	    // the accumulator variable. The second parameter is the step-expression
	    KeyTranslationResult initExpressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	     KeyTranslationResult stepExpressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(1));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term setTerm = setTranslation.getTerm();

	    Type initExpressionType = tree.getNodeType((Node) parameterNodes.elementAt(0));
 	    Term initExpressionTerm = initExpressionTranslation.getTerm();
	    Type stepExpressionType = tree.getNodeType((Node) parameterNodes.elementAt(1));
 	    Term stepExpressionTerm = null;

	    // Convert the representation of the expression to be computed in
	    // some suitable form ...
	    if (stepExpressionType == Basic.BOOLEAN) {
		// We have to ensure to use boolean terms (instead of formulas) ...
		stepExpressionTerm = convertFormulaToTerm(stepExpressionTranslation.getTerm(), 
							  stepExpressionTranslation);
	    } else {
		stepExpressionTerm = stepExpressionTranslation.getTerm();
	    }
	       
	    
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);

	    Sort exprSort = stepExpressionTerm.sort();
	    Sort exprElementSort;
	    
	    Sort initSort = initExpressionTerm.sort();

	    Sort retSort = exprSort;
	    
	    // Currently we only support the basic translation method for iterate ...

		// ==================== TRANSLATION AS TERM ==============================
		
	    
		// Get the iterator variable (and the corresponding term) ...

		// We have use the first iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		
		// And now the accumulator ...

		// We have use the second iterator variable here !!!!! ...
		String accumulatorVarName = (String)iteratorNames.elementAt(1);

		Type accumulatorType = initExpressionType; 
		String accumulatorTypeName = accumulatorType.toString().trim();
		
		// Get the translation of the accumulator variable ...
		KeyTranslationResult accumulatorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(accumulatorVarName,
								     accumulatorTypeName);
		
		Term accumulatorTerm = accumulatorTranslation.getTerm();
		LogicVariable accumulatorVar = (LogicVariable) accumulatorTerm.op();
		
		
		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable freeVarsStepExpr =  stepExpressionTerm.freeVars();
		SetOfQuantifiableVariable freeVarsInitExpr =  initExpressionTerm.freeVars();

		SetOfQuantifiableVariable paramsSetSmall = 
		    freeVarsStepExpr.remove(iteratorVar).remove(accumulatorVar);
		SetOfQuantifiableVariable paramsSetBig = 
		    freeVarsStepExpr.union(freeVarsInitExpr).remove(iteratorVar).remove(accumulatorVar);
	
		Term[] paramTermsStep;
		Sort[] paramSortsStep;
		Term[] variableParamTermsSmall;
		Sort[] variableParamSortsSmall;
		
		int freeVarsCntSmall = paramsSetSmall.size();
		
		paramTermsStep = new Term[freeVarsCntSmall + 2];
		paramSortsStep = new Sort[freeVarsCntSmall + 2];
		
		variableParamTermsSmall = new Term[freeVarsCntSmall];
		variableParamSortsSmall = new Sort[freeVarsCntSmall];

		// The first parameter is the current element in the iteration ...
		paramTermsStep[0] = iteratorTerm;
		paramSortsStep[0] = elemSort;

		// The second parameter is the current accumulator value in the iteration ...
		paramTermsStep[1] = accumulatorTerm;
		paramSortsStep[1] = accumulatorTerm.sort();

		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsSetSmall.iterator();
		LogicVariable currentVar;
		int i=2;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTermsStep[i] = tf.createVariableTerm(currentVar);
		    paramSortsStep[i] = currentVar.sort();
		    variableParamTermsSmall[i-2] = paramTermsStep[i];
		    variableParamSortsSmall[i-2] = paramSortsStep[i];
		    i++;
		}
		 
		Term[] paramTermsIterate;
		Sort[] paramSortsIterate;
		Term[] variableParamTermsBig;
		Sort[] variableParamSortsBig;
		
		int freeVarsCntBig = paramsSetBig.size();
		
		paramTermsIterate = new Term[freeVarsCntBig + 1];
		paramSortsIterate = new Sort[freeVarsCntBig + 1];
		
		variableParamTermsBig = new Term[freeVarsCntBig];
		variableParamSortsBig = new Sort[freeVarsCntBig];

		// The first parameter is the collection to be iterated over ...
		paramTermsIterate[0] = setTerm;
		paramSortsIterate[0] = collSort;

		 // The rest are the necessary parameters from the other expressions ...
		varIt = paramsSetBig.iterator();
		i=1;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTermsIterate[i] = tf.createVariableTerm(currentVar);
		    paramSortsIterate[i] = currentVar.sort();
		    variableParamTermsBig[i-1] = paramTermsIterate[i];
		    variableParamSortsBig[i-1] = paramSortsIterate[i];
		    i++;
		}

		String stepFunctionName = newFunctionNameDeliverer.getNewFunctionNameForName("it-step");
		Function stepFunction =  proverFacade.getCollectionFunctionForFeature(stepFunctionName,
											paramSortsStep,
											retSort);



		String iterateFunctionName = newFunctionNameDeliverer.getNewFunctionNameForName("iterate");
		Function iterateFunction =  proverFacade.getCollectionFunctionForFeature(iterateFunctionName,
											   paramSortsIterate,
											   retSort);

	
		resultTerm = tf.createFunctionTerm(iterateFunction,paramTermsIterate);
		
		log("INTO :" + resultTerm.toString());
		
		
		// Create additional axioms for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2, newAxiom3;
		Term[] currentArgs;
		Sort[] currentSorts;

		// Now get the corresponding constant for the empty collection ...
		Function emptySetFunction = proverFacade.getCollectionFunctionForFeature(SET_EMPTYSET,
											   new Sort[0],
											   collSort);
		
		Function setInsertFunction = proverFacade.getCollectionFunctionForFeature(SET_INSERT,
											    new Sort[]{collSort,elemSort},
											    collSort);

		Function setRemoveFunction = proverFacade.getCollectionFunctionForFeature(SET_REMOVE,
											    new Sort[]{collSort,elemSort},
											    collSort);
		
	
		Term emptySetTerm = tf.createFunctionTerm(emptySetFunction);
		
		
		
		// Create necessary variables and variable terms ...		
		LogicVariable collVar = newVariableDeliverer.getNewVariableForSort(collSort);
		Term collVarTerm = tf.createVariableTerm(collVar);
		
		Term elementVarTerm = iteratorTerm;

		// Now we can build the additional axioms ...
		
		// No.1 ...
		Term stpTerm = tf.createFunctionTerm(stepFunction,paramTermsStep);
		Term eqTerm0 = tf.createEqualityTerm(stpTerm, stepExpressionTerm);

		newAxiom1 = constraintTranslator.createAllClosureOfTerm(eqTerm0);
		
		// No. 2 ...
		currentArgs = prependTerm(emptySetTerm,variableParamTermsBig);
		Term itTerm = tf.createFunctionTerm(iterateFunction,currentArgs);
		Term eqTerm1 = tf.createEqualityTerm(itTerm, initExpressionTerm);

		newAxiom2 = constraintTranslator.createAllClosureOfTerm(eqTerm1);
		
 
		// No.3 ...
		Term insertTerm1 =  tf.createFunctionTerm(setInsertFunction, collVarTerm, elementVarTerm);
		Term removeTerm = tf.createFunctionTerm(setRemoveFunction, collVarTerm, elementVarTerm);
		currentArgs = prependTerm(insertTerm1, variableParamTermsBig);
		Term itTerm1 =  tf.createFunctionTerm(iterateFunction, currentArgs);

		currentArgs = prependTerm(removeTerm, variableParamTermsBig);
		Term itTerm2 =  tf.createFunctionTerm(iterateFunction, currentArgs);

		currentArgs = prependTerm(itTerm2, variableParamTermsSmall);
		currentArgs = prependTerm(elementVarTerm, currentArgs);

		Term stpTerm2 = tf.createFunctionTerm(stepFunction, currentArgs);

		Term eqTerm2 = tf.createEqualityTerm(itTerm1,stpTerm2);
		
		newAxiom3 = constraintTranslator.createAllClosureOfTerm(eqTerm2);
		
		
		log("Adding new axiom (1)  for " + iterateFunctionName  + " : " + newAxiom1);
		log("Adding new axiom (2)  for " + iterateFunctionName  + " : " + newAxiom2);
		log("Adding new axiom (3)  for " + iterateFunctionName  + " : " + newAxiom3);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(setTranslation.getAxioms());
		result.addAxioms(stepExpressionTranslation.getAxioms());
		result.addAxioms(initExpressionTranslation.getAxioms());
		result.addAxiom(newAxiom1);
		result.addAxiom(newAxiom2);
		result.addAxiom(newAxiom3);
		
	   

	} else if(featureName.equals("asBag")){
	    // Translate asBag on sets ...
 
	    KeyTranslationResult setTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term setTerm = setTranslation.getTerm();
 	  
	    NonCollectionSort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort =  proverFacade.getSortForOCLBag(elemSort);
	    Sort intSort =  proverFacade.getSortForOCLInteger();
	    String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);

	     if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		 
		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable paramsForNewFunction = setTerm.freeVars();
	
		Term[] paramTerms;
		Sort[] paramSorts;		
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt];
		paramSorts = new Sort[freeVarsCnt];
		
		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=0;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    i++;
		}
		    

		 Function setFeatureFunction =
		     proverFacade.getCollectionFunctionForFeature(functionName,
								    paramSorts,
								    retSort);

		 resultTerm = tf.createFunctionTerm(setFeatureFunction,paramTerms);	
		 log("INTO :" + resultTerm.toString());
		 
		 // Create an additional axiom for the interpretation of the new function symbol ...
		 Term newAxiom1, newAxiom2;
		 
		 // use a predicative translation for the set on which the feature is applied ...
		 setTranslation = 
		     getPredicativeTranslationOfTranslation(setTranslation,collType);
		 
		 Term setPredicate = setTranslation.getTerm();
		 LogicVariable predicateVar = setTranslation.getPredicateVariable();
		 Term predicateVarTerm = tf.createVariableTerm(predicateVar);
		 
		 
		 Function countFunction = proverFacade.getCollectionFunctionForFeature(COUNT,
											 new Sort[]{retSort,elemSort},
											 intSort);
		 
		 
		 Function bagContainsFunction =  proverFacade.getCollectionFunctionForFeature(CONTAINS,
												new Sort[]{retSort,elemSort},
												Sort.FORMULA);
		 
		 Function leqFunction =  proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN,
									      new Sort[]{intSort,intSort},
									      Sort.FORMULA);
		 
		 Term oneTerm = this.computeTranslationForIntegerLiteral("1");
		 Term countTerm = tf.createFunctionTerm(countFunction,resultTerm,predicateVarTerm);
		 Term leqTerm =  tf.createFunctionTerm(leqFunction,countTerm,oneTerm);
		 Term eqTerm = tf.createFunctionTerm(countFunction,countTerm,oneTerm);	
		 Term aTerm = tf.createJunctorTerm(Op.EQV,setPredicate, eqTerm);

		 aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);
		 newAxiom1 = constraintTranslator.createAllClosureOfTerm(aTerm);
		 
		 log("Adding new axiom for " + functionName + " : " + newAxiom1);

		 aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , leqTerm);
		 newAxiom2 = constraintTranslator.createAllClosureOfTerm(aTerm);
		 
		 log("Adding new axiom for " + functionName + " : " + newAxiom2);

		 result = new KeyTranslationResult();
		 result.setTerm(resultTerm);
		 result.addAxioms(setTranslation.getAxioms());
		 result.addAxiom(newAxiom1);
		 result.addAxiom(newAxiom2);
	     
	     } else {
		 // ==================== TRANSLATION AS A PREDICATE ==============================
		 
		 setTranslation = 
		     getPredicativeTranslationOfTranslation(setTranslation,collType);
			
		 Term setPredicate = setTranslation.getTerm();
		 LogicVariable predicateVar = setTranslation.getPredicateVariable();

		 log("INTO :" + setPredicate.toString());
		
		 result = new KeyTranslationResult();
		 result.setTerm(setPredicate);
		 result.addAxioms(setTranslation.getAxioms());
		 result.setPredicateVariable(predicateVar);
		 result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	     }
	    
	    
	} 



	else {
	    error("Translation of SET feature " + featureName + " NOT supported yet !!!");
	}
	

	// Finally set the translation result for the current node ...
	setTranslationForNode(n, result);

	
    }



   // =========================================================================================================
   // ================================= The feature calls on OCL type Bag[T] ==================================
   // =========================================================================================================

	protected void translateFeatureCallonBag(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType , boolean  featureCallWithAtPre){
	    log("Feature call on Bag : " + featureName);
	    log("Applied on node :" + appliedTo.toString());
	    
	    
	    KeyTranslationResult result = null;
	    Term resultTerm = null;
	    
	    // Get the kind of translation to be performed for the given node ...
	    TranslationType translationType = getTranslationTypeForNode(n);
	    log("Current feature call should be translated in the following way : " + translationType.toString());
	
	   
	    Sort iteratorSort;

	   
	  if(featureName.equals("including") || featureName.equals("excluding") ){
	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult elementTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term bagTerm = bagTranslation.getTerm();
 	    Term elementTerm = elementTranslation.getTerm();

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = collSort;

	    if(elemSort == proverFacade.getSortForOCLBoolean()){
		// Ensure the representation of the expression to be a boolean term (not a formula) ...
		elementTerm = convertFormulaToTerm(elementTerm, elementTranslation);
	    }

	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================

		String funcName = (featureName.equals("including")) ? BAG_INSERT :  BAG_REMOVE; 
		
		Function bagFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(funcName,
								   new Sort[]{collSort,elemSort},
								   retSort);
	    
		resultTerm = tf.createFunctionTerm(bagFeatureFunction,bagTerm,elementTerm);
		
		
		
		log("INTO :" + resultTerm.toString());
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(bagTranslation.getAxioms());
		result.addAxioms(elementTranslation.getAxioms());
	    
	    } else {
		// ==================== PREDICATIVE TRANSLATION ==========================

		// use a predicative translation for the bag on which 'including' is applied ...
		bagTranslation = 
		    getPredicativeTranslationOfTranslation(bagTranslation,collType);
		
		Term bagPredicate = bagTranslation.getTerm();
		
		LogicVariable predicateVar = bagTranslation.getPredicateVariable();
		Term varTerm = tf.createVariableTerm(predicateVar);
		Term eqTerm = null;
		
		if (featureName.equals("including")){
		    // Create new predicate for including ...
		    eqTerm = tf.createEqualityTerm(varTerm, elementTerm);
		    resultTerm = tf.createJunctorTermAndSimplify(Op.OR, bagPredicate, eqTerm);
		} else {
		    // Create new predicate for excluding ...
		    eqTerm = tf.createEqualityTerm(varTerm, elementTerm);
		    eqTerm = tf.createJunctorTerm(Op.NOT,eqTerm);
		    resultTerm = tf.createJunctorTermAndSimplify(Op.AND, bagPredicate, eqTerm);
		}
		
		log("INTO :" + resultTerm.toString());
				
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(bagTranslation.getAxioms());
		result.addAxioms(elementTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);

	    }

	} else if(featureName.equals("forAll") ||featureName.equals("exists") ){
	    // Translate forAll and exists on Bags ...

	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult formulaTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    Collection collType = (Collection) tree.getNodeType(appliedTo);

	    // use a predicative translation for the bag on which 'forAll' is applied ...
 	    bagTranslation = 
		getPredicativeTranslationOfTranslation(bagTranslation,collType);
	    
	    Term bagPredicate = bagTranslation.getTerm();
	    Term formulaTerm = convertTermToFormula(formulaTranslation.getTerm());
	    LogicVariable predicateVar = bagTranslation.getPredicateVariable();
	    
	    log(featureName + "formulaTerm : " + formulaTerm.toString());

	    

	    // We have exactely one iterator variable here !!!!! ...
	    String iteratorVarName = (String)iteratorNames.elementAt(0);
	    String iteratorTypeName = iteratorType.toString().trim();

	    // Get the translation of the iteratorVariable ...
	    KeyTranslationResult iteratorTranslation = 
		(KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								 iteratorTypeName);
	    
	    Term iteratorTerm = iteratorTranslation.getTerm();
	    LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
	    

	    formulaTerm = renameVariableInTerm(formulaTerm, iteratorVar, predicateVar);

	    log(featureName + "formulaTerm after renaming: " + formulaTerm.toString());
	    Term aTerm = null;

	    if (featureName.equals("forAll")){
		// Create formula for forAll ...
		aTerm = tf.createJunctorTermAndSimplify(Op.IMP,bagPredicate,formulaTerm);
		// resultTerm = tf.createQuantifierTerm(Op.ALL,predicateVar,aTerm);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);

	    } else {
		// Create formula for exists ...
		aTerm = tf.createJunctorTermAndSimplify(Op.AND,bagPredicate,formulaTerm);
		// resultTerm = tf.createQuantifierTerm(Op.EX,predicateVar,aTerm);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , aTerm);
	    }

	    
	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(bagTranslation.getAxioms());
	    result.addAxioms(formulaTranslation.getAxioms());

	} else if(featureName.equals("isEmpty") ||featureName.equals("notEmpty") ){
	    // Translate isEmpty and notEmpty on Bags ...

	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);

	    // use a predicative translation for the bag on which 'forAll' is applied ...
 	    bagTranslation = 
		getPredicativeTranslationOfTranslation(bagTranslation,collType);
	    
	    Term bagPredicate = bagTranslation.getTerm();
	    LogicVariable predicateVar = bagTranslation.getPredicateVariable();
	    
	    Term aTerm;


	    if (featureName.equals("isEmpty")){
		// Create formula for isEmpty ...
		aTerm = tf.createJunctorTerm(Op.NOT,bagPredicate);
		// resultTerm = tf.createQuantifierTerm(Op.ALL,predicateVar,aTerm);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);
	    } else {
		// Create formula for notEmpty ...
		resultTerm = tf.createQuantifierTerm(Op.EX,predicateVar,bagPredicate);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , bagPredicate);
	    }

	    
	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(bagTranslation.getAxioms());
	    
    
	} else if(featureName.equals("includes")|| (featureName.equals("excludes") )){
	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult elementTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term bagTerm = bagTranslation.getTerm();
 	    Term elementTerm = elementTranslation.getTerm();

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = proverFacade.getSortForOCLBoolean();
	   
	    Function bagFeatureFunction =
 		proverFacade.getCollectionFunctionForFeature(CONTAINS,
 						     new Sort[]{collSort,elemSort},
 						     Sort.FORMULA);
	    
	    resultTerm = tf.createFunctionTerm(bagFeatureFunction,bagTerm,elementTerm);
		
	    if(featureName.equals("excludes") ){
		// we have to negate the formula ...
		resultTerm = tf.createJunctorTerm(Op.NOT,resultTerm);
	    }
 	 

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(bagTranslation.getAxioms());
	    result.addAxioms(elementTranslation.getAxioms());
	   
	    
	}  else if(featureName.equals("includesAll") || (featureName.equals("excludesAll") )){
	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);

	    Node collArgNode = (Node) parameterNodes.elementAt(0);

	    KeyTranslationResult collTranslation =
		(KeyTranslationResult) getTranslationForNode(collArgNode);
	    
	    Collection collType1 = (Collection) tree.getNodeType(appliedTo);
	    Collection collType2 = (Collection) tree.getNodeType(collArgNode);
	    
	    // use a predicative translation for the bag on which the feature is applied ...
 	    bagTranslation = 
		getPredicativeTranslationOfTranslation(bagTranslation,collType1);
	    collTranslation = 
		getPredicativeTranslationOfTranslation(collTranslation,collType2);

	    Term bagPredicate = bagTranslation.getTerm();
 	    Term collPredicate = collTranslation.getTerm();

	    Sort elemSort1 = getElementSortForCollectionType(collType1);
	    Sort collSort1 = getDLSortForType(collType1);
	    Sort elemSort2 = getElementSortForCollectionType(collType2);
	    Sort collSort2 = getDLSortForType(collType2);

	    Sort retSort = proverFacade.getSortForOCLBoolean();
	   
	    if (elemSort1 != elemSort2) {
		log("includesAll/excludesAll on collections with DIFFERENT element types is currently not supported !!!!");
	    }

	    LogicVariable predicateVar1 = bagTranslation.getPredicateVariable();
	    LogicVariable predicateVar2 = collTranslation.getPredicateVariable();
	    LogicVariable predicateVar = predicateVar1;

	    collPredicate = renameVariableInTerm(collPredicate, predicateVar2, predicateVar);
	    
	    // log(featureName + "collPredicate after renaming: " + collPredicate.toString());
	    Term aTerm = null;
	    
	    if (featureName.equals("excludesAll")){
		// Create formula for excludesAll ...
		aTerm = tf.createJunctorTerm(Op.NOT,bagPredicate);
	    }  else {
		aTerm = bagPredicate;
	    }
	    
	    aTerm = tf.createJunctorTermAndSimplify(Op.IMP,collPredicate,aTerm);
	    // resultTerm = tf.createQuantifierTerm(Op.ALL,predicateVar,aTerm);
	    // NEW:
	    resultTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);
	   

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(bagTranslation.getAxioms());
	    result.addAxioms(collTranslation.getAxioms());

	} else if(featureName.equals("size") ){
	    // ==================== TRANSLATION AS TERM ==============================
	    
	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Term bagTerm = bagTranslation.getTerm();

 	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = proverFacade.getSortForOCLInteger();

	    Function bagFeatureFunction =
		proverFacade.getCollectionFunctionForFeature(SIZE,
						     new Sort[]{collSort},
						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(bagFeatureFunction,bagTerm);
		

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(bagTranslation.getAxioms());
	        
	} else if(featureName.equals("count")){
	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult elementTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term bagTerm = bagTranslation.getTerm();
 	    Term elementTerm = elementTranslation.getTerm();

	    // ==================== TRANSLATION AS TERM ==============================
	    
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = proverFacade.getSortForOCLInteger();
	   
	    if(elemSort == proverFacade.getSortForOCLBoolean()){
		// Ensure the representation of the expression to be a boolean term (not a formula) ...
		elementTerm = convertFormulaToTerm(elementTerm, elementTranslation);
	    }

	    Function bagFeatureFunction =
 		proverFacade.getCollectionFunctionForFeature(COUNT,
 						     new Sort[]{collSort,elemSort},
 						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(bagFeatureFunction,bagTerm,elementTerm);

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(bagTranslation.getAxioms());
	    result.addAxioms(elementTranslation.getAxioms());
	   
	    
	} else if(featureName.equals("sum") ){
	    // ==================== TRANSLATION AS TERM ==============================
	    
	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Term bagTerm = bagTranslation.getTerm();

 	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Sort collSort = getDLSortForType(collType);
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort retSort = elemSort;

	    if ( (elemSort != proverFacade.getSortForOCLInteger()) && 
		 (elemSort != proverFacade.getSortForOCLReal()) ){
		// We don't allow the sum feature for other element types than OclInteger or OclReal ...
		error("The feature sum on collections is only allowed for collections with elements of type Integer or Real !!!!");
	    }

	    Function bagFeatureFunction =
		proverFacade.getCollectionFunctionForFeature(SUM,
						     new Sort[]{collSort},
						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(bagFeatureFunction,bagTerm);
		

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(bagTranslation.getAxioms());
	    
	        
	} else if(featureName.equals("union") || (featureName.equals("intersection") )){
	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);

	    Node collArgNode = (Node) parameterNodes.elementAt(0);

	    KeyTranslationResult collTranslation =
		(KeyTranslationResult) getTranslationForNode(collArgNode);
	    
	    Collection collType1 = (Collection) tree.getNodeType(appliedTo);
	    Collection collType2 = (Collection) tree.getNodeType(collArgNode);
	    
	    
	    Term bagTerm = bagTranslation.getTerm();
 	    Term collTerm = collTranslation.getTerm();

	    
	    Sort elemSort1 = getElementSortForCollectionType(collType1);
	    Sort collSort1 = getDLSortForType(collType1);
	    Sort elemSort2 = getElementSortForCollectionType(collType2);
	    Sort collSort2 = getDLSortForType(collType2);

	    Sort retSort;

	    if (elemSort1 != elemSort2) {
		error( featureName + " on bags/collections with DIFFERENT element types is currently not supported !!!!");
	    }

	    if(!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		
		String funcName;
		if (featureName.equals("union")){
		    retSort = collSort1;
		    funcName = BAG_UNION;
		} else {
		    retSort = collSort2;
		    funcName = BAG_INTERSECTION;
		}
		
		Function bagFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(funcName,
								   new Sort[]{collSort1,collSort2},
								   retSort);
	    
		resultTerm = tf.createFunctionTerm(bagFeatureFunction,bagTerm,collTerm);
		
		log("INTO :" + resultTerm.toString());
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(bagTranslation.getAxioms());
		result.addAxioms(collTranslation.getAxioms());
		
	    } else {

		// ==================== TRANSLATION AS PREDICATE  ==============================
		bagTranslation = 
		    getPredicativeTranslationOfTranslation(bagTranslation,collType1);
		collTranslation = 
		    getPredicativeTranslationOfTranslation(collTranslation,collType2);

		Term bagPredicate = bagTranslation.getTerm();
		Term collPredicate = collTranslation.getTerm();

		retSort = proverFacade.getSortForOCLBoolean();
	   
		LogicVariable predicateVar1 = bagTranslation.getPredicateVariable();
		LogicVariable predicateVar2 = collTranslation.getPredicateVariable();
		LogicVariable predicateVar = predicateVar1;

		collPredicate = renameVariableInTerm(collPredicate, predicateVar2, predicateVar);
	    
		// log(featureName + "collPredicate after renaming: " + collPredicate.toString());
		Term aTerm = null;
	    
		if (featureName.equals("union")){
		    // Create formula for union ...
		    aTerm = tf.createJunctorTermAndSimplify(Op.OR,bagPredicate, collPredicate);
		}  else {
		    // Create formula for intersection ...
		    aTerm = tf.createJunctorTermAndSimplify(Op.AND,bagPredicate, collPredicate);
		}
	    
	   
		resultTerm = aTerm;

		log("INTO :" + resultTerm.toString());

		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(bagTranslation.getAxioms());
		result.addAxioms(collTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    
	    
	    }


	} else if(featureName.equals("select") ||featureName.equals("reject") ){
	    // Translate select and reject on Bags ...

	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult formulaTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term bagTerm = bagTranslation.getTerm();
	    Term formulaTerm = convertTermToFormula(formulaTranslation.getTerm());

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = collSort;
	   
	     if(!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		 // Get the iterator variable (and the corresponding term) ...

		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		

		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable freeVarsExpr = formulaTerm.freeVars();
		SetOfQuantifiableVariable paramsForNewFunction = freeVarsExpr.remove(iteratorVar);
		
	
		Term[] paramTerms;
		Term[] variableParamTerms;
		Sort[] paramSorts;
		Sort[] variableParamSorts;
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt + 1];
		paramSorts = new Sort[freeVarsCnt + 1];
		
		variableParamTerms = new Term[freeVarsCnt];
		variableParamSorts = new Sort[freeVarsCnt];

		// The first parameter is the collection the feature is applied to ...
		paramTerms[0] = bagTerm; 
		paramSorts[0] = collSort;

		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=1;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    variableParamTerms[i-1] = paramTerms[i];
		    variableParamSorts[i-1] = paramSorts[i];
		    i++;
		}
		    

		 String functionName = newFunctionNameDeliverer.getNewFunctionNameForName("select");

		 Function bagFeatureFunction =
		     proverFacade.getCollectionFunctionForFeature(functionName,
								    paramSorts,
								    retSort);
 
		 resultTerm = tf.createFunctionTerm(bagFeatureFunction,paramTerms);
		 
		 log("INTO :" + resultTerm.toString());
		 
		 // Create additional axioms for the interpretation of the new function symbol ...
		 Term newAxiom1, newAxiom2, newAxiom3;
		 Term[] currentArgs;
		 

		 // Now get the corresponding constant for the empty collection ...
		 Function emptyCollFunction = proverFacade.getCollectionFunctionForFeature(BAG_EMPTYBAG,
											     new Sort[0],
											     retSort);
		 
		 Function collInsertFunction = proverFacade.getCollectionFunctionForFeature(BAG_INSERT,
											      new Sort[]{retSort,elemSort},
											      retSort);
		 
		 
		 Term emptyCollTerm = tf.createFunctionTerm(emptyCollFunction);
		 
		 
		 
		 // Create necessary variables and variable terms ...
		 LogicVariable elementVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		 Term elementVarTerm = tf.createVariableTerm(elementVar);
		 
		 LogicVariable collVar = newVariableDeliverer.getNewVariableForSort(collSort);
		 Term collVarTerm = tf.createVariableTerm(collVar);
		 
		 
		 // Set up the formuls term with the 'right' iteratorvariable ...
		 
		 log(featureName + "formulaTerm : " + formulaTerm.toString());
	
		 formulaTerm = renameVariableInTerm(formulaTerm, iteratorVar, elementVar);
		 
		 log(featureName + "formulaTerm after renaming: " + formulaTerm.toString());
		 
		 
		 // Create the boolean expression to be used for the test ...
		 Term expressionTerm, negExpressionTerm; 
		 if (featureName.equals("select")) {
		     expressionTerm = formulaTerm;
		     negExpressionTerm = tf.createJunctorTerm(Op.NOT,formulaTerm);
		 } else {
		     expressionTerm = tf.createJunctorTerm(Op.NOT,formulaTerm);
		     negExpressionTerm = formulaTerm;
		 }
		 
		 // Now we can build the additional axioms ...
		 
		 // No.1 ...
		 currentArgs = prependTerm(emptyCollTerm,variableParamTerms);
		 Term selectTerm = tf.createFunctionTerm(bagFeatureFunction, currentArgs);
		 newAxiom1 = tf.createEqualityTerm(selectTerm, emptyCollTerm);
		 
		 // No.2 ...
		 Term insertTerm1 =  tf.createFunctionTerm(collInsertFunction, collVarTerm, elementVarTerm);
		 currentArgs = prependTerm(insertTerm1,variableParamTerms);
		 Term selectTerm1 =  tf.createFunctionTerm(bagFeatureFunction, currentArgs);
		 currentArgs = prependTerm(collVarTerm,variableParamTerms);
		 Term selectTerm2 =  tf.createFunctionTerm(bagFeatureFunction, currentArgs);
		 Term insertTerm2 =  tf.createFunctionTerm(collInsertFunction, selectTerm2, elementVarTerm);
		 Term eqTerm = tf.createEqualityTerm(selectTerm1,insertTerm2);
		 
		 Term impTerm = tf.createJunctorTermAndSimplify(Op.IMP, expressionTerm, eqTerm);
		 newAxiom2 = constraintTranslator.createAllClosureOfTerm(impTerm);
		 
		 // No.3 ...
		 eqTerm = tf.createEqualityTerm(selectTerm1,selectTerm2);
		 
		 impTerm = tf.createJunctorTermAndSimplify(Op.IMP, negExpressionTerm, eqTerm);
		 newAxiom3 = constraintTranslator.createAllClosureOfTerm(impTerm);
		 
		 
		 log("Adding new axiom (1)  for " + functionName + " : " + newAxiom1);
		 log("Adding new axiom (2)  for " + functionName + " : " + newAxiom2);
		 log("Adding new axiom (3)  for " + functionName + " : " + newAxiom3);
		 
		 result = new KeyTranslationResult();
		 result.setTerm(resultTerm);
		 result.addAxioms(bagTranslation.getAxioms());
		 result.addAxioms(formulaTranslation.getAxioms());
		 result.addAxiom(newAxiom1);
		 result.addAxiom(newAxiom2);
		 result.addAxiom(newAxiom3);
	     } else {
		// ==================== TRANSLATION AS PREDICATE  ==============================
		bagTranslation = 
		    getPredicativeTranslationOfTranslation(bagTranslation,collType);
		
		Term bagPredicate = bagTranslation.getTerm();
		
		LogicVariable predicateVar = bagTranslation.getPredicateVariable();
		
		// Set up the formuls term with the 'right' iteratorvariable for the formula term ...
		
		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		
		formulaTerm = renameVariableInTerm(formulaTerm, iteratorVar, predicateVar);
	    
		Term aTerm = null;
	    
		if (featureName.equals("select")){
		    // Create filter formula for select  ...
		    aTerm = formulaTerm;
		}  else {
		    // Create filter formula for reject  ...
		    aTerm = tf.createJunctorTerm(Op.NOT,formulaTerm);
		}
	    
	   
		resultTerm = tf.createJunctorTermAndSimplify(Op.AND, bagPredicate, aTerm);

		log("INTO :" + resultTerm.toString());

		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(bagTranslation.getAxioms());
		result.addAxioms(formulaTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    	    
	    }
	    
	}  else if(featureName.equals("collect")){
	     // Translate collect on bags ...

	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult expressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term bagTerm = bagTranslation.getTerm();

	    Type expressionType = tree.getNodeType((Node) parameterNodes.elementAt(0));
 	    Term expressionTerm;

	    // Convert the representation of the expression to be computed in
	    // some suitable form ...
	    if (expressionType == Basic.BOOLEAN) {
		// We have to ensure to use boolean terms (instead of formulas) ...
		expressionTerm = convertFormulaToTerm(expressionTranslation.getTerm(), 
						      expressionTranslation);
	    } else {
		expressionTerm = expressionTranslation.getTerm();
	    }

	    boolean flatteningRequired = false;
	    if (expressionType instanceof Collection) {
		log("Working on a collect producing collections => Flattening is required ...");
		flatteningRequired = true;
	    }

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);

	    Sort exprSort = expressionTerm.sort();
	    NonCollectionSort exprElementSort;
	    Sort retSort;
	    
	    if (!flatteningRequired){
		// the given collect expression has a BASIC type and therefore no flattening is required ...
		exprElementSort = (NonCollectionSort)exprSort;
		retSort = proverFacade.getSortForOCLBag(exprElementSort);
	    } else {
		exprElementSort = getElementSortForCollectionType((Collection) expressionType);
		retSort = proverFacade.getSortForOCLBag(exprElementSort);
	    }

	    if(!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		// Get the iterator variable (and the corresponding term) ...

		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		

		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable freeVarsExpr =  expressionTerm.freeVars();
		SetOfQuantifiableVariable paramsForNewFunction = freeVarsExpr.remove(iteratorVar);
		
	
		Term[] paramTerms;
		Term[] variableParamTerms;
		Sort[] paramSorts;
		Sort[] variableParamSorts;
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt + 1];
		paramSorts = new Sort[freeVarsCnt + 1];
		
		variableParamTerms = new Term[freeVarsCnt];
		variableParamSorts = new Sort[freeVarsCnt];

		// The first parameter is the collection the feature is applied to ...
		paramTerms[0] = bagTerm;
		paramSorts[0] = collSort;

		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=1;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    variableParamTerms[i-1] = paramTerms[i];
		    variableParamSorts[i-1] = paramSorts[i];
		    i++;
		}
		
		String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);
		
		Function bagFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(functionName,
								   paramSorts,
								   retSort);
		
		
		resultTerm = tf.createFunctionTerm(bagFeatureFunction,paramTerms);
		
		log("INTO :" + resultTerm.toString());
		
		
		// Create additional axioms for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2;
		Term[] currentArgs;
		
		// Now get the corresponding constant for the empty collection ...
		Function emptyBagFunction1 = proverFacade.getCollectionFunctionForFeature(BAG_EMPTYBAG,
											    new Sort[0],
											    collSort);
		
		Function emptyBagFunction2 = proverFacade.getCollectionFunctionForFeature(BAG_EMPTYBAG,
											    new Sort[0],
											    retSort);
		
		Function bagInsertFunction1 = proverFacade.getCollectionFunctionForFeature(BAG_INSERT,
											     new Sort[]{collSort,elemSort},
											     collSort);
		
		
		
		Term emptyBagTerm1 = tf.createFunctionTerm(emptyBagFunction1);
		Term emptyBagTerm2 = tf.createFunctionTerm(emptyBagFunction2);
		
		
		// Create necessary variables and variable terms ...
		LogicVariable elementVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		Term elementVarTerm = tf.createVariableTerm(elementVar);
		
		LogicVariable collVar = newVariableDeliverer.getNewVariableForSort(collSort);
		Term collVarTerm = tf.createVariableTerm(collVar);
		
		
		// Set up the formulas term with the 'right' iteratorvariable ...
		
		log(featureName + "expressionTerm : " + expressionTerm.toString());
		
				
		expressionTerm = renameVariableInTerm(expressionTerm, iteratorVar, elementVar);
		
		log(featureName + "expressionTerm after renaming: " + expressionTerm.toString());
		
		
		// Now we can build the additional axioms ...
		
		// No.1 ...
		currentArgs = prependTerm(emptyBagTerm1, variableParamTerms);

		Term collectTerm = tf.createFunctionTerm(bagFeatureFunction, currentArgs);
		newAxiom1 = tf.createEqualityTerm(collectTerm, emptyBagTerm2);
		
		// No.2 ...
		Term insertTerm1 =  tf.createFunctionTerm(bagInsertFunction1, collVarTerm, elementVarTerm);
		currentArgs = prependTerm(insertTerm1, variableParamTerms);
		Term collectTerm1 =  tf.createFunctionTerm(bagFeatureFunction, currentArgs);
		currentArgs = prependTerm(collVarTerm, variableParamTerms);
		Term collectTerm2 =  tf.createFunctionTerm(bagFeatureFunction, currentArgs);
		
		Term insertOrUnionTerm;

		if (!flatteningRequired){
		    
		    Function bagInsertFunction2 = 
			proverFacade.getCollectionFunctionForFeature(BAG_INSERT,
								       new Sort[]{retSort,
										  exprSort},
								       retSort);
		    
		    insertOrUnionTerm = tf.createFunctionTerm(bagInsertFunction2,collectTerm2,expressionTerm);
		    
		} else {
		    Function bagUnionFunction = 
			proverFacade.getCollectionFunctionForFeature(BAG_UNION,
								       new Sort[]{retSort,
										  exprSort},
								       retSort);
		    insertOrUnionTerm = tf.createFunctionTerm(bagUnionFunction,collectTerm2,expressionTerm);
		}
		Term eqTerm = tf.createEqualityTerm(collectTerm1,insertOrUnionTerm);
		
		newAxiom2 = constraintTranslator.createAllClosureOfTerm(eqTerm);
		
		
		log("Adding new axiom (1)  for " + functionName + " : " + newAxiom1);
		log("Adding new axiom (2)  for " + functionName + " : " + newAxiom2);
		
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(bagTranslation.getAxioms());
		result.addAxioms(expressionTranslation.getAxioms());
		result.addAxiom(newAxiom1);
		result.addAxiom(newAxiom2);

	    } else {
		// ==================== TRANSLATION AS PREDICATE  ==============================
		
		bagTranslation = 
		    getPredicativeTranslationOfTranslation(bagTranslation,collType);
			
		Term bagPredicate = bagTranslation.getTerm();
		
		LogicVariable predicateVar = bagTranslation.getPredicateVariable();
		
		// Set up the formuls term with the 'right' iteratorvariable for the formula term ...
		
		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		
		
		// Create necessary new variables and variable terms ...
		LogicVariable newPredicateVar  = newVariableDeliverer.getNewVariableForSort(exprElementSort);
		Term newPredicateVarTerm = tf.createVariableTerm(newPredicateVar);
		
		Term aTerm = null;
	    
		if (!flatteningRequired){
		    // the given collect expression has a BASIC type and therefore no flattening is required ...
		   
		    expressionTerm = renameVariableInTerm(expressionTerm, iteratorVar, predicateVar);
		    
		    Term eqTerm = tf.createEqualityTerm(newPredicateVarTerm, expressionTerm);
		    aTerm = tf.createJunctorTermAndSimplify(Op.AND, bagPredicate, eqTerm);
		    
		    resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , aTerm);


		} else {
		    // the given collecting expression has a COLLECTION type => Flattening is required ...
		   
		    if (expressionTranslation.isPredicativeTranslation()) {
			expressionTranslation = 
			    getPredicativeTranslationOfTranslation(expressionTranslation, (Collection) expressionType);
		    } else {
			expressionTerm = renameVariableInTerm(expressionTerm, iteratorVar, predicateVar);
			expressionTranslation.setTerm(expressionTerm);
			expressionTranslation = 
			    getPredicativeTranslationOfTranslation(expressionTranslation, (Collection) expressionType);
		    }
			
		    Term expressionPredicate = expressionTranslation.getTerm();
		    LogicVariable expressionPredicateVar = expressionTranslation.getPredicateVariable();
	       
		    expressionPredicate = renameVariableInTerm(expressionPredicate, expressionPredicateVar, newPredicateVar);
		
		    aTerm = tf.createJunctorTermAndSimplify(Op.AND, bagPredicate, expressionPredicate);
		    
		    resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , aTerm);

		    
		}

		log("INTO :" + resultTerm.toString());

		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(bagTranslation.getAxioms());
		result.addAxioms(expressionTranslation.getAxioms());
		result.setPredicateVariable(newPredicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    
		
	    }

	} else if(featureName.equals("iterate")){
	    // Translate iterate on bags ...

	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);

	    // The first parameter node belongs to the OCL expression that is used for initializing
	    // the accumulator variable. The second parameter is the step-expression
	    KeyTranslationResult initExpressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	     KeyTranslationResult stepExpressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(1));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term bagTerm = bagTranslation.getTerm();

	    Type initExpressionType = tree.getNodeType((Node) parameterNodes.elementAt(0));
 	    Term initExpressionTerm = initExpressionTranslation.getTerm();
	    Type stepExpressionType = tree.getNodeType((Node) parameterNodes.elementAt(1));
 	    Term stepExpressionTerm = null;

	    // Convert the representation of the expression to be computed in
	    // some suitable form ...
	    if (stepExpressionType == Basic.BOOLEAN) {
		// We have to ensure to use boolean terms (instead of formulas) ...
		stepExpressionTerm = convertFormulaToTerm(stepExpressionTranslation.getTerm(), 
							  stepExpressionTranslation);
	    } else {
		stepExpressionTerm = stepExpressionTranslation.getTerm();
	    }
	       
	    
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);

	    Sort exprSort = stepExpressionTerm.sort();
	    Sort exprElementSort;
	    
	    Sort initSort = initExpressionTerm.sort();

	    Sort retSort = exprSort;
	    
	    // Currently we only support the basic translation method for iterate ...

		// ==================== TRANSLATION AS TERM ==============================
		
	    
		// Get the iterator variable (and the corresponding term) ...

		// We have use the first iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		
		// And now the accumulator ...

		// We have use the second iterator variable here !!!!! ...
		String accumulatorVarName = (String)iteratorNames.elementAt(1);

		Type accumulatorType = initExpressionType; 
		String accumulatorTypeName = accumulatorType.toString().trim();
		
		// Get the translation of the accumulator variable ...
		KeyTranslationResult accumulatorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(accumulatorVarName,
								     accumulatorTypeName);
		
		Term accumulatorTerm = accumulatorTranslation.getTerm();
		LogicVariable accumulatorVar = (LogicVariable) accumulatorTerm.op();
		
		
		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable freeVarsStepExpr =  stepExpressionTerm.freeVars();
		SetOfQuantifiableVariable freeVarsInitExpr =  initExpressionTerm.freeVars();

		SetOfQuantifiableVariable paramsSetSmall = 
		    freeVarsStepExpr.remove(iteratorVar).remove(accumulatorVar);
		SetOfQuantifiableVariable paramsSetBig = 
		    freeVarsStepExpr.union(freeVarsInitExpr).remove(iteratorVar).remove(accumulatorVar);
	
		Term[] paramTermsStep;
		Sort[] paramSortsStep;
		Term[] variableParamTermsSmall;
		Sort[] variableParamSortsSmall;
		
		int freeVarsCntSmall = paramsSetSmall.size();
		
		paramTermsStep = new Term[freeVarsCntSmall + 2];
		paramSortsStep = new Sort[freeVarsCntSmall + 2];
		
		variableParamTermsSmall = new Term[freeVarsCntSmall];
		variableParamSortsSmall = new Sort[freeVarsCntSmall];

		// The first parameter is the current element in the iteration ...
		paramTermsStep[0] = iteratorTerm;
		paramSortsStep[0] = elemSort;

		// The second parameter is the current accumulator value in the iteration ...
		paramTermsStep[1] = accumulatorTerm;
		paramSortsStep[1] = accumulatorTerm.sort();

		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsSetSmall.iterator();
		LogicVariable currentVar;
		int i=2;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTermsStep[i] = tf.createVariableTerm(currentVar);
		    paramSortsStep[i] = currentVar.sort();
		    variableParamTermsSmall[i-2] = paramTermsStep[i];
		    variableParamSortsSmall[i-2] = paramSortsStep[i];
		    i++;
		}
		 
		Term[] paramTermsIterate;
		Sort[] paramSortsIterate;
		Term[] variableParamTermsBig;
		Sort[] variableParamSortsBig;
		
		int freeVarsCntBig = paramsSetBig.size();
		
		paramTermsIterate = new Term[freeVarsCntBig + 1];
		paramSortsIterate = new Sort[freeVarsCntBig + 1];
		
		variableParamTermsBig = new Term[freeVarsCntBig];
		variableParamSortsBig = new Sort[freeVarsCntBig];

		// The first parameter is the collection to be iterated over ...
		paramTermsIterate[0] = bagTerm;
		paramSortsIterate[0] = collSort;

		 // The rest are the necessary parameters from the other expressions ...
		varIt = paramsSetBig.iterator();
		i=1;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTermsIterate[i] = tf.createVariableTerm(currentVar);
		    paramSortsIterate[i] = currentVar.sort();
		    variableParamTermsBig[i-1] = paramTermsIterate[i];
		    variableParamSortsBig[i-1] = paramSortsIterate[i];
		    i++;
		}

		String stepFunctionName = newFunctionNameDeliverer.getNewFunctionNameForName("it-step");
		Function stepFunction =  proverFacade.getCollectionFunctionForFeature(stepFunctionName,
											paramSortsStep,
											retSort);



		String iterateFunctionName = newFunctionNameDeliverer.getNewFunctionNameForName("iterate");
		Function iterateFunction =  proverFacade.getCollectionFunctionForFeature(iterateFunctionName,
											   paramSortsIterate,
											   retSort);

	
		resultTerm = tf.createFunctionTerm(iterateFunction,paramTermsIterate);
		
		log("INTO :" + resultTerm.toString());
		
		
		// Create additional axioms for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2, newAxiom3;
		Term[] currentArgs;
		Sort[] currentSorts;

		// Now get the corresponding constant for the empty collection ...
		Function emptyBagFunction = proverFacade.getCollectionFunctionForFeature(BAG_EMPTYBAG,
											   new Sort[0],
											   collSort);
		
		Function bagInsertFunction = proverFacade.getCollectionFunctionForFeature(BAG_INSERT,
											    new Sort[]{collSort,elemSort},
											    collSort);

	
		Term emptyBagTerm = tf.createFunctionTerm(emptyBagFunction);
		
		
		
		// Create necessary variables and variable terms ...		
		LogicVariable collVar = newVariableDeliverer.getNewVariableForSort(collSort);
		Term collVarTerm = tf.createVariableTerm(collVar);
		
		Term elementVarTerm = iteratorTerm;

		// Now we can build the additional axioms ...
		
		// No.1 ...
		Term stpTerm = tf.createFunctionTerm(stepFunction,paramTermsStep);
		Term eqTerm0 = tf.createEqualityTerm(stpTerm, stepExpressionTerm);

		newAxiom1 = constraintTranslator.createAllClosureOfTerm(eqTerm0);
		
		// No. 2 ...
		currentArgs = prependTerm(emptyBagTerm,variableParamTermsBig);
		Term itTerm = tf.createFunctionTerm(iterateFunction,currentArgs);
		Term eqTerm1 = tf.createEqualityTerm(itTerm, initExpressionTerm);

		newAxiom2 = constraintTranslator.createAllClosureOfTerm(eqTerm1);
		
 
		// No.3 ...
		Term insertTerm1 =  tf.createFunctionTerm(bagInsertFunction, collVarTerm, elementVarTerm);
		currentArgs = prependTerm(insertTerm1, variableParamTermsBig);
		Term itTerm1 =  tf.createFunctionTerm(iterateFunction, currentArgs);

		currentArgs = prependTerm(collVarTerm , variableParamTermsBig);
		Term itTerm2 =  tf.createFunctionTerm(iterateFunction, currentArgs);

		currentArgs = prependTerm(itTerm2, variableParamTermsSmall);
		currentArgs = prependTerm(elementVarTerm, currentArgs);

		Term stpTerm2 = tf.createFunctionTerm(stepFunction, currentArgs);

		Term eqTerm2 = tf.createEqualityTerm(itTerm1,stpTerm2);
		
		newAxiom3 = constraintTranslator.createAllClosureOfTerm(eqTerm2);
		
		
		log("Adding new axiom (1)  for " + iterateFunctionName  + " : " + newAxiom1);
		log("Adding new axiom (2)  for " + iterateFunctionName  + " : " + newAxiom2);
		log("Adding new axiom (3)  for " + iterateFunctionName  + " : " + newAxiom3);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(bagTranslation.getAxioms());
		result.addAxioms(stepExpressionTranslation.getAxioms());
		result.addAxioms(initExpressionTranslation.getAxioms());
		result.addAxiom(newAxiom1);
		result.addAxiom(newAxiom2);
		result.addAxiom(newAxiom3);
		
	   

	} else if(featureName.equals("asSet")){
	    // Translate asSet on bags ...

	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term bagTerm = bagTranslation.getTerm();
 	  
	   
	    NonCollectionSort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort =  proverFacade.getSortForOCLSet(elemSort);

	     if(!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		 // ==================== TRANSLATION AS TERM ==============================
		 // Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable paramsForNewFunction = bagTerm.freeVars();
	
		Term[] paramTerms;
		Sort[] paramSorts;		
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt];
		paramSorts = new Sort[freeVarsCnt];
		
		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=0;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    i++;
		}
		
		 String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);
		 
		 Function bagFeatureFunction =
		     proverFacade.getCollectionFunctionForFeature(functionName,
								    paramSorts,
								    retSort);
		 
		 resultTerm = tf.createFunctionTerm(bagFeatureFunction,paramTerms);	
		 log("INTO :" + resultTerm.toString());
		 
		 // Create an additional axiom for the interpretation of the new function symbol ...
		 Term newAxiom;
		 
		 // use a predicative translation for the bag on which the feature is applied ...
		 bagTranslation = 
		     getPredicativeTranslationOfTranslation(bagTranslation,collType);
		 
		 Term bagPredicate = bagTranslation.getTerm();
		 LogicVariable predicateVar = bagTranslation.getPredicateVariable();
		 Term predicateVarTerm = tf.createVariableTerm(predicateVar);
		 
		 Function bagContainsFunction =  proverFacade.getCollectionFunctionForFeature(CONTAINS,
												new Sort[]{retSort,elemSort},
												Sort.FORMULA);
		 
		 Term containsTerm = tf.createFunctionTerm(bagContainsFunction,resultTerm,predicateVarTerm);	
		 Term aTerm = tf.createJunctorTerm(Op.EQV,bagPredicate,containsTerm);
		
		 aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);
	    
		 newAxiom = constraintTranslator.createAllClosureOfTerm(aTerm);
	  
		 log("Adding new axiom for " + functionName + " : " + newAxiom);
		 
		 result = new KeyTranslationResult();
		 result.setTerm(resultTerm);
		 result.addAxioms(bagTranslation.getAxioms());
		 result.addAxiom(newAxiom);
		 result.setTranslationType(KeyTranslationResult.NORMAL_TRANSLATION);
	   
	   } else {
		 // ==================== TRANSLATION AS A PREDICATE ==============================
		 
		 bagTranslation = 
		     getPredicativeTranslationOfTranslation(bagTranslation,collType);
			
		 Term bagPredicate = bagTranslation.getTerm();
		 LogicVariable predicateVar = bagTranslation.getPredicateVariable();

		 log("INTO :" + bagPredicate.toString());
		
		 result = new KeyTranslationResult();
		 result.setTerm(bagPredicate);
		 result.setPredicateVariable(predicateVar);
		 result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
		 result.addAxioms(bagTranslation.getAxioms());

	   }
	    
	}  else if(featureName.equals("asSequence")){
	    // Translate asSequence on bags ...
 
	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term bagTerm = bagTranslation.getTerm();
 	  
	    NonCollectionSort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort =  proverFacade.getSortForOCLSequence(elemSort);
	    Sort intSort =  proverFacade.getSortForOCLInteger();
	    String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);
	    
	     if(!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		 
		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable paramsForNewFunction = bagTerm.freeVars();
	
		Term[] paramTerms;
		Sort[] paramSorts;		
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt];
		paramSorts = new Sort[freeVarsCnt];
		
		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=0;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    i++;
		}
		 
		 Function bagFeatureFunction =
		     proverFacade.getCollectionFunctionForFeature(functionName,
								    paramSorts,
								    retSort);
		 
		 resultTerm = tf.createFunctionTerm(bagFeatureFunction,paramTerms);	
		 log("INTO :" + resultTerm.toString());
		 
		 // Create an additional axiom for the interpretation of the new function symbol ...
		 Term newAxiom;
		 
		 
		 LogicVariable predicateVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		 Term predicateVarTerm = tf.createVariableTerm(predicateVar);
		 
		 
		 Function seqCountFunction = proverFacade.getCollectionFunctionForFeature(COUNT,
											    new Sort[]{retSort,elemSort},
											    intSort);
		 
		 
		 
		 Function bagCountFunction =  proverFacade.getCollectionFunctionForFeature(COUNT,
											     new Sort[]{collSort,elemSort},
											     intSort);
		 
		 Term countTerm1 = tf.createFunctionTerm(seqCountFunction,resultTerm,predicateVarTerm);
		 Term countTerm2 = tf.createFunctionTerm(bagCountFunction,bagTerm,predicateVarTerm);	
		 
		 Term eqTerm = tf.createEqualityTerm(countTerm1,countTerm2);
		 Term aTerm; 
		 
		 aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , eqTerm);
		 
		 newAxiom = constraintTranslator.createAllClosureOfTerm(aTerm);
		 
		 log("Adding new axiom for " + functionName + " : " + newAxiom);
		 	
		 result = new KeyTranslationResult();
		 result.setTerm(resultTerm);
		 result.addAxioms(bagTranslation.getAxioms());
		 result.addAxiom(newAxiom);
		 
	     } else {
		 // ==================== TRANSLATION AS A PREDICATE ==============================
		 
		 bagTranslation = 
		     getPredicativeTranslationOfTranslation(bagTranslation,collType);
			
		 Term bagPredicate = bagTranslation.getTerm();
		 LogicVariable predicateVar = bagTranslation.getPredicateVariable();

		 result = new KeyTranslationResult();
		 result.setTerm(bagPredicate);
		 result.addAxioms(bagTranslation.getAxioms());
		 result.setPredicateVariable(predicateVar);
		 result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);

	     }

	} else if(featureName.equals("isUnique")){
	    // Translate isUnique on bags ...

	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult expressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term bagTerm = bagTranslation.getTerm();
 	    Term expressionTerm = expressionTranslation.getTerm();


	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort exprSort = expressionTerm.sort();
	    
	    // use a predicative translation for the bag on which 'isUnique' is applied ...
 	    bagTranslation = 
		getPredicativeTranslationOfTranslation(bagTranslation,collType);
	    
	    Term bagPredicate = bagTranslation.getTerm();
	    LogicVariable predicateVar = bagTranslation.getPredicateVariable();
	    
	    // Create necessary variables and variable terms ...
	    LogicVariable elementVar1 = newVariableDeliverer.getNewVariableForSort(elemSort);
	    Term elementVarTerm1 = tf.createVariableTerm(elementVar1);
	    LogicVariable elementVar2 = newVariableDeliverer.getNewVariableForSort(elemSort);
	    Term elementVarTerm2 = tf.createVariableTerm(elementVar2);

	    // We have exactely one iterator variable here !!!!! ...
	    String iteratorVarName = (String)iteratorNames.elementAt(0);
	    String iteratorTypeName = iteratorType.toString().trim();

	    // Get the translation of the iteratorVariable ...
	    KeyTranslationResult iteratorTranslation = 
		(KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								 iteratorTypeName);
	    
	    Term iteratorTerm = iteratorTranslation.getTerm();
	    LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
	    
	    Term containsTerm1 = renameVariableInTerm(bagPredicate, predicateVar, elementVar1);
	    Term containsTerm2 = renameVariableInTerm(bagPredicate, predicateVar, elementVar2);
	    Term exprTerm1 = renameVariableInTerm(expressionTerm, iteratorVar, elementVar1);
	    Term exprTerm2 = renameVariableInTerm(expressionTerm, iteratorVar, elementVar2);
	    Term eqTerm1 = tf.createEqualityTerm(exprTerm1,exprTerm2);
	    Term andTerm = tf.createJunctorTermAndSimplify(Op.AND, containsTerm1, containsTerm2);
	    andTerm = tf.createJunctorTermAndSimplify(Op.AND, andTerm, eqTerm1);
	    Term eqTerm2 =  tf.createEqualityTerm(elementVarTerm1,elementVarTerm2);
	    Term impTerm =  tf.createJunctorTermAndSimplify(Op.IMP, andTerm, eqTerm2);
	    
	    // resultTerm = tf.createQuantifierTerm(Op.ALL,new LogicVariable[]{elementVar1,elementVar2},impTerm);
	    // NEW:
	    resultTerm = getQuantifiedTermWith(Op.ALL, elementVar2 , impTerm);
	    resultTerm = getQuantifiedTermWith(Op.ALL, elementVar1 , resultTerm);

	   
	    log("INTO :" + resultTerm.toString());

	 
	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(bagTranslation.getAxioms());
	    result.addAxioms(expressionTranslation.getAxioms());
	  
	    
	} else if(featureName.equals("sortedBy")){
	    // Translate sortedBy on bags ...
 
	    KeyTranslationResult bagTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	     KeyTranslationResult expressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term bagTerm = bagTranslation.getTerm();
	    Term expressionTerm = expressionTranslation.getTerm();

	    NonCollectionSort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort =  proverFacade.getSortForOCLSequence(elemSort);
	    Sort intSort =  proverFacade.getSortForOCLInteger();

	    Sort exprSort = expressionTerm.sort();
	    if ( (exprSort != proverFacade.getSortForOCLInteger()) && 
		 (exprSort != proverFacade.getSortForOCLReal()) ){
		// We don't allow the sortedBy expression for other element types than OclInteger or OclReal ...
		error("The expression used to describe the order for the 'sortedBy' feature is only allowed to have Integer or Real as its type !!!!");
	    }

	    if(!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);
		
		
		// Collect all the necessary parameters for the new function symbol:
		// These are all free variables in the translation of the collection and the expression ...
		// except of the iteratorvariable ...
		
		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		// This is the variable we have to care about within the expression (which specifies the order !) ...
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
	    
	    
		SetOfQuantifiableVariable freeVarsBag = bagTerm.freeVars();
		SetOfQuantifiableVariable freeVarsExpr = expressionTerm.freeVars();
		
		SetOfQuantifiableVariable paramsForNewFunction = freeVarsBag.union(freeVarsExpr.remove(iteratorVar));
		
		log("FreeVars for new sorted collection '" + functionName + "' == " + paramsForNewFunction );
		
		Term[] paramVars;
		Sort[] paramSorts;
		
		int freeVarsCnt = paramsForNewFunction.size();
		if ( freeVarsCnt > 0){
		    paramVars = new Term[freeVarsCnt];
		    paramSorts = new Sort[freeVarsCnt];
		    
		    IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		    LogicVariable currentVar;
		    int i=0;
		    
		    while (varIt.hasNext()){
			currentVar = (LogicVariable) varIt.next();
			paramVars[i] = tf.createVariableTerm(currentVar);
			paramSorts[i] = currentVar.sort();
			i++;
		    }
		    
		} else {
		    paramVars = new Term[0];
		    paramSorts = new Sort[0];
		}
		
		
		// Now we can create a new function symbol for the sorted collection ...
		
		
		Function bagFeatureFunction =
		    proverFacade.getFunctionForFeature(functionName,
							 paramSorts,
							 retSort);
		
		
		// We take this new function term as the result of the translation ...
		resultTerm = 
		    tf.createFunctionTerm(bagFeatureFunction, paramVars);
		
		
		// Create additional axioms for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2;
		
		
		// 1st axiom ... The resulting sequence has the same elements as the original collection !
		
		LogicVariable predicateVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		Term predicateVarTerm = tf.createVariableTerm(predicateVar);
		
		
		Function seqCountFunction = proverFacade.getCollectionFunctionForFeature(COUNT,
											   new Sort[]{retSort,elemSort},
											   intSort);
		
		
		
		Function bagCountFunction =  proverFacade.getCollectionFunctionForFeature(COUNT,
											    new Sort[]{collSort,elemSort},
											    intSort);
		
		Term countTerm1 = tf.createFunctionTerm(seqCountFunction,resultTerm,predicateVarTerm);
		Term countTerm2 = tf.createFunctionTerm(bagCountFunction,bagTerm,predicateVarTerm);	
		
		Term eqTerm = tf.createEqualityTerm(countTerm1,countTerm2);
		Term aTerm;
		// aTerm = tf.createQuantifierTerm(Op.ALL,predicateVar, eqTerm);
		// NEW:
		aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , eqTerm);
		
		newAxiom1 = constraintTranslator.createAllClosureOfTerm(aTerm);
		
		// 2nd axiom ... The resulting sequence is ordered according to the given expression ...
	
		LogicVariable intVar1 = new LogicVariable(new Name("i"), intSort );
		LogicVariable intVar2 = new LogicVariable(new Name("j"), intSort );
		
		Term intVarTerm1 = tf.createVariableTerm(intVar1);
		Term intVarTerm2 = tf.createVariableTerm(intVar2);
		Term oneTerm = this.computeTranslationForIntegerLiteral("1");
		
		Function intLessOrEqualThanFunction = proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN,
											   new Sort[]{intSort, intSort},
											   Sort.FORMULA);
		
		Function exprLessOrEqualThanFunction = proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN,
											    new Sort[]{exprSort, exprSort},
											    Sort.FORMULA);
		
		Function collSizeFunction = 
		    proverFacade.getFunctionForFeature(SIZE,
							 new Sort[]{retSort},
							 intSort);
		
		Function collAtFunction = 
		    proverFacade.getFunctionForFeature(SEQUENCE_AT,
							 new Sort[]{retSort, intSort},
							 elemSort);
		
		Term sizeTerm = tf.createFunctionTerm(collSizeFunction, resultTerm);
		Term lessTerm1 = tf.createFunctionTerm(intLessOrEqualThanFunction, oneTerm, intVarTerm1);
		Term lessTerm2 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm1, sizeTerm);
		Term lessTerm3 = tf.createFunctionTerm(intLessOrEqualThanFunction, oneTerm, intVarTerm2);
		Term lessTerm4 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm2, sizeTerm);
		Term lessTerm5 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm1, intVarTerm2);
		
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, lessTerm1 , lessTerm2 );
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm3 );
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm4 );
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm5 );
		
		Term atTerm1 = tf.createFunctionTerm(collAtFunction, resultTerm, intVarTerm1);
		Term atTerm2 = tf.createFunctionTerm(collAtFunction, resultTerm, intVarTerm2);
		
		// Now we want to build terms like : expression(at(i))
		// Therefore we have to substitute occurences of  the iteratorvariable with the 'at(i)' term ...
		
		ClashFreeSubst subst1 = new ClashFreeSubst(iteratorVar, atTerm1);
		Term exprAtTerm1 = subst1.apply(expressionTerm);
		ClashFreeSubst subst2 = new ClashFreeSubst(iteratorVar, atTerm2);
		Term exprAtTerm2 = subst2.apply(expressionTerm);
		
		Term lessTerm6 = tf.createFunctionTerm(exprLessOrEqualThanFunction,exprAtTerm1,exprAtTerm2);
		
		Term impTerm = tf.createJunctorTermAndSimplify(Op.IMP, aTerm , lessTerm6 );
		
		Term allTerm = tf.createQuantifierTerm(Op.ALL, new LogicVariable[]{intVar1, intVar2}, impTerm);
		newAxiom2 = constraintTranslator.createAllClosureOfTerm(allTerm);
		
		
		log("Adding new axiom (1) for " + functionName + " : " + newAxiom1);
		log("Adding new axiom (2) for " + functionName + " : " + newAxiom2);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(bagTranslation.getAxioms());
		result.addAxioms(expressionTranslation.getAxioms());
		result.addAxiom(newAxiom1);
		result.addAxiom(newAxiom2);
	   
	    } else {
		 // ==================== TRANSLATION AS A PREDICATE ==============================
		 
		 bagTranslation = 
		     getPredicativeTranslationOfTranslation(bagTranslation,collType);
			
		 Term bagPredicate = bagTranslation.getTerm();
		 LogicVariable predicateVar = bagTranslation.getPredicateVariable();

		 log("INTO :" + bagPredicate.toString());
		
		 result = new KeyTranslationResult();
		 result.setTerm(bagPredicate);
		 result.addAxioms(bagTranslation.getAxioms());
		 result.setPredicateVariable(predicateVar);
		 result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	     }

	    
	    
	} else {
	    error("Translation of BAG feature " + featureName + " NOT supported yet !!!");
	}
	

	// Finally set the translation result for the current node ...
	setTranslationForNode(n, result);

   }



   // =========================================================================================================
   // ============================ The feature calls on OCL type Sequence[T] ==================================
   // =========================================================================================================



    protected void translateFeatureCallonSequence(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType, boolean  featureCallWithAtPre){
	log("Feature call on Sequence : " + featureName);
	log("Applied on node :" + appliedTo.toString());
    
	 KeyTranslationResult result = null;
	 Term resultTerm = null;
	    
	 // Get the kind of translation to be performed for the given node ...
	TranslationType translationType = getTranslationTypeForNode(n);
	log("Current feature call should be translated in the following way : " + translationType.toString());
	
	 Sort iteratorSort;

	 
	 if(featureName.equals("including") || featureName.equals("excluding") ){
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult elementTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();
 	    Term elementTerm = elementTranslation.getTerm();

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = collSort;

	    if(elemSort == proverFacade.getSortForOCLBoolean()){
		// Ensure the representation of the expression to be a boolean term (not a formula) ...
		elementTerm = convertFormulaToTerm(elementTerm, elementTranslation);
	    }

	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		
		String funcName = (featureName.equals("including")) ? SEQUENCE_INSERT : SEQUENCE_REMOVE; 
		
		Function sequenceFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(funcName,
								   new Sort[]{collSort,elemSort},
								   retSort);
	    
		resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,sequenceTerm,elementTerm);
	
		log("INTO :" + resultTerm.toString());
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxioms(elementTranslation.getAxioms());
		
	    } else {
		// ==================== PREDICATIVE TRANSLATION ==========================

		// use a predicative translation for the sequence on which 'including' is applied ...
		sequenceTranslation = 
		    getPredicativeTranslationOfTranslation(sequenceTranslation,collType);
		
		Term sequencePredicate = sequenceTranslation.getTerm();
		
		LogicVariable predicateVar = sequenceTranslation.getPredicateVariable();
		Term varTerm = tf.createVariableTerm(predicateVar);
		Term eqTerm = null;
		
		if (featureName.equals("including")){
		    // Create new predicate for including ...
		    eqTerm = tf.createEqualityTerm(varTerm, elementTerm);
		    resultTerm = tf.createJunctorTermAndSimplify(Op.OR, sequencePredicate, eqTerm);
		} else {
		    // Create new predicate for excluding ...
		    eqTerm = tf.createEqualityTerm(varTerm, elementTerm);
		    eqTerm = tf.createJunctorTerm(Op.NOT,eqTerm);
		    resultTerm = tf.createJunctorTermAndSimplify(Op.AND, sequencePredicate, eqTerm);
		}
		
		log("INTO :" + resultTerm.toString());
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxioms(elementTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);

	    }

	    
	} else if(featureName.equals("forAll") ||featureName.equals("exists") ){
	    // Translate forAll and exists on Sequences ...

	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult formulaTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    Collection collType = (Collection) tree.getNodeType(appliedTo);

	    // use a predicative translation for the sequence on which 'forAll' is applied ...
 	    sequenceTranslation = 
		getPredicativeTranslationOfTranslation(sequenceTranslation,collType);
	    
	    Term sequencePredicate = sequenceTranslation.getTerm();
	    Term formulaTerm = convertTermToFormula(formulaTranslation.getTerm());

	    LogicVariable predicateVar = sequenceTranslation.getPredicateVariable();
	    
	    log(featureName + "formulaTerm : " + formulaTerm.toString());

	    

	    // We have exactely one iterator variable here !!!!! ...
	    String iteratorVarName = (String)iteratorNames.elementAt(0);
	    String iteratorTypeName = iteratorType.toString().trim();

	    // Get the translation of the iteratorVariable ...
	    KeyTranslationResult iteratorTranslation = 
		(KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								 iteratorTypeName);
	    
	    Term iteratorTerm = iteratorTranslation.getTerm();
	    LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
	    

	    formulaTerm = renameVariableInTerm(formulaTerm, iteratorVar, predicateVar);

	    log(featureName + "formulaTerm after renaming: " + formulaTerm.toString());
	    Term aTerm = null;

	    if (featureName.equals("forAll")){
		// Create formula for forAll ...
		aTerm = tf.createJunctorTermAndSimplify(Op.IMP,sequencePredicate,formulaTerm);
		// resultTerm = tf.createQuantifierTerm(Op.ALL,predicateVar,aTerm);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);
	    } else {
		// Create formula for exists ...
		aTerm = tf.createJunctorTermAndSimplify(Op.AND,sequencePredicate,formulaTerm);
		// resultTerm = tf.createQuantifierTerm(Op.EX,predicateVar,aTerm);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , aTerm);
	    }

	    
	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(sequenceTranslation.getAxioms());
	    result.addAxioms(formulaTranslation.getAxioms());

	} else if(featureName.equals("isEmpty") ||featureName.equals("notEmpty") ){
	    // Translate isEmpty and notEmpty on Sequences ...

	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);

	    // use a predicative translation for the sequence on which 'forAll' is applied ...
 	    sequenceTranslation = 
		getPredicativeTranslationOfTranslation(sequenceTranslation,collType);
	    
	    Term sequencePredicate = sequenceTranslation.getTerm();
	    LogicVariable predicateVar = sequenceTranslation.getPredicateVariable();
	    
	    Term aTerm;


	    if (featureName.equals("isEmpty")){
		// Create formula for isEmpty ...
		aTerm = tf.createJunctorTerm(Op.NOT,sequencePredicate);
		// resultTerm = tf.createQuantifierTerm(Op.ALL,predicateVar,aTerm);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);
	    } else {
		// Create formula for notEmpty ...
		// resultTerm = tf.createQuantifierTerm(Op.EX,predicateVar,sequencePredicate);
		// NEW:
		resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , sequencePredicate);
	    }

	    
	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(sequenceTranslation.getAxioms());
    
	} else if(featureName.equals("includes")|| (featureName.equals("excludes") )){
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult elementTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();
 	    Term elementTerm = elementTranslation.getTerm();

	    // ==================== TRANSLATION AS TERM ==============================
	    
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = proverFacade.getSortForOCLBoolean();
	   
	    Function sequenceFeatureFunction =
 		proverFacade.getCollectionFunctionForFeature(CONTAINS,
 						     new Sort[]{collSort,elemSort},
 						     Sort.FORMULA);
	    
	    resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,sequenceTerm,elementTerm);
		
	    if(featureName.equals("excludes") ){
		// we have to negate the formula ...
		resultTerm = tf.createJunctorTerm(Op.NOT,resultTerm);
	    }
 	 

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(sequenceTranslation.getAxioms());
	    result.addAxioms(elementTranslation.getAxioms());
	   
	    
	}  else if(featureName.equals("includesAll") || (featureName.equals("excludesAll") )){
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);

	    Node collArgNode = (Node) parameterNodes.elementAt(0);

	    KeyTranslationResult collTranslation =
		(KeyTranslationResult) getTranslationForNode(collArgNode);
	    
	    Collection collType1 = (Collection) tree.getNodeType(appliedTo);
	    Collection collType2 = (Collection) tree.getNodeType(collArgNode);
	    
	    // use a predicative translation for the sequence on which the feature is applied ...
 	    sequenceTranslation = 
		getPredicativeTranslationOfTranslation(sequenceTranslation,collType1);
	    collTranslation = 
		getPredicativeTranslationOfTranslation(collTranslation,collType2);

	    Term sequencePredicate = sequenceTranslation.getTerm();
 	    Term collPredicate = collTranslation.getTerm();

	    Sort elemSort1 = getElementSortForCollectionType(collType1);
	    Sort collSort1 = getDLSortForType(collType1);
	    Sort elemSort2 = getElementSortForCollectionType(collType2);
	    Sort collSort2 = getDLSortForType(collType2);

	    Sort retSort = proverFacade.getSortForOCLBoolean();
	   
	    if (elemSort1 != elemSort2) {
		log("includesAll/excludesAll on collections with DIFFERENT element types is currently not supported !!!!");
	    }

	    LogicVariable predicateVar1 = sequenceTranslation.getPredicateVariable();
	    LogicVariable predicateVar2 = collTranslation.getPredicateVariable();
	    LogicVariable predicateVar = predicateVar1;

	    collPredicate = renameVariableInTerm(collPredicate, predicateVar2, predicateVar);
	    
	    // log(featureName + "collPredicate after renaming: " + collPredicate.toString());
	    Term aTerm = null;
	    
	    if (featureName.equals("excludesAll")){
		// Create formula for excludesAll ...
		aTerm = tf.createJunctorTerm(Op.NOT,sequencePredicate);
	    }  else {
		aTerm = sequencePredicate;
	    }
	    
	    aTerm = tf.createJunctorTermAndSimplify(Op.IMP,collPredicate,aTerm);
	    // resultTerm = tf.createQuantifierTerm(Op.ALL,predicateVar,aTerm);
	    // NEW:
	    resultTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(sequenceTranslation.getAxioms());
	    result.addAxioms(collTranslation.getAxioms());

	} else if(featureName.equals("size") ){
	    
	    // ==================== TRANSLATION AS TERM ==============================
	    
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Term sequenceTerm = sequenceTranslation.getTerm();

 	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = proverFacade.getSortForOCLInteger();

	    Function sequenceFeatureFunction =
		proverFacade.getCollectionFunctionForFeature(SIZE,
						     new Sort[]{collSort},
						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,sequenceTerm);
		

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(sequenceTranslation.getAxioms());
	        
	} else if(featureName.equals("count")){
	    
	    // ==================== TRANSLATION AS TERM ==============================
	    
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult elementTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();
 	    Term elementTerm = elementTranslation.getTerm();

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = proverFacade.getSortForOCLInteger();
	   
	    if(elemSort == proverFacade.getSortForOCLBoolean()){
		// Ensure the representation of the expression to be a boolean term (not a formula) ...
		elementTerm = convertFormulaToTerm(elementTerm, elementTranslation);
	    }

	    Function sequenceFeatureFunction =
 		proverFacade.getCollectionFunctionForFeature(COUNT,
 						     new Sort[]{collSort,elemSort},
 						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,sequenceTerm,elementTerm);

	    log("INTO :" + resultTerm.toString());
	   
	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(sequenceTranslation.getAxioms());
	    result.addAxioms(elementTranslation.getAxioms());
	   
	    
	} else if(featureName.equals("sum") ){
	    
	    // ==================== TRANSLATION AS TERM ==============================
	    
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Term sequenceTerm = sequenceTranslation.getTerm();

 	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Sort collSort = getDLSortForType(collType);
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort retSort = elemSort;

	    if ( (elemSort != proverFacade.getSortForOCLInteger()) && 
		 (elemSort != proverFacade.getSortForOCLReal()) ){
		// We don't allow the sum feature for other element types than OclInteger or OclReal ...
		error("The feature sum on collections is only allowed for collections with elements of type Integer or Real !!!!");
	    }

	    Function sequenceFeatureFunction =
		proverFacade.getCollectionFunctionForFeature(SUM,
						     new Sort[]{collSort},
						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,sequenceTerm);
		

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(sequenceTranslation.getAxioms());
	    
	        
	} else if(featureName.equals("union")){
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);

	    Node collArgNode = (Node) parameterNodes.elementAt(0);

	    KeyTranslationResult collTranslation =
		(KeyTranslationResult) getTranslationForNode(collArgNode);
	    
	    Collection collType1 = (Collection) tree.getNodeType(appliedTo);
	    Collection collType2 = (Collection) tree.getNodeType(collArgNode);
	    
	    
	    Term sequenceTerm = sequenceTranslation.getTerm();
 	    Term collTerm = collTranslation.getTerm();

	    Sort elemSort1 = getElementSortForCollectionType(collType1);
	    Sort collSort1 = getDLSortForType(collType1);
	    Sort elemSort2 = getElementSortForCollectionType(collType2);
	    Sort collSort2 = getDLSortForType(collType2);

	    Sort retSort = collSort2;
	   
	    if (elemSort1 != elemSort2) {
		error( featureName + " on sequences/collections with DIFFERENT element types is currently not supported !!!!");
	    }


	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================

		Function sequenceFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(SEQUENCE_UNION,
								   new Sort[]{collSort1,collSort2},
								   retSort);
		
		resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,sequenceTerm,collTerm);
		
		log("INTO :" + resultTerm.toString());
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxioms(collTranslation.getAxioms());

	    } else {

		// ==================== TRANSLATION AS PREDICATE  ==============================
		sequenceTranslation = 
		    getPredicativeTranslationOfTranslation(sequenceTranslation,collType1);
		collTranslation = 
		    getPredicativeTranslationOfTranslation(collTranslation,collType2);

		Term sequencePredicate = sequenceTranslation.getTerm();
		Term collPredicate = collTranslation.getTerm();

		retSort = proverFacade.getSortForOCLBoolean();
	   
		LogicVariable predicateVar1 = sequenceTranslation.getPredicateVariable();
		LogicVariable predicateVar2 = collTranslation.getPredicateVariable();
		LogicVariable predicateVar = predicateVar1;

		collPredicate = renameVariableInTerm(collPredicate, predicateVar2, predicateVar);
	    
		// log(featureName + "collPredicate after renaming: " + collPredicate.toString());
		Term aTerm = null;
	    		
		aTerm = tf.createJunctorTermAndSimplify(Op.OR,sequencePredicate, collPredicate);
		
		resultTerm = aTerm;

		log("INTO :" + resultTerm.toString());

		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxioms(collTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    
	    
	    }

	} else if(featureName.equals("select") ||featureName.equals("reject") ){
	    // Translate select and reject on Sequences ...

	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult formulaTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();
	    Term formulaTerm = convertTermToFormula(formulaTranslation.getTerm());

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = collSort;
	   
	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		 // Get the iterator variable (and the corresponding term) ...

		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		

		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable freeVarsExpr = formulaTerm.freeVars();
		SetOfQuantifiableVariable paramsForNewFunction = freeVarsExpr.remove(iteratorVar);
		
	
		Term[] paramTerms;
		Term[] variableParamTerms;
		Sort[] paramSorts;
		Sort[] variableParamSorts;
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt + 1];
		paramSorts = new Sort[freeVarsCnt + 1];
		
		variableParamTerms = new Term[freeVarsCnt];
		variableParamSorts = new Sort[freeVarsCnt];

		// The first parameter is the collection the feature is applied to ...
		paramTerms[0] = sequenceTerm ;
		paramSorts[0] = collSort;

		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=1;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    variableParamTerms[i-1] = paramTerms[i];
		    variableParamSorts[i-1] = paramSorts[i];
		    i++;
		}
		    

		String functionName = newFunctionNameDeliverer.getNewFunctionNameForName("select");
		
		Function sequenceFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(functionName,
								   paramSorts,
								   retSort);
		
		resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,paramTerms);
		
		log("INTO :" + resultTerm.toString());
		
		// Create additional axioms for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2, newAxiom3;
		Term[] currentArgs;

		// Now get the corresponding constant for the empty collection ...
		Function emptyCollFunction = proverFacade.getCollectionFunctionForFeature(SEQUENCE_EMPTYSEQUENCE,
											    new Sort[0],
											    retSort);
		
		Function collInsertFunction = proverFacade.getCollectionFunctionForFeature(SEQUENCE_INSERT,
											     new Sort[]{retSort,elemSort},
											     retSort);
		
		
		Term emptyCollTerm = tf.createFunctionTerm(emptyCollFunction);
		
		
		
		// Create necessary variables and variable terms ...
		LogicVariable elementVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		Term elementVarTerm = tf.createVariableTerm(elementVar);
		
		LogicVariable collVar = newVariableDeliverer.getNewVariableForSort(collSort);
		Term collVarTerm = tf.createVariableTerm(collVar);
		
		// Set up the formuls term with the 'right' iteratorvariable ...
		
		log(featureName + "formulaTerm : " + formulaTerm.toString());
		
	
		formulaTerm = renameVariableInTerm(formulaTerm, iteratorVar, elementVar);
		
		log(featureName + "formulaTerm after renaming: " + formulaTerm.toString());
		
		
		// Create the boolean expression to be used for the test ...
		Term expressionTerm, negExpressionTerm; 
		if (featureName.equals("select")) {
		    expressionTerm = formulaTerm;
		    negExpressionTerm = tf.createJunctorTerm(Op.NOT,formulaTerm);
		} else {
		    expressionTerm = tf.createJunctorTerm(Op.NOT,formulaTerm);
		    negExpressionTerm = formulaTerm;
		}
		
		// Now we can build the additional axioms ...
		
		// No.1 ...
		currentArgs = prependTerm(emptyCollTerm,variableParamTerms);
		Term selectTerm = tf.createFunctionTerm(sequenceFeatureFunction, currentArgs);
		newAxiom1 = tf.createEqualityTerm(selectTerm, emptyCollTerm);
		
		// No.2 ...
		Term insertTerm1 =  tf.createFunctionTerm(collInsertFunction, collVarTerm, elementVarTerm);
		currentArgs = prependTerm(insertTerm1, variableParamTerms);
		Term selectTerm1 =  tf.createFunctionTerm(sequenceFeatureFunction, currentArgs);
		currentArgs = prependTerm(collVarTerm , variableParamTerms);
		Term selectTerm2 =  tf.createFunctionTerm(sequenceFeatureFunction,currentArgs);
		Term insertTerm2 =  tf.createFunctionTerm(collInsertFunction, selectTerm2, elementVarTerm);
		Term eqTerm = tf.createEqualityTerm(selectTerm1,insertTerm2);
		
		Term impTerm = tf.createJunctorTermAndSimplify(Op.IMP, expressionTerm, eqTerm);
		newAxiom2 = constraintTranslator.createAllClosureOfTerm(impTerm);
		
		// No.3 ...
		eqTerm = tf.createEqualityTerm(selectTerm1,selectTerm2);
		
		impTerm = tf.createJunctorTermAndSimplify(Op.IMP, negExpressionTerm, eqTerm);
		newAxiom3 = constraintTranslator.createAllClosureOfTerm(impTerm);
		
		
		log("Adding new axiom (1)  for " + functionName + " : " + newAxiom1);
		log("Adding new axiom (2)  for " + functionName + " : " + newAxiom2);
		log("Adding new axiom (3)  for " + functionName + " : " + newAxiom3);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxioms(formulaTranslation.getAxioms());
		result.addAxiom(newAxiom1);
		result.addAxiom(newAxiom2);
		result.addAxiom(newAxiom3);
	    
	    } else {
		// ==================== TRANSLATION AS PREDICATE  ==============================
		sequenceTranslation = 
		    getPredicativeTranslationOfTranslation(sequenceTranslation,collType);
		
		Term sequencePredicate = sequenceTranslation.getTerm();
		
		LogicVariable predicateVar = sequenceTranslation.getPredicateVariable();
		
		// Set up the formuls term with the 'right' iteratorvariable for the formula term ...
		
		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		
		formulaTerm = renameVariableInTerm(formulaTerm, iteratorVar, predicateVar);
	    
		Term aTerm = null;
	    
		if (featureName.equals("select")){
		    // Create filter formula for select  ...
		    aTerm = formulaTerm;
		}  else {
		    // Create filter formula for reject  ...
		    aTerm = tf.createJunctorTerm(Op.NOT,formulaTerm);
		}
	    
	   
		resultTerm = tf.createJunctorTermAndSimplify(Op.AND, sequencePredicate, aTerm);

		log("INTO :" + resultTerm.toString());

		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxioms(formulaTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    	    
	    }


	    
	}   else if(featureName.equals("collect")){
	    // Translate collect on sequences ...
	    
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult expressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();

	    Type expressionType = tree.getNodeType((Node) parameterNodes.elementAt(0));
 	    Term expressionTerm;

	    // Convert the representation of the expression to be computed in
	    // some suitable form ...
	    if (expressionType == Basic.BOOLEAN) {
		// We have to ensure to use boolean terms (instead of formulas) ...
		expressionTerm = convertFormulaToTerm(expressionTranslation.getTerm(), 
						      expressionTranslation);
	    } else {
		expressionTerm = expressionTranslation.getTerm();
	    }

	    boolean flatteningRequired = false;
	    if (expressionType instanceof Collection) {
		log("Working on a collect producing collections => Flattening is required ...");
		flatteningRequired = true;
	    }

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);

	    Sort exprSort = expressionTerm.sort();
	    NonCollectionSort exprElementSort;
	    Sort retSort;
	    
	    if (!flatteningRequired){
		// the given collect expression has a BASIC type and therefore no flattening is required ...
		exprElementSort = (NonCollectionSort)exprSort;
		retSort = proverFacade.getSortForOCLSequence(exprElementSort);
	    } else {
		exprElementSort = getElementSortForCollectionType((Collection) expressionType);
		retSort = proverFacade.getSortForOCLSequence(exprElementSort);
	    }

	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		// Get the iterator variable (and the corresponding term) ...

		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		

		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable freeVarsExpr =  expressionTerm.freeVars();
		SetOfQuantifiableVariable paramsForNewFunction = freeVarsExpr.remove(iteratorVar);
		
	
		Term[] paramTerms;
		Term[] variableParamTerms;
		Sort[] paramSorts;
		Sort[] variableParamSorts;
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt + 1];
		paramSorts = new Sort[freeVarsCnt + 1];
		
		variableParamTerms = new Term[freeVarsCnt];
		variableParamSorts = new Sort[freeVarsCnt];

		// The first parameter is the collection the feature is applied to ...
		paramTerms[0] = sequenceTerm;
		paramSorts[0] = collSort;

		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=1;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    variableParamTerms[i-1] = paramTerms[i];
		    variableParamSorts[i-1] = paramSorts[i];
		    i++;
		}
		
		String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);
		
		Function sequenceFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(functionName,
								   paramSorts,
								   retSort);
		
		
		resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,paramTerms);
		
		log("INTO :" + resultTerm.toString());
		
		
		// Create additional axioms for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2;
		Term[] currentArgs;
		
		// Now get the corresponding constant for the empty collection ...
		Function emptySequenceFunction1 = proverFacade.getCollectionFunctionForFeature(SEQUENCE_EMPTYSEQUENCE,
												 new Sort[0],
												 collSort);
		
		Function emptySequenceFunction2 = proverFacade.getCollectionFunctionForFeature(SEQUENCE_EMPTYSEQUENCE,
												 new Sort[0],
												 retSort);
		
		Function sequenceInsertFunction1 = 
		    proverFacade.getCollectionFunctionForFeature(SEQUENCE_INSERT,
								   new Sort[]{collSort,elemSort},
								   collSort);
		
		
		
		Term emptySequenceTerm1 = tf.createFunctionTerm(emptySequenceFunction1);
		Term emptySequenceTerm2 = tf.createFunctionTerm(emptySequenceFunction2);
		
	     
		// Create necessary variables and variable terms ...
		LogicVariable elementVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		Term elementVarTerm = tf.createVariableTerm(elementVar);
		
		LogicVariable collVar = newVariableDeliverer.getNewVariableForSort(collSort);
		Term collVarTerm = tf.createVariableTerm(collVar);
		
		
		// Set up the formulas term with the 'right' iteratorvariable ...
		
		log(featureName + "expressionTerm : " + expressionTerm.toString());
		
	
		expressionTerm = renameVariableInTerm(expressionTerm, iteratorVar, elementVar);
		
		log(featureName + "expressionTerm after renaming: " + expressionTerm.toString());
		
		
		// Now we can build the additional axioms ...
		
		// No.1 ...
		currentArgs = prependTerm(emptySequenceTerm1, variableParamTerms);
		Term collectTerm = tf.createFunctionTerm(sequenceFeatureFunction, currentArgs);
		newAxiom1 = tf.createEqualityTerm(collectTerm, emptySequenceTerm2);
		
		// No.2 ...
		Term insertTerm1 =  tf.createFunctionTerm(sequenceInsertFunction1, collVarTerm, elementVarTerm);
		currentArgs = prependTerm(insertTerm1, variableParamTerms);
		Term collectTerm1 =  tf.createFunctionTerm(sequenceFeatureFunction,currentArgs);
		currentArgs = prependTerm(collVarTerm, variableParamTerms);
		Term collectTerm2 =  tf.createFunctionTerm(sequenceFeatureFunction, currentArgs);
		
		Term insertOrUnionTerm;
		
		if (!flatteningRequired){
		    
		    Function sequenceInsertFunction2 = 
			proverFacade.getCollectionFunctionForFeature(SEQUENCE_INSERT,
								       new Sort[]{retSort,
										  exprSort},
								       retSort);
		    
		    insertOrUnionTerm = tf.createFunctionTerm(sequenceInsertFunction2,collectTerm2,expressionTerm);
		    
		} else {
		    Function sequenceUnionFunction = 
			proverFacade.getCollectionFunctionForFeature(SEQUENCE_UNION,
								       new Sort[]{retSort,
										  exprSort},
								       retSort);
		    insertOrUnionTerm = tf.createFunctionTerm(sequenceUnionFunction,collectTerm2,expressionTerm);
		}
		Term eqTerm = tf.createEqualityTerm(collectTerm1,insertOrUnionTerm);
		
		newAxiom2 = constraintTranslator.createAllClosureOfTerm(eqTerm);
		
		
		log("Adding new axiom (1)  for " + functionName + " : " + newAxiom1);
		log("Adding new axiom (2)  for " + functionName + " : " + newAxiom2);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxioms(expressionTranslation.getAxioms());
		result.addAxiom(newAxiom1);
		result.addAxiom(newAxiom2);
		

	    } else {
		// ==================== TRANSLATION AS PREDICATE  ==============================
		
		sequenceTranslation = 
		    getPredicativeTranslationOfTranslation(sequenceTranslation,collType);
			
		Term sequencePredicate = sequenceTranslation.getTerm();
		
		LogicVariable predicateVar = sequenceTranslation.getPredicateVariable();
		
		// Set up the formuls term with the 'right' iteratorvariable for the formula term ...
		
		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		
		
		// Create necessary new variables and variable terms ...
		LogicVariable newPredicateVar  = newVariableDeliverer.getNewVariableForSort(exprElementSort);
		Term newPredicateVarTerm = tf.createVariableTerm(newPredicateVar);
		
		Term aTerm = null;
	    
		if (!flatteningRequired){
		    // the given collect expression has a BASIC type and therefore no flattening is required ...
		   
		    expressionTerm = renameVariableInTerm(expressionTerm, iteratorVar, predicateVar);
		    
		    Term eqTerm = tf.createEqualityTerm(newPredicateVarTerm, expressionTerm);
		    aTerm = tf.createJunctorTermAndSimplify(Op.AND, sequencePredicate, eqTerm);
		    
		    resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , aTerm);


		} else {
		    // the given collecting expression has a COLLECTION type => Flattening is required ...
		   
		    if (expressionTranslation.isPredicativeTranslation()) {
			expressionTranslation = 
			    getPredicativeTranslationOfTranslation(expressionTranslation, (Collection) expressionType);
		    } else {
			expressionTerm = renameVariableInTerm(expressionTerm, iteratorVar, predicateVar);
			expressionTranslation.setTerm(expressionTerm);
			expressionTranslation = 
			    getPredicativeTranslationOfTranslation(expressionTranslation, (Collection) expressionType);
		    }
			
		    Term expressionPredicate = expressionTranslation.getTerm();
		    LogicVariable expressionPredicateVar = expressionTranslation.getPredicateVariable();
	       
		    expressionPredicate = renameVariableInTerm(expressionPredicate, expressionPredicateVar, newPredicateVar);
		
		    aTerm = tf.createJunctorTermAndSimplify(Op.AND, sequencePredicate, expressionPredicate);
		    
		    resultTerm = getQuantifiedTermWith(Op.EX, predicateVar , aTerm);

		    
		}

		log("INTO :" + resultTerm.toString());

		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxioms(expressionTranslation.getAxioms());
		result.setPredicateVariable(newPredicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    
		
	    }

	} else if(featureName.equals("iterate")){
	    // Translate iterate on sequences ...

	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);

	    // The first parameter node belongs to the OCL expression that is used for initializing
	    // the accumulator variable. The second parameter is the step-expression
	    KeyTranslationResult initExpressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	     KeyTranslationResult stepExpressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(1));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();

	    Type initExpressionType = tree.getNodeType((Node) parameterNodes.elementAt(0));
 	    Term initExpressionTerm = initExpressionTranslation.getTerm();
	    Type stepExpressionType = tree.getNodeType((Node) parameterNodes.elementAt(1));
 	    Term stepExpressionTerm = null;

	    // Convert the representation of the expression to be computed in
	    // some suitable form ...
	    if (stepExpressionType == Basic.BOOLEAN) {
		// We have to ensure to use boolean terms (instead of formulas) ...
		stepExpressionTerm = convertFormulaToTerm(stepExpressionTranslation.getTerm(), 
							  stepExpressionTranslation);
	    } else {
		stepExpressionTerm = stepExpressionTranslation.getTerm();
	    }
	       
	    
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);

	    Sort exprSort = stepExpressionTerm.sort();
	    Sort exprElementSort;
	    
	    Sort initSort = initExpressionTerm.sort();

	    Sort retSort = exprSort;
	    
	    // Currently we only support the basic translation method for iterate ...

		// ==================== TRANSLATION AS TERM ==============================
		
	    
		// Get the iterator variable (and the corresponding term) ...

		// We have use the first iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
		
		// And now the accumulator ...

		// We have use the second iterator variable here !!!!! ...
		String accumulatorVarName = (String)iteratorNames.elementAt(1);

		Type accumulatorType = initExpressionType; 
		String accumulatorTypeName = accumulatorType.toString().trim();
		
		// Get the translation of the accumulator variable ...
		KeyTranslationResult accumulatorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(accumulatorVarName,
								     accumulatorTypeName);
		
		Term accumulatorTerm = accumulatorTranslation.getTerm();
		LogicVariable accumulatorVar = (LogicVariable) accumulatorTerm.op();
		
		
		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable freeVarsStepExpr =  stepExpressionTerm.freeVars();
		SetOfQuantifiableVariable freeVarsInitExpr =  initExpressionTerm.freeVars();

		SetOfQuantifiableVariable paramsSetSmall = 
		    freeVarsStepExpr.remove(iteratorVar).remove(accumulatorVar);
		SetOfQuantifiableVariable paramsSetBig = 
		    freeVarsStepExpr.union(freeVarsInitExpr).remove(iteratorVar).remove(accumulatorVar);
	
		Term[] paramTermsStep;
		Sort[] paramSortsStep;
		Term[] variableParamTermsSmall;
		Sort[] variableParamSortsSmall;
		
		int freeVarsCntSmall = paramsSetSmall.size();
		
		paramTermsStep = new Term[freeVarsCntSmall + 2];
		paramSortsStep = new Sort[freeVarsCntSmall + 2];
		
		variableParamTermsSmall = new Term[freeVarsCntSmall];
		variableParamSortsSmall = new Sort[freeVarsCntSmall];

		// The first parameter is the current element in the iteration ...
		paramTermsStep[0] = iteratorTerm;
		paramSortsStep[0] = elemSort;

		// The second parameter is the current accumulator value in the iteration ...
		paramTermsStep[1] = accumulatorTerm;
		paramSortsStep[1] = accumulatorTerm.sort();

		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsSetSmall.iterator();
		LogicVariable currentVar;
		int i=2;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTermsStep[i] = tf.createVariableTerm(currentVar);
		    paramSortsStep[i] = currentVar.sort();
		    variableParamTermsSmall[i-2] = paramTermsStep[i];
		    variableParamSortsSmall[i-2] = paramSortsStep[i];
		    i++;
		}
		 
		Term[] paramTermsIterate;
		Sort[] paramSortsIterate;
		Term[] variableParamTermsBig;
		Sort[] variableParamSortsBig;
		
		int freeVarsCntBig = paramsSetBig.size();
		
		paramTermsIterate = new Term[freeVarsCntBig + 1];
		paramSortsIterate = new Sort[freeVarsCntBig + 1];
		
		variableParamTermsBig = new Term[freeVarsCntBig];
		variableParamSortsBig = new Sort[freeVarsCntBig];

		// The first parameter is the collection to be iterated over ...
		paramTermsIterate[0] = sequenceTerm;
		paramSortsIterate[0] = collSort;

		 // The rest are the necessary parameters from the other expressions ...
		varIt = paramsSetBig.iterator();
		i=1;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTermsIterate[i] = tf.createVariableTerm(currentVar);
		    paramSortsIterate[i] = currentVar.sort();
		    variableParamTermsBig[i-1] = paramTermsIterate[i];
		    variableParamSortsBig[i-1] = paramSortsIterate[i];
		    i++;
		}

		String stepFunctionName = newFunctionNameDeliverer.getNewFunctionNameForName("it-step");
		Function stepFunction =  proverFacade.getCollectionFunctionForFeature(stepFunctionName,
											paramSortsStep,
											retSort);



		String iterateFunctionName = newFunctionNameDeliverer.getNewFunctionNameForName("iterate");
		Function iterateFunction =  proverFacade.getCollectionFunctionForFeature(iterateFunctionName,
											   paramSortsIterate,
											   retSort);

	
		resultTerm = tf.createFunctionTerm(iterateFunction,paramTermsIterate);
		
		log("INTO :" + resultTerm.toString());
		
		
		// Create additional axioms for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2, newAxiom3;
		Term[] currentArgs;
		Sort[] currentSorts;

		// Now get the corresponding constant for the empty collection ...
		Function emptySequenceFunction = proverFacade.getCollectionFunctionForFeature(SEQUENCE_EMPTYSEQUENCE,
											   new Sort[0],
											   collSort);
		
		Function sequenceInsertFunction = proverFacade.getCollectionFunctionForFeature(SEQUENCE_INSERT,
											    new Sort[]{collSort,elemSort},
											    collSort);

	
		Term emptySequenceTerm = tf.createFunctionTerm(emptySequenceFunction);
		
		
		
		// Create necessary variables and variable terms ...		
		LogicVariable collVar = newVariableDeliverer.getNewVariableForSort(collSort);
		Term collVarTerm = tf.createVariableTerm(collVar);
		
		Term elementVarTerm = iteratorTerm;

		// Now we can build the additional axioms ...
		
		// No.1 ...
		Term stpTerm = tf.createFunctionTerm(stepFunction,paramTermsStep);
		Term eqTerm0 = tf.createEqualityTerm(stpTerm, stepExpressionTerm);

		newAxiom1 = constraintTranslator.createAllClosureOfTerm(eqTerm0);
		
		// No. 2 ...
		currentArgs = prependTerm(emptySequenceTerm,variableParamTermsBig);
		Term itTerm = tf.createFunctionTerm(iterateFunction,currentArgs);
		Term eqTerm1 = tf.createEqualityTerm(itTerm, initExpressionTerm);

		newAxiom2 = constraintTranslator.createAllClosureOfTerm(eqTerm1);
		
 
		// No.3 ...
		Term insertTerm1 =  tf.createFunctionTerm(sequenceInsertFunction, collVarTerm, elementVarTerm);
		currentArgs = prependTerm(insertTerm1, variableParamTermsBig);
		Term itTerm1 =  tf.createFunctionTerm(iterateFunction, currentArgs);

		currentArgs = prependTerm(collVarTerm , variableParamTermsBig);
		Term itTerm2 =  tf.createFunctionTerm(iterateFunction, currentArgs);

		currentArgs = prependTerm(itTerm2, variableParamTermsSmall);
		currentArgs = prependTerm(elementVarTerm, currentArgs);

		Term stpTerm2 = tf.createFunctionTerm(stepFunction, currentArgs);

		Term eqTerm2 = tf.createEqualityTerm(itTerm1,stpTerm2);
		
		newAxiom3 = constraintTranslator.createAllClosureOfTerm(eqTerm2);
		
		
		log("Adding new axiom (1)  for " + iterateFunctionName  + " : " + newAxiom1);
		log("Adding new axiom (2)  for " + iterateFunctionName  + " : " + newAxiom2);
		log("Adding new axiom (3)  for " + iterateFunctionName  + " : " + newAxiom3);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxioms(stepExpressionTranslation.getAxioms());
		result.addAxioms(initExpressionTranslation.getAxioms());
		result.addAxiom(newAxiom1);
		result.addAxiom(newAxiom2);
		result.addAxiom(newAxiom3);
		
	   

	} else if(featureName.equals("asSet")){
	    // Translate asSet on sequences ...
 
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();
 	  
	    NonCollectionSort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort =  proverFacade.getSortForOCLSet(elemSort);
	    String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);

	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		 // Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable paramsForNewFunction = sequenceTerm.freeVars();
	
		Term[] paramTerms;
		Sort[] paramSorts;		
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt];
		paramSorts = new Sort[freeVarsCnt];
		
		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=0;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    i++;
		}
		
		Function sequenceFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(functionName,
								   paramSorts,
								   retSort);

		resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,paramTerms);	
		log("INTO :" + resultTerm.toString());
		
		// Create an additional axiom for the interpretation of the new function symbol ...
		Term newAxiom;
		
		// use a predicative translation for the sequence on which the feature is applied ...
		sequenceTranslation = 
		    getPredicativeTranslationOfTranslation(sequenceTranslation,collType);
		
		Term sequencePredicate = sequenceTranslation.getTerm();
		LogicVariable predicateVar = sequenceTranslation.getPredicateVariable();
		Term predicateVarTerm = tf.createVariableTerm(predicateVar);
		
		Function sequenceContainsFunction =  proverFacade.getCollectionFunctionForFeature(CONTAINS,
												    new Sort[]{retSort,elemSort},
												    Sort.FORMULA);
		
		Term containsTerm = tf.createFunctionTerm(sequenceContainsFunction,resultTerm,predicateVarTerm);	
		Term aTerm = tf.createJunctorTerm(Op.EQV,sequencePredicate,containsTerm);
		
		aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , aTerm);
		
		newAxiom = constraintTranslator.createAllClosureOfTerm(aTerm);
		
		log("Adding new axiom for " + functionName + " : " + newAxiom);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxiom(newAxiom);
		result.setTranslationType(KeyTranslationResult.NORMAL_TRANSLATION);
	   
	     } else {
		 // ==================== TRANSLATION AS A PREDICATE ==============================
		 
		 sequenceTranslation = 
		     getPredicativeTranslationOfTranslation(sequenceTranslation,collType);
			
		 Term sequencePredicate = sequenceTranslation.getTerm();
		 LogicVariable predicateVar = sequenceTranslation.getPredicateVariable();

		 log("INTO :" + sequencePredicate.toString());
		
		 result = new KeyTranslationResult();
		 result.setTerm(sequencePredicate);
		 result.addAxioms(sequenceTranslation.getAxioms());
		 result.setPredicateVariable(predicateVar);
		 result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	     }
	    
	    
	}  else if(featureName.equals("asBag")){
	    // Translate asBag on sequences ...
 
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();
 	  
	    NonCollectionSort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort =  proverFacade.getSortForOCLBag(elemSort);
	    Sort intSort =  proverFacade.getSortForOCLInteger();
	   
	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
		// Now compute the necessary parameters ...
       
		SetOfQuantifiableVariable paramsForNewFunction = sequenceTerm.freeVars();
	
		Term[] paramTerms;
		Sort[] paramSorts;		
		
		int freeVarsCnt = paramsForNewFunction.size();
		
		paramTerms = new Term[freeVarsCnt];
		paramSorts = new Sort[freeVarsCnt];
		
		// The rest are the necessary parameters from the other expressions ...
		IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		LogicVariable currentVar;
		int i=0;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramTerms[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    i++;
		}
		 

		String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);
 
		Function sequenceFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(functionName,
								   paramSorts,
								   retSort);
		
		resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,paramTerms);	
		log("INTO :" + resultTerm.toString());
		
		// Create an additional axiom for the interpretation of the new function symbol ...
		Term newAxiom;
		
		
		LogicVariable predicateVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		Term predicateVarTerm = tf.createVariableTerm(predicateVar);
		
		
		Function bagCountFunction = proverFacade.getCollectionFunctionForFeature(COUNT,
											   new Sort[]{retSort,elemSort},
											   intSort);
		
		
		
		Function seqCountFunction =  proverFacade.getCollectionFunctionForFeature(COUNT,
											    new Sort[]{collSort,elemSort},
											    intSort);
		
		Term countTerm1 = tf.createFunctionTerm(bagCountFunction,resultTerm,predicateVarTerm);
		Term countTerm2 = tf.createFunctionTerm(seqCountFunction,sequenceTerm,predicateVarTerm);	
		
		Term eqTerm = tf.createEqualityTerm(countTerm1,countTerm2);
		Term aTerm;
		
		aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , eqTerm);
		
		newAxiom = constraintTranslator.createAllClosureOfTerm(aTerm);
		
		log("Adding new axiom for " + functionName + " : " + newAxiom);
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxiom(newAxiom);
		
	    } else {
		 // ==================== TRANSLATION AS A PREDICATE ==============================
		 
		 sequenceTranslation = 
		     getPredicativeTranslationOfTranslation(sequenceTranslation,collType);
			
		 Term sequencePredicate = sequenceTranslation.getTerm();
		 LogicVariable predicateVar = sequenceTranslation.getPredicateVariable();

		 log("INTO :" + sequencePredicate.toString());
		 
		 result = new KeyTranslationResult();
		 result.setTerm(sequencePredicate);
		 result.addAxioms(sequenceTranslation.getAxioms());
		 result.setPredicateVariable(predicateVar);
		 result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	     }
	    
	    
	} else if(featureName.equals("first") || featureName.equals("last") ){
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	    Term sequenceTerm = sequenceTranslation.getTerm();

 	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = elemSort;
	    
	    // ==================== TRANSLATION AS TERM ==============================
	    
	    String funcName = featureName.equals("first") ? SEQUENCE_FIRST : SEQUENCE_LAST;
	    
	    Function sequenceFeatureFunction =
		proverFacade.getCollectionFunctionForFeature(funcName,
						     new Sort[]{collSort},
						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,sequenceTerm);
		

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(sequenceTranslation.getAxioms());
	        
	} else if(featureName.equals("at")){
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult indexTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();
 	    Term indexTerm = indexTranslation.getTerm();

	    // ==================== TRANSLATION AS TERM ==============================
	    
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort intSort = proverFacade.getSortForOCLInteger();
	    Sort retSort = elemSort;
	   
	    Function sequenceFeatureFunction =
 		proverFacade.getCollectionFunctionForFeature(SEQUENCE_AT,
 						     new Sort[]{collSort,intSort},
 						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,sequenceTerm,indexTerm);

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(sequenceTranslation.getAxioms());
	    result.addAxioms(indexTranslation.getAxioms());
	   
	    
	} else if(featureName.equals("prepend") || featureName.equals("append") ){
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult elementTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();
 	    Term elementTerm = elementTranslation.getTerm();

	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort = collSort;

	    if(elemSort == proverFacade.getSortForOCLBoolean()){
		// Ensure the representation of the expression to be a boolean term (not a formula) ...
		elementTerm = convertFormulaToTerm(elementTerm, elementTranslation);
	    }
	    
	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
	    
		String funcName = featureName.equals("prepend") ? SEQUENCE_PREPEND : SEQUENCE_APPEND ;
		
		Function sequenceFeatureFunction =
		    proverFacade.getCollectionFunctionForFeature(funcName,
								   new Sort[]{collSort,elemSort},
								   retSort);
		
		resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,sequenceTerm,elementTerm);
		
		log("INTO :" + resultTerm.toString());
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxioms(elementTranslation.getAxioms());
		
	    } else {
		// ==================== PREDICATIVE TRANSLATION ==========================

		// use a predicative translation for the sequence on which 'including' is applied ...
		sequenceTranslation = 
		    getPredicativeTranslationOfTranslation(sequenceTranslation,collType);
		
		Term sequencePredicate = sequenceTranslation.getTerm();
		
		LogicVariable predicateVar = sequenceTranslation.getPredicateVariable();
		Term varTerm = tf.createVariableTerm(predicateVar);
		Term eqTerm = null;
		
		
		eqTerm = tf.createEqualityTerm(varTerm, elementTerm);
		resultTerm = tf.createJunctorTermAndSimplify(Op.OR, sequencePredicate, eqTerm);
		
		log("INTO :" + resultTerm.toString());
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(sequenceTranslation.getAxioms());
		result.addAxioms(elementTranslation.getAxioms());
		result.setPredicateVariable(predicateVar);
		result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);

	    }

	    
	    
	}  else if(featureName.equals("subSequence")){
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult indexTranslation1 =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    KeyTranslationResult indexTranslation2 =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(1));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();
 	    Term indexTerm1 = indexTranslation1.getTerm();
	    Term indexTerm2 = indexTranslation2.getTerm();

	    // ==================== TRANSLATION AS TERM ==============================
	    
	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort intSort = proverFacade.getSortForOCLInteger();
	    Sort retSort = collSort;
	   
	    Function sequenceFeatureFunction =
 		proverFacade.getCollectionFunctionForFeature(SEQUENCE_SUBSEQUENCE,
 						     new Sort[]{collSort,intSort, intSort},
 						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(sequenceFeatureFunction,
					       new Term[]{sequenceTerm, indexTerm1, indexTerm2});

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(sequenceTranslation.getAxioms());
	    result.addAxioms(indexTranslation1.getAxioms());
	    result.addAxioms(indexTranslation2.getAxioms());
	   
	    
	}  else if(featureName.equals("isUnique")){
	    // Translate isUnique on sequences ...

	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult expressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();
 	    Term expressionTerm = expressionTranslation.getTerm();


	    Sort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort exprSort = expressionTerm.sort();
	    
	    // use a predicative translation for the sequence on which 'isUnique' is applied ...
 	    sequenceTranslation = 
		getPredicativeTranslationOfTranslation(sequenceTranslation,collType);
	    
	    Term sequencePredicate = sequenceTranslation.getTerm();
	    LogicVariable predicateVar = sequenceTranslation.getPredicateVariable();
	    
	    // Create necessary variables and variable terms ...
	    LogicVariable elementVar1 = newVariableDeliverer.getNewVariableForSort(elemSort);
	    Term elementVarTerm1 = tf.createVariableTerm(elementVar1);
	    LogicVariable elementVar2 = newVariableDeliverer.getNewVariableForSort(elemSort);
	    Term elementVarTerm2 = tf.createVariableTerm(elementVar2);

	    // We have exactely one iterator variable here !!!!! ...
	    String iteratorVarName = (String)iteratorNames.elementAt(0);
	    String iteratorTypeName = iteratorType.toString().trim();

	    // Get the translation of the iteratorVariable ...
	    KeyTranslationResult iteratorTranslation = 
		(KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								 iteratorTypeName);
	    
	    Term iteratorTerm = iteratorTranslation.getTerm();
	    LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
	    
	    Term containsTerm1 = renameVariableInTerm(sequencePredicate, predicateVar, elementVar1);
	    Term containsTerm2 = renameVariableInTerm(sequencePredicate, predicateVar, elementVar2);
	    Term exprTerm1 = renameVariableInTerm(expressionTerm, iteratorVar, elementVar1);
	    Term exprTerm2 = renameVariableInTerm(expressionTerm, iteratorVar, elementVar2);
	    Term eqTerm1 = tf.createEqualityTerm(exprTerm1,exprTerm2);
	    Term andTerm = tf.createJunctorTermAndSimplify(Op.AND, containsTerm1, containsTerm2);
	    andTerm = tf.createJunctorTermAndSimplify(Op.AND, andTerm, eqTerm1);
	    Term eqTerm2 =  tf.createEqualityTerm(elementVarTerm1,elementVarTerm2);
	    Term impTerm =  tf.createJunctorTermAndSimplify(Op.IMP, andTerm, eqTerm2);
	    
	    // resultTerm = tf.createQuantifierTerm(Op.ALL,new LogicVariable[]{elementVar1,elementVar2},impTerm);
	    // NEW:
	    resultTerm = getQuantifiedTermWith(Op.ALL, elementVar2 , impTerm);
	    resultTerm = getQuantifiedTermWith(Op.ALL, elementVar1 , resultTerm);

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(sequenceTranslation.getAxioms());
	    result.addAxioms(expressionTranslation.getAxioms());
	  
	    
	}  else if(featureName.equals("sortedBy")){
	    // Translate sortedBy on sequences ...
 
	    KeyTranslationResult sequenceTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    
	     KeyTranslationResult expressionTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Collection collType = (Collection) tree.getNodeType(appliedTo);
	    Term sequenceTerm = sequenceTranslation.getTerm();
	    Term expressionTerm = expressionTranslation.getTerm();

	    NonCollectionSort elemSort = getElementSortForCollectionType(collType);
	    Sort collSort = getDLSortForType(collType);
	    Sort retSort =  proverFacade.getSortForOCLSequence(elemSort);
	    Sort intSort =  proverFacade.getSortForOCLInteger();

	    Sort exprSort = expressionTerm.sort();
	    if ( (exprSort != proverFacade.getSortForOCLInteger()) && 
		 (exprSort != proverFacade.getSortForOCLReal()) ){
		// We don't allow the sortedBy expression for other element types than OclInteger or OclReal ...
		error("The expression used to describe the order for the 'sortedBy' feature is only allowed to have Integer or Real as its type !!!!");
	    }

	    if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		// ==================== TRANSLATION AS TERM ==============================
	    
		String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(featureName);
	    
		// Collect all the necessary parameters for the new function symbol:
		// These are all free variables in the translation of the collection and the expression ...
		
		// except of the iteratorvariable ...
		
		// We have exactely one iterator variable here !!!!! ...
		String iteratorVarName = (String)iteratorNames.elementAt(0);
		String iteratorTypeName = iteratorType.toString().trim();
		
		// Get the translation of the iteratorVariable ...
		KeyTranslationResult iteratorTranslation = 
		    (KeyTranslationResult) getTranslationForVariable(iteratorVarName,
								     iteratorTypeName);
		
		Term iteratorTerm = iteratorTranslation.getTerm();
		// This is the variable we have to care about within the expression (which specifies the order !) ...
		LogicVariable iteratorVar = (LogicVariable) iteratorTerm.op();
	    

		SetOfQuantifiableVariable freeVarsSequence = sequenceTerm.freeVars();
		SetOfQuantifiableVariable freeVarsExpr = expressionTerm.freeVars();
		
		SetOfQuantifiableVariable paramsForNewFunction = freeVarsSequence.union(freeVarsExpr.remove(iteratorVar));
		
		log("FreeVars for new sorted collection '" + functionName + "' == " + paramsForNewFunction );
		
		Term[] paramVars;
		Sort[] paramSorts;
		
		int freeVarsCnt = paramsForNewFunction.size();
		if ( freeVarsCnt > 0){
		    paramVars = new Term[freeVarsCnt];
		    paramSorts = new Sort[freeVarsCnt];
		    
		    IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		    LogicVariable currentVar;
		    int i=0;
		    
		    while (varIt.hasNext()){
			currentVar = (LogicVariable) varIt.next();
			paramVars[i] = tf.createVariableTerm(currentVar);
			paramSorts[i] = currentVar.sort();
			i++;
		    }
		    
		} else {
		    paramVars = new Term[0];
		    paramSorts = new Sort[0];
		}
		
		
		// Now we can create a new function symbol for the sorted collection ...
		
		
		Function sequenceFeatureFunction =
		    proverFacade.getFunctionForFeature(functionName,
							 paramSorts,
							 retSort);
		
 
		// We take this new function term as the result of the translation ...
		resultTerm = 
		    tf.createFunctionTerm(sequenceFeatureFunction, paramVars);
		
		
		// Create additional axioms for the interpretation of the new function symbol ...
		Term newAxiom1, newAxiom2;

		
		// 1st axiom ... The resulting sequence has the same elements as the original collection !
		
		LogicVariable predicateVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		Term predicateVarTerm = tf.createVariableTerm(predicateVar);
		
		
		Function seqCountFunction = proverFacade.getCollectionFunctionForFeature(COUNT,
											   new Sort[]{retSort,elemSort},
											   intSort);
		
	    
		Function sequenceCountFunction =  proverFacade.getCollectionFunctionForFeature(COUNT,
												 new Sort[]{collSort,elemSort},
												 intSort);
		
		Term countTerm1 = tf.createFunctionTerm(seqCountFunction,resultTerm,predicateVarTerm);
		Term countTerm2 = tf.createFunctionTerm(sequenceCountFunction,sequenceTerm,predicateVarTerm);	
		
		Term eqTerm = tf.createEqualityTerm(countTerm1,countTerm2);
		Term aTerm;
		// aTerm = tf.createQuantifierTerm(Op.ALL,predicateVar, eqTerm);
		// NEW:
		aTerm = getQuantifiedTermWith(Op.ALL, predicateVar , eqTerm);
		
		newAxiom1 = constraintTranslator.createAllClosureOfTerm(aTerm);
		
		// 2nd axiom ... The resulting sequence is ordered according to the given expression ...
		
		
		
		LogicVariable intVar1 = new LogicVariable(new Name("i"), intSort );
		LogicVariable intVar2 = new LogicVariable(new Name("j"), intSort );
		
		Term intVarTerm1 = tf.createVariableTerm(intVar1);
		Term intVarTerm2 = tf.createVariableTerm(intVar2);
		Term oneTerm = this.computeTranslationForIntegerLiteral("1");
		
		Function intLessOrEqualThanFunction = proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN,
											   new Sort[]{intSort, intSort},
											   Sort.FORMULA);
		
		Function exprLessOrEqualThanFunction = proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN,
											    new Sort[]{exprSort, exprSort},
											    Sort.FORMULA);
		
		Function collSizeFunction = 
		    proverFacade.getFunctionForFeature(SIZE,
							 new Sort[]{retSort},
							 intSort);
		
		Function collAtFunction = 
		    proverFacade.getFunctionForFeature(SEQUENCE_AT,
							 new Sort[]{retSort, intSort},
							 elemSort);
		
		Term sizeTerm = tf.createFunctionTerm(collSizeFunction, resultTerm);
		Term lessTerm1 = tf.createFunctionTerm(intLessOrEqualThanFunction, oneTerm, intVarTerm1);
		Term lessTerm2 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm1, sizeTerm);
		Term lessTerm3 = tf.createFunctionTerm(intLessOrEqualThanFunction, oneTerm, intVarTerm2);
		Term lessTerm4 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm2, sizeTerm);
		Term lessTerm5 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm1, intVarTerm2);
		
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, lessTerm1 , lessTerm2 );
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm3 );
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm4 );
		aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm5 );
		
		Term atTerm1 = tf.createFunctionTerm(collAtFunction, resultTerm, intVarTerm1);
		Term atTerm2 = tf.createFunctionTerm(collAtFunction, resultTerm, intVarTerm2);
		
		// Now we want to build terms like : expression(at(i))
		// Therefore we have to substitute occurences of  the iteratorvariable with the 'at(i)' term ...
		
		ClashFreeSubst subst1 = new ClashFreeSubst(iteratorVar, atTerm1);
		Term exprAtTerm1 = subst1.apply(expressionTerm);
		ClashFreeSubst subst2 = new ClashFreeSubst(iteratorVar, atTerm2);
		Term exprAtTerm2 = subst2.apply(expressionTerm);
		
		Term lessTerm6 = tf.createFunctionTerm(exprLessOrEqualThanFunction,exprAtTerm1,exprAtTerm2);
		
		Term impTerm = tf.createJunctorTermAndSimplify(Op.IMP, aTerm , lessTerm6 );
		
		Term allTerm = tf.createQuantifierTerm(Op.ALL, new LogicVariable[]{intVar1, intVar2}, impTerm);
		newAxiom2 = constraintTranslator.createAllClosureOfTerm(allTerm);
		
		
		log("Adding new axiom (1) for " + functionName + " : " + newAxiom1);
		log("Adding new axiom (2) for " + functionName + " : " + newAxiom2);
		
		 result = new KeyTranslationResult();
		 result.setTerm(resultTerm);
		 result.addAxioms(sequenceTranslation.getAxioms());
		 result.addAxioms(expressionTranslation.getAxioms());
		 result.addAxiom(newAxiom1);
		 result.addAxiom(newAxiom2);
	    } else {
		 // ==================== TRANSLATION AS A PREDICATE ==============================
		 
		 sequenceTranslation = 
		     getPredicativeTranslationOfTranslation(sequenceTranslation,collType);
			
		 Term sequencePredicate = sequenceTranslation.getTerm();
		 LogicVariable predicateVar = sequenceTranslation.getPredicateVariable();

		 log("INTO :" + sequencePredicate.toString());
		
		 result = new KeyTranslationResult();
		 result.setTerm(sequencePredicate);
		 result.addAxioms(sequenceTranslation.getAxioms());
		 result.setPredicateVariable(predicateVar);
		 result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    }
	    
	    
	} else {
	    error("Translation of SEQUENCE feature " + featureName + " NOT supported yet !!!");
	}
	

	// Finally set the translation result for the current node ...
	setTranslationForNode(n, result);

    }
    

   
   // =========================================================================================================
   // ================================= The feature calls on UML model types ==================================
   // =========================================================================================================



    protected void translateFeatureCallonModelType(Node n, String featureName, Node appliedTo, 
						   Vector parameterNodes, Vector iteratorNames, 
						   boolean  featureCallWithAtPre, boolean featureCallWithArrow, 
						   boolean isMethodCall){
	log("Translating feature call on UML Model type : " + featureName);
	log("Applied on node :" + appliedTo.toString());
	
	// Get the kind of translation to be performed for the given node ...
	TranslationType translationType = getTranslationTypeForNode(n);
	log("Current feature call should be translated in the following way : " + translationType.toString());
	

	// get the return type of the feature call ...
	Type returnType = tree.getNodeType(n);
	
	log("Return type = " + returnType.toString());

	KeyTranslationResult resultTranslation;
	Term resultTerm;
	
	if (featureName.equals("oclIsKindOf")) {
	    KeyTranslationResult objectTranslation, typeTranslation;
	    
	    objectTranslation = (KeyTranslationResult) getTranslationForNode(appliedTo);
	    Term objectTerm = objectTranslation.getTerm();
	    Sort objectSort = objectTerm.sort();
	    

	    typeTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    Term oclTypeTerm = typeTranslation.getTerm();
	    Function oclTypeConstant = (Function) oclTypeTerm.op();

	    String oclTypeName =   oclTypeConstant.name().toString();

	    int index = oclTypeName.indexOf("::");
	    if (index != -1){
		oclTypeName = oclTypeName.substring(index + 2);
	    }
	   
	   
	    Sort sortForGivenOclType =  getDLSortForNonCollectionType(oclTypeName); 
	      
	    LogicVariable predicateVar = newVariableDeliverer.getNewVariableForSort(sortForGivenOclType);
	    Term predicateVarTerm = tf.createVariableTerm(predicateVar);
	    
	    Term eqTerm = tf.createEqualityTerm(predicateVarTerm, objectTerm);
	    
	    resultTerm = getQuantifiedTermWith(Op.EX, predicateVar, eqTerm);

	    log("INTO :" + resultTerm.toString());
	    
	    
	    resultTranslation = new KeyTranslationResult();
	    resultTranslation.setTerm(resultTerm);
	    resultTranslation.addAxioms(typeTranslation.getAxioms());
	    resultTranslation.addAxioms(objectTranslation.getAxioms());
	
	} else if (featureName.equals("oclIsTypeOf")){

	    // Currently this is not the desired version, since
	    // we need to access the UML model in order to create a formula as for 'oclIsKindOf' !!!!

	    KeyTranslationResult objectTranslation, typeTranslation;
	    
	    objectTranslation = (KeyTranslationResult) getTranslationForNode(appliedTo);
	    Term objectTerm = objectTranslation.getTerm();
	    Sort objectSort = objectTerm.sort();
	    

	    typeTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    Term oclTypeTerm = typeTranslation.getTerm();

	    Sort oclTypeSort =  proverFacade.getSortForOCLType(); 
	    Sort boolSort = proverFacade.getSortForOCLBoolean();
 
	    String featureFunctionName;
	    
		featureFunctionName = IS_TYPE_OF;
	   
	    
	    // Now get the corresponding function symbol ...
	    Function featureFunction = proverFacade.getFunctionForFeature(featureFunctionName,
									    new Sort[]{objectSort, oclTypeSort},
									    boolSort);

	    resultTerm = tf.createFunctionTerm(featureFunction,objectTerm,oclTypeTerm);
	    
	    log("INTO :" + resultTerm.toString());
	    
	    
	    resultTranslation = objectTranslation;
	    resultTranslation.setTerm(resultTerm);
	    resultTranslation.addAxioms(typeTranslation.getAxioms());
		
	   
	} else if (featureName.equals("oclAsType")){
	    KeyTranslationResult objectTranslation, typeTranslation;
	    
	    objectTranslation = (KeyTranslationResult) getTranslationForNode(appliedTo);
	    Term objectTerm = objectTranslation.getTerm();
	    Sort objectSort = objectTerm.sort();
	    

	    typeTranslation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    Term oclTypeTerm = typeTranslation.getTerm();
	    Function oclTypeConstant = (Function) oclTypeTerm.op();

	    String oclTypeName =   oclTypeConstant.name().toString();

	    int index = oclTypeName.indexOf("::");
	    if (index != -1){
		oclTypeName = oclTypeName.substring(index + 2);
	    }
	   
	   
	    Sort sortForGivenOclType =  getDLSortForNonCollectionType(oclTypeName); 
	      
	    
	    String featureFunctionName = AS_TYPE + "-" + objectSort.toString()  + "-" + sortForGivenOclType.toString();
	    
	    // Now get the corresponding function symbol ...
	    Function featureFunction = proverFacade.getFunctionForFeature(featureFunctionName,
									    new Sort[]{objectSort},
									    sortForGivenOclType);

	    resultTerm = tf.createFunctionTerm(featureFunction,objectTerm);
	    
	    log("INTO :" + resultTerm.toString());
	    
	    // We have to describe the new symbol by an additional axiom ...
	    Term newAxiom;

	    LogicVariable predicateVar = newVariableDeliverer.getNewVariableForSort(sortForGivenOclType);
	    Term predicateVarTerm = tf.createVariableTerm(predicateVar);
	    Term funcTerm = tf.createFunctionTerm(featureFunction, predicateVarTerm);
	    Term eqTerm = tf.createEqualityTerm(funcTerm, predicateVarTerm);
	    
	    newAxiom = getQuantifiedTermWith(Op.ALL, predicateVar, eqTerm);

	    log("Adding additional axiom for '"+featureFunction +"' :" + newAxiom.toString());
	    
	   
	    resultTranslation = new KeyTranslationResult();
	    resultTranslation.setTerm(resultTerm);
	    resultTranslation.addAxioms(typeTranslation.getAxioms());
	    resultTranslation.addAxioms(objectTranslation.getAxioms());
	    resultTranslation.addAxiom(newAxiom);
	   
	} else if (featureName.equals("oclIsNew")){
	
	    KeyTranslationResult objectTranslation;
	    
	    objectTranslation = (KeyTranslationResult) getTranslationForNode(appliedTo);
	    Term objectTerm = objectTranslation.getTerm();
	    Sort objectSort = objectTerm.sort();
	    
	    Sort boolSort = Sort.FORMULA; // proverFacade.getSortForOCLBoolean();
 
	    String featureFunctionName1 = CREATED;
	    String featureFunctionName2 = CREATED + "@pre";
	    
	    
	    // Now get the corresponding function symbols ...
	    Function featureFunction1 = proverFacade.getFunctionForFeature(featureFunctionName1,
									    new Sort[]{objectSort},
									    boolSort);

	    Function featureFunction2 = proverFacade.getFunctionForFeature(featureFunctionName2,
									     new Sort[]{objectSort},
									     boolSort);

	    // This is a nonrigid function symbol therefore store it ...
	    occuringNonrigidFunctionSymbols.add(featureFunction1);
	    occuringNonrigidFunctionSymbols.add(featureFunction2);
	    

	    Term createdTerm1 = tf.createFunctionTerm(featureFunction1,objectTerm);
	    Term createdTerm2 = tf.createFunctionTerm(featureFunction2,objectTerm);
	    Term notTerm = tf.createJunctorTerm(Op.NOT, createdTerm2);

	    resultTerm = tf.createJunctorTermAndSimplify(Op.AND,createdTerm1,notTerm);
	    


	    log("INTO :" + resultTerm.toString());
	    
	    
	    resultTranslation = new KeyTranslationResult();
	    resultTranslation.setTerm(resultTerm);
	    resultTranslation.addAxioms(objectTranslation.getAxioms());
	    
	   
	} else {
	    // We have a feature call for a feature in the UML model (attribute, methodcall or association) ...
	    
	    String featureFunctionName = featureName;
	    if(featureCallWithAtPre){
		featureFunctionName += "@pre";
	    }
	    
	    KeyTranslationResult appliedToTranslation, currentParameterTranslation;
	    KeyTranslationResult[] parameterTranslations;
	    Node currentParameterNode;
	    
	    Sort appliedToSort;
	    Sort[] argSorts;
	    Term[] argTerms;
	    
	    appliedToTranslation = (KeyTranslationResult) getTranslationForNode(appliedTo);
	    appliedToSort = appliedToTranslation.getTerm().sort();
	    
	    parameterTranslations = new KeyTranslationResult[parameterNodes.size()];
	    argSorts = new Sort[parameterNodes.size() + 1];
	    argTerms = new Term[parameterNodes.size() + 1];
	    
	    argSorts[0] = appliedToSort;
	    argTerms[0] = appliedToTranslation.getTerm();
	    
	    for(int i=0;i<parameterNodes.size();i++){
		currentParameterNode = (Node) parameterNodes.elementAt(i);
		currentParameterTranslation = (KeyTranslationResult) getTranslationForNode(currentParameterNode);
		parameterTranslations[i] = currentParameterTranslation;
		argSorts[i+1] = currentParameterTranslation.getTerm().sort();
		argTerms[i+1] = currentParameterTranslation.getTerm(); 
	    }
	    
	    
	    // Get the corresponding sort for the return type;
	    Sort returnSort;
	    if ((returnType == Basic.BOOLEAN) && 
		(!translationType.containsMark(TranslationMark.AS_BOOLEAN_TERM)))  {
		returnSort = Sort.FORMULA;
	    } else {
		returnSort = getDLSortForType(returnType);
	    }
	    
	    if ((!translationType.containsMark(TranslationMark.AS_NONRIGID_FUNCTION_SYMBOL))) {
		// Use a representation using attribute and method-call operators (iff possible) ...
		//		System.err.println(returnType.getClass()+"  "+returnType);
		if (!featureCallWithAtPre && !(returnType instanceof Collection)) {
		    // We have a simple attribute or method call ...
		    
		    if (isMethodCall) {
			// We have to handle some method call...
			
			// Now get the corresponding function symbol ...

			TermSymbol featureFunction 
			    = proverFacade.getQueryForFeature
			    (featureFunctionName, argTerms);
			// This is a nonrigid function symbol therefore store it ...
			occuringNonrigidFunctionSymbols.add(featureFunction);

			if ((featureFunction instanceof ProgramMethod) &&
			    ((ProgramMethod)featureFunction).isStatic()) {
			    Term[] argTerms0 = new Term[argTerms.length-1];
			    System.arraycopy(argTerms, 1, argTerms0, 
					     0, argTerms0.length);
			    argTerms=argTerms0;
			}
			
			resultTerm = tf.createFunctionTerm(featureFunction,argTerms);
			
			log("INTO :" + resultTerm.toString());
			
			
			resultTranslation = new KeyTranslationResult(resultTerm);
			
			// Now take the union of the axioms of all subtranslations ...
			resultTranslation.addAxioms(appliedToTranslation.getAxioms());
			for(int j=0;j<parameterTranslations.length;j++){
			    resultTranslation.addAxioms(parameterTranslations[j].getAxioms()); 
			}
			
		    } else {
			// We have to handle some attribute call ...
			
			
			// We don't use an unary nonrigid function symbol here anymore.
			// Instead we really use a attribute  in the corresponding DL class.
			
			String appliedToClassWithName = appliedToSort.name().toString();
			KeYJavaType classReference =  getKeYJavaTypeForNonCollectionType(appliedToClassWithName); 
			
			// Now get the corresponding attribute operator ...
			ProgramVariable attributeOp = proverFacade.getAttribute(featureFunctionName, classReference);
			
			Term objTerm = appliedToTranslation.getTerm();
			resultTerm = proverFacade.getAttributeTerm(attributeOp, objTerm);
			
			log("INTO :" + resultTerm.toString());
			
			
			resultTranslation = new KeyTranslationResult(resultTerm);
			
			// Now take the union of the axioms of all subtranslations ...
			resultTranslation.addAxioms(appliedToTranslation.getAxioms());
			
			
		    }
		    
		} else if (featureCallWithAtPre) {
		    // We have some feature that is tagged with '@pre' therefore
		    // create a nonrigid-function symbol ...
		    
		    // Now get the corresponding function symbol ...
		    if (isMethodCall) { //because of equal names of possibly
			                // attributes and methods
			featureFunctionName=featureFunctionName+"_";
		    }
		    Function featureFunction = proverFacade.getFunctionForFeature(featureFunctionName,
										    argSorts,
										    returnSort);
		    // This is a nonrigid function symbol therefore store it ...
		    occuringNonrigidFunctionSymbols.add(featureFunction);
		    
		    resultTerm = tf.createFunctionTerm(featureFunction,argTerms);
		    
		    log("INTO :" + resultTerm.toString());
		    
		    
		    resultTranslation = new KeyTranslationResult(resultTerm);
		    
		    // Now take the union of the axioms of all subtranslations ...
		    resultTranslation.addAxioms(appliedToTranslation.getAxioms());
		    for(int j=0;j<parameterTranslations.length;j++){
			resultTranslation.addAxioms(parameterTranslations[j].getAxioms()); 
		    }
		    
		    
		    if (!isMethodCall && !(returnType instanceof Collection)){
			// Mark the corresponding nonrigid function symbol as
			// corresponding to some attribute on a model type ...
			
			
			String appliedToClassWithName = appliedToSort.name().toString();
			KeYJavaType classReference =  getKeYJavaTypeForNonCollectionType(appliedToClassWithName); 
			
			// Now get the corresponding attribute operator ...
			ProgramVariable progVar = proverFacade.getAttribute(featureName, classReference);
			
			Term objTerm = appliedToTranslation.getTerm();
			Term t = proverFacade.getAttributeTerm(progVar, objTerm);
			
			ProgramVariable attributeOp = (t.op() instanceof ProgramVariable ? 
						       (ProgramVariable) t.op() : 
						       (ProgramVariable) ((AttributeOp)t.op()).attribute());
			
			occuringAttributeAtPreFunctions.put(featureFunction, attributeOp);
		    } else {			
			TermSymbol methodOp = proverFacade.getQueryForFeature(featureName, argTerms);

			occuringMethodAtPreFunctions.put(featureFunction, methodOp);
		    }
		    
		} else { //collection not @pre
		    Function featureFunction = proverFacade.getFunctionForFeature(featureFunctionName,
										    argSorts,
										    returnSort);
		    // This is a nonrigid function symbol therefore store it ...
		    occuringNonrigidFunctionSymbols.add(featureFunction);
		    
		    resultTerm = tf.createFunctionTerm(featureFunction,argTerms);
		    
		    log("INTO :" + resultTerm.toString());
		    
		    
		    resultTranslation = new KeyTranslationResult(resultTerm);
		    
		    // Now take the union of the axioms of all subtranslations ...
		    resultTranslation.addAxioms(appliedToTranslation.getAxioms());
		    for(int j=0;j<parameterTranslations.length;j++){
			resultTranslation.addAxioms(parameterTranslations[j].getAxioms()); 
		    }
		    
		}
	    } else {
		// Use a representation using non-rigid function symbols ...
		
		// create a nonrigid-function symbol ...
		TermSymbol featureFunction;

		// %%% this is fuzzy: actually there should only be the latter call
		// to getFunctionForFeature. But there were situations where
		// a query should be the result. However, it's certainly not the
		// right place to fix... /AR
		try {
		    featureFunction = proverFacade.getQueryForFeature
			(featureFunctionName, argTerms);	      
		} catch (IllegalArgumentException e) {
		    featureFunction=null;
		}		
		if (featureFunction==null) {
		    featureFunction 
			= proverFacade.getFunctionForFeature(featureFunctionName, 
							       argSorts, returnSort);
		}
		// This is a nonrigid function symbol therefore store it ...
		occuringNonrigidFunctionSymbols.add(featureFunction);
		
		resultTerm = tf.createFunctionTerm(featureFunction,argTerms);
		
		log("INTO :" + resultTerm.toString());
		
		
		resultTranslation = new KeyTranslationResult(resultTerm);
		
		// Now take the union of the axioms of all subtranslations ...
		resultTranslation.addAxioms(appliedToTranslation.getAxioms());
		for(int j=0;j<parameterTranslations.length;j++){
		    resultTranslation.addAxioms(parameterTranslations[j].getAxioms()); 
		}
		
	    }
	    
	}
	
	// Finally set the new translation result ...
	setTranslationForNode(n,resultTranslation);
	
    }


   // =========================================================================================================
   // ================================= The feature calls on OCLInteger ==== ==================================
   // =========================================================================================================


    protected void translateFeatureCallonInteger(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow ){
	log("Feature call on Integer : " + featureName);
	log("Applied on node :" + appliedTo.toString());

	KeyTranslationResult result = null;
	Term resultTerm = null;
	
	if(featureName.equals("mod") || featureName.equals("div") ){
	    KeyTranslationResult int1Translation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult int2Translation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Term int1Term = int1Translation.getTerm();
 	    Term int2Term = int2Translation.getTerm();

	    // ==================== TRANSLATION AS TERM ==============================
	    Sort intSort = proverFacade.getSortForOCLInteger();
	    
	    String funcName = featureName.equals("mod") ? INT_MOD : INT_DIV ;

	    Function intFeatureFunction =
		proverFacade.getFunctionForFeature(funcName,
						     new Sort[]{intSort,intSort},
						     intSort);
	    
	    resultTerm = tf.createFunctionTerm(intFeatureFunction,int1Term,int2Term);
		

	    

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(int1Translation.getAxioms());
	    result.addAxioms(int2Translation.getAxioms());
	    
	} else if(featureName.equals("max") || featureName.equals("min") ){
	    // This could be handeled 'directly' with a new constant/function symbol
	    // and additional axioms ...
	    KeyTranslationResult int1Translation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult int2Translation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Term int1Term = int1Translation.getTerm();
 	    Term int2Term = int2Translation.getTerm();

	    // ==================== TRANSLATION AS TERM ==============================
	    Sort intSort = proverFacade.getSortForOCLInteger();
	     
	    String funcName = featureName.equals("max") ? INT_MAX : INT_MIN ;

	    Function intFeatureFunction =
		proverFacade.getFunctionForFeature(funcName,
						     new Sort[]{intSort,intSort},
						     intSort);
	    
	    resultTerm = tf.createFunctionTerm(intFeatureFunction,int1Term,int2Term);
		

	    

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(int1Translation.getAxioms());
	    result.addAxioms(int2Translation.getAxioms());
	    
	} else if(featureName.equals("abs")) {
	    // This could be handeled 'directly' with a new constant/function symbol
	    // and additional axioms ...
	    KeyTranslationResult int1Translation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	   
	    Term int1Term = int1Translation.getTerm();
 	    
	    // ==================== TRANSLATION AS TERM ==============================
	    Sort intSort = proverFacade.getSortForOCLInteger();
	    
	    Function intFeatureFunction =
		proverFacade.getFunctionForFeature(INT_ABS,
						     new Sort[]{intSort},
						     intSort);
	    
	    resultTerm = tf.createFunctionTerm(intFeatureFunction,int1Term);
		

	    

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(int1Translation.getAxioms());
	   
	    
	} else {
	    error("Feature '" + featureName + "' on OCLInteger not supported yet !!!");
	}

	// Finally set the new translation result ...
	setTranslationForNode(n,result);
	
    }


   // =========================================================================================================
   // ================================= The feature calls on OCLReal == =======================================
   // =========================================================================================================

    

    protected void translateFeatureCallonReal(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow ){
	log("Feature call on Real : " + featureName);
	log("Applied on node :" + appliedTo.toString());

	
	KeyTranslationResult result = null;
	Term resultTerm = null;
	
	if(featureName.equals("max") || featureName.equals("min") ){
	    // This could be handeled 'directly' with a new constant/function symbol
	    // and additional axioms ...
	    KeyTranslationResult real1Translation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult real2Translation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Term real1Term = real1Translation.getTerm();
 	    Term real2Term = real2Translation.getTerm();

	    // ==================== TRANSLATION AS TERM ==============================
	    Sort realSort = proverFacade.getSortForOCLReal();
	    
	    String funcName = featureName.equals("max") ? REAL_MAX : REAL_MIN ;

	    Function realFeatureFunction =
		proverFacade.getFunctionForFeature(funcName,
						     new Sort[]{realSort,realSort},
						     realSort);
	    
	    resultTerm = tf.createFunctionTerm(realFeatureFunction,real1Term,real2Term);
		

	    

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(real1Translation.getAxioms());
	    result.addAxioms(real2Translation.getAxioms());
	    
	} else if(featureName.equals("abs")) {
	    // This could be handeled 'directly' with a new constant/function symbol
	    // and additional axioms ...
	    KeyTranslationResult real1Translation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	   
	    Term real1Term = real1Translation.getTerm();
 	    
	    // ==================== TRANSLATION AS TERM ==============================
	    Sort realSort = proverFacade.getSortForOCLReal();
	    
	    Function realFeatureFunction =
		proverFacade.getFunctionForFeature(REAL_ABS,
						     new Sort[]{realSort},
						     realSort);
	    
	    resultTerm = tf.createFunctionTerm(realFeatureFunction,real1Term);
		

	    

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(real1Translation.getAxioms());
	   
	    
	} else if(featureName.equals("floor") || featureName.equals("round") ){
	    // This could be handeled 'directly' with a new constant/function symbol
	    // and additional axioms ...
	    KeyTranslationResult real1Translation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	   
	    Term real1Term = real1Translation.getTerm();
 	    
	    // ==================== TRANSLATION AS TERM ==============================
	    Sort realSort = proverFacade.getSortForOCLReal();
	    Sort intSort =  proverFacade.getSortForOCLInteger();
	    
	     
	    String funcName = featureName.equals("floor") ? REAL_FLOOR : REAL_ROUND;

	    Function realFeatureFunction =
		proverFacade.getFunctionForFeature(featureName,
						     new Sort[]{realSort},
						     intSort);
	    
	    resultTerm = tf.createFunctionTerm(realFeatureFunction,real1Term);
		
	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(real1Translation.getAxioms());
	   
	        
	} else {
	    error("Feature '" + featureName + "' on OCLReal not supported yet !!!");
	}

	// Finally set the new translation result ...
	setTranslationForNode(n,result);
	
    }

    
   // =========================================================================================================
   // ================================= The feature calls on OCLString ========================================
   // =========================================================================================================

    protected void translateFeatureCallonString(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow ){
	log("Feature call on String : " + featureName);
	log("Applied on node :" + appliedTo.toString());

	KeyTranslationResult result = null;
	Term resultTerm = null;
	
	if(featureName.equals("concat")){
	    // This could be handeled 'directly' with a new constant/function symbol
	    // and additional axioms ...
	    KeyTranslationResult string1Translation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	    KeyTranslationResult string2Translation =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    
	    Term string1Term = string1Translation.getTerm();
 	    Term string2Term = string2Translation.getTerm();

	    // ==================== TRANSLATION AS TERM ==============================
	    Sort stringSort = proverFacade.getSortForOCLString();
	    
	    Function stringFeatureFunction =
		proverFacade.getFunctionForFeature(STRING_CONCAT,
						     new Sort[]{stringSort,stringSort},
						     stringSort);
	    
	    resultTerm = tf.createFunctionTerm(stringFeatureFunction,string1Term,string2Term);
		

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(string1Translation.getAxioms());
	    result.addAxioms(string2Translation.getAxioms());
	    
	} else if(featureName.equals("toUpper") || featureName.equals("toLower")) {
	    
	    KeyTranslationResult stringTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	   
	    Term stringTerm = stringTranslation.getTerm();
 	    
	    // ==================== TRANSLATION AS TERM ==============================
	    Sort stringSort = proverFacade.getSortForOCLString();
	    
	     
	    String funcName = featureName.equals("toUpper") ? STRING_TO_UPPER : STRING_TO_LOWER ;


	    Function stringFeatureFunction =
		proverFacade.getFunctionForFeature(funcName,
						     new Sort[]{stringSort},
						     stringSort);
	    
	    resultTerm = tf.createFunctionTerm(stringFeatureFunction,stringTerm);
		
	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(stringTranslation.getAxioms());
	   
	    
	} else if(featureName.equals("size")){
	   
	    KeyTranslationResult stringTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	   
	    Term stringTerm = stringTranslation.getTerm();
 	    
	    // ==================== TRANSLATION AS TERM ==============================
	    Sort stringSort = proverFacade.getSortForOCLString();
	    Sort intSort =  proverFacade.getSortForOCLInteger();
	    
	    Function stringFeatureFunction =
		proverFacade.getFunctionForFeature(SIZE,
						     new Sort[]{stringSort},
						     intSort);
	    
	    resultTerm = tf.createFunctionTerm(stringFeatureFunction,stringTerm);
	       

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(stringTranslation.getAxioms());
	   
	        
	} else if(featureName.equals("substring")){
	   
	    KeyTranslationResult stringTranslation = 
		(KeyTranslationResult) getTranslationForNode(appliedTo);
	   
	    KeyTranslationResult indexTranslation1 =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(0));
	    KeyTranslationResult indexTranslation2 =
		(KeyTranslationResult) getTranslationForNode((Node) parameterNodes.elementAt(1));
	    
	   
 	    Term indexTerm1 = indexTranslation1.getTerm();
	    Term indexTerm2 = indexTranslation2.getTerm();	    

	    Term stringTerm = stringTranslation.getTerm();
 	    
	    // ==================== TRANSLATION AS TERM ==============================
	    Sort stringSort = proverFacade.getSortForOCLString();
	    Sort intSort =  proverFacade.getSortForOCLInteger();
	    Sort retSort = stringSort;
	   
	    Function stringFeatureFunction =
 		proverFacade.getFunctionForFeature(STRING_SUBSTRING,
 						     new Sort[]{stringSort,intSort, intSort},
 						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(stringFeatureFunction,
					       new Term[]{stringTerm, indexTerm1, indexTerm2});

	    log("INTO :" + resultTerm.toString());

	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(stringTranslation.getAxioms());
	    result.addAxioms(indexTranslation1.getAxioms());
	    result.addAxioms(indexTranslation2.getAxioms());
	   
	} else {
	    error("Feature '" + featureName + "' on OCLString not supported yet !!!");
	}

	// Finally set the new translation result ...
	setTranslationForNode(n,result);
	
    }



    
   // =========================================================================================================
   // ================================= The feature calls on OCLType ==========================================
   // =========================================================================================================

    protected void translateFeatureCallonOclType(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow,  boolean isMethodCall){
	log("Feature call on OclType : " + featureName);
	log("Applied on node :" + appliedTo.toString());

	// Get the kind of translation to be performed for the given node ...
	TranslationType translationType = getTranslationTypeForNode(n);
	log("Current feature call should be translated in the following way : " + translationType.toString());
       
	KeyTranslationResult result = null;
	Term resultTerm = null;
	
	Type resultType = tree.getNodeType(n);
	OclType oclType = (OclType) tree.getNodeType(appliedTo);
	Type typeObjectToCurrentOCLType = oclType.getType();

	log("Result type is :" + resultType);
	log("Current type object to OCLType is :" + typeObjectToCurrentOCLType);
	
	boolean oclTypeIsUMLModelType = (!typeObjectToCurrentOCLType.equals(Basic.INTEGER) &&
					 !typeObjectToCurrentOCLType.equals(Basic.REAL) &&
					 !typeObjectToCurrentOCLType.equals(Basic.BOOLEAN) &&
					 !typeObjectToCurrentOCLType.equals(Basic.STRING) &&
					 !typeObjectToCurrentOCLType.equals(Basic.ENUMERATION) &&
					 !(typeObjectToCurrentOCLType instanceof Collection) &&
					 !(typeObjectToCurrentOCLType instanceof OclType) &&
					 !(typeObjectToCurrentOCLType instanceof OclState) );
	
	log("Current type object is UML modeltype :" + oclTypeIsUMLModelType);

	boolean featureIsStaticFeature = ( !featureName.equals("allInstances") && 
					   !featureName.equals("name") && 
					   !featureName.equals("attributes") &&
					   !featureName.equals("associationEnds") &&
					   !featureName.equals("operations") &&
					   !featureName.equals("supertypes") &&
					   !featureName.equals("allSupertypes") );

	log("Current feature call on type object is a static call :" + featureIsStaticFeature);

	if(featureName.equals("allInstances")){
	    
	    if(oclTypeIsUMLModelType){
		// We currently handle 'allInstances' only for UML model types (classes) ...
		
		
		if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
		    // ==================== TRANSLATION AS TERM ==============================
		    
		    // Create a constant of type Set[T];

		    // String allInstancesConstantName = typeObjectToCurrentOCLType.toString() + ".allInstances";
		    String allInstancesConstantName = ALL_INSTANCES + typeObjectToCurrentOCLType.toString();

		    // allInstances is a nonrigid feature in the DL (it can change from state to state)
		    // therefore we have to handle '@pre':
		    if(featureCallWithAtPre){
			allInstancesConstantName += "@pre";
		    }
		    
		    log("Get a constant with name (for allInstances feature) '" + allInstancesConstantName + 
			"' of type " + resultType);
		    
		    Sort retSort = getDLSortForType(resultType);
		    
		    Function allInstancesConstant =
			proverFacade.getFunctionForFeature(allInstancesConstantName,
							     new Sort[]{},
							     retSort);
		    resultTerm = tf.createFunctionTerm(allInstancesConstant);
		     
		    log("INTO :" + resultTerm.toString());

		    result = new KeyTranslationResult();
		    result.setTerm(resultTerm);

		    Term newAxiom;

		    // When we have an @pre-occurence we have to be careful !
		    // We may not create an defining axiom in the Post-Part,
		    // instead is has to be placed in the Pre-Part, since
		    // otherwise in this special case we would have an 
		    // incorrect description of allInstancesOfT@pre !!!!
		    // Also the axioms has to describe the non-@pre constant.
		    if(!featureCallWithAtPre){
			
			// allInstances is a nonrigid feature in the DL (it can change from state to state)
			// This is a nonrigid function symbol therefore store it ...
			boolean alreadUsed = occuringNonrigidFunctionSymbols.contains(allInstancesConstant);
			if (!alreadUsed) {
			    // Create a corresponding axiom and store the symbol (mark as used) ...
			    occuringNonrigidFunctionSymbols.add(allInstancesConstant);
			    
			    
			    // Create an additional axiom constraining the interpretation of the new constant ...
			    
			    
			    Sort elemSort = getElementSortForCollectionType(resultType);
			    Sort collSort = retSort;
			    
			    LogicVariable predicateVar = newVariableDeliverer.getNewVariableForSort(elemSort);
			    Term predicateVarTerm = tf.createVariableTerm(predicateVar);
			    
			    
			    Function setContainsFunction = 
				proverFacade.getCollectionFunctionForFeature(CONTAINS,
									       new Sort[]{collSort, elemSort},
									       Sort.FORMULA);
			    
			    
			    
			    Term containsTerm = tf.createFunctionTerm(setContainsFunction,resultTerm,predicateVarTerm);
			    newAxiom = getQuantifiedTermWith(Op.ALL, predicateVar , containsTerm);
			    
			    log("Additional axiom for allInstances constant:" + newAxiom.toString());
			    
			    result.addAxiom(newAxiom);
			    
			} else {
			    log("FunctionSymbol '"+ allInstancesConstant + 
				" has already been used, therefore skip\n creating the same axiom !");
			}
		       
		    } else {
			log("Axiomatization of functionSymbol '"+ allInstancesConstant + 
			    " has to be done in the pre-Part, therefore we do it later !");
			
		    }

		
		} else {
		    // ==================== TRANSLATION AS PREDICATE ==============================
		    
		    Sort elemSort = getElementSortForCollectionType(resultType);
		   
		    LogicVariable predicateVar = newVariableDeliverer.getNewVariableForSort(elemSort);
		    
		    // When we have an @pre-occurence we have to be careful, to get a correct characterization,
		    // since functions in the Post-part are evaluated in the Post-State (not in the Pre-State).
		    
		    if(!featureCallWithAtPre){
			// Note, that in the Post-part 'allInstancesOfT' may be charcterized by :
			// all' t:T ( contains(allInstancesOfT@pre,t) <-> true).
			resultTerm = tf.createJunctorTerm(Op.TRUE);

		    } else {
			// Note, that in the Post-part 'allInstancesOfT@pre' may be charcterized by :
			// all' t:T ( contains(allInstancesOfT@pre,t) <-> created@pre(t)).
			// Therfore, we use the predicate: created@pre(t).
			    Sort boolSort = proverFacade.getSortForOCLBoolean();
 
			    String functionName = CREATED + "@pre";
	    
	    
			    // Now get the corresponding function symbols ...
			    Function createdFunction = proverFacade.getFunctionForFeature(functionName,
											    new Sort[]{elemSort},
											    boolSort);

			    // This is a nonrigid function symbol therefore store it ...
			    occuringNonrigidFunctionSymbols.add(createdFunction);
	    
			    Term predVarTerm = tf.createVariableTerm(predicateVar);
			    Term createdTerm = tf.createFunctionTerm(createdFunction,predVarTerm);
			    
			    resultTerm = createdTerm;
	    
		    }

		    log("INTO :" + resultTerm.toString());
		
		    result = new KeyTranslationResult();
		    result.setTerm(resultTerm);	    
		    result.setPredicateVariable(predicateVar);
		    result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);

		}
	    

	    } else {
		error("Feature '" + featureName + "' on OCLType ("+ oclType  +") not supported yet !!!");
	    }

	} else if(oclTypeIsUMLModelType && featureIsStaticFeature){
	    // Handle a static method or attribute ...
	     log("Translating static feature of UML model type : '" + featureName + "' ...");
	
	     // The following is very similar to the translation of non-static features of UML-Modeltypes ...

	     // get the return type of the feature call ...
	     Type returnType = tree.getNodeType(n);
	     
	     log("Return type = " + returnType.toString());

	     String featureFunctionName = featureName;
	     if(featureCallWithAtPre){
		 featureFunctionName += "@pre";
	     }
	
	     KeyTranslationResult currentParameterTranslation;
	     KeyTranslationResult[] parameterTranslations;
	     Node currentParameterNode;

	     Sort[] argSorts;
	     Term[] argTerms;
	
	     parameterTranslations = new KeyTranslationResult[parameterNodes.size()];
	     argSorts = new Sort[parameterNodes.size()];
	     argTerms = new Term[parameterNodes.size()];
	
	
	     for(int i=0;i<parameterNodes.size();i++){
		 currentParameterNode = (Node) parameterNodes.elementAt(i);
		 currentParameterTranslation = (KeyTranslationResult) getTranslationForNode(currentParameterNode);
		 parameterTranslations[i] = currentParameterTranslation;
		 argSorts[i] = currentParameterTranslation.getTerm().sort();
		 argTerms[i] = currentParameterTranslation.getTerm(); 
	     }
	
	
	     // Get the corresponding sort for the return type;
	     Sort returnSort = getDLSortForType(returnType);
	

	     // Now get the corresponding function symbol ...
	     Function featureFunction = proverFacade.getFunctionForFeature(featureFunctionName,
									     argSorts,
									     returnSort);
	     // This is a nonrigid function symbol therefore store it ...
	     occuringNonrigidFunctionSymbols.add(featureFunction);

	     resultTerm = tf.createFunctionTerm(featureFunction,argTerms);
	     
	     log("INTO :" + resultTerm.toString());
	

	     result = new KeyTranslationResult(resultTerm);

	     // Now take the union of the axioms of all subtranslations ...
	     for(int j=0;j<parameterTranslations.length;j++){
		 result.addAxioms(parameterTranslations[j].getAxioms()); 
	     }
	   
	} else {
	    error("Feature '" + featureName + "' on OCLType ("+ oclType  +") not supported yet !!!");
	}

	// Finally set the new translation result ...
	setTranslationForNode(n,result);
	

    }
    




    protected void translateVariable(Node n, String varName , Type varType){
	String varTypeString = varType.toString();

	log("Translating variable '"+ varName +"' of type '"+ varTypeString +"'.");

	KeyTranslationResult varTranslation = null;
	try {
	    // check whether there is already a translation for the given variable and
	    // use an already existing translation ...
	    varTranslation = 
		(KeyTranslationResult) getTranslationForVariable(varName,varTypeString);
	} catch(RuntimeException rte){
	    // There is no already defined translation for that variable ...
	    ParsableVariable lv;
	    Term varTerm;
	    Sort varSort;
	    
	    KeYJavaType kjt = getKeYJavaTypeForNonCollectionType(varType.toString());
	    lv = proverFacade.getProgramVariable(varName,kjt);	    
	    if (lv==null) {
		varSort = getDLSortForType(varType);
		lv = new LogicVariable(new Name(varName),varSort);	    
		varTerm = tf.createVariableTerm((LogicVariable)lv);
	    } else {
		varTerm = tf.createVariableTerm((ProgramVariable)lv);
	    }
	    varTranslation = new KeyTranslationResult(varTerm);
	    
	    log("Created the first translation for variable '"+ varName +"' of type '"+ varTypeString +"'");
	    setTranslationForVariable(varName,varTypeString, varTranslation);
	    log("Saved translation for variable '"+ varName +"' of type '"+ varTypeString +"'");
	    
	}
	
	log("INTO : " + varTranslation.toString());
	setTranslationForNode(n,varTranslation);
	
    }

    protected void translateVariableDeclaration(String varName, Type varType){
    	log("Translating Variable declarationition for Var: '" + varName + "'.");

	String varTypeString = varType.toString();

	LogicVariable lv;
	Function pf;
	Term varTerm;
	Sort varSort;
	
	varSort = getDLSortForType(varType); 
	
	// Create a free variable of that type ...
	lv = new LogicVariable(new Name(varName),varSort);
	varTerm = tf.createVariableTerm(lv);
	
	KeyTranslationResult varTranslation = new KeyTranslationResult(varTerm);

	setTranslationForVariable(varName,varTypeString, varTranslation);

    } 

    protected void removeVariableDeclaration(String varName, Type varType){
    	log("Removing Variable declaration for Var: '" + varName + "'.");
	
	String varTypeString = varType.toString();
	removeTranslationForVariable(varName,varTypeString);
    } 
    

    

    protected void translateLetVariableDeclaration(Node currentLetExpression, String letVarName, Type letVarType){
    	log("Translating LET-Variable definition for let-Var: '" + letVarName + "'.");

	// At this point the expression which defines the value of the declared
	// variable has been already translated ...
	ALetExpression letExpression = (ALetExpression) currentLetExpression;
	AExpression valueExpression = (AExpression) letExpression.getExpression();
	KeyTranslationResult valueTranslation =
	    (KeyTranslationResult)getTranslationForNode(valueExpression);
	
	// Get the kind of translation that was used for the defined expression ...
	TranslationType translationType = getTranslationTypeForNode(valueExpression);
	log("Current let: defined expression has been translated in the following way : " + 
	    translationType.toString() + "\n and is : " + valueExpression);
	
	Term valueTerm = valueTranslation.getTerm();
	Sort retSort;

	if (letVarType == Basic.BOOLEAN) {
	    if (!translationType.containsMark(TranslationMark.AS_BOOLEAN_TERM)) {
		retSort = Sort.FORMULA;
	    } else {
		retSort = getDLSortForType(letVarType);
	    }
	} else {
	    retSort = getDLSortForType(letVarType);
	}

	KeyTranslationResult letVarTranslation = new KeyTranslationResult();
	Term resultTerm;

	if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
	    // ==================== TRANSLATION AS TERM ==============================
	    // Now compute the necessary parameters ...
	    
	    SetOfQuantifiableVariable paramsForNewFunction = valueTerm.freeVars();
		
	    Term[] paramTerms;
	    Sort[] paramSorts;		
	    
	    int freeVarsCnt = paramsForNewFunction.size();
	    
	    paramTerms = new Term[freeVarsCnt];
	    paramSorts = new Sort[freeVarsCnt];
	    
	    // The rest are the necessary parameters from the other expressions ...
	    IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
	    LogicVariable currentVar;
	    int i=0;
	    
	    while (varIt.hasNext()){
		currentVar = (LogicVariable) varIt.next();
		paramTerms[i] = tf.createVariableTerm(currentVar);
		paramSorts[i] = currentVar.sort();
		i++;
	    }
	    

	    String functionName = newFunctionNameDeliverer.getNewFunctionNameForName("let");
	    
	    Function letFunction =
		proverFacade.getFunctionForFeature(functionName,
						     paramSorts,
						     retSort);
	    
	    resultTerm = tf.createFunctionTerm(letFunction,paramTerms);	
	    log("INTO :" + resultTerm.toString());
	    
	    // Create an additional axiom for the interpretation of the new function symbol ...
	    Term newAxiom;
		
	    Term eqTerm = tf.createEqualityTerm(resultTerm, valueTerm);
	    newAxiom = constraintTranslator.createAllClosureOfTerm(eqTerm);
	    
	    log("Adding new axiom for " + functionName + " (in valueTranslation) : " + newAxiom);
		
	    // Now create the translation result for the current letVariable.
	    // The corresponding term will be syntactically replaced for the 
	    // variable.
	    letVarTranslation.setTerm(resultTerm);
	    letVarTranslation.addAxiom(newAxiom);
	    

	} else {
	    // ==================== TRANSLATION AS PREDICATE ==============================
	    
	    // Just use the generated predicate directly (or if a term has actually been used
	    // convert the translation result to a predicative one!) ...
	      
	    valueTranslation = getPredicativeTranslationOfTranslation(valueTranslation, (Collection) letVarType);
	    resultTerm = valueTranslation.getTerm();
	    

	    // Now create the translation result for the current letVariable.
	    // The corresponding term will be syntactically replaced for the 
	    // variable.
	    letVarTranslation.setTerm(resultTerm);
	    // Set the same translation type as for the defined expression ...
	    letVarTranslation.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	    letVarTranslation.setPredicateVariable(valueTranslation.getPredicateVariable());
	    
	}		
	
	
	// Now we set the translation for the new let variable ...
	setTranslationForVariable(letVarName, letVarType.toString().trim(),letVarTranslation);
	log("Will replace LET-Variable syntactically as: " + letVarTranslation.getTerm() );




    } 


    protected void removeLetVariableDeclaration(String letVarName, Type letVarType){
    	log("Removing LET-Variable definition for Var: '" + letVarName + "'.");
	// Now we have to remove the local variable definition (and it's corresponding
	// translation) ...
	removeTranslationForVariable(letVarName, letVarType.toString().trim());
    }

    protected void translateLetExpression(Node le, String letVarName,Type letVarType, Node e, Node inExpression){
	log("Translating LET-construct with Variable '" + letVarName + "'.");

	KeyTranslationResult result = new KeyTranslationResult();
	KeyTranslationResult inTranslation = 
	    (KeyTranslationResult) getTranslationForNode(inExpression);

	AExpression valueExpression = (AExpression) (((ALetExpression) le).getExpression());
	KeyTranslationResult valueTranslation =
	    (KeyTranslationResult)getTranslationForNode(valueExpression);
	

	result.setTerm(inTranslation.getTerm());
 	result.addAxioms(inTranslation.getAxioms());
 	result.addAxioms(valueTranslation.getAxioms());

	
	setTranslationForNode(le,result);
	    
	
    }

    
   // =========================================================================================================
   // ================================= The a literal of OCL type Integer =====================================
   // =========================================================================================================


    protected void translateIntegerLiteral(Node sl, String lit){
	log("Translating Integer literal: " + lit);
	
	KeyTranslationResult result;
	Term resultTerm;
	
	resultTerm = computeTranslationForIntegerLiteral(lit);
	
	
	log("INTO : " + resultTerm.toString());

	result =  new KeyTranslationResult(resultTerm);
	setTranslationForNode(sl,result);
    }


    protected Term computeTranslationForIntegerLiteral(String lit){
	Term resultTerm;

	// Translate an integer literal into an digit-representation ADT ...
	Sort intSort = proverFacade.getSortForOCLInteger();
	Sort numberSort = proverFacade.getNumberSort();
	Function intZero = 
  	    proverFacade.getFunctionForFeature("#",new Sort[0],numberSort);
	resultTerm = tf.createFunctionTerm(intZero, new Term[0]);

	Function nextDigitFunction;
	for(int i=0; i < lit.length(); i++){
	    String nextDigit = lit.substring(i,i+1);
	    nextDigitFunction = proverFacade.getFunctionForFeature(nextDigit,
								     new Sort[]{numberSort},
								     numberSort);
	    resultTerm = tf.createFunctionTerm(nextDigitFunction,resultTerm);
	}
	Function numberDilimiter = 
  	    proverFacade.getFunctionForFeature("Z",new Sort[]{numberSort},intSort);
	resultTerm = tf.createFunctionTerm(numberDilimiter,resultTerm);
	return resultTerm;
    }



   // =========================================================================================================
   // ================================= The a literal of OCL type Real ========================================
   // =========================================================================================================    

    protected void translateRealLiteral(Node sl, String lit){
 	log("Translating Real literal: " + lit);

	KeyTranslationResult result;
	Term resultTerm;
	
	
	// Currently: Translate a real literal into an constant of ADT real ...
	Sort realSort = proverFacade.getSortForOCLReal();
	Function realConstant = 
	    proverFacade.getFunctionForFeature(lit,
						 new Sort[0],
						 realSort);
	resultTerm = tf.createFunctionTerm(realConstant);

	log("INTO : " + resultTerm.toString());

	result =  new KeyTranslationResult(resultTerm);
	setTranslationForNode(sl,result);
    }



   // =========================================================================================================
   // ================================= The a literal of OCL type String ======================================
   // =========================================================================================================


    
     protected void translateStringLiteral(Node sl, String lit){
	log("Translating String literal: " + lit);
	
	//  // At the moment we translate each string literal into
//  	// a constant (with the same name) of the corresponing DL sort.
	
//  	Sort stringSort = proverFacade.getSortForOCLString();
//  	Function stringConstant = 
//  	    proverFacade.getFunctionForFeature(lit,
//  						 new Sort[0],
//  						 stringSort);
//  	Term resultTerm = tf.createFunctionTerm(stringConstant);

//  	log("INTO : " + resultTerm.toString());
	
//  	KeyTranslationResult result =  new KeyTranslationResult(resultTerm);
//  	setTranslationForNode(sl,result);


	// Use a ADT term  representation with the letters and 'emptyString' as constants and constructor 'append' ...
	
	Sort stringSort = proverFacade.getSortForOCLString();
	Function emptyStringConstant = 
  	    proverFacade.getFunctionForFeature(STRING_EMPTYSTRING,new Sort[0],stringSort);
	Term resultTerm = tf.createFunctionTerm(emptyStringConstant);

	Function stringAppendFunction = proverFacade.getFunctionForFeature(STRING_APPEND,
									     new Sort[]{stringSort, stringSort},stringSort);

	Function nextLetterConstant;
	Term nextLetterTerm;
	for(int i=0; i < lit.length(); i++){
	    String nextLetter = lit.substring(i,i+1);
	    nextLetterConstant = proverFacade.getFunctionForFeature(nextLetter,
								     new Sort[0],
								     stringSort);
	    nextLetterTerm = tf.createFunctionTerm(nextLetterConstant);
	    resultTerm = tf.createFunctionTerm(stringAppendFunction,resultTerm,nextLetterTerm);
	}

	log("INTO : " + resultTerm.toString());
	
 	KeyTranslationResult result =  new KeyTranslationResult(resultTerm);
  	setTranslationForNode(sl,result);


     }

   
    // =========================================================================================================
    // ================================= The a literal of OCL type OclType =====================================
    // =========================================================================================================    

    protected void translateOclTypeLiteral(Node sl, String lit){
 	log("Translating OclType literal: " + lit);

	KeyTranslationResult result;
	Term resultTerm;
	
	// Create a constant symbol of type OCLType with the same name as the literal ...

	Sort oclTypeSort = proverFacade.getSortForOCLType();
	Function oclTypeConstant = 
	    proverFacade.getFunctionForFeature(lit,
						 new Sort[0],
						 oclTypeSort);
	resultTerm = tf.createFunctionTerm(oclTypeConstant);
	
	log("INTO : " + resultTerm.toString());
	
	result =  new KeyTranslationResult(resultTerm);
	setTranslationForNode(sl,result);
    }





   // =========================================================================================================
   // ======================= The a literal of OCL type Set[T],Bag[T], Sequence[T] ============================
   // =========================================================================================================

     protected void translateCollectionLiteralWithExpression(Node cl, Type collType, Vector expressions){
	 log("Translating Collection literal with expressions for type: " + collType.toString());
	 
	 
	 // Get the kind of translation to be performed for the given node ...
	 TranslationType translationType = getTranslationTypeForNode(cl);
	 log("Current feature call should be translated in the following way : " + translationType.toString());
       

	 KeyTranslationResult result = new KeyTranslationResult();

	 Term resultTerm = null;
	
	 
	 Type returnType = collType;
	    
	
	 KeyTranslationResult currentExpressionTranslation;
	 Node currentExpressionNode;
	 
	 // Get the corresponding sort for the return type;
	 Sort returnSort = getDLSortForType(returnType);
	 Sort elemSort =  getElementSortForCollectionType(returnType);
	 
	 Collection c=(Collection) returnType;

	 if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
	     // ==================== TRANSLATION AS TERM ==============================
		   
	     String emptyCollConstantName;
	     String insertFunctionName, unionFunctionName;
	     switch (c.getCollectionKind()) {
	     case Collection.SET: 
		 emptyCollConstantName = SET_EMPTYSET ;
		 insertFunctionName = SET_INSERT;
		 unionFunctionName = SET_UNION;
		 break;
	     case Collection.BAG:
		 emptyCollConstantName = BAG_EMPTYBAG;
		 insertFunctionName = BAG_INSERT;
		 unionFunctionName = BAG_UNION;
		 break;
	     case Collection.SEQUENCE:
		 emptyCollConstantName = SEQUENCE_EMPTYSEQUENCE;
		 insertFunctionName = SEQUENCE_INSERT;
		 unionFunctionName = SEQUENCE_UNION;
		 break;
	     default: throw new RuntimeException("illegal collection type");
	     }
	     
	     
	     // Now get the corresponding constant for the empty collection ...
	     Function emptyCollFunction = proverFacade.getCollectionFunctionForFeature(emptyCollConstantName,
											 new Sort[0],
											 returnSort);
	     
	     Function collInsertFunction = 
		 proverFacade.getCollectionFunctionForFeature(insertFunctionName,
								new Sort[]{returnSort,elemSort},
								returnSort);
	     
	     Function collUnionFunction; 
	     
	     
	     resultTerm = tf.createFunctionTerm(emptyCollFunction);
	     
	     Type currentExpressionType;
	     Sort currentExpressionSort;
	     Term currentExpressionTerm;

	     for(int i=0;i<expressions.size();i++){
		 currentExpressionNode = (Node) expressions.elementAt(i);
		 currentExpressionType = tree.getNodeType(currentExpressionNode);
		 currentExpressionTranslation = (KeyTranslationResult) getTranslationForNode(currentExpressionNode);

		 if (currentExpressionType == Basic.BOOLEAN) {
		     currentExpressionTerm = convertFormulaToTerm(currentExpressionTranslation.getTerm(), 
								  currentExpressionTranslation);
		 } else {
		     currentExpressionTerm = currentExpressionTranslation.getTerm();
		 }

		 if (!(currentExpressionType instanceof Collection)){
		     // Insert next element in the collection ...
		     
		     resultTerm = tf.createFunctionTerm(collInsertFunction, resultTerm, 
							currentExpressionTerm);
		 } else {
		     // we have a collection as an element => care about flattening ...
		     currentExpressionSort = getDLSortForType(currentExpressionType);
		     collUnionFunction = 
			 proverFacade.getCollectionFunctionForFeature(unionFunctionName,
									new Sort[]{returnSort, currentExpressionSort},
									returnSort);
		     
		     resultTerm = tf.createFunctionTerm(collUnionFunction, resultTerm, 
							currentExpressionTerm);
		 }
		 
		 // Add all axioms that have been generated during the translation of currentExpressionNode ...
		 result.addAxioms(currentExpressionTranslation.getAxioms());
	     }
	     
	     
	     log("INTO :" + resultTerm.toString());
	     
	     result.setTerm(resultTerm);
	 } else {
	     // ==================== TRANSLATION AS PREDICATE ==============================
	    
	     resultTerm = tf.createJunctorTerm(Op.FALSE);
	     Type currentExpressionType;
	     Sort currentExpressionSort;
	     Term currentExpressionTerm;
	     Term currentExpressionPredicate;
	     Term eqTerm;
	     Term aTerm;

	     LogicVariable expressionPredicateVar;
	     LogicVariable newPredicateVar = newVariableDeliverer.getNewVariableForSort(elemSort);
	     Term newPredicateVarTerm = tf.createVariableTerm(newPredicateVar);
	     Collection currentExpressionCollType;

	     for(int i=0;i<expressions.size();i++){
		 currentExpressionNode = (Node) expressions.elementAt(i);
		 currentExpressionType = tree.getNodeType(currentExpressionNode);
		 currentExpressionTranslation = (KeyTranslationResult) getTranslationForNode(currentExpressionNode);
		 
		 if (!(currentExpressionType instanceof Collection)){
		     // Insert next element in the collection ...
		     if (currentExpressionType == Basic.BOOLEAN) {
			 currentExpressionTerm = convertFormulaToTerm(currentExpressionTranslation.getTerm(), 
								      currentExpressionTranslation);
		     } else {
			 currentExpressionTerm = currentExpressionTranslation.getTerm();
		     }

		     eqTerm = tf.createEqualityTerm(newPredicateVarTerm,currentExpressionTerm);
		     resultTerm = tf.createJunctorTermAndSimplify(Op.OR,resultTerm,eqTerm);
		 } else {
		     // we have a collection as an element => care about flattening ...
		     currentExpressionCollType =  (Collection) currentExpressionType;
		     currentExpressionTranslation = 
			 getPredicativeTranslationOfTranslation(currentExpressionTranslation, currentExpressionCollType);
			
		     currentExpressionPredicate = currentExpressionTranslation.getTerm();
		     expressionPredicateVar = currentExpressionTranslation.getPredicateVariable();

		     // Now replace the expression predicate var by the new predicate var ...		     
		     aTerm = renameVariableInTerm(currentExpressionPredicate, expressionPredicateVar, newPredicateVar);
		     resultTerm = tf.createJunctorTermAndSimplify(Op.OR,resultTerm,aTerm);

		 }
		 
		 // Add all axioms that have been generated during the translation of currentExpressionNode ...
		 result.addAxioms(currentExpressionTranslation.getAxioms());
	     }
	     
	     
	     log("INTO :" + resultTerm.toString());
	     
	     
	     result.setTerm(resultTerm);
	     result.setPredicateVariable(newPredicateVar);
	     result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);

	 }
	 

	 // Finally set the new translation result ...
	 setTranslationForNode(cl,result);
	
    }
    
    
    protected void translateCollectionLiteralWithRange(Node cl, Type collType, Node lowerBound,
						       Node upperBound){
	log("Translating Collection literal with range for type: " + collType.toString());
	 // Get the kind of translation to be performed for the given node ...
	 TranslationType translationType = getTranslationTypeForNode(cl);
	 log("Current feature call should be translated in the following way : " + translationType.toString());
       
	KeyTranslationResult result = new KeyTranslationResult();
	Term resultTerm = null;
	 
	 Type returnType = collType;
	    
	
	 KeyTranslationResult lowerBoundTranslation = 
	     (KeyTranslationResult) getTranslationForNode(lowerBound);
	 KeyTranslationResult upperBoundTranslation = 
	      (KeyTranslationResult) getTranslationForNode(upperBound);

	 Term lowerBoundTerm = lowerBoundTranslation.getTerm();
	 Term upperBoundTerm = upperBoundTranslation.getTerm();

	 // Get the corresponding sort for the return type;
	 Sort returnSort = getDLSortForType(returnType);
	 Sort elemSort =  getElementSortForCollectionType(returnType);
	 
	 if (elemSort != proverFacade.getSortForOCLInteger()){
	     error("A literal definition by use of a range specification is only allowed for INTEGER collection !!!");
	 }
	
	 Collection c=(Collection) returnType;
	
	 if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
	     // ==================== TRANSLATION AS TERM ==============================
		
	     String funcName;
	     
	     switch (c.getCollectionKind()) {
	     case Collection.SET: 
		 funcName = "set";
		 break;
	     case Collection.BAG:
		 funcName = "bag";
		 break;
	     case Collection.SEQUENCE:
		 funcName = "sequence";
		 break;
	     default: throw new RuntimeException("illegal collection type");
	     }
	     
	     String functionName = newFunctionNameDeliverer.getNewFunctionNameForName(funcName);
	     
	     // Get free vars in translation of the upperBound and lowerBound part ...
	     
	     SetOfQuantifiableVariable freeVarsLower = lowerBoundTerm.freeVars();
	     SetOfQuantifiableVariable freeVarsUpper = upperBoundTerm.freeVars();
	     
	     SetOfQuantifiableVariable paramsForNewFunction = freeVarsLower.union(freeVarsUpper);
	     log("FreeVars for new collection literal function '" + functionName + "' == " + paramsForNewFunction );
	     
	     Term[] paramVars;
	     Sort[] paramSorts;
	     
	     int freeVarsCnt = paramsForNewFunction.size();
	     if ( freeVarsCnt > 0){
		 paramVars = new Term[freeVarsCnt];
		 paramSorts = new Sort[freeVarsCnt];
		 
		 IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		 LogicVariable currentVar;
		 int i=0;
		 
		 while (varIt.hasNext()){
		     currentVar = (LogicVariable) varIt.next();
		     paramVars[i] = tf.createVariableTerm(currentVar);
		     paramSorts[i] = currentVar.sort();
		     i++;
		 }
		 
	     } else {
		 paramVars = new Term[0];
		 paramSorts = new Sort[0];
	     }
	     
	     Function newCollLiteralFunction =
		 proverFacade.getFunctionForFeature(functionName,
						      paramSorts,
						      returnSort);
	     
	     Term newCollLiteralFunctionTerm = 
		 tf.createFunctionTerm(newCollLiteralFunction, paramVars);
	     
	     // We take this new function term as the result of the translation ...
	     resultTerm = newCollLiteralFunctionTerm;
	     
	     
	     log("INTO :" + resultTerm.toString());
	     
	     // Now we have to create additional axioms restricting the interpretation of the new
	     // function symbol according to its intended meaning ...
	     
	     Term newAxiom1 = null;
	     Term newAxiom2 = null;
	     Term newAxiom3 = null;


	     // we have to distinguish the possible cases ...
	     Sort intSort = proverFacade.getSortForOCLInteger();
	     LogicVariable intVar1 = new LogicVariable(new Name("i"), intSort );
	     LogicVariable intVar2 = new LogicVariable(new Name("j"), intSort );
	     
	     Term intVarTerm1 = tf.createVariableTerm(intVar1);
	     Term intVarTerm2 = tf.createVariableTerm(intVar2);
	     
	     Function intLessOrEqualThanFunction = proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN,
											new Sort[]{intSort, intSort},
											Sort.FORMULA);
	     
	     
	     Function collContainsFunction = 
		 proverFacade.getFunctionForFeature(CONTAINS,
						      new Sort[]{returnSort, intSort},
						      Sort.FORMULA);
	     
	     Term aTerm;
	     Term containsTerm = tf.createFunctionTerm(collContainsFunction, resultTerm, intVarTerm1);
	     Term lessTerm1 = tf.createFunctionTerm(intLessOrEqualThanFunction,lowerBoundTerm, intVarTerm1);
	     Term lessTerm2 = tf.createFunctionTerm(intLessOrEqualThanFunction,intVarTerm1, upperBoundTerm);
	     Term rangeTerm = tf.createJunctorTermAndSimplify(Op.AND, lessTerm1 ,  lessTerm2 );
	     Term eqvTerm = tf.createJunctorTerm(Op.EQV, containsTerm ,  rangeTerm );
	     
	     Term oneTerm = this.computeTranslationForIntegerLiteral("1");

	     newAxiom1 = constraintTranslator.createAllClosureOfTerm(eqvTerm);


	     if ((c.getCollectionKind() == Collection.BAG) || 
		 (c.getCollectionKind() == Collection.SEQUENCE)) {

		 // No multiple occurences exist ...

		 Function collCountFunction = 
		     proverFacade.getFunctionForFeature(COUNT,
							  new Sort[]{returnSort, elemSort},
							  intSort);
		 
		 Term countTerm = tf.createFunctionTerm(collCountFunction, resultTerm, intVarTerm1);
		 Term lessTerm2b = tf.createFunctionTerm(intLessOrEqualThanFunction,countTerm, oneTerm);
  
		 newAxiom2 = constraintTranslator.createAllClosureOfTerm(lessTerm2b);
		 
		 if (c.getCollectionKind() == Collection.SEQUENCE) {
		     // The resulting sequence is ordered ...

		     Function collSizeFunction = 
			 proverFacade.getFunctionForFeature(SIZE,
							      new Sort[]{returnSort},
							      intSort);
		     
		     Function collAtFunction = 
			 proverFacade.getFunctionForFeature(SEQUENCE_AT,
							      new Sort[]{returnSort, intSort},
							      elemSort);
		     
		    
		     Term sizeTerm = tf.createFunctionTerm(collSizeFunction, resultTerm);
		     Term lessTerm3 = tf.createFunctionTerm(intLessOrEqualThanFunction, oneTerm, intVarTerm1);
		     Term lessTerm4 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm1, sizeTerm);
		     Term lessTerm5 = tf.createFunctionTerm(intLessOrEqualThanFunction, oneTerm, intVarTerm2);
		     Term lessTerm6 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm2, sizeTerm);
		     Term lessTerm7 = tf.createFunctionTerm(intLessOrEqualThanFunction, intVarTerm1, intVarTerm2);
		     
		     aTerm = tf.createJunctorTermAndSimplify(Op.AND, lessTerm3 , lessTerm4 );
		     aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm5 );
		     aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm6 );
		     aTerm = tf.createJunctorTermAndSimplify(Op.AND, aTerm , lessTerm7 );
		 
		     Term atTerm1 = tf.createFunctionTerm(collAtFunction, resultTerm, intVarTerm1);
		     Term atTerm2 = tf.createFunctionTerm(collAtFunction, resultTerm, intVarTerm2);
		     Term lessTerm8 = tf.createFunctionTerm(intLessOrEqualThanFunction, atTerm1, atTerm2);
		     
		     Term impTerm = tf.createJunctorTermAndSimplify(Op.IMP, aTerm , lessTerm8 );
		     
		     Term allTerm2 = tf.createQuantifierTerm(Op.ALL, new LogicVariable[]{intVar1, intVar2}, impTerm);
		  
		     newAxiom3 = constraintTranslator.createAllClosureOfTerm(allTerm2);
		 
		 }
	     
	     }
	     
	     log("Generated new axiom for '" + functionName + "' == " + newAxiom1);

	     if(newAxiom2 != null) {
		 log("Generated new axiom for '" + functionName + "' == " + newAxiom2);
	     }
	     if(newAxiom3 != null) {
		 log("Generated new axiom for '" + functionName + "' == " + newAxiom3);
	     }
	     // and now set the result ...
	     
	     result = new KeyTranslationResult();
	     result.setTerm(resultTerm);
	     result.addAxioms(lowerBoundTranslation.getAxioms());
	     result.addAxioms(upperBoundTranslation.getAxioms());
	     
	     
	     // add additional axioms ...
	     result.addAxiom(newAxiom1);
	     if(newAxiom2 != null) {
		  result.addAxiom(newAxiom2);
	     }
	     if(newAxiom3 != null) {
		  result.addAxiom(newAxiom3);
	     }

	 } else {
	     // ==================== TRANSLATION AS PREDICATE ==============================
	     Sort intSort = proverFacade.getSortForOCLInteger(); 
	     LogicVariable newPredicateVar = newVariableDeliverer.getNewVariableForSort(intSort);
	     Term newPredicateVarTerm = tf.createVariableTerm(newPredicateVar);
	     
	     
	     
	     Function intLessOrEqualThanFunction = proverFacade.getFunctionForFeature(LESS_OR_EQUAL_THAN,
											new Sort[]{intSort, intSort},
											Sort.FORMULA);
	     
	     Term lessOrEqTerm1 = tf.createFunctionTerm(intLessOrEqualThanFunction, lowerBoundTerm, newPredicateVarTerm);
	     Term lessOrEqTerm2 = tf.createFunctionTerm(intLessOrEqualThanFunction, newPredicateVarTerm, upperBoundTerm);
	     resultTerm = tf.createJunctorTerm(Op.AND, lessOrEqTerm1, lessOrEqTerm2);
 
	     log("INTO :" + resultTerm.toString());
	     
	      // and now set the result ...
	     
	     result = new KeyTranslationResult();
	     result.setTerm(resultTerm);
	     result.addAxioms(lowerBoundTranslation.getAxioms());
	     result.addAxioms(upperBoundTranslation.getAxioms());     
	     result.setPredicateVariable(newPredicateVar);
	     result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);

	     

	  }

	 // Finally set the new translation result ...
	 setTranslationForNode(cl,result);


    }
    


    protected void translateIfThenElseExpression(Node ie, Node ifExpression, Node thenExpression,
						 Node elseExpression){
	log("Translating if-then-else-construct.");
	
	// Get the kind of translation to be performed for the given node ...
	TranslationType translationType = getTranslationTypeForNode(ie);
	log("Current feature call should be translated in the following way : " + translationType.toString());
	
	Type thenExpressionType = tree.getNodeType(thenExpression);
	Type elseExpressionType = tree.getNodeType(elseExpression);
	
	KeyTranslationResult result;
	Term resultTerm;
	
	KeyTranslationResult ifTrans = 
	    (KeyTranslationResult) getTranslationForNode(ifExpression);
	Term ifTerm = ifTrans.getTerm();
	
	ifTerm = convertTermToFormula(ifTerm);

	KeyTranslationResult thenTrans = 
	    (KeyTranslationResult) getTranslationForNode(thenExpression);
	Term thenTerm = thenTrans.getTerm();
	
	KeyTranslationResult elseTrans = 
	    (KeyTranslationResult) getTranslationForNode(elseExpression);
	Term elseTerm = elseTrans.getTerm();


	if (!translationType.containsMark(TranslationMark.AS_PREDICATE)) {
	    // ==================== TRANSLATION AS TERM ==============================
	   
	    Term axiom1,axiom2;

	    if (thenTerm.sort() == elseTerm.sort()){
		Sort resSort = thenTerm.sort();
		
		if ( (thenExpressionType == Basic.BOOLEAN) && (elseExpressionType == Basic.BOOLEAN)) {
		    if (!translationType.containsMark(TranslationMark.AS_BOOLEAN_TERM)) {
			resSort = Sort.FORMULA;
			thenTerm = convertTermToFormula(thenTerm);
			elseTerm = convertTermToFormula(elseTerm);
		    } else {
			resSort = proverFacade.getSortForOCLBoolean();
			thenTerm = convertFormulaToTerm(thenTerm, thenTrans);
			elseTerm = convertFormulaToTerm(elseTerm, elseTrans);
		    }
		    
		    
		}

		// Get free vars in translation of then and else part and of the boolean statement ...
		SetOfQuantifiableVariable freeVarsIf = ifTerm.freeVars();
		log("FreeVars in ifTerm == " + freeVarsIf );
		
		SetOfQuantifiableVariable freeVarsThen = thenTerm.freeVars();
		log("FreeVars in thenTerm == " + freeVarsThen );
		
		SetOfQuantifiableVariable freeVarsElse = elseTerm.freeVars();
		log("FreeVars in elseTerm == " + freeVarsElse );
		
		SetOfQuantifiableVariable paramsForNewFunction =  freeVarsIf.union(freeVarsThen).union(freeVarsElse);
		log("FreeVars for new if-res-function == " + paramsForNewFunction );
		
		String newFuncName = newFunctionNameDeliverer.getNewFunctionNameForName("if");
		
		
		Term[] paramVars;
		Sort[] paramSorts;
		
		int freeVarsCnt = paramsForNewFunction.size();
		if ( freeVarsCnt > 0){
		    paramVars = new Term[freeVarsCnt];
		    paramSorts = new Sort[freeVarsCnt];
		    
		    IteratorOfQuantifiableVariable varIt = paramsForNewFunction.iterator();
		    LogicVariable currentVar;
		    int i=0;
		    
		    while (varIt.hasNext()){
			currentVar = (LogicVariable) varIt.next();
			paramVars[i] = tf.createVariableTerm(currentVar);
			paramSorts[i] = currentVar.sort();
			i++;
		    }
		    
		} else {
		    paramVars = new Term[0];
		    paramSorts = new Sort[0];
		}
		
		Function newIfFunction =
		    proverFacade.getFunctionForFeature(newFuncName,
							 paramSorts,
							 resSort);
		
		Term newIfFunctionTerm = 
		    tf.createFunctionTerm(newIfFunction, paramVars);
		
		// We take this new function term as the result of the translation ...
		resultTerm = newIfFunctionTerm;
		
		
		// Now we can create the new axioms ...
		
		Term eqTerm, implTerm, notTerm;
		Equality eqFunction;
		eqFunction = resSort.getEqualitySymbol();
		
		// the first one ...		    
		eqTerm = tf.createEqualityTerm(eqFunction,
					       newIfFunctionTerm,
					       thenTerm);
		implTerm = tf.createJunctorTermAndSimplify(Op.IMP, ifTerm, eqTerm); 
		axiom1 = constraintTranslator.createAllClosureOfTerm(implTerm);
		
		// the second one ...
		
		eqTerm = tf.createEqualityTerm(eqFunction,
					       newIfFunctionTerm,
					       elseTerm);
		notTerm = tf.createJunctorTerm(Op.NOT, ifTerm);
		implTerm = tf.createJunctorTermAndSimplify(Op.IMP, notTerm, eqTerm);	    
		axiom2 = constraintTranslator.createAllClosureOfTerm(implTerm);
		
		
		log("Generated new axiom (1) == " + axiom1);
		log("Generated new axiom (2) == " + axiom2);
		
		
		
		// and now set the result ...
		
		result = new KeyTranslationResult();
		result.setTerm(resultTerm);
		result.addAxioms(ifTrans.getAxioms());
		result.addAxioms(thenTrans.getAxioms());
		result.addAxioms(elseTrans.getAxioms());
		
		// add two additional axioms ...
		result.addAxiom(axiom1);
		result.addAxiom(axiom2);
	    

		setTranslationForNode(ie,result);
		
	    } else {
		// we have two different sorts in the DL and thus can't be handeled yet ...
		error("IF expression with two different types for 'then' and 'else' can't be handeled yet !!!");
	    }
	} else {
	    // ==================== TRANSLATION AS A FORMULA ==============================
	    // Translate the if-feature more directly without
	    // introducing a new symbol 'if-k' for boolean then- & else-expressions:
	    // We can simply use the formula: 
	    //    (transIfExp -> transThenExp) & (!transIfExp -> transElseExp)
	
	    // But we have to distinguish the cases when a boolean term or a formula should
	    // be created ...

 
	    if (!translationType.containsMark(TranslationMark.AS_BOOLEAN_TERM)) {
		// Create a formula ...
		thenTerm = convertTermToFormula(thenTerm);
		elseTerm = convertTermToFormula(elseTerm);
		
		Term impTerm1, impTerm2, notTerm;
	   
		// The first case ...
		impTerm1 = tf.createJunctorTermAndSimplify(Op.IMP,
							   ifTerm,
							   thenTerm);
		
		// The second case ...
		notTerm = tf.createJunctorTerm(Op.NOT,ifTerm);
		impTerm2 = tf.createJunctorTermAndSimplify(Op.IMP, notTerm, elseTerm); 
		
		resultTerm = tf.createJunctorTermAndSimplify(Op.AND, impTerm1, impTerm2); 

	    } else {
		// Create a term over sort 'boolean' ...

		thenTerm = convertFormulaToTerm(thenTerm, thenTrans);
		elseTerm = convertFormulaToTerm(elseTerm, elseTrans);
		
		Sort boolSort = proverFacade.getSortForOCLBoolean();

		Function impSymbol = proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_IMPLIES,
									  new Sort[]{boolSort,boolSort},
									  boolSort); 

		Function andSymbol = proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_AND,
									  new Sort[]{boolSort,boolSort},
									  boolSort); 
	       
		Function notSymbol = proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_NOT,
									  new Sort[]{boolSort},
									  boolSort); 

		Term impTerm1, impTerm2, notTerm;
	   
		// The first case ...
		impTerm1 = tf.createFunctionTerm(impSymbol, ifTerm, thenTerm);
		
		// The second case ...
		notTerm = tf.createFunctionTerm(notSymbol,ifTerm);
		impTerm2 = tf.createFunctionTerm(impSymbol, notTerm, elseTerm); 
		
		resultTerm = tf.createFunctionTerm(andSymbol, impTerm1, impTerm2); 
	    }
		    

	    // and now set the result ...
		
	    result = new KeyTranslationResult();
	    result.setTerm(resultTerm);
	    result.addAxioms(ifTrans.getAxioms());
	    result.addAxioms(thenTrans.getAxioms());
	    result.addAxioms(elseTrans.getAxioms());
		
	    setTranslationForNode(ie,result);
	}

    }
    




    // =============================== Some auxilliary methods ... =============================
    
    /**
     * Maps some OCL type onto the logical sort that represents this OCL type in
     * the target language.
     * @param t - the OCL type to be mapped.
     * @return the sort (in the target languge) that is used to represent the
     * given OCL type. 
     */ 

    public Sort getDLSortForType(Type t) {
	
	Sort returnSort = null;

	// initialize proverFacade when necessary ...
	if (proverFacade == null){
	    this.proverFacade = constraintTranslator.getProverFacade();
	}

	if (t instanceof Basic) {
	    if (t==Basic.INTEGER) returnSort = proverFacade.getSortForOCLInteger();
	    else if (t==Basic.REAL) returnSort = proverFacade.getSortForOCLReal();
	    else if (t==Basic.BOOLEAN) returnSort = proverFacade.getSortForOCLBoolean();
	    else if (t==Basic.STRING) returnSort = proverFacade.getSortForOCLString();
	    else throw new RuntimeException("illegal basic type");
	} else if (t instanceof Collection) {
	    Collection c=(Collection) t;
	    String elementType = getElementTypeForCollectionType(c);
	   
	    NonCollectionSort elemSort
		= (NonCollectionSort)getDLSortForNonCollectionType(elementType);

	    switch (c.getCollectionKind()) {
	    case Collection.SET: 
		returnSort = proverFacade.getSortForOCLSet(elemSort);
		break;
	    case Collection.BAG:
		returnSort = proverFacade.getSortForOCLBag(elemSort);
		break;
	    case Collection.SEQUENCE:
		returnSort = proverFacade.getSortForOCLSequence(elemSort);
		break;
	    default: throw new RuntimeException("illegal collection type");
	    }
	} else if (t instanceof OclType) {
	    error("OCL-Type not handled yet !!!!!!!!!!!!!!!!");
	} else {
	    // we have a UML model type ...
	    returnSort = proverFacade.getSortForClass(t.toString());
	}
	return returnSort;

    }

    /**
     * Maps some non collection OCL type onto the KeYJavaType that represents this OCL type in
     * the target language.
     * @param elementType - the name of the non-collection OCL type to be mapped.
     * @return the KeYJavaType that is used to represent the given OCL type within KeY. 
     */ 
    public KeYJavaType getKeYJavaTypeForNonCollectionType(String elementType) {	
	// initialize proverFacade when necessary ...
	if (proverFacade == null){
	    this.proverFacade = constraintTranslator.getProverFacade();
	}
	if (elementType.equals("Integer")) return proverFacade.getKeYJavaTypeForOCLInteger();
	else if (elementType.equals("Real")) return proverFacade.getKeYJavaTypeForOCLReal();
	else if (elementType.equals("Boolean")) return proverFacade.getKeYJavaTypeForOCLBoolean();
	else if (elementType.equals("String")) return proverFacade.getKeYJavaTypeForOCLString();
	else if (elementType.equals("OclType")) error("OCL-Type not handled yet !!!!!!!!!!!!!!!!");
	else {
	    // we have an UML model type ...
	    KeYJavaType kjt = proverFacade.getKeYJavaTypeForClass(elementType);
	    //System.err.println(elementType+"-->"+kjt);
	    return kjt;
	}
	return null;
    }


    /**
     * Maps some non-collection OCL type onto the logical sort that represents this OCL type in
     * the target language.
     * @param elementType - the name of the non-collection OCL type to be mapped.
     * @return the sort (in the target languge) that is used to represent the
     * given OCL type. 
     */ 
    public Sort getDLSortForNonCollectionType(String elementType){
	return getKeYJavaTypeForNonCollectionType(elementType).getSort();
    }

    /**
     * Maps a typename onto the corresponding OCL type object that is used by the parser 
     * (and therefore in the OclTree object of some constraint) to represent this OCL type.
     * @param elementType - the name of the non-collection OCL type to be mapped.
     * @return the Type object that represents the OCL type with the given name or null
     *         if no such type exists.
     */ 
    
    public Type getTypeForNonCollectionTypeName(String elementType){
	Type result = null;
	if (elementType.equals("Integer")) result = Basic.INTEGER;
	else if (elementType.equals("Real")) result = Basic.REAL;
	else if (elementType.equals("Boolean")) result = Basic.BOOLEAN;
	else if (elementType.equals("String")) result = Basic.STRING;
	else if (elementType.equals("OclType")) error("OCL-Type not handled yet !!!!!!!!!!!!!!!!");
	else {
	    // we have an UML model type ...
	    ModelFacade mf = null;
	   
	    //REMARK: THE INITIALIZATION of the mf variable has to
	    // be done first, but is NOT implemented currently.
	    // Therefore a NullPointerException can be thrown !!!!

	    // In order to initialize mf we have to get the ModelFacade
	    // that will be used by the parser in the constructor call,
	    // because the OclTree doesn't allow us to get the
	    // Object of the Facade that was used during parsing the
	    // the constraint ...

	    try {
		result =  mf.getClassifier(elementType);
	    } catch(Exception e){
		log("ERROR WHEN trying to get the Type-Object of the ModelType '"+ elementType +"' from the ModelFacade'");
		e.printStackTrace();
	    }
	    
	}
	return(result);
    }

    /**
     * Maps the given OCL collection type onto the corresponding type 
     * of the elements of the collection.
     * The given type has to be a collection type!
     * @param ct - the OCL collection type to be mapped.
     * @return the name of the type of the elements in the collection.
     */

    public String getElementTypeForCollectionType(Type ct){
	Collection c = (Collection) ct;
	String cTypeName = c.toString();
	int index1 = cTypeName.indexOf("(");
	int index2 = cTypeName.indexOf(")");
	String elementTypeName = cTypeName.substring(index1+1,index2);
	return elementTypeName;
    }
    
   /**
    * Maps the given OCL collection type onto the corresponding sort that
    * belongs to the type of the elements of the collection.
    * The given type has to be a collection type!
    * @param ct - the OCL collection type to be mapped.
    * @return the name of the type of the elements in the collection.
    */

    public NonCollectionSort getElementSortForCollectionType(Type ct){
	String elementType = getElementTypeForCollectionType(ct);
	NonCollectionSort result = (NonCollectionSort)
	    getDLSortForNonCollectionType(elementType);
	return result;
    }

     /**
     * If the given TranslationResult is not predicative (therefore
     * we have a constructive description by use of an ADT term !)
     * then a corresponding predicative translation will be created.
     * If we already have an predicative translation, it will be
     * cloned ...
     * Should only be used for the predicative representation of
     * sets and bags.
     * UniverseSort is the Sort of the Universe the set/bag elements
     * originate from !
     */
    public KeyTranslationResult getPredicativeTranslationOfTranslation(KeyTranslationResult tr,
								       Collection collType){
	KeyTranslationResult result = new KeyTranslationResult();
	Term resultTerm = null;
	
	if(tr.isPredicativeTranslation()){
	    // copy the given translation result
	    result.setTerm(tr.getTerm());
	    result.addAxioms(tr.getAxioms());
	    result.setPredicateVariable(tr.getPredicateVariable());
	    result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);    
	    return result;
	}

	Sort universeSort = getElementSortForCollectionType(collType);

	LogicVariable newVar = newVariableDeliverer.getNewVariableForSort(universeSort);
	Term varTerm = tf.createVariableTerm(newVar);
	
	Term repTerm = tr.getTerm();
	Function containsFunction = 
	    proverFacade.getCollectionFunctionForFeature(CONTAINS,
							   new Sort[] {repTerm.sort(), varTerm.sort()},
							   Sort.FORMULA);
	
	resultTerm = tf.createFunctionTerm(containsFunction, repTerm, varTerm);
	
	result.setTerm(resultTerm);
	result.addAxioms(tr.getAxioms());
	result.setPredicateVariable(newVar);
	result.setTranslationType(KeyTranslationResult.PREDICATIVE_TRANSLATION);
	   
	
	return result;
	
    }

    /**
     * Converts the given (boolean) Term into an equivalent formula.
     * If the given term is already a formula, the operand will be returned unchanged.
     * Returns null if the given terms has no boolean sort or represents a formula.
     */
    Term convertTermToFormula(Term opTerm){
	Sort opSort = opTerm.sort();
	if (opTerm.op() instanceof ProgramMethod) {
	    opSort=((ProgramMethod)opTerm.op()).sort();
	}
	Term result;
	Sort boolSort = proverFacade.getSortForOCLBoolean();

	if(opSort == Sort.FORMULA){
	    result = opTerm;
	} else if (opSort == boolSort){
	    // Create the formula : (opTerm = true-boolean)
	    Function trueSymbol;
	    

	    trueSymbol =   proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_TRUE,
								new Sort[0],
								boolSort); 
	    Term trueTerm = tf.createFunctionTerm(trueSymbol);
	    result = tf.createEqualityTerm(opTerm, trueTerm);
	    
	    log("Converting boolean-term '" + opTerm +"' into formula: " +result);
	    
	} else {
	    result = null;
	    log("!!!! Converting NON-boolean-term into a formula is ILLEGAL !!!!!");
	}

	return result;
    }


    /**
     * Converts the given formula into an equivalent term of sort 'boolean'.
     * An additional axiom will be introduced in the axiom set of the given
     * KeyTranslationResult.
     * If the given expression is already a term, the operand will be returned unchanged.
     * Returns null if the given expression has no boolean sort or represents a formula.
     */
    Term convertFormulaToTerm(Term opTerm, KeyTranslationResult translation){
	Sort opSort = opTerm.sort();
	Term result;
	Sort boolSort = proverFacade.getSortForOCLBoolean();

	if(opSort == Sort.FORMULA){
	    // Create a new function symbol that represents the boolean value of the
	    // given formula ...
	    String newSymbolName = newFunctionNameDeliverer.getNewFunctionNameForName("toBoolean");

	    // Get free vars in the given formula ...
	    SetOfQuantifiableVariable freeVars = opTerm.freeVars();

	    log("FreeVars in given formula (for conversion to boolean term) == " + freeVars);
		
	    Term[] paramVars;
	    Sort[] paramSorts;
		
	    int freeVarsCnt = freeVars.size();
	    if ( freeVarsCnt > 0){
		paramVars = new Term[freeVarsCnt];
		paramSorts = new Sort[freeVarsCnt];
		    
		IteratorOfQuantifiableVariable varIt = freeVars.iterator();
		LogicVariable currentVar;
		int i=0;
		
		while (varIt.hasNext()){
		    currentVar = (LogicVariable) varIt.next();
		    paramVars[i] = tf.createVariableTerm(currentVar);
		    paramSorts[i] = currentVar.sort();
		    i++;
		}
		
	    } else {
		paramVars = new Term[0];
		paramSorts = new Sort[0];
	    }
		
	    Function newBooleanFunction = proverFacade.getFunctionForFeature(newSymbolName,
									       paramSorts,
									       boolSort);
		
	    Term newFunctionTerm = tf.createFunctionTerm(newBooleanFunction, paramVars);
		
	    // We take this new function term as the result of the translation ...
	    result = newFunctionTerm;
		

	    // Now we can create the new axioms ...
	    Term eqTerm1, eqTerm2, newAxiom;

	    Function trueSymbol;
	    trueSymbol =   proverFacade.getFunctionForFeature(TranslationConstants.BOOLEAN_TRUE,
								new Sort[0],
								boolSort); 

	    Term trueTerm = tf.createFunctionTerm(trueSymbol);
	    eqTerm1 = tf.createEqualityTerm(newFunctionTerm, trueTerm);
	    
	    eqTerm2 = tf.createEqualityTerm(eqTerm1, opTerm);
	    newAxiom = constraintTranslator.createAllClosureOfTerm(eqTerm2); 

	    // Add the new axiom into the axiom set of the given translation result ...
	    translation.addAxiom(newAxiom);

	    log("Converting boolean-term '" + opTerm +"' into formula: " +result);
	    log("Adding axiom: " + newAxiom);
	    
	} else if (opSort == boolSort){
	    result = opTerm;
	} else {
	    result = null;
	    log("!!!! Converting NON-boolean-expression into a boolean term is ILLEGAL !!!!!");
	}

	return result;
    }


    /**
     * Replaces all occurrences of a given variable by another variable in
     * the specified term.
     * @param t - the term for which the variable replacement should be applied.
     * @param oldVar - the variable to replace.
     * @param newVar - the variable to be used instead of 'oldVar'.
     * @return the term that results after application of the substitution.
     */

    Term renameVariableInTerm(Term t, LogicVariable oldVar, LogicVariable newVar){
	Term result;
	MapFromLogicVariableToTerm map = 
	    MapAsListFromLogicVariableToTerm.EMPTY_MAP;

	
	log("renameVariableInTerm -- term ==" + t);
	log("renameVariableInTerm -- oldVar ==" + oldVar);
	log("renameVariableInTerm -- newVar ==" + newVar);

	Term newVarTerm = tf.createVariableTerm(newVar);
	ClashFreeSubst subst = 
	    new ClashFreeSubst(oldVar, newVarTerm);
	return subst.apply(t);
    }

    /**
     * returns the set of occuring non-rigid function symbols for the last translation
     * of an OCL expression ...
     */
    Set getOccuringNonRigidFunctionSymbols(){
	return occuringNonrigidFunctionSymbols;
    }
    
    /**
     * returns the map that maps  occuring '@pre'-tagged non-rigid function symbols that 
     * correspond to some attribute of some model type onto the Operator in to logic
     * (for the last translation of an OCL expression)
     */
    Map getOccuringAttributeAtPreFunctions() {
	return occuringAttributeAtPreFunctions;
    }

     /**
     * returns the map that maps occuring '@pre'-tagged non-rigid function symbols that 
     * correspond to some method of some model type onto the Operator in to logic
     * (for the last translation of an OCL expression)
     */
    Map getOccuringMethodAtPreFunctions() {
	return occuringMethodAtPreFunctions;
    }


    /**
     * returns a quantified term with the given variables as quantified variables
     * and a special quantor corresponding to the given quantor, when the logical
     * variables to be quantified have a class type ...
     */
    public Term getQuantifiedTermWith(Quantifier q, LogicVariable v, Term t){
	Term resultTerm;
	Quantifier quantor = null;
	Sort varSort = v.sort();  
	if (!proverFacade.isSortForClass(varSort)){   //possible for a logic variable????
	    // The given variable has no CLASS type ...
	    quantor = q;
	} else {
	    
	    // The correct version ...
	    //   if (q == Op.ALL){
//  		quantor = ALL_PRIME;
//  	    } else if (q == Op.EX){
//  		quantor = EX_PRIME;
//  	    }

	    // The incorrect version ...
	    quantor = q;
	}

	resultTerm = tf.createQuantifierTerm(quantor,v,t);
	return resultTerm;

    }
    
    /**
     * Return an array that has the given term prepended to
     * the given term array.
     */
    Term[] prependTerm(Term t,Term[] terms){
	Term[] result = new Term[terms.length + 1];
	result[0] = t;
	for (int i=0;i<terms.length;i++){
	    result[i+1] = terms[i];
	}
	return result;
    }
    
    // ================= Some helper methods ============================================
    

    /**
     * @return the ProverFacade object which should be used
     *         in order to build logical terms or access sorts. 
     */
    public ProverFacade getProverFacade(){
	return proverFacade;
    }

    /**
     * Sets the translation control object that will be used to
     * determine how each subexpression in the constraint tree will be
     * treated by the transltion.
     * @param tControl - the object to be used for controling the translation process. 
     */
    public void setTranslationControl(DefaultTranslationControl tControl){
	this.translationControl = tControl;
    }

    /**
     * Enables or disables logging output in the command line.
     * @param log - if set to true, logging will be enabled.
     */
    public void setLoggingEnabled(boolean log){
	logEnabled = log;
    }

    /**
     * Logs the given message on the command line if 
     * logging is currently enabled.
     */
    
     protected void log(String msg){
	// do nothing ...
	 if (logEnabled){
	     super.log(msg);
	 }
     }
    
}




